from collections import defaultdict


class Graph(object):
    def __init__(self, doc_id, entry_id, sent_id, doc_text, tokens, anchors, triggers, mentions, vocabs, start, end):
        self.doc_id = doc_id
        self.entry_id = entry_id
        self.sent_id = sent_id
        self.doc_text = doc_text
        self.tokens = tokens  # with document-level span information
        self.main_trigger_id = -1
        self.anchors = []
        self.triggers = []
        self.mentions = []
        self.roles = []

        self.vocabs = vocabs
        self.event_type_itos = {i: s for s, i in vocabs['event_type'].items()}

        self.anchor_num = len(anchors)
        self.trigger_num = len(triggers)
        self.mention_num = len(mentions)

        self.write_entry = {
            'doc-id': self.doc_id,
            'entry-id': self.entry_id,
            'sent-id': self.sent_id,
            'segment-text': doc_text,
            'annotation-sets': {
                'basic-events': {
                    'granular-templates': {},
                    'events': {},
                    'span-sets': {}
                }
            },
            'start': start, 'end': end
        }
        self.anchor_id_2_template_id = {}
        self.trigger_id_2_event_id = {}
        self.mention_id_2_span_id = {}
        # initialize info
        for anchor in anchors:
            self.add_anchor(
                start=anchor['start'], end=anchor['end'], template_type=anchor['template-type'],
                type_label=anchor['type'], completion_label=anchor['completion'],
                overtime_label=anchor['over-time'], coordinated_label=anchor['coordinated']
            )

        for trigger in triggers:
            self.add_trigger(trigger[0], trigger[1], trigger[2], trigger[3])

        for mention in mentions:
            self.add_mention(mention[0], mention[1], mention[2])

    def add_anchor(self, start, end, template_type, type_label, completion_label, overtime_label, coordinated_label):
        self.anchors.append((start, end, template_type))
        self.anchor_num = len(self.anchors)

        anchor_span = (self.tokens[start]['start'], self.tokens[end - 1]['end'])
        if self.write_entry['start'] <= anchor_span[0] and anchor_span[1] <= self.write_entry['end']:
            template_id = f"template-{len(self.write_entry['annotation-sets']['basic-events']['granular-templates']) + 1}"
            span_id = f"ss-{len(self.write_entry['annotation-sets']['basic-events']['span-sets']) + 1}"
            self.write_entry['annotation-sets']['basic-events']['granular-templates'][template_id] = {
                "template-anchor": span_id,
                "template-id": template_id,
                "template-type": template_type
            }
            if type_label != 'O':
                self.write_entry['annotation-sets']['basic-events']['granular-templates'][template_id][
                    'type'] = type_label
            if completion_label != 'O':
                self.write_entry['annotation-sets']['basic-events']['granular-templates'][template_id][
                    'completion'] = completion_label
            if overtime_label != 'O':
                self.write_entry['annotation-sets']['basic-events']['granular-templates'][template_id][
                    'over-time'] = True
            if coordinated_label != 'O':
                self.write_entry['annotation-sets']['basic-events']['granular-templates'][template_id][
                    'coordinated'] = True

            self.add_span(span_id, start, end, span_type='template-anchor')
            self.anchor_id_2_template_id['{}-{}'.format(start, end)] = template_id

    def add_mention(self, start, end, label):
        self.mentions.append((start, end, label))
        self.mention_num = len(self.mentions)

        mention_span = (self.tokens[start]['start'], self.tokens[end - 1]['end'])
        if self.write_entry['start'] <= mention_span[0] and mention_span[1] <= self.write_entry['end']:
            span_id = f"ss-{len(self.write_entry['annotation-sets']['basic-events']['span-sets']) + 1}"
            self.add_span(span_id, start, end, span_type=label)

    def add_span(self, span_id, start_token, end_token, span_type):
        self.write_entry['annotation-sets']['basic-events']['span-sets'][span_id] = {
            'spans': [
                {
                    "start": self.tokens[start_token]['start'],
                    "end": self.tokens[end_token - 1]['end'],
                    "string": self.doc_text[self.tokens[start_token]['start']:
                                            self.tokens[end_token - 1]['end']],
                    "synclass": span_type
                }
            ],
            'ssid': span_id
        }

    def add_trigger(self, start, end, label, score):
        self.triggers.append((start, end, label))
        self.trigger_num = len(self.triggers)

        trigger_span = (self.tokens[start]['start'], self.tokens[end - 1]['end'])
        if self.write_entry['start'] <= trigger_span[0] and trigger_span[1] <= self.write_entry['end']:
            event_id = f"event-{len(self.write_entry['annotation-sets']['basic-events']['events']) + 1}"
            span_id = f"ss-{len(self.write_entry['annotation-sets']['basic-events']['span-sets']) + 1}"
            self.write_entry['annotation-sets']['basic-events']['events'][event_id] = {
                "agents": [],
                "money": [],
                "anchors": span_id,
                "event-type": label,
                'event-confidence-score': score,
                "eventid": event_id,
                "patients": [],
                "ref-events": []
            }
            self.add_span(span_id, start, end, span_type='event-anchor')
            self.trigger_id_2_event_id['{}-{}'.format(start, end)] = event_id

    @staticmethod
    def empty_graph(doc_id, entry_id, sent_id, doc_text, tokens, vocabs):
        """Create a graph without any node and edge.
        :param vocabs (dict): Vocabulary object.
        """
        return Graph(doc_id, entry_id, sent_id, doc_text, tokens, [], [], vocabs)


class MentionGraph(object):
    def __init__(self, doc_id, entry_id, sent_id, doc_text, tokens, triggers, mentions, vocabs, start, end):
        self.doc_id = doc_id
        self.entry_id = entry_id
        self.sent_id = sent_id
        self.doc_text = doc_text
        self.tokens = tokens  # with document-level span information
        self.main_trigger_id = -1
        self.triggers = []
        self.mentions = []
        self.roles = []

        self.vocabs = vocabs
        self.event_type_itos = {i: s for s, i in vocabs['event_type'].items()}

        self.trigger_num = len(triggers)
        self.mention_num = len(mentions)

        self.write_entry = {
            'doc-id': self.doc_id,
            'entry-id': self.entry_id,
            'sent-id': self.sent_id,
            'segment-text': doc_text,
            'annotation-sets': {
                'basic-events': {
                    'events': {},
                    'span-sets': {}
                }
            },
            'start': start, 'end': end
        }
        self.trigger_id_2_event_id = {}
        self.mention_id_2_span_id = {}
        # initialize info
        for trigger in triggers:
            self.add_trigger(trigger[0], trigger[1], trigger[2], trigger[3])
        for mention in mentions:
            self.add_mention(mention[0], mention[1])

    def add_mention(self, start, end):
        self.mentions.append((start, end))
        self.mention_num = len(self.mentions)

    def add_span(self, span_id, start_token, end_token, span_type):

        self.write_entry['annotation-sets']['basic-events']['span-sets'][span_id] = {
            'spans': [
                {
                    "start": self.tokens[start_token]['start'],
                    "end": self.tokens[end_token - 1]['end'],
                    "string": self.doc_text[self.tokens[start_token]['start']:
                                            self.tokens[end_token - 1]['end']],
                    "synclass": span_type
                }
            ],
            'ssid': span_id
        }

    def add_trigger(self, start, end, label, score):
        self.triggers.append((start, end, label))
        self.trigger_num = len(self.triggers)

        trigger_span = (self.tokens[start]['start'], self.tokens[end - 1]['end'])
        if self.write_entry['start'] <= trigger_span[0] and trigger_span[1] <= self.write_entry['end']:
            event_id = f"event-{len(self.write_entry['annotation-sets']['basic-events']['events']) + 1}"
            span_id = f"ss-{len(self.write_entry['annotation-sets']['basic-events']['span-sets']) + 1}"
            self.write_entry['annotation-sets']['basic-events']['events'][event_id] = {
                "agents": [],
                "agents-confidence-scores": [],
                "money": [],
                "money-confidence-scores": [],
                "anchors": span_id,
                "event-type": label,
                'event-confidence-score': score,
                "eventid": event_id,
                "patients": [],
                "patients-confidence-scores": [],
                "ref-events": [],
                "ref-events-confidence-scores": []
            }
            self.add_span(span_id, start, end, span_type='event-anchor')
            self.trigger_id_2_event_id['{}-{}'.format(start, end)] = event_id

    def add_role(self, start, end, mention_label, role_label, score):
        assert self.main_trigger_id >= 0

        mention_span = (self.tokens[start]['start'], self.tokens[end - 1]['end'])
        if self.write_entry['start'] <= mention_span[0] and mention_span[1] <= self.write_entry['end']:
            event_id = self.trigger_id_2_event_id[
                '{}-{}'.format(self.triggers[self.main_trigger_id][0], self.triggers[self.main_trigger_id][1])]

            if role_label in ['agents', 'money', 'patients', 'time-attachments']:
                span_id = f"ss-{len(self.write_entry['annotation-sets']['basic-events']['span-sets']) + 1}"
                self.add_span(span_id, start, end, span_type=mention_label)

                if role_label not in self.write_entry['annotation-sets']['basic-events']['events'][event_id]:
                    self.write_entry['annotation-sets']['basic-events']['events'][event_id][role_label] = []
                    self.write_entry['annotation-sets']['basic-events']['events'][event_id][
                        role_label + '-confidence-scores'] = []

                self.write_entry['annotation-sets']['basic-events']['events'][event_id][role_label].append(span_id)
                self.write_entry['annotation-sets']['basic-events']['events'][event_id][
                    role_label + '-confidence-scores'].append(score)
            elif role_label == 'ref-events' and '{}-{}'.format(start, end) in self.trigger_id_2_event_id:
                ref_event_id = self.trigger_id_2_event_id['{}-{}'.format(start, end)]

                self.write_entry['annotation-sets']['basic-events']['events'][event_id]["ref-events"].append(
                    ref_event_id)
                self.write_entry['annotation-sets']['basic-events']['events'][event_id][
                    "ref-events-confidence-scores"].append(
                    score)

    @staticmethod
    def empty_graph(doc_id, entry_id, sent_id, doc_text, tokens, vocabs):
        """Create a graph without any node and edge.
        :param vocabs (dict): Vocabulary object.
        """
        return Graph(doc_id, entry_id, sent_id, doc_text, tokens, [], [], vocabs)


class SlotGraph(object):
    def __init__(self, doc_id, entry_id, doc_ori_text, tokens, triggers, mentions, roles, vocabs):
        self.doc_id = doc_id
        self.entry_id = entry_id
        self.doc_ori_text = doc_ori_text
        self.tokens = tokens
        self.triggers = []
        self.mentions = []
        self.roles = []
        self.event_relations = set()
        self.relations = set()

        self.vocabs = vocabs
        self.event_type_itos = {i: s for s, i in vocabs['event_type'].items()}
        self.mention_type_itos = {i: s for s, i in vocabs['mention_type'].items()}
        self.role_type_itos = {i: s for s, i in vocabs['role_type'].items()}

        self.trigger_num = len(triggers)
        self.mention_num = len(mentions)

        self.role_num = len(roles)

        self.write_entry = {
            'doc-id': self.doc_id,
            'entry-id': self.entry_id,
            'start': 0,
            'end': 0,
            'segment-text': doc_ori_text,
            'annotation-sets': {
                'basic-events': {
                    'events': {},
                    'includes-relations': defaultdict(list),
                    'span-sets': {}
                }
            }
        }
        self.trigger_id_2_event_id = {}
        self.mention_id_2_span_id = {}
        # initialize info
        for trigger in triggers:
            self.add_trigger(trigger[0], trigger[1], trigger[2])
        for mention in mentions:
            self.add_mention(mention[0], mention[1], mention[2])
        for role in roles:
            self.add_role(role[0], role[1], role[2])

    def add_span(self, span_id, start_token, end_token, span_type):
        self.write_entry['annotation-sets']['basic-events']['span-sets'][span_id] = {
            'spans': [
                {
                    "end": self.tokens[end_token - 1]['end'],
                    "start": self.tokens[start_token]['start'],
                    "string": self.doc_ori_text[self.tokens[start_token]['start']:
                                                self.tokens[end_token - 1]['end']],
                    "synclass": span_type
                }
            ],
            'ssid': span_id
        }

    def add_trigger(self, start, end, label):
        self.triggers.append((start, end, label))
        self.trigger_num = len(self.triggers)
        event_id = f"event-{len(self.write_entry['annotation-sets']['basic-events']['events']) + 1}"
        span_id = f"ss-{len(self.write_entry['annotation-sets']['basic-events']['span-sets']) + 1}"
        self.write_entry['annotation-sets']['basic-events']['events'][event_id] = {
            "agents": [],
            "anchors": span_id,
            "event-type": self.event_type_itos[label],
            "eventid": event_id,
            "patients": [],
            "ref-events": []
        }
        self.add_span(span_id, start, end, span_type='event-anchor')
        self.trigger_id_2_event_id['{}-{}'.format(start, end)] = event_id

    def add_mention(self, start, end, label):
        self.mentions.append((start, end, label))
        self.mention_num = len(self.mentions)

        span_id = f"ss-{len(self.write_entry['annotation-sets']['basic-events']['span-sets']) + 1}"
        self.add_span(span_id, start, end, span_type=self.mention_type_itos[label])
        self.mention_id_2_span_id['{}-{}'.format(start, end)] = span_id

    def add_role(self, idx1, idx2, label):
        """Add an event-argument link edge to the graph.
        :param idx1 (int): Index of the trigger node.
        :param idx2 (int): Index of the mention node.
        """
        self.roles.append((idx1, idx2, label))
        self.role_num = len(self.roles)
        event_id = self.trigger_id_2_event_id['{}-{}'.format(self.triggers[idx1][0], self.triggers[idx1][1])]
        span_id = self.mention_id_2_span_id['{}-{}'.format(
            self.mentions[idx2][0],
            self.mentions[idx2][1]
        )]
        if self.role_type_itos[label] == 'Agent':
            argument = 'agents'
        else:
            assert self.role_type_itos[label] == 'Patient'
            argument = 'patients'

        self.write_entry['annotation-sets']['basic-events']['events'][event_id][argument].append(span_id)

    def add_relation(self, idx1, idx2, label, is_label_1):
        """Add an relation edge to the graph.
        :param idx1 (int): Index of the trigger node 1.
        :param idx2 (int): Index of the trigger node 2.
        :param label (int): Index of the event relation type label.
        :param score (float): Label score.
        """
        relation_type = self.relation_type_itos[label]
        if relation_type == 'Symmetric':
            if idx1 < idx2:
                arg1 = idx1
                arg2 = idx2
            else:
                arg1 = idx2
                arg2 = idx1
            self.relations.add((arg1, arg2, 'Symmetric'))  # arg1 < arg2
        elif relation_type == 'Includes':
            if is_label_1:
                arg1 = idx1
                arg2 = idx2
            else:
                arg1 = idx2
                arg2 = idx1
            self.relations.add((arg1, arg2, 'Includes'))  # implies that "arg1" includes "arg2"
        elif relation_type == 'Included-IN':
            if is_label_1:
                arg1 = idx2
                arg2 = idx1
            else:
                arg1 = idx1
                arg2 = idx2
            self.relations.add((arg1, arg2, 'Includes'))  # implies that "arg1" includes "arg2"
        else:
            print('basic_graph.py:line-91 sent this message: Undefined reference type: {}'.format(
                relation_type))
            exit(1)

    def add_event_relation(self, idx1, idx2, label, is_label_1):
        """Add an event relation edge to the graph.
        :param idx1 (int): Index of the trigger node 1.
        :param idx2 (int): Index of the trigger node 2.
        :param label (int): Index of the event relation type label.
        :param score (float): Label score.
        """
        event_relation_type = self.event_relation_type_itos[label]
        if event_relation_type == 'Includes':
            if is_label_1:
                arg1 = idx1
                arg2 = idx2
            else:
                arg1 = idx2
                arg2 = idx1
        elif event_relation_type == 'Included-IN':
            if is_label_1:
                arg1 = idx2
                arg2 = idx1
            else:
                arg1 = idx1
                arg2 = idx2
        else:
            print('basic_graph.py:line-91 sent this message: Undefined event reference type: {}'.format(
                event_relation_type))
            exit(1)

        self.event_relations.add((arg1, arg2))  # implies that "arg1" includes "arg2"
        arg1_event_id = f"event-{arg1 + 1}"
        arg2_event_id = f"event-{arg2 + 1}"
        self.write_entry['annotation-sets']['basic-events']['events'][arg1_event_id]["ref-events"].append(arg2_event_id)

    @staticmethod
    def empty_graph(doc_id, entry_id, doc_ori_text, tokens, vocabs):
        """Create a graph without any node and edge.
        :param vocabs (dict): Vocabulary object.
        """
        return Graph(doc_id, entry_id, doc_ori_text, tokens, [], [], [], vocabs)
