import copy
import itertools
import json, os
import random
import torch
from torch.utils.data import Dataset
from collections import Counter, namedtuple, defaultdict
from graph import *
from util import parse_entry, get_mid2token_span
from copy import deepcopy
from constants import *

# for trigger
instance_fields = [
    'doc_id', 'entry_id', 'sent_id', 'doc_text', 'tokens', 'piece_idxs', 'attention_mask', 'token_lens',
    'trigger_label_idxs', 'mention_label_idxs',
    'event_type_idxs', 'role_type_idxs', 'relation_type_idxs',
    'graph', 'mention_num', 'trigger_num', 'trigger', 'upos_ids', 'deprel_ids', 'head_ids', 'start', 'end',
    'main_trigger_id',
    'start_token', 'end_token',
    'type_ids', 'subj_positions', 'obj_positions', 'eventids', 'ssids', 'weight',
    'template_type_idxs', 'type_slot_idxs', 'completion_slot_idxs', 'overtime_slot_idxs', 'coordinated_slot_idxs',
    'template_id', 'event_ids', 'mention_ssids', 'slot_type_idxs', 'irrealis_type_idxs', 'template_type'
]

batch_fields = [
    'doc_ids', 'entry_ids', 'sent_ids', 'doc_texts', 'tokens', 'piece_idxs', 'attention_masks', 'token_lens',
    'trigger_label_idxs', 'mention_label_idxs',
    'event_type_idxs', 'role_type_idxs', 'relation_type_idxs',
    'graphs', 'token_nums', 'triggers', 'upos_ids', 'deprel_ids', 'head_ids', 'start', 'end', 'main_trigger_id',
    'start_tokens', 'end_tokens',
    'type_ids', 'subj_positions', 'obj_positions', 'eventids', 'ssids',
    'argument_mask', 'weights',
    'template_type_idxs', 'type_slot_idxs', 'completion_slot_idxs', 'overtime_slot_idxs', 'coordinated_slot_idxs',
    'template_ids', 'event_ids', 'mention_ssids', 'slot_type_idxs', 'irrealis_type_idxs', 'template_types'
]

Instance = namedtuple('Instance', field_names=instance_fields)

Batch = namedtuple('Batch', field_names=batch_fields)


def get_mention_labels(mentions, token_num):
    labels = ['O'] * token_num
    for mention in mentions:
        start, end = mention['start'], mention['end']
        mention_type = mention['mention_type']
        if any([labels[i] != 'O' for i in range(start, end)]):
            continue
        labels[start] = 'B-{}'.format(mention_type)
        for i in range(start + 1, end):
            labels[i] = 'I-{}'.format(mention_type)
    return labels


def remove_overlap_mentions(mentions):
    tokens = [None] * 10000
    mentions_ = []
    id_map = {}
    for mention in mentions:
        start, end = mention['start'], mention['end']
        for i in range(start, end):
            if tokens[i]:
                id_map[mention['id']] = tokens[i]  # tokens[i] returns the id of the overlapping entity
                continue
        mentions_.append(mention)
        for i in range(start, end):
            tokens[i] = mention['id']
    return mentions_, id_map


def get_template_labels(templates, token_num):
    labels = ['O'] * token_num
    for template in templates:
        anchor = template['anchor']
        start, end = anchor['start'], anchor['end']
        event_type = template['template-type']
        labels[start] = 'B-{}'.format(event_type)
        for i in range(start + 1, end):
            labels[i] = 'I-{}'.format(event_type)
    return labels


def get_trigger_labels(events, token_num):
    labels = ['O'] * token_num
    for event in events:
        trigger = event['trigger']
        start, end = trigger['start'], trigger['end']
        event_type = event['event_type']
        labels[start] = 'B-{}'.format(event_type)
        for i in range(start + 1, end):
            labels[i] = 'I-{}'.format(event_type)
    return labels


class TriggerDataset(Dataset):
    def __init__(self, path, config):
        self.path = path
        self.config = config
        self.max_sent_length = 300
        self.data = []
        self.load_data()

    def __len__(self):
        return len(self.data)

    def __getitem__(self, item):
        return self.data[item]

    def load_data(self):
        with open(self.path, 'r', encoding='utf-8') as r:
            for line in r:
                inst = json.loads(line)
                self.data.append(inst)

                if 'train' in self.path and self.config.debug is True and len(self.data) > 500:
                    break

        print('Loaded {} examples from {}'.format(len(self.data), os.path.basename(self.path)))

    def numberize(self, tokenizer, vocabs):
        template_label_stoi = vocabs['template_label']
        type_slot_stoi = vocabs['type_slot']
        completion_slot_stoi = vocabs['completion_slot']
        overtime_slot_stoi = vocabs['overtime_slot']
        coordinated_slot_stoi = vocabs['coordinated_slot']

        trigger_label_stoi = vocabs['trigger_label']
        event_type_stoi = vocabs['event_type']

        data = []
        skip_num = 0
        # debug = []
        for inst in self.data:
            tokens = [tkn['text'] for tkn in inst['tokens']]
            s_pieces = [[p for p in tokenizer.tokenize(w) if p != '▁'] for w in
                        [t['text'] for t in inst['left_tokens']] + tokens + [t['text'] for t in inst['right_tokens']]]
            for ps in s_pieces:
                if len(ps) == 0:
                    ps += ['-']
            s_token_lens = [len(ps) for ps in s_pieces]
            s_pieces = [p for ps in s_pieces for p in ps]
            # Pad word pieces with special tokens
            piece_idxs = tokenizer.encode(
                s_pieces,
                add_special_tokens=True,
                max_length=self.max_sent_length
            )
            if len(piece_idxs) > self.max_sent_length or len(piece_idxs) == 0:
                skip_num += 1
                continue

            pad_num = self.max_sent_length - len(piece_idxs)
            attn_masks = [1] * len(piece_idxs) + [0] * pad_num
            piece_idxs = piece_idxs + [0] * pad_num
            doc_id = inst['doc_id']
            entry_id = inst['entry_id']

            events = inst['events']
            events.sort(key=lambda x: x['trigger']['start'])
            ori_token_num = len(tokens)

            # Triggers
            trigger_labels = get_trigger_labels(events, ori_token_num)
            trigger_label_idxs = [trigger_label_stoi.get(l, 0)
                                  for l in trigger_labels]
            event_types = [e['event_type'] for e in events]
            event_type_idxs = [event_type_stoi.get(l, 0) for l in event_types]
            trigger_list = [(e['trigger']['start'], e['trigger']['end'], e['event_type'], e['event-confidence-score'])
                            for e in events]

            # Graph
            graph = Graph(
                doc_id=inst['doc_id'],
                entry_id=inst['entry_id'],
                sent_id=inst['sent_id'],
                doc_text=inst['doc_ori_text'],
                tokens=inst['tokens'],
                anchors=[],
                triggers=trigger_list,
                mentions=[],
                vocabs=vocabs,
                start=inst['start'],
                end=inst['end']
            )

            instance = Instance(
                doc_id=doc_id,
                entry_id=entry_id,
                sent_id=inst['sent_id'],
                doc_text=inst['doc_ori_text'],
                tokens=inst['tokens'],
                piece_idxs=piece_idxs,
                attention_mask=attn_masks,
                token_lens=s_token_lens,
                trigger_label_idxs=trigger_label_idxs,
                mention_label_idxs=[],
                event_type_idxs=event_type_idxs,
                role_type_idxs=[],
                relation_type_idxs=[],
                graph=graph,
                mention_num=0,
                trigger_num=len(events),
                trigger=0,
                upos_ids=[],
                deprel_ids=[],
                head_ids=[],
                start=inst['start'], end=inst['end'],
                main_trigger_id=-1,
                start_token=inst['start_token'],
                end_token=inst['end_token'],
                type_ids=[],
                subj_positions=[],
                obj_positions=[],
                eventids=[], ssids=[],
                weight=self.config.p1_data_weight if 'phase' in inst and inst['phase'] in [1, 2] else 1.0,
                template_type_idxs=[],
                type_slot_idxs=[], completion_slot_idxs=[],
                overtime_slot_idxs=[], coordinated_slot_idxs=[],
                template_id='', event_ids=[], mention_ssids=[], slot_type_idxs=[], irrealis_type_idxs=[],
                template_type=''
            )
            data.append(instance)

        self.data = data

    def collate_fn(self, batch):
        batch_piece_idxs = []
        batch_attention_masks = []
        batch_token_lens = []

        batch_mention_labels, batch_trigger_labels = [], []

        batch_event_types, batch_role_types, batch_relation_types = [], [], []

        batch_graphs = []
        batch_tokens = []
        batch_token_nums = [len(inst.tokens) for inst in batch]

        doc_ids = [inst.doc_id for inst in batch]
        entry_ids = [inst.entry_id for inst in batch]
        sent_ids = [inst.sent_id for inst in batch]

        max_token_num = max(batch_token_nums)
        max_trigger_num = max([inst.trigger_num for inst in batch] + [1])

        for inst in batch:
            token_num = len(inst.tokens)
            batch_piece_idxs.append(inst.piece_idxs)
            batch_attention_masks.append(inst.attention_mask)
            batch_token_lens.append(inst.token_lens)
            batch_graphs.append(inst.graph)
            batch_tokens.append(inst.tokens)
            # for identification
            batch_trigger_labels.append(inst.trigger_label_idxs +
                                        [0] * (max_token_num - token_num))
            # for classification
            batch_event_types.extend(inst.event_type_idxs +
                                     [-100] * (max_trigger_num - inst.trigger_num))

        batch_piece_idxs = torch.cuda.LongTensor(batch_piece_idxs)
        batch_attention_masks = torch.cuda.FloatTensor(
            batch_attention_masks)

        batch_trigger_labels = torch.cuda.LongTensor(batch_trigger_labels)

        batch_event_types = torch.cuda.LongTensor(batch_event_types)
        batch_token_nums = torch.cuda.LongTensor(batch_token_nums)

        batch_weights = torch.cuda.FloatTensor([inst.weight for inst in batch])

        return Batch(
            doc_ids=doc_ids,
            entry_ids=entry_ids,
            sent_ids=sent_ids,
            doc_texts=[inst.doc_text for inst in batch],
            piece_idxs=batch_piece_idxs,
            attention_masks=batch_attention_masks,
            token_lens=batch_token_lens,
            mention_label_idxs=batch_mention_labels,
            trigger_label_idxs=batch_trigger_labels,
            event_type_idxs=batch_event_types,
            role_type_idxs=batch_role_types,
            relation_type_idxs=batch_relation_types,
            graphs=batch_graphs,
            tokens=[inst.tokens for inst in batch],
            token_nums=batch_token_nums,
            triggers=[],
            upos_ids=[],
            deprel_ids=[],
            head_ids=[],
            start=[inst.start for inst in batch], end=[inst.end for inst in batch],
            main_trigger_id=[inst.main_trigger_id for inst in batch],
            start_tokens=[inst.start_token for inst in batch],
            end_tokens=[inst.end_token for inst in batch],
            type_ids=[],
            subj_positions=[],
            obj_positions=[],
            eventids=[], ssids=[], argument_mask=[], weights=batch_weights,
            template_type_idxs=[],
            type_slot_idxs=[], completion_slot_idxs=[],
            overtime_slot_idxs=[], coordinated_slot_idxs=[],
            template_ids=[], event_ids=[], mention_ssids=[], slot_type_idxs=[], irrealis_type_idxs=[],
            template_types=[]
        )


class ArgumentDataset(Dataset):
    def __init__(self, path, config, eval=False):
        self.path = path
        self.config = config
        self.eval = eval
        self.max_sent_length = 200
        self.data = []
        self.load_data()

    def __len__(self):
        return len(self.data)

    def __getitem__(self, item):
        return self.data[item]

    def load_data(self):
        with open(self.path, 'r', encoding='utf-8') as r:
            for line in r:
                inst = json.loads(line)
                self.data.append(inst)

                if 'train' in self.path and self.config.debug is True and len(self.data) > 500:
                    break

        print('Loaded {} examples from {}'.format(len(self.data), os.path.basename(self.path)))

    def numberize(self, tokenizer, vocabs):
        mention_label_stoi = vocabs['mention_label']
        trigger_label_stoi = vocabs['trigger_label']
        event_type_stoi = vocabs['event_type']
        role_type_stoi = vocabs['role_type']

        data = []
        skip_num = 0
        # debug = []
        for inst in self.data:
            # if self.eval:
            #     inst['mentions'] = []

            tokens = [tkn['text'] for tkn in inst['tokens']]
            doc_id = inst['doc_id']
            entry_id = inst['entry_id']
            mentions = inst['mentions']

            events = inst['events']
            events.sort(key=lambda x: x['trigger']['start'])

            trigger_list = [(e['trigger']['start'], e['trigger']['end'],
                             e['event_type'], e['event-confidence-score'])
                            for e in events]

            ori_token_num = len(tokens)

            trigger_labels = get_trigger_labels(events, ori_token_num)
            trigger_label_idxs = [trigger_label_stoi.get(l, 0)
                                  for l in trigger_labels]
            event_types = [e['event_type'] for e in events]
            event_type_idxs = [event_type_stoi.get(l, 0) for l in event_types]

            spanid_2_span = {}

            for event in events:
                trigger = event['trigger']
                spanid_2_span[event['id']] = [trigger['start'], trigger['end'], 'anchor']

            for mention in mentions:
                spanid_2_span[mention['id']] = [mention['start'], mention['end'], mention['mention_type']]

            for ev_id, cur_event in enumerate(events):
                cur_mention_labels = ['O'] * ori_token_num
                cur_arguments = []
                for arg in events:
                    if arg['eventid'] in cur_event['ref-events']:
                        start, end, mention_label = spanid_2_span[arg['id']]
                        cur_mention_labels[start] = 'B-{}'.format(mention_label)
                        for k in range(start + 1, end):
                            cur_mention_labels[k] = 'I-{}'.format(mention_label)

                        cur_arguments.append((start, end, 'ref-events'))
                for arg in mentions:
                    if arg['ssid'] in cur_event['agents']:
                        start, end, mention_label = spanid_2_span[arg['id']]
                        cur_mention_labels[start] = 'B-{}'.format(mention_label)
                        for k in range(start + 1, end):
                            cur_mention_labels[k] = 'I-{}'.format(mention_label)

                        cur_arguments.append((start, end, 'agents'))
                    elif arg['ssid'] in cur_event['money']:
                        start, end, mention_label = spanid_2_span[arg['id']]
                        cur_mention_labels[start] = 'B-{}'.format(mention_label)
                        for k in range(start + 1, end):
                            cur_mention_labels[k] = 'I-{}'.format(mention_label)

                        cur_arguments.append((start, end, 'money'))
                    elif arg['ssid'] in cur_event['patients']:
                        start, end, mention_label = spanid_2_span[arg['id']]
                        cur_mention_labels[start] = 'B-{}'.format(mention_label)
                        for k in range(start + 1, end):
                            cur_mention_labels[k] = 'I-{}'.format(mention_label)

                        cur_arguments.append((start, end, 'patients'))
                    elif arg['ssid'] in cur_event['time-attachments']:
                        start, end, mention_label = spanid_2_span[arg['id']]
                        cur_mention_labels[start] = 'B-{}'.format(mention_label)
                        for k in range(start + 1, end):
                            cur_mention_labels[k] = 'I-{}'.format(mention_label)

                        cur_arguments.append((start, end, 'time-attachments'))

                cur_arguments.sort(key=lambda x: (x[0], x[1]))

                cur_mention_label_idxs = [mention_label_stoi.get(l, 0) for l in cur_mention_labels]
                # Argument role
                cur_role_type_idxs = [role_type_stoi.get(l, 0) for _, _, l in cur_arguments]

                # contextualized input

                context_tokens = ['</s>', '<s>'] + cur_event['trigger']['text'].split()
                s_pieces = [[p for p in tokenizer.tokenize(w) if p != '▁'] for w in tokens + context_tokens]
                for ps in s_pieces:
                    if len(ps) == 0:
                        ps += ['-']
                s_token_lens = [len(ps) for ps in s_pieces[: len(tokens)]]
                s_pieces = [p for ps in s_pieces for p in ps]
                # Pad word pieces with special tokens
                piece_idxs = tokenizer.encode(
                    s_pieces,
                    add_special_tokens=True,
                    max_length=self.max_sent_length
                )
                if len(piece_idxs) > self.max_sent_length or len(piece_idxs) == 0:
                    skip_num += 1
                    continue

                pad_num = self.max_sent_length - len(piece_idxs)
                attn_masks = [1] * len(piece_idxs) + [0] * pad_num
                piece_idxs = piece_idxs + [0] * pad_num

                # Graph
                graph = MentionGraph(
                    doc_id=inst['doc_id'],
                    entry_id=inst['entry_id'],
                    sent_id=inst['sent_id'],
                    doc_text=inst['doc_ori_text'],
                    tokens=inst['tokens'],
                    triggers=trigger_list,
                    mentions=cur_arguments if not self.eval else [],
                    vocabs=vocabs,
                    start=inst['start'], end=inst['end']
                )
                graph.main_trigger_id = ev_id

                instance = Instance(
                    doc_id=doc_id,
                    entry_id=entry_id,
                    sent_id=inst['sent_id'],
                    doc_text=inst['doc_ori_text'],
                    tokens=inst['tokens'],
                    piece_idxs=piece_idxs,
                    attention_mask=attn_masks,
                    token_lens=s_token_lens,
                    trigger_label_idxs=trigger_label_idxs,
                    mention_label_idxs=cur_mention_label_idxs,
                    event_type_idxs=event_type_idxs,
                    role_type_idxs=cur_role_type_idxs,
                    relation_type_idxs=[],
                    graph=graph,
                    mention_num=len(cur_role_type_idxs),
                    trigger_num=len(trigger_list),
                    trigger=cur_event['trigger']['start'],
                    upos_ids=[vocabs['upos'].get(t['upos'], 1) for t in inst['tokens']],
                    deprel_ids=[vocabs['deprel'].get(t['deprel'], 1) for t in inst['tokens']],
                    head_ids=[t['head'] for t in inst['tokens']],
                    start=inst['start'], end=inst['end'],
                    main_trigger_id=ev_id,
                    start_token=inst['start_token'],
                    end_token=inst['end_token'],
                    type_ids=[],
                    subj_positions=[],
                    obj_positions=[],
                    eventids=[], ssids=[],
                    weight=self.config.p1_data_weight if 'phase' in inst and inst['phase'] in [1, 2] else 1.0,
                    template_type_idxs=[],
                    type_slot_idxs=[], completion_slot_idxs=[],
                    overtime_slot_idxs=[], coordinated_slot_idxs=[],
                    template_id='', event_ids=[], mention_ssids=[], slot_type_idxs=[], irrealis_type_idxs=[],
                    template_type=''
                )
                data.append(instance)
                # debug.append({'inst': inst, 'event': cur_event, 'mentions': cur_mentions, 'labels': cur_mention_labels,
                #               'tokens': tokens})
        self.data = data

        # with open('debu-arg.json', 'w') as f:
        #     json.dump(debug, f, ensure_ascii=False)

    def collate_fn(self, batch):
        batch_piece_idxs = []
        batch_attention_masks = []
        batch_token_lens = []

        batch_mention_labels, batch_trigger_labels = [], []

        batch_event_types, batch_role_types, batch_relation_types = [], [], []

        batch_graphs = []
        batch_tokens = []
        batch_token_nums = [len(inst.tokens) for inst in batch]

        doc_ids = [inst.doc_id for inst in batch]
        entry_ids = [inst.entry_id for inst in batch]
        sent_ids = [inst.sent_id for inst in batch]

        max_token_num = max(batch_token_nums)
        max_mention_num = max([inst.mention_num for inst in batch] + [1])
        max_trigger_num = max([inst.trigger_num for inst in batch] + [1])

        batch_upos_ids, batch_deprel_ids, batch_head_ids = [], [], []

        for inst in batch:
            token_num = len(inst.tokens)
            batch_piece_idxs.append(inst.piece_idxs)
            batch_attention_masks.append(inst.attention_mask)
            batch_token_lens.append(inst.token_lens)
            batch_graphs.append(inst.graph)
            batch_tokens.append(inst.tokens)
            # for identification
            batch_mention_labels.append(inst.mention_label_idxs +
                                        [0] * (max_token_num - token_num))
            batch_trigger_labels.append(inst.trigger_label_idxs +
                                        [0] * (max_token_num - token_num))
            # for classification
            batch_event_types.extend(inst.event_type_idxs +
                                     [-100] * (max_trigger_num - inst.trigger_num))
            batch_role_types.extend(inst.role_type_idxs +
                                    [-100] * (max_mention_num - inst.mention_num))

            batch_upos_ids.append(inst.upos_ids + [0] * (max_token_num - token_num))
            batch_deprel_ids.append(inst.deprel_ids + [0] * (max_token_num - token_num))
            batch_head_ids.append(inst.head_ids + [0] * (max_token_num - token_num))

        batch_piece_idxs = torch.cuda.LongTensor(batch_piece_idxs)
        batch_attention_masks = torch.cuda.FloatTensor(
            batch_attention_masks)

        batch_mention_labels = torch.cuda.LongTensor(batch_mention_labels)
        batch_role_types = torch.cuda.LongTensor(batch_role_types)
        batch_token_nums = torch.cuda.LongTensor(batch_token_nums)
        batch_triggers = torch.cuda.LongTensor([inst.trigger for inst in batch])
        batch_upos_ids = torch.cuda.LongTensor(batch_upos_ids)
        batch_deprel_ids = torch.cuda.LongTensor(batch_deprel_ids)
        batch_head_ids = torch.cuda.LongTensor(batch_head_ids)

        batch_trigger_labels = torch.cuda.LongTensor(batch_trigger_labels)
        batch_event_types = torch.cuda.LongTensor(batch_event_types)
        return Batch(
            doc_ids=doc_ids,
            entry_ids=entry_ids,
            sent_ids=sent_ids,
            doc_texts=[inst.doc_text for inst in batch],
            piece_idxs=batch_piece_idxs,
            attention_masks=batch_attention_masks,
            token_lens=batch_token_lens,
            mention_label_idxs=batch_mention_labels,
            trigger_label_idxs=batch_trigger_labels,
            event_type_idxs=batch_event_types,
            role_type_idxs=batch_role_types,
            relation_type_idxs=batch_relation_types,
            graphs=batch_graphs,
            tokens=[inst.tokens for inst in batch],
            token_nums=batch_token_nums,
            triggers=batch_triggers,
            upos_ids=batch_upos_ids,
            deprel_ids=batch_deprel_ids,
            head_ids=batch_head_ids,
            start=[inst.start for inst in batch], end=[inst.end for inst in batch],
            main_trigger_id=[inst.main_trigger_id for inst in batch],
            start_tokens=[inst.start_token for inst in batch],
            end_tokens=[inst.end_token for inst in batch],
            type_ids=[],
            subj_positions=[],
            obj_positions=[],
            eventids=[], ssids=[], argument_mask=[], weights=torch.cuda.FloatTensor([inst.weight for inst in batch]),
            template_type_idxs=[],
            type_slot_idxs=[], completion_slot_idxs=[],
            overtime_slot_idxs=[], coordinated_slot_idxs=[],
            template_ids=[], event_ids=[], mention_ssids=[], slot_type_idxs=[], irrealis_type_idxs=[],
            template_types=[]
        )


def collapse_templates(granular_templates):
    type2template = {}
    for template in granular_templates:
        if template['template-type'] not in type2template:
            type2template[template['template-type']] = deepcopy(template)
        else:
            # categorical slots
            if 'type' in template and template['type'] != 'O':
                type2template[template['template-type']]['type'] = template['type']
            if 'completion' in template and template['completion'] != 'O':
                type2template[template['template-type']]['completion'] = template['completion']
            if 'over-time' in template and template['over-time'] != 'O':
                type2template[template['template-type']]['over-time'] = template['over-time']
            if 'coordinated' in template and template['coordinated'] != 'O':
                type2template[template['template-type']]['coordinated'] = template['coordinated']

            # span-based slots
            filler_ids = defaultdict(set)

            for slot_name, fillers in type2template[template['template-type']]['slots'].items():
                if isinstance(fillers, list):
                    for filler in fillers:
                        if 'event-id' in filler:
                            filler_ids[slot_name].add(filler['event-id'])
                        elif 'ssid' in filler:
                            filler_ids[slot_name].add(filler['ssid'])

            for slot_name, fillers in template['slots'].items():
                if isinstance(fillers, list):
                    for filler_id in fillers:
                        adding = False
                        if filler_id not in filler_ids[slot_name]:
                            adding = True

                        if adding:
                            if slot_name not in type2template[template['template-type']]['slots']:
                                type2template[template['template-type']]['slots'][slot_name] = []

                            type2template[template['template-type']]['slots'][slot_name].append(filler_id)
                            filler_ids[slot_name].add(filler_id)

                            if filler_id in template['irrealis'] and filler_id not in \
                                    type2template[template['template-type']]['irrealis']:
                                type2template[template['template-type']]['irrealis'][filler_id] = template['irrealis'][
                                    filler_id]

    return list(type2template.values())


class TemplateDataset(Dataset):
    def __init__(self, path, config):
        self.path = path
        self.config = config
        self.max_sent_length = 512
        self.data = []
        self.load_data()

    def __len__(self):
        return len(self.data)

    def __getitem__(self, item):
        return self.data[item]

    def load_data(self):
        with open(self.path, 'r', encoding='utf-8') as r:
            for line in r:
                inst = json.loads(line)
                self.data.append(inst)

                if 'train' in self.path and self.config.debug is True and len(self.data) > 500:
                    break

        print('Loaded {} examples from {}'.format(len(self.data), os.path.basename(self.path)))

    def numberize(self, tokenizer, vocabs):
        self.vocabs = vocabs
        template_type_stoi = vocabs['template-type']
        type_slot_stoi = vocabs['type_slot']
        completion_slot_stoi = vocabs['completion_slot']
        overtime_slot_stoi = vocabs['overtime_slot']
        coordinated_slot_stoi = vocabs['coordinated_slot']
        basic_type_stoi = vocabs['type']

        data = []
        skip_num = 0
        # debug = []
        for document in self.data:
            if len(document['sentences']) == 0:
                continue

            granular_templates = document['granular-templates']
            granular_templates = collapse_templates(granular_templates)

            inst = {
                'tokens': [token for sent in document['sentences'] for token in sent['tokens']],
                'doc_id': document['sentences'][0]['doc_id'],
                'entry_id': document['sentences'][0]['entry_id'],
                'templates': granular_templates,
                'sent_id': 0,
                'doc_ori_text': document['sentences'][0]['doc_ori_text'],
                'phase': document['sentences'][0]['phase'] if 'phase' in document['sentences'][0] else 2
            }

            event_tags = []
            for sent in document['sentences']:
                tags = ['O' for _ in sent['tokens']]
                for event in sent['events']:
                    for k in range(event['trigger']['start'], event['trigger']['end']):
                        tags[k] = event['event_type']
                event_tags.extend(tags)
            type_ids = [basic_type_stoi.get(tag, 0) for tag in event_tags]
            #######################################################
            tokens = [tkn['text'] for tkn in inst['tokens']]
            s_pieces = [[p for p in tokenizer.tokenize(w) if p != '▁'] for w in tokens]
            for ps in s_pieces:
                if len(ps) == 0:
                    ps += ['-']
            s_token_lens = [len(ps) for ps in s_pieces]
            marks = [[1] + [0] * (tlen - 1) for tlen in s_token_lens]
            marks = [p for ps in marks for p in ps]
            s_pieces = [p for ps in s_pieces for p in ps]
            # Pad word pieces with special tokens
            piece_idxs = tokenizer.encode(
                s_pieces,
                add_special_tokens=True,
                max_length=self.max_sent_length,
                truncation=True
            )

            marks = marks[: len(piece_idxs)]
            cropped_token_num = sum(marks) - 1
            type_ids = type_ids[: cropped_token_num]
            s_token_lens = s_token_lens[: cropped_token_num]

            pad_num = self.max_sent_length - len(piece_idxs)
            attn_masks = [1] * len(piece_idxs) + [0] * pad_num
            piece_idxs = piece_idxs + [0] * pad_num
            doc_id = inst['doc_id']
            entry_id = inst['entry_id']

            templates = inst['templates']
            type2template = {}
            for template in templates:
                type2template[template['template-type']] = template

            template_type_idxs = {}
            type_slot_idxs = {}
            completion_slot_idxs = {}
            overtime_slot_idxs = {}
            coordinated_slot_idxs = {}

            for template_type in template_type_stoi:
                if template_type == 'O':
                    continue

                if template_type in type2template:
                    cur_template = type2template[template_type]

                    template_type_idxs[template_type] = 1
                    type_slot_idxs[template_type] = type_slot_stoi.get(cur_template['type'], 0)
                    completion_slot_idxs[template_type] = completion_slot_stoi.get(cur_template['completion'], 0)
                    overtime_slot_idxs[template_type] = overtime_slot_stoi.get(cur_template['over-time'], 0)
                    coordinated_slot_idxs[template_type] = coordinated_slot_stoi.get(cur_template['coordinated'], 0)
                else:
                    template_type_idxs[template_type] = 0
                    type_slot_idxs[template_type] = 0
                    completion_slot_idxs[template_type] = 0
                    overtime_slot_idxs[template_type] = 0
                    coordinated_slot_idxs[template_type] = 0

            instance = Instance(
                doc_id=doc_id,
                entry_id=entry_id,
                sent_id=inst['sent_id'],
                doc_text=inst['doc_ori_text'],
                tokens=[0 for _ in range(cropped_token_num)],
                piece_idxs=piece_idxs,
                attention_mask=attn_masks,
                token_lens=s_token_lens,
                trigger_label_idxs=[],
                mention_label_idxs=[],
                event_type_idxs=[],
                role_type_idxs=[],
                relation_type_idxs=[],
                graph=None,
                mention_num=0,
                trigger_num=0,
                trigger=0,
                upos_ids=[],
                deprel_ids=[],
                head_ids=[],
                start=inst['tokens'][0]['start'], end=inst['tokens'][-1]['end'],
                main_trigger_id=-1,
                start_token=0,
                end_token=cropped_token_num,
                type_ids=type_ids,
                subj_positions=[],
                obj_positions=[],
                eventids=[], ssids=[],
                weight=self.config.p1_data_weight if 'phase' in inst and inst['phase'] in [1, 2] else 1.0,
                template_type_idxs=template_type_idxs,
                type_slot_idxs=type_slot_idxs, completion_slot_idxs=completion_slot_idxs,
                overtime_slot_idxs=overtime_slot_idxs, coordinated_slot_idxs=coordinated_slot_idxs,
                template_id='', event_ids=[], mention_ssids=[], slot_type_idxs=[], irrealis_type_idxs=[],
                template_type=''
            )
            data.append(instance)

        self.data = data

    def collate_fn(self, batch):
        batch_piece_idxs = []
        batch_attention_masks = []
        batch_token_lens = []

        batch_mention_labels, batch_trigger_labels = [], []

        batch_event_types, batch_role_types, batch_relation_types = [], [], []

        batch_graphs = []
        batch_tokens = []

        batch_token_nums = [len(inst.tokens) for inst in batch]
        max_token_num = max(batch_token_nums)

        doc_ids = [inst.doc_id for inst in batch]
        entry_ids = [inst.entry_id for inst in batch]
        sent_ids = [inst.sent_id for inst in batch]

        batch_template_type_idxs = defaultdict(list)
        batch_type_slot_idxs = defaultdict(list)
        batch_completion_slot_idxs = defaultdict(list)
        batch_overtime_slot_idxs = defaultdict(list)
        batch_coordinated_slot_idxs = defaultdict(list)
        batch_type_ids = []

        for inst in batch:
            batch_piece_idxs.append(inst.piece_idxs)
            batch_attention_masks.append(inst.attention_mask)
            batch_token_lens.append(inst.token_lens)
            batch_graphs.append(inst.graph)
            batch_tokens.append(inst.tokens)
            batch_type_ids.append(inst.type_ids + [0] * (max_token_num - len(inst.tokens)))

            for template_type in self.vocabs['template-type']:
                if template_type == 'O':
                    continue

                batch_template_type_idxs[template_type].append(inst.template_type_idxs[template_type])
                batch_type_slot_idxs[template_type].append(inst.type_slot_idxs[template_type])
                batch_completion_slot_idxs[template_type].append(inst.completion_slot_idxs[template_type])
                batch_overtime_slot_idxs[template_type].append(inst.overtime_slot_idxs[template_type])
                batch_coordinated_slot_idxs[template_type].append(inst.coordinated_slot_idxs[template_type])

        batch_piece_idxs = torch.cuda.LongTensor(batch_piece_idxs)
        batch_attention_masks = torch.cuda.FloatTensor(
            batch_attention_masks)

        batch_token_nums = torch.cuda.LongTensor(batch_token_nums)

        batch_weights = torch.cuda.FloatTensor([inst.weight for inst in batch])

        batch_type_ids = torch.cuda.FloatTensor(batch_type_ids)

        for template_type in self.vocabs['template-type']:
            batch_template_type_idxs[template_type] = torch.cuda.LongTensor(batch_template_type_idxs[template_type])
            batch_type_slot_idxs[template_type] = torch.cuda.LongTensor(batch_type_slot_idxs[template_type])
            batch_completion_slot_idxs[template_type] = torch.cuda.LongTensor(batch_completion_slot_idxs[template_type])
            batch_overtime_slot_idxs[template_type] = torch.cuda.LongTensor(batch_overtime_slot_idxs[template_type])
            batch_coordinated_slot_idxs[template_type] = torch.cuda.LongTensor(
                batch_coordinated_slot_idxs[template_type])

        return Batch(
            doc_ids=doc_ids,
            entry_ids=entry_ids,
            sent_ids=sent_ids,
            doc_texts=[inst.doc_text for inst in batch],
            piece_idxs=batch_piece_idxs,
            attention_masks=batch_attention_masks,
            token_lens=batch_token_lens,
            mention_label_idxs=batch_mention_labels,
            trigger_label_idxs=[],
            event_type_idxs=[],
            role_type_idxs=batch_role_types,
            relation_type_idxs=batch_relation_types,
            graphs=batch_graphs,
            tokens=[inst.tokens for inst in batch],
            token_nums=batch_token_nums,
            triggers=[],
            upos_ids=[],
            deprel_ids=[],
            head_ids=[],
            start=[inst.start for inst in batch], end=[inst.end for inst in batch],
            main_trigger_id=[inst.main_trigger_id for inst in batch],
            start_tokens=[inst.start_token for inst in batch],
            end_tokens=[inst.end_token for inst in batch],
            type_ids=batch_type_ids,
            subj_positions=[],
            obj_positions=[],
            eventids=[], ssids=[], argument_mask=[], weights=batch_weights,
            template_type_idxs=batch_template_type_idxs,
            type_slot_idxs=batch_type_slot_idxs, completion_slot_idxs=batch_completion_slot_idxs,
            overtime_slot_idxs=batch_overtime_slot_idxs, coordinated_slot_idxs=batch_coordinated_slot_idxs,
            template_ids=[], event_ids=[], mention_ssids=[], slot_type_idxs=[], irrealis_type_idxs=[],
            template_types=[]
        )


slot_instance_fields = [
    'doc_id', 'entry_id', 'doc_ori_text', 'tokens', 'piece_idxs', 'attention_masks', 'token_lens',
    'template_type', 'template_id', 'slot_ids', 'irrealis_idxs',
    'graph', 'mention_num', 'trigger_num', 'slot_type_idxs',
    'trigger_features', 'mention_features', 'start_token', 'end_token',
    'basic_graph', 'weight'
]

slot_batch_fields = [
    'doc_ids', 'entry_ids', 'doc_ori_texts', 'piece_idxs', 'attention_masks', 'token_lens',
    'template_types', 'template_ids', 'slot_ids', 'irrealis_idxs',
    'graphs', 'tokens', 'token_nums', 'slot_type_idxs',
    'trigger_features', 'mention_features', 'start_tokens', 'end_tokens',
    'basic_graphs', 'weights'
]

SlotInstance = namedtuple('SlotInstance', field_names=slot_instance_fields)

SlotBatch = namedtuple('SlotBatch', field_names=slot_batch_fields)


def get_trigger_features(events, vocabs):
    out_features = []
    for event in events:
        feat_vec = get_zero_features(vocabs)
        feat_vec[0] = 1.
        feat_vec[1 + vocabs['event_type'][event['event_type']]] = 1.
        feat_vec[1 + len(vocabs['event_type'])] = len(event['agents'])
        feat_vec[1 + len(vocabs['event_type']) + 1] = len(event['patients'])

        out_features.append(feat_vec)
    return out_features


def get_zero_features(vocabs):
    features = [0. for _ in
                range(1 + len(vocabs['event_type']) + 2 + len(vocabs['mention_type']) + 2 * len(vocabs['event_type']))]
    return features


def get_mention_features(mentions, vocabs):
    out_features = []
    event_range = 1 + len(vocabs['event_type']) + 2
    event_type_num = len(vocabs['event_type'])
    mention_type_num = len(vocabs['mention_type'])
    for mention in mentions:
        feat_vec = get_zero_features(vocabs)
        feat_vec[0] = 0.
        feat_vec[event_range + vocabs['mention_type'].get(mention['mention_type'], 0)] = 1.

        out_features.append(feat_vec)
    return out_features


def create_basic_graph(events, mentions):
    trigger_num = len(events)
    mention_num = len(mentions)
    graph = [[0] * (trigger_num + mention_num) for _ in range(trigger_num + mention_num)]
    eventid2idx = {}
    ssid2idx = {}

    for idx, event in enumerate(events):
        eventid2idx[event['eventid']] = idx

    for idx, mention in enumerate(mentions):
        ssid2idx[mention['ssid']] = trigger_num + idx

    for event in events:
        idx1 = eventid2idx[event['eventid']]
        for ssid in event['agents']:
            if ssid in ssid2idx:
                idx2 = ssid2idx[ssid]
                graph[idx1][idx2] = 1.0
                graph[idx2][idx1] = 1.0

        for ssid in event['patients']:
            if ssid in ssid2idx:
                idx2 = ssid2idx[ssid]
                graph[idx1][idx2] = 1.0
                graph[idx2][idx1] = 1.0

        for eventid in event['ref-events']:
            if eventid in eventid2idx:
                idx2 = eventid2idx[eventid]
                graph[idx1][idx2] = 1.0
                graph[idx2][idx1] = 1.0
    return graph


class SlotDataset(Dataset):
    def __init__(self, path, config):
        self.path = path
        self.config = config
        self.max_sent_length = 300
        self.data = []
        self.load_data()

    def __len__(self):
        return len(self.data)

    def __getitem__(self, item):
        return self.data[item]

    def load_data(self):
        with open(self.path, 'r', encoding='utf-8') as r:
            for line in r:
                inst = json.loads(line)
                self.data.append(inst)
                if 'train' in self.path and self.config.debug is True and len(self.data) > 500:
                    break

    def numberize(self, tokenizer, vocabs):
        print(self.path)
        self.vocabs = vocabs
        event_type_stoi = vocabs['event_type']
        mention_type_stoi = vocabs['mention_type']
        irrealis_stoi = vocabs['irrealis_type']
        data = []
        skip_num = 0
        for document in self.data:
            granular_templates = document['granular-templates']
            granular_templates = collapse_templates(granular_templates)
            sent_list = document['sentences']

            for template in granular_templates:
                context_tokens = []
                context_tokens += TEMPLATE_TYPE2TOKENS[template['template-type']].split()
                context_tokens += ['</s>', '<s>']

                for inst in sent_list:
                    tokens = [tkn['text'] for tkn in inst['tokens']]
                    s_pieces = [[p for p in tokenizer.tokenize(w) if p != '▁'] for w in context_tokens + tokens]
                    for ps in s_pieces:
                        if len(ps) == 0:
                            ps += ['-']
                    s_token_lens = [len(ps) for ps in s_pieces]
                    s_pieces = [p for ps in s_pieces for p in ps]

                    # Pad word pieces with special tokens
                    piece_idxs = tokenizer.encode(
                        s_pieces,
                        add_special_tokens=True,
                        max_length=self.max_sent_length
                    )
                    if len(piece_idxs) > self.max_sent_length:
                        skip_num += 1
                        continue

                    attn_masks = [1] * len(piece_idxs)
                    piece_idxs = piece_idxs
                    if len(piece_idxs) == 0:
                        skip_num += 1
                        continue
                    doc_id = inst['doc_id']
                    entry_id = inst['entry_id']
                    mentions = inst['mentions']
                    mentions.sort(key=lambda x: x['start'])
                    events = inst['events']
                    events.sort(key=lambda x: x['trigger']['start'])

                    # Triggers
                    trigger_list = [(e['trigger']['start'], e['trigger']['end'],
                                     event_type_stoi.get(e['event_type'], 0))
                                    for e in events]
                    trigger_features = get_trigger_features(events, vocabs)
                    # Mentions
                    mention_list = [(m['start'], m['end'], mention_type_stoi.get(m['mention_type'], 0))
                                    for m in mentions]
                    mention_features = get_mention_features(mentions, vocabs)

                    basic_graph = create_basic_graph(events, mentions)

                    # slot types and slot attributes
                    slot_type_idxs = []
                    irrealis_idxs = []
                    slot_ids = []
                    for slot_id, slot_name in enumerate(vocabs['template-type2slots'][template['template-type']]):
                        tmp_slot = [[], []]
                        tmp_id = [[], []]
                        tmp_irrealis = [[], []]
                        for evid, event in enumerate(events):
                            if slot_name in template['slots'] and event['eventid'] in template['slots'][slot_name]:
                                tmp_slot[0].append(1)
                            else:
                                tmp_slot[0].append(0)
                            tmp_id[0].append(event['eventid'])
                            if event['eventid'] in template['irrealis']:
                                tmp_irrealis[0].append(irrealis_stoi.get(template['irrealis'][event['eventid']], 0))
                            else:
                                tmp_irrealis[0].append(0)

                        for mid, mention in enumerate(mentions):
                            if slot_name in template['slots'] and mention['ssid'] in template['slots'][slot_name]:
                                tmp_slot[1].append(1)
                            else:
                                tmp_slot[1].append(0)
                            tmp_id[1].append(mention['ssid'])
                            if mention['ssid'] in template['irrealis']:
                                tmp_irrealis[1].append(irrealis_stoi.get(template['irrealis'][mention['ssid']], 0))
                            else:
                                tmp_irrealis[1].append(0)

                        slot_type_idxs.append(tmp_slot)
                        irrealis_idxs.append(tmp_irrealis)
                        slot_ids.append(tmp_id)

                    # Graph
                    graph = SlotGraph(
                        doc_id=inst['doc_id'],
                        entry_id=inst['entry_id'],
                        doc_ori_text=inst['doc_ori_text'],
                        tokens=inst['tokens'],
                        triggers=trigger_list,
                        mentions=mention_list,
                        roles=[],
                        vocabs=vocabs
                    )
                    graph.write_entry['start'] = inst['start']
                    graph.write_entry['end'] = inst['end']

                    instance = SlotInstance(
                        doc_id=doc_id,
                        entry_id=entry_id,
                        doc_ori_text=inst['doc_ori_text'],
                        tokens=inst['tokens'],
                        piece_idxs=piece_idxs,
                        attention_masks=attn_masks,
                        token_lens=s_token_lens,
                        template_type=template['template-type'],
                        template_id=template['template-id'],
                        slot_type_idxs=slot_type_idxs,
                        slot_ids=slot_ids,
                        irrealis_idxs=irrealis_idxs,
                        graph=graph,
                        mention_num=len(mentions),
                        trigger_num=len(events),
                        trigger_features=trigger_features,
                        mention_features=mention_features,
                        start_token=inst['start_token'] + len(context_tokens),
                        end_token=inst['end_token'] + len(context_tokens),
                        basic_graph=basic_graph,
                        weight=self.config.p1_data_weight if 'phase' in document and document['phase'] in [1,
                                                                                                           2] else 1.0,
                    )
                    data.append(instance)
        self.data = data

    def collate_fn(self, batch):
        batch_piece_idxs = []
        batch_attention_masks = []
        batch_token_lens = []

        batch_graphs = []
        batch_tokens = []
        batch_token_nums = [len(inst.tokens) for inst in batch]

        doc_ids = [inst.doc_id for inst in batch]
        entry_ids = [inst.entry_id for inst in batch]
        doc_ori_texts = [inst.doc_ori_text for inst in batch]

        max_mention_num = max([inst.mention_num for inst in batch] + [1])
        max_trigger_num = max([inst.trigger_num for inst in batch] + [1])
        max_candidate_num = max_trigger_num + max_mention_num
        max_wordpiece_num = max([len(inst.piece_idxs) for inst in batch])

        batch_basic_graphs = []

        batch_slot_type_idxs = []
        batch_irrealis_idxs = []
        batch_slot_ids = []

        batch_trigger_features = []
        batch_mention_features = []

        for inst in batch:
            batch_piece_idxs.append(inst.piece_idxs + [0] * (max_wordpiece_num - len(inst.piece_idxs)))
            batch_attention_masks.append(inst.attention_masks + [0] * (max_wordpiece_num - len(inst.piece_idxs)))
            batch_token_lens.append(inst.token_lens)
            batch_graphs.append(inst.graph)
            batch_tokens.append(inst.tokens)
            # trigger and mention features
            batch_trigger_features.append(
                inst.trigger_features + [get_zero_features(self.vocabs)] * (max_trigger_num - inst.trigger_num))
            batch_mention_features.append(
                inst.mention_features + [get_zero_features(self.vocabs)] * (max_mention_num - inst.mention_num))

            # basic graphs
            graph = []  # [max trg num + max ent num, max trg num + max ent num]
            for trgid in range(inst.trigger_num):
                row = inst.basic_graph[trgid][:inst.trigger_num]
                row.extend([0] * (max_trigger_num - inst.trigger_num))
                row.extend(inst.basic_graph[trgid][inst.trigger_num:])
                row.extend([0] * (max_mention_num - inst.mention_num))
                graph.append(row)

            for _ in range(max_trigger_num - inst.trigger_num):
                graph.append([1] * (max_trigger_num + max_mention_num))

            for entid in range(inst.trigger_num, inst.trigger_num + inst.mention_num):
                row = inst.basic_graph[entid][:inst.trigger_num]
                row.extend([0] * (max_trigger_num - inst.trigger_num))
                row.extend(inst.basic_graph[entid][inst.trigger_num:])
                row.extend([0] * (max_mention_num - inst.mention_num))
                graph.append(row)

            for _ in range(max_mention_num - inst.mention_num):
                graph.append([1] * (max_trigger_num + max_mention_num))

            batch_basic_graphs.append(graph)

            # for classification
            slot_type_idxs = []
            irrealis_idxs = []
            slot_ids = []

            for slid, slot_name in enumerate(self.vocabs['template-type2slots'][inst.template_type]):
                trigger_types = inst.slot_type_idxs[slid][0]
                mention_types = inst.slot_type_idxs[slid][1]
                candidate_types = trigger_types + [-100] * (max_trigger_num - inst.trigger_num) + \
                                  mention_types + [-100] * (max_mention_num - inst.mention_num)
                slot_type_idxs.append(candidate_types)

                # slot id
                trigger_slot_ids = inst.slot_ids[slid][0]
                mention_slot_ids = inst.slot_ids[slid][1]
                candidate_slot_ids = trigger_slot_ids + ['skip'] * (max_trigger_num - inst.trigger_num) + \
                                     mention_slot_ids + ['skip'] * (max_mention_num - inst.mention_num)
                slot_ids.append(candidate_slot_ids)

                # irrealis
                trigger_irrealis_ids = inst.irrealis_idxs[slid][0]
                mention_irrealis_ids = inst.irrealis_idxs[slid][1]
                candidate_irrealis_ids = trigger_irrealis_ids + [-100] * (max_trigger_num - inst.trigger_num) + \
                                         mention_irrealis_ids + [-100] * (max_mention_num - inst.mention_num)
                irrealis_idxs.append(candidate_irrealis_ids)

            slot_type_idxs = torch.cuda.LongTensor(slot_type_idxs)
            irrealis_idxs = torch.cuda.LongTensor(irrealis_idxs)

            batch_slot_type_idxs.append(slot_type_idxs)
            batch_irrealis_idxs.append(irrealis_idxs)
            batch_slot_ids.append(slot_ids)

        batch_piece_idxs = torch.cuda.LongTensor(batch_piece_idxs)
        batch_attention_masks = torch.cuda.FloatTensor(
            batch_attention_masks)
        batch_token_nums = torch.cuda.LongTensor(batch_token_nums)

        batch_trigger_features = torch.cuda.FloatTensor(batch_trigger_features)
        batch_mention_features = torch.cuda.FloatTensor(batch_mention_features)

        batch_basic_graphs = torch.cuda.FloatTensor(batch_basic_graphs)

        batch_weights = [float(inst.weight) for inst in batch]

        return SlotBatch(
            doc_ids=doc_ids,
            entry_ids=entry_ids,
            doc_ori_texts=doc_ori_texts,
            piece_idxs=batch_piece_idxs,
            attention_masks=batch_attention_masks,
            token_lens=batch_token_lens,
            graphs=batch_graphs,
            tokens=[inst.tokens for inst in batch],
            token_nums=batch_token_nums,
            slot_type_idxs=batch_slot_type_idxs,
            irrealis_idxs=batch_irrealis_idxs,
            slot_ids=batch_slot_ids,
            template_types=[inst.template_type for inst in batch],
            template_ids=[inst.template_id for inst in batch],
            trigger_features=batch_trigger_features,
            mention_features=batch_mention_features,
            start_tokens=[inst.start_token for inst in batch],
            end_tokens=[inst.end_token for inst in batch], basic_graphs=batch_basic_graphs,
            weights=batch_weights
        )


class EventCorefDataset(Dataset):
    def __init__(self, path, config):
        self.path = path
        self.config = config
        self.max_sent_length = 300
        self.data = []
        self.load_data()

    def __len__(self):
        return len(self.data)

    def __getitem__(self, item):
        return self.data[item]

    def load_data(self):
        with open(self.path, 'r', encoding='utf-8') as r:
            for line in r:
                inst = json.loads(line)
                self.data.append(inst)

                if 'train' in self.path and self.config.debug is True and len(self.data) > 500:
                    break

        print('Loaded {} examples from {}'.format(len(self.data), os.path.basename(self.path)))

    def numberize(self, tokenizer, vocabs):
        event_type_stoi = vocabs['event_type']
        coref_type_stoi = {
            'O': 0, 'RefEvent': 1, 'CorefEvent': 2
        }
        type_stoi = vocabs['type']

        data = []
        skip_num = 0
        for inst in self.data:
            tokens = [tkn['text'] for tkn in inst['tokens']]
            s_pieces = [[p for p in tokenizer.tokenize(w) if p != '▁'] for w in
                        [t['text'] for t in inst['left_tokens']] + tokens + [t['text'] for t in inst['right_tokens']]]
            for ps in s_pieces:
                if len(ps) == 0:
                    ps += ['-']
            s_token_lens = [len(ps) for ps in s_pieces]
            s_pieces = [p for ps in s_pieces for p in ps]
            # Pad word pieces with special tokens
            piece_idxs = tokenizer.encode(
                s_pieces,
                add_special_tokens=True,
                max_length=self.max_sent_length
            )
            if len(piece_idxs) > self.max_sent_length or len(piece_idxs) == 0:
                skip_num += 1
                continue

            pad_num = self.max_sent_length - len(piece_idxs)
            attn_masks = [1] * len(piece_idxs) + [0] * pad_num
            piece_idxs = piece_idxs + [0] * pad_num
            doc_id = inst['doc_id']
            entry_id = inst['entry_id']

            events = inst['events']
            events.sort(key=lambda x: x['trigger']['start'])
            ori_token_num = len(tokens)

            # Triggers
            event_types = [e['event_type'] for e in events]
            event_type_idxs = [event_type_stoi.get(l, 0) for l in event_types]
            trigger_list = [(e['trigger']['start'], e['trigger']['end'], e['event_type'], e['event-confidence-score'])
                            for e in events]

            # Graph
            graph = Graph(
                doc_id=inst['doc_id'],
                entry_id=inst['entry_id'],
                sent_id=inst['sent_id'],
                doc_text=inst['doc_ori_text'],
                tokens=inst['tokens'],
                anchors=[],
                triggers=trigger_list,
                mentions=[],
                vocabs=vocabs,
                start=inst['start'],
                end=inst['end']
            )

            # type ids
            type_ids = [0] * ori_token_num
            for event in events:
                for k in range(event['trigger']['start'], event['trigger']['end']):
                    type_ids[k] = type_stoi[event['event_type']]

            # event coreference links
            coref_type_idxs = []
            for hid, head_event in enumerate(events):
                row = []
                for tid, tail_event in enumerate(events):
                    if tid == hid:
                        row.append(-100)
                        continue

                    if head_event['eventid'] == tail_event['eventid']:
                        row.append(coref_type_stoi['CorefEvent'])
                    elif tail_event['eventid'] in head_event['ref-events']:
                        row.append(coref_type_stoi['RefEvent'])
                    else:
                        row.append(coref_type_stoi['O'])
                coref_type_idxs.append(row)

            instance = Instance(
                doc_id=doc_id,
                entry_id=entry_id,
                sent_id=inst['sent_id'],
                doc_text=inst['doc_ori_text'],
                tokens=inst['tokens'],
                piece_idxs=piece_idxs,
                attention_mask=attn_masks,
                token_lens=s_token_lens,
                trigger_label_idxs=[],
                mention_label_idxs=[],
                event_type_idxs=event_type_idxs,
                role_type_idxs=coref_type_idxs,
                relation_type_idxs=[],
                graph=graph,
                mention_num=0,
                trigger_num=len(events),
                trigger=0,
                upos_ids=[],
                deprel_ids=[],
                head_ids=[],
                start=inst['start'], end=inst['end'],
                main_trigger_id=-1,
                start_token=inst['start_token'],
                end_token=inst['end_token'],
                type_ids=type_ids,
                subj_positions=[],
                obj_positions=[],
                eventids=[event['eventid'] for event in events], ssids=[],
                weight=self.config.p1_data_weight if 'phase' in inst and inst['phase'] in [1, 2] else 1.0,
                template_type_idxs=[],
                type_slot_idxs=[], completion_slot_idxs=[],
                overtime_slot_idxs=[], coordinated_slot_idxs=[],
                template_id='', event_ids=[], mention_ssids=[], slot_type_idxs=[], irrealis_type_idxs=[],
                template_type=''
            )
            data.append(instance)

        self.data = data

    def collate_fn(self, batch):
        batch_piece_idxs = []
        batch_attention_masks = []
        batch_token_lens = []

        batch_mention_labels, batch_trigger_labels = [], []

        batch_event_types, batch_role_types, batch_relation_types = [], [], []

        batch_graphs = []
        batch_tokens = []
        batch_token_nums = [len(inst.tokens) for inst in batch]

        doc_ids = [inst.doc_id for inst in batch]
        entry_ids = [inst.entry_id for inst in batch]
        sent_ids = [inst.sent_id for inst in batch]

        max_token_num = max(batch_token_nums)
        max_trigger_num = max([inst.trigger_num for inst in batch] + [1])

        batch_type_ids = []

        for inst in batch:
            token_num = len(inst.tokens)
            batch_piece_idxs.append(inst.piece_idxs)
            batch_attention_masks.append(inst.attention_mask)
            batch_token_lens.append(inst.token_lens)
            batch_graphs.append(inst.graph)
            batch_tokens.append(inst.tokens)

            batch_type_ids.append(inst.type_ids + [0] * (max_token_num - token_num))

            # for classification
            batch_event_types.extend(inst.event_type_idxs +
                                     [-100] * (max_trigger_num - inst.trigger_num))
            for row in inst.role_type_idxs:
                batch_role_types.extend(
                    row + [-100] * (max_trigger_num - inst.trigger_num))
            batch_role_types.extend(
                [-100] * max_trigger_num * (max_trigger_num - inst.trigger_num))

        batch_piece_idxs = torch.cuda.LongTensor(batch_piece_idxs)
        batch_attention_masks = torch.cuda.FloatTensor(
            batch_attention_masks)

        batch_event_types = torch.cuda.LongTensor(batch_event_types)
        batch_token_nums = torch.cuda.LongTensor(batch_token_nums)
        batch_role_types = torch.cuda.LongTensor(batch_role_types)
        batch_type_ids = torch.cuda.LongTensor(batch_type_ids)

        return Batch(
            doc_ids=doc_ids,
            entry_ids=entry_ids,
            sent_ids=sent_ids,
            doc_texts=[inst.doc_text for inst in batch],
            piece_idxs=batch_piece_idxs,
            attention_masks=batch_attention_masks,
            token_lens=batch_token_lens,
            mention_label_idxs=batch_mention_labels,
            trigger_label_idxs=batch_trigger_labels,
            event_type_idxs=batch_event_types,
            role_type_idxs=batch_role_types,
            relation_type_idxs=batch_relation_types,
            graphs=batch_graphs,
            tokens=[inst.tokens for inst in batch],
            token_nums=batch_token_nums,
            triggers=[],
            upos_ids=[],
            deprel_ids=[],
            head_ids=[],
            start=[inst.start for inst in batch], end=[inst.end for inst in batch],
            main_trigger_id=[inst.main_trigger_id for inst in batch],
            start_tokens=[inst.start_token for inst in batch],
            end_tokens=[inst.end_token for inst in batch],
            type_ids=batch_type_ids,
            subj_positions=[],
            obj_positions=[],
            eventids=[inst.eventids for inst in batch], ssids=[], argument_mask=[],
            weights=torch.cuda.FloatTensor([inst.weight for inst in batch]),
            template_type_idxs=[],
            type_slot_idxs=[], completion_slot_idxs=[],
            overtime_slot_idxs=[], coordinated_slot_idxs=[],
            template_ids=[], event_ids=[], mention_ssids=[], slot_type_idxs=[], irrealis_type_idxs=[],
            template_types=[]
        )
