import json
import logging
import subprocess
import sys
import argparse

from enum import Enum
from collections import OrderedDict 


COMMANDS_JSON = "/code/docker/commands.json"
TASK_JSON = "/app/tasks.json"
TRAIN_DATA = "/app/train_data.bp.json"
TEST_DATA = "/app/test_data.bp.json"
RESULTS = "/app/results.json"
LOG_FILE = "/app/tasks.log"

# These are ordered
TASKS = ['extract-granular-events']


# This is a convenience function that maps from the tasks defined in
# /app/tasks.json to actual commands with arguments.  The commands and arguments
# are stored in the path pointed to by the COMMANDS variable.
def load_commands():
    """
    Load the list of command line processes to start for a given task.
    """
    task_commands = {}
    with open(COMMANDS_JSON, 'r') as fin:
        data = json.load(fin)
        for task in data:
            if task not in TASKS:
                raise RuntimeError(
                    f'Unrecognized task in commands.json: {task}')
            task_commands[task] = data[task].split()
    logging.info(f'Loaded {len(task_commands.keys())} task commands.')
    return task_commands


# Per the evaluation plan, /app/tasks.json will have a format like:
# {"test-id": "used by T&E to track evaluation runs",
#  "task-set":
#  {"extract-abstract-events": {"perform?": true,
#                              "train-on-hidden-data?": true
#                             },
#  "extract-basic-events": {"perform?": false},
#  "extract-granular-events": {"perform?": false},
#  "find-relevant-docs.automatic": {"perform?": false}}
def parse_tasks_json(json_file, train_data, test_data, results):
    """
    Extract tasks from json task file.
    """
    task_commands = load_commands()
    with open(json_file, 'r') as fin:
        data = json.load(fin)
        # Make sure we recognize all the tasks
        for task in data['task-set']:
            if task not in task_commands and data['task-set'][task]['perform?']:
                raise NotImplementedError(
                    f'Task {task} is not implemented.')
        # Run tasks in order
        for task in TASKS:
            if task in data['task-set']:
                if data['task-set'][task]['perform?'] is True:
                    run_task(
                        task,
                        data['task-set'][task],
                        task_commands[task],
                        train_data,
                        test_data,
                        results
                    )


def parse_task_args(task_args):
    args_to_join = []

    task_arg_order = OrderedDict()
    task_arg_order['train-on-hidden-data?'] = {
        'defaultvalue': False
    }

    for task_arg_key, task_arg_value in task_arg_order.items():
        if task_arg_key in task_args:
            args_to_join.append(
                str(task_args[task_arg_key])
            )
        else:
            args_to_join.append(
                str(task_arg_value['defaultvalue'])
            )

    return args_to_join


def run_task(task, task_args, task_command, train_data, test_data, results_file):
    logging.info(f'Starting task: {task} with command '
                 f'{str(task_command)} '
                 f'and args {str(task_args)}')
    cmd = task_command + [train_data, test_data, results_file] + parse_task_args(task_args)
    ret = subprocess.call(cmd)
    if ret == 0:
        logging.info("Completed task: " + task)


def parse_setup():
    parser = argparse.ArgumentParser(
        description="BBN BETTER Solution"
    )
    parser.add_argument(
        "--tasks",
        type=str,
        help="JSON task specification file",
        default=TASK_JSON
    )
    parser.add_argument(
        "--train_data",
        type=str,
        help="BP JSON hidden training data file",
        default=TRAIN_DATA
    )
    parser.add_argument(
        "--test_data",
        type=str,
        help="BP JSON test data",
        default=TEST_DATA
    )
    parser.add_argument(
        "--results",
        type=str,
        help="BP JSON results output file",
        default=RESULTS
    )
    parser.add_argument(
        "--logfile",
        type=str,
        help="Logfile",
        default=LOG_FILE
    )
    return parser


def main(args):
    """
    Execute the tasks specified in tasks.json.

    args - contains program specifed command-line arguments (tasks, train_data, test_data, results)
    """
    logging.basicConfig(filename=args.logfile, level=logging.INFO)

    parse_tasks_json(args.tasks, args.train_data, args.test_data, args.results)


if __name__ == "__main__":
    main(parse_setup().parse_args())
