#!/usr/bin/env python3

####################################################################################################
### -----------------------------------------  NOTICE  ----------------------------------------- ###
###                                                                                              ###
### This software (or technical data) was produced for the U.S. Government under contract        ###
### 2015-14120200002-002, and is subject to the Rights in Data-General Clause 52.227-14 , ALT    ###
### IV (MAY 2014) or (DEC 2007).                                                                 ###
###                                                                                              ###
### © 2019 The MITRE Corporation. All Rights Reserved.                                           ###
###                                                                                              ###
### -------------------------------------------------------------------------------------------- ###
####################################################################################################

"""
Converts MAT JSON, BP JSON, and other document formats between each other.

---

# Head-Finding Information

This applies to Basic, Granular, and Abstract.

We're looking for (or assuming) Head or No-Head annotations within the bounds of a
_span_, not to be confused with the ``bounds`` argument passed into this function. Here,
a 'span' refers to the text bounded by an Argument Mention annotation, which can be
looked up via the ``mention_by_annot`` dictionary. The ``bounds`` argument to this
function restricts the search space to only a specific sentence within the text signal
in ``mat_doc``. Every span (regardless of the dialect) needs a Head.

Cases we're looking for:

1. A Head annotation somewhere within the span
2. A No-Head annotation somewhere within the span
3. Neither of these (infer if allowed)

There's an edge condition where a span may overlap (either partially or completely)
another span. In this case, Head (or No-Head) annotations found within two (or more)
spans will be ambiguous. In such cases, we treat the smallest span that a Head or
No-Head is found within as its corresponding span.
"""

__author__  = "Leland E. Vakarian <lsperl@mitre.org>, David S. Day <day@mitre.org>"
__date__    = "05 April, 2022"


import click

from lib import bp
from lib.utils import BP_ANNOTATOR_CLASS_TYPES
from lib.utils import CONTEXT_SETTINGS
from lib.utils import LOGGER
from lib.utils import MAT_DIALECT_CHOICES
from lib.utils import TQDM_PREFIX

import datetime

"""bool: Global flags handled by cli()."""
DEBUG = False
FORCE = False
VERBOSE = False


## Helper Functions


def __spansets_are_mergeable(ss_a, ss_b):
    """Comparator for checking whether a spanset should be merged.

    Conditions for equivalence:

    - if a span _is_ another span
        - note that this isn't the same as "span_a == span_b"
    - if a span's string is the same as another span's string
    - if a span's hstring is the same as another span's hstring

    Conditions for nonequivalence:

    - if the two spansets are the same

    Args:
        ss_a (lib.bp.BPDocumentSpanSet): a ``SpanSet``
        ss_b (lib.bp.BPDocumentSpanSet): a ``SpanSet``

    Returns:
        bool: whether the ``SpanSet``s should be merged
    """
    if ss_a == ss_b:
        return False

    for span_a in ss_a.spans.value:
        for span_b in ss_b.spans.value:
            if span_a is span_b:
                return True

    return False


def __granulate(
    mat_doc,
    evids_by_anchor,
    ssids_by_annot,
    event_set,
    spans_by_annot,
    annots_by_span,
    span_sets,
    ss_counter,
    head_assignments,
    no_validation=False
):
    """Handles Granular annotations not within the Basic dialect.

    This function primarily generates Granular Template objects. Meant to be a supplementary helper
    to :function:`__basic2entry`.

    Args:
        mat_doc (MAT.Document.AnnotatedDoc): the MAT document, in the Basic or Granular dialects, to
            convert into a BP Document entry.
        evids_by_anchor (dict): a dictionary keyed with MAT annotations whose values are BP Basic
            Events created from those annotations.
        ssids_by_annot (dict): a dictionary keyed with MAT annotations whose values are BP Span Sets
            created from those annotations.
        event_set (dict): a dictionary keyed with Event IDs (strings) whose values are the
            corresponding :class:`BPDocumentBasicEvent` objects.
        spans_by_annot (dict): a dictionary keyed with MAT Annotations whose values are the
            corresponding :class:`BPDocumentSpan` objects.
        annots_by_span (dict): a dictionary keyed with :class:`BPDocumentSpan` objects whose values
            are the corresponding MAT Annotation objects.
        span_sets (dict): a dictionary keyed with SSIDs (strings) whose values are the corresponding
            :class:`BPDocumentSpanSet` objects.
        ss_counter (int): the current value of the next available SSID integer.
        head_assignments (dict): keyed with MAT Annotions of type ``Head``, whose values are a
            single :class:`BPDocumentSpan` object.
        no_validation (bool): If True, skip validating constructed BP objects. Defaults to False.

    Returns:
        dict: the granular data members participant to a Basic Event Set.
    """
    from copy import deepcopy

    from MAT.Annotation import AttributeValueSet

    from lib import bp
    from lib.utils import BASIC_MAT_ADDITIONAL_ANCHOR_TYPES
    from lib.utils import BASIC_MAT_ADDITIONAL_MENTION_TYPES
    from lib.utils import BASIC_MAT_KEY_ANCHOR_TYPES
    from lib.utils import BASIC_MAT_KEY_MENTION_TYPES
    from lib.utils import GRANULAR_MAT_ATTACHMENT_TYPES
    from lib.utils import GRANULAR_MAT_COREF_TYPES
    from lib.utils import GRANULAR_MAT_TEMPLATE_TYPES
    from lib.utils import PUNCTUATION
    from lib.utils import pretty_annot

    global DEBUG
    global LOGGER
    global VERBOSE

    corefs = mat_doc.getAnnotations(atypes=GRANULAR_MAT_COREF_TYPES)
    granular_templates = {}

    ## build templates

    # For checking if ``BPDocumentGranularTemplateProperty.ref_id_type`` should populate "ssid" or
    # "event-id".
    event_types = BASIC_MAT_KEY_ANCHOR_TYPES + BASIC_MAT_ADDITIONAL_ANCHOR_TYPES
    entity_types = BASIC_MAT_KEY_MENTION_TYPES + BASIC_MAT_ADDITIONAL_MENTION_TYPES + [
        "Head", "Time-Mention", "Duration-Mention",
    ]

    if DEBUG:
        LOGGER.debug(f"Building templates")

    gtid_counter = 1
    for template_type in GRANULAR_MAT_TEMPLATE_TYPES:
        for annot in mat_doc.getAnnotations(atypes=[template_type,]):

            gtid = f"template-{gtid_counter}"
            gtid_counter += 1

            klass = bp.BPDocumentGranularTemplate.get_child_class(template_type)
            if not klass:
                LOGGER.error(
                    f"Could not find child class for template type '{template_type}'. Is it "
                    "implemented in `bp.BPDocumentGranularTemplate.get_child_class()`?"
                )

            granular_templates[gtid] = klass(
                gtid,
                ssids_by_annot[annot],
                no_validation=no_validation
            )

            if DEBUG:
                LOGGER.debug(
                    f"    Building granular template from {pretty_annot(annot, None, doc=mat_doc)}"
                )
                LOGGER.debug(f"        Creating {granular_templates[gtid]}")

            for annot_prop_name in GRANULAR_MAT_TEMPLATE_TYPES[template_type]:

                prop_annots = annot.get(annot_prop_name)
                if prop_annots:

                    if isinstance(prop_annots, AttributeValueSet):

                        for prop_annot in prop_annots:

                            ## handle pointing directly to Heads

                            if prop_annot.atype.lab == "Head":

                                if prop_annot not in head_assignments:
                                    if DEBUG:
                                        LOGGER.debug(("            "
                                            "Head property isn't found in head_assignments! "
                                            "Skipping!"
                                        ))
                                    continue

                                span = head_assignments[prop_annot]

                                if span not in annots_by_span:
                                    if DEBUG:
                                        LOGGER.debug(("            "
                                            "Can't find MAT annotation Head was assigned to! "
                                            f"Skipping! | {repr(span)}"
                                        ))
                                    continue

                                if annots_by_span[span] not in ssids_by_annot:
                                    if DEBUG:
                                        LOGGER.debug(("            "
                                            "Can't find SSID for MAT annotation Head was assigned "
                                            "to! Skipping!"
                                        ))
                                    continue

                                ssid = ssids_by_annot[annots_by_span[span]]

                                template_prop = bp.BPDocumentGranularTemplateProperty(
                                    ssid=ssid,
                                    no_validation=no_validation
                                )

                                bp_prop = getattr(
                                    granular_templates[gtid],
                                    GRANULAR_MAT_TEMPLATE_TYPES[template_type][annot_prop_name]
                                )

                                if not bp_prop.value:
                                    bp_prop.value = []

                                bp_prop.value.append(template_prop)
                                if DEBUG:
                                    LOGGER.debug(
                                        f"            Added {annot_prop_name} -> {template_prop}."
                                    )

                            ## handle Entity Irrealis assertions

                            elif prop_annot.atype.lab == "Entity-Irrealis":

                                item = prop_annot.get("item") # an annotation
                                if not item:
                                    if DEBUG:
                                        LOGGER.warning(("            "
                                            "Entity-Irrealis property has no item! Skipping!"
                                        ))
                                    continue

                                status = prop_annot.get("status")
                                if not status:
                                    if DEBUG:
                                        LOGGER.warning(("            "
                                            "Entity-Irrealis property has no status! Skipping!"
                                        ))
                                    continue

                                ssid = ssids_by_annot[item]

                                template_prop = bp.BPDocumentGranularTemplateProperty(
                                    ssid=ssid,
                                    irrealis=status,
                                    no_validation=no_validation
                                )

                                bp_prop = getattr(
                                    granular_templates[gtid],
                                    GRANULAR_MAT_TEMPLATE_TYPES[template_type][annot_prop_name]
                                )

                                if not bp_prop.value:
                                    bp_prop.value = []

                                bp_prop.value.append(template_prop)
                                if DEBUG:
                                    LOGGER.debug(
                                        f"            Added {annot_prop_name} -> {template_prop}."
                                    )

                            ## handle Event Irrealis assertions

                            elif prop_annot.atype.lab == "Event-Irrealis":

                                item = prop_annot.get("item") # an annotation
                                if not item:
                                    if DEBUG:
                                        LOGGER.warning(("            "
                                            "Event-Irrealis property has no item! Skipping!"
                                        ))
                                    continue

                                status = prop_annot.get("status")
                                if not status:
                                    if DEBUG:
                                        LOGGER.warning(("            "
                                            "Event-Irrealis property has no status! Skipping!"
                                        ))
                                    continue

                                event = event_set[evids_by_anchor[item]]

                                template_prop = bp.BPDocumentGranularTemplateProperty(
                                    event_id=event.event_id.value,
                                    irrealis=status,
                                    no_validation=no_validation
                                )

                                bp_prop = getattr(
                                    granular_templates[gtid],
                                    GRANULAR_MAT_TEMPLATE_TYPES[template_type][annot_prop_name]
                                )

                                if not bp_prop.value:
                                    bp_prop.value = []

                                bp_prop.value.append(template_prop)
                                if DEBUG:
                                    LOGGER.debug(
                                        f"            Added {annot_prop_name} -> {template_prop}."
                                    )

                            ## build Event-based template properties

                            elif prop_annot.atype.lab in event_types:

                                if prop_annot not in evids_by_anchor:
                                    if DEBUG:
                                        LOGGER.warning(("            "
                                            f"{pretty_annot(prop_annot, None, doc=mat_doc)} not in "
                                            "evids_by_anchor! Skipping!"
                                        ))
                                    continue

                                template_prop = bp.BPDocumentGranularTemplateProperty(
                                    event_id=evids_by_anchor[prop_annot],
                                    no_validation=no_validation
                                )

                                bp_prop = getattr(
                                    granular_templates[gtid],
                                    GRANULAR_MAT_TEMPLATE_TYPES[template_type][annot_prop_name]
                                )

                                if not bp_prop.value:
                                    bp_prop.value = []

                                bp_prop.value.append(template_prop)
                                if DEBUG:
                                    LOGGER.debug(
                                        f"            Added {annot_prop_name} -> {template_prop}."
                                    )

                            ## build Entity-based template properties

                            elif prop_annot.atype.lab in entity_types:

                                if prop_annot not in ssids_by_annot:
                                    if DEBUG:
                                        LOGGER.warning(("            "
                                            f"{pretty_annot(prop_annot, None, doc=mat_doc)} not in "
                                            "ssids_by_annot! Skipping!"
                                        ))
                                    continue

                                template_prop = bp.BPDocumentGranularTemplateProperty(
                                    ssid=ssids_by_annot[prop_annot],
                                    no_validation=no_validation
                                )

                                bp_prop = getattr(
                                    granular_templates[gtid],
                                    GRANULAR_MAT_TEMPLATE_TYPES[template_type][annot_prop_name]
                                )

                                if not bp_prop.value:
                                    bp_prop.value = []

                                bp_prop.value.append(template_prop)
                                if DEBUG:
                                    LOGGER.debug(
                                        f"            Added {annot_prop_name} -> {template_prop}."
                                    )

                            ## catch-all error

                            else:

                                if DEBUG:
                                    LOGGER.warning(("            "
                                        "Found unknown annotation type in property "
                                        f"reference: {prop_annot.atype.lab}. Skipping!"
                                    ))

                    else: # not an iterable, just a string literal or bool (likely)

                        prop_value = prop_annots # readability

                        bp_prop = getattr(
                            granular_templates[gtid],
                            GRANULAR_MAT_TEMPLATE_TYPES[template_type][annot_prop_name]
                        )

                        if bp_prop.type_ == bool:
                            if prop_value == "true":
                                prop_value = True
                            elif prop_value == "false":
                                prop_value = False
                            else:
                                if DEBUG:
                                    LOGGER.warning(("            "
                                        f"{pretty_annot(item, None, mat_doc)}['{annot_prop_name}'] "
                                        f"is not a boolean type: '{prop_value}' | Skipping!"
                                    ))
                                continue

                        if DEBUG:
                            LOGGER.debug(f"            Setting {bp_prop} to {prop_value}.")

                        bp_prop.value = prop_value

    ## build time attachments

    if DEBUG:
        LOGGER.debug(f"Building time attachments")

    for annot in mat_doc.getAnnotations(
        spanlessOnly=True, atypes=["Time-attachment",], ordered=False
    ):

        ## collect key annotation SSID

        key_item = annot.get("time") # an annotation
        if not key_item:
            if DEBUG:
                LOGGER.warning(("    "
                    f"{annot.atype.lab} has no key-item (Time-attachment) value! Skipping!"
                ))
            continue

        if key_item not in ssids_by_annot:
            if DEBUG:
                LOGGER.warning(("    "
                    f"{annot.atype.lab} key-item {pretty_annot(key_item, None, mat_doc)} "
                    "not in ssids_by_annot! Skipping!"
                ))
            continue

        key_ssid = ssids_by_annot[key_item]

        if key_ssid not in span_sets:
            if DEBUG:
                LOGGER.warning(("    "
                    f"{annot.atype.lab} key-item {pretty_annot(key_item, None, mat_doc)} "
                    f"(ssid {key_ssid}) not in span_sets!"
                ))
            continue

        if DEBUG:
            LOGGER.debug(
                f"    Gathering attachments for {annot.atype.lab} key-item "
                f"{pretty_annot(key_item, None, mat_doc)} (ssid {key_ssid})"
            )

        ## collect attachment annotation spans

        items = annot.get("item") # an annotation set
        if not items:
            if DEBUG:
                LOGGER.warning(("        "
                    f"{annot.atype.lab} has no items! Skipping!"
                ))
            continue

        for item in items:

            mention_types = BASIC_MAT_KEY_MENTION_TYPES + BASIC_MAT_ADDITIONAL_MENTION_TYPES + [
                "Time-Mention", "Duration-Mention", "Head",
            ]

            if item.atype.lab in mention_types:

                if item.atype.lab == "Head":

                    if item not in head_assignments:
                        if DEBUG:
                            LOGGER.debug(("            "
                                "Head property isn't found in head_assignments! "
                                "Skipping!"
                            ))
                        continue

                    span = head_assignments[item]

                    if span not in annots_by_span:
                        if DEBUG:
                            LOGGER.debug(("            "
                                "Can't find MAT annotation Head was assigned to! "
                                f"Skipping! | {repr(span)}"
                            ))
                        continue

                    if annots_by_span[span] not in ssids_by_annot:
                        if DEBUG:
                            LOGGER.debug(("            "
                                "Can't find SSID for MAT annotation Head was assigned "
                                "to! Skipping!"
                            ))
                        continue

                    item_ssid = ssids_by_annot[annots_by_span[span]]

                elif item not in ssids_by_annot:
                    if DEBUG:
                        LOGGER.warning(("        "
                            f"{annot.atype.lab} item {pretty_annot(item, None, mat_doc)} "
                            "not in ssids_by_annot! Ignoring!"
                        ))
                    continue

                else:
                    item_ssid = ssids_by_annot[item]

                if item_ssid not in span_sets:
                    if DEBUG:
                        LOGGER.warning(("        "
                            f"{annot.atype.lab} item {pretty_annot(item, None, mat_doc)} "
                            f"(ssid {item_ssid}) not in span_sets! Ignoring!"
                        ))
                    continue

                if item_ssid == key_ssid:

                    if DEBUG:
                        LOGGER.debug(("        "
                            f"{annot.atype.lab} item {pretty_annot(item, None, mat_doc)} "
                            f"(ssid {item_ssid}) is identical to key-item "
                            f"{pretty_annot(item, None, mat_doc)} (ssid {key_ssid}). Skipping "
                            "as expected!"
                        ))
                    continue

                for gtid in granular_templates:
                    template = granular_templates[gtid]
                    for attr in dir(template):
                        attr = getattr(template, attr)
                        if type(attr) == bp.BPProperty:
                            if attr.type_ == list and attr.value:
                                for prop in attr.value:
                                    if prop.ssid.value and prop.ssid.value == item_ssid:

                                        if not prop.time_attachments.value:
                                            prop.time_attachments.value = []
                                        prop.time_attachments.value.append(key_ssid)

                                        if DEBUG:
                                            LOGGER.debug(("        "
                                                f"Adding time attachment to {gtid}'s {attr.name}."
                                            ))

            elif item.atype.lab in BASIC_MAT_KEY_ANCHOR_TYPES:

                if item not in evids_by_anchor:
                    if DEBUG:
                        LOGGER.warning(("        "
                            f"{annot.atype.lab} item {pretty_annot(item, None, mat_doc)} "
                            "not in evids_by_anchor! Ignoring!"
                        ))
                    continue

                item_evid = evids_by_anchor[item]

                for gtid in granular_templates:
                    template = granular_templates[gtid]
                    for attr in dir(template):
                        attr = getattr(template, attr)
                        if type(attr) == bp.BPProperty:
                            if attr.type_ == list and attr.value:
                                for prop in attr.value:
                                    if prop.event_id.value and prop.event_id.value == item_evid:

                                        if not prop.time_attachments.value:
                                            prop.time_attachments.value = []
                                        prop.time_attachments.value.append(key_ssid)

                                        if DEBUG:
                                            LOGGER.debug(("        "
                                                f"Adding time attachment to {gtid}'s {attr.name}."
                                            ))

            else:

                if DEBUG:
                    LOGGER.warning(("        "
                        f"{annot.atype.lab} item {pretty_annot(item, None, mat_doc)} isn't a "
                        "Mention or Event type! Ignoring!"
                    ))
                continue

    ## build corefs

    if DEBUG:
        LOGGER.debug(f"Building corefs")

    mergers = {}
    for coref_type in GRANULAR_MAT_COREF_TYPES:
        for annot in mat_doc.getAnnotations(spanlessOnly=True, atypes=[coref_type,], ordered=False):

            ## collect key annotation SSID

            key_item = annot.get("key-item") # an annotation
            if not key_item:
                if DEBUG:
                    LOGGER.warning(("    "
                        f"{annot.atype.lab} has no key-item value! Skipping!"
                    ))
                continue

            if key_item not in ssids_by_annot:
                if DEBUG:
                    LOGGER.warning(("    "
                        f"{annot.atype.lab} key-item {pretty_annot(key_item, None, mat_doc)} "
                        "not in ssids_by_annot! Skipping!"
                    ))
                continue

            key_ssid = ssids_by_annot[key_item]

            if key_ssid not in span_sets:
                if DEBUG:
                    LOGGER.warning(("    "
                        f"{annot.atype.lab} key-item {pretty_annot(key_item, None, mat_doc)} "
                        f"(ssid {key_ssid}) not in span_sets!"
                    ))
                continue

            if DEBUG:
                LOGGER.debug(
                    f"    Gathering corefs for {annot.atype.lab} key-item "
                    f"{pretty_annot(key_item, None, mat_doc)} (ssid {key_ssid})"
                )

            key_span_set = span_sets[key_ssid]

            ## collect coreferential annotation spans

            items = annot.get("items") # an annotation set
            if not items:
                if DEBUG:
                    LOGGER.warning(("        "
                        f"{annot.atype.lab} has no items! Skipping!"
                    ))
                continue

            spansets_to_merge = []
            for item in items:

                if item not in ssids_by_annot:
                    if DEBUG:
                        LOGGER.debug(("        "
                            f"{annot.atype.lab} item {pretty_annot(item, None, mat_doc)} "
                            "not in ssids_by_annot! Looking for unassigned span!"
                        ))

                    ## check if a span was created for this item, and add it to the key's span set

                    if item not in spans_by_annot:
                        if DEBUG:
                            LOGGER.warning(("            "
                                f"{annot.atype.lab} item {pretty_annot(item, None, mat_doc)} "
                                "not in spans_by_annot! Ignoring!"
                            ))
                        continue
                    elif DEBUG:
                        LOGGER.debug(("            "
                            f"Found span for item, adding to the key item's span set."
                        ))

                    item_span = spans_by_annot[item]
                    key_span_set.spans.value.append(item_span)
                    ssids_by_annot[item] = key_ssid
                    continue

                item_ssid = ssids_by_annot[item]

                if item_ssid != key_ssid:

                    if item_ssid not in span_sets:
                        if DEBUG:
                            LOGGER.warning(("        "
                                f"{annot.atype.lab} item {pretty_annot(item, None, mat_doc)} "
                                f"ssid ({item_ssid}) not in span_sets! Ignoring!"
                            ))
                        continue

                    # span_sets[item_ssid]

                    if item not in spans_by_annot:
                        if DEBUG:
                            LOGGER.warning(("        "
                                f"{annot.atype.lab} item {pretty_annot(item, None, mat_doc)} "
                                f"assigned to span ({item_ssid}) not in spans_by_annot! Ignoring!"
                            ))
                            from pprint import pformat as p
                            LOGGER.critical(("            "
                                f"Current spans in {item_ssid}: {p(span_sets[item_ssid].to_dict())}"
                            ))
                        continue

                    item_span = spans_by_annot[item]
                    key_span_set.spans.value.append(item_span)
                    spansets_to_merge.append(span_sets[item_ssid])
                    ssids_by_annot[item] = key_ssid

                else:

                    if DEBUG:
                        LOGGER.debug(("        "
                            f"{annot.atype.lab} item {pretty_annot(item, None, mat_doc)} "
                            f"ssid ({item_ssid}) is identical to key-item "
                            f"{pretty_annot(item, None, mat_doc)} ssid ({key_ssid}). No SpanSet to "
                            "merge for this item!"
                        ))
                    continue

            ## merge item spansets

            for span_set in spansets_to_merge:

                ## merge the spans

                for span in span_set.spans.value:
                    if span not in key_span_set.spans.value:
                        key_span_set.spans.value.append(span)
                        try:
                            ssids_by_annot[annots_by_span[span]] = key_ssid
                        except KeyError:
                            continue

                # the rest is handled outside of this function, after it's returned in __basic2entry

            mergers[key_ssid] = spansets_to_merge

    ## emit

    return {
        "templates": granular_templates,
        "spansets-to-merge": mergers,
    }


def __basic2entry(
    mat_doc,
    doc_id,
    head_first=False,
    language=None,
    no_infer_heads=False,
    no_annotation_sets=False,
    granular=False,
    no_validation=False,
    annotator_label=None,
    dataset_label=None
):
    """Converts a Basic MAT document into a BP Document entry object, in the Basic dialect.

    The constructed entry represents the whole document, complete with annotations.

    Args:
        mat_doc (MAT.Document.AnnotatedDoc): the MAT document, in the Basic or Granular dialects, to
            convert into a BP Document entry.
        doc_id (str): the ID of the MAT document, used in the BP Document entry's bookkeeping
            information.
        head_first (bool): infer the head as the first token in a span, rather than the last.
        language (str): an ISO 639-3 language code label to apply to each entry in the emitted BP
            document. This code should reflect the language of all of the incoming MAT documents.
        no_infer_heads (bool): if True, don't infer Heads on spans without explicitly-tagged Heads
            (default is to do so).
        no_annotation_sets (bool): if True, don't include annotation-sets.
        granular (bool): if True, process Granular annotations and template information in addition
            to Basic annotation information.
        no_validation (bool): If True, skip validating constructed BP objects. Defaults to False.
        annotator_label (str): if passed, apply this label to all Provenance objects indicating the
            source of the anntotations. Defaults to None.
        dataset_label (str): if passed, apply a this label to all Provenance objects indicating what
            dataset these annoations belong to.

    Returns:
        lib.bp.BPDocumentEntry: an entry for representing the input MAT document.

    ---

    Strategy:

    0. Create ``BPDocumentSegmentSection`` from MAT annotations of structural element types.
    1. Create ``BPDocumentSpan`` objects for every instance of a MAT annotation of a type that
       should be converted into a BP span.
    2. Resolve "Head" information for every ``BPDocumentSpan``.
    3. Resolve coreferential spans, creating ``BPDocumentSpanSet`` objects.
    4. Resolve includes-relations.
    5. Create ``BPDocumentBasicEvent`` objects from anchor spans.
    6. Handle ref_events.
    7. Create and yield a ``BPDocumentEntry`` object to house it all.
    """
    from datetime import date

    from MAT.Annotation import Annotation
    from MAT.Annotation import SpanlessAnnotation

    from lib import bp
    from lib.utils import BASIC_MAT_ADDITIONAL_ANCHOR_TYPES
    from lib.utils import BASIC_MAT_ADDITIONAL_MENTION_TYPES
    from lib.utils import BASIC_MAT_KEY_ANCHOR_TYPES
    from lib.utils import BASIC_MAT_KEY_MENTION_TYPES
    from lib.utils import BASIC_MAT_NAME_MENTION_TYPES
    from lib.utils import BASIC_MAT_NOMINAL_MENTION_TYPES
    from lib.utils import BASIC_MAT_PRONOUN_MENTION_TYPES
    from lib.utils import BASIC_MAT_SENTENCE_TYPES
    from lib.utils import BASIC_MAT_STRUCTURAL_TYPES
    from lib.utils import BP_ANNOTATOR_CLASS_TYPES
    from lib.utils import GRANULAR_MAT_TEMPLATE_TYPES
    from lib.utils import MAT_DIALECT_CHOICES
    from lib.utils import PUNCTUATION
    from lib.utils import annot_in_bounds
    from lib.utils import annot_window
    from lib.utils import detect_overlapping_mat_annots
    from lib.utils import fix_synclass
    from lib.utils import is_overlapped_by
    from lib.utils import pretty_annot
    from lib.utils import signal

    global DEBUG
    global LOGGER
    global VERBOSE

    assert annotator_label in BP_ANNOTATOR_CLASS_TYPES+[None,], \
        f"Bad annotator label: {repr(annotator_label)}."

    def __find_smallest_unassigned_span(annot, spans_by_annot):
        """Finds the smalled unassigned mention span that the given annot is found within.

        Spans with an already-assigned ``hstring`` or ``no-head`` value will be ignored since
        they've already been assigned a Head or No-Head.
        """

        def __len(annot):
            return annot.end - annot.start

        smallest_span = None

        for span_an in spans_by_annot:

            # check if span is already assigned a Head or No-Head
            if spans_by_annot[span_an].hstring.value or spans_by_annot[span_an].no_head.value:
                continue

            # make sure annot is within bounds of span
            if annot.start >= span_an.start and annot.end <= span_an.end:

                # check if smallest span so far
                if not smallest_span:
                    smallest_span = span_an
                elif __len(span_an) <= __len(smallest_span):
                    smallest_span = span_an

        return smallest_span

    def __event_args(key_annot, type_, ss_counter, span_sets):
        """Builds a list of agent, patient, or money SSIDs."""
        args = []
        annots = key_annot.get(type_)
        if annots:
            for annot in annots:

                if type(annot) == SpanlessAnnotation:

                    if annot.atype.lab == "Entity-Irrealis":

                        if not granular:
                            if DEBUG:
                                LOGGER.error((
                                    "Encountered granular Irrealis Annotation "
                                    f"{pretty_annot(annot)} in the {type_} list of "
                                    f"{pretty_annot(key_annot, None, mat_doc)} the task was set to "
                                    "Basic! Processing it anyway."
                                ))

                        if DEBUG:
                            LOGGER.debug(("    "
                                "Processing Entity-Irrealis arg in "
                                f"{pretty_annot(annot, None, mat_doc)} {type_} list"
                            ))

                        item = annot.get("item") # an annotation
                        if not item:
                            if DEBUG:
                                LOGGER.warning(("        "
                                    "Entity-Irrealis property has no item! Skipping!"
                                ))
                            continue

                        ssid = f"ss-{ss_counter}"
                        ss_counter += 1

                        span_sets[ssid] = bp.BPDocumentSpanSet(
                            ssid=ssid,
                            spans=[
                                bp.BPDocumentSpan(
                                    string_=signal(mat_doc, item),
                                    start=item.start,
                                    end=item.end,
                                    hstring=signal(mat_doc, item),
                                    hstart=item.start,
                                    hend=item.end,
                                    no_validation=no_validation
                                )
                            ],
                            no_validation=no_validation
                        )

                        if DEBUG:
                            LOGGER.debug(("        "
                                f"Created span {span_sets[ssid]} from Entity-Irrealis item"
                                f" {pretty_annot(item, None, mat_doc)}"
                            ))

                        args.append(ssid)

                else:

                    if annot in ssids_by_annot:

                        ssid = ssids_by_annot[annot]
                        args.append(ssid)

                    elif DEBUG:

                        LOGGER.warning((
                            "Event argument not found in ssids_by_annot dict!\n"
                            f"    Annot ({annot.atype.lab}): {signal(mat_doc, annot)}\n"
                            f"    Location: {annot.start}: {annot.end}"
                        ))
        return args

    def __merge_span_sets(ss_a, ss_b):
        """Merges ss_b into ss_a."""
        for span in ss_b.spans.value:
            if span not in ss_a.spans.value:
                ss_a.spans.value.append(span)
        for annot in ssids_by_annot:
            if ssids_by_annot[annot] == ss_b.ssid.value:
                ssids_by_annot[annot] = ss_a.ssid.value

    def __sanitize_segsec_type(label):
        if label == "S":
            return "Sentence"
        elif label == "Delete-This-Text":
            return "Ignore"
        else:
            return label

    ## (0) get structural elements

    if VERBOSE:
        sentence_overlaps = detect_overlapping_mat_annots(mat_doc, "S")

        if sentence_overlaps["overlaps"] or sentence_overlaps["duplicates"]:
            LOGGER.warning(f"Sentence tag overlaps detected in {doc_id}!")

        for annot1, annot2, message in sentence_overlaps['overlaps']:
            LOGGER.warning(("    "
                f"Partial overlap: ({annot1.atype.lab})[{annot1.start}, {annot1.end}] with "
                f"({annot2.atype.lab})[{annot2.start}, {annot2.end}], {message}"
            ))

        for dupe in sentence_overlaps['duplicates']:
            LOGGER.warning(("    "
                f"Exact duplicate: [{dupe[0]}, {dupe[1]}] repeated "
                f"{len(sentence_overlaps['duplicates'][dupe])} times."
            ))

    structural_elements = []
    for annot in mat_doc.getAnnotations(atypes=BASIC_MAT_STRUCTURAL_TYPES):

        segsec = bp.BPDocumentSegmentSection(
            start=annot.start,
            end=annot.end,
            structural_element=__sanitize_segsec_type(annot.atype.lab),
            no_validation=no_validation
        )
        if not no_validation:
            assert segsec.is_valid()
        structural_elements.append(segsec)

        # create accompanying sentences for all Headlines
        if annot.atype.lab == "Headline" and not is_overlapped_by(annot, "S", mat_doc):
            segsec = bp.BPDocumentSegmentSection(
                start=annot.start,
                end=annot.end,
                structural_element="Sentence",
                no_validation=no_validation
            )
            if not no_validation:
                assert segsec.is_valid()
            structural_elements.append(segsec)

    if not no_annotation_sets:

        ## (1) build spans from spanned annotations (not spanless)

        types = BASIC_MAT_ADDITIONAL_MENTION_TYPES + \
            BASIC_MAT_KEY_MENTION_TYPES

        if granular:
            types += ["Time-Mention", "Duration-Mention", "Head",]

        spans_by_annot = {}
        annots_by_span = {}
        for annot in mat_doc.getAnnotations(atypes=types):
            spans_by_annot[annot] = bp.BPDocumentSpan(
                string_=signal(mat_doc, annot),
                start=annot.start,
                end=annot.end,
                synclass=fix_synclass("Basic", annot.get("Mention_Type", "")),
                no_validation=no_validation
            )

            if annot.atype.lab in BASIC_MAT_NOMINAL_MENTION_TYPES:
                spans_by_annot[annot].synclass.value = "nominal"
            if annot.atype.lab in BASIC_MAT_NAME_MENTION_TYPES:
                spans_by_annot[annot].synclass.value = "name"
            if annot.atype.lab in BASIC_MAT_PRONOUN_MENTION_TYPES:
                spans_by_annot[annot].synclass.value = "pronoun"

            annots_by_span[spans_by_annot[annot]] = annot

        anchor_types = BASIC_MAT_KEY_ANCHOR_TYPES + BASIC_MAT_ADDITIONAL_ANCHOR_TYPES

        for annot in mat_doc.getAnnotations(atypes=anchor_types):
            spans_by_annot[annot] = bp.BPDocumentSpan(
                string_=signal(mat_doc, annot),
                start=annot.start,
                end=annot.end,
                synclass="event-anchor",
                hstring=signal(mat_doc, annot), # Anchors are, by definition always their own head.
                hstart=annot.start,
                hend=annot.end,
                is_anchor=True,
                no_validation=no_validation
            )

            annots_by_span[spans_by_annot[annot]] = annot

            if annot.atype.lab in BASIC_MAT_KEY_ANCHOR_TYPES:
                # Only actual anchors should be turned into event anchor spans
                spans_by_annot[annot].synclass.value = "event-anchor"
                spans_by_annot[annot].is_anchor.value = True

        if granular:
            for annot in mat_doc.getAnnotations(atypes=GRANULAR_MAT_TEMPLATE_TYPES):

                spans_by_annot[annot] = bp.BPDocumentSpan(
                    string_=signal(mat_doc, annot),
                    start=annot.start,
                    end=annot.end,
                    synclass="template-anchor",
                    no_validation=no_validation
                )
                annots_by_span[spans_by_annot[annot]] = annot

        ## (2) Add head information, if it exists

        heads = mat_doc.getAnnotations(atypes=['Head',], ordered=True)
        no_heads = mat_doc.getAnnotations(atypes=['No-Head',], ordered=True)

        assigned_heads = []
        assigned_no_heads = []
        head_assignments = {}
        for annot in sorted(spans_by_annot.keys(), key=lambda a: a.end-a.start):

            if annot.atype.lab not in anchor_types and type(annot) == Annotation and \
                annot.atype.lab not in ["Head", "No-Head",]:

                if DEBUG:
                    LOGGER.debug(
                        f"Resolving Head value for {pretty_annot(annot, indent=None, doc=mat_doc)}"
                    )

                # instance of BPDocumentSpan
                span = spans_by_annot[annot]

                # find all No-Head annotations within the bounds of annot
                candidates = [
                    a for a in annot_window(annot.start, annot.end, no_heads) \
                        if a not in assigned_no_heads
                ]

                # look for `Head` annots since no `No-Head`s were found
                if len(candidates) == 0:

                    # find all Head annotations within the bounds of annot
                    candidates = [
                        a for a in annot_window(annot.start, annot.end, heads) \
                            if a not in assigned_heads
                    ]

                    # case 3
                    if len(candidates) == 0 and not no_infer_heads:

                        if annot.atype.lab in ["Time-Mention", "Duration-Mention",]:

                            head = mat_doc.createAnnotation(
                                annot.start,
                                annot.end,
                                "Head"
                            )
                            if annot.atype.lab == "Time-Mention":
                                span.synclass.value = "time-mention"
                            else:
                                span.synclass.value = "duration-mention"
                        else:

                            signal_ = signal(mat_doc, annot)
                            token = signal_.split()[0 if head_first else -1]
                            head = mat_doc.createAnnotation(
                                annot.start + signal_.rfind(token),
                                annot.start + signal_.rfind(token) + len(token),
                                "Head"
                            )

                        span.hstring.value = signal(mat_doc, head)
                        span.hstart.value = head.start
                        span.hend.value = head.end
                        span.hinferred.value = True
                        head_assignments[head] = span
                        if DEBUG:
                            LOGGER.debug(
                                "    Creating Head since no Head or No-Head annotations found: "
                                f"{pretty_annot(head, None, mat_doc)}"
                            )

                    # case 1
                    elif len(candidates) == 1:

                        head = candidates[0]
                        assigned_heads.append(head)
                        span.hstring.value = signal(mat_doc, head)
                        span.hstart.value = head.start
                        span.hend.value = head.end
                        span.hinferred.value = False
                        head_assignments[head] = span

                        if DEBUG:
                            LOGGER.debug(
                                f"    Found Head annotation: {pretty_annot(head, None, mat_doc)}"
                            )

                    # edge condition for Heads
                    elif len(candidates) > 1:

                        if DEBUG:
                            LOGGER.debug("    Found multiple Head annotations!")

                        for head in candidates:

                            smallest_annot = __find_smallest_unassigned_span(head, spans_by_annot)

                            if not smallest_annot:
                                if VERBOSE:
                                    LOGGER.warning((
                                        "Couldn't find an unassigned span for: "
                                        f"{pretty_annot(head, indent=1, doc=mat_doc)}\n"
                                        f">>> {doc_id}"
                                    ))
                                continue

                            if DEBUG:
                                LOGGER.debug(
                                    f"        Resolving Head {pretty_annot(head, None, mat_doc)} "
                                    f"and {pretty_annot(smallest_annot, None, mat_doc)}"
                                )

                            span = spans_by_annot[smallest_annot]
                            span.hstring.value = signal(mat_doc, head)
                            span.hstart.value = head.start
                            span.hend.value = head.end
                            span.hinferred.value = False
                            assigned_heads.append(head)
                            head_assignments[head] = span

                # case 2
                elif len(candidates) == 1:

                    if DEBUG:
                        LOGGER.debug("    Found No-Head annotation.")

                    span.no_head.value = True
                    span.hinferred.value = False
                    assigned_no_heads.append(candidates[0])

                # edge condition for No-Heads
                elif len(candidates) > 1:

                    if DEBUG:
                        LOGGER.debug("    Found multiple No-Head annotations!")

                    for no_head in candidates:

                        smallest_annot = __find_smallest_unassigned_span(no_head, spans_by_annot)

                        if not smallest_annot:
                            if VERBOSE:
                                LOGGER.warning((
                                    "Couldn't find an unassigned span for: "
                                    f"{pretty_annot(no_head, indent=1)}\n"
                                    f">>> {doc_id}"
                                ))
                            continue

                        if DEBUG:
                            LOGGER.debug(
                                f"        Resolving No-Head {pretty_annot(no_head, None, mat_doc)} "
                                f"and {pretty_annot(smallest_annot, None, mat_doc)}"
                            )
                        assigned_no_heads.append(no_head)
                        span = spans_by_annot[smallest_annot]
                        span.no_head.value = True
                        span.hinferred.value = False

        ## (3) build spansets

        ssids_by_annot = {}
        span_sets = {}
        ss_counter = 1
        memo = []
        for annot in spans_by_annot:

            if annot not in memo and annot.atype.lab not in BASIC_MAT_ADDITIONAL_MENTION_TYPES + \
                BASIC_MAT_ADDITIONAL_ANCHOR_TYPES + ["Head", "No-Head",]:

                ssid = f"ss-{ss_counter}"
                ss_counter += 1

                span_sets[ssid] = bp.BPDocumentSpanSet(
                    ssid=ssid,
                    spans=[spans_by_annot[annot],],
                    no_validation=no_validation
                )
                memo.append(annot)
                ssids_by_annot[annot] = ssid
                if DEBUG:
                    LOGGER.debug((
                        f"Creating span set {ssid} with annotation: "
                        f"{pretty_annot(annot, indent=None, doc=mat_doc)}"
                    ))

                if annot.get("addtl_anchors") is not None:

                    for coref_annot in annot.get("addtl_anchors"):
                        span_sets[ssid].spans.value.append(spans_by_annot[coref_annot])
                        if DEBUG:
                            LOGGER.debug((
                                "    Adding Additional Anchor annotation to span set: "
                                f"{pretty_annot(coref_annot, indent=None, doc=mat_doc)}"
                            ))

                        memo.append(coref_annot)
                        ssids_by_annot[coref_annot] = ssid

                elif annot.get("additional-mentions") is not None:

                    for coref_annot in annot.get("additional-mentions"):
                        span_sets[ssid].spans.value.append(spans_by_annot[coref_annot])
                        if DEBUG:
                            LOGGER.debug((
                                "    Adding Additional Mention annotation to span set: "
                                f"{pretty_annot(coref_annot, indent=None, doc=mat_doc)}"
                            ))
                        memo.append(coref_annot)
                        ssids_by_annot[coref_annot] = ssid

        for annot in spans_by_annot:
            if annot not in memo and annot.atype.lab in BASIC_MAT_ADDITIONAL_MENTION_TYPES + \
               BASIC_MAT_ADDITIONAL_ANCHOR_TYPES:

                ssid = f"ss-{ss_counter}"
                ss_counter += 1

                span_sets[ssid] = bp.BPDocumentSpanSet(
                    ssid=ssid,
                    spans=[spans_by_annot[annot],],
                    no_validation=no_validation
                )
                memo.append(annot)
                ssids_by_annot[annot] = ssid
                if DEBUG:
                    LOGGER.debug((
                        f"Creating span set {ssid} with annotation: "
                        f"{pretty_annot(annot, indent=None, doc=mat_doc)} since it has no "
                        "non-additional annotation type coreferences."
                    ))

        ## (3.5) merge spansets (transitive closure problem)

        if DEBUG:
            LOGGER.debug("Merging span-sets...")
        found_merger = True # lets us enter the loop
        while found_merger:
            found_merger = False
            for ssid_a in span_sets:
                ss_a = span_sets[ssid_a]
                for ssid_b in span_sets:
                    ss_b = span_sets[ssid_b]
                    if __spansets_are_mergeable(ss_a, ss_b):
                        if DEBUG:
                            LOGGER.debug(f"    Merging {ss_b} into {ss_a}")
                        __merge_span_sets(ss_a, ss_b)
                        span_sets.pop(ssid_b)
                        found_merger = True
                        break
                if found_merger:
                    break

        ## (4) build includes-relations

        includes_relations = {}
        for annot in ssids_by_annot:

            includes = annot.get('includes')
            if includes:

                if DEBUG:
                    LOGGER.debug(
                        f"Annot {pretty_annot(annot, None, mat_doc)} has includes relations!"
                    )
                    LOGGER.debug(
                        f"    Includes: {[pretty_annot(inc, None, mat_doc) for inc in includes]}"
                    )

                try:
                    ssid = ssids_by_annot[annot]
                except KeyError:
                    if VERBOSE:
                        LOGGER.warning(
                            "Annotation never parsed into ssids_by_annot, skipping: "
                            f"{pretty_annot(annot, indent=1, doc=mat_doc)}\n"
                            f">>> {doc_id}"
                        )
                    continue

                if ssid in includes_relations:
                    if DEBUG:
                        LOGGER.debug((
                            f"SSID ({ssid}) for Annot already used in "
                            f"includes_relations dict, skipping: {span_sets[ssid]}"
                        ))
                    continue

                include_ssids = []
                for include in includes:
                    try:
                        include_ssid = ssids_by_annot[include]
                    except KeyError:
                        if VERBOSE:
                            LOGGER.warning(
                                "Include Annot not found in ssids_by_annot dict! Skipping: "
                                f"{pretty_annot(include, indent=1, doc=mat_doc)}\n"
                                f">>> {doc_id}"
                            )
                        continue
                    include_ssids.append(include_ssid)

                includes_relations[ssid] = include_ssids

        ## (5) build events

        event_set = {}
        anchors_by_evid = {}
        evids_by_anchor = {}
        event_counter = 1
        for annot in ssids_by_annot:

            if spans_by_annot[annot].is_anchor.value == True and \
               annot.atype.lab != "Additional-Event-Anchor":

                try:
                    assert annot in ssids_by_annot
                except:
                    if DEBUG:
                        LOGGER.warning(
                            "Annotation never parsed into ssids_by_annot: "
                            f"{pretty_annot(annot, indent=1, doc=mat_doc)}"
                            f">>> {doc_id}"
                        )
                    continue

                event_id = "event-" + str(event_counter)
                event_counter += 1

                if DEBUG:
                    LOGGER.debug(
                        f"Building Event {event_id} from {pretty_annot(annot, None, mat_doc)}"
                    )

                anchors_by_evid[event_id] = annot
                evids_by_anchor[annot] = event_id

                soa = annot.get('state-of-affairs')

                # break some events into their sub-event labels
                if annot.atype.lab == "Violence":
                    event_type = annot.get("replace-with", annot.atype.lab)
                    if event_type == "NoChange":
                        event_type = annot.atype.lab
                elif annot.atype.lab == "Cybersecurity-Response":
                    event_type = "Cybersecurity-Measure"
                else:
                    event_type = annot.atype.lab

                event = bp.BPDocumentBasicEvent(
                    event_id,
                    ssids_by_annot[annot], # `anchors`
                    __event_args(annot, "agent", ss_counter, span_sets),
                    __event_args(annot, "patient", ss_counter, span_sets),
                    [], # `ref_events`
                    event_type,
                    money=__event_args(annot, "money", ss_counter, span_sets),
                    state_of_affairs=soa == "true",
                    event_template_filler_coref_events=[],
                    no_validation=no_validation
                )
                if not no_validation:
                    assert event.is_valid(), \
                        f"Event {event_id} ({ssids_by_annot[annot]}) in {doc_id} is invalid!"

                if event.has_duplicate_args(debug=DEBUG):
                    if DEBUG:
                        LOGGER.debug(">>> Discovered duplicate elements will be squashed!")
                        LOGGER.debug(f">>> {doc_id}, Event {event_id}")

                event_set[event_id] = event

        # (6) build ref_events and template_filler_coref_events lists

        for event_id, event in event_set.items():

            key_annot = anchors_by_evid[event_id]

            ref_annots = key_annot.get("ref_event")
            if ref_annots:
                for annot in ref_annots:

                    if annot in evids_by_anchor:
                        event.ref_events.value.append(evids_by_anchor[annot])
                    elif DEBUG:
                        LOGGER.warning((
                            "Found ref_event annot that doesn't exist as a key in "
                            f"evids_by_anchor!\n    Annot ({annot.atype.lab}): "
                            f"{signal(mat_doc, annot)}\n    Location: {annot.start}:{annot.end} "
                            f"{doc_id}"
                        ))

            etf_annots = key_annot.get("equivalent_template_fillers")
            if etf_annots:
                if DEBUG:
                    LOGGER.debug(f"Found equivalent template fillers on {event_id}.")
                for etf in etf_annots:
                    event.event_template_filler_coref_events.append(evids_by_anchor[etf])
                    if DEBUG:
                        LOGGER.debug(("    "
                            f"Adding {evids_by_anchor[etf]} "
                            f"({pretty_annot(etf, indent=None, doc=mat_doc)})."
                        ))

        # (7) Construct a list of events that are co-referential across the document to support
        # Granular scoring.  This table is built from the transitive closure of events that are
        # linked through the 'equivalent_template_fillers' slot of MAT events.  The resulting
        # table of equivalence classes is captured in the 'event-sets' field in BP JSON.
        # Example:  "events-sets": [["e-1", "e-2"], ["e-42", "e-59", "e-103"]]
        template_filler_coref_events = __construct_template_filler_transitive_closure(
            event_set, anchors_by_evid, evids_by_anchor
        )

    # (8) build and yield entry

    entry = bp.BPDocumentEntry(
        entry_id=doc_id,
        doc_id=doc_id,
        lang=language,
        segment_type="document",
        segment_text=mat_doc.signal,
        provenance=bp.BPDocumentProvenance(
            annotation_procedure="mat",
            corpus_creation_date=str(date.today()),
            source=doc_id,
            no_validation=no_validation,
            annotator_class=annotator_label
        ),
        no_validation=no_validation
    )

    if dataset_label:
        entry.provenance.value.add_misc_key("dataset-label", str(dataset_label))

    entry.segment_sections.value = structural_elements

    if not no_annotation_sets:

        if granular:

            gran_data = __granulate(
                mat_doc,
                evids_by_anchor,
                ssids_by_annot,
                event_set,
                spans_by_annot,
                annots_by_span,
                span_sets,
                ss_counter,
                head_assignments,
                no_validation=no_validation
            )

            ## replace references to item's SpanSet's SSID with key-item's SpanSet's SSID

            for key_ssid in gran_data['spansets-to-merge']:

                spansets_to_merge = gran_data['spansets-to-merge'][key_ssid]

                if DEBUG:
                    LOGGER.debug(f"Handling {key_ssid}'s replacements")

                for span_set in spansets_to_merge:

                    if DEBUG:
                        LOGGER.debug(f"    Merging {span_set}")

                    remove_ssid = span_set.ssid.value

                    ## perform replacements in includes-relations

                    clone = {}
                    for ssid in includes_relations:

                        for candidate_ssid in includes_relations[ssid]:
                            if candidate_ssid == remove_ssid:
                                includes_relations[ssid].remove(candidate_ssid)
                                includes_relations[ssid].append(key_ssid)
                                if DEBUG:
                                    LOGGER.debug(("        "
                                        f"Replacing ssid {candidate_ssid} in {ssid}'s includes-"
                                        "relations."
                                    ))

                        if ssid == remove_ssid:
                            clone[key_ssid] = includes_relations[ssid]
                            if DEBUG:
                                LOGGER.debug(("        "
                                    f"Replacing ssid {ssid} in includes-relations keys."
                                ))
                        else:
                            clone[ssid] = includes_relations[ssid]

                    includes_relations = clone

                    ## perform replacements in Events

                    for evid in event_set:

                        event = event_set[evid]

                        if event.anchors.value == remove_ssid:
                            event.anchors.value = key_ssid
                            if DEBUG:
                                LOGGER.debug(("        "
                                    f"Replacing {evid}'s anchors value."
                                ))

                        for ssid in event.agents.value:
                            if ssid == remove_ssid:
                                event.agents.value.remove(ssid)
                                event.agents.value.append(key_ssid)
                                if DEBUG:
                                    LOGGER.debug(("        "
                                        f"Replacing {ssid} in {evid}'s agents list."
                                    ))

                        for ssid in event.patients.value:
                            if ssid == remove_ssid:
                                event.patients.value.remove(ssid)
                                event.patients.value.append(key_ssid)
                                if DEBUG:
                                    LOGGER.debug(("        "
                                        f"Replacing {ssid} in {evid}'s patients list."
                                    ))

                        if event.money.value:
                            for ssid in event.money.value:
                                if ssid == remove_ssid:
                                    event.patients.value.remove(ssid)
                                    event.patients.value.append(key_ssid)
                                    if DEBUG:
                                        LOGGER.debug(("        "
                                            f"Replacing {ssid} in {evid}'s money list."
                                        ))

                    ## perform replacements in Granular-Templates

                    for gtid in gran_data["templates"]:

                        template = gran_data["templates"][gtid]

                        if template.anchor.value == remove_ssid:
                            template.anchor.value = key_ssid

                        for attr in dir(template):
                            attr = getattr(template, attr)
                            if type(attr) == bp.BPProperty:
                                if attr.type_ == list and attr.value:
                                    for prop in attr.value:
                                        if prop.ssid.value == remove_ssid:
                                            prop.ssid.value = key_ssid
                                            if DEBUG:
                                                LOGGER.debug(("        "
                                                    f"Replacing an ssid for {gtid}'s {attr.name}."
                                                ))

                    ## remove merged-in SpanSet

                    span_sets.pop(remove_ssid, None)

            ## make entry

            entry.annotation_sets.value = bp.BPDocumentAnnotationSet(
                basic_events=bp.BPDocumentBasicEventSet(
                    events=event_set,
                    span_sets=span_sets,
                    includes_relations=includes_relations,
                    granular_templates=gran_data["templates"],
                    template_filler_coref_events=template_filler_coref_events,
                    no_validation=no_validation
                ),
                no_validation=no_validation
            )

        else:

            entry.annotation_sets.value = bp.BPDocumentAnnotationSet(
                basic_events=bp.BPDocumentBasicEventSet(
                    events=event_set,
                    span_sets=span_sets,
                    includes_relations=includes_relations,
                    no_validation=no_validation
                ),
                no_validation=no_validation
            )

    if not no_validation:
        assert entry.is_valid(), f"BP Doc created from {doc_id} is invalid!"

    return entry


def __construct_template_filler_transitive_closure(event_set, anchors_by_evid, evids_by_anchor):
    # (7) Construct a list of events that are co-referential across the document to support
    # Granular scoring.  This table is built from the transitive closure of events that are
    # linked through the 'equivalent_template_fillers' slot of MAT events.  The resulting
    # table of equivalence classes is captured in the 'event-sets' field in BP JSON.
    # Example:  "events-sets": [["e-1", "e-2"], ["e-42", "e-59", "e-103"]]

    global DEBUG
    global LOGGER

    if DEBUG:
        LOGGER.debug("Building initial ETF table.")

    coref_events = {}
    # pre-populate this table of equivalence sets
    for event_id, event in event_set.items():

        coref_events[event_id] = set([event_id])
        for coref_event_id in event.event_template_filler_coref_events:

            if event_id not in coref_events:
                coref_events[event_id].add(event_id)

            coref_events[event_id].add(coref_event_id)


    if DEBUG:
        LOGGER.debug(f"    initial co-referring events table: {coref_events}")

    # Pairwise relations: E1 --> E2, E3 --> E2, E4 --> E5, E4 --> E6
    # Defines two equivalence classes: [E1, E2, E3] and [E4, E5, E6]

    updated_coref = True  # lets us enter the loop
    while updated_coref:
        updated_coref = False
        changes = []
        for evt_id1, coref_set1 in coref_events.items():
            for coref_id1 in coref_set1:
                for coref_id2 in coref_events[coref_id1]:
                    if evt_id1 not in coref_events[coref_id2]:
                        updated_coref = True
                        changes.append([coref_id2, evt_id1])
                        # coref_events[coref_id2].add(evt_id1)
                        if DEBUG:
                            LOGGER.debug("    Updated template filler event equivalence: " + \
                                         f"{coref_events[coref_id2]}")
                    if coref_id2 not in coref_events[evt_id1]:
                        updated_coref = True
                        # coref_events[evt_id1].add(coref_id2)
                        changes.append([evt_id1, coref_id2])
                        if DEBUG:
                            LOGGER.debug("    Updated template filler event equivalence: " + \
                                         f"{coref_events[evt_id1]}")
                    if updated_coref:
                        break
        for change in changes:
            if DEBUG:
                LOGGER.debug(f"        Executing the postponed changes: {change}")
            coref_events[change[0]].add(change[1])

    reduced_coref_events = []
    for evt_id, event_coref_set in coref_events.items():
        # for every event_coref_set, some of which could be identical in content, see if it has
        # been entered into the reduced list; if not add it, providing a single copy of each unique
        # list
        reduced = False
        for reduced_event_set in reduced_coref_events:
            if event_coref_set == reduced_event_set:
                reduced = True
        if not reduced:
            event_coref_set = list(event_coref_set)
            event_coref_set.sort()
            if len(event_coref_set) > 1 and event_coref_set not in reduced_coref_events:
                reduced_coref_events.append(event_coref_set)
    if DEBUG:
        LOGGER.debug(f"Final ETF table: {reduced_coref_events}")
    return reduced_coref_events


def __abstract2entries(
    mat_doc,
    doc_id,
    no_documents=False,
    noquad=False,
    head_first=False,
    language=None,
    no_infer_heads=False,
    no_annotation_sets=False,
    no_validation=False,
    annotator_label=None,
    dataset_label=None
):
    """Converts an Abstract MAT document into a series of entries.

    Entries constructed include:
    - One entry representing the whole document, with no annotations.
    - One entry for each sentence in the document, with annotations.

    Args:
        mat_doc (MAT.Document.AnnotatedDoc): the MAT document, in the Abstract dialect, to convert
            into a BP Document entry.
        doc_id (str): the ID of the MAT document, used in each BP Document entry's bookkeeping
            information.
        no_documents (bool): if True, don't include document-type entries.
        noquad (bool): if True, don't build QuadClass information.
        head_first (bool): infer the head as the first token in a span, rather than the last.
        language (str): an ISO 639-3 language code label to apply to each entry in the emitted BP
            document. This code should reflect the language of all of the incoming MAT documents.
        no_infer_heads (bool): if True, don't infer Heads on spans without explicitly-tagged Heads
            (default is to do so).
        no_annotation_sets (bool): if True, don't include annotation-sets.
        no_validation (bool): If True, skip validating constructed BP objects. Defaults to False.
        annotator_label (str): if passed, apply this label to all Provenance objects indicating the
            source of the anntotations. Defaults to None.
        dataset_label (str): if passed, apply a this label to all Provenance objects indicating what
            dataset these annoations belong to.

    Returns:
        dict: keyed with Entry ID strings pointing to lib.bp.BPDocumentEntry instances, each
        representing portions of the input MAT document.
    """
    from datetime import date

    from lib import bp
    from lib.utils import ABSTRACT_MAT_ANCHOR_TYPES
    from lib.utils import ABSTRACT_MAT_KEY_MENTION_TYPES
    from lib.utils import ABSTRACT_MAT_ADDITIONAL_MENTION_TYPES
    from lib.utils import BP_ANNOTATOR_CLASS_TYPES
    from lib.utils import PUNCTUATION
    from lib.utils import annot_in_bounds
    from lib.utils import annot_window
    from lib.utils import find_sentence_regions
    from lib.utils import fix_synclass
    from lib.utils import pretty_annot
    from lib.utils import signal

    global DEBUG
    global LOGGER
    global VERBOSE

    assert annotator_label in BP_ANNOTATOR_CLASS_TYPES+[None,], \
        f"Bad annotator label: {repr(annotator_label)}."

    def __find_smallest_unassigned_span(annot, spans_by_annot):
        """Finds the smalled unassigned mention span that the given annot is found within.

        Spans with an already-assigned ``hstring`` or ``no-head`` value will be ignored since
        they've already been assigned a Head or No-Head.
        """

        def __len(annot):
            return annot.end - annot.start

        smallest_span = None

        for span_an in spans_by_annot:

            # check if span is already assigned a Head or No-Head
            if spans_by_annot[span_an].hstring.value or spans_by_annot[span_an].no_head.value:
                continue

            # make sure annot is within bounds of span
            if annot.start >= span_an.start and annot.end <= span_an.end:

                # check if smallest span so far
                if not smallest_span:
                    smallest_span = span_an
                elif __len(span_an) <= __len(smallest_span):
                    smallest_span = span_an

        return smallest_span

    def __merge_span_sets(ss_a, ss_b, ssids_by_annot):
        """Merges ss_b into ss_a."""
        for span in ss_b.spans.value:
            if span not in ss_a.spans.value:
                ss_a.spans.value.append(span)
        for annot in ssids_by_annot:
            if ssids_by_annot[annot] == ss_b.ssid.value:
                ssids_by_annot[annot] = ss_a.ssid.value

    entries = {}

    ## construct whole-document entry

    if not no_documents:
        entry = bp.BPDocumentEntry(
            entry_id=doc_id,
            doc_id=doc_id,
            lang=language,
            segment_type="document",
            segment_text=mat_doc.signal,
            segment_start=0,
            segment_end=len(mat_doc.signal),
            start_in_source=0,
            end_in_source=len(mat_doc.signal),
            provenance=bp.BPDocumentProvenance(
                annotation_procedure="mat",
                corpus_creation_date=str(date.today()),
                source=doc_id,
                no_validation=no_validation,
                annotator_class=annotator_label
            ),
            no_validation=no_validation
        )

        if dataset_label:
            entry.provenance.value.add_misc_key("dataset-label", str(dataset_label))

        if not no_validation:
            assert entry.is_valid()

        entries[doc_id] = entry

    ## construct sentence-level entries

    ignore_bounds = mat_doc.getAnnotations(atypes="Skip_This_Sentence")

    def __check_skip_sentence(ignore_bounds, sent_bounds):
        for annot in ignore_bounds:
            if sent_bounds[0] >= annot.start and sent_bounds[1] <= annot.end:
                return True
        return False

    num_events = 0
    sent_counter = 0
    for bounds in find_sentence_regions(mat_doc):

        if __check_skip_sentence(ignore_bounds, bounds):
            continue

        if DEBUG:
            LOGGER.debug(f"Sentence bounds: {bounds}")

        if not no_annotation_sets:

            ## build spans

            spans_by_annot = {}
            annots_by_span = {}
            for annot in annot_window(bounds[0], bounds[1], mat_doc.getAnnotations(
                atypes=ABSTRACT_MAT_KEY_MENTION_TYPES+ABSTRACT_MAT_ADDITIONAL_MENTION_TYPES
            )):
                spans_by_annot[annot] = bp.BPDocumentSpan(
                    string_=signal(mat_doc, annot),
                    start=annot.start-bounds[0],
                    end=annot.end-bounds[0],
                    synclass=fix_synclass("Abstract", annot.get("Mention_Type")),
                    no_validation=no_validation
                )
                annots_by_span[spans_by_annot[annot]] = annot

            for annot in annot_window(bounds[0], bounds[1], mat_doc.getAnnotations(
                atypes=ABSTRACT_MAT_ANCHOR_TYPES
            )):
                spans_by_annot[annot] = bp.BPDocumentSpan(
                    string_=signal(mat_doc, annot),
                    start=annot.start-bounds[0],
                    end=annot.end-bounds[0],
                    synclass="event-anchor", # Anchors are, by definition always their own head.
                    hstring=signal(mat_doc, annot),
                    hstart=annot.start-bounds[0],
                    hend=annot.end-bounds[0],
                    is_anchor=True,
                    no_validation=no_validation
                )
                annots_by_span[spans_by_annot[annot]] = annot

            ## Add head information, if it exists

            heads = mat_doc.getAnnotations(atypes=['Head',], ordered=True)
            no_heads = mat_doc.getAnnotations(atypes=['No-Head',], ordered=True)

            assigned_heads = []
            assigned_no_heads = []

            for annot in sorted(spans_by_annot.keys(), key=lambda a: a.end-a.start):

                if annot.atype.lab not in ABSTRACT_MAT_ANCHOR_TYPES + ["Head", "No-Head",]:

                    if DEBUG:
                        LOGGER.debug((
                            "Resolving Head value for "
                            f"{pretty_annot(annot, indent=None, doc=mat_doc)}"
                        ))

                    # instance of BPDocumentSpan
                    span = spans_by_annot[annot]

                    # find all No-Head annotations within the bounds of annot
                    candidates = [
                        a for a in annot_window(annot.start, annot.end, no_heads) \
                            if a not in assigned_no_heads
                    ]

                    # look for `Head` annots since no `No-Head`s were found
                    if len(candidates) == 0:

                        # find all Head annotations within the bounds of annot
                        candidates = [
                            a for a in annot_window(annot.start, annot.end, heads) \
                                if a not in assigned_heads
                        ]

                        # case 3
                        if len(candidates) == 0 and not no_infer_heads:

                            signal_ = signal(mat_doc, annot)
                            token = signal_.split()[0 if head_first else -1]
                            head = mat_doc.createAnnotation(
                                annot.start + signal_.rfind(token),
                                annot.start + signal_.rfind(token) + len(token),
                                "Head"
                            )
                            span.hstring.value = signal(mat_doc, head)
                            span.hstart.value = head.start - bounds[0]
                            span.hend.value = head.end - bounds[0]
                            span.hinferred.value = True
                            if DEBUG:
                                LOGGER.debug(
                                    "    Creating Head since no Head or No-Head annotations found: "
                                    f"{pretty_annot(head, None, mat_doc)}"
                                )

                        # case 1
                        elif len(candidates) == 1:

                            head = candidates[0]
                            assigned_heads.append(head)
                            span.hstring.value = signal(mat_doc, head)
                            span.hstart.value = head.start - bounds[0]
                            span.hend.value = head.end - bounds[0]
                            span.hinferred.value = False

                            if DEBUG:
                                LOGGER.debug(("    "
                                    f"Found Head annotation: {pretty_annot(head, None, mat_doc)}"
                                ))

                        # edge condition for Heads
                        elif len(candidates) > 1:

                            if DEBUG:
                                LOGGER.debug("    Found multiple Head annotations!")

                            for head in candidates:

                                smallest_annot = __find_smallest_unassigned_span(
                                    head,
                                    spans_by_annot
                                )

                                if not smallest_annot:
                                    if VERBOSE:
                                        LOGGER.warning((
                                            "Couldn't find an unassigned span for: "
                                            f"{pretty_annot(head, indent=1, doc=mat_doc)}\n"
                                            f">>> {doc_id}"
                                        ))
                                    continue

                                if DEBUG:
                                    LOGGER.debug(("        "
                                        f"Resolving Head {pretty_annot(head, None, mat_doc)} "
                                        f"and {pretty_annot(smallest_annot, None, mat_doc)}"
                                    ))

                                span = spans_by_annot[smallest_annot]
                                assigned_heads.append(head)
                                span.hstring.value = signal(mat_doc, head)
                                span.hstart.value = head.start - bounds[0]
                                span.hend.value = head.end - bounds[0]
                                span.hinferred.value = False

                    # case 2
                    elif len(candidates) == 1:

                        if DEBUG:
                            LOGGER.debug("    Found No-Head annotation.")

                        assigned_no_heads.append(candidates[0])
                        span.no_head.value = True
                        span.hinferred.value = False

                    # edge condition for No-Heads
                    elif len(candidates) > 1:

                        if DEBUG:
                            LOGGER.debug("    Found multiple No-Head annotations!")

                        for no_head in candidates:

                            smallest_annot = __find_smallest_unassigned_span(
                                no_head,
                                spans_by_annot
                            )

                            if not smallest_annot:
                                if VERBOSE:
                                    LOGGER.warning((
                                        "Couldn't find an unassigned span for: "
                                        f"{pretty_annot(no_head, indent=1)}\n"
                                        f">>> {doc_id}"
                                    ))
                                continue

                            if DEBUG:
                                LOGGER.debug(("        "
                                    f"Resolving No-Head {pretty_annot(no_head, None, mat_doc)} "
                                    f"and {pretty_annot(smallest_annot, None, mat_doc)}"
                                ))
                            span = spans_by_annot[smallest_annot]
                            assigned_no_heads.append(no_head)
                            span.no_head.value = True
                            span.hinferred.value = False

            ## build spansets

            ssids_by_annot = {}
            span_sets = {}
            ss_counter = 1
            memo = []
            for annot in spans_by_annot:

                if annot.atype.lab == "Additional_Anchor":
                    continue

                if annot not in memo:

                    ssid = f"ss-{ss_counter}"
                    ss_counter += 1

                    span_sets[ssid] = bp.BPDocumentSpanSet(
                        ssid=ssid,
                        spans=[spans_by_annot[annot],],
                        no_validation=no_validation
                    )
                    memo.append(annot)
                    ssids_by_annot[annot] = ssid
                    if DEBUG:
                        LOGGER.debug((
                            f"Creating span set {ssid} with annotation: "
                            f"{pretty_annot(annot, indent=None, doc=mat_doc)}"
                        ))

                    # From the abstract_v12.xml task file, the "Identical-To" property of the
                    # `Agent-or-Patient` object.
                    if annot.get("Identical-To") is not None:
                        for coref_annot in annot.get("Identical-To"):
                            try:
                                span_sets[ssid].spans.value.append(spans_by_annot[coref_annot])
                                if DEBUG:
                                    LOGGER.debug((
                                        "    Adding coref annotation to span set: "
                                        f"{pretty_annot(coref_annot, indent=None, doc=mat_doc)}"
                                    ))
                            except KeyError:
                                LOGGER.warning((
                                    "Found coref annotation outside of sentence boundaries (will "
                                    f"ignore): {pretty_annot(coref_annot, indent=1, doc=mat_doc)}\n"
                                    f"    Bounds: {bounds}"
                                ))
                                continue

                            memo.append(coref_annot)
                            ssids_by_annot[coref_annot] = ssid

                    if annot.get("addtl_anchor") is not None:
                        for coref_annot in annot.get("addtl_anchor"):
                            try:
                                span_sets[ssid].spans.value.append(spans_by_annot[coref_annot])
                                if DEBUG:
                                    LOGGER.debug((
                                        "    Adding addition anchor annotation to span set: "
                                        f"{pretty_annot(coref_annot, indent=None, doc=mat_doc)}"
                                    ))
                            except KeyError:
                                LOGGER.warning((
                                    "Found additional anchor annotation outside of sentence "
                                    "boundaries (will ignore): "
                                    f"{pretty_annot(coref_annot, indent=1, doc=mat_doc)}\n"
                                    f"    Bounds: {bounds}"
                                ))
                                continue

                            memo.append(coref_annot)
                            ssids_by_annot[coref_annot] = ssid

            # make span sets from the stragglers (like additional-anchors with no "parent" anchor)
            for annot in spans_by_annot:
                if annot not in memo:
                    ssid = f"ss-{ss_counter}"
                    ss_counter += 1
                    span_sets[ssid] = bp.BPDocumentSpanSet(
                        ssid=ssid,
                        spans=[spans_by_annot[annot],],
                        no_validation=no_validation
                    )
                    memo.append(annot)
                    ssids_by_annot[annot] = ssid

            ## merge spansets (transitive closure problem)

            if DEBUG:
                LOGGER.debug("Merging span-sets...")
            found_merger = True # lets us enter the loop
            while found_merger:
                found_merger = False
                for ssid_a in span_sets:
                    ss_a = span_sets[ssid_a]
                    for ssid_b in span_sets:
                        ss_b = span_sets[ssid_b]
                        if __spansets_are_mergeable(ss_a, ss_b):
                            if DEBUG:
                                LOGGER.debug(f"        Merging {ss_b} into {ss_a}")
                            __merge_span_sets(ss_a, ss_b, ssids_by_annot)
                            span_sets.pop(ssid_b)
                            found_merger = True
                            break
                    if found_merger:
                        break

            ## look for annotations with includes-relations that we didn't know about

            warn_counter = 0
            mention_types = ["Agent-or-Patient", 'SEGMENT'] # ignore SEGMENTs
            for annot in annot_window(
                bounds[0],
                bounds[1],
                mat_doc.getAnnotations(spannedOnly=True)
            ):

                if annot.atype.lab not in mention_types:

                    if annot.get('Includes'):
                        LOGGER.warning((
                            f"Found unaccounted-for annot type with includes-relations: "
                            f"{pretty_annot(annot, indent=1, doc=mat_doc)}\n"
                            f">>> {doc_id}"
                        ))
                        warn_counter += 1

            if warn_counter:
                LOGGER.warning((
                    f"Found {warn_counter} unaccounted-for annot types with includes-relations.\n"
                    f">>> {doc_id}"
                ))

            ## build includes-relations

            includes_relations = {}
            mentions = annot_window(
                bounds[0],
                bounds[1],
                mat_doc.getAnnotations(atypes=["Agent-or-Patient",], ordered=True)
            )

            for annot in mentions:

                includes = annot.get('Includes') # capitalized in Abstract, lowercase in Basic
                if includes:

                    if DEBUG:
                        LOGGER.debug((
                            f"Found includes-relations on annot: "
                            f"{pretty_annot(annot, indent=None, doc=mat_doc)}"
                        ))

                    try:
                        ssid = ssids_by_annot[annot]
                    except KeyError:
                        LOGGER.warning((
                            "Mention Annot not found in ssids_by_annot dict! Skipping: "
                            f"{pretty_annot(annot, indent=None, doc=mat_doc)}\n"
                            f">>> {doc_id}"
                        ))
                        continue

                    if ssid in includes_relations:
                        LOGGER.warning((
                            f"SSID ({ssid}) for Mention Annot already used in "
                            "includes_relations dict! Skipping: "
                            f"{pretty_annot(annot, indent=None, doc=mat_doc)}\n"
                            f">>> {doc_id}"
                        ))
                        continue

                    if DEBUG:
                        LOGGER.debug(f"    Parent SSID: {ssid}")
                    include_ssids = []
                    for include in includes:

                        if DEBUG:
                            LOGGER.debug(
                                f"    Include: {pretty_annot(include, indent=None, doc=mat_doc)}"
                            )

                        try:
                            include_ssid = ssids_by_annot[include]
                            if DEBUG:
                                LOGGER.debug(f"        SSID: {include_ssid}")
                        except KeyError:
                            LOGGER.warning(
                                "Include Annot not found in ssids_by_annot dict! Skipping: "
                                f"{pretty_annot(include, indent=None, doc=mat_doc)}"
                                f">>> {doc_id}"
                            )
                            continue

                        include_ssids.append(include_ssid)

                    includes_relations[ssid] = include_ssids

            if DEBUG:
                LOGGER.debug(f"Includes Map: {includes_relations}")

            ## build events

            event_set = {}
            event_counter = 1
            for annot in annot_window(
                bounds[0],
                bounds[1],
                mat_doc.getAnnotations(atypes=["AbstractEvent",])
            ):

                try:
                    assert annot in ssids_by_annot
                except:
                    LOGGER.warning((
                        f"Anchor annotation never parsed: "
                        f"{pretty_annot(annot, indent=1, doc=mat_doc)}\n"
                        f">>> {doc_id}"
                    ))
                    continue

                if DEBUG:
                    LOGGER.debug(f"Building Event from {pretty_annot(annot, None, mat_doc)}")
                    LOGGER.debug(f"    Anchor: {ssids_by_annot[annot]}")

                agents = []
                if annot.get("agent") is not None:
                    agents = [ssids_by_annot[agent] for agent in annot.get("agent")]
                if DEBUG:
                    LOGGER.debug(f"    Agents: {agents}")
                patients = []
                if annot.get("patient") is not None:
                    patients = [ssids_by_annot[patient] for patient in annot.get("patient")]
                if DEBUG:
                    LOGGER.debug(f"    Patients: {patients}")

                if noquad:

                    event = bp.BPDocumentAbstractEvent(
                        event_id=f"event-{event_counter}",
                        anchors=ssids_by_annot[annot],
                        agents=agents,
                        patients=patients,
                        noquad=True,
                        no_validation=no_validation
                    )

                else:

                    assert annot.get("QuadClass")

                    quadclass = annot.get("QuadClass")
                    if quadclass == "Unknown-QuadClass":
                        mv = "unk"
                        hh = "neutral"
                    else:
                        mv = quadclass.split("_")[0]
                        hh = quadclass.split("_")[1]

                    if DEBUG:
                        LOGGER.debug(f"    QuadClass: {mv} {hh}")

                    event = bp.BPDocumentAbstractEvent(
                        event_id=f"event-{event_counter}",
                        anchors=ssids_by_annot[annot],
                        agents=agents,
                        patients=patients,
                        helpful_harmful=hh.lower(),
                        material_verbal=mv.lower(),
                        no_validation=no_validation
                    )

                if event.has_duplicate_args(debug=DEBUG):
                    if DEBUG:
                        LOGGER.debug(">>> Discovered duplicate elements will be squashed!")
                        LOGGER.debug(f">>> {doc_id}, Event {event_id}")

                event_set[f"event-{event_counter}"] = event
                event_counter += 1

        ## yield entry

        entry = bp.BPDocumentEntry(
            entry_id=f"{doc_id}_{bounds[0]}_{bounds[1]}",
            doc_id=doc_id,
            lang=language,
            sent_id=str(sent_counter),
            segment_type="sentence",
            segment_text=mat_doc.signal[bounds[0]:bounds[1]],
            no_validation=no_validation
        )
        if not no_documents:
            entry.start_in_source.value = bounds[0]
            entry.end_in_source.value = bounds[1]
            entry.provenance.value = bp.BPDocumentProvenance(
                source=doc_id,
                no_validation=no_validation,
                annotator_class=annotator_label
            )
            if dataset_label:
                entry.provenance.value.add_misc_key("dataset-label", str(dataset_label))

        if not no_annotation_sets:
            entry.annotation_sets.value = bp.BPDocumentAnnotationSet(
                abstract_events=bp.BPDocumentAbstractEventSet(
                    events=event_set,
                    span_sets=span_sets,
                    includes_relations=includes_relations,
                    no_validation=no_validation
                ),
                no_validation=no_validation
            )

        if not no_validation:
            assert entry.is_valid(), f"{entry}"

        entries[entry.entry_id.value] = entry
        sent_counter += 1

    return entries


## Importable Functions


def mat2bp(
    paths,
    mat_dialect,
    corpus_id,
    no_documents=False,
    noquad=False,
    head_first=False,
    language=None,
    no_infer_heads=False,
    no_annotation_sets=False,
    no_validation=False,
    annotator_label=None,
    dataset_label=None
):
    """Converts a MAT document with the specified dialect (task) into a BP document.

    Args:
        paths (list | str): either a list of string paths to MAT documents of the given dialect, or
            a single string path to a MAT document of the given dialect.
        mat_dialect (str): the MAT task the MAT document was created with. Must be one of
            ``utils.MAT_DIALECT_CHOICES``.
        corpus_id (str): the new ID of the resulting BP document corpus.
        no_documents (bool): if True, don't include document-type entries.
        noquad (bool): if True, for Abstract, don't build QuadClass information.
        head_first (bool): infer the head as the first token in a span, rather than the last.
        language (str): an ISO 639-3 language code label to apply to each entry in the emitted BP
            document. This code should reflect the language of all of the incoming MAT documents.
        no_infer_heads (bool): if True, don't infer Heads on spans without explicitly-tagged Heads
            (default is to do so).
        no_annotation_sets (bool): if True, don't include annotation-sets.
        no_validation (bool): If True, skip validating constructed BP objects. Defaults to False.
        annotator_label (str): if passed, apply this label to all Provenance objects indicating the
            source of the anntotations. Defaults to None.
        dataset_label (str): if passed, apply a this label to all Provenance objects indicating what
            dataset these annoations belong to.

    Returns:
        bp.BPDocument: a newly constructed BP document object tree.
    """
    from datetime import date
    import os
    import sys

    from MAT.DocumentIO import getDocumentIO

    from lib import bp
    from lib.utils import BP_ANNOTATOR_CLASS_TYPES
    from lib.utils import MAT_DIALECT_CHOICES
    from lib.utils import get_tqdm

    global DEBUG
    global LOGGER
    global VERBOSE

    if mat_dialect not in MAT_DIALECT_CHOICES:
        LOGGER.error(f"Selected dialect {mat_dialect} is invalid!")
        sys.exit()

    tqdm = get_tqdm(VERBOSE, DEBUG)

    if type(paths) != list:

        assert type(paths) == str, \
            f"Value passed for 'paths' is neither list nor string: {type(paths)}"

        paths = [paths,]

    if not VERBOSE or (VERBOSE and DEBUG):
        def tqdm(iterable, *args, **kwargs):
            return iterable

    mat_io = getDocumentIO("mat-json")

    ## bookkeeping information

    assert annotator_label in BP_ANNOTATOR_CLASS_TYPES+[None,], \
        f"Bad annotator label: {repr(annotator_label)}."

    top_level_provenance = bp.BPDocumentProvenance(
        annotation_procedure="mat",
        corpus_creation_date=str(date.today()),
        source="MAT document collection",
        annotator_class=annotator_label,
        no_validation=no_validation
    )

    if dataset_label:
        top_level_provenance.add_misc_key("dataset-label", str(dataset_label))

    if not no_validation:
        assert top_level_provenance.is_valid(), \
            f"Invalid top-level provenance object: {repr(top_level_provenance)}"

    ## construct document

    bp_doc = bp.BPDocument(
        corpus_id,
        {},
        provenance=top_level_provenance,
        no_validation=no_validation
    )

    for path in tqdm(paths, desc=f"{TQDM_PREFIX} Processing MAT Documents ", unit="doc"):

        if DEBUG:
            LOGGER.debug(f"=== Processing {path} ===")

        mat_doc = mat_io.readFromSource(path)
        id_ = os.path.splitext(os.path.basename(path))[0]

        if mat_dialect == "Abstract":

            abstr_entries = __abstract2entries(
                mat_doc,
                id_,
                no_documents=no_documents,
                noquad=noquad,
                head_first=head_first,
                language=language,
                no_infer_heads=no_infer_heads,
                no_annotation_sets=no_annotation_sets,
                no_validation=no_validation,
                annotator_label=annotator_label,
                dataset_label=dataset_label
            )
            for entry_id in abstr_entries:
                bp_doc.entries.value[entry_id] = abstr_entries[entry_id]

        elif mat_dialect == "Basic":

            bp_doc.entries.value[id_] = __basic2entry(
                mat_doc,
                id_,
                head_first=head_first,
                language=language,
                no_infer_heads=no_infer_heads,
                no_annotation_sets=no_annotation_sets,
                no_validation=no_validation,
                annotator_label=annotator_label,
                dataset_label=dataset_label
            )

        elif mat_dialect == "Granular":

            bp_doc.entries.value[id_] = __basic2entry(
                mat_doc,
                id_,
                head_first=head_first,
                language=language,
                no_infer_heads=no_infer_heads,
                no_annotation_sets=no_annotation_sets,
                granular=True,
                no_validation=no_validation,
                annotator_label=annotator_label,
                dataset_label=dataset_label
            )

    if VERBOSE:
        LOGGER.info("Trimming spans...")
    bp_doc.trim_spans()
    if VERBOSE:
        LOGGER.info("Checking for bad sentence boundaries...")
    bp_doc.check_sentence_puncts() # bp_lib::Issue#31
    if VERBOSE:
        LOGGER.info("Removing duplicate event argument values...")
    bp_doc.dedupe_args() # bp_data::Issue#7

    return bp_doc


def jl2bp(corpus_id, path):
    """Munges IR document data from JSONL to BP JSON.

    Args:
        corpus_id (str): the new ID of the resulting BP document corpus.
        path (str): a path to an input JSONL file to load.

    Returns:
        bp.BPDocument: a newly constructed BP document object tree.
    """

    import json

    from lib.bp import BPDocument
    from lib.bp import BPDocumentEntry
    from lib.bp import BPDocumentProvenance
    from lib.bp import BPDocumentSegmentSection
    from lib.utils import get_tqdm

    global DEBUG
    global VERBOSE

    tqdm = get_tqdm(VERBOSE, DEBUG)

    provenance = BPDocumentProvenance(source="IR document corpus")
    doc = BPDocument(corpus_id, {}, provenance=provenance)

    with open(path, 'r') as f:
        for line in tqdm(f, desc=f"{TQDM_PREFIX} Parsing documents", unit="json"):
            ir_doc = json.loads(line)

            entry = BPDocumentEntry(
                ir_doc['derived-metadata']['id'],
                ir_doc['derived-metadata']['id'],
                sent_id=None,
                segment_type="document",
                segment_text=ir_doc['derived-metadata']['text'],
                segment_sections=[
                    BPDocumentSegmentSection.from_dict(segsec) \
                        for segsec in ir_doc['derived-metadata']['segment-sections']
                ],
                start_in_source=None,
                end_in_source=None,
                segment_start=None,
                segment_end=None,
                annotation_sets=None,
                provenance=BPDocumentProvenance(
                    source=ir_doc['derived-metadata']['warc-file']
                ),
                no_validation=False
            )

            doc.entries.value[ir_doc['derived-metadata']['id']] = entry

    return doc


## CLI Functions


@click.group(context_settings=CONTEXT_SETTINGS)
@click.option('-v', '--verbose', count=True, help="Show verbose output. Pass again for debug.")
@click.option('-f', '--force', is_flag=True, help="Skip asking to overwrite output.")
@click.option(
    '-l',
    '--save-logs',
    'save_logs',
    type=click.Path(),
    help="Capture logging output (at whatever verbosity level) to the specified path."
)
def cli(verbose, force, save_logs):
    """Convert data to various formats.

    \f

    This function serves as the entrypoint for the program when called on the CLI. It also handles
    global flag ``verbose``.

    Args:
        verbose (int): if 1, pretty-print output as the command runs. If 2, print debugging output
            in addition to level 1 output. If 0, prints nothing as the command runs. Defaults to 0.
        force (bool): if True, skip asking for permission to overwrite a file with output. Defaults
            to False.
        save_logs (str): a path to write captured logging output to.

    """
    # do this before anything else
    from lib.utils import capture_logs
    if save_logs:
        capture_logs(save_logs, verbose, force)

    global DEBUG
    global FORCE
    global LOGGER
    global VERBOSE

    if verbose:
        LOGGER.info("Verbose output is on!")
        if verbose >= 1:
            VERBOSE = True
        if verbose > 1:
            DEBUG = True
            LOGGER.debug("Debug output is on!")

    if verbose and force:
        LOGGER.warning("Forced overwrite is on!")
    elif verbose and not force:
        LOGGER.info("Forced overwrite is off!")

    FORCE = force


@cli.command('mat2bp', context_settings=CONTEXT_SETTINGS)
@click.option('--allow-duplicates', is_flag=True, help="Don't dedupe the input file list.")
@click.option(
    '--annotator-label', type=click.Choice(BP_ANNOTATOR_CLASS_TYPES),
    help="Indicates annotation source in Provenance objects."
)
@click.option(
    '--dataset-label', type=str,
    help="Indicates dataset membership in Provenance objects."
)
@click.option('--flat', is_flag=True, help="Emit a one-line file instead of indented.")
@click.option(
    '--head-first', is_flag=True,
    help=(
        "Infer the head as the first token in a span, rather than the last (default). Use for "
        "lanuages where Heads syntactically occur first in a span (like Arabic)."
    )
)
@click.option(
    '--language', type=str,
    help=(
        "Specify an ISO 639-3 language code label to apply to each entry in the emitted BP "
        "document. This code should reflect the language of all of the incoming MAT documents."
    )
)
@click.option(
    '--no-annotation-sets', is_flag=True,
    help="Don't include annotation-sets."
)
@click.option(
    '--no-documents', is_flag=True,
    help="Don't include document entries (Abstract only)."
)
@click.option(
    '--noquad', is_flag=True,
    help="For Abstract data, create BP documents without QuadClass information."
)
@click.option(
    '--no-infer-heads', is_flag=True,
    help="Don't infer Heads on spans without explicitly-tagged Heads (default is to do so)."
)
@click.option('--no-recurse', is_flag=True, help="Don't look for files in subdirectories.")
@click.option('--no-validation', is_flag=True, help="Don't validate the output BP document.")
@click.option(
    '-s', '--show-summary', count=True,
    help="Print summary statistics. Pass again for more statistics."
)
@click.argument('dialect', type=click.Choice(MAT_DIALECT_CHOICES))
@click.argument('corpus-id', type=str)
@click.argument('mat-path', type=click.Path(exists=True), nargs=-1)
@click.argument('bp-path', type=click.Path())
def mat2bp_cli(
    allow_duplicates,
    annotator_label,
    dataset_label,
    flat,
    head_first,
    language,
    no_annotation_sets,
    no_documents,
    noquad,
    no_infer_heads,
    no_recurse,
    no_validation,
    show_summary,
    dialect,
    corpus_id,
    mat_path,
    bp_path
):
    """Converts one or more MAT documents to a BP document.

    The dialect (task) of the MAT document must be specified to correctly construct the BP document.
    All MAT documents passed must be of the same dialect. Only the most recent versions of the MAT
    task definition will be considered when parsing dialect. Constructed BP document will follow the
    schema defined in ``schema/bp.schema.json``. Dialect schemas can be found in ``schema/``.

    Pass multiple paths for MAT_PATH. These can either be file paths or directories, in which case
    all MAT JSON files in the directory will be used. Unless ``--no-recurse`` is passed,
    subdirectories for input directory paths will be searched as well.

    Nota bene: MAT JSON files are identified by their ".mat.json" file extension. If file names
    exclude this extension, they will be ignored.

    \f

    Args:
        allow_duplicates (bool): if True, don't dedupe the files.
        annotator_label (str): if passed, apply this label to all Provenance objects indicating the
            source of the anntotations.
        dataset_label (str): if passed, apply a this label to all Provenance objects indicating what
            dataset these annoations belong to.
        flat (bool): if True, emit a one-liner instead of indented JSON.
        head_first (bool): infer the head as the first token in a span, rather than the last. Use
            for lanuages where Heads syntactically occur first in a span (like Arabic).
        language (str): an ISO 639-3 language code label to apply to each entry in the emitted BP
            document. This code should reflect the language of all of the incoming MAT documents.
        no_annotation_sets (bool): if True, don't include annotation-sets.
        no_documents (bool): if True, don't include document-type entries. For Abstract.
        noquad (bool): if True, for Abstract, don't build QuadClass information.
        no_infer_heads (bool): if True, don't infer Heads on spans without explicitly-tagged Heads
            (default is to do so).
        no_recurse (bool): if True, don't look for files in subdirectories passed in ``mat_path``.
            Flag passed to ``lib.utils.resolve_paths()``, not used anywhere else.
        no_validation (bool): If True, skip validating constructed BP objects. Defaults to False.
        show_summary (int): if 0, print nothing. If 1, print top-level counts. If 2, print entry
            counts as well.
        dialect (str): the MAT task the MAT document was created with. Must be one of
            ``utils.MAT_DIALECT_CHOICES``, which is verified by the Click decorator.
        corpus_id (str): the new ID of the resulting BP document corpus.
        mat_path (tuple): a list of paths to input MAT documents or directories of MAT documents to
            load.
        bp_path (str): a path to output a new BP document.
    """

    import os
    import sys

    from tabulate import tabulate

    from lib.utils import resolve_paths

    global FORCE
    global LOGGER
    global VERBOSE

    def __vpath(p):
        assert len(p) > 0, f"No elements in iterable! Did you forget CORPUS_ID? -> {p}"
        str_ = ""
        if type(p) == tuple:
            p = list(p)
            if os.path.isfile(p[0]):
                str_ += f"...{os.path.basename(p[0])}" if os.path.basename(p[0]) != p[0] else p[0]
            else:
                if p[0].endswith("/") or p[0].endswith("\\"):
                    p[0] = p[0][:-1]
                str_ += f"...{os.path.normpath(p[0])}" if os.path.normpath(p[0]) != p[0] else p[0]
            if len(p) > 1:
                str_ += "\n..."
            return str_
        else:
            return f"...{os.path.basename(p)}" if os.path.basename(p) != p else p

    if DEBUG:
        LOGGER.debug("Invocation Arguments:")

        for line in tabulate([
            ["annotator_label", type(annotator_label).__name__, annotator_label],
            ["allow_duplicates", type(allow_duplicates).__name__, allow_duplicates],
            ["dataset_label", type(dataset_label).__name__, dataset_label],
            ["flat", type(flat).__name__, flat],
            ["language", type(language).__name__, language],
            ["head_first", type(head_first).__name__, head_first],
            ["no_annotation_sets", type(no_annotation_sets).__name__, no_annotation_sets],
            ["no_documents", type(no_documents).__name__, no_documents],
            ["noquad", type(noquad).__name__, noquad],
            ["no_infer_heads", type(no_infer_heads).__name__, no_infer_heads],
            ["no_recurse", type(no_recurse).__name__, no_recurse],
            ["no_validation", type(no_validation).__name__, no_validation],
            ["show_summary", type(show_summary).__name__, show_summary],
            ["dialect", type(dialect).__name__, dialect],
            ["corpus_id", type(corpus_id).__name__, corpus_id],
            ["mat_path", type(mat_path).__name__, __vpath(mat_path)],
            ["bp_path", type(bp_path).__name__, __vpath(bp_path)],
        ], headers=["Argument", "Type", "Value"], tablefmt="fancy_grid").split("\n"):
            LOGGER.debug(line)

    try:
        if os.path.dirname(bp_path):
            assert os.path.exists(os.path.dirname(bp_path))
    except AssertionError:
        LOGGER.error(f"Directory {os.path.dirname(bp_path)} doesn't exist, can't write file!")
        sys.exit()

    paths = resolve_paths(
        mat_path,
        ".mat.json",
        no_recurse=no_recurse,
        allow_duplicates=allow_duplicates,
        verbose=VERBOSE,
        debug=DEBUG
    )

    if not paths:
        LOGGER.error("No MAT documents found in input paths!")
        sys.exit(1)

    ## crunch data

    bp_doc = mat2bp(
        paths,
        dialect,
        corpus_id,
        no_documents=no_documents,
        noquad=noquad,
        head_first=head_first,
        language=language,
        no_infer_heads=no_infer_heads,
        no_annotation_sets=no_annotation_sets,
        no_validation=no_validation,
        annotator_label=annotator_label,
        dataset_label=dataset_label
    )

    ## emit BP doc

    if show_summary:
        if VERBOSE:
            LOGGER.info("Summary Results:")
            for line in bp_doc.stats(
                pretty=True,
                abstract_events=show_summary>1 and dialect=="Abstract",
                basic_events=show_summary>1 and dialect in ["Basic", "Granular",],
                granular_templates=show_summary>1 and dialect=="Granular"
            ).split("\n"):
                LOGGER.info(line)
        else:
            click.echo(bp_doc.stats(
                abstract_events=show_summary>1 and dialect=="Abstract",
                basic_events=show_summary>1 and dialect in ["Basic", "Granular",],
                granular_templates=show_summary>1 and dialect=="Granular"
            ))

    if VERBOSE:
        LOGGER.info("Serializing...")

    bp_doc.serialize(bp_path, flat=flat, verbose=VERBOSE, force=FORCE)

    if VERBOSE:
        LOGGER.info(f"New BP doc saved to {bp_path}")


@cli.command('txt2mat', context_settings=CONTEXT_SETTINGS)
@click.argument('txt-path', type=click.Path(exists=True))
@click.argument('mat-path', type=click.Path())
def txt2mat_cli(txt_path, mat_path):
    """Converts one UTF-8 plain text file to a MAT document.

    \f

    Args:
        txt_path (str): a path to an input plain text file to load.
        mat_path (str): a path to output a new MAT document.
    """

    import MAT

    global LOGGER
    global VERBOSE

    raw_io = MAT.DocumentIO.getDocumentIO("raw", encoding="utf-8")
    json_io = MAT.DocumentIO.getDocumentIO("mat-json")
    json_io.writeToTarget(raw_io.readFromSource(txt_path), mat_path)

    if VERBOSE:
        LOGGER.info(f"New document saved to '{mat_path}'.")


@cli.command('txt2bp', context_settings=CONTEXT_SETTINGS)
@click.argument('corpus-id', type=str)
@click.argument('txt-path', type=click.Path(exists=True))
@click.argument('bp-path', type=click.Path())
def txt2bp_cli(corpus_id, txt_path, bp_path):
    """Converts one UTF-8 plain text file to a BP document.

    \f

    Args:
        corpus_id (str): the new ID of the resulting BP document corpus.
        txt_path (str): a path to an input plain text file to load.
        bp_path (str): a path to output a new BP document.
    """
    import os

    from lib.bp import BPDocument
    from lib.bp import BPDocumentEntry
    from lib.bp import BPDocumentProvenance

    global FORCE
    global LOGGER
    global VERBOSE

    text = open(txt_path, 'r', encoding='utf-8').read()

    bp_doc = bp.BPDocument(
        corpus_id,
        {
            "entry-1": BPDocumentEntry(
                "entry-1",
                os.path.splitext(os.path.basename(txt_path))[0],
                segment_text=text
            )
        },
        provenance=BPDocumentProvenance(
            corpus_creation_date=str(datetime.date.today()),
            source="Plain text file"
        )
    )

    bp_doc.serialize(bp_path, verbose=VERBOSE, force=FORCE)


@cli.command('jl2bp', context_settings=CONTEXT_SETTINGS)
@click.argument('corpus-id', type=str)
@click.argument('jl-path', type=click.Path(exists=True))
@click.argument('bp-path', type=click.Path())
def jl2bp_cli(corpus_id, jl_path, bp_path):
    """Munges IR document data from JSONL to BP JSON.

    IR Documents are JSONL collections where each JSON document contains information that can
    be used to create a BP document. This function extracts the documents out and builds a new BP
    file from the collection.

    \f

    Args:
        corpus_id (str): the new ID of the resulting BP document corpus.
        jl_path (str): a path to an input JSONL file to load.
        bp_path (str): a path to output a new BP document.
    """

    global FORCE
    global LOGGER
    global VERBOSE

    bp_doc = jl2bp(corpus_id, jl_path)

    if VERBOSE:
        LOGGER.info("Serializing...")

    bp_doc.serialize(bp_path, verbose=VERBOSE, force=FORCE)

    if VERBOSE:
        LOGGER.info(f"New BP doc saved to {bp_path}")


if __name__ == '__main__':

    ## program proper

    cli()
