# Copyright (C) 2014 - 2021 The MITRE Corporation. See the toplevel
# file LICENSE.txt for license terms.

from __future__ import print_function

# This skeleton has been pared down from an original, much more
# complex infrastructure which recapitulated a fair bit of the
# workflow and model training mechanics. At the moment, it serves
# as a framework for building specific implementations, such as
# engines for the jCarafe maxent classifier, or for MALLET or Weka.
# Less specific classes in this file will probably be suitable for
# managing interactions with deep-learning classifiers like BERT.

#
# Attribute classifiers
#

# So we're going to steal the infrastructure from the trivial relation tagger
# to build a wrapper for a maxent classifier. The classifiers target
# actual attributes, but these attribute sets can be handled as a
# group, or individually. We're going to focus on the classes which
# handle attributes one-by-one. 

import os, sys, re, tempfile

from .Annotation import AnnotationError, AnnotationCore, AttributeValueSet
from .ModelBuilder import ModelBuilderError
from . import ModelBuilder
from . import ExecutionContext
from .PluginMgr import CmdlineTagStep, PluginError, FindPluginClass, FindPluginObject
from .Operation import OpArgument, Option, OptionBearer, XMLOpArgumentAggregator, OptionTemplate
from .Error import TaggerConfigurationError

from .PyCompat import _strtypes

# So the idea is that there will be items that will be classified.
# In many cases, the item will be a single span annotation. In other cases,
# we'll have tools to extract these items. There must be one item for each
# line in the classifier.

from .AnnotationTree import TreeBuilder, ETYPE_ANNOT

#
# Toplevel
#

# We start with an AttributeClassifierEngine, which takes a set of
# labels and the relevant attributes. Each classifier is responsible
# for how it manages the set. The one I'm going to focus on is
# the one which treats each attribute separately.

# First, reading and writing metadata, so I can store all 
# the relevant information.

def _writeMetadata(modelOutputDir, metadata):
    # Write the decodes, so we can grab them when we apply.
    import json, codecs
    fp = codecs.open(os.path.join(modelOutputDir, "model_metadata.json"), "w", "utf8")
    fp.write(json.dumps(metadata))
    fp.close()

def _readMetadata(modelDir):
    p = os.path.join(modelDir, "model_metadata.json")
    if not os.path.isfile(p):
        raise PluginError("no model metadata file for %s" % modelDir)
    import json, codecs
    fp = codecs.open(p, "r", "utf8")
    metadata = json.loads(fp.read())
    fp.close()
    return metadata

class AttributeClassifierEngine(object):

    def __init__(self, task, modelDir, labelsAndAttrs, treeBuilder = None,
                 treeLevels = None,
                 featureExtractorResources = None):
        self.task = task
        # dictionary from label string to list of attribute objects.
        self.labelsAndAttrs = labelsAndAttrs
        self.modelDir = modelDir
        self.treeBuilder = None
        self._tbObj = None
        self.treeLevels = treeLevels
        self.featureExtractorResources = featureExtractorResources

    def getFeatureExtractorResources(self):
        return self.featureExtractorResources

    # I used to be able to pass in a tree builder, but I've eliminated that
    # for the moment, because I'm creat
    def getTreeBuilder(self):
        if self._tbObj is None:            
            if self.treeBuilder:
                self._tbObj = self.treeBuilder(self.task)
            else:
                tb = TreeBuilder(self.task, levels = self.treeLevels)
                # Just tokens and all content, for the moment.
                if self.treeLevels is None:
                    tb.setLevels()
                self._tbObj = tb
        return self._tbObj

    def buildTrees(self, docPairList):
        return self.getTreeBuilder().buildTrees(docPairList)
            
    # If applyModel is True, the label will be UNK. We need to extract
    # the tokens for each annotation before we do anything.
    
    def trainToplevel(self, docPairList, outDir, outStream = None, **kw):
        modelDir = self.modelDir
        # Corpus preparation has to happen before anything, including
        # tree generation.
        self.prepareCorpusForTrain(docPairList, modelDir)
        # Each tuple is (doc, annot, tokList, cache).
        # The cache can be used cross-classifier.
        allDocs = self.buildTrees(docPairList)
        self.train(allDocs, modelDir, outDir, outStream = outStream, **kw)
        
    def applyModelToplevel(self, docPairList, outDir, **kw):
        modelDir = self.modelDir
        self.prepareCorpusForApplyModel(docPairList, modelDir)
        # Each tuple is (doc, annot, tokList, cache).
        # The cache can be used cross-classifier.
        allDocs = self.buildTrees(docPairList)
        self.applyModel(allDocs, modelDir, outDir, **kw)

    def generateFeatureData(self, docPairList):
        res = []
        modelDir = self.modelDir        
        self.prepareCorpusForGenerateFeatureData(docPairList, modelDir)        
        # Each tuple is (doc, annot, tokList, cache).
        # The cache can be used cross-classifier.
        allDocs = self.buildTrees(docPairList)
        return self.generateFeatureDataFromItems(allDocs, modelDir)

    def writeModelMetadata(self, mDir):
        _writeMetadata(mDir, self.getModelMetadata())

    # We can't save the tree builder. I hope this doesn't get me in trouble
    # down the line. It probably will...
    def getModelMetadata(self):
        c = self.__class__
        return {"mclass": c.__module__ + "." + c.__name__,
                "labels_and_attrs": {label: [attr.name for attr in attrs] for (label, attrs) in self.labelsAndAttrs.items()},
                "tree_levels": self.treeLevels,
                "feature_extractor_resources": self.featureExtractorResources}

    @classmethod
    def fromMetadata(cls, task, modelDir, metadata):
        args, kw = cls.classCreationArgs(task, metadata)
        return cls(*([task, modelDir] + args), **kw)

    @classmethod
    def classCreationArgs(cls, task, metadata):
        repo = task.getAnnotationTypeRepository()
        labelsAndAttrs = {}

        for label, attrList in metadata["labels_and_attrs"].items():
            labObj = repo.findAnnotationType(label, None, create = False)
            allAttrs = []
            if labObj is None:
                # I don't know how this would be possible, but lets just check.
                raise ModelBuilderError("label '%s' for model builder unknown" % label)
            for attribute in attrList:
                if attribute not in labObj.attr_table:
                    raise ModelBuilderError("attribute '%s' for label '%s' unknown for model builder" % (attribute, label))
                allAttrs.append(labObj.attr_list[labObj.attr_table[attribute]])
            labelsAndAttrs[label] = allAttrs

        d = {}
        if metadata.get("tree_levels"):
            d["treeLevels"] = metadata["tree_levels"]
        if metadata.get("feature_extractor_resources"):
            d["featureExtractorResources"] = metadata["feature_extractor_resources"]
        return [labelsAndAttrs], d

    #
    # Specializations
    #

    # The default implementation does nothing.    
    def prepareCorpusForTrain(self, docPairList, modelDir):
        pass    

    def train(self, allDocs, modelDir, outDir, outStream = None, **kw):
        raise Exception("not implemented")
    
    # Default implementation does nothing.
    def prepareCorpusForApplyModel(self, docPairList, modelDir):
        pass

    def applyModel(self, allDocs, modelDir, outDir, **kw):
        raise Exception("not implemented")

    # Default implementation does nothing.
    def prepareCorpusForGenerateFeatureData(self, docPairList, modelDir):
        pass    

    def generateFeatureDataFromItems(self, allDocs, modelDir):
        raise Exception("not implemented")

# Per attribute. Everything has to be segregated by label and attribute,
# and the classifier class has to be applied.

# I need some helpers to encapsulate the behavior of each attribute,
# because I don't want multiple instances of the classifier subclass.

class ClassifierAttributeManager(object):

    def __init__(self, label, attrObj, nullLabel = "NULL"):
        self.label = label
        self.attr = attrObj
        self.nullLabel = nullLabel

    def getSubpath(self, p):
        return os.path.join(p, self.getLabel())

    def getLabel(self):
        return self.label + "_" + self.attr.name

class BooleanClassifierAttributeManager(ClassifierAttributeManager):

    def extractLabel(self, item):
        return ((item.annot.get(self.attr.name) and "yes") or "no")

    # So this is a bit complicated. The default value may be
    # True or False or None. None and False are equivalent from
    # the point of view of extractLabel(), so I think we only
    # want to set True, unless the default is True and the value
    # we get is "no".
    def applyLabel(self, lab, item):
        v = (lab == "yes")
        if (v is True) or (self.attr.default is True):
            item.annot[self.attr.name] = v

class SingletonValueClassifierAttributeManager(ClassifierAttributeManager):

    def extractLabel(self, item):
        v = item.annot.get(self.attr.name)
        if v is None:
            return self.nullLabel
        elif self.attr._typename_ == "int":
            return str(v)
        else:
            return v

    def applyLabel(self, lab, item):
        annot = item.annot
        attrName = self.attr.name
        if lab == self.nullLabel:
            annot[attrName] = None
        elif self.attr._typename_ == "int":
            annot[attrName] = int(lab)
        else:
            annot[attrName] = lab            

class AggregateValueClassifierAttributeManager(ClassifierAttributeManager):

    def __init__(self, label, attrObj, val):
        ClassifierAttributeManager.__init__(self, label, attrObj)
        self.val = val

    def getLabel(self):
        return self.label + "_" + self.attr.name + "_" + str(self.val)

    def extractLabel(self, item):
        a = item.annot
        v = a.get(self.attr.name)
        if v is None:
            return self.nullLabel
        elif self.val not in v:
            return self.nullLabel
        else:
            return self.val

    def applyLabel(self, lab, item):
        annot = item.annot
        attrName = self.attr.name
        if lab != self.nullLabel:
            if self.attr._typename_ == "int":
                lab = int(lab)
            curVal = annot.get(attrName)
            if curVal is None:
                annot[attrName] = AttributeValueSet([lab])
            else:
                curVal.add(lab)

class PerAttributeClassifierEngine(AttributeClassifierEngine):

    def __init__(self, task, modelDir, labelsAndAttrs, cclass, cclassSettings, **kw):
        AttributeClassifierEngine.__init__(self, task, modelDir, labelsAndAttrs, **kw)
        # This will have to be decoded somehow.
        self.managers = []
        for label, attrs in self.labelsAndAttrs.items():            
            for attr in attrs:
                self._addManagers(label, attr)
        if not issubclass(cclass, PerAttributeClassifier):
            raise AnnotationError("Per-attribute classifier engine classifier '%s' is not a PerAttributeClassifier" % (cclass.__module__ + "." + cclass.__name__))
        self.cclass = cclass
        self.cclassSettings = cclassSettings
        self.cInstance = self.cclass(self.task, featureExtractorResources = self.featureExtractorResources, **self.cclassSettings)
        self.cInstance.setEngine(self)

    def _addManagers(self, label, attrObj):
        if attrObj._typename_ == "boolean":
            if attrObj.aggregation:
                raise AnnotationError("boolean attribute '%s' of '%s' for span attribute classifier is an aggregation" % (attrObj.name, label))
            self.managers.append(BooleanClassifierAttributeManager(label, attrObj))
        elif attrObj._typename_ in ("float", "annotation"):
            raise AnnotationError("attribute '%s' of '%s' for span attribute classifier is a float or annotation" % (attrObj.name, label))
        elif not attrObj.choices:
            # DON'T test for _choiceAttribute - that's only present if it's a singleton.
            raise AnnotationError("attribute '%s' of '%s' for span attribute classifier is not a choice attribute" % (attrObj.name, label))
        elif attrObj.aggregation == "list":
            raise AnnotationError("attribute '%s' of '%s' for span attribute classifier is a list aggregation" % (attrObj.name, label))
        elif attrObj.aggregation == "set":
            for choice in attrObj.choices:
                self.managers.append(AggregateValueClassifierAttributeManager(label, attrObj, choice))
        else:
            self.managers.append(SingletonValueClassifierAttributeManager(label, attrObj))
                
    def getModelMetadata(self):
        d = AttributeClassifierEngine.getModelMetadata(self)
        d["cclass"] = self.cclass.__module__ + "." + self.cclass.__name__
        d["cclass_settings"] = self.cclassSettings
        return d

    @classmethod
    def classCreationArgs(cls, task, metadata):
        args, kw = AttributeClassifierEngine.classCreationArgs(task, metadata)
        args.append(FindPluginClass(metadata["cclass"], task.name, parentClass = PerAttributeClassifier, tpe = "per-attribute classifier class"))
        args.append(metadata["cclass_settings"])
        return args, kw
        
    def prepareCorpusForTrain(self, docPairList, storageDir):
        for manager in self.managers:
            self.cInstance.prepareCorpusForTrain(manager, docPairList, manager.getSubpath(storageDir))
                
    def prepareCorpusForApplyModel(self, docPairList, storageDir):
        for manager in self.managers:
            self.cInstance.prepareCorpusForApplyModel(manager, docPairList, manager.getSubpath(storageDir))
                
    def prepareCorpusForGenerateFeatureData(self, docPairList, storageDir):
        for manager in self.managers:
            self.cInstance.prepareCorpusForGenerateFeatureData(manager, docPairList, manager.getSubpath(storageDir))
    
    # There's a hierarchy of places where we can stop. The issue here
    # is that in some cases, we might want to collect a pile of data from
    # the docs before we generate the items. This is INDEPENDENT of the
    # label and attribute.
    
    def extractItemsFromCorpus(self, allDocs, items):
        for treeRoot in allDocs:
            self.extractItems(treeRoot, items)

    def extractItems(self, treeRoot, itemDict):
        # Look in the appropriate level.
        for label in self.labelsAndAttrs.keys():
            try:
                itemList = itemDict[label]
            except KeyError:
                itemList = []
                itemDict[label] = itemList
            l = treeRoot.builder.levelHash[label]
            content = [elt for elt in treeRoot.getSequence(treeRoot.builder.levels[l][0])
                       if elt.annot.atype.lab == label]
            resources = self.getFeatureExtractorResources()
            for tNode in content:
                itemList.append(ClassifierItem(tNode, resources = resources))

    # Let's say you run the trainer, and there's no data. jCarafe maxent will fail;
    # and if there are multiple classifiers, the whole chain will fail. Should that
    # happen? What that would mean is that nothing should be applied when the
    # model is applied. Since I know how many rows I have when I save out the
    # sequence data.

    def train(self, allDocs, modelDir, outDir, outStream = None, **kw):
        itemDict = {}
        self.extractItemsFromCorpus(allDocs, itemDict)
        for manager in self.managers:
            items = itemDict[manager.label]
            mgrModelDir = manager.getSubpath(modelDir)
            if not os.path.exists(mgrModelDir):
                os.makedirs(mgrModelDir)
            # We pass in the task because the classifiers are still
            # potentially standalone objects, and they may not point
            # to the engine.
            dataPath = None
            if outDir:
                dataPath = manager.getSubpath(outDir)
                if not os.path.exists(dataPath):
                    os.makedirs(dataPath)
            self.cInstance.train(manager, items,
                                 os.path.join(mgrModelDir, "model"),
                                 dataPath,
                                 outStream = outStream, **kw)

    def applyModel(self, allDocs, modelDir, outDir, **kw):
        itemDict = {}
        self.extractItemsFromCorpus(allDocs, itemDict)
        for manager in self.managers:
            items = itemDict[manager.label]
            mgrModelDir = manager.getSubpath(modelDir)
            if not os.path.exists(mgrModelDir):
                os.makedirs(mgrModelDir)
            dataPath = os.path.join(manager.getSubpath(outDir), "data")
            os.makedirs(dataPath)
            self.cInstance.applyModel(manager, items,
                                      os.path.join(mgrModelDir, "model"),
                                      dataPath = dataPath,
                                      **kw)

    def generateFeatureDataFromItems(self, allDocs, modelDir):
        itemDict = {}
        self.extractItemsFromCorpus(allDocs, itemDict)
        res = []
        labelDict = {}
        for manager in self.managers:
            items = itemDict[manager.label]
            r = self.cInstance.generateFeatureDataFromItems(manager, items, os.path.join(manager.getSubpath(modelDir), "model"))
            if r is not None:
                try:
                    labelDict[manager.label] += r
                except KeyError:
                    labelDict[manager.label] = r[:]
        # For consistency with previous return value.
        return tuple([(label, None, r) for label, r in labelDict.items()])

#
# Next level down in the organization of classes: the cclass which handles the
# per-attribute implementation. This can be anything from jCarafe maxent to
# deep-learning BERT, at least I hope so.
#

# This class needs to be a child of OptionBearer so that it can use class-defined OpArgument
# args lists to populate the classifier from the build settings in task.xml.

            
class Classifier(OptionBearer):

    # Let's have an empty list here, just for convenience.
    # The option processing code doesn't require it, but children
    # of this class might want to append things to the list.
    args = []

    def __init__(self, task, featureExtractorResources = None, **kw):
        OptionBearer.__init__(self)
        self.task = task
        self.engine = None
        self.featureExtractorResources = featureExtractorResources

    # Defined for the benefit of extractItems, if needed.
    def getFeatureExtractorResources(self):
        return self.featureExtractorResources or \
               (self.engine and self.engine.getFeatureExtractorResources())

    # An item processor is set.
    def setEngine(self, p):
        self.engine = p

    #
    # Specializations
    #

    # The default implementation does nothing.    
    def prepareCorpusForTrain(self, docPairList, modelDir):
        pass    

    # Default implementation does nothing.
    def prepareCorpusForGenerateFeatureData(self, docPairList, modelDir):
        pass    
    
    # Default implementation does nothing.
    def prepareCorpusForApplyModel(self, docPairList, modelDir):
        pass

    def train(self, items, modelDir, dataDir, outStream = None, **kw):
        raise Exception("not implemented")

    def applyModel(self, items, modelDir, outDir, **kw):
        raise Exception("not implemented")

    def generateFeatureDataFromItems(self, items, modelDir):
        raise Exception("not implemented")

# The PerAttributeClassifier class is still an unimplementable class, but it's intended
# to be instantiated by label and attr. At some point, I may want to have a label-scope
# classifier, but for now, we collapse the behavior for label and attr.


class PerAttributeClassifier(Classifier):

    # All of these methods now take a manager as an initial argument.

    # The default implementation does nothing.    
    def prepareCorpusForTrain(self, manager, docPairList, modelDir):
        pass    

    # Default implementation does nothing.
    def prepareCorpusForGenerateFeatureData(self, manager, docPairList, modelDir):
        pass    
    
    # Default implementation does nothing.
    def prepareCorpusForApplyModel(self, manager, docPairList, modelDir):
        pass

    def train(self, manager, items, modelDir, dataDir, outStream = None, **kw):
        raise Exception("not implemented")

    def applyModel(self, manager, items, modelDir, outDir, **kw):
        raise Exception("not implemented")

    def generateFeatureDataFromItems(self, manager, items, modelDir):
        raise Exception("not implemented")

# Now we're getting even more specific. This is a class which
# is intended to be used with explicit feature extractors.
# I'm NOT going to allow the feature extractors to be defined on the
# command line, but rather only in the task configuration.

# And a helper class to manage the manipulation of features.

class ProcessedClassifierItemEntry(object):

    def __init__(self, label, item, features, encodedLabel = None):
        self.label = label
        self.encodedLabel = encodedLabel
        self.item = item
        self.features = features
        self.encodedFeatures = features
        self.resultData = None

class PerAttributeFeatureExtractorClassifier(PerAttributeClassifier):

    # The feature extractors are dealt with when the engine is set.
    def __init__(self, task, feature_extractors = None, **kw):
        PerAttributeClassifier.__init__(self, task, **kw)
        self.featureExtractors = []
        # Decode the list of feature extractors, if present.
        if feature_extractors is not None:
            for fExtractor in [s.strip() for s in feature_extractors.split(",")]:
                extractor = getRegisteredFeatureExtractor(fExtractor)
                if extractor is None:
                    raise ModelBuilderError("feature extractor '%s' not found for PerAttributeClassifier" % fExtractor)
                self.featureExtractors.append(extractor)

    #
    # Methods to train and apply
    #

    def train(self, manager, items, modelDir, dataDir, outStream = None, saveFeatureVectors = False, saveArff = False):
        # The model directory is modelDir/label/"model"

        labelDir = os.path.dirname(modelDir)

        if saveFeatureVectors:
            dataPath = os.path.join(labelDir, "fvecs.out")
        else:
            dataPath = None
            if dataDir:
                dataPath = os.path.join(dataDir, "fvecs.out")
                
        if saveArff:
            arffPath = os.path.join(labelDir, "fvecs.arff")
        else:
            arffPath = None

        metadata = {"label_mappings": None,
                    "label_count": None,
                    "item_count": len(items),
                    "trainer_error": False}

        if len(items) == 0:
            _writeMetadata(labelDir, metadata)
            return

        # The OTHER situation where it'll barf is when there's only one label
        # target.

        dataPathExists = (dataPath is not None)
        
        if not dataPathExists:
            tmpd = tempfile.mkdtemp()
            dataPath = os.path.join(tmpd, "fvecs")

        try:

            featureItems = [ProcessedClassifierItemEntry(manager.extractLabel(item), item, self.extractFeatures(item))
                            for item in items]
            metadata["label_mappings"] = {}
            metadata["label_count"] = self.createFeatureFile(dataPath, featureItems, encodedLabelsToPublicLabels = metadata["label_mappings"])

            if saveArff:
                self.createArffFile(arffPath, featureItems)

            if metadata["label_count"] < 2:
                _writeMetadata(labelDir, metadata)
                return

            self.executeModelBuild(dataPath, modelDir, outStream, metadata)

            _writeMetadata(labelDir, metadata)

        finally:
            if not dataPathExists:
                import shutil
                try:
                    shutil.rmtree(tmpd)
                except:
                    pass
            
    def applyModel(self, manager, items, model, dataPath, save_feature_vectors_for_scoring = False):

        # The model is modelDir/label/"model"

        labelDir = os.path.dirname(model)
        
        if save_feature_vectors_for_scoring:
            saveFeatureVectorsForScoring = manager.label
            for item in items:
                doc = item.annot.doc
                if "feature_vec_output" not in doc.metadata:
                    # I THINK.
                    doc.metadata["feature_vec_output"] = {"modelDir": os.path.dirname(labelDir), "fvecs": {}}
        else:
            saveFeatureVectorsForScoring = None

        if len(items) == 0:
            # jCarafe will barf. Don't even bother.
            return

        metadata = _readMetadata(labelDir)

        if metadata["item_count"] == 0:
            print("WARNING: no model %s because no data was available when the model was created" % model, file = sys.stderr)
            return
        elif metadata["label_count"] < 2:
            print("WARNING: no model %s because only one target label was present in the training data" % model, file = sys.stderr)
            return
        elif metadata["trainer_error"]:
            raise PluginError("no model %s because model build failed" % model)

        featureItems = [ProcessedClassifierItemEntry("UNK", item, self.extractFeatures(item))
                        for item in items]

        dataPathExists = (dataPath is not None)
        
        if not dataPathExists:
            tmpd = tempfile.mkdtemp()
            dataPath = os.path.join(tmpd, "fvecs")            

        try:
            inPath = dataPath + "_input"
            outPath = dataPath + "_output"

            self.createFeatureFile(inPath, featureItems)

            self.executeModelApplication(inPath, outPath, model)

            self.decodeModelApplicationResult(featureItems, outPath, metadata["label_mappings"])

            for featureItem in featureItems:
                manager.applyLabel(featureItem.label, featureItem.item)

            if saveFeatureVectorsForScoring:
                # Let's figure out where we are.
                # It's the last two components of the model.
                # And what we want to save is the postprocessed feature bag.
                classifierName = os.path.basename(model)
                labName = saveFeatureVectorsForScoring
                for featureItem in featureItems:
                    self._maybeRecordFeatureVectorTagResult(labName, classifierName, featureItem)

        finally:
            if not dataPathExists:
                import shutil
                try:
                    shutil.rmtree(tmpd)
                except:
                    pass            

    def generateFeatureDataFromItems(self, manager, items, model):

        # The model is modelDir/label/"model"

        labelDir = os.path.dirname(model)

        if len(items) == 0:
            # jCarafe will barf. Don't even bother.
            return []

        metadata = _readMetadata(labelDir)

        if metadata["item_count"] == 0:
            print("WARNING: no model %s because no data was available when the model was created" % model, file = sys.stderr)
            return []
        elif metadata["label_count"] < 2:
            print("WARNING: no model %s because only one target label was present in the training data" % model, file = sys.stderr)
            return []
        elif metadata["trainer_error"]:
            raise PluginError("no model %s because model build failed" % model)

        featureItems = [ProcessedClassifierItemEntry(manager.extractLabel(item), item, self.extractFeatures(item))
                        for item in items]

        self.encodeFeaturesForClassifier(featureItems)
        return featureItems

    # When do I record the info? If the item has an annot.
    # If the item has an annot, I go to the document, make sure
    # the annot has an ID (well, a sequence of IDs, since
    # it may be a pair or something), and then record the
    # relevant data.
    
    def _maybeRecordFeatureVectorTagResult(self, labName, classifierName, featureItem):
        line = featureItem.resultData
        taggedLab = featureItem.encodedLabel
        item = featureItem.item
        featureBag = featureItem.encodedFeatures
        annots = item.getAnnotationSequenceForScoring()
        if annots:
            aids = [annot.getID() for annot in annots]
            # This has already been primed in applyModel() at the toplevel.
            fvecOutput = annots[0].doc.metadata["feature_vec_output"]["fvecs"]
            try:
                labEntry = fvecOutput[labName]
            except KeyError:
                labEntry = {}
                fvecOutput[labName] = labEntry
            try:
                classifierEntry = labEntry[classifierName]
            except KeyError:
                classifierEntry = []
                labEntry[classifierName] = classifierEntry
            classifierEntry.append((aids, item.getDescriptionForScore(truncateString = 30),
                                    line, taggedLab, list(featureBag)))

    def extractFeatures(self, item):
        featureBag = set()
        featureDict = {}        
        # This had better not fail. I don't see how it can, frankly -
        # The feature extractors are only allowed in fromXML if they've been
        # declared, and if they've been declared, they'll be in the cache.
        # And if they're not cached, we run the extractor.
        for e in self.featureExtractors:
            features = e(item)
            for f in features:
                if type(f) in _strtypes:
                    if f in featureDict:
                        print("WARNING: feature extractors produced feature %s both with and without a weight; treating absent weight as 1" % f, file=sys.stderr)
                        featureDict[f] += 1
                    else:
                        featureBag.add(f)
                elif f[0] in featureBag:
                    print("WARNING: feature extractors produced feature %s both with and without a weight; treating absent weight as 1.0" % f[0], file=sys.stderr)
                    featureDict[f[0]] = 1 + f[1]
                    featureBag.remove(f[0])
                elif f[0] in featureDict:
                    print("WARNING: feature extractors produced weighted feature %s multiple times; summing weights" % f[0], file = sys.stderr)
                    featureDict[f[0]] += f[1]
                else:
                    featureDict[f[0]] = f[1]
        featureBag.update(featureDict.items())
        return featureBag
            
    def createArffFile(self, outFile, featureItems):
        # It's a little crazy to reconstruct the contents of the feature
        # file, but frankly, it's easier than modifying createFeatureFile to accommodate it.
        import codecs, string
        # For Weka, I'm going to use a sparse distribution, with a dummy string
        # attribute at the top. The target label is index 1.
        indexMap = {}
        actualAttrNames = set()
        classVals = set()
        headers = ["@RELATION rel", ""]
        otherHeaders = []
        dataLines = []
        curIdx = 2
        for featureItem in featureItems:
            lab = featureItem.encodedLabel
            item = featureItem.item
            featureBag = featureItem.encodedFeatures
            classVals.add(lab)
            sparseEntries = [(1, lab)]
            for t in featureBag:
                # Each token has to start with an alphabetic. if the token
                # doesn't start with something like that, add f at the front.
                # The restrictions on the attribute names are really poorly described.
                # It looks like single quotes are significant. There's a Weka
                # wiki which says that % is a comment character and must be quoted,
                # and that double-qoutes are acceptable, and that backslash is
                # an escape character. Also, I can't end up with multiple instances
                # of the same feature. 
                tSplit = t.split(":", 1)
                if len(tSplit) > 1:
                    feature = tSplit[0]
                    val = tSplit[1]
                else:
                    feature = t
                    val = "1.0"
                try:
                    pos = indexMap[feature]
                except KeyError:
                    pos = curIdx
                    indexMap[feature] = pos
                    if feature[0] not in string.ascii_letters:
                        feature = "f"+feature
                    if "'" in feature:
                        feature = feature.replace("'", "_")
                    if "\\" in feature:
                        feature = feature.replace("\\", "_")
                    i = 0
                    fName = feature                    
                    while fName in actualAttrNames:
                        i += 1
                        fName = feature + "_" + str(i)
                    otherHeaders.append("@ATTRIBUTE '%s' NUMERIC" % fName)
                    actualAttrNames.add(fName)
                    curIdx += 1
                sparseEntries.append((pos, val))
            sparseEntries.sort(key = lambda x: x[0])            
            dataLines.append("{" + ", ".join(["%d %s" % p for p in sparseEntries]) + "} % " + featureItem.item.getComment())
        dummyName = "dummy"
        i = 1
        while dummyName in actualAttrNames:
            dummyName = "dummy_" + str(i)
            i += 1
        headers.append("@ATTRIBUTE %s STRING" % dummyName)
        className = "class"
        i = 1
        while className in actualAttrNames:
            className = "class_" + str(i)
            i += 1
        headers.append("@ATTRIBUTE %s {%s}" % (className, ",".join(classVals)))
        fp = codecs.open(outFile, "w", "utf8")
        fp.write("\n".join(headers))
        fp.write("\n")
        fp.write("\n".join(otherHeaders))
        fp.write("\n@DATA\n")
        fp.write("\n".join(dataLines))
        fp.close()            

    # Specializations

    def createFeatureFile(self, outFile, featureItems, encodedLabelsToPublicLabels = None):
        raise Exception("not implemented")

    def decodeModelApplicationResult(self, featureItems, outPath, encodedLabelsToPublicLabels):
        raise Exception("not implemented")

    def encodeFeaturesForClassifier(self, featureItems):
        raise Exception("not implemented")
    
    def executeModelBuild(self, dataPath, modelDir, outStream, metadata):
        raise Exception("not implemented")

    def executeModelApplication(self, inPath, outPath, model):
        raise Exception("not implemented")
    
# Most of these seem to be in Java, so let's have a separate parent for that.
    
class PerAttributeJavaFeatureExtractorClassifier(PerAttributeFeatureExtractorClassifier):

    args = PerAttributeFeatureExtractorClassifier.args + \
           [OpArgument("heap_size", hasArg = True,
                       help = "If present, specifies the -Xmx argument for the Java JVM"),
            OpArgument("stack_size", hasArg = True,
                       help = "If present, specifies the -Xss argument for the Java JVM")]

    def __init__(self, task, heap_size = None, stack_size = None, **kw):
        PerAttributeFeatureExtractorClassifier.__init__(self, task, **kw)
        self.heapSize = heap_size
        self.stackSize = stack_size

#
# And finally, the items that get created.
#
            
# All items will have to have an annot to hang things off of, if
# we want to generate an actual vector-level score.

class ClassifierItem:

    def __init__(self, tNode, resources = None):
        self.treeNode = tNode
        self.annot = tNode.annot
        self.resources = resources

    def getResource(self, name):
        if self.resources is not None:
            return self.resources.get(name)
        else:
            return None

    def getContext(self, **kw):
        return self.treeNode.getContext(**kw)

    def getComment(self):
        return "%s %s(%d-%d)" % (self.treeNode.root.fileName, self.annot.atype.lab, self.annot.start, self.annot.end)

    def getDescriptionForScore(self, truncateString = None):
        s = self.treeNode.getString()
        if (truncateString is not None) and (len(s) > truncateString):
            s = s[:(truncateString/2) - 2] + "..." + s[(truncateString/2) + 1:]
        bestLabel = self.annot.getEffectiveLabel()
        return "%s (%s, %d-%d)" % (s, bestLabel, self.annot.start, self.annot.end)

    # If you're trying to make a decision about a pair, this would be
    # different.

    # NOT SURE THIS IS STILL RIGHT. It may have to be a tree node.
    
    def getAnnotationSequenceForScoring(self):
        return [self.annot]

#
# And now, the feature extractors.
#
    
_REGISTERED_FEATURE_EXTRACTORS = {}

def registerFeatureExtractor(name, fn):
    global _REGISTERED_FEATURE_EXTRACTORS
    if not callable(fn):
        raise ModelBuilderError("can't register non-callable feature extractor '%s'" % name)
    _REGISTERED_FEATURE_EXTRACTORS[name] = fn

def getRegisteredFeatureExtractor(name):
    global _REGISTERED_FEATURE_EXTRACTORS
    return _REGISTERED_FEATURE_EXTRACTORS.get(name)

#
# Here are some feature extractors.
#

# Do I want to remove everything that isn't alphabetic?
# I don't think so. Let's replace it with _. But not here;
# do it later, when we pull together all the features.

# item is a ClassifierItem.

def _ngramsOfWords(item, ngramSize, context = None, level_scope = None,
                   token_context_size = None, **kw):
    if (context is None) and (level_scope is None) and (token_context_size is None):
        tokList = [t.getString()
                   for t in item.treeNode.root.getSequence("token")[item.treeNode.tokenStart:item.treeNode.tokenEnd]]
        return _ngramsOfElements(tokList, ngramSize, **kw)
    else:
        c = item.treeNode.getContext("token", context, parentType = level_scope,
                                     tokenContextSize = token_context_size)
        tokList = [a.getString() for a in c]
        return _ngramsOfElements(tokList, ngramSize, **kw)

def _bagOfWords(item, **kw):
    return _ngramsOfWords(item, 1, **kw)

registerFeatureExtractor("_bagOfWords", _bagOfWords)

def _caseNormalizedBagOfWords(item, normalize_case = False, **kw):
    return _bagOfWords(item, normalize_case = True, **kw)

registerFeatureExtractor("_caseNormalizedBagOfWords", _caseNormalizedBagOfWords)

def _weightedBagOfWords(item, weighted = False, **kw):
    return _bagOfWords(item, weighted = True, **kw)

registerFeatureExtractor("_weightedBagOfWords", _weightedBagOfWords)

def _caseNormalizedWeightedBagOfWords(item, normalize_case = False, weighted = False, **kw):
    return _bagOfWords(item, normalize_case = True, weighted = True, **kw)

registerFeatureExtractor("_caseNormalizedWeightedBagOfWords", _caseNormalizedWeightedBagOfWords)

def _ngramsOfElements(tokList, ngramSize, prefix = None, multiplier = None, normalize_case = False, weighted = False, **kw):
    sList = []        
    if len(tokList) >= ngramSize:
        if (normalize_case == "yes") or (normalize_case is True):
            tokList = [s.lower() for s in tokList]
        if multiplier is not None:
            multiplier = int(multiplier)
        if prefix is None:
            prefix = ""
        else:
            prefix = prefix + "_"
        bagCount = {}
        for i in range(0, len(tokList) - ngramSize + 1):
            tokSeq = "+".join(tokList[i:i+ngramSize])
            try:
                bagCount[tokSeq] += 1
            except KeyError:
                bagCount[tokSeq] = 1
        for t, count in bagCount.items():
            if multiplier is not None:
                count = count * multiplier
            if (count > 1) and weighted:
                sList.append((prefix + t, count))
            else:
                sList.append(prefix + t)
    return sList

def _ngrams(item, ngramSize = 2, **kw):
    return _ngramsOfWords(item, ngramSize, **kw)

def _caseNormalizedNgrams(item, normalize_case = False, **kw):
    return _ngrams(item, normalize_case = True, **kw)

def _weightedNgrams(item, weighted = False, **kw):
    return _ngrams(item, weighted = True, **kw)

def _caseNormalizedWeightedNgrams(item, normalize_case = False, weighted = False, **kw):
    return _ngrams(item, normalize_case = True, weighted = True, **kw)

registerFeatureExtractor("_bigrams", _ngrams)
registerFeatureExtractor("_caseNormalizedBigrams", _caseNormalizedNgrams)
registerFeatureExtractor("_weightedBigrams", _weightedNgrams)
registerFeatureExtractor("_caseNormalizedWeightedBigrams", _caseNormalizedWeightedNgrams)

def _trigrams(item, **kw):
    return _ngrams(item, ngramSize = 3, **kw)

def _caseNormalizedTrigrams(item, **kw):
    return _caseNormalizedNgrams(item, ngramSize = 3,  **kw)

def _weightedTrigrams(item, **kw):
    return _weightedNgrams(item, ngramSize = 3, **kw)

def _caseNormalizedWeightedTrigrams(item, **kw):
    return _caseNormalizedWeightedNgrams(item, ngramSize = 3, **kw)

registerFeatureExtractor("_trigrams", _trigrams)
registerFeatureExtractor("_caseNormalizedTrigrams", _caseNormalizedTrigrams)
registerFeatureExtractor("_weightedTrigrams", _weightedTrigrams)
registerFeatureExtractor("_caseNormalizedWeightedTrigrams", _caseNormalizedWeightedTrigrams)

def _bagOfLabels(item, prefix = None, context = None,
                 entity_level = None, label_filter =None,
                 level_scope = None, add_nearest = False,
                 add_nearest_distance = False):
    if (entity_level is None) or (level_scope is None):
        raise Exception("must define entity_level and level_scope for _bagOfLabels")
    c = item.treeNode.getContext(entity_level, context, parentType = level_scope)
    if label_filter:
        c = [node for node in c if node.annot.atype.lab in label_filter]
    labels = [node.annot.atype.lab for node in c]
    l = _ngramsOfElements(labels, 1, prefix = prefix)
    # See if they're different.
    specLabels = [node.annot.getEffectiveLabel() for node in c]
    setLabels = set(labels)
    if set(specLabels) - setLabels:
        l += _ngramsOfElements([lb for lb in specLabels if lb not in setLabels], 1, prefix = prefix)
    if ((add_nearest == "yes") or (add_nearest is True)) and c and context:
        if context in ("right", "left"):
            if context == "right":
                thisC = c[0]
            else:
                thisC = c[-1]
            l.append(prefix + "_nearest_" + thisC.annot.atype.lab)
            if thisC.annot.getEffectiveLabel() != thisC.annot.atype.lab:
                l.append(prefix + "_nearest_" + thisC.annot.getEffectiveLabel())
        if (add_nearest_distance == "yes") or (add_nearest_distance is True):            
            # Can't use tokens here, because the document may not be tokenized.
            signal = item.annot.doc.signal
            if context == "left":
                l.append((prefix + "_nearestdistance", len(re.findall("\s+", signal[c[-1].annot.end:item.annot.start]))))
            elif context == "right":
                l.append((prefix + "_nearestdistance", len(re.findall("\s+", signal[item.annot.end:c[0].annot.start]))))
    return l            

#
# MAT WRAPPERS
#

# Finally, the model builder and step wrappers for the classifier processor.

# Note that this model builder requires a setting in its build_settings which
# is not available on the command line. So it's not listed in the args.

class PerAttributeClassifierEngineModelBuilder(ModelBuilder.ModelBuilder):

    depClass = PerAttributeClassifier
    eClass = PerAttributeClassifierEngine

    def __init__(self, task, trainingStep, buildInfo, save_feature_vectors = False, save_arff = False, **kw):
        ModelBuilder.ModelBuilder.__init__(self, task, trainingStep, buildInfo, **kw)
        classifier_class = self.buildInfo.rawBuildSettings.get("classifier_class")
        if classifier_class is None:
            raise ModelBuilderError("no classifier_class specified in model config build settings for PerAttributeClassifierEngineModelBuilder in task '%s'" % task.name)
        # Find the class.
        cclass = FindPluginClass(classifier_class, task.name, parentClass = self.depClass, tpe = "per-attribute classifier class")
        if not issubclass(cclass, PerAttributeClassifier):
            raise ModelBuilderError("model config specifies per-attribute classifier class %s which is not a subclass of PerAttributeClassifier in task %s" % (classifier_class, task.name))
        self.cclass = cclass
        # Now that I have the cclass, I should be able to use it to extract options to pass to it.
        self.cclassSettings = cclass.enhanceAndExtract(XMLOpArgumentAggregator(kw))
        # And this isn't available on the command line, but is present in the raw settings.
        self.cclassSettings["feature_extractors"] = self.buildInfo.rawBuildSettings.get("feature_extractors")
        # We have to be very clear on what we're retrieving. We're looking to classify ATTRIBUTES.
        # So we retrieve the labels which bear the attributes. The third argument is
        # attributes which are present but their labels are not listed; if includeAttributeOnly
        # is true, this will be empty. The fourth is attribute exclusions.
        # I now realize that I do NOT want to includeAttributeOnly, because I need
        # to know that there are no labels in the set, only attributes.
        labels, ignore, attrInclusions, ignore = task.getAnnotationSetDetailsForSetNames(self.trainingAnnotationSets, attributeDetails = True)
        if labels:
            raise ModelBuilderError("attribute classifier training annotation set includes labels (%s) as well as attributes in task %s" % (", ".join(labels), task.name))
        # So to construct my list, I go through the attrInclusions.

        repo = task.getAnnotationTypeRepository()
        self.labelsAndAttrs = {}

        for label, includedAttrs in attrInclusions.items():
            labObj = repo.findAnnotationType(label, None, create = False)
            allAttrs = []
            if labObj is None:
                # I don't know how this would be possible, but lets just check.
                raise ModelBuilderError("label '%s' for model builder unknown" % label)
            for attr in labObj.attr_list:
                if attr.name in includedAttrs:
                    allAttrs.append(attr)
            self.labelsAndAttrs[label] = allAttrs

    def _run(self, modelOutputDir, fileList, docTmpDir, tmpDir, oStream):

        if os.path.exists(modelOutputDir):
            if not os.path.isdir(modelOutputDir):
                # It's a file. Remove it.
                os.remove(modelOutputDir)
                os.makedirs(modelOutputDir)
        else:
            os.makedirs(modelOutputDir)

        p = self.eClass(self.task, modelOutputDir, self.labelsAndAttrs, self.cclass, self.cclassSettings)
        self._trainToplevel(p, [(os.path.basename(path), self.reader.readFromSource(path)) for path in fileList], docTmpDir,
                            outStream = oStream)        
        # Write the relevant information as a JSON file.
        p.writeModelMetadata(modelOutputDir)
        return [(path, path) for path in fileList]

    def _trainToplevel(self, engine, docPairList, docTmpDir, **kw):
        engine.trainToplevel(docPairList, docTmpDir, **kw)

    # Here's how we use the classifier_class to extend the available options.
    @classmethod
    def addDependentOptions(cls, buildInfo, aggregator, **kw):
        classifier_class = buildInfo.rawBuildSettings.get("classifier_class")
        task = buildInfo.task
        if classifier_class is None:
            raise ModelBuilderError("no classifier_class specified in model config build settings for PerAttributeClassifierEngineModelBuilder in task '%s'" % task.name)
        # Find the class.
        cclass = FindPluginClass(classifier_class, task.name, parentClass = cls.depClass, tpe = "per-attribute classifier class")
        if not issubclass(cclass, cls.depClass):
            raise ModelBuilderError("model config specifies per-attribute classifier class %s which is not a subclass of %s in task %s" % (classifier_class, cls.depClass.__name__, task.name))
        cclass.addOptions(aggregator, **kw)

class PerAttributeClassifierEngineFeatureExtractionModelBuilder(PerAttributeClassifierEngineModelBuilder):

    depClass = PerAttributeFeatureExtractorClassifier

    args = PerAttributeClassifierEngineModelBuilder.args + \
           [OpArgument("save_feature_vectors",
                       help = "If present, the feature vector file which is the input to the classifier will be saved alongside the model."),
            OpArgument("save_arff",
                       help = "If present, an ARFF file suitable for loading into Weka will be saved alongside the model.")]

    def __init__(self, task, trainingStep, buildInfo, save_feature_vectors = False, save_arff = False, **kw):
        PerAttributeClassifierEngineModelBuilder.__init__(self, task, trainingStep, buildInfo, **kw)
        self.saveFeatureVectors = save_feature_vectors
        self.saveArff = save_arff

    def _trainToplevel(self, engine, docPairList, docTmpDir, **kw):
        PerAttributeClassifierEngineModelBuilder._trainToplevel(self, engine, docPairList, docTmpDir, saveFeatureVectors = self.saveFeatureVectors, saveArff = self.saveArff, **kw)

# This is kind of hideous: the model processor can take some command-line
# options, like heap and stack, but only if they're the right class, and
# there's no way to get those to appear only when the model is a particular
# type. I have to pass it to the call and to the constructor. Bah.

class PerAttributeClassifierTagStep(CmdlineTagStep):    

    def do(self, annotSet, **kw):
        self._doBatch([("<file>", annotSet)], **kw)
        return annotSet
    
    def doBatch(self, iDataPairs, **kw):
        self._doBatch([(os.path.basename(x[0]), x[1]) for x in iDataPairs], **kw)
        return iDataPairs
    
    def _doBatch(self, docPairList, model = None, language = None, **kw):

        tagger_model = model
        # Make sure we have a model.
        if tagger_model is None:
            defaultModel = self.wrapperStep.getDefaultModel()
            if defaultModel and os.path.exists(defaultModel):
                tagger_model = defaultModel
        if tagger_model is None:
            raise TaggerConfigurationError("no model specified")

        with ExecutionContext.Tmpdir() as tmpDir:
            # I'm making this a separate method so I can specialize it and check the class information.
            processor = self._loadProcessor(tagger_model, **kw)
            processor.applyModelToplevel(docPairList, tmpDir, **kw)

    def _loadProcessor(self, taggerModel, **kw):
        md = _readMetadata(taggerModel)
        mClass = FindPluginClass(md["mclass"], self.descriptor.name, parentClass = PerAttributeClassifierEngine, tpe = "per-attribute classifier engine")
        if not issubclass(mClass, PerAttributeClassifierEngine):
            raise AnnotationError("Stored model is not a PerAttributeClassifierEngine")
        md["cclass_settings"].update(kw)
        return mClass.fromMetadata(self.descriptor, taggerModel, md)

class PerAttributeJavaFeatureExtractorClassifierTagStep(PerAttributeClassifierTagStep):

    args = PerAttributeClassifierTagStep.args + \
           [OpArgument("heap_size", hasArg = True, help = "If present, specifies the -Xmx argument for the Java JVM"),
            OpArgument("stack_size", hasArg = True, help = "If present, specifies the -Xss argument for the Java JVM")] + \
           OptionTemplate([OpArgument("save_feature_vectors_for_scoring", help = "If present, input and output are saved as JSON in the file, for the scorer to find later")],
                          use_activator_value_prefix = True)

    def _loadProcessor(self, taggerModel, **kw):
        processor = PerAttributeClassifierTagStep._loadProcessor(self, taggerModel, **kw)
        if not issubclass(processor.cclass, PerAttributeJavaFeatureExtractorClassifier):
            raise AnnotationError("Stored classifier is not a PerAttributeJavaFeatureExtractorClassifier")
        return processor
        

