# Copyright (C) 2021 The MITRE Corporation. See the toplevel
# file LICENSE.txt for license terms.

from __future__ import absolute_import

# Here's where I generate a comparison document. This document will store
# all the comparison information in metadata. We couldn't do that with Callisto,
# but we CAN do that with the MAT UI. What we need is a set of pairs,
# with a pivot and comparison for each pair, a boolean indicating if the pair
# matches or not, and a list of matching dimensions for use if the pair doesn't
# match. We also need to know what dimensions were considered during the match,
# and we can save the similarity profile information to handle that.

from .Annotation import (AnnotationCore, AttributeValueSequence,
                         AttributeValueList, AnnotationError, AnnotationAttributeType)
from .Document import AnnotatedDoc

from . import Pair

class ComparisonError(Exception):
    pass

def generateComparisonDocument(task, pivot, others, similarityProfile = None,
                               pivotLabel = None, otherLabels = None,
                               forReconciliation = False):

    if pivot.metadata.get("comparison"):
        raise ComparisonError("pivot document is already a comparison document")
    for d in others:
        if d.metadata.get("comparison"):
            raise ComparisonError("one of the documents to be compared is already a comparison document")
        if d.signal != pivot.signal:
            raise ComparisonError("the signal of one of the documents to be compared doesn't match the pivot")
        
    # Set up the outbound document.
    outD = AnnotatedDoc(signal = pivot.signal, globalTypeRepository = task.getAnnotationTypeRepository())
    pairer = Pair.PairState(task = task, similarityProfile = similarityProfile, skipTokens = True)

    # It's entirely possible that there may be some duplication in the given labels.
    if pivotLabel or otherLabels:
        labelsTaken = set()
        if pivotLabel:
            labelsTaken.add(pivotLabel)
        if otherLabels:
            oldOtherLabels = otherLabels
            otherLabels = []
            for otherLabel in oldOtherLabels:
                if otherLabel in labelsTaken:
                    i = 1
                    while True:
                        cand = otherLabel + ("_(%d)" % i)
                        if cand not in labelsTaken:
                            otherLabels.append(cand)
                            labelsTaken.add(cand)
                            break
                        else:
                            i += 1
                else:
                    otherLabels.append(otherLabel)
                    labelsTaken.add(otherLabel)

    docMap = [(((otherLabels and (len(otherLabels) > i) and otherLabels[i]) or "d"+str(i+1)), other) \
              for (other, i) in zip(others, range(len(others)))]

    pairer.addDocumentTuples([(((pivotLabel or "d0"), pivot), p) for p in docMap])
    
    # Now, let's set up a cache for the compared annotations in each doc.
    copyCache = dict([(o, set()) for o in others])
    # We're going to import all the pivot annotations at once.
    pivotSet = set()
    copyCache[pivot] = pivotSet

    # So we've paired everything up, and we have an output document.
    for d in pairer.resultEntries:
        (pivotDname, pivotD), (compDname, compD) = d["tuple"]
        pairs = d["pairs"]
        # How do we create the new document? First, we want to import all
        # the compared annotations. But what about the case where
        # some of the arguments aren't compared? I think that we have to
        # ensure that if you're copied, everything you point to is copied.
        # And how do we produce the mapping between the originals and the copies?
        # Voila! importAnnotationGroup deals with all of this.
        # So the first thing we need to do is find all the annotations which
        # have been paired for each doc.
        compSet = copyCache[compD]
        for p in pairs:
            [label, ann, refMatchStatus, hLabel, hAnn, hypMatchStatus] = p.asList()
            if ann: pivotSet.add(ann)
            if hAnn: compSet.add(hAnn)

    # Now, let's see if the import works. Be sure to capture the mappings!
    # When we have annotations that point to annotations that aren't being
    # compared, we have two choices: either we can ignore the extra annotations,
    # or we can force them to be copied. It seems to me that if we're reconciling,
    # we probably want to remove them, or, at the very least, barf; if we're just
    # comparing, we can probably get away with pulling them in (although it'll be a
    # lot of extra ink on the page). Well, it turns out that if this is for
    # reconciliation, it'll complain because you've got annotations which have
    # no pairs in the reconciled document.

    def recursivelyAddAnnotationAttrValues(annot, s):
        # Assume that annot is already in s, and we
        # need to check its children.
        atype = annot.atype
        if atype.hasAnnotationValuedAttributes:
            for attrObj, attr in zip(annot.atype.attr_list, annot.attrs):
                if isinstance(attr, AnnotationCore):
                    if attr not in s:
                        if forReconciliation:
                            raise ComparisonError("found an annotation among the annotations to be reconciled which points to annotations which aren't covered in the similarity profile")
                        s.add(attr)
                        recursivelyAddAnnotationAttrValues(attr, s)
                elif isinstance(attr, AttributeValueSequence) and attr.ofDocAndAttribute and \
                     isinstance(attr.ofDocAndAttribute[1], AnnotationAttributeType):
                    for subval in attr:
                        if subval not in s:
                            if forReconciliation:
                                raise ComparisonError("found an annotation among the annotations to be reconciled which points to annotations which aren't covered in the similarity profile")
                            s.add(subval)
                            recursivelyAddAnnotationAttrValues(subval, s)
                            
    mappingCache = {}
    for d, s in copyCache.items():
        for annot in list(s):
            recursivelyAddAnnotationAttrValues(annot, s)
        mappingCache[d] = outD.importAnnotationGroup(list(s))

    # When I assemble the comparisons, I need to make sure that everything mentioned
    # in the comparisons has a public ID. So let's go through the results again,
    # and collect the outcomes.
    allPairings = []

    # local hash to keep track of the entries for each pivot as we assemble them
    pivotDict = {}
    # local variable to keep track of doc order
    docOrder = []

    pivotMapping = mappingCache[pivot]
    for d in pairer.resultEntries:
        (pivotDname, pivotD), (compDname, compD) = d["tuple"]
        compMapping = mappingCache[compD]
        if len(docOrder) == 0:
            docOrder.append(pivotDname)
        docOrder.append(compDname)
        pairs = d["pairs"]
        for p in pairs:
            [label, ann, refMatchStatus, hLabel, hAnn, hypMatchStatus] = p.asList()
            if ann and hAnn:
                # rescue the dimension similarity.
                ignore, dimSim, errToks = pairer.simEngine.similarityCache[(ann, hAnn)]
                dimMatch = [dim.dimName for (dim, status) in dimSim.items() if status[0] == 1.0]
            else:
                dimMatch = None
            # Any pairs will be in the same stratum, so we only need to check the
            # stratum for one side or the other. And remember, we can't use
            # label or hLabel, because the strata are in terms of TRUE labels.
            stratum = pairer.simEngine.labelToStratumIdx[(ann and ann.atype.lab) or hAnn.atype.lab]
            if ann:
                ann = pivotMapping[ann]
                pivotID = ann.getID()
                # find or create the entry
                if pivotID in pivotDict:
                    pivotEntry = pivotDict[pivotID]
                else:
                    pivotEntry = {"match": True, "stratum": stratum, "entries": {}}
                    pivotEntry["entries"][pivotDname] = {"annot": pivotID, "pivot": True}
                    pivotDict[pivotID] = pivotEntry
                    allPairings.append(pivotEntry)
            else:
                # spurious (no entry in pivotDict)
                pivotEntry = {"match": False, "stratum": stratum, "entries": {}}
                allPairings.append(pivotEntry)
            pivotEntry["match"] = pivotEntry["match"] and (refMatchStatus == "match")
            if hAnn:
                hAnn = compMapping[hAnn]
                pivotEntry["entries"][compDname] = {"annot": hAnn.getID(), "match": refMatchStatus == "match",
                                                    "dimensions_match": dimMatch}
            
    # Now, we record the pairings in the metadata, along with the dimensions considered.

    dimensionsConsidered = {}
    for (lLabel, rLabel), d in pairer.simEngine.methodMap.items():
        # d now contains "dimensions" and possibly "dead_weight".
        dimsConsidered = d.get("dimensions", [])
        attrsConsidered = []
        for dim in dimsConsidered:
            if isinstance(dim, Pair.LabelComparisonClass):
                attrsConsidered.append("_label")
            elif isinstance(dim, Pair.SpanComparisonClass):
                attrsConsidered.append("_span")
            elif isinstance(dim, Pair.NonAnnotationAttributeRemainderClass):
                pass # not sure what to do with this yet.
            elif isinstance(dim, Pair.AnnotationAttributeRemainderClass):
                attrsConsidered.append("_annotation_attribute_remainder")
            else:
                # It might have a dimension map. We're recording
                # only the right label here.
                if dim.dimMap:
                    attrsConsidered.append(dim.dimMap[rLabel])
                else:
                    attrsConsidered.append(dim.dimName)
                    
        # We're going to need to record this for labels in both directions, but
        # the symmetry is already present in the method map.
        try:
            lEntry = dimensionsConsidered[lLabel]
        except KeyError:
            lEntry = {}
            dimensionsConsidered[lLabel] = lEntry
        lEntry[rLabel] = attrsConsidered

    # SAM 7/24/13: The exported reconciliation documents will end up requiring
    # segments which declare that certain sets are reconciled. But we can't
    # infer that from the document; if there are no annotations for the set in the output,
    # then we won't find it. So basically, we need a union of the sets done for the
    # documents. Doesn't matter if they're gold or not - everything present
    # is reconciled.

    inputSetsDone = set(pivot.getSetsDone())
    for doc in others:
        inputSetsDone.update(doc.getSetsDone()) 
    
    outD.metadata["comparison"] = {"similarity_profile": similarityProfile or "<default>",
                                   "strata": pairer.simEngine.strata,
                                   "labelToStratumIndex": pairer.simEngine.labelToStratumIdx,
                                   "doc_labels": docOrder,
                                   "input_sets_done": list(inputSetsDone),
                                   "global_pairings": allPairings,
                                   "dimensions_considered": dimensionsConsidered}
        
    # And what about importing zones and tokens? Pick the first document that has some.
    tokTypes = task.getAnnotationTypesByCategory("token")
    if tokTypes:
        for d in [pivot] + others:
            toks = d.getAnnotations(tokTypes)
            if toks:
                outD.importAnnotationGroup(toks)
                break
    zoneTypes = task.getAnnotationTypesByCategory("zone")
    if zoneTypes:
        for d in [pivot] + others:
            zones = d.getAnnotations(zoneTypes)
            if zones:
                outD.importAnnotationGroup(zones)
                break

    if forReconciliation:
        comparisonToReconciliation(outD)
            
    return outD

# When we copy new elements that match, it had better be because
# all its "children" match. You could certainly do reconciliation based on
# a forgiving similarity profile, but there wouldn't be any point.
# And it better be that if we copy a match annotation, its
# "children" must have a match too.

# I need this both in converting a comparison doc to a reconciliation doc,
# and in updating the reconciliation doc.

class _ReconciliationCopier(object):

    def __init__(self, outD, pivotLabel, choiceLabel):
        self.copyCache = {}
        self.matchCopyCache = set()
        self.outD = outD
        self.pivotLabel = pivotLabel
        self.choiceLabel = choiceLabel
        self.pointerReplacementCache = []

    def _copyAnnot(self, annId):
        copyCache = self.copyCache
        outD = self.outD
        try:
            return copyCache[annId]
        except KeyError:
            newAnnot = outD.getAnnotationByID(annId).copy(copyID = False)
            if newAnnot.atype.hasAnnotationValuedAttributes:
                for attrObj, v in zip(newAnnot.atype.attr_list, newAnnot.attrs):
                    if attrObj._typename_ == "annotation":
                        if isinstance(v, AnnotationCore):
                            newAnnot[attrObj.name] = self._copyAnnot(v.getID())
                        elif isinstance(v, AttributeValueSequence) and (len(v) > 0):
                            newAnnot[attrObj.name] = v.__class__([self._copyAnnot(elt.getID()) for elt in v])
            outD._addAnnotation(newAnnot)
            copyCache[annId] = newAnnot
            return newAnnot

    def convertPair(self, pair):
        # add a copy of the matching annotation from the pivot doc to the choices doc
        # and mark the row as reconciled.
        pivotID = pair["entries"][self.pivotLabel]["annot"]
        newAnnot = self._copyAnnot(pivotID)
        self.matchCopyCache.add(pivotID)
        # Update the argument fillers, if any.
        # We need to shift the pivot from the pivotLabel to the choiceLabel to
        # get the right display in the spanless sidebar (we think), and we don't
        # use the pivot setting otherwise (we claim). We also don't look at the
        # dimension-level "match" values, so it doesn't matter that the pivot's
        # entry doesn't have one.
        pair["entries"][self.choiceLabel] = {"annot": newAnnot.getID(), "pivot": True}
        pair["reconciled"] = True        
        # When I create a new copy, I have to update the pointers.
        # The things that point to them have to now point to the
        # new one. Doesn't matter if it was previously copied or not.
        for docLabel, docEntry in pair["entries"].items():
            if docLabel != self.choiceLabel:
                rId = docEntry.get("annot")
                if rId:
                    rAnnot = self.outD.getAnnotationByID(rId)
                    self.pointerReplacementCache.append((rAnnot, newAnnot))

    # This is an element that already matches. Since we're precaching,
    # and we're not going to call finish(), I can overgenerate
    # in the copyCache, which is perfect, because if the pair is
    # reconciled, it doesn't matter which element was chosen;
    # for the purposes of precaching, any annotation in this pair
    # counts as being replaced by the element in the choice dimension.
    # In addition, if we encounter the choice annotation itself
    # already, we should know that it's a "replacement" for itself.
    
    def precache(self, pair):
        choiceID = pair["entries"][self.choiceLabel]["annot"]
        newAnnot = self.outD.getAnnotationByID(choiceID)
        self.copyCache[choiceID] = newAnnot
        for k, e in pair["entries"].items():
            if (k != self.choiceLabel) and e.get("annot"):
                self.copyCache[e["annot"]] = newAnnot
                self.matchCopyCache.add(e["annot"])

    def finish(self):
        if set(self.copyCache.keys()) - self.matchCopyCache:
            raise ComparisonError("some annotations were copied during match creation which don't correspond to matches")

    def updateCachedPointers(self):
        if self.pointerReplacementCache:
            self.updatePointers(self.outD, self.pointerReplacementCache)
            self.pointerReplacementCache = []

    @staticmethod
    def updatePointers(recDoc, replacementPairs):
        recDoc.atypeRepository._buildInverseIdDict()
        positionsToUpdate = []

        # Collect them all first, because changing annotation-valued
        # attributes clears out the _inverseIdDict.
        
        for rAnnot, choiceAnnot in replacementPairs:
            refs = recDoc.atypeRepository._inverseIdDict.get(rAnnot.id)
            if refs:
                for ref in refs:
                    positionsToUpdate.append((rAnnot, ref, choiceAnnot))
                    
        for orig, (parentAnnot, parentAName, aggrType), replacement in positionsToUpdate:
            # Don't forget about the possibility that the types clash,
            # in which case it has to be a remove.
            if aggrType is None:
                try:
                    parentAnnot[parentAName] = replacement
                except AnnotationError as e:
                    # We got an error. This must mean a type clash,
                    # so you must remove.
                    parentAnnot[parentAName] = None
            else:
                # Replace it in the attribute value sequence.
                v = parentAnnot[parentAName]
                if replacement is None:
                    v.remove(orig)
                elif isinstance(v, AttributeValueList):
                    idx = v.index(orig)
                    try:
                        v[idx] = replacement
                    except AnnotationError:
                        v[idx:idx+1] = []
                else:
                    v.remove(orig)
                    try:
                        v.add(replacement)
                    except AnnotationError:
                        pass

def comparisonToReconciliation(outD):
    if not outD.metadata.get("reconciliation_doc"):
        outD.metadata["reconciliation_doc"] = True
        comp = outD.metadata["comparison"]        
        pivotLabel = comp["doc_labels"][0]
        choiceLabel = "choices_label"
        # Add these to the beginning of these lists.
        comp["doc_labels"][0:0] = [choiceLabel]
        pairings = comp["global_pairings"]
        
        copier = _ReconciliationCopier(outD, pivotLabel, choiceLabel)
                    
        for pair in pairings:
            docEntry = pair["entries"].get(pivotLabel)
            # All pivots need to be removed.
            if docEntry and docEntry.get("pivot"):
                del docEntry["pivot"]
            if pair.get("match"):
                copier.convertPair(pair)

        copier.updateCachedPointers()

        copier.finish()

def reconciliationDocumentFullyReconciled(recDoc):
    if not recDoc.metadata.get("reconciliation_doc"):
        raise ComparisonError("input document is not a reconciliation document")
    # If it's a reconciliation document, everything better be reconciled.
    comp = recDoc.metadata["comparison"]
    pairings = comp["global_pairings"]
    choicesLabel = comp["doc_labels"][0]
    for entry in pairings:
        if not entry.get("reconciled"):
            return False
        # If we haven't cleared out all the votes via updateReconciliationDocument,
        # fail.
        if entry.get("voted"):
            return False
    return True

def reconciliationDocumentStarted(recDoc):
    if not recDoc.metadata.get("reconciliation_doc"):
        raise ComparisonError("input document is not a reconciliation document")
    # If it's a reconciliation document, at least one of the nonmatch pairs must be reconciled or voted.
    # The difference between a pair that starts out matched automatically and
    # a pair that needs to be reconciled is that the automatically matched pair
    # has match: true at the top.
    comp = recDoc.metadata["comparison"]
    pairings = comp["global_pairings"]
    choicesLabel = comp["doc_labels"][0]
    for entry in pairings:
        if (not entry.get("match")) and (entry.get("reconciled") or entry.get("voted")):
            return True
    return False    
        
def exportReconciliationDocument(task, recDoc):    
    if not recDoc.metadata.get("reconciliation_doc"):
        raise ComparisonError("input document is not a reconciliation document")
    # If it's a reconciliation document, everything better be reconciled.
    comp = recDoc.metadata["comparison"]
    pairings = comp["global_pairings"]
    choicesLabel = comp["doc_labels"][0]
    annotsMentioned = set()
    annotsToKeep = set()
    for entry in pairings:
        if entry.get("voted") or (not entry.get("reconciled")):
            raise ComparisonError("input document is not fully reconciled")
        for docLabel, v in entry["entries"].items():
            a = v.get("annot")
            if a is not None:
                annot = recDoc.getAnnotationByID(a)
                annotsMentioned.add(annot)
                if docLabel == choicesLabel:
                    annotsToKeep.add(annot)
    # Remove everything mentioned that we're not keeping.
    recDoc.removeAnnotationGroup(annotsMentioned - annotsToKeep)
    del recDoc.metadata["comparison"]
    del recDoc.metadata["reconciliation_doc"]
    # Now, we want to make sure the segments are marked correctly.
    # First, we need to infer the phases done, and then we need to
    # mark them all reconciled.
    recDoc.removeAnnotations(["SEGMENT"])
    if "input_sets_done" in comp:
        recDoc.setSetsDone(comp["input_sets_done"][:])
    else:
        if "phasesDone" in recDoc:
            del recDoc.metadata["phasesDone"]
        task._inferPhasesDone(recDoc)
    # Now that we have the inferred or recorded sets, we can create the zones.
    managedPresentSets = set(recDoc.getSetsDone()) & task.getManagedSets()
    for zoneAnnot in recDoc.getAnnotations(task.getAnnotationTypesByCategory("zone")):
        for managedSet in managedPresentSets:
            recDoc.createAnnotation(zoneAnnot.start, zoneAnnot.end, "SEGMENT", {"annotator": "GOLD_STANDARD", "status": "reconciled", "set": managedSet})
    return recDoc
                

# When we come back, voted will be set, reconciled will be true,
# match will be false, and there will be an entry for the choices label.

def updateReconciliationDocument(task, recDoc):

    # If a stratum is complete but not processed, process it.
    # We can tell a stratum is done but not processed if every
    # pair is reconciled, but some have "voted" on them. In this case,
    # we remove the "voted" elements, and we prepare the next
    # stratum or stratum slice. Don't forget to split the
    # stratum into spanned/spanless for the purposes of this
    # computation.

    # Once we identify the stratum to process, we update the pointers
    # for the elements which point to the resolved pairs, and
    # then redo the similarity for the next stratum.

    if not recDoc.metadata.get("reconciliation_doc"):
        raise ComparisonError("input document is not a reconciliation document")
    
    comp = recDoc.metadata["comparison"]
    simProfile = comp["similarity_profile"]
    if simProfile == "<default>":
        simProfile = None
    pairings = comp["global_pairings"]
    choicesLabel = comp["doc_labels"][0]
    otherDocs = comp["doc_labels"][1:]

    stratumInfo = {}
    someVotedSet = set()

    for pair in pairings:
        # Determine the stratum.
        thisStratum = pair["stratum"]
        try:
            stratumEntry = stratumInfo[thisStratum]
        except KeyError:
            stratumEntry = {"spanned": {"pairs": [], "allReconciled": True, "someVoted": False},
                            "spanless": {"pairs": [], "allReconciled": True, "someVoted": False}}
            stratumInfo[thisStratum] = stratumEntry
        record = None
        # We'll determine spanned/spanless by finding the first annot.
        for v in pair["entries"].values():
            a = v.get("annot")                          
            if a is not None:
                annot = recDoc.getAnnotationByID(a)
                if annot.atype.hasSpan:
                    record = stratumEntry["spanned"]
                else:
                    record = stratumEntry["spanless"]
                record["pairs"].append(pair)
                if not pair.get("reconciled"):
                    record["allReconciled"] = False
                if pair.get("voted"):
                    record["someVoted"] = True
                    someVotedSet.add((thisStratum, annot.atype.hasSpan))
                break

        # Check, just in case.
        if record is None:
            raise ComparisonError("found a pairing entry which is neither spanned nor spanless")
    
    # OK. If there's nothing which has been voted, nothing needs to be updated.
    # If there's more than one set, it's an error.

    if len(someVotedSet) == 0:
        return
    elif len(someVotedSet) != 1:
        raise ComparisonError("found multiple strata with votes")
    else:
        # Here's where the action is.
        thisStratum, hasSpan = someVotedSet.pop()
        # All the things which have been voted on need to update
        # the things that point to them.
        stratumEntry = stratumInfo[thisStratum]
        
        if hasSpan:
            record = stratumEntry["spanned"]
        else:
            record = stratumEntry["spanless"]

        if not record["allReconciled"]:
            # Not done yet.
            return

        # We're going to cache the positions to update, because when 
        # we remove annotations from a position, it blanks out the
        # inverseIdDict.
        
        positionsToUpdate = []
                
        for pair in record["pairs"]:
            if pair.get("voted"):
                # Any annotation that points to anything
                # in the non-choice labels should now
                # be replaced by the chosen one. If the choice is
                # none, "voted" will be "none".
                choiceDocEntry = pair["entries"].get(choicesLabel)
                if (choiceDocEntry is None) and (pair["voted"] != "none"):
                    raise ComparisonError("no chosen annot in a voted pair, but vote is not marked as none")
                
                del pair["voted"]
                # Do NOT make match True. 

                choiceAnnot = None
                if choiceDocEntry:
                    choiceAnnot = recDoc.getAnnotationByID(choiceDocEntry["annot"])
                
                for docLabel, docEntry in pair["entries"].items():
                    if docLabel != choicesLabel:
                        rId = docEntry.get("annot")
                        if rId:
                            rAnnot = recDoc.getAnnotationByID(rId)
                            positionsToUpdate.append((rAnnot, choiceAnnot))

        _ReconciliationCopier.updatePointers(recDoc, positionsToUpdate)

        # Update is done. Now, let's move on to the next stratum, and
        # redo its similarity. We should keep doing that for each stratum
        # that is fully reconciled.

        maxStratum = max(stratumInfo.keys())

        p = copier = None

        contentAnnotationSets = task.getNonAdminAnnotationSetNames()
        
        while True:
            if hasSpan:
                hasSpan = False
            else:
                thisStratum += 1
                while thisStratum not in stratumInfo:
                    if thisStratum > maxStratum:
                        # We're all out. Return from the function.
                        return
                    thisStratum += 1
                stratumEntry = stratumInfo[thisStratum]
                hasSpan = True
                
            if hasSpan:
                record = stratumEntry["spanned"]
            else:
                record = stratumEntry["spanless"]
                
            if len(record["pairs"]) > 0:

                # OK, we should check this stratum.

                # First, create the pairer and copier if we haven't yet,
                # and preload them.

                if p is None:
                    # Our goal is not to try to redo the pairing - our goal is to
                    # try to update the similarity. In particular, if it isn't
                    # a match already, see if it is now.

                    PairState = Pair.PairState

                    p = PairState(task, similarityProfile = simProfile)
                    copier = _ReconciliationCopier(recDoc, otherDocs[0], choicesLabel)

                    # We have to precache all the pivot -> choice pairs in
                    # preexisting matches in PREVIOUS STRATA. We ALSO have
                    # to record all the annotations in the choices_label document
                    # in previous strata as PAIRING WITH THEMSELVES in the pair state,
                    # so they'll be recognized as matches. Finally, we have to
                    # record the pairs between the annotations that were determined
                    # to match with each other, whether or not they're in an
                    # overall match.

                    def recordPreviousPair(pair):
                        # It's not just if they're reconciled - they have to be reconciled
                        # to an annotation rather than None. If they're reconciled to None,
                        # there won't be a choices_label.
                        if pair.get("reconciled") and pair["entries"].get("choices_label"):
                            copier.precache(pair)
                            recAnn = recDoc.getAnnotationByID(pair["entries"]["choices_label"]["annot"])
                            p.simEngine.similarityCache[(recAnn, recAnn)] = 1.0, None, None
                            p.simEngine.recordPair(recAnn, recAnn)
                        pivotID = pair["entries"].get(otherDocs[0], {}).get("annot")
                        if pivotID:
                            pivotAnnot = recDoc.getAnnotationByID(pivotID)
                            for lab, e in pair["entries"].items():
                                if (lab != "choices_label") and (lab != otherDocs[0]) and \
                                   e.get("annot") and e.get("match"):
                                    hAnnot = recDoc.getAnnotationByID(e["annot"])
                                    p.simEngine.similarityCache[(pivotAnnot, hAnnot)] = 1.0, None, None
                                    p.simEngine.recordPair(pivotAnnot, hAnnot)

                    for stratum, sEntry in stratumInfo.items():
                        if (stratum < thisStratum) or ((stratum == thisStratum) and (not hasSpan)):
                            # If this is a previous stratum, or the current stratum but
                            # we're currently working on the spanless side, do something.
                            for pair in sEntry["spanned"]["pairs"]:
                                recordPreviousPair(pair)

                            # Only do spanless if it's a previous stratum.
                            if stratum < thisStratum:
                                for pair in sEntry["spanless"]["pairs"]:
                                    recordPreviousPair(pair)                        

                pairsToCheck = []

                for pair in record["pairs"]:
                    if pair.get("match"):
                        continue
                    # If not every document is represented, it can't
                    # match.
                    allAnnots = [pair["entries"].get(docLabel, {}).get("annot")
                                 for docLabel in otherDocs]
                    if None not in allAnnots:
                        pairsToCheck.append(pair)

                
                for pair in pairsToCheck:

                    # OK. See how these annotations match. If there are
                    # more than two, check it pair by pair.
                    refAnnot = recDoc.getAnnotationByID(pair["entries"][otherDocs[0]]["annot"])
                    refLab = task.getEffectiveAnnotationLabel(refAnnot, 
                                                              restrictToAnnotationSetNames = contentAnnotationSets)

                    allMatch = True
                    for docLabel in otherDocs[1:]:
                        hID = pair["entries"][docLabel]["annot"]
                        hAnnot = recDoc.getAnnotationByID(hID)
                        hLab = task.getEffectiveAnnotationLabel(hAnnot, 
                                                                restrictToAnnotationSetNames = contentAnnotationSets)
                        r, dimSim, errToks = p.simEngine.computeSimilarity((refLab, refAnnot), (hLab, hAnnot))
                        dimMatch = [dim.dimName for (dim, status) in dimSim.items() if status[0] == 1.0]
                        # Does it match? Call computePairStatuses, just in case this
                        # needs to be future-proofed.
                        rStatus, hStatus = p._computePairStatuses(r, errToks)
                        entry = pair["entries"][docLabel]
                        if rStatus == "match":
                            entry["match"] = True
                        else:
                            entry["match"] = False
                            allMatch = False
                        entry["dimensions_match"] = dimMatch

                    if allMatch:
                        # It matches!
                        pair["match"] = True
                        copier.convertPair(pair)
                        
                copier.updateCachedPointers()

                if [pair for pair in record["pairs"] if not pair.get("reconciled")]:
                    # If they're not all reconciled now, stop.
                    break
                
        # I was checking finish() here, but I can't because I've
        # prepopulated it with all sorts of things I need when I'm
        # cleaning up the pairs.
