# Copyright (C) 2007 - 2021 The MITRE Corporation. See the toplevel
# file LICENSE.txt for license terms.

from __future__ import print_function, absolute_import

# We need to be able to deal globally with execution context:
# debug vs. nondebug, subprocess monitoring, tempfile location.
# There's a default context, which can be updated by various executables.

_TMPDIR = None
_SUBPROCESS_DEBUG = None
_PRESERVE_TEMPFILES = False
_DEBUG = False
_VERBOSE_CONFIG = False

import tempfile, os, shutil, codecs

# MAT imports

from .Config import MATConfig
from .PyCompat import Py3, _strtypes

# The logic is simple: if you specify a directory,
# then it doesn't get removed. Otherwise, it gets removed unless
# the global context says don't remove it.

# usage:
# with Tmpdir() as dir:
#   ...

class Tmpdir(object):

    def __init__(self, specifiedDir = None, specifiedParent = None, preserveTempfiles = None):
        global _PRESERVE_TEMPFILES
        self.specifiedDir = specifiedDir
        self.specifiedParent = None
        if preserveTempfiles is None:
            self.preserveTempfiles = _PRESERVE_TEMPFILES
        else:
            self.preserveTempfiles = preserveTempfiles
        if self.preserveTempfiles:
            # Leave a little gift. Where are we? Extract 2
            # stack frames (because __exit__ is one of them)
            # and grab the first (top) frame. The first two
            # elements are the file and line.
            import traceback
            self.location = traceback.extract_stack(limit = 2)[0][:2]

    def __enter__(self):
        global _TMPDIR
        if self.specifiedDir:
            self.dir = self.specifiedDir
        elif self.specifiedParent:
            self.dir = tempfile.mkdtemp(dir = self.specifiedParent)
        elif _TMPDIR:
            self.dir = tempfile.mkdtemp(dir = _TMPDIR)
        elif MATConfig.get("TMPDIR_ROOT"):
            self.dir = tempfile.mkdtemp(dir = MATConfig.get("TMPDIR_ROOT"))
        else:
            self.dir = tempfile.mkdtemp()
        return self.dir

    def __exit__(self, type, value, traceback):
        if (self.specifiedDir is None) and (not self.preserveTempfiles):
            shutil.rmtree(self.dir)
        elif self.preserveTempfiles:            
            fp = codecs.open(os.path.join(self.dir, ".source"), "w", "utf-8")
            fp.write("%s, line %d" % self.location)
            fp.close()

def setDebug(boolVal):
    global _DEBUG
    _DEBUG = boolVal

def setVerboseConfig(boolVal):
    global _VERBOSE_CONFIG, MATConfig
    _VERBOSE_CONFIG = boolVal
    MATConfig.setVerboseConfig(boolVal)

#
# You can use "with" to catch errors! This is exactly what I want for my
# debug handling. But you can't really scope the debugs - you can
# pass it a value, but you can't climb the stack and find out which
# debug you're in the context of. 
#

# But I don't want to be that clever right now. I've reviewed all the cases
# where debug is handled, and at the moment I don't see the value in
# generalizing.

def setTempfilePreservation(boolVal):
    global _PRESERVE_TEMPFILES
    _PRESERVE_TEMPFILES = boolVal

def setTmpdirRoot(path):
    global _TMPDIR
    import os
    tRoot = os.path.abspath(path)
    if not os.path.exists(tRoot):
        os.makedirs(tRoot)
    elif not os.path.isdir(tRoot):
        raise OSError("requested tempdir exists, but is not a directory")
    _TMPDIR = tRoot

# This can be called from the command line.

def setSubprocessDebug(val):
    global _SUBPROCESS_DEBUG
    try:
        _SUBPROCESS_DEBUG = int(val)
    except ValueError:
        pass

# Set a default from the configuration.

def getSubprocessDebug():
    global _SUBPROCESS_DEBUG
    if _SUBPROCESS_DEBUG is not None:
        return _SUBPROCESS_DEBUG
    elif MATConfig.get("SUBPROCESS_DEBUG") == "yes":
        return 10
    else:
        return 0

# So. On to the command line processing.

# No idea where this will ultimately live, but for now, we put it here.

_HERE = os.path.dirname(os.path.abspath(__file__))
def getMATVersion():
    vFile = os.path.join(_HERE, "MAT_VERSION")
    fp = codecs.open(vFile, "r", "utf-8")
    v = fp.read().strip()
    fp.close()
    return v

def _printVersionAndExit(option, opt, value, parser):    
    print("MAT version is %s." % getMATVersion())
    import sys
    sys.exit(1)

def addOptions(optionBearer):
    optionBearer.add_option("--version",
                            help = "Print version number and exit",
                            action = "callback",
                            callback = _printVersionAndExit)
    optionBearer.add_option("--debug", dest = "debug",
                            action = "store_true",
                            help = "Enable debug output.")
    optionBearer.add_option("--subprocess_debug", dest = "subprocess_debug",
                            type = "int",
                            metavar = "int",
                            help = "Set the subprocess debug level to the value provided, overriding the global setting. 0 disables, 1 shows some subprocess activity, 2 shows all subprocess activity.")
    optionBearer.add_option("--tmpdir_root", dest = "tmpdir_root",
                            metavar = "dir",
                            help = "Override the default system location for temporary files. If the directory doesn't exist, it will be created. Use this feature to control where temporary files are created, for added security, or in conjunction with --preserve_tempfiles, as a debugging aid.")
    optionBearer.add_option("--preserve_tempfiles", dest = "preserve_tempfiles",
                            action = "store_true",
                            help = "Preserve the temporary files created, as a debugging aid.")
    optionBearer.add_option("--verbose_config", dest = "verbose_config",
                            action = "store_true",
                            help = "If specified, print to stderr the source of each MAT configuration variable the first time it's accessed.")

def extractOptions(options):
    if options.preserve_tempfiles is not None:
        setTempfilePreservation(True)
    if options.tmpdir_root is not None:
        setTmpdirRoot(options.tmpdir_root)
    if options.subprocess_debug is not None:
        setSubprocessDebug(options.subprocess_debug)
    if options.debug is not None:
        setDebug(True)
    if options.verbose_config is not None:
        setVerboseConfig(True)

#
# Here's a new one. I want to have the option of capturing
# all subprocess output, for the purposes of inspecting it for
# the test suite. This has to work in multiple situations,
# including in the test suite when I'm calling a command line.
# So this needs to be followed up in the Command.py package, and
# in runCmdblock in the test suite (if that chooses not to use
# anything in Command.py).
#

# I'm going to use thread-local storage. The rest of MAT
# might not be thread-safe, but this will be.

import threading

_OUTPUT_CAPTURE_LOCAL = threading.local()

class OutputCapture(object):

    def __init__(self):
        self.buf = []

    # This should be a stack, just in case.
    
    def __enter__(self):
        global _OUTPUT_CAPTURE_LOCAL
        o = getattr(_OUTPUT_CAPTURE_LOCAL, "output_capture", None)
        if o is None:
            _OUTPUT_CAPTURE_LOCAL.output_capture = [self]
        else:
            o.append(self)
        return self

    def __exit__(self, tpe, value, traceback):
        global _OUTPUT_CAPTURE_LOCAL
        _OUTPUT_CAPTURE_LOCAL.output_capture[-1:] = []

    # I need to ensure that what's stored has
    # been converted to Unicode strings. It may be a list
    # of items.
    def addText(self, txt):
        if type(txt) in _strtypes:
            self.buf.append(txt)
        elif type(txt) in (list, tuple):
            for t in txt:
                self.addText(t)
        elif Py3 and (type(txt) is bytes):
            self.buf.append(txt.decode("utf-8"))

    def getText(self):
        return u"".join(self.buf)

# At one point, I tried to set it up so that this
# was a Python file-like object so it could be passed
# to subprocess.call(), but subprocess.call() really needs
# a real file object.

class OutputCaptureProxy(list):

    def __init__(self, seed):
        if seed is not None:
            list.__init__(self, seed)
        else:
            list.__init__(self)
    
    def write(self, o):
        for oc in self:
            oc.addText(o)

def getOutputCapture():
    global _OUTPUT_CAPTURE_LOCAL
    v = getattr(_OUTPUT_CAPTURE_LOCAL, "output_capture", None)
    if v is None:
        return None
    else:
        return OutputCaptureProxy(v)
