# Copyright (C) 2007 - 2021 The MITRE Corporation. See the toplevel
# file LICENSE.txt for license terms.

from __future__ import absolute_import
from .PyCompat import _strtypes, Py3, _unicodestring

# It turns out that a number of places in MAT have the same need,
# namely for an operation object which allows us to
# export argument lists to things like command line tools
# and CGI scripts. The operations should support
# argument validation, and provide support for
# obligatory, optional and repeatable arguments.

# The idea is that there will be a group of arguments,
# some of which might be shared by multiple operations.
# If multiple operations have different interpretations,
# then we signal an error, but otherwise, Conflicts are
# checked globally (across all arguments) at parse
# time. So we have to check to see if the argument is
# already present.

import optparse, types, re

class OperationError(Exception):
    pass

# We piggyback off the existing options.  We only
# add an option if the argument isn't already present;
# if it is, we make sure it's compatible with the
# argument that's already present, and record
# it in the global element. Then, later, we
# retrieve the relevant keys and values by filtering
# on the name of the step.

# First step: everybody needs to be using my OptionParser, Option,
# and OptionGroup.

#
# Option
# 

def check_boolean(option, opt, value):
    if value == "yes":
        return True
    elif value == "no":
        return False
    else:
        raise optparse.OptionValueError("option %s: value '%s' is neither 'yes' or 'no'" % (opt, value))

def check_multiple_choice(option, opt, value):
    if option.multiple_choice_special_values and (value in option.multiple_choice_special_values):
        return value
    choices = [s.strip() for s in value.split(",")]
    if choices == ['']:
        choices = []
    if not choices:
        raise optparse.OptionValueError("option %s: value '%s' lists no elements" % (opt, value))
    else:
        for c in choices:
            optparse.check_choice(option, opt, c)
        return choices

def help_is_template(help):
    return (help is not None) and (help.find("%(prefix)s") > -1)

class Option(optparse.Option):

    # Add the Boolean type. But it's only available with a long opt,
    # because of how we implement the workaround.
    
    TYPES = ("boolean", "multiple_choice") + optparse.Option.TYPES
    TYPE_CHECKER = optparse.Option.TYPE_CHECKER.copy()
    TYPE_CHECKER["boolean"] = check_boolean
    TYPE_CHECKER["multiple_choice"] = check_multiple_choice

    def _check_choice(self):
        if self.type == "multiple_choice":
            if self.choices is None:
                raise OptionError(
                    "must supply a list of choices for type 'choice'", self)
            elif type(self.choices) not in (tuple, list):
                raise OptionError(
                    "choices must be a list of strings ('%s' supplied)"
                    % str(type(self.choices)).split("'")[1], self)
        else:
            optparse.Option._check_choice(self)

    # In Python 3, there's no unbound method, so no __func__ property,
    # just the method itself.
    if Py3:
        CHECK_METHODS = [m for m in optparse.Option.CHECK_METHODS
                        if m is not optparse.Option._check_choice] + [_check_choice]
    else:
        CHECK_METHODS = [m for m in optparse.Option.CHECK_METHODS
                        if m is not optparse.Option._check_choice.__func__] + [_check_choice]

    ATTRS = optparse.Option.ATTRS + ["multiple_choice_special_values"]

    def __init__(self, *opts, **kw):
        self.useActivatorValuePrefix = False
        if kw.get("use_activator_value_prefix") is True:
            self.useActivatorValuePrefix = True
        try:
            del kw["use_activator_value_prefix"]
        except KeyError:
            pass
        optparse.Option.__init__(self, *opts, **kw)
        if self.type == "boolean":
            if self.default is optparse.NO_DEFAULT:
                self.default = False
            if self._short_opts:
                raise optparse.OptionError("boolean type not permitted with short opts: %s" % \
                                           ", ".join(self._short_opts), self)
        if (self.metavar is None) and (self.choices is not None):
            if self.type == "choice":
                self.metavar = " | ".join(self.choices)
            elif self.type == "multiple_choice":
                self.metavar = "val(,val,...) from " + " | ".join(self.choices)

        # Default values.
        # THE NAME IS NOT THE DEST.
        self._matName = self.get_opt_string()[2:]
        # This is a container for options which can't be registered
        # because they're duplicates.
        self.phantom_container = None

    def process(self, opt, value, values, parser):
        if self.type == "boolean":
            if isinstance(self.container, OptionGroup):
                p = self.container.parser
            else:
                p = self.container
            if isinstance(p, OptionParser):
                # Better add a value.
                # The default is OPPOSITE of the occurrence.
                if self.default is False:
                    value = "yes"
                else:
                    value = "no"                 
        r = optparse.Option.process(self, opt, value, values, parser)
        # We have to keep track of the options we've SEEN,
        # because later, when we extract in the aggregator, it's
        # gonna want to know.
        self._after_process(value, values, parser)
        return r

    def _after_process(self, value, values, parser):
        if not hasattr(values, "_optionsSeen"):
            values._optionsSeen = set([])
        values._optionsSeen.add(self)
        
    def _copyForPrefix(self, prefix):
        # Let's mangle the prefix.
        optName = "--" + prefix + self.get_opt_string()[2:]        
        # We don't preserve the destination. Just infer it.
        opts = dict([(a, getattr(self, a, None)) for a in self.ATTRS if getattr(self, a, None) is not None])
        try:
            del opts["dest"]
        except KeyError:
            pass
        if help_is_template(opts.get("help")):
            opts["help"] = opts["help"] % {"prefix": prefix}
        o = Option(optName, **opts)
        if prefix:
            o._optionPrefix = prefix
        return o

    # The hash option parsers all want a guaranteed value.
    
    def _copyForBoolean(self):
        opts = dict([(a, getattr(self, a, None)) for a in self.ATTRS if getattr(self, a, None) is not None])
        # The default is the opposite of the expected value.
        if self.action == "store_true":
            opts.update({"action": "store", "type": "boolean", "default": False})
        elif self.action == "store_false":
            opts.update({"action": "store", "type": "boolean", "default": True})
        return Option(self.get_opt_string(), **opts)

    def copy(self):
        opts = dict([(a, getattr(self, a, None)) for a in self.ATTRS if getattr(self, a, None) is not None])
        if self.useActivatorValuePrefix:
            opts["use_activator_value_prefix"] = True
        return Option(self.get_opt_string(), **opts)
          
    # When the parser is an option parser (see below), boolean
    # can't take an argument. See also process() above.

    def get_container(self):
        return (hasattr(self, "container") and self.container) or self.phantom_container

    def get_parser(self):
        c = self.get_container()
        if isinstance(c, OptionGroup):
            return c.parser
        else:
            return c
    
    def takes_value(self):
        if self.type == "boolean":
            p = self.get_parser()
            # It takes a value if it's not a commandline parser.
            return not isinstance(p, OptionParser)
        else:
            return optparse.Option.takes_value(self)

    # The option, at this point, will not have a container, I
    # don't think.
    
    def activate(self, parser, aggregator, activator = None, **kw):
        # I always need to copy this, it turns out, because of
        # how I handle help templates.
        if activator and (activator[4] or self.useActivatorValuePrefix):
            prefix = activator[4]
            if prefix and self.useActivatorValuePrefix:
                prefix += re.sub("\W+", "_", activator[3]) + "_"
            elif self.useActivatorValuePrefix:
                prefix = re.sub("\W+", "_", activator[3]) + "_"
            o = self._copyForPrefix(prefix)
        elif help_is_template(self.help):
            o = self._copyForPrefix("")
        else:
            o = self
        if activator is not None:
            try:
                aggregator.activationMap[o.dest].add(activator)
            except KeyError:
                aggregator.activationMap[o.dest] = set([activator])
        return o
            
    
#
# OptionGroup
#

# I want to defer to the parser for a number of things, for instance,
# checking the option.

class OptionGroup(optparse.OptionGroup):

    def __init__(self, *args, **kw):
        optparse.OptionGroup.__init__(self, *args, **kw)
        self.phantom_options = set()

    def add_phantom_option(self, op):
        op.phantom_container = self
        self.phantom_options.add(op)

    def format_description(self, formatter):
        # The issue here is that I want to ensure that any newlines I've inserted are preserved.
        desc = self.get_description()
        lines = desc.split("\n")
        return "".join([formatter.format_description(l) or '\n' for l in lines])

    def add_option(self, *args, **kwargs):

        # First step: force the option digestion. This duplicates
        # the work in add_option() in OptionGroup (and I'll have to
        # do the same thing in the parser), but it's cleaner than
        # dealing with the internal APIs.

        if type(args[0]) in _strtypes:
            option = self.option_class(*args, **kwargs)
        elif len(args) == 1 and not kwargs:
            option = args[0]
            if not isinstance(option, Option):
                raise TypeError("not an Option instance: %r" % option)
        else:
            raise TypeError("invalid arguments")

        return self.parser._add_option(option, group = self)

    def add_dynamic_choice_option(self, *args, **kw):
        kw["group"] = self
        return self.parser.add_dynamic_choice_option(*args, **kw)

#
# OptionTemplate
#

# The option template is the internal (or not so internal)
# representation of the options that the optionbearer holds. They
# don't really count as real options yet, not until they're added to the
# parser and activated; this means we can do all sorts of useful things
# about postponing, e.g., figuring out what the list of document IO
# options is.

class OptionTemplate(object):

    # Options can be a list, or, if not available, kw should be the
    # hash of types to lists of options. options can also be a function,
    # as can any of the values in kw. kw can also have heading values,
    # if the key is <k>_heading. Remember, they only need to be
    # sorted if you ask for the sort. So all the arguments should be
    # in the upper list.

    def __init__(self, options = None, heading = None, use_activator_value_prefix = False, **kw):
        # Each of the elements in options must be either an Option
        # or an DynamicChoiceOption.
        self.heading = heading
        self.otherTemplates = {}
        if options:
            self.options = self._digestOptions("options", options)
        else:
            self.options = []
        otherHashes = {}
        for k, w in kw.items():
            key = "options"
            if k.endswith("_heading"):
                k = k[:-8]
                key = "heading"
            try:
                otherHashes[k][key] = w
            except KeyError:
                otherHashes[k] = {key: w}
        for h, d in otherHashes.items():
            self.otherTemplates[h] = OptionTemplate(**d)
            # Accumulate all the options.
            self.options += self.otherTemplates[h].options
        if use_activator_value_prefix:
            for o in self.options:
                o.useActivatorValuePrefix = True

    def _digestOptions(self, v, listOrFn):
        if callable(listOrFn):
            listOrFn = listOrFn()
        if type(listOrFn) not in (list, tuple):
            raise TypeError("options are not a sequence")
        oList = []
        for o in listOrFn:
            if not (isinstance(o, Option) or isinstance(o, DynamicChoiceOption)):
                raise TypeError("neither an option nor an option activation: %s" % o)
            oList.append(o)
        return oList

    def activate(self, parser, aggregator, **kw):
        # Return a list of options.
        return [o.activate(parser, aggregator, **kw) for o in self.options]

    # This really needs to copy itself, and then add the options, and then return the copy.
    # But that causes another problem elsewhere - if this argument now appears in multiple
    # places, its default is ("NO", "DEFAULT"), because its original default was None.

    # This is now a bit trickier, now that we have digestable options. We need
    # to make sure that the right subtemplates get updated.
    
    def __add__(self, o):
        cpy = self.copy()
        if type(o) in (tuple, list):
            # Add the options.
            opts = o
            subtemplates = {}
        elif isinstance(o, OptionTemplate):
            opts = o.options
            subtemplates = o.otherTemplates
        else:
            raise ValueError("argument of option template addition is neither an option template nor a sequence of options")
        for o in opts:
            if not (isinstance(o, Option) or isinstance(o, DynamicChoiceOption)):
                raise TypeError("neither an option nor an option activation: %s" % o)
            cpy.options.append(o)
        for k, v in subtemplates.items():
            if k not in cpy.otherTemplates:
                cpy.otherTemplates[k] = v
            else:
                cpy.otherTemplates[k].options += v.options
        return cpy

    # How do we copy the other templates?
    
    def copy(self, heading = None):
        # Use a global list of options for the toplevel template and its subordinates,
        # to mimic what happens in __init__.
        oMap = dict([(o, o.copy()) for o in self.options])
        # Preserve the order when we copy.
        t = OptionTemplate(options = [oMap[o] for o in self.options], heading = heading or self.heading)
        for k, v in self.otherTemplates.items():
            t.otherTemplates[k] = OptionTemplate(options = [oMap[o] for o in v.options], heading = v.heading)
        return t

    # This is used exactly once up to this point, in
    # ToolChain.py.
    
    def addOptions(self, aggregator):
        aggregator.initializeActivationQueue()
        self._addOptions(aggregator)
        aggregator.processActivationQueue()

    def _addOptions(self, aggregator, heading = None, heading_suffix = None, source_class = None, **kw):
        if heading is None:
            heading = self.heading
            if heading and (heading_suffix is not None):
                heading = heading + heading_suffix
        aggregator._addOptions(self.activate(aggregator.parser, aggregator, **kw),
                               # This will only be used in the cmdline.
                               source_class = source_class,
                               name = (source_class and source_class.name) or None,
                               heading = heading, **kw)


# This is where lazy activation comes in. It's used for values which
# can be assembled from a tree of OptionBearers, and cause other
# options to be added.

class DynamicChoiceOption(object):

    def __init__(self, optionstr, cls, subtype = None, prefix = None, dest = None,
                 classFilters = None, specialValues = None,
                 recurse = True, multiple = False, storeLiteral = False,
                 use_activator_value_prefix = False, **kw):
        
        if dest is not None:
            # Because it must be computed by itself.
            raise TypeError("dest is not permitted in DynamicChoiceOption")        

        self.optionstr = optionstr
        self.activationClass = cls
        self.subtype = subtype
        self.classFilters = classFilters
        self.prefix = prefix
        self.recurse = recurse
        self.multiple = multiple
        self.specialValues = specialValues
        self.storeLiteral = storeLiteral
        self.useActivatorValuePrefix = use_activator_value_prefix
        self.kw = kw

    # The stored prefix is the TOPLEVEL prefix. It will not be present anywhere else.
    # So I use it, and only it, to pass down if it's present. I NEVER use it on
    # the optionstr. The prefix that's passed in, on the other hand, I use as the
    # local and recursive prefixes.

    # This method must parallel the Option.activate method. So everything
    # that method does, this one must do. So it has to do all the right things
    # wrt the activator value prefix: the local option string must
    # be generated, etc.

    def activate(self, parser, aggregator, activator = None, **kw):
        prefix = None
        if activator is not None:
            prefix = activator[4]
            
        # First, we prep the option name, because we're about to make an option.
        # This is the same thing that happens when we copy for a prefix.
        # Don't put your own prefix on yourself!

        if self.prefix is not None:
            localPrefix = None
            remotePrefix = self.prefix
        else:
            localPrefix = remotePrefix = prefix

        # The information we need to pass to the activation queue is the
        # activator information for this element.

        optionstr = self.optionstr
        dest = self.optionstr[2:]
        if activator and self.useActivatorValuePrefix:
            if localPrefix:
                localPrefix += re.sub("\W+", "_", activator[3]) + "_"
            else:
                localPrefix = re.sub("\W+", "_", activator[3]) + "_"
            if remotePrefix:
                remotePrefix += re.sub("\W+", "_", activator[3]) + "_"
            else:
                remotePrefix = re.sub("\W+", "_", activator[3]) + "_"
        if localPrefix:
            optionstr = "--" + localPrefix + optionstr[2:]
            
        # Patch up the help before it gets created.
        if help_is_template(self.kw.get("help")):
            selfKw = kw.copy()
            selfKw["help"] = selfKw["help"] % {"prefix": localPrefix or ""}
        else:
            selfKw = self.kw

        filters = set()
        if self.subtype:
            filters.add(self.subtype)
        if self.classFilters:
            filters.update(self.classFilters)
        choiceDict = self.activationClass.findDynamicOptionMapping(filters = filters)

        # This only works because choices doesn't check to see if the
        # default is among the choices.
        if selfKw.get("default"):
            selfKw["default"] = choiceDict[selfKw["default"]]
            
        # We want the optionstr, but we also want the original "dest".
        if self.recurse:
            aggregator.activationQueue.append((optionstr, self.activationClass, choiceDict, self.subtype, remotePrefix, dest))
        # We really want the class back.
        oType = "choice"
        if self.multiple:
            oType = "multiple_choice"
        # Use list() to freeze dictionary keys in Python 3.
        if self.specialValues:
            o = Option(optionstr, type = oType, choices = list(choiceDict.keys()),
                       action = "callback", callback = self._chooseValue,
                       use_activator_value_prefix = self.useActivatorValuePrefix,
                       multiple_choice_special_values = self.specialValues,
                       callback_args = (choiceDict, aggregator, optionstr, dest, remotePrefix), **selfKw)
        else:
            o = Option(optionstr, type = oType, choices = list(choiceDict.keys()),
                       use_activator_value_prefix = self.useActivatorValuePrefix,
                       action = "callback", callback = self._chooseValue,
                       callback_args = (choiceDict, aggregator, optionstr, dest, remotePrefix), **selfKw)
        # Mark it as a dynamic option so that _argValidate knows what to do.
        o.dynamicChoiceOption = self
        if localPrefix:
            o._optionPrefix = localPrefix
        if activator is not None:
            try:
                aggregator.activationMap[o.dest].add(activator)
            except KeyError:
                aggregator.activationMap[o.dest] = set([activator])
        return o

    # It's possible that value will be None, if this is coming from the HashOptionParser.
    # Worry about that case.
    
    def _chooseValue(self, option, opt, value, parser, choiceDict, aggregator, source, dest, prefix):
        if self.multiple:
            if value is None:
                recurseClasses = []
                value = []
            # This is an "escape".
            elif self.specialValues and (value in self.specialValues):
                recurseClasses = []
            else:
                recurseClasses = [(v, choiceDict[v]) for v in value]
                if not self.storeLiteral:
                    value = [p[1] for p in recurseClasses]
        elif value is None:
            recurseClasses = []
        else:
            recurseClasses = [(value, choiceDict[value])]
            if not self.storeLiteral:
                value = recurseClasses[0][1]
        setattr(parser.values, option.dest, value)
        # And, while we're at it, tell the aggregator that
        # this activation happened.
        if self.recurse:
            for v, oClass in recurseClasses:
                activator = (source, oClass, dest, v, prefix)
                aggregator.activatedSet.add(activator)
    

#
# OptionParser
#

class UnactivatedOptionError(optparse.BadOptionError):

    def __str__(self):
        return "unenabled option: %s" % self.opt_str

    def __unicode__(self):
        return _unicodestring(self.__str__())

class _OptionParserRoot(optparse.OptionParser):

    def __init__(self, option_class = Option, **kw):
        self.aggregator = None
        # Make sure that the default option class is my own.
        optparse.OptionParser.__init__(self, option_class = option_class, **kw)

    def add_option(self, *args, **kwargs):
        # First step: force the option digestion. This duplicates
        # the work in add_option() in OptionGroup (and I'll have to
        # do the same thing in the parser), but it's cleaner than
        # dealing with the internal APIs.

        if type(args[0]) in _strtypes:
            option = self.option_class(*args, **kwargs)
        elif len(args) == 1 and not kwargs:
            option = args[0]
            if not isinstance(option, Option):
                raise TypeError("not an Option instance: %r" % option)
        else:
            raise TypeError("invalid arguments")

        return self._add_option(option)

    # We're going to be forgiving - you can add an option twice, but
    # the second one will be ignored, unless your types don't match,
    # in which case you get an error.

    def _add_option(self, option, group = None):
        
        opt = self.get_option(option.get_opt_string())
        if opt is not None:
            localAction = option.action or "store"
            optAction = opt.action or "store"
            localType = option.type or "string"
            optType = opt.type or "string"
            # Note that in order to do this, we have to be careful, when we define the original
            # Option, not to pass the default key if default is None.
            if (localAction != optAction) or (localType != optType) or (option.default != opt.default):
                raise optparse.OptionError("attempt to define multiple options %s with different properties" % option.get_opt_string(), option)
            # Why add this? Because I need it during usage printing.
            # Otherwise, as long as one of them has been added, I
            # don't need another.
            if group is not None:
                group.add_phantom_option(option)
        elif group is not None:
            optparse.OptionGroup.add_option(group, option)
        else:
            optparse.OptionParser.add_option(self, option)

    # classFilters is a list of terms which, along with the subtype, will be
    # passed as filter information to findDynamicOptionMapping, which can be specialized.
    
    def add_dynamic_choice_option(self, optionstr, aggregator, activationCls, subtype = None, activatedPrefix = None,
                                  group = None, classFilters = None, **kw):
        # This is one of the entry points for lazy activation of a tree of options.
        # We make an option activation, empty the queue, activate the option, and
        # process the queue.
        aggregator.initializeActivationQueue()
        o = DynamicChoiceOption(optionstr, activationCls, subtype = subtype, prefix = activatedPrefix, classFilters = classFilters, **kw)
        (group or self).add_option(o.activate(self, aggregator))
        aggregator.processActivationQueue()

    def parse_args(self, *args, **kw):
        values, args = optparse.OptionParser.parse_args(self, *args, **kw)
        # Check to see which options were seen, and whether they
        # were activated.
        if hasattr(values, "_optionsSeen") and self.aggregator:
            for opt in values._optionsSeen:
                needsActivation, activators = self.aggregator.findActivation(opt.dest)
                if needsActivation and (not activators):
                    raise UnactivatedOptionError(opt.get_opt_string())
        return values, args

    def print_help(self, *args, **kw):
        # We have to reconfigure, temporarily, the options and groups
        # here. The only things we should be printing are
        # the things that are activated.
        # So --help will print out only the things which require
        # no activation, plus additional recommendations about what to
        # do for more help. --foo bar --help will print out
        # the things which are enabled by --foo bar.
        # And we have to be sure that we watch to make sure there's
        # an aggregator, because if there isn't, there's no way
        # to do anything besides the normal behavior.
        if not self.aggregator:
            optparse.OptionParser.print_help(self, *args, **kw)
            return
        _cache = {}
        furtherHelp = set()
        allOK = set()
        self._cacheOptionsForHelp(self, _cache, furtherHelp, allOK)
        okGroups = []
        _cachedGroups = self.option_groups
        # So how do we decide if a group has been activated? One of
        # its options needs to have been activated. This INCLUDES
        # phantom elements - otherwise, groups which we accidently
        # empty because their options were described elsewhere
        # wouldn't be listed (and I should make these normal options
        # for the time being).
        for grp in self.option_groups:
            if self._cacheOptionsForHelp(grp, _cache, furtherHelp, allOK):
                okGroups.append(grp)
        self.option_groups = okGroups
        oldEpilog = self.epilog
        trueFurther = (allOK & furtherHelp) - set([activator[0] for activator in self.aggregator.activatedSet])
        if trueFurther:
            self.epilog = "Further help may be available for %s.\nSpecify a value for these options to activate further help." % \
                          ", ".join(trueFurther)
            if oldEpilog:
                self.epilog = oldEpilog + "\n" + self.epilog
        optparse.OptionParser.print_help(self, *args, **kw)
        self.epilog = oldEpilog
        for k, origOptions in _cache.items():
            k.option_list = origOptions
        self.option_groups = _cachedGroups

    # If the source class is in the automatic activations,
    # don't bother checking the option list.
    
    def _cacheOptionsForHelp(self, o, _cache, furtherHelp, allOK):
        okOpts = []
        optList = o.option_list
        if isinstance(o, OptionGroup):
            optList += list(o.phantom_options)
        activate = False
        if o in self.aggregator.automaticActivations:
            okOpts = optList
            # In case there are no options
            activate = True
            allOK.update([opt.get_opt_string() for opt in optList])
        else:
            sourceClass = self.aggregator.sourceClasses.get(o, None)
            for opt in optList:
                # The activators are the ACTUAL things that appear in the
                # command line that cause the group to be shown. So if multiple
                # classes introduced it, the one corresponding to what's on
                # the command line will be present. ONLY that activator
                # counts.
                needsActivation, activators = self.aggregator.findActivation(opt.dest)
                activated = not needsActivation
                if (not activated) and activators:
                    for activator in activators:
                        if activator[1] == sourceClass:
                            activated = True
                            break
                if activated:
                    okOpts.append(opt)
                    allOK.add(opt.get_opt_string())
                else:
                    # Who gets further help? The elements which ARE 
                    # presented which have rejected children.
                    for activator in self.aggregator.activationMap[opt.dest]:
                        furtherHelp.add(activator[0])
        _cache[o] = o.option_list
        o.option_list = okOpts
        return activate or okOpts

# This is the COMMAND-LINE option parser. In order to
# work, it needs to do something hideous and sneaky with the
# boolean type, because nargs is going to be 1. There's a brief
# opportunity in _match_long_opt to add a "yes" or "no" to the
# command line.

# Actually, let's not do that at all. Let's modify option
# instead.

class OptionParser(_OptionParserRoot):

    pass

# Now, we inherit from Option above. We follow the
# following rules. The name we pass to the optparse.Option
# object is "--" + the name. 

class OpArgument(Option):

    def __init__(self, name, hasArg = False,
                 # Now, the attrs for Option.
                 action = None, default = None, type = None, 
                 metavar = None, **attrs):
        
        optparseName = "--" + name
        if action in ["store_const", "count", "append_const"]:
            # We'll follow this up with a way of handling "boolean"
            # as a type. The type will be changed when it gets
            # installed in a parser.
            raise OperationError("none of store_const, append_const, count are permitted")
        
        if hasArg is False:
            if action is not None:
                raise OperationError("can't specify action with hasArg = False")
            if type is not None:
                raise OperationError("can't specify type with hasArg = False")
            action = "store"
            type = "boolean"
            default = False
        elif metavar is None:
            metavar = name.upper()

        # We have to be careful about default. Passing default = None is NOT
        # the same as not passing a value for default. See __add__ in OptionTemplate
        # for the problem.

        if action is not None:
            attrs["action"] = action
        if type is not None:
            attrs["type"] = type
        if metavar is not None:
            attrs["metavar"] = metavar
        if default is not None:
            attrs["default"] = default

        Option.__init__(self, optparseName, **attrs)

        self._matName = name

# Now, option parsers. Neither of these default to the command line.
# For each of these, we stash the args as we're passed them, and return
# [] from get_args, because this will be surgically altered to
# be the final list of nonoption arguments, and neither of these
# have any.

# For both of these, we want to anticipate the possibility that
# we'll be expanding the arguments. So if there's no option yet,
# we should set it aside and return to it later.

# This class is never instantiated; only its children are.

class HashOptionParser(_OptionParserRoot):

    pType = "hash"
    
    def error(self, msg):
        raise OperationError(msg)

    def _get_keys(self):
        raise Exception("not implemented")

    def _get_value(self, option):
        raise Exception("not implemented")

    def parse_args(self, *args, **kw):
        try:
            return _OptionParserRoot.parse_args(self, *args, **kw)
        except (UnactivatedOptionError, OperationError) as e:
            strE = str(e)
            # We have to remove the -- before the options.
            strE = re.sub("--(\S)", r"\1", strE)
            raise OperationError("while digesting arguments from %s: %s" % (self.pType, strE))

    def _add_option(self, option, group = None):        
        # Convert any store_true or store_false options
        # into booleans, right here.
        if option.action in ["store_true", "store_false"]:
            option = option._copyForBoolean()
        return _OptionParserRoot._add_option(self, option, group = group)

    def _process_args(self, largs, rargs, values):

        # For each key, see if the argument is present.
        # If it isn't, return to it later; we may have
        # enhanced the list of arguments. If the loop
        # ever doesn't decrease in length, then we're done.
        keys = self._get_keys()
        lenKeys = len(keys)
        while True:
            if lenKeys == 0:
                return
            nextKeys = []
            for k in keys:
                try:
                    option = self._long_opt["--" + k]
                except KeyError:
                    # Kick it down the road.
                    nextKeys.append(k)
                    continue
                if not option.takes_value():
                    self.error("option '%s' must take a value" % k)
                if option.nargs != 1:
                    self.error("option '%s' requires other than 1 argument" % k)
                v = self._get_value(option)
                self._process_option(option, "--" + k, v, values, self)
            if len(nextKeys) == lenKeys:
                # Actually, at this point, we should just break
                # out of the loop; we want it to ignore the
                # things it doesn't know about, and it's
                # had the chance to do all the progressive
                # enhancement it wants to.
                # raise optparse.BadOptionError(nextKeys[0])
                break
            keys = nextKeys
            lenKeys = len(keys)

    def _process_option(self, option, opt, value, values, parser):
        option.process(opt, value, values, parser)

class DictOptionParser(HashOptionParser):

    pType = "dictionary"

    def __init__(self, *args, **kw):
        HashOptionParser.__init__(self, *args, **kw)
        self.dictInput = None

    def _get_args(self, args):
        self.dictInput = args
        return []

    def _get_keys(self):
        # Use list() to freeze dictionary keys in Python 3.
        return list(self.dictInput.keys())

    def _get_value(self, option):
        return self.dictInput[option._matName]

    def _process_option(self, option, opt, value, values, parser):
        # We don't want to convert the value; we go directly to
        # take_action, and then _after_process.
        r = option.take_action(option.action, option.dest, opt, value, values, parser)
        option._after_process(value, values, parser)
        return r        
        
class XMLOptionParser(HashOptionParser):

    pType = "xml"

    def __init__(self, *args, **kw):
        HashOptionParser.__init__(self, *args, **kw)
        self.xmlInput = None            

    def _get_args(self, args):
        self.xmlInput = args
        return []

    def _get_keys(self):
        # Use list() to freeze dictionary keys in Python 3.
        return list(self.xmlInput.keys())

    def _get_value(self, option):
        return self.xmlInput[option._matName]

class CGIOptionParser(HashOptionParser):

    pType = "CGI"

    def __init__(self, *args, **kw):
        HashOptionParser.__init__(self, *args, **kw)
        self.cgiInput = None

    def _get_args(self, args):
        self.cgiInput = args
        return []

    def _get_keys(self):
        # Use list() to freeze dictionary keys in Python 3.
        return list(self.cgiInput.keys())

    def _get_value(self, option):
        if option.action == "append":
            return self.cgiInput.getlist(option._matName)
        else:
            return self.cgiInput.getfirst(option._matName)

# We want this to be a state machine, basically. But we have to
# tackle the issue of how to manage it when it's a commandline.

# So basically, this guy needs to keep track of its state.
# It's initialized possibly with values, possibly with a parser.

class OpArgumentAggregator(object):

    parserClass = None

    def __init__(self, inputs, parser = None, values = None):
        self.inputs = inputs
        # Values is an option object, if provided.
        self.values = values
        self.parser = parser
        # This is the queue of activations.
        self.activationQueue = []
        # This is the record of what's already been
        # activated, so you don't do it more than once.
        self.activationRecord = {}
        # This is the mapping from destinations to activations.
        self.activationMap = {}
        self.activatedSet = set()
        if self.parser:
            self.parser.aggregator = self

    def _getParser(self):
        if not self.parser:
            self.parser = self.parserClass()
            self.parser.aggregator = self
        return self.parser

    def _getValues(self, defaults = None):
        if not self.values:
            values = None
            parser = self._getParser()
            if defaults is not None:
                # We need to ensure that the side effect callbacks are called.
                # In order to do that, we call a DictOptionParser on the defaults.
                # And we have to rescue the arguments which are added, and
                # add them to the current parser. Grrr.                
                p = DictOptionParser()
                p.aggregator = self
                for o in parser._get_all_options():
                    p.add_option(o)
                values, ignore = p.parse_args(defaults)
                # Now, REUSE THE VALUES. We'll end up with all the values
                # in the right place.
                for o in set(p._get_all_options()) - set(parser._get_all_options()):
                    parser.add_option(o)
            self.values = parser.parse_args(self.inputs, values = values)[0]
        return self.values

    # Instead of being initialized with an option object
    # (which isn't possible sometimes), we initialize with a parser,
    # store the option later, and disable addOptions,
    # because it should have already happened. The only
    # use of this so far in the system is in MATWorkspaceEngine,
    # where the operations are parsed and later augmented with
    # operation settings.
    
    def storeValues(self, values):
        self.values = values

    # This can be called numerous times, and it should be a no-op
    # if the values have already been created.

    # If there's a prefix, I need to duplicate the argument, but
    # keep the SAME DEST.

    # If we're doing progressive enhancement, and we're adding an
    # option, we'd better add the default to the values.

    # Nixing the progressive enhancements. Instead, addOptions knows
    # how to deal with a queue of things to activate. So this is the
    # main entry point. No one should be calling this except at
    # the toplevel.

    # SAM 4/2/13: It USED to be the case that this was a callable
    # public function. But now, this is called via the addOptions
    # method on the OptionBearer. So I've changed it to _addOptions
    # to indicate this.
    
    def _addOptions(self, cmdArgs, group = None, values = None, **kw):
        if cmdArgs and (self.values is None):
            for arg in cmdArgs:
                if group:
                    group.add_option(arg)
                else:
                    self._getParser().add_option(arg)
                if values is not None:
                    # We KNOW we don't have it yet. Adding it is cheap.
                    setattr(values, arg.dest, arg.default)

    def initializeActivationQueue(self):
        self.activationQueue = []
        self.activationRecord = set()
        
    def processActivationQueue(self):
        while self.activationQueue:            
            source, cls, cDict, subtype, prefix, dest = self.activationQueue[0]
            self.activationQueue[0:1] = []
            recordKey = (cls, subtype, prefix)
            if recordKey in self.activationRecord:
                continue
            self.activationRecord.add(recordKey)
            for vname, oClass in cDict.items():
                activator = (source, oClass, dest, vname, prefix)
                oClass._addOptions(self, activator = activator, 
                                   heading_suffix = (source and " (via %s %s)" % (source, vname)) or None,
                                   subtype = subtype)

    def findActivation(self, dest):        
        needsActivation = dest in self.activationMap
        if needsActivation:
            return True, (self.activationMap[dest] & self.activatedSet)
        else:
            return False, None

    def extract(self, reportDefaults = False, defaults = None):
        if self.values is None:
            # Use it in getting the values. But also we need to be able to pass through
            # any defaults which weren't digested. I'm pretty sure reusing defaults
            # in both places will get me this. The engine default tests pass, and
            # that's a good sign.
            return self.convertToKW(self._getValues(defaults = defaults), reportDefaults = reportDefaults, defaults = defaults)
        else:
            # Use it in extracting the kw.
            return self.convertToKW(self._getValues(), reportDefaults = reportDefaults, defaults = defaults)

    # Make sure that all the defaults are appropriately reported.
    # Don't want to return [] in a situation where it might be
    # altered.

    # We can pass defaults to convertToKW, and they won't get
    # salted into the values, unlike extract() above. reportDefaults
    # means to report the defaults from the OPTIONS.

    # SAM 4/3/13: the default params have to interact appropriately
    # with the prefixes. So the order should be
    # values -> defaults -> (option defaults if reportDefaults is True) -> None.
    
    def convertToKW(self, values, reportDefaults = False, defaults = None):
        # If reportDefaults is False, then make sure the option
        # has been optionsSeen if the value is not None. If reportDefaults
        # is True, then make sure the proper defaults are provided.
        
        optArgs = {}
                
        allKeys = dict([(arg.dest, arg) for arg in self.parser._get_all_options()
                        if arg.dest is not None])
        optsSeen = None
        if hasattr(values, "_optionsSeen"):
            optsSeen = values._optionsSeen

        # The optArgs respect the prefixes on the activators on the options.
        for arg, obj in allKeys.items():
            
            # So in this case, we're digesting the options object.
            # And the options have the DEFAULTS FROM ALL THE UNACTIVATED
            # PARAMETERS. Remember, the new DynamicChoiceOption stuff
            # adds the parameters, but marks them as unactivated until
            # the parameter that caused them to be added is filled.
            # So if we're reporting defaults, we may have a key/value pair
            # here which has no activators. So I need to check
            # if there actually ARE activators, before I proceed.
            # If there are no activators, we don't add the value.
            # This should include any default values that are passed in.
            needsActivation, activators = self.findActivation(arg)
            if needsActivation and (not activators):
                continue

            v = getattr(values, arg, None)
            # We actually have to make sure that when we retrieve the
            # value, we make sure that we rule out optparse.NO_DEFAULT.
            if (v is None) or (v is optparse.NO_DEFAULT):
                if obj.type == "boolean":
                    v = True
                elif obj.action == "append":
                    v = []
                else:
                    v = None
            # v is now the value we got from the options
            # themselves, including the option defaults.

            valueToRecord = None
            recordValue = True

            if optsSeen and (obj in optsSeen):
                # Use the value. It won't be None.
                valueToRecord = v
            elif defaults and (defaults.get(arg) is not None):
                valueToRecord = defaults[arg]
            elif reportDefaults:
                valueToRecord = v
            else:
                recordValue = False
            
            if recordValue:
                container = optArgs
                if needsActivation:
                    # We know there's an activator, since we checked above.
                    prefix = getattr(obj, "_optionPrefix", None)
                    if prefix is not None:
                        containerName = prefix
                        arg = arg[len(prefix):]
                        try:
                            container = optArgs[containerName]
                        except KeyError:
                            container = {}
                            optArgs[containerName] = container
                container[arg] = valueToRecord

        return optArgs

    # This method used to be on the OptionBearer, but in the new
    # regime, it has to be here. It's also used ONLY in enhanceAndValidate,
    # which itself is only used once. DO NOT USE THIS FURTHER.
    # I'm exploiting the fact that it only adds arguments from a single
    # class to be able to ignore the fact that the classes no longer
    # know what arguments they yielded.
    
    # We used to check obligatorihood, but no one was using it
    # so I killed it. 
    
    def validate(self, cls, **argDict):
        
        # argDict is a dictionary, where the
        # values are either strings, lists of strings, or
        # booleans.

        argsByKw = {}
        for o in self.parser._get_all_options():
            argsByKw[o._matName] = o

        for arg, val in argDict.items():
            try:
                argObj = argsByKw[arg]
            except KeyError:
                # The argument may not exist, in which case we skip it.
                # We used to barf, but now we're accumulating lots of
                # arguments, possibly for multiple commandline engine steps.
                continue
            
            # This may raise an operation error.
            self._argValidate(argObj, val)

        return argDict
    
    #
    # This is special for the use of validate() in the option bearer below.
    # 

    def _argValidate(self, option, val):
        if val is None:
            if option.action == "append":
                raise OperationError("option %s default must be []" % option._matName)
            else:
                return
        # Prepare for testing all vals.
        if option.action != "append":
            valList = [val]
        else:
            valList = val
        for val in valList:
            err = None
            if option.type == "boolean":
                if val not in (True, False):
                    err = "True or False"
            elif option.type == "int":
                if type(val) is not int:
                    err = "an integer"
            elif option.type == "float":
                if type(val) is not float:
                    err = "a float"
            elif option.type == "string":
                if type(val) not in _strtypes:
                    err = "a string"
            elif option.type == "complex":
                if type(val) is not complex:
                    err = "a complex number"
            elif option.type == "long":
                # There's no "long" type in optparse in Python 3.
                if Py3:
                    err = "a long integer (type no longer exists)"
                elif type(val) is not long:
                    err = "a long integer"
            elif option.type == "choice":
                if hasattr(option, "dynamicChoiceOption") and (not option.dynamicChoiceOption.storeLiteral):
                    # See DynamicChoiceOption above.
                    if not issubclass(val, option.dynamicChoiceOption.activationClass):
                        err = "a class corresponding to one of the strings " + ", ".join(["'"+a+"'" for a in option.choices])
                elif val not in option.choices:
                    err = "one of the strings " + ", ".join(["'"+a+"'" for a in option.choices])
            if err:
                if option.action == "append":                    
                    raise OperationError("each value of option %s must be %s" % (option._matName, err))
                else:
                    raise OperationError("the value of option %s must be %s" % (option._matName, err))

# This one is the exception, because it interacts with the parser
# options, and the addOptions method needs to set things up for
# parsing command lines, which requires more information than usual.

class CmdlineOpArgumentAggregator(OpArgumentAggregator):

    def __init__(self, parser):
        OpArgumentAggregator.__init__(self, None, parser = parser)
        self.sourceClasses = {}
        self.automaticActivations = set()
        
    # Right now, we're using this for the workspaces,
    # where groups are created for each folder action in some cases.
    # And sometimes they don't have options.
    
    def forceActivation(self, group):
        self.automaticActivations.add(group)

    def _addOptions(self, cmdArgs, name = None, heading = None, group = None, description = None, source_class = None, **kw):
        if cmdArgs and (self.values is None):
            if group is None:
                # I'm not using the name in the PluginStep case, because I'm updating
                # the title of the group later.
                if heading is None:
                    heading = "Arguments for the '%s' operation" % (name or "",)
                group = OptionGroup(self.parser, heading)
                self.sourceClasses[group] = source_class
                if description is not None:
                    group.set_description(description)
                # Add it FIRST.
                self.parser.add_option_group(group)
            OpArgumentAggregator._addOptions(self, cmdArgs, group = group, **kw)

# The CGI form is what gets stored when this is created.

class CGIOpArgumentAggregator(OpArgumentAggregator):

    parserClass = CGIOptionParser

class XMLOpArgumentAggregator(OpArgumentAggregator):

    parserClass = XMLOptionParser

class DictOpArgumentAggregator(OpArgumentAggregator):

    parserClass = DictOptionParser

# What about sharing arguments on the command line?
# The problem arises only when multiple operations
# need the same value, but the interpretation of that
# value is different. How do we deal with that?
# Let's add operation name prefixes.

# One more thing: I think I'm going to set things up
# so that you can instantiate separate instances of operations.
# That is, all the argument information will be CLASS information
# now, and addOptions will be a class method.
# That way, we can have the class operation still
# able to add options to the command line, if desired,
# and the individual operations can hold state. If desired.

class OptionBearer(object):

    @classmethod
    def _getInstantiableName(cls):
        # Ugh. This is so horrid. We can find local declarations
        # by looking at __dict__ of the class. This does NOT
        # include inheritance.
        if ("instantiable" not in cls.__dict__) or (cls.__dict__["instantiable"] is True):
            if "ov_name" in cls.__dict__:
                return cls.__dict__["ov_name"]
        return None

    @classmethod
    def findDynamicOptionMapping(cls, d = None, filters = None):
        if d is None: d = {}
        n = cls._getInstantiableName()
        if n is not None:
            d[n] = cls
        for subCls in cls.__subclasses__():
            subCls.findDynamicOptionMapping(d = d)
        return d        

    name = None

    # args = [] or OptionTemplate([])

    # We're going to force everyone who
    # uses Operations to use the new optparse
    # module. It also needs to be called
    # recursively on all the operations.

    # And although it's the aggregator that accumulates things,
    # this is the entry point for the recursive activation.    

    @classmethod
    def addOptions(cls, aggregator, subtype = None, heading = None, **kw):
        aggregator.initializeActivationQueue()
        cls._addOptions(aggregator, subtype = subtype, heading = heading, **kw)
        aggregator.processActivationQueue()

    @classmethod
    def getOptionTemplate(cls, subtype = None):
        if hasattr(cls, "args"):
            args = getattr(cls, "args")
            if args:
                if type(args) in (list, tuple):
                    cls.args = OptionTemplate(args)
                    args = cls.args
                elif callable(args):
                    # It's a class which needs to be instantiated now,
                    # rather than at the time the class that bears
                    # the args was defined.
                    cls.args = args()
                    args = cls.args
                if isinstance(args, OptionTemplate):
                    if subtype is not None:
                        return args.otherTemplates.get(subtype)
                    else:
                        return args
        return None


    # This is the element that all calls besides the toplevel should make.
    @classmethod
    def _addOptions(cls, aggregator, subtype = None, heading = None, heading_suffix = None, **kw):
        args = cls.getOptionTemplate(subtype = subtype)
        if isinstance(args, OptionTemplate):
            args._addOptions(aggregator, heading = heading, heading_suffix = heading_suffix,
                             source_class = cls, **kw)

    def __init__(self):
        pass

    @classmethod
    def enhanceAndExtract(cls, aggregator, reportDefaults = False, **params):
        cls.addOptions(aggregator)
        return aggregator.extract(reportDefaults = reportDefaults, defaults = params or None)

class Operation(OptionBearer):

    # By the time we get to "do", the arguments
    # should already have been vetted. The **args
    # should be the postprocessed dictionary that
    # is the output of validate.

    def do(self, *args, **kw):
        pass
