# Copyright (C) 2007 - 2021 The MITRE Corporation. See the toplevel
# file LICENSE.txt for license terms.

from __future__ import print_function, absolute_import
from .PyCompat import _inttypes, _numtypes, _strtypes

# Here's where we manage plugins. The plugins contain tasks.

# The plugins are in a sister directory to this module.

import os, sys, glob, shutil, re, subprocess, codecs

# MAT imports

from .Error import (PluginError, TaggerConfigurationError)
from .PluginDocInstaller import PluginDocInstaller
from .Operation import (Operation, OpArgument, OptionTemplate, XMLOpArgumentAggregator)
from . import Annotation
from . import Document
from . import Error
from .XMLNode import (XMLNode, XMLNodeDescFromFile, XMLNodeFromFile)

from . import ExecutionContext
from .ModelBuilder import ModelBuilder
from .Config import MATConfig
from .TaskUtilities import TaskAnnotationTypeSorter

# NonAsciiError 

class NonAsciiError(Exception):
    pass

# Originally, I had a distinction between applications
# and document formats. In the long run, because I need
# inheritance across the boundary between them, and because
# I need to distribute different format configurations separately,
# it makes sense to combine them again.

# The inheritance, as a result, isn't actually Python
# inheritance except when there's an accompanying class to
# the XML task specification. 

#
# Utilities
#

from xml.dom import Node

TASK_DESC = XMLNodeDescFromFile(os.path.join(os.path.dirname(os.path.abspath(__file__)), "task_template.xml"))

#
# Task definition
#

def LoadPluginTask(xmlPath):
    try:
        xmlNode = XMLNodeFromFile(xmlPath, TASK_DESC)
    except Exception as e:
        if True or ExecutionContext._DEBUG:
            raise
        else:
            raise PluginError("encountered error parsing XML task file %s; error was: %s" % (xmlPath, e))
    if xmlNode is None:
        if ExecutionContext._DEBUG:
            raise
        else:
            raise PluginError("could not find tasks in XML task file %s; error was: %s" % xmlPath)
    if xmlNode.label == "tasks":
        tasks = xmlNode.children["task"]
    else:
        tasks = [xmlNode]
    return tasks

class ModelInfo(object):

    def __init__(self, engine, step, clsName, configName, buildSettings):
        self.engine = engine
        # Might be None.
        self.step = step
        task = engine.task
        self.task = task
        self.modelClass = FindPluginClass(clsName, task.name, parentClass = ModelBuilder, tpe = "mode build class")
        if not issubclass(self.modelClass, ModelBuilder):
            raise PluginError("task %s specifies model build class %s which is not a subclass of ModelBuilder" % (task.name, clsName))
        self.configName = configName
        # I need these in case we need to pass the build settings to a child class, as in AttributeClassifier,
        # and to try to find various attributes in the configuration which aren't overrideable
        # on the command line.
        self.rawBuildSettings = buildSettings
        self.modelBuildSettings = self.modelClass.enhanceAndExtract(XMLOpArgumentAggregator(buildSettings))
        
    # Returns an absolute model path or None

    def getDefaultModel(self, step = None, language = None):
        return self.engine.getDefaultModel(step = step, language = language)

    def getModelClass(self):
        return self.modelClass

    def getModelBuildSettings(self):
        return self.modelBuildSettings

    def buildModelBuilder(self, **kw):
        # It's a bit perverse to have the self.step and annotationSets be None, but
        # it turns out that there are a number
        # of places where we create a ModelInfo object
        # outside the task engines and steps. So.
        # Actually, I'm trying to get rid of all of these.
        if self.step is None:
            annotationStep = self.task.getTrainableStep()
        else:
            annotationStep = self.step  
        settings = self.modelBuildSettings.copy()
        # So I want to override the defaults with
        # the results from **kw.
        settings.update(kw)
        return self.getModelClass()(self.task, annotationStep, self, **settings)

    # These two methods allow the model builder to deal with cases where the
    # model builder dispatches additional arguments based on some of its
    # settings. This can be done before the class is instantiated.

    def addOptions(self, aggregator, **kw):
        mClass = self.getModelClass()
        mClass.addOptions(aggregator, **kw)
        mClass.addDependentOptions(self, aggregator, **kw)

    # This method is identical to the OptionBearer method, except it
    # calls the local addOptions() instead of the class one.
    
    def enhanceAndExtract(self, aggregator, **kw):
        self.addOptions(aggregator)
        return aggregator.extract(**kw)

class PluginTaskDescriptor(object):

    def __init__(self, name, dir):
        self.name = name
        self.taskRoot = dir
        self.webDir = None
        if dir:
            self.webDir = os.path.basename(dir)
        self.settings = {}
        self.visible = True
        if dir:
            self.resourceDir = os.path.join(dir, "resources")
        # Documentation enhancement
        self.docEnhancementClass = None
        self.docEnhancement = None
        # Workflows.
        self.localWorkflowCache = {"inherit": None,
                                   "inherit_all": False,
                                   "workflows": {}}
        self._workflowDescCache = None
        self._workflowCache = None
        # Web customization.
        self.displayConfig = None
        self.inheritJS = True
        self.inheritCSS = True
        self.alphabetizeUILabels = True
        self.localLanguageCache = {"languages": [], "inherit_all": False, "inherit": None}
        self._languageCache = None
        self._languageAndCodeCache = None
        self.localJSFiles = []
        self.localCSSFiles = []
        self.shortWebName = None
        self.longWebName = None
        self.localEngineCache = {"java_subprocess_parameters": {},
                                 "inherit_all": False,
                                 "inherit": None,
                                 "enhance": None,
                                 "engines": []}
        self._engineDescCache = None
        self._engineCache = None
        # Java engine settings.
        self._javaEngineSettingsCache = None
        # Steps.
        self.localStepCache = {"steps": {},
                               "inherit_all": False,
                               "inherit": None}
        self._stepDescCache = None
        self._stepCache = None
        # Parentage.
        self.parentName = None
        self.parentObj = None
        self.children = []
        # Tags.
        # I need to keep track of the tag order
        # in case that's the presentation order
        # in the UI someone wants.
        self._tagHierarchyCache = None
        # New set descriptors.
        self.localAnnotationSetDescriptors = {"all_annotations_known": False,
                                              "inherit": [],
                                              "descriptors":
                                              [{"category": "admin",
                                                "name": "admin",
                                                "annotations":
                                                [{"label": "SEGMENT",
                                                  "all_attributes_known": True,
                                                  "span": True}],
                                                "attributes":
                                                [{"name": "annotator",
                                                  "type": "string",
                                                  "of_annotations": ["SEGMENT"]},
                                                 {"name": "set",
                                                  "type": "string",
                                                  "of_annotations": ["SEGMENT"]},
                                                 {"name": "status",
                                                  "type": "string",
                                                  "of_annotations": ["SEGMENT"],
                                                  "choices": ["non-gold", "human gold", "reconciled", "ignore during reconciliation"]}
                                                 ]}]}
        self._cachedAnnotationSetDescriptors = None
        self.localAnnotationDisplays = {"labels": {}, "groups": {}, "order": ["SEGMENT"],
                                        "attributes": {}, "gestures": []}
        self._cachedAnnotationDisplays = None
        self._annotationTypeRepository = None
        self._annotationCategoriesToSets = {}
        self._annotationSetsToCategories = {}
        self._managedSets = set()
        # Workspaces
        self.localWorkspaceCache = {"workspaces": [],
                                    "inherit": None,
                                    "enhance": None, 
                                    "inherit_all": False,
                                    "default_config": None}
        self._workspaceDescCache = None
        self._workspaceDefaultConfig = None
        # Similarity profiles.
        self.similarityProfiles = None
        self.scoreProfiles = None

    def addChild(self, child):
        child.parentObj = self
        self.children.append(child)

    def _digestXMLSettings(self, xmlNode):
        if xmlNode is None:
            return {}
        d = xmlNode.wildcardAttrs.copy()
        for setting in xmlNode.children["setting"]:
            if setting.children["name"].text in d:
                raise PluginError("explicit <setting> for %s clashes with parent attribute" % setting.children["name"].text)
            d[setting.children["name"].text] = setting.children["value"].text
        return d

    def _digestAttrListValue(self, node, v):
        return (node.attrs[v] and [s.strip() for s in node.attrs[v].split(",")]) or None

    def _digestDocEnhancementChildren(self, xmlNode):
        tree = []
        for entry in xmlNode.orderedChildren:
            if entry.label == "doc_entry":
                tree.append((entry.attrs["url"], entry.attrs["text"]))
            elif entry.label == "doc_section":
                tree.append((entry.attrs["url"], entry.attrs["text"],
                             self._digestDocEnhancementChildren(entry)))
        return tree

    # lazyClassLoad is set up for the cases where people want
    # to use tasks which have MAT.JavaCarafe steps defined, but
    # they're using the pip package.
    
    def fromXML(self, taskDesc, lazyClassLoad = False):

        attrs = taskDesc.attrs
        children = taskDesc.children
        delayedPackages = None
        if lazyClassLoad:
            delayedPackages = ["MAT"]
        
        if taskDesc.attrs["visible"] == "no":
            self.visible = False

        if children["doc_enhancement_class"]:
            clsName = taskDesc.children["doc_enhancement_class"].text
            cls = FindPluginClass(clsName, self.name, parentClass = PluginDocInstaller,
                                  tpe = "doc enhancement class", delayedPackages = delayedPackages)
            self.docEnhancementClass = cls

        if children["doc_enhancement"]:
            deSpec = {}
            deNode = children["doc_enhancement"]
            if deNode.children["app_header"]:                    
                appHeaders = []
                for deApp in deNode.children["app_header"]:
                    appHeaders.append({"id": deApp.attrs["id"], "text": deApp.attrs["text"],
                                       "brand_url": deApp.attrs["brand_url"],
                                       "tree": self._digestDocEnhancementChildren(deApp)})
                if appHeaders:
                    deSpec["app_headers"] = appHeaders
            if deNode.children["customization_header"]:
                customizationHeaders = []
                for deCust in deNode.children["customization_header"]:
                    customizationHeaders.append({"id": deCust.attrs["id"], "text": deCust.attrs["text"],
                                                 "tree": self._digestDocEnhancementChildren(deCust)})
                if customizationHeaders:
                    deSpec["customization_headers"] = customizationHeaders
            if deNode.children["external_id_header"]:
                otherHeaders = []
                for deOther in deNode.children["external_id_header"]:
                    otherHeaders.append({"id": deOther.attrs["id"],
                                         "tree": self._digestDocEnhancementChildren(deOhter)})
                if otherHeaders:
                    deSpec["other_headers"] = otherHeaders
            if deSpec:
                self.docEnhancement = deSpec

        if self.docEnhancement and self.docEnhancementClass:
            raise PluginError("for task %s, only one of <doc_enhancement> and <doc_enhancement_class> are permitted" % self.name)

        self.parentName = attrs["parent"]

        # Languages.

        if children["languages"]:
            lgBlock = children["languages"]
            self.localLanguageCache["inherit_all"] = (lgBlock.attrs["inherit_all"] == "yes")
            self.localLanguageCache["inherit"] = self._digestAttrListValue(lgBlock, "inherit")
            for lg in lgBlock.children["language"]:
                self.localLanguageCache["languages"].append({"name": lg.attrs["name"],
                                                             "code": lg.attrs["code"],
                                                             "text_right_to_left": (lg.attrs["text_right_to_left"] == "yes"),
                                                             "tokenless_autotag_delimiters": lg.attrs["tokenless_autotag_delimiters"],
                                                             "tokenless_autotag_respects_delimiters": (lg.attrs["tokenless_autotag_respects_delimiters"] != "no")})

        # Do the web customizations.
        if children["web_customization"] is not None:
            cust = children["web_customization"]
            self.displayConfig = cust.attrs["display_config"]
            self.localJSFiles = [a.text for a in cust.children["js"]]
            self.localCSSFiles = [a.text for a in cust.children["css"]]
            if cust.attrs["inherit_js"] == "no":
                self.inheritJS = False
            if cust.attrs["inherit_css"] == "no":
                self.inheritCSS = False
            if cust.attrs["alphabetize_labels"] == "no":
                self.alphabetizeUILabels = False                
            if cust.children["short_name"]:
                self.shortWebName = cust.children["short_name"].text
            if cust.children["long_name"]:
                self.longWebName = cust.children["long_name"].text
            if ((self.shortWebName is not None) or (self.longWebName is not None)) and \
               not ((self.shortWebName is not None) and (self.longWebName is not None)):
                raise PluginError("for task %s, short_name and long_name must be both defined, or neither" % self.name)
        
        # Do the settings.

        self.settings = self._digestXMLSettings(children["settings"])

        # Do the annotation sets. 

        # Let's just get us back to where we are. Don't even bother trying to take them
        # apart yet.

        # I need this in the similarity profiles, too.

        unpermittedAttrSpecs = {
            "string": ["minval", "maxval", "label_restrictions"],
            "int": ["label_restrictions"],
            "float": ["choices", "effective_labels", "label_restrictions"],
            "boolean": ["choices", "effective_labels",
                        "minval", "maxval", "label_restrictions", "default_is_text_span"],
            "annotation": ["choices", "effective_labels",
                           "minval", "maxval", "default", "default_is_text_span"]}

        # I was hoping that I wouldn't have to rewrite the annotation_set_descriptors
        # processing, but if order depends on the order of definition, then I should
        # definitely process it second. So let's decompose the construction into
        # some useful functions.
        
        def _coerceValue(attr, kName, v, tList, desc):
            if v is None:
                return None
            for t in tList:
                try:
                    return t(v)
                except ValueError:
                    continue
            raise PluginError("for task %s, %s '%s' of attribute %s is %s" %
                                (self.name, kName, v, attr["name"], desc))

        def _boolCoerce(s):
            if s == "yes":
                return True
            elif s == "no":
                return False
            else:
                raise ValueError

        def _modifyASDs(asds, all_annotations_known = None, inherit = None, **kw):
            if all_annotations_known is not None:
                asds["all_annotations_known"] = (all_annotations_known == "yes") or asds["all_annotations_known"]
            if inherit is not None:
                asds["inherit"] = list(set(asds["inherit"]) | set([s.strip() for s in inherit.split(",")]))

        # So if there's no set name, the set name is "content". If the chosen set already
        # exists, ensure that the category that's present is right, and return it. Otherwise,
        # create a new one.
        def _getOrAddASD(newDescriptors, name = None, category = None, managed = None, **kw):
            if name is None: name = "content"
            for d in newDescriptors:
                if d["name"] == name:
                    if (category is not None) and (d["category"] != category):
                        # It's the same descriptor, but the category is different.
                        raise PluginError("for task %s, the descriptor set name '%s' is used twice with different categories" % (self.name, name))
                    return d
            # Now that it isn't an existing category, add a default.
            if category is None: category = "content"
            # All nonadmin categories are managed by default.
            newDescriptor = {"category": category, 
                             "name": name,
                             "managed": (category not in ('zone', 'admin', 'token')) and (managed != "no"),
                             "annotations": [],
                             "attributes": []}
            newDescriptors.append(newDescriptor)
            return newDescriptor

        def _addLabelToASD(asd, label = None, all_attributes_known = None, span = None, processable = None, **kw):
            # Don't check for duplicates right now.
            asd["annotations"].append({"label": label,
                                       "all_attributes_known": all_attributes_known == "yes",
                                       "span": span != "no",
                                       "processable": processable != "no"})

        def _addAttributeToASD(asd, name = None, of_annotation = None, aggregation = None,
                               optional = None, default = None, default_is_text_span = None,
                               choices = None, effective_labels = None, minval = None, maxval = None,
                               label_restrictions = None, **kw):
            # Don't check for duplicates right now.
            attr = {"name": name,
                    "of_annotations": (of_annotation and [s.strip() for s in of_annotation.split(",")]) or [],
                    "type": kw["type"] or "string",
                    "aggregation": aggregation,
                    "optional": optional == "yes",
                    "default": default,
                    "default_is_text_span": default_is_text_span == "yes",
                    "choices": choices or [],
                    "effective_labels": effective_labels or None,
                    "minval": minval or None,
                    "maxval": maxval or None,
                    "label_restrictions": label_restrictions or None}
            asd["attributes"].append(attr)
            return attr

        # And here are the functions we need for managing the displays.
        # We're gonna deal with the order here too.

        def _recordLabelDisplay(name = None, css = None, presented_name = None,
                                edit_immediately = None, accelerator = None,
                                rendering_style = None,
                                overlap_rank = None, **kw):
            convertedADisplays = self.localAnnotationDisplays["labels"]
            if overlap_rank is not None:
                try:
                    overlap_rank = int(overlap_rank)
                except ValueError:
                    raise PluginError("overlap_rank value '%s' in task %s is not an integer" % (overlap_rank, self.name))
            if rendering_style is not None:
                if rendering_style not in ("normal", "background_span"):
                    raise PluginError("unknown rendering_style value '%s' in task %s" % rendering_style)
                
            convertedADisplays[name] = {"name": name,
                                        "css": css,
                                        "rendering_style": rendering_style,
                                        "presented_name": presented_name,
                                        "gestures": [],
                                        "edit_immediately": edit_immediately == "yes",
                                        "overlap_rank": overlap_rank,
                                        "accelerator": accelerator}
            # The group names can also be label names. So remove those things
            # which are groups and also labels. But it's only the label and label_group
            # elements which form this; not the attribute info.
            self.localAnnotationDisplays["order"].append(name)

        # This must be called after _recordLabelDisplay.
        def _recordGestureForLabel(name, label = None, function = None, writeable_only = None,
                                   attribute_writeable_only = None, **kw):
            attributeWriteableOnly = []
            if attribute_writeable_only:
                attributeWriteableOnly = [s.strip() for s in attribute_writeable_only.split(",")]
            convertedADisplays = self.localAnnotationDisplays["labels"]
            convertedADisplays[name]["gestures"].append({"label": label,
                                                         "function": function,
                                                         "writeable_only": (writeable_only == "yes"),
                                                         "attribute_writeable_only": attributeWriteableOnly})

        def _recordLabelGroup(name, children, css):
            convertedADisplays = self.localAnnotationDisplays["groups"]

            if name in convertedADisplays:
                raise PluginError("label group '%s' defined more than once in task %s" % (name, self.name))

            if not children:
                raise PluginError("label group %s has no children in task %s" % (name, self.name))

            convertedADisplays[name] = {"name": name,
                                        "children": children,
                                        "css": css}
            if name not in self.localAnnotationDisplays["labels"]:
                self.localAnnotationDisplays["order"].append(name)

        def _recordAttributeDisplay(aLabel, choice_displays = None, max_choice_size = None,
                                    name = None, editor_style = None, read_only = None,
                                    custom_editor = None, custom_editor_button_label = None,
                                    url_link = None, custom_editor_is_multiattribute = None,
                                    custom_editor_multiattributes = None,
                                    annotation_table_sort_fn = None, **kw):
            # max_choice_size.
            if max_choice_size is not None:
                try:
                    max_choice_size = int(max_choice_size)
                except ValueError:
                    raise PluginError("max_choice_size display specification '%s' in task %s is not an integer" %
                                        (max_choice_size, self.name))
            self.localAnnotationDisplays["attributes"][(name, aLabel)] = {
                "editor_style": editor_style,
                "max_choice_size": max_choice_size,
                "read_only": read_only == "yes",
                "custom_editor": custom_editor,
                "custom_editor_button_label": custom_editor_button_label,
                "url_link": url_link,
                "custom_editor_is_multiattribute": custom_editor_is_multiattribute == "yes",
                "custom_editor_multiattributes": (custom_editor_multiattributes and [s.strip() for s in custom_editor_multiattributes.split(",")]) or None,
                "choice_displays": choice_displays,
                "annotation_table_sort_fn": annotation_table_sort_fn}

        # Hold onto this until all the annotations are defined.
        newDescriptors = []
        
        asds = children["annotation_set_descriptors"]
        eLabelDict = {}
        if asds:
            convertedASDs = self.localAnnotationSetDescriptors
            # First, convert it to JSON for local storage.
            _modifyASDs(convertedASDs, **asds.attrs)
            for asd in asds.children["annotation_set_descriptor"]:
                newASD = _getOrAddASD(newDescriptors, **asd.attrs)
                for annot in asd.children["annotation"]:
                    _addLabelToASD(newASD, **annot.attrs)
                for attr in asd.children["attribute"]:
                    _addAttributeToASD(newASD, choices = [c.text for c in attr.children["choice"]],
                                       effective_labels = dict([(c.text, c.attrs["effective_label"])
                                                                for c in attr.children["choice"]
                                                                if c.attrs["effective_label"] is not None]) or None,
                                       minval = (attr.children["range"] and attr.children["range"].attrs["from"]) or None,
                                       maxval = (attr.children["range"] and attr.children["range"].attrs["to"]) or None,
                                       label_restrictions = [(t.attrs["label"], ((t.children["attributes"] and t.children["attributes"].wildcardAttrs.copy()) or None)) for t in attr.children["label_restriction"]],
                                       **attr.attrs)

        annotations = children["annotations"]

        if annotations:
            convertedASDs = self.localAnnotationSetDescriptors
            _modifyASDs(convertedASDs, **annotations.attrs)
            
            def _recurseThroughAnnotations(parentNode):
                # Order of definition counts.
                for node in parentNode.orderedChildren:
                    if node.label in ("span", "spanless"):
                        # Here's where all the real work happens.
                        span = ((node.label == "spanless") and "no")
                        asdSet = node.attrs["of_set"]
                        asdCategory = node.attrs["of_category"]
                        newASD = _getOrAddASD(newDescriptors, name = asdSet, category = asdCategory)

                        # OK, add the label.
                        _addLabelToASD(newASD, span = span, **node.attrs)

                        # Now, display information for the label.
                        displayAttrs = dict([(k[2:], v) for (k, v) in node.attrs.items()
                                             if k.startswith("d_")])
                        if set(displayAttrs.values()) != set([None]):
                            _recordLabelDisplay(name = node.attrs["label"], **displayAttrs)
                        
                        # Now, the attributes. This is tricky. Here are the children:
                        # int int_set, int_list, float, float_set, float_list,
                        # effective_labels, string, string_set, string_list,
                        # boolean, filler, filler_set, filler_list
                        # Order matters here, too.
                        
                        of_annotation = node.attrs["label"]
                        for attrChild in node.orderedChildren:
                            if attrChild.label == "d_gesture":
                                # It's actually a gesture annotation.
                                _recordGestureForLabel(of_annotation, **attrChild.attrs)
                                continue
                            attrASD = newASD
                            if attrChild.attrs["of_set"] or attrChild.attrs["of_category"]:
                                attrASD = _getOrAddASD(newDescriptors, name = attrChild.attrs["of_set"],
                                                       category = attrChild.attrs["of_category"])
                            d = {}
                            d.update(attrChild.attrs)
                            if attrChild.label in ("int", "int_set", "int_list"):
                                d["type"] = "int"
                                d["minval"] = attrChild.attrs["range_from"]
                                d["maxval"] = attrChild.attrs["range_to"]
                                choices = []
                                if attrChild.attrs["choices"]:
                                    # Split by commas. If that doesn't work, split by whitespace.
                                    choices = [s.strip() for s in attrChild.attrs["choices"].split(",")]
                                    if len(choices) == 1:
                                        choices = choices[0].split()
                                # Now look at the children.
                                for choiceChild in attrChild.children["choice"]:
                                    choices.append(choiceChild.attrs["value"])
                                d["choices"] = choices
                            elif attrChild.label in ("float", "float_set", "float_list"):
                                d["type"] = "float"
                                d["minval"] = attrChild.attrs["range_from"]
                                d["maxval"] = attrChild.attrs["range_to"]
                            elif attrChild.label == "boolean":
                                d["type"] = "boolean"
                            elif attrChild.label in ("string", "string_set", "string_list", "effective_labels"):
                                d["type"] = "string"
                                # choices, effective labels
                                choices = []
                                effectiveLabels = None
                                if attrChild.label == "effective_labels":
                                    effectiveLabels = dict([(c.attrs["value"], c.attrs["label"] or c.attrs["value"])
                                                            for c in attrChild.children["label"]])
                                    # Check display properties.
                                    for c in attrChild.children["label"]:
                                        displayAttrs = dict([(k[2:], v) for (k, v) in c.attrs.items()
                                                            if k.startswith("d_")])
                                        choices.append(c.attrs["value"])
                                        if displayAttrs:
                                            _recordLabelDisplay(name = c.attrs["label"] or c.attrs["value"], **displayAttrs)
                                        if c.children["d_gesture"]:
                                            for g in c.children["d_gesture"]:
                                                _recordGestureForLabel(c.attrs["label"] or c.attrs["value"], **g.attrs)
                                    if attrChild.children["display_group"]:
                                        # Use list() to freeze dictionary values in Python 3.
                                        _recordLabelGroup(attrChild.children["display_group"].attrs["name"] or of_annotation,
                                                          list(effectiveLabels.values()),
                                                          attrChild.children["display_group"].attrs["css"])
                                else:
                                    if attrChild.attrs["choices"]:
                                        # Split by commas. If that doesn't work, split by whitespace.
                                        choices = [s.strip() for s in attrChild.attrs["choices"].split(",")]
                                        if len(choices) == 1:
                                            choices = choices[0].split()
                                    # Now look at the children.
                                    for choiceChild in attrChild.children["choice"]:
                                        choices.append(choiceChild.attrs["value"])
                                d["choices"] = choices
                                d["effective_labels"] = effectiveLabels
                            elif attrChild.label in ("filler", "filler_set", "filler_list"):
                                d["type"] = "annotation"
                                fillerTypes = []
                                if attrChild.attrs["filler_types"]:
                                    # Split by commas. If that doesn't work, split by whitespace.
                                    tList = [s.strip() for s in attrChild.attrs["filler_types"].split(",")]
                                    if len(tList) == 1:
                                        tList = tList[0].split()
                                    fillerTypes += [(t, None) for t in tList]
                                for fillerChild in attrChild.children["filler_type"]:
                                    fillerTypes.append((fillerChild.attrs["label"], fillerChild.wildcardAttrs.copy()))
                                d["label_restrictions"] = fillerTypes
                            _addAttributeToASD(attrASD, of_annotation = of_annotation, **d)
                            # And deal with the display attrs.. but NOT for effective label children.
                            if attrChild.label != "effective_labels":
                                displayAttrs = dict([(k[2:], v) for (k, v) in attrChild.attrs.items()
                                                     if k.startswith("d_")])
                                choiceDisplays = {}
                                # Figure out the choice accelerators and display, if present.
                                if attrChild.label == "boolean":
                                    posChild = attrChild.children.get("positive")
                                    negChild = attrChild.children.get("negative")
                                    if posChild:
                                        choiceDisplays["yes"] = {"accelerator": posChild.attrs["d_accelerator"], "css": posChild.attrs["d_css"]}
                                    if negChild:
                                        choiceDisplays["no"] = {"accelerator": negChild.attrs["d_accelerator"], "css": negChild.attrs["d_css"]}                                        
                                elif attrChild.children.get("choice"):
                                    choiceDisplays = dict([(choice.attrs["value"], {"accelerator": choice.attrs["d_accelerator"], "css": choice.attrs["d_css"]})
                                                           for choice in attrChild.children["choice"]])
                                if choiceDisplays or (set(displayAttrs.values()) != set([None])):
                                    _recordAttributeDisplay(of_annotation, name = attrChild.attrs["name"],
                                                            choice_displays = choiceDisplays, **displayAttrs)
                        
                    elif node.label == "display_group":
                        _recurseThroughAnnotations(node)
                        # If there's a name, use it at the root,
                        # otherwise the first element is the parent.
                        # But finding the names of each of the children is tricky, because the display_group children
                        # may have either a name or a first child.
                        def _findChildNames(cList):
                            names = []
                            for child in cList:
                                if child.label == "display_group":
                                    if child.attrs["name"]:
                                        names.append(child.attrs["name"])
                                    elif not child.orderedChildren:
                                        # It's the name of the first child.
                                        raise PluginError("non-effective-label <display_group> has no children in task %s" % self.name)
                                    elif child.orderedChildren[0].label == "display_group":
                                        raise PluginError("<display_group> first child is <display_group> in task %s" % self.name)
                                    else:
                                        names.append(child.orderedChildren[0].attrs["label"])
                                else:
                                    names.append(child.attrs["label"])
                            return names
                                        
                        if node.attrs["name"]:
                            _recordLabelGroup(node.attrs["name"],
                                              _findChildNames(node.orderedChildren),
                                              node.attrs["css"])
                        elif not node.orderedChildren:
                            raise PluginError("non-effective-label <display_group> has no children in task %s" % self.name)
                        elif node.orderedChildren[0].label == "display_group":
                            raise PluginError("<display_group> first child is <display_group> in task %s" % self.name)
                        else:
                            _recordLabelGroup(node.orderedChildren[0].attrs["label"],
                                              _findChildNames(node.orderedChildren[1:]),
                                              None)
                    elif node.label == "annotation_set":
                        _getOrAddASD(newDescriptors, **node.attrs)

            _recurseThroughAnnotations(annotations)
            
        
        if newDescriptors:
            # Convert the values.
            
            for d in newDescriptors:

                for e in d["annotations"]:
                    if e["label"] in ("untaggable", "SEGMENT"):
                        raise PluginError("reserved annotation name '%s' mentioned in task %s" % (e["label"], self.name))
                localLabels = set([e["label"] for e in d["annotations"]])

                for attr in d["attributes"]:
                    
                    t = attr["type"]
                    if t not in unpermittedAttrSpecs:
                        raise PluginError("for task %s, attribute %s for labels %s has unknown type %s" %
                                            (self.name, attr["name"], attr["of_annotations"], t))
                    for e in unpermittedAttrSpecs[t]:
                        if attr[e]:
                            raise PluginError("for task %s, %s attribute %s for labels %s is not permitted a value for %s" %
                                                (self.name, t, attr["name"], attr["of_annotations"], e))
                        # It's null, but we don't want the key there at all.
                        del attr[e]
                    if t == "int":
                        attr["minval"] = _coerceValue(attr, "minval", attr["minval"], _inttypes, "neither int nor long")
                        attr["maxval"] = _coerceValue(attr, "maxval", attr["maxval"], _inttypes, "neither int nor long")
                        attr["default"] = _coerceValue(attr, "default", attr["default"], _inttypes, "neither int nor long")
                        attr["choices"] = [_coerceValue(attr, "choices", c, _inttypes, "neither int nor long")
                                           for c in attr["choices"]]
                        if attr["effective_labels"]:
                            attr["effective_labels"] = dict([(_coerceValue(attr, "effective_labels", k,
                                                                           _inttypes,
                                                                           "neither int nor long"), v)
                                                             for (k, v) in attr["effective_labels"].items()])
                    elif t  == "float":
                        attr["minval"] = _coerceValue(attr, "minval", attr["minval"], _numtypes,
                                                      "neither int nor long nor float")
                        attr["maxval"] = _coerceValue(attr, "maxval", attr["maxval"], _numtypes,
                                                      "neither int nor long nor float")
                        attr["default"] = _coerceValue(attr, "default", attr["default"], _numtypes,
                                                       "neither int nor long nor float")
                    elif t == "boolean":
                        attr["default"] = _coerceValue(attr, "default", attr["default"], (_boolCoerce,),
                                                       "neither 'yes' or 'no'")
                        
            # Check uniqueness.
            allDescriptorNames = set()
            allLabelNames = set()
            allAttributes = set()
            self._checkASDUniqueness(newDescriptors, allDescriptorNames, allLabelNames, allAttributes)
            convertedASDs["descriptors"] += newDescriptors

        adisplay = children["annotation_display"]
        if adisplay is not None:
            convertedADisplays = self.localAnnotationDisplays
            # First, convert it to something storeable.
            for lbl in adisplay.children["label"]:
                _recordLabelDisplay(**lbl.attrs)
                for g in lbl.children["gesture"]:
                    _recordGestureForLabel(lbl.attrs["name"], **g.attrs)
            for grp in adisplay.children["label_group"]:
                _recordLabelGroup(grp.attrs["name"],
                                  (grp.attrs["children"] and [s.strip() for s in grp.attrs["children"].split(",")]),
                                  grp.attrs["css"])
            for g in adisplay.children["gesture"]:
                if g.attrs["for_label"] is not None:
                    _recordGestureForLabel(g.attrs["for_label"], **g.attrs)
                else:
                    convertedADisplays["gestures"].append({"label": g.attrs["label"],
                                                           "function": g.attrs["function"],
                                                           "url_link": g.attrs["url_link"]})            
            for e in adisplay.children["attribute"]:
                # Accelerators for the choice values. The CSS option is not available in the old style.
                choiceAccelerators = dict([(choice.text, {"accelerator": choice.attrs["accelerator"], "css": None})
                                           for choice in e.children["choice"]])
                for aLabel in e.attrs["of_annotation"].split(","):
                    _recordAttributeDisplay(aLabel, choice_displays = choiceAccelerators, **e.attrs)

        # Similarity profiles
        simProfiles = children["similarity_profile"]
        if simProfiles:
            self.similarityProfiles = []
            for simProfile in simProfiles:
                p = {"name": simProfile.attrs["name"],
                     "tag_profiles": [{"true_labels": [l.strip() for l in tp.attrs["true_labels"].split(",")],
                                       "attr_equivalences": dict([(aq.attrs["name"], [s.strip() for s in aq.attrs["attrs"].split(",")]) for aq in tp.children["attr_equivalences"]]),
                                       "dimensions": [{"name": dim.attrs["name"],
                                                       "weight": float(dim.attrs["weight"]),
                                                       "method": dim.attrs["method"],
                                                       "aggregator_method": dim.attrs["aggregator_method"],
                                                       "param_digester_method": dim.attrs["param_digester_method"],
                                                       "params": dim.wildcardAttrs}
                                                      for dim in tp.children["dimension"]]}
                                      for tp in simProfile.children["tag_profile"]]}
                
                # Look at each remainder dimension. Its extra recognized params are
                # attrs, exclude_attrs, type, aggregation. It's gotta have one of these.
                
                for profile in p["tag_profiles"]:
                    for d in profile["dimensions"]:
                        if d["name"] == "_remainder":
                            if not (d["params"].get("attrs") or d["params"].get("exclude_attrs") or \
                                    d["params"].get("type") or d["params"].get("aggregation")):
                                raise PluginError("_remainder dimension in similarity profile %s in task %s doesn't specify attrs, exclude_attrs, type, or aggregation" % (p["name"], self.name))
                            d["remainder_params"] = {}
                            if d["params"].get("attrs"):
                                d["remainder_params"]["attrs"] = [s.strip() for s in d["params"]["attrs"].split(",")]
                                del d["params"]["attrs"]
                            if d["params"].get("exclude_attrs"):
                                d["remainder_params"]["exclude_attrs"] = [s.strip() for s in d["params"]["exclude_attrs"].split(",")]
                                del d["params"]["exclude_attrs"]
                            if d["params"].get("type"):
                                if d["params"]["type"] not in unpermittedAttrSpecs:
                                    raise PluginError("for task %s, _remainder dimension in similarity profile %s has unknown attribute type '%s'" %
                                                        (self.name, p["name"], d["params"]["type"]))
                                d["remainder_params"]["type"] = d["params"]["type"]
                                del d["params"]["type"]
                            if d["params"].get("aggregation"):
                                if d["params"]["aggregation"] not in ("none", "set", "list"):
                                    raise PluginError("for task %s, _remainder dimension in similarity profile %s has unknown attribute aggregation '%s'" %                                                        
                                            (self.name, p["name"], d["params"]["aggregation"]))
                                d["remainder_params"]["aggregation"] = d["params"]["aggregation"]
                                del d["params"]["aggregation"]
                
                if p["name"] == "<matdefault>":
                    raise PluginError("task %s uses the reserved name <matdefault> as a similarity profile name" % self.name)
                # Strata should not even be set if there aren't any in the input.
                if simProfile.children["stratum"]:
                    p["strata"] = [[l.strip() for l in s.attrs["true_labels"].split(",")] for s in simProfile.children["stratum"]]
                self.similarityProfiles.append(p)

        # Score profiles.
        scoreProfiles = children["score_profile"]
        if scoreProfiles:
            self.scoreProfiles = []
            for scoreProfile in scoreProfiles:
                p = {"name": scoreProfile.attrs["name"],
                     "label_limitation": (scoreProfile.children["label_limitation"] and [l.strip() for l in scoreProfile.children["label_limitation"].attrs["true_labels"].split(",")]) or None,
                     "aggregations": [{"name": aggr.attrs["name"],
                                       "true_labels": [l.strip() for l in aggr.attrs["true_labels"].split(",")]}
                                      for aggr in scoreProfile.children["aggregation"]],
                     "attr_decompositions": [{"attrs": [l.strip() for l in attr.attrs["attrs"].split(",")],
                                              "true_labels": [l.strip() for l in attr.attrs["true_labels"].split(",")]}
                                             for attr in scoreProfile.children["attr_decomposition"]],
                     "partition_decompositions": [{"method": pt.attrs["method"],
                                                   "true_labels": [l.strip() for l in pt.attrs["true_labels"].split(",")]}
                                                  for pt in scoreProfile.children["partition_decomposition"]],
                     "attrs_alone": (scoreProfile.children["attrs_alone"] and [l.strip() for l in scoreProfile.children["attrs_alone"].attrs["true_labels"].split(",")]) or None}
                if p["name"] == "<matdefault>":
                    raise PluginError("task %s uses the reserved name <matdefault> as a score profile name" % self.name)
                self.scoreProfiles.append(p)

        # I've tried to make sure all the annotation-related stuff happens before
        # any of the step-related stuff.
        
        # Engines.

        if children["engines"]:
            eBlock = children["engines"]            
            eCache = self.localEngineCache
            eCache["inherit_all"] = (eBlock.attrs["inherit_all"] == "yes")
            eCache["inherit"] = self._digestAttrListValue(eBlock, "inherit")
            eCache["enhance"] = self._digestAttrListValue(eBlock, "enhance")

            # First, check the subprocess parameters.
            if eBlock.children["java_subprocess_parameters"]:
                attrs = eBlock.children["java_subprocess_parameters"].attrs
                eCache["java_subprocess_parameters"]["stack_size"] = attrs["stack_size"]
                eCache["java_subprocess_parameters"]["heap_size"] = attrs["heap_size"]

            # Now, let's look at the engines.
            for e in eBlock.children["engine"]:
                eName = e.attrs["name"]
                newE = {"name": eName,
                        "languages": self._digestAttrListValue(e, "languages"),
                        "default_model": None,
                        "model_configs": {},
                        "step_config_class": None,
                        "step_config_settings": {}}
                eCache["engines"].append(newE)

                # Capture the default models.
                if e.children["default_model"]:
                    newE["default_model"] = e.children["default_model"].text

                # Capture the model configs.

                if e.children["model_config"]:
                    settingsList = e.children["model_config"]
                    for settings in settingsList:
                        configName = settings.attrs["config_name"]
                        # configName may be None. If there's already a default,
                        # or if there's already a model by that configName, complain.
                        if (configName is None) and (None in newE["model_configs"]):
                            raise PluginError("for task %s and engine %s, two local default model build settings defined" % (self.name, eName))
                        elif configName in newE["model_configs"]:
                            raise PluginError("for task %s and engine %s, two local model build settings named %s" % (self.name, eName, configName))
                        newE["model_configs"][configName] = (settings.attrs["class"], configName, 
                                                             self._digestXMLSettings(settings.children["build_settings"]))

                stepImpl = e.children["step_config"]
                sClassName = stepImpl.attrs["class"]
                sClass = FindPluginClass(sClassName, self.name, parentClass = PluginStep, tpe = "step config class", delayedPackages = delayedPackages)
                newE["step_config_class"] = sClass
                newE["step_config_settings"] = self._digestXMLSettings(stepImpl.children["create_settings"])

        # Steps.

        sBlock = children["steps"]
        if sBlock:
            sCache = self.localStepCache
            sCache["inherit_all"] = (sBlock.attrs["inherit_all"] == "yes")
            sCache["inherit"] = self._digestAttrListValue(sBlock, "inherit")
            for signalStep in sBlock.children["signal_step"]:
                sName = signalStep.attrs["name"]
                if sName in sCache["steps"]:
                    raise PluginError("multiple steps named '%s' in task %s" % (sName, self.name))
                sType = signalStep.attrs["type"]
                if sType and (sType not in ("hand", "auto")):
                    raise PluginError("type for signal step '%s' in task %s is neither 'hand' nor 'auto'" % (sName, self.name))
                sCache["steps"][sName] = {"name": sName,
                                          "step_type": "signal",
                                          "type": sType or "auto",
                                          "engine": signalStep.attrs["engine"]}
            for annotStep in sBlock.children["annotation_step"]:
                sName = annotStep.attrs["name"]
                if sName in sCache["steps"]:
                    raise PluginError("multiple steps named '%s' in task %s" % (sName, self.name))
                sType = annotStep.attrs["type"]
                if sType and (sType not in ("hand", "auto", "mixed", "auto-with-correction")):
                    raise PluginError("type for annotation step '%s' in task %s is not one of 'hand', 'auto', 'mixed', 'auto-with-correction'" % (sName, self.name))
                sEngine = annotStep.attrs["engine"]
                if (sType != 'hand') and (not sEngine):
                    raise PluginError("engine required for non-hand annotation step '%s' in task %s" % (sName, self.name))
                sCache["steps"][sName] = {"name": sName,
                                          "step_type": "annotation",
                                          "type": sType or "mixed",
                                          "engine": sEngine,
                                          "sets_required": self._digestAttrListValue(annotStep, "sets_required"),
                                          "sets_forbidden": self._digestAttrListValue(annotStep, "sets_forbidden"),
                                          "sets_modified": self._digestAttrListValue(annotStep, "sets_modified"),
                                          "sets_added": self._digestAttrListValue(annotStep, "sets_added"),
                                          "sets_desired": self._digestAttrListValue(annotStep, "sets_desired")}
            for tStep in sBlock.children["transform_step"]:
                sName = tStep.attrs["name"]
                if sName in sCache["steps"]:
                    raise PluginError("multiple steps named '%s' in task %s" % (sName, self.name))
                sCache["steps"][sName] = {"name": sName,
                                          "step_type": "transform",
                                          "engine": tStep.attrs["engine"],
                                          "sets_required": self._digestAttrListValue(tStep, "sets_required"),
                                          "sets_forbidden": self._digestAttrListValue(tStep, "sets_forbidden"),
                                          "sets_desired": self._digestAttrListValue(tStep, "sets_desired")}
                
        # Do the workflows. The elements in the workflows will
        # be dictionaries, and they'll guide the creation of the step
        # objects. 

        workflows = children["workflows"]
        if workflows:
            wCache = self.localWorkflowCache
            wCache["inherit_all"] = (workflows.attrs["inherit_all"] == "yes")
            wCache["inherit"] = self._digestAttrListValue(workflows, "inherit")
            for workflow in workflows.children["workflow"]:
                stepsList = []
                for step in workflow.children["step"]:
                    createSettings = self._digestXMLSettings(step.children["create_settings"])
                    runSettings = self._digestXMLSettings(step.children["run_settings"])
                    uiSettings = self._digestXMLSettings(step.children["ui_settings"])
                    # name, pretty_name, type
                    attrs = step.attrs.copy()                    
                    stepsList.append((attrs, createSettings, runSettings, uiSettings))
                wfAttrs = {"uiSettings": self._digestXMLSettings(workflow.children["ui_settings"]),
                           "undoable": workflow.attrs["undoable"] == "yes"}
                if workflow.attrs["name"] in wCache["workflows"]:
                    raise PluginError("multiple workflows named '%s' in task '%s'" % (workflow.attrs["name"], self.name))
                wCache["workflows"][workflow.attrs["name"]] = (wfAttrs, stepsList)
            
        # Do the workspaces.

        wsBlock = children["workspaces"]
        if wsBlock:
            wsCache = self.localWorkspaceCache
            wsCache["inherit_all"] = (wsBlock.attrs["inherit_all"] == "yes")
            wsCache["inherit"] = self._digestAttrListValue(wsBlock, "inherit")
            wsCache["enhance"] = self._digestAttrListValue(wsBlock, "enhance")
            wsCache["default_config"] = wsBlock.attrs["default_config"]
            foundNames = []
            for workspace in wsBlock.children["workspace"]:

                configName = workspace.attrs["config_name"]
                wfName = workspace.attrs["workflow"]
                
                if configName is not None:
                    if configName in foundNames:
                        raise PluginError("workspaces for task %s define a named configuration %s which redefines another local workspace configuration, either by config name or workflow" % (self.name, configName))
                    else:
                        foundNames.append(configName)
                elif wfName in foundNames:
                    raise PluginError("workspaces for task %s define an unnamed configuration for workflow %s which redefines another local workspace configuration, either by config name or workflow" % (self.name, wfName))
                    pass
                else:
                    foundNames.append(wfName)
                                        
                d = {"workflow": wfName,
                     "config_name": configName,
                     "operation_settings": {}}

                for action in workspace.children["operation"]:
                    aName = action.attrs["name"]
                    actionSettings = self._digestXMLSettings(action.children["settings"])
                    if aName in d["operation_settings"]:
                        raise PluginError("workspace definition in task %s has multiple operation settings for %s" % (self.name, aName))
                    d["operation_settings"][aName] = actionSettings
                
                wsCache["workspaces"].append(d)

    #
    # Managing resource files
    #

    def getResourceFile(self, pathSuff):
        return os.path.join(self.resourceDir, pathSuff)

    #
    # Settings. Config can override.
    #

    #
    # Retrieving local workflows, tag tables, and step implementations.
    #

    # This just checks uniqueness. Later, I need to ensure that the true label
    # for effective labels and the of_annotations for attributes actually exist.
    # We do that when we create the annotation type repository, in order to
    # ensure that inherited labels can be used.
    
    def _checkASDUniqueness(self, descriptors, allDescriptorNames, allLabelNames, allAttributes):

        for d in descriptors:
            if d["name"] in allDescriptorNames:
                raise PluginError("for task %s, duplicate annotation set descriptor name %s" % (self.name, d["name"]))
            allDescriptorNames.add(d["name"])
            for a in d["annotations"]:
                if a["label"] in allLabelNames:
                    raise PluginError("for task %s, duplicate annotation or effective label name %s" % (self.name, a["label"]))
                allLabelNames.add(a["label"])
            localLabelNames = [a["label"] for a in d["annotations"]]
            localAttributes = set()
            for attr in d["attributes"]:
                for tName in attr["of_annotations"]:
                    if (attr["name"], tName) in allAttributes:
                        raise PluginError("for task %s, duplicate attribute %s of annotation %s" %
                                            (self.name, attr["name"], tName))
                    allAttributes.add((attr["name"], tName))
                    localAttributes.add((attr["name"], tName))
                if ("effective_labels" in attr) and attr["effective_labels"]:
                    for v in attr["effective_labels"].values():
                        if v in allLabelNames:
                            raise PluginError("for task %s, duplicate annotation or effective label name %s" % (self.name, v))
                        allLabelNames.add(v)

    # The original version of the inheritor simple inherited ASDs. But the problem
    # with that is that you can inherit individual labels, but there's no support
    # for checking it. Really, the child ASDs should be viewed as adding
    # labels, not ASDs, and extending existing ASDs if necessary. 

    def _cacheASDs(self):
        if self._cachedAnnotationSetDescriptors is None:
            asdCache = []
            if self.localAnnotationSetDescriptors:
                asds = self.localAnnotationSetDescriptors
                if asds["inherit"] and self.parentObj:
                    asdCache = self.parentObj._inheritASDs(asds["inherit"])
                allLabelNames = set()
                allAttributes = set()
                # Don't need to check uniqueness for what's in the parent.
                for d in asdCache:
                    allLabelNames |= set([a["label"] for a in d["annotations"]])
                    for attr in d["attributes"]:
                        if ("effective_labels" in attr) and attr["effective_labels"]:
                            allLabelNames |= set(attr["effective_labels"].values())
                    for attr in d["attributes"]:
                        allAttributes |= set([(attr["name"], tName) for tName in attr["of_annotations"]])
                # But we DO for what's in the child. We know the names are unique within
                # each local set, because we check that when we define them, but now
                # when we merge, we DON'T want to check the descriptor names, because
                # they might overlap (i.e., the child may be adding labels to the parent).
                # But we do want to check the labels and attributes.
                self._checkASDUniqueness(asds["descriptors"], set(), allLabelNames, allAttributes)
                # Now, we need to enforce that of_annotations all exist, as well
                # as the truenames and attributes in the effective_labels. 
                for d in asdCache:
                    for attr in d["attributes"]:
                        for tName in attr["of_annotations"]:
                            if tName not in allLabelNames:
                                raise PluginError("for task %s and annotation set descriptor %s, attribute %s refers to undefined label %s" %
                                                    (self.name, d["name"], attr["name"], tName))
                # If it passes, merge them.
                parentASDs = dict([(d["name"], d) for d in asdCache])
                for childASD in asds["descriptors"]:
                    if childASD["name"] not in parentASDs:
                        asdCache.append(childASD)
                    else:
                        parentASD = parentASDs[childASD["name"]]
                        if parentASD["category"] != childASD["category"]:
                            raise PluginError ("for task %s, annotation set descriptor %s has a different category (%s) than that of its inherited parent (%s)" %
                                               (self.name, childASD["name"], childASD["category"], parentASD["category"]))
                        # We have to merge these. Which means replacing the parent with the
                        # child and copying in the appropriate information. An ASD has
                        # category, name, managed, annotations, attributes. We know that
                        # category and name match, and the child managed value is the correct one.
                        # So we just need to merge the annotations and attributes lists.
                        asdCache[asdCache.index(parentASD)] = childASD
                        # I don't think I need this, but just in case...
                        parentASDs[childASD["name"]] = childASD
                        childASD["annotations"] += parentASD["annotations"]
                        childASD["attributes"] += parentASD["attributes"]
            self._cachedAnnotationSetDescriptors = asdCache

    # If I inherit sets or categories, I take them as given; if the
    # attributes for the labels in them are specified in other sets
    # that aren't inherited, not my problem. But when I specify labels,
    # I want all their attributes, no matter what set they're in.
    # And I can't inherit labels from the admin set. And if
    # the label is specified in a set and by itself, still inherit
    # all the attributes.
    
    def _inheritASDs(self, inheritList):
        nameList = []
        categoryList = []
        setList = []
        for i in inheritList:
            if i in ("category:admin", "set:admin"):
                raise PluginError("for task, %s, can't inherit admin set or category" % self.name)
            if i.startswith("category:"):
                categoryList.append(i[9:])
            elif i.startswith("set:"):
                setList.append(i[4:])
            else:
                nameList.append(i)
        self._cacheASDs()
        # If you try to inherit from the nameList, you have to paw through
        # the individual sets.        
        setSeeds = [d for d in self._cachedAnnotationSetDescriptors
                    if (d["category"] and (d["category"] in categoryList)) or (d["name"] in setList)]
        trueLabelsFound = set()
        for cSeed in setSeeds:
            trueLabelsFound.update([a["label"] for a in cSeed["annotations"]])

        if nameList:
            setsAlreadyPresent = [d["name"] for d in setSeeds]
            nameList = set(nameList)
            addedSets = {}
            # Group the annotations and attributes into additional sets.
            for cSeed in self._cachedAnnotationSetDescriptors:
                if cSeed["name"] not in setsAlreadyPresent:
                    for aDesc in cSeed["annotations"]:
                        if aDesc["label"] in nameList:
                            if (cSeed["name"] == "admin") or (cSeed["category"] == "admin"):
                                raise PluginError("for task %s, can't inherit admin label %s" % (self.name, cSeed["name"]))
                            # Add the set if needed.
                            try:
                                thisSet = addedSets[cSeed["name"]]
                            except KeyError:
                                thisSet = {"name": cSeed["name"],
                                           "category": cSeed["category"],
                                           "annotations": [],
                                           "attributes": []}
                                addedSets[cSeed["name"]] = thisSet
                            thisSet["annotations"].append(aDesc)
                    # If somebody puts an annotation attribute in the admin set,
                    # we're not about to inherit it.
                    if cSeed["category"] != "admin":
                        for aDesc in cSeed["attributes"]:
                            if set(aDesc["of_annotations"]) & nameList:
                                # Add the set if needed.
                                try:
                                    thisSet = addedSets[cSeed["name"]]
                                except KeyError:
                                    thisSet = {"name": cSeed["name"],
                                               "category": cSeed["category"],
                                               "annotations": [],
                                               "attributes": []}
                                    addedSets[cSeed["name"]] = thisSet
                                thisSet["attributes"].append(aDesc)
            setSeeds += list(addedSets.values())

        # By this point, I should have all the sets and categories, and
        # have collected all the attributes and annotations that correspond
        # do the labels listed, in whatever set they appear.                        
                      
        return setSeeds

    def getAnnotationTypeRepository(self):
        if self._annotationTypeRepository is None:
            atp = Annotation.GlobalAnnotationTypeRepository()
            # Here's how we do it. First, we inherit all the relevant types
            # from the parent. Then, we add the local types. You CANNOT
            # redefine something that's already defined; that's an error.
            # I want to be able to inherit specific sets, so we really want
            # to cache the accumulated annotation set descriptors. Any
            # accumulated parent set will already have the appropriate uniqueness
            # conditions enforced already (no duplicate annots, no duplicate
            # attrs in the annots). We only need to augment the parent
            # with the local additions. And once we've accumulated all the
            # elements, we create the types. I think this is cleaner than
            # asking the types to copy themselves, etc.
            self._cacheASDs()
            allKnown = self.localAnnotationSetDescriptors and self.localAnnotationSetDescriptors.get("all_annotations_known", False)
            atp.fromJSONDescriptorList(self._cachedAnnotationSetDescriptors,
                                       allAnnotationsKnown = allKnown)
            self._annotationCategoriesToSets = {}
            self._annotationSetsToCategories = {}
            for d in self._cachedAnnotationSetDescriptors:
                cat = d.get("category")
                if cat is not None:
                    self._annotationSetsToCategories[d["name"]] = cat
                    try:
                        self._annotationCategoriesToSets[cat].add(d["name"])
                    except KeyError:
                        self._annotationCategoriesToSets[cat] = set([d["name"]])
                if d.get("managed"):
                    self._managedSets.add(d["name"])
            self._annotationTypeRepository = atp
        return self._annotationTypeRepository

    # Here's a convenience function.
    def newDocument(self, signal = None, docClass = Document.AnnotatedDoc):
        return docClass(signal = signal, globalTypeRepository = self.getAnnotationTypeRepository())

    def getAnnotationDisplayInfo(self):
        if self._cachedAnnotationDisplays is None:
            # Use the inheritance info from the local annotation set descriptors,
            # and then override and filter. The only label display info should
            # refer to an effective label or true label in the annotation type repository.
            atp = self.getAnnotationTypeRepository()
            # If there's any inheritance information at all, then get the display
            # info from the parent. The information has to be ordered, because
            # I need to compute the local tag info. Overriding locally overrides
            # the order.

            localInfo = self.localAnnotationDisplays
            
            if localInfo:
                localOverlapOrder = localInfo["order"][:]
                # Then, find all the elements in the local label dict which
                # are ordered. Note that overlap_rank may carelessly have been omitted
                # from the locally defined annotations...
                ranked = [e for e in localInfo["labels"].values() if e.get("overlap_rank") is not None]
                if ranked:
                    ranked.sort(key = lambda e: e["overlap_rank"])
                    orderedLabels = [e["name"] for e in ranked]
                    localOverlapOrder = [lab for lab in localOverlapOrder if lab not in orderedLabels] + orderedLabels
                
            if self.parentObj and self.localAnnotationSetDescriptors and \
               self.localAnnotationSetDescriptors["inherit"]:
                parentInfo = self.parentObj.getAnnotationDisplayInfo()
                tagOrder = parentInfo["order"][:]
                overlapOrder = parentInfo["overlapOrder"][:]
                labelDict = parentInfo["labels"].copy()
                groupDict = parentInfo["groups"].copy()
                attrDict = parentInfo["attributes"].copy()
                if "choice_displays" in attrDict:
                    # This has to be recursively copied.
                    attrDict["choice_displays"] = dict([(k, v.copy()) for (k, v) in attrDict["choice_displays"].items()])
                gestures = parentInfo["gestures"][:]
                if localInfo:
                    labelDict.update(localInfo["labels"])
                    groupDict.update(localInfo["groups"])
                    # Be sure to update the choice accelerators.
                    attrDictChoiceDisplays = attrDict.get("choice_displays")
                    attrDict.update(localInfo["attributes"])
                    if attrDictChoiceDisplays is not None:
                        for k, v in localInfo["attributes"]["choice_displays"].items():
                            if k not in attrDictChoiceDisplays:
                                attrDictChoiceDisplays[k] = v
                            else:
                                attrDictChoiceDisplays[k].update(v)
                        attrDict["choice_displays"] = attrDictChoiceDisplays
                    gestures += localInfo["gestures"]
                    tagOrder = [o for o in tagOrder if o not in localInfo["order"]] + localInfo["order"]
                    overlapOrder = [o for o in overlapOrder if o not in localOverlapOrder] + localOverlapOrder
            elif localInfo:
                labelDict = localInfo["labels"].copy()
                groupDict = localInfo["groups"].copy()
                attrDict = localInfo["attributes"].copy()
                if "choice_displays" in attrDict:
                    attrDict["choice_displays"] = dict([(k, v.copy()) for (k, v) in attrDict["choice_displays"].items()])
                gestures = localInfo["gestures"][:]
                tagOrder = localInfo["order"][:]
                overlapOrder = localOverlapOrder
            else:
                labelDict = {}
                groupDict = {}
                attrDict = {}
                tagOrder = []
                gestures = []
                overlapOrder = []
            # We'll build the order backward. Anything that's in a group's children
            # which isn't a mentioned group or a known label will be dropped;
            # any labels which aren't known labels will be dropped.
            labelDict = dict([(k, v) for (k, v) in labelDict.items() if atp.labelKnown(k)])
            # For the groups, first remove all the unknown labels. THEN
            # iteratively remove all the groups which have no children, and repeat
            # as long as you find something to remove. This won't catch cycles.
            # I'll deal with that when I actually build the tree.
            groupNames = groupDict.keys()
            for g in groupDict.values():
                g["children"] = [e for e in g["children"] if (e in groupNames) or atp.labelKnown(e)]
            while True:
                toRemove = set()
                for g in groupDict.values():
                    if len(g["children"]) == 0:
                        toRemove.add(g["name"])
                if not toRemove:
                    break
                for t in toRemove:
                    del groupDict[t]
                for g in groupDict.values():
                    g["children"] = [e for e in g["children"] if e not in toRemove]
            groupNames = groupDict.keys()
            tagOrder = [t for t in tagOrder if atp.labelKnown(t) or (t in groupNames)]
            info = {"labels": labelDict, "groups": groupDict, "order": tagOrder, "attributes": attrDict,
                    "gestures": gestures, "overlapOrder": overlapOrder}
            self._cachedAnnotationDisplays = info
        return self._cachedAnnotationDisplays

    # We're now going to try to construct this directly off the display info,
    # which has already been inherited.
    
    def getTagHierarchy(self):
        if self._tagHierarchyCache is None:
            atp = self.getAnnotationTypeRepository()
            displayInfo = self.getAnnotationDisplayInfo()

            if not displayInfo["groups"]:
                # If there's no groups, just leave the tag hierarchy none.
                return

            # Now, the displayInfo has all the elements I need.

            # So the idea of tag groups is that they can be inherited, but only the
            # entire family of groupings. No interleaving, one from column A, one from
            # column B nonsense. All groups without parents are toplevel groups;
            # no cycles are permitted; ui children are permitted to be defined
            # ONLY if the group doesn't already exist in the localTags list;
            # no accelerators are permitted for the ui children; 
            # every element which is referenced has to be a content tag.
            # Children are obligatory. The structure must be a tree - nothing can
            # be mentioned more than once, and all tag group names must be mentioned.

            # So first, collect the visible names and their categories.
            # This includes the effective labels.

            allNames = dict([(k, self.getCategoryForLabel(k)) for k in displayInfo["labels"]])

            # Next, check to see which groups are already defined, and
            # find the roots.

            allGroupNamesOrig = set(displayInfo["groups"].keys())
            allGroupNames = allGroupNamesOrig.copy()
            allTagNames = set(allNames.keys())

            for g in displayInfo["groups"].values():
                tName = g["name"]
                children = g["children"]
                uiCSS = g["css"]

                for child in children:
                    if (child not in allGroupNamesOrig) and \
                       (child not in allNames):
                        raise PluginError("tag group %s references child %s in task %s which is either not defined or has no defined CSS" %
                                            (tName, child, self.name))
                    if (child not in allGroupNames) and \
                       (child not in allTagNames):
                        raise PluginError("child tag %s is named more than once in tag hierarchy in task %s" % \
                                            (child, self.name))
                    if child in allGroupNames:
                        allGroupNames.remove(child)
                    else:
                        allTagNames.remove(child)
                if tName in allNames:
                    if allNames[tName] != "content":
                        raise PluginError("tag group %s references non-content tag in task %s" % \
                                            tName, self.name)
                    if uiCSS is not None:
                        raise PluginError("tag group %s references content tag but also tries to define its own CSS in task %s" % \
                                            (tName, self.name))
                elif atp.labelKnown(tName):
                    # There's no display definition for the label group name, but
                    # it's a known annotation, and there's CSS here for it. You can't
                    # do that; if you want this group name to be recognized
                    # as an annotation, you must define its style separately.
                    raise PluginError("in task '%s', the name of tag group '%s' is a known label which doesn't define its own CSS; this is not permitted. To use a known label as a tag group, you must declare it separately in the annotation display." % (self.name, tName))
                
            if len(allGroupNames) == 0:
                raise PluginError("tag groups in task %s are cyclic" % self.name)

            def _checkCycles(tbl, root, prefix):
                if root in prefix:
                    raise PluginError("tag group %s is cyclic in task %s" % (root, self.name))
                if root in tbl:
                    for child in tbl[root]["children"]:
                        _checkCycles(tbl, child, [root] + prefix)

            # These are the roots.
            for r in allGroupNames:
                _checkCycles(displayInfo["groups"], r, [])

            # All the checking is done. The tag groups constitute a tree, and no
            # name is present more than once. The elements remaining in allGroupNames are
            # the root tag groups, and the elements remaining in allTagNames, if content tags, are
            # the remaining tags (which must also be "roots"), AS LONG AS THEY'RE NOT IN THE
            # GROUPS AS WELL.

            self._tagHierarchyCache = {}
            for name in allTagNames:
                if (allNames[name] == "content") and (name not in displayInfo["groups"]):
                    self._tagHierarchyCache[name] = {}
            def _populateTagHierarchy(names, d):
                for name in names:
                    try:
                        g = displayInfo["groups"][name]
                        children = g["children"]
                        uiCSS = g["css"]
                        childD = {}
                        d[name] = {"css": uiCSS,
                                   "children": childD}
                        _populateTagHierarchy(children, childD)
                    except KeyError:
                        d[name] = {}
            _populateTagHierarchy(allGroupNames, self._tagHierarchyCache)

        return self._tagHierarchyCache

    # Let's redefine this, for the time being, from the annotation type
    # repository, etc. 

    def getJSONAnnotationRepositoryForDisplay(self):
        toplevelATR = self.getAnnotationTypeRepository().toJSON()
        jsonATR = toplevelATR["types"]
        # And I want to salt this repository with presentation information.
        # And category information, for that matter.
        displayInfo = self.getAnnotationDisplayInfo()
        for k, v in jsonATR.items():
            if k in displayInfo["labels"]:
                # There's display info for this label.
                v["display"] = displayInfo["labels"][k]
            v["category"] = self._annotationSetsToCategories.get(v.get("set_name"))
            v["managed"] = v.get("set_name") in self._managedSets
            for attr in v["attrs"]:
                attr["category"] = self._annotationSetsToCategories.get(attr.get("set_name"))
                attr["managed"] = attr.get("set_name") in self._managedSets
            if "effective_labels" in v:
                for eName, eData in v["effective_labels"].items():
                    if eName in displayInfo["labels"]:
                        eData["display"] = displayInfo["labels"][eName]
                    eData["category"] = self._annotationSetsToCategories.get(eData.get("set_name"))
        for (attrName, labelName), attrD in displayInfo["attributes"].items():
            for attr in jsonATR[labelName]["attrs"]:
                if attr["name"] == attrName:
                    attr["display"] = attrD
                    break
        return toplevelATR

    # In MATAnnotationInfoToJSON, I want to further modify the
    # output of getJSONAnnotationRepositoryForDisplay. But I can't
    # do it in the same call, because of how the code is constructed.

    @staticmethod
    def simplifyJSONDisplayAnnotationRepository(toplevelATR, removeNoncontentAnnotations = False, removeRedundantInfo = False):
        if not (removeNoncontentAnnotations or removeRedundantInfo):
            return toplevelATR
        
        import copy
        typeData = copy.deepcopy(toplevelATR["types"])
        toplevelATR["types"] = typeData
        if removeNoncontentAnnotations:
            # Use list() to freeze dictionary items in Python 3.
            for label, typeInfo in list(typeData.items()):
                if typeInfo.get("category") in ("zone", "token", "admin"):
                    del typeData[label]
        if removeRedundantInfo:
            for label, typeInfo in list(typeData.items()):
                if typeInfo.get("hasSpan"):
                    del typeInfo["hasSpan"]
                if "allAttributesKnown" in typeInfo:
                    if not typeInfo["allAttributesKnown"]:
                        del typeInfo["allAttributesKnown"]
                # This can be pruned because the UI doesn't use it.
                if "processable" in typeInfo:
                    if typeInfo["processable"] is True:
                        del typeInfo["processable"]
                if typeInfo.get("type"):
                    del typeInfo["type"]
                if typeInfo.get("attrs") == []:
                    del typeInfo["attrs"]
                elif typeInfo.get("attrs"):
                    for attr in typeInfo["attrs"]:
                        if ("aggregation" in attr) and  \
                           (attr["aggregation"] in ("none", None)):
                            del attr["aggregation"]
                        if attr.get("type") == "string":
                            del attr["type"]
                        if "display" in attr:
                            if attr["display"].get("editor_style") == "long_string":
                                del attr["display"]["editor_style"]
                            for k in ("custom_editor", "custom_editor_is_multiattribute", "custom_editor_multiattributes", "url_link",
                                      "editor_style", "max_choice_size", "read_only", "custom_editor_button_label",
                                      "annotation_table_sort_fn"):
                                if k in attr["display"]:
                                    if not attr["display"][k]:
                                        del attr["display"][k]
                            if attr["display"].get("choice_displays"):
                                for k, v in attr["display"]["choice_displays"].items():
                                    for subk in ("accelerator", "css"):
                                        if subk in v:
                                            if not v[subk]:
                                                del v[subk]
                            if not attr["display"]["choice_displays"]:
                                del attr["display"]["choice_displays"]                                    
                            if not attr["display"]:
                                del attr["display"]
                if "display" in typeInfo:
                    if typeInfo["display"].get("name"):
                        del typeInfo["display"]["name"]
                    for k in ["edit_immediately", "presented_name", "accelerator", "overlap_rank"]:
                        if k in typeInfo["display"]:
                            if not typeInfo["display"][k]:
                                del typeInfo["display"][k]
                    if typeInfo["display"].get("rendering_style") in (None, "normal"):
                        del typeInfo["display"]["rendering_style"]
                    if "gestures" in typeInfo["display"]:
                        if not typeInfo["display"]["gestures"]:
                            del typeInfo["display"]["gestures"]
                        else:
                            for g in typeInfo["display"]["gestures"]:
                                if g.get("writeable_only") is False:
                                    del g["writeable_only"]
                                if ("attribute_writeable_only" in g) and (not g["attribute_writeable_only"]):
                                    del g["attribute_writeable_only"]
                    if not typeInfo["display"]:
                        # dict is now empty.
                        del typeInfo["display"]
                if "effective_labels" in typeInfo:
                    if not typeInfo.get("effective_labels"):
                        del typeInfo["effective_labels"]
                    else:
                        for eLab, eLabData in typeInfo["effective_labels"].items():
                            if "display" in eLabData:
                                if eLabData["display"].get("name"):
                                    del eLabData["display"]["name"]
                                for k in ["edit_immediately", "presented_name", "accelerator", "overlap_rank"]:
                                    if k in eLabData["display"]:
                                        if not eLabData["display"][k]:
                                            del eLabData["display"][k]
                                if not eLabData["display"]:
                                    del eLabData["display"]
        return toplevelATR
    
    # These are now defined in the annotation type repository.
    # The main source of distinguishing attributes is the record of
    # attribute sets, but this may not be sufficient for our
    # purposes (e.g., the scorer ought to want all of them, even
    # if the attribute sets aren't named).
    
    def getEffectiveAnnotationLabel(self, annot, restrictToAnnotationSetNames = None, **kw):
        return self.getAnnotationTypeRepository().getEffectiveAnnotationLabel(annot, restrictToAnnotationSetNames = restrictToAnnotationSetNames, **kw)

    # Probably want to phase this out. The name is stupid, anyway.
    def getAnnotationTypesByCategory(self, cat):
        if self._annotationTypeRepository is None:
            self.getAnnotationTypeRepository()
        sets = self._annotationCategoriesToSets.get(cat)
        if sets is None:
            return []
        # Believe it or not, you can pass an actual set() object using *. The order
        # is arbitrary, but that's correct for this function.
        # This returns only the true labels.
        return self._annotationTypeRepository.getLabelsForAnnotationSetNames(sets)[0]

    def getNonAdminAnnotationSetNames(self):
        # Find the sets which aren't category:zone, category:admin or category:token.
        setNames = self.getAnnotationSetNames()
        for cat in ('zone', 'admin', 'token'):
            sets = self._annotationCategoriesToSets.get(cat)
            if sets is not None:
                setNames -= sets
        return setNames

    def getAnnotationSetsForCategory(self, cat):
        return self._annotationCategoriesToSets.get(cat)

    def getHandAnnotatableAnnotationSetNames(self):
        if self._annotationTypeRepository is None:
            self.getAnnotationTypeRepository()
        # Now, we loop through the steps, and find all the sets which
        # are modified or added.
        annotatableSets = set()
        for step in self.getSteps().values():
            if isinstance(step, AnnotationStep) and (step.sType != "auto"):
                annotatableSets |= step.setsAdded
                annotatableSets |= step.setsModified
        return annotatableSets

    def getAnnotationSetNames(self):
        # Get them all. If I want to merge them, I call getAnnotationLabelsForSetNames.
        return self.getAnnotationTypeRepository().getAnnotationSetNames()

    def convertAnnotationSetNames(self, baseSets, additionalSets = None, setsToExclude = None):
        baseSets = self.checkCategoriesOrSets(baseSets)
        if additionalSets is not None:
            baseSets |= self.checkCategoriesOrSets(additionalSets)
        if setsToExclude is not None:
            # This function barfs if a set to exclude doesn't exist, which is
            # a good check. However, if it's a built-in set, and it doesn't exist
            # (e.g., category:token, category:zone, category:admin) it
            # shouldn't complain.
            for setToExclude in setsToExclude:
                try:
                    trueSets = self.checkCategoriesOrSets([setToExclude])
                except PluginError:
                    if setToExclude in ("category:token", "category:zone", "category:admin"):
                        continue
                    else:
                        raise
                baseSets -= trueSets
        return baseSets

    def getAnnotationLabelsForSetNames(self, baseSets, additionalSets = None, setsToExclude = None, includeAttributeOnly = False):
        baseSets = self.convertAnnotationSetNames(baseSets, additionalSets = additionalSets, setsToExclude = setsToExclude)
        # Amazingly enough, Python 2.6 and 2.7 allows anything that can be
        # coerced into a sequence to be passed using the *, including sets.
        return self._annotationTypeRepository.getLabelsForAnnotationSetNames(baseSets, includeAttributeOnly = includeAttributeOnly)[0]

    def getCategoryForLabel(self, lab):
        if self._annotationTypeRepository is None:
            self.getAnnotationTypeRepository()
        sName = self._annotationTypeRepository.getAnnotationSetNameForLabel(lab)
        if sName is None:
            return None
        return self._annotationSetsToCategories.get(sName)

    def getLabelsForCategory(self, cat):
        if self._annotationTypeRepository is None:
            self.getAnnotationTypeRepository()
        sets = self._annotationCategoriesToSets.get(cat)
        if sets is None:
            return ([], [])
        # Believe it or not, you can pass an actual set() object using *. The order
        # is arbitrary, but that's correct for this function.
        return self._annotationTypeRepository.getLabelsForAnnotationSetNames(sets)

    def getAnnotationSetDetailsForCategory(self, cat, includeAttributeOnly = False,
                                           attributeDetails = False,
                                           effectiveLabelMapping = False):
        if self._annotationTypeRepository is None:
            self.getAnnotationTypeRepository()
        sets = self._annotationCategoriesToSets.get(cat)
        if sets is None:
            if attributeDetails:
                return [], [], {}, {}
            else:
                return [], []
        # Believe it or not, you can pass an actual set() object using *. The order
        # is arbitrary, but that's correct for this function.
        return self._annotationTypeRepository.getAnnotationSetDetails(sets, includeAttributeOnly = includeAttributeOnly,
                                                                      attributeDetails = attributeDetails,
                                                                      effectiveLabelMapping = effectiveLabelMapping)

    def getAnnotationSetDetailsForSetNames(self, baseSets, includeAttributeOnly = False,
                                           attributeDetails = False,
                                           effectiveLabelMapping = False):
        return self.getAnnotationTypeRepository().getAnnotationSetDetails(self.checkCategoriesOrSets(baseSets),
                                                                          includeAttributeOnly = includeAttributeOnly,
                                                                          attributeDetails = attributeDetails,
                                                                          effectiveLabelMapping = effectiveLabelMapping)

    def checkCategoriesOrSets(self, categoriesOrSets):
        if self._annotationTypeRepository is None:
            self.getAnnotationTypeRepository()
        finalSets = set()
        for c in categoriesOrSets:
            if c.startswith("category:"):
                sets = self._annotationCategoriesToSets.get(c[9:])
                if sets is None:
                    raise PluginError("unknown annotation category %s for task %s" % (c, self.name))
                finalSets.update(sets)
            elif not self._annotationTypeRepository.annotationSetNameKnown(c):
                raise PluginError("unknown annotation set name %s for task %s" % (c, self.name))
            else:
                finalSets.add(c)
        return finalSets

    def _getWorkflowDescTable(self):
        if self._workflowDescCache is None:
            wCache = self.localWorkflowCache
            if self.parentObj and (wCache["inherit_all"] or wCache["inherit"]):
                t = self.parentObj._getWorkflowDescTable()
                if wCache["inherit_all"]:
                    t = t.copy()
                else:
                    t = dict([(k, v) for (k, v) in t.items() if k in wCache["inherit"]])
            else:
                t = {}
            t.update(wCache["workflows"])
            self._workflowDescCache = t
        return self._workflowDescCache

    def getWorkflows(self):
        if self._workflowCache is None:
            # Compute the two local tables, and create objects
            # for each step.
            t = self._getWorkflowDescTable()
            impls = self.getSteps()
            wfTable = {}
            for wfName, (params, stepList) in t.items():
                finalSteps = []
                for declaredAttrs, createSettings, runSettings, uiSettings in stepList:
                    stepName = declaredAttrs["name"]
                    try:
                        globalStep = impls[stepName]
                    except KeyError:
                        raise PluginError("step %s mentioned in workflow %s unknown in task %s" % (stepName, wfName, self.name))
                    sImpl = globalStep.cloneForWorkflow(wfName, 
                                                        createSettings, runSettings, uiSettings,
                                                        prettyName = declaredAttrs.get("pretty_name"),
                                                        sType = declaredAttrs.get("type"))
                    finalSteps.append(sImpl)
                wfTable[wfName] = PluginWorkflow(wfName, self, finalSteps, **params)
            self._workflowCache = wfTable
        return self._workflowCache

    # Getting steps.

    def _getStepDescs(self):
        if self._stepDescCache is None:
            # Inherit if you can.
            sCache = self.localStepCache
            if self.parentObj and (sCache["inherit_all"] or sCache["inherit"]):
                sDict = self.parentObj._getStepDescs()
                if sCache["inherit_all"]:
                    sDict = sDict.copy()
                else:
                    # Inherit must be set.
                    sDict = dict([(k, v) for (k, v) in sDict.items() if k in sCache["inherit"]])
            else:
                sDict = {}
            sDict.update(sCache["steps"])
            self._stepDescCache = sDict
        return self._stepDescCache
    
    def getSteps(self):
        if self._stepCache is None:
            desc = self._getStepDescs()
            d = {}
            for k in desc.values():
                if k["step_type"] == "signal":
                    d[k["name"]] = SignalStep(self, k["name"], k["engine"], sType = k["type"])
                elif k["step_type"] == "annotation":
                    d[k["name"]] = AnnotationStep(self, k["name"], k["engine"], sType = k["type"], 
                                                  setsRequired = k["sets_required"],
                                                  setsForbidden = k["sets_forbidden"],
                                                  setsModified = k["sets_modified"],
                                                  setsAdded = k['sets_added'],
                                                  setsDesired = k["sets_desired"])
                elif k["step_type"] == "transform":
                    d[k["name"]] = TransformStep(self, k["name"], k["engine"],
                                                 setsRequired = k["sets_required"],
                                                 setsForbidden = k["sets_forbidden"],
                                                 setsDesired = k["sets_desired"])
                else:
                    # We should never, ever, ever be here.
                    raise PluginError("unknown step type %s" % k["step_type"])
            self._stepCache = d
        return self._stepCache

    def getManagedSets(self):
        if self._annotationTypeRepository is None:
            # This has to be created, because _managedSets is populated
            # as a side effect.
            self.getAnnotationTypeRepository()
        return self._managedSets
        
    def getStep(self, workflow, stepName):
        try:
            wf = self.getWorkflows()[workflow]
        except KeyError:
            raise PluginError("no workflow named %s" % workflow)
        s = wf.getStepByName(stepName)
        if s:
            return s
        else:
            raise PluginError("no step named %s in workflow %s" % (stepName, workflow))

    # Java subprocess setting defaults.
        
    def getJavaSubprocessParameters(self):
        if self._javaEngineSettingsCache is None:
            lCache = self.localEngineCache
            if self.parentObj and (lCache["inherit_all"] or lCache["enhance"] or lCache["inherit"]):
                # If there's a parent, and we're inheriting some engines,
                # inherit the subprocess parameters.
                p = self.parentObj.getJavaSubprocessParameters().copy()
            else:
                p = {}
            p["stack_size"] = lCache["java_subprocess_parameters"].get("stack_size") or p.get("stack_size")
            p["heap_size"] = lCache["java_subprocess_parameters"].get("heap_size") or p.get("heap_size")
            self._javaEngineSettingsCache = p
        return self._javaEngineSettingsCache

    # And now, the engines, complete with inheritance and override.

    def _getEngineDescs(self):
        if self._engineDescCache is None:
            # First, we inherit everything, and then once we have everything, we
            # create a PluginEngine instance.
            lCache = self.localEngineCache
            
            def _copyEngineDesc(v):
                v = v.copy()
                if v["languages"]:
                    v["languages"] = v["languages"][:]
                v["default_model"] = v["default_model"]
                v["step_config_settings"] = v["step_config_settings"].copy()
                v["model_configs"] = dict([(cfgName, (cName, cfgName, mSettings.copy()))
                                           for (cName, cfgName, mSettings) in v["model_configs"].values()])
                return v
            
            enhanceList = []
            if self.parentObj and (lCache["inherit_all"] or lCache["enhance"] or lCache["inherit"]):
                inheritList = lCache["inherit"] or []
                enhanceList = lCache["enhance"] or []
                eDict = dict([(k, _copyEngineDesc(v))
                              for (k, v) in self.parentObj._getEngineDescs().items()
                              if lCache["inherit_all"] or (k in inheritList) or (k in enhanceList)])
            else:
                eDict = {}
            
            for e in lCache["engines"]:
                if (e["name"] in enhanceList) and (e["name"] in eDict):
                    # We augment rather than overwrite.
                    parent = eDict[e["name"]]
                    # Augment the language list.
                    if e["languages"]:
                        if parent["languages"]:
                            parent["languages"] = list(set(e["languages"]) | set(parent["languages"]))
                        else:
                            parent["languages"] = e["languages"].copy()
                    # If the step config is overwritten, overwrite the create settings.
                    if e["step_config_class"]:
                        parent["step_config_class"] = e["step_config_class"]
                        parent["step_config_settings"].update(e["step_config_settings"])
                    # Update the default models and model configs.
                    if e["default_model"]:
                        parent["default_model"] = e["default_model"]
                    if e["model_configs"]:
                        for cName, cfgName, mSettings in e["model_configs"].values():
                            parent["model_configs"][cfgName] = (cName, cfgName, mSettings.copy())
                else:
                    eDict[e["name"]] = _copyEngineDesc(e)

            self._engineDescCache = eDict
            
        return self._engineDescCache

    def getEngines(self):
        if self._engineCache is None:
            cache = self._getEngineDescs()
            self._engineCache = {}
            # Build an engine object for everything in the cache.
            for k, v in cache.items():
                defaultModel = v["default_model"] or None
                self._engineCache[k] = PluginEngine(self, k, defaultModel = defaultModel,
                                                    modelConfigs = v["model_configs"] or None,
                                                    stepConfigClass = v["step_config_class"],
                                                    stepConfigSettings = v["step_config_settings"] or None,
                                                    languages = v["languages"] or None)
        return self._engineCache

    def getTrainableEngines(self):
        return dict([(k, v) for (k, v) in self.getEngines().items() if v.isTrainable()])

    # We need to find the steps which have engines which are trainable.
    
    def getTrainableSteps(self):
        return dict([(k, s) for (k, s) in self.getSteps().items() if s.isTrainable()])

    # None indicates a default model. There may not be one. If there's
    # no local default model, climb until you find one. Ditto for named
    # configs. But if we climb, we need to copy it and create the proper
    # backpointer to THIS task.

    # The idea is that the name of the default model is inheritable,
    # but if it's a relative pathname, it'll be interpreted relative
    # to the child which was asked for the default model.

    # The idea is that we ask for a default model for a particular
    # language and engine. We inherit from the parents if some of the engines
    # are inherited.

    # Actually, you can use the same engine for multiple
    # steps. So we should pass in a step, not an engine.

    def getTrainableStep(self, step = None):
                
        trainableSteps = self.getTrainableSteps()
        if not trainableSteps:
            raise PluginError("no trainable steps")

        if (step is None) and (len(trainableSteps.values()) > 1):
            raise PluginError("getTrainableStep() requires an explicit step because task %s has multiple trainable steps defined" % self.name)

        if step is None:
            # In Python 3, values() isn't indexable because it's a view.
            stepEntry = list(trainableSteps.values())[0]
        else:
            try:
                stepEntry = trainableSteps[step]
            except KeyError:
                raise PluginError("%s is not a trainable step in task %s" % (step, self.name))
        return stepEntry            

    def getDefaultModel(self, language = None, step = None):

        # So the idea is that the default models are a hash from
        # engines to hashes from languages to engine names. So we get the
        # cache, and then we consider the language and engine.

        if (language is None) and (len(self.getLanguages().keys()) > 1):
            raise PluginError("getDefaultModel() requires an explicit language because task %s has multiple languages defined" % self.name)
        try:
            s = self.getTrainableStep(step = step)
        except PluginError as e:
            raise PluginError("getDefaultModel() failed while retrieving trainable step: %s" % str(e))

        return s.getDefaultModel(language = language)

    def getLanguage(self, language = None):
        if language is None:
            lgHash = self.getLanguages()
            if len(lgHash.keys()) > 1:
                raise PluginError("getLanguages() failed because no language was specified and the task has more than one language")
            return list(self.getLanguages().values())[0]
        else:
            try:
                return self.getLanguagesAndCodes()[language]
            except KeyError:
                raise PluginError("getLanguages() failed because the language name or code '%s' is unknown in the task" % language)

    # Model info works similarly.
    
    def getModelInfo(self, configName = None, step = None):
        return self.getTrainableStep(step = step).getModelInfo(configName = configName)

    # Ditto workspaces.

    def getWorkspaceOperations(self):
        raise PluginError("Workspaces don't work yet")

    # Very much like _getEngineDescs.
    
    def _getWorkspaceDescCache(self):
        if self._workspaceDescCache is None:
            wsCache = self.localWorkspaceCache
            
            def _copyWorkspaceDesc(v):
                v = v.copy()
                v["operation_settings"] = v["operation_settings"].copy()
                # TURN OFF DEFAULT SETTINGS WHEN YOU COPY.
                v["default"] = False
                return v
            
            enhanceList = []
            if self.parentObj and (wsCache["inherit_all"] or wsCache["enhance"] or wsCache["inherit"]):
                inheritList = wsCache["inherit"] or []
                enhanceList = wsCache["enhance"] or []
                wsDict = dict([(k, _copyWorkspaceDesc(v))
                               for (k, v) in self.parentObj._getWorkspaceDescCache().items()
                               if wsCache["inherit_all"] or (k in inheritList) or (k in enhanceList)])
            else:
                wsDict = {}

            for ws in wsCache["workspaces"]:

                # If something has a config name, it can only overwrite
                # a config name.

                if ws.get("config_name") is not None:
                    if ws["config_name"] in wsDict:
                        if wsDict[ws["config_name"]].get("config_name") is None:
                            # Can't override a bare workspace with a config name.
                            raise PluginError("named workspace configuration %s in task %s attempts to update or replace unnamed parent workspace configuration" % (ws["config_name"], self.name))
                        elif ws["config_name"] in enhanceList:
                            overrideBlock = wsDict[ws["config_name"]]
                            overrideBlock["operation_settings"].update(ws["operation_settings"])
                        else:
                            wsDict[ws["config_name"]] = _copyWorkspaceDesc(ws)
                    else:
                        wsDict[ws["config_name"]] = _copyWorkspaceDesc(ws)
                elif ws["workflow"] in wsDict:
                    if wsDict[ws["workflow"]].get("config_name") is not None:
                        # Can't override a config name with a bare workspace.
                        raise PluginError("unnamed workspace configuration for workflow %s in task %s attempts to update or replace named parent workspace configuration" % (ws["workflow"], self.name))
                    elif ws["workflow"] in enhanceList:
                        overrideBlock = wsDict[ws["workflow"]]
                        overrideBlock["operation_settings"].update(ws["operation_settings"])
                    else:
                        wsDict[ws["workflow"]] = _copyWorkspaceDesc(ws)
                else:
                    wsDict[ws["workflow"]] = _copyWorkspaceDesc(ws)

            # Try to inherit.
            defaultConfig = wsCache["default_config"]
            if (defaultConfig is None) and self.parentObj:
                for (k, v) in self.parentObj._getWorkspaceDescCache().items():
                    if v["default"]:
                        defaultConfig = k
                        break
            
            if defaultConfig is not None:
                # Don't check this yet - it may be the name of a workflow.
                self._workspaceDefaultConfig = defaultConfig                
            self._workspaceDescCache = wsDict
            
        return self._workspaceDescCache

    def getWorkspaceDescription(self, wfOrConfigName = None, language = None):
        descs = self._getWorkspaceDescCache()
        # If there's no description by the given name, see
        # if there's a workflow by that name. We return a triple
        # of workflow, operation settings, step settings, where
        # we know that the workflow is workspace-eligible and compatible
        # with the language.
        opSettings = None
        stepSettings = None
        workflow = None        
        eligibilityChecked = False
        wfDict = self.getWorkflows()
        if (wfOrConfigName is None) and self._workspaceDefaultConfig:
            wfOrConfigName = self._workspaceDefaultConfig
        if wfOrConfigName:
            if wfOrConfigName in descs:
                desc = descs[wfOrConfigName]
                if desc["workflow"] not in wfDict:
                    raise PluginError("requested workspace configuration %s specifies a workflow %s which is unknown in task %s" % (wfOrConfigName, desc["workflow"], self.name))
                else:
                    opSettings = desc["operation_settings"]
                    workflow = wfDict[desc["workflow"]]
            elif wfOrConfigName in wfDict:
                workflow = wfDict[wfOrConfigName]
            else:
                raise PluginError("requested workspace configuration %s is unknown in task %s as a workspace configuration or workflow" % (wfOrConfigName, self.name))
        else:
            # See how many workflows you have. If it's one, try it.
            # Otherwise, find an eligible one.
            wfList = wfDict.values()
            if wfList:
                if len(wfList) == 1:
                    # Must coerce this to a list for Python 3
                    workflow = list(wfList)[0]
                    eligibilityChecked = True
                else:
                    eligibleWfs = [wf for wf in wfList if wf.workspaceEligible()]
                    if len(eligibleWfs) == 1:
                        workflow = eligibleWfs[0]
                        eligibilityChecked = True                        
                    else:
                        # Too many of them.
                        raise PluginError("a workspace name must be provided to getWorkspaceDescription() because the task %s has multiple eligible workspaces" % self.name)
            else:
                raise PluginError("getWorkspaceDescription() could not find any workflows in task %s" % self.name)

        if not workflow:
            raise PluginError("couldn't find eligible workflow for workspace")

        if (not eligibilityChecked) and (not workflow.workspaceEligible()):
            raise PluginError("requested workflow %s is not workspace-eligible" % workflow.name)

        if (language is not None) and (not workflow.languageEligible(language)):
            raise PluginError("requested workflow %s does not support language %s" % (workflow.name, language))

        return workflow, opSettings or {}

    # This can be specialized. The backend doesn't
    # care at all what's in this; it's just for passing to
    # the frontend Javascript, as a JSON object.

    def getCGIWorkflowMetadata(self, wfObj):
        p = wfObj.params.copy()
        p["uiSettings"] = wfObj.uiSettings or {}
        return p

    # Catch-all for tasks, if they need to add something.
    def enhanceCGIMetadata(self, mData):
        pass

    # This can be specialized. It should be a list of
    # lines to be printed out in the usage mssage.

    def getCmdlineTaskMetadata(self):
        return []

    # In MAT 3.0, we look at the step implementations.

    def _getStepClassTable(self):
        visited = {}
        for wf in self.getWorkflows().values():
            for step in wf.stepList:
                if step.stepImplementation is None:
                    continue
                if isinstance(step, AnnotationStep) and (step.sType == "hand"):
                    # The mixed case, specialized to hand annotation.
                    continue
                cls = step.stepImplementation.__class__
                try:
                    try:
                        visited[cls][(step.stepName, step.prettyName)].append(wf.name)
                    except KeyError:
                        visited[cls][(step.stepName, step.prettyName)] = [wf.name]
                except KeyError:
                    visited[cls] = {(step.stepName, step.prettyName): [wf.name]}
        return visited

    def addOptions(self, aggregator):
        # For the various workflows, for the various steps,
        # collect all the classes, and ask each of them
        # for their options.
        # Update: since we can reuse the implementation classes
        # in multiple steps, I can't collapse it across step names
        # anymore.
        visited = self._getStepClassTable()
        for cls, stepDict in visited.items():
            byStepName = {}
            for (stepName, prettyName), wfList in stepDict.items():
                try:
                    entries = byStepName[stepName]
                except KeyError:
                    entries = []
                    byStepName[stepName] = entries
                if prettyName and (prettyName != stepName):
                    entries.append("step '%s' (true name '%s', workflows %s)" % (prettyName, stepName, ", ".join(wfList)))
                else:
                    entries.append("step '%s' (workflows %s)" % (stepName, ", ".join(wfList)))
            for stepName, entries in byStepName.items():
                stepStr = ", ".join(entries)
                activator = (None, cls, None, stepName, None)
                cls.addOptions(aggregator, heading = "Options for " + stepStr, activator = activator)
                # And then, register the activator. I'd love this to actually be a
                # dynamic choice option, but I can't figure out how to make that happen
                # when the trigger is an instance rather than a class.
                aggregator.activatedSet.add(activator)

    # This probably won't be specialized. Enforces
    # inheritance for display config.

    def _inheritValue(self, attr, default = None):
        localV = getattr(self, attr)
        if localV == "":
            return default
        elif localV is not None:
            return localV
        elif self.parentObj:
            return self.parentObj._inheritValue(attr, default)
        else:
            return default

    def getDisplayConfig(self):
        return self._inheritValue("displayConfig")

    def getShortWebName(self):
        return self._inheritValue("shortWebName")

    def getLongWebName(self):
        return self._inheritValue("longWebName")
        
    def getLanguages(self):
        if self._languageCache is None:
            lCache = self.localLanguageCache
            if self.parentObj and (lCache["inherit_all"] or lCache["inherit"]):
                h = self.parentObj.getLanguages()
                if not lCache["inherit_all"]:
                    h = dict([(k, v) for (k, v) in h.items() if k in lCache["inherit"]])
                else:
                    h = h.copy()
            else:
                h = {}
            for lg in lCache["languages"]:
                h[lg["name"]] = lg
            self._languageCache = h
            h2 = h.copy()
            # Use list() to freeze dictionary values in Python 3.
            for lg in list(h2.values()):
                h2[lg["code"]] = lg
            self._languageAndCodeCache = h2
            if len(self._languageCache.keys()) == 0:
                # Only flag this if the task is visible.
                if self.visible:
                    raise PluginError("task %s doesn't specify any languages" % self.name)
            elif len(self._languageCache.keys()) > 1:
                # raise PluginError, "MAT doesn't yet support multiple languages"
                pass
        return self._languageCache

    def getLanguagesAndCodes(self):
        # Cache it.
        self.getLanguages()
        return self._languageAndCodeCache
        
    # Looks for an appropriate leaf.

    # If eligibleLeafTaskImplementation throws a plugin error,
    # then we should catch the error and cache it as a possible
    # failure reason.

    def getTaskImplementation(self, wfName, steps, language = None, **params):
        failureReasons = []
        if (language is not None) and (language not in self.getLanguagesAndCodes().keys()):
            raise PluginError("requested language or code '%s' not supported by task" % language)
        res = self._getTaskImplementation(wfName, steps, failureReasons, **params)
        if res is None:
            if failureReasons:
                raise PluginError("task implementation not found; possible reasons: %s" % ("; ".join(failureReasons)))
        elif self.visible and (res is not self):
            # Pretty unfriendly, since whoever's asking for it probably
            # intended to ask for this task, and you'll never get this task.
            raise PluginError("task %s is visible, but it's not accessible for operations because it's not a leaf in the task tree" % self.name)
        return res            

    def _getTaskImplementation(self, wfName, steps, failureReasons, **params):
        if not self.children:
            # I don't think this throws the relevant error anymore.
            # I think it all happens in eligibleLeafTaskImplementation.
            try:
                return self.eligibleLeafTaskImplementation(wfName, steps, failureReasons, **params)
            except PluginError as e:
                if str(e) not in failureReasons:
                    failureReasons.append(str(e))
                return None
        else:
            for child in self.children:
                a = child._getTaskImplementation(wfName, steps, failureReasons, **params)
                if a is not None:
                    return a
            return None

    # If there are no steps, return yourself.
    
    def eligibleLeafTaskImplementation(self, wfName, steps, failureReasons, **params):
        try:
            if not steps:
                return self
            else:
                wfObj = self.getWorkflows()[wfName]
                # Give each step that's being invoked an opportunity
                # to reject the params.
                for step in wfObj.stepList:
                    if step.stepImplementation and \
                       ((step.stepName in steps) or (step.prettyName and (step.prettyName in steps))) and \
                       not step.stepImplementation.paramsSatisfactory(wfName, failureReasons, **params):
                        return None
                return self
        except KeyError:
            return None

    # Returns a list of four-tuples (taskName, serverName, lg, serverCmd).
    # Only leaves have servers.

    def findServers(self):
        if not self.children:
            serverTable = {}
            # Look through the workflows.
            for wfObj in self.getWorkflows().values():
                for step in wfObj.stepList:
                    # The key is (serverName, language). With the
                    # taskName, that's the address of the server.
                    k = step.getServerKeys()
                    if k is not None:
                        for key in k:
                            if key not in serverTable:
                                try:
                                    serverTable[key] = step.createServer(*key)
                                except TaggerConfigurationError:
                                    pass
            return [(self.name, serverName, lg, inst) for ((serverName, lg), inst) in serverTable.items()]
        else:
            serverTable = []
            for c in self.children:
                serverTable += c.findServers()
            return serverTable
    
    # Subclass this one. Used in the tokenization and tagging methods.
    # The zone types that are returned here must be the same zone types
    # assigned in the zone step (i.e., by the addZones() method of
    # the PrepStep).

    # SAM 5/26/11: Actually, the problem is that I'm trying to
    # overload the zones, because I don't have enough categories of
    # annotations. In 2.0, I'll fix this right. But for now, I'm
    # going to have the task do a more thorough job. Instead of
    # just returning the zone attribute and region type, I'm
    # going to return the true zone info.

    def getTrueZoneInfo(self):
        # We can only use this if the task has a single zone type
        # with a single attribute, which has a single value.
        tLabs, eLabs = self.getLabelsForCategory("zone")
        if len(tLabs) > 1:
            raise PluginError("retrieving default zone info failed because there's more than one zone label")
        elif not tLabs:
            raise PluginError("retrieving default zone info failed because there aren't any zone labels")
        else:
            tLab = tLabs[0]
            thisAttr = None
            theseVals = None
            atype = self.getAnnotationTypeRepository()[tLab]
            if len(atype.attr_list) > 1:
                raise PluginError("retrieving default zone info failed because the default zone label has more than one attribute")
            elif atype.attr_list:
                thisAttr = atype.attr_list[0]
                if thisAttr._typename_ != "string":
                    raise PluginError("retrieving default zone info failed because the single attribute of the default zone label is not a string")
                if not thisAttr._choiceAttribute:
                    raise PluginError("retrieving default zone info failed because the single attribute of the default zone label is not a choice attribute")
                theseVals = list(thisAttr.choices)
                thisAttr = thisAttr.name
            return tLab, thisAttr, theseVals
        
    # For the experiment engine, so we can specialize how the tables
    # are computed.

    def augmentTagSummaryScoreTable(self, t):
        return t

    def augmentTokenSummaryScoreTable(self, t):
        return t

    def augmentDetailScoreTable(self, t):
        return t

    # For the workspace engine, so we can specialize the operations
    # and folders.

    def workspaceCustomize(self, workspace, create = False):
        pass

    # Stub for task updating:

    def workspaceUpdate2To3(self, newWs, oldWorkspace, basenames):
        pass
        
    # Similarity profiles. The profiles need to be compiled when
    # the scorer uses them, and also when they're shipped to the
    # UI. But I can't compile them in advance, because the compilation
    # process sometimes requires runtime information (e.g., when 
    # specific content annotation sets are included or included
    # in the scorer). 

    def getSimilarityProfile(self, name = None):
        if name == "<matdefault>":
            # Reserved. Force the pairer to use the default profile.
            return None
        elif self.similarityProfiles is None:
            if name is not None:
                raise PluginError("No similarity profile named '%s'" % name)
            return None
        elif (name is None) and (len(self.similarityProfiles) == 1):
            return self.similarityProfiles[0]
        else:
            for p in self.similarityProfiles:
                if (name is None) and (p.get("name") is None):
                    return p
                elif p.get("name") == name:
                    return p
            if name is None:
                raise PluginError("Task has multiple similarity profiles, but none are identified as the default; profile must be explicitly specified")
            else:
                raise PluginError("No similarity profile named '%s'" % name)
    
    # This is called when a profile is shipped to the UI or used in the
    # pairer. As noted above, we can't cache this because the contentAnnotationSets
    # may differ from the default.
    
    def _compileSimilarityProfile(self, profile, annotationTypeSorter):
        # First, copy the profile recursively. It's just lists and dicts.
        import copy
        profile = copy.deepcopy(profile)
        # Now, things need to happen. First, the set: elements in the
        # strata need to be converted. Second, the same has to happen for the
        # true label lists in the tag profiles. Third, we need to add a mapping
        # from the labels to the stratum index. Finally, the strata need to be
        # partitioned into spanned and spanless.
        
        globalTable = self.getAnnotationTypeRepository()
        if "strata" in profile:
            finalStrata, labelToStratumIdx = self._compileKnownSimilarityProfile(annotationTypeSorter, profile)
        else:
            finalStrata, labelToStratumIdx = self._constructSimilarityProfile(annotationTypeSorter)

        # Record the converted strata.
        profile["strata"] = finalStrata
        profile["labelToStratumIndex"] = labelToStratumIdx

        for tp in profile.get("tag_profiles", []):
            # Convert sets to labels.
            tp["true_labels"] = globalTable.expandTrueLabelSets(tp["true_labels"])
        return profile

    def _compileKnownSimilarityProfile(self, annotationTypeSorter, profile):
        globalTable = self.getAnnotationTypeRepository()
        strata = [globalTable.expandTrueLabelSets(stratum) for stratum in profile["strata"]]
        # Each compiled stratum is a pair of (spanned, spanless). They may never
        # pair with each other. The spanned are always done first; that should
        # make for a useful default if we ever develop a default comparison
        # for spanless annotations.
        strataToCheck = []
        finalStrata = []
        labelToStratumIdx = {}
        for s in strata:
            stratum = ([], [])
            aLabStratum = ([], [])
            for aLab in s:
                if aLab in labelToStratumIdx:
                    raise PluginError("label '%s' appears more than once in similarity stratum for task %s" % \
                                         (aLab, self.name))
                labelToStratumIdx[aLab] = len(finalStrata)
                atype = globalTable.get(aLab)
                if atype is None:
                    raise PluginError("unknown label '%s' in similarity stratum for task %s" % (aLab, self.name))
                # Don't need the other arguments. This may be redundant, because if there are no
                # strata, we don't need to check this, since we know that it's already
                # content. But this will get it into the table, which is also useful.
                if not annotationTypeSorter.isContentType(atype):
                    raise PluginError("label '%s' in similarity stratum for task %s is not a content tag" % \
                                         (aLab, self.name))
                if atype.hasSpan:
                    stratum[0].append(atype)
                    aLabStratum[0].append(aLab)
                else:
                    stratum[1].append(atype)
                    aLabStratum[1].append(aLab)
            strataToCheck.append(stratum)
            finalStrata.append(aLabStratum)
        # Check the stratification.
        self._checkStratification(strataToCheck)
        return finalStrata, labelToStratumIdx

    def _constructSimilarityProfile(self, annotationTypeSorter):

        globalTable = self.getAnnotationTypeRepository()
        # We have a task, so we're going to have annotation sets. Note that in this case,
        # we want to get the labels which host attributes which may have been mentioned.
        labs = globalTable.getLabelsForAnnotationSetNames(annotationTypeSorter.contentAnnotationSets, includeAttributeOnly = True)[0]

        # So to construct a default profile, we have to make sure there are no
        # type sequences.

        spannedLabPointsTo = {}
        spanlessLabPointsTo = {}
        strata = []
        labelToStratumIdx = {}

        for lab in labs:
            atype = globalTable.get(lab)
            pointsTo = set()
            if atype.hasAnnotationValuedAttributes:                
                for attr in atype.attr_list:
                    if isinstance(attr, Annotation.AnnotationAttributeType):
                        if attr.atomicLabelRestrictions:
                            pointsTo.update(attr.atomicLabelRestrictions)
                        if attr.complexLabelRestrictions:
                            pointsTo.update([p[0] for p in attr.complexLabelRestrictions])
            if atype.hasSpan:
                spannedLabPointsTo[lab] = pointsTo
            else:
                spanlessLabPointsTo[lab] = pointsTo

        # OK, those are our inputs.
        lastRound = 0
        allFound = set()
        while spannedLabPointsTo or spanlessLabPointsTo:
            thisStratum = ([], [])
            # Use list() to freeze dictionary items in Python 3.
            for k, v in list(spannedLabPointsTo.items()):
                v = v - allFound
                if not v:
                    thisStratum[0].append(k)
                    labelToStratumIdx[k] = len(strata)
                    del spannedLabPointsTo[k]
            # Update allFound after each subround.
            allFound.update(thisStratum[0])
            # Use list() to freeze dictionary items in Python 3.
            for k, v in list(spanlessLabPointsTo.items()):
                v = v - allFound
                if not v:
                    thisStratum[1].append(k)
                    labelToStratumIdx[k] = len(strata)
                    del spanlessLabPointsTo[k]
            allFound.update(thisStratum[1])
            # If I didn't find anything on this round,
            # it's a problem.
            if not (thisStratum[0] or thisStratum[1]):
                # Use list() to convert dictionary keys to something summable in Python 3.
                raise PluginError("encountered annotation type circularity in constructing default similarity profile for task %s (%s labels remaining)" %
                                  (self.name, ", ".join(list(spanlessLabPointsTo.keys()) + list(spannedLabPointsTo.keys()))))
            strata.append(thisStratum)

        return strata, labelToStratumIdx        

    # These strata are lists of lists of atype objects. Each atype
    # which points to another atype has to have that atype in a previous
    # stratum (stratified) or in the current stratum (unstratified). Otherwise,
    # it's an error. For that matter, right now, unstratified won't be handled.
    # Also, we're going to always process spanned before spanless, so
    # we can take that into account here.
    
    def _checkStratification(self, strata):
        alreadyFound = set()
        def checkAtypeSetStratification(atypeSet):
            atypeLabSet = set([atype.lab for atype in atypeSet])
            for atype in atypeSet:
                if atype.hasAnnotationValuedAttributes:
                    for attr in atype.attr_list:
                        if isinstance(attr, Annotation.AnnotationAttributeType):
                            labs = attr.atomicLabelRestrictions
                            if attr.complexLabelRestrictions:
                                if not labs:
                                    labs = set([p[0] for p in attr.complexLabelRestrictions])
                                else:
                                    labs = labs.copy()
                                    labs.update([p[0] for p in attr.complexLabelRestrictions])
                            if labs:
                                for l in labs:
                                    if l not in alreadyFound:
                                        if l in atypeLabSet:
                                            raise PluginError("label %s is stratified in task %s with an annotation type which refers to it" % (l, self.name))
                                        else:
                                            raise PluginError("label %s is referenced in task %s in a stratum before it's paired" % (l, self.name))
            alreadyFound.update(atypeLabSet)                                    
        for (spanned, spanless) in strata:
            checkAtypeSetStratification(spanned)
            checkAtypeSetStratification(spanless)
    
    def getScoreProfile(self, name = None):
        if name == "<matdefault>":
            # Reserved. Force the scorer to use the default profile.
            return None
        elif self.scoreProfiles is None:
            if name is not None:
                raise PluginError("No score profile named '%s'" % name)
            return None
        elif (name is None) and (len(self.scoreProfiles) == 1):
            return self.scoreProfiles[0]
        else:
            for p in self.scoreProfiles:
                if (name is None) and (p.get("name") is None):
                    return p
                elif p.get("name") == name:
                    return p                
            if name is None:
                raise PluginError("No default score profile")
            else:
                raise PluginError("No score profile named '%s'" % name)

    # This was in DocumentIO, but it turns out I need it in exporting
    # reconciliation documents. I need to check SEGMENTs here, too.

    # And here's a final glitch: sometimes, when we import documents,
    # we have the zones set, and in those cases, even if we don't have
    # a zoning step in the task, as long as we have evidence that the
    # zoning has happened, we want to put that in. Is this true of
    # any annotation set? We seem to be doing it step by step here, but
    # I don't exactly understand why. This may be historical, but I
    # don't have the courage to remove the code right now. So I'm
    # going to add a final condition.
    
    def _inferPhasesDone(self, doc):
        phasesChecked = set([])
        stepsPresent = set([s.get("set") for s in doc.getAnnotations(["SEGMENT"])
                            if s.get("annotator") is not None and s.get("set") is not None])
        for step in self.getSteps().values():
            phaseNames = step.getPhaseNames()
            allNamesChecked = True
            if phaseNames is not None:
                for phaseName in phaseNames:
                    if phaseName not in phasesChecked:
                        allNamesChecked = False
                        phasesChecked.add(phaseName)
                # If they haven't all been checked,
                # see if the document is done.
                # Remember, phasesDone marks something where at least SOMEthing
                # has happened, not that the phase is complete.
                if not allNamesChecked:
                    for s in (set(step.getSetsDone(doc)) | stepsPresent):
                        # What if the step is done? I should record it, right?
                        # Not so fast. I really want to know if its SETS
                        # are present - after all, perhaps not all the
                        # sets are present, and we don't want to pretend
                        # that they all are.
                        doc.recordSet(s)
        # Now, for all remaining sets, we should check the annotation
        # presence directly. But NOT admin.
        phasesChecked.add("admin")
        remainingNames = set(self.getAnnotationSetNames()) - phasesChecked
        if remainingNames:
            for presentR in self.getAnnotationTypeRepository().filterSetsPresent(doc, remainingNames):                
                doc.recordSet(presentR)            

class PluginEngine(object):

    # defaultModel is a string
    # modelConfigs is a hash from config names to triples that can be used to
    # create ModelInfo objects.
    # stepConfigClass is a subclass of PluginStep. It is required.
    
    def __init__(self, task, name, defaultModel = None, modelConfigs = None,
                 stepConfigClass = None, 
                 stepConfigSettings = None,
                 languages = None):
        self.task = task
        self.name = name
        self.stepConfigClass = stepConfigClass
        if not stepConfigClass:
            raise PluginError("engine '%s' defined in task %s without step config class" % (name, task.name))
        self.stepConfigSettings = stepConfigSettings
        if modelConfigs:
            self.modelConfigs = dict([(configName, (clsName, configName, buildSettings))
                                      for (clsName, configName, buildSettings) in modelConfigs.values()])
        else:
            self.modelConfigs = None
        self.languages = None
        self.languagesAndCodes = None
        if languages:
            # Convert the languages to contain all the codes and names.
            finalLangs = set()
            langNames = set()
            taskLangsAndCodes = self.task.getLanguagesAndCodes()
            for lg in languages:
                try:
                    lgObj = taskLangs[lg]
                except KeyError:
                    raise PluginError("language '%s' for engine '%s' is not among the languages or language codes for task %s" % (lg, name, task.name))
                finalLangs.add(lgObj["name"])
                finalLangs.add(lgObj["code"])
                langNames.add(lgObj["name"])
            self.languages = langNames
            self.languagesAndCodes = finalLangs
        
        if defaultModel and (not modelConfigs):
            raise PluginError("engine '%s' in task %s defines a default model, but has no model building configuration" % (name, task.name))
        # defaultModel is a string. It is differentiated by step and language when it's retrieved.
        self.defaultModel = defaultModel

    def isTrainable(self):
        if self.modelConfigs:
            return True
        else:
            return False

    def isRetaggable(self):
        return self.stepConfigClass and self.stepConfigClass.retaggable

    def isPartiallyTaggable(self):
        return self.stepConfigClass and self.stepConfigClass.partiallyTaggable

    def languageEligible(self, lg):
        return (not self.languagesAndCodes) or (lg in self.languagesAndCodes)

    def getDefaultModel(self, step = None, language = None):

        if not self.defaultModel:
            return None

        if not self.modelConfigs:
            raise PluginError("the engine %s is not trainable in task %s" % (self.name, self.task.name))

        if language is None:
            # In this case, there better be only one available language.
            availableLangs = self.languages or self.task.getLanguages().keys()
            if len(availableLangs) > 1:
                # Should this return None, or raise an error? Returning None for now.
                return None
            # At this point, we can grab the first language code we find.
            m = self.defaultModel
            if self.languages:
                lCode = self.task.getLanguages()[self.languages[0]]["code"]
            else:
                # In Python 3, values() isn't indexable because it's a view.
                lCode = list(self.task.getLanguages().values())[0]["code"]
        elif language not in (self.languagesAndCodes or self.task.getLanguagesAndCodes().keys()):
            raise PluginError("the requested language or code %s is not supported by the engine %s" % (language, self.name))
        else:
            m = self.defaultModel
            lCode = self.task.getLanguagesAndCodes()[language]["code"]
            
        # First, let's make sure there's a step.
        if step is None:
            trainableSteps = [s for s in self.task.getTrainableSteps().values() if
                              s.engine is self]
            if not trainableSteps:
                raise PluginError("no default model for engine %s because it isn't used in any trainable step" % self.name)
            elif len(trainableSteps) > 1:
                raise PluginError("no step provided to getDefaultModel(), but multiple steps use the engine %s" % self.name)
            step = trainableSteps[0]
        m += "." + re.sub("\s+", "_", step.stepName) + "_" + re.sub("\s+", "_", lCode)
        if os.path.isabs(m):
            return m
        else:
            return os.path.join(self.task.taskRoot, m)

    # Note the None will be a key. If there's only one model info, and it's
    # named, you should still retrieve it with None.
    
    def getModelInfo(self, configName = None, step = None):

        if not self.modelConfigs:
            return None

        configTuple = None
        if configName is None and (len(self.modelConfigs.keys()) == 1):
            # In Python 3, values() isn't indexable because it's a view.
            configTuple = list(self.modelConfigs.values())[0]
        else:
            try:
                configTuple = self.modelConfigs[configName]
            except KeyError:
                return None
        clsName, configName, buildSettings = configTuple
        return ModelInfo(self, step, clsName, configName, buildSettings)

    def instantiateStepImplementation(self, wrapperStep, workflowName = None,
                                      prettyName = None,
                                      createSettings = None,
                                      runSettings = None, uiSettings = None):
        if self.stepConfigSettings:
            creationDict = self.stepConfigSettings.copy()
        else:
            creationDict = {}
        if createSettings:
            creationDict.update(createSettings)
        # I need to untangle this later. Right now, this is how the
        # pretty name makes it down to the PluginStep.
        if prettyName:
            creationDict["pretty_name"] = prettyName
        
        return self.stepConfigClass(wrapperStep, self.task, workflowName or "<none>",
                                    runSettings = runSettings, uiSettings = uiSettings,
                                    **creationDict)

# These steps can exist by themselves, or as part of a workflow. The
# steps by themselves are instances, and they clone themselves for
# the workflows, I think.

# At what point do we instantiate the step class in the engine? It's
# gotta get the creation info. This info is passed when there's a
# workflow. My feeling is that it should instantiate one when it
# loads the engine, and create another when it clones for the workspace.
# When it clones, you can specialize the type (for the ones that support
# the type) and you can add a prettyName.

class BasicStep(object):

    def __init__(self, task, name, engineName, workflowName = None, **kw):
        self.task = task
        self.stepName = name
        # Only when it's in a workflow.
        self.prettyName = None
        self.engineName = engineName
        self.engine = None
        self._stepImplementation = None
        self._stepImplementationPopulated = False
        if engineName:
            try:
                self.engine = task.getEngines()[engineName]
            except KeyError:
                raise PluginError("engine %s unknown for step %s in task %s" % (engineName, name, task.name))
        self.workflowName = workflowName
        self._workflowInits = {}
        self._localInit(**kw)

    @property
    def stepImplementation(self):
        if not self._stepImplementationPopulated:
            self._stepImplementationPopulated = True
            if self.engine:
                if not self.workflowName:
                    self._stepImplementation = self.engine.instantiateStepImplementation(self)
                else:
                    self._stepImplementation = self.engine.instantiateStepImplementation(self, workflowName = self.workflowName,
                                                                                         prettyName = self.prettyName,
                                                                                         **self._workflowInits)
        return self._stepImplementation                

    def _localInit(self, **kw):
        pass

    def _localClone(self, templateStep, childType):
        pass

    # sType is available for all of them, but only some children support it.
    def cloneForWorkflow(self, wfName, createSettings, runSettings, uiSettings,
                         prettyName = None, sType = None):
        newStep = self.__class__(self.task, self.stepName, self.engineName, workflowName = wfName)
        newStep.prettyName = prettyName
        newStep._localClone(self, sType)
        newStep._workflowInits = {"createSettings": createSettings, "runSettings": runSettings, "uiSettings": uiSettings}
        return newStep

    def _categoriesToSets(self, categoriesOrSets):
        if categoriesOrSets is None:
            return set()
        return self.task.checkCategoriesOrSets(categoriesOrSets)

    #
    # Public utilities. Many of these have to be implemented
    # differently for each child type.
    #
    
    # In most cases, if an operation is done, it can't be redone.
    # However, in certain exceptional cases (redoable taggers) this
    # is not true. See ToolChain.py.
    
    def stepCanBeDone(self, annotSet):
        return True

    # Return True if you can determine that the step has been done,
    # independently of whether the phase is recorded.
    
    def getSetsDone(self, annotSet):
        return []

    def recordStep(self, doc):
        return

    def stepUndone(self, doc):
        return

    def getPhaseNames(self):
        return None

    def believedAtLeastPartiallyDone(self, iData):
        return False

    def believedCompletelyDone(self, iData):
        return False

    def getCGIMetadata(self):
        if self.stepImplementation:
            d = self.stepImplementation.getCGIMetadata()
        else:
            d = {}
        if "initSettings" not in d:
            d["initSettings"] = {"name": self.stepName}
            if self.prettyName:
                d["initSettings"]["pretty_name"] = self.prettyName
        return d

    # And applying the steps. If there's no step implementation,
    # what happens? Seems to me we need to leave this to the children, maybe.

    # language can always be among the kw.
    
    def do(self, annotSet, **kw):
        if self.stepImplementation:
            return self.stepImplementation.do(annotSet, **kw)
        else:
            return annotSet

    def undo(self, annotSet, **kw):
        if self.stepImplementation and hasattr(self.stepImplementation, "undo"):
            self.stepImplementation.undo(annotSet, **kw)

    # language can always be among the kw.
    
    def doBatch(self, iDataPairs, **kw):
        if self.stepImplementation:
            return self.stepImplementation.doBatch(iDataPairs, **kw)
        else:
            return iDataPairs

    # Only the annotation steps are trainable, and then only
    # if the engine is trainable and the step is auto-taggable.
    
    def isTrainable(self):
        return False

    def getDefaultModel(self, language = None):
        return None

    def getModelInfo(self, configName = None):
        return None

    def markGold(self, doc):
        raise PluginError("step can't be marked gold")

    def markReconciled(self, doc):
        raise PluginError("step can't be marked reconciled")

    # So the server key is always two elements: a "name"
    # and a "language". In jCarafe, the name will be a step and
    # the language will be the model language key. These are
    # passed to createServer if the key hasn't been found yet.
    
    def getServerKeys(self):
        if self.stepImplementation:
            return self.stepImplementation.getServerKeys()
        else:
            return None

    def createServer(self, serverName, serverLanguage):
        if self.stepImplementation:
            return self.stepImplementation.createServer(serverName, serverLanguage)
        else:
            raise TaggerConfigurationError
            
    def getVisibleName(self):
        return self.prettyName or self.stepName

class SignalStep(BasicStep):

    def _localInit(self, sType = None):
        if sType is None:
            sType = "auto"
        self.sType = sType

    def _localClone(self, templateStep, childType):
        if childType is not None:
            if childType != templateStep.sType:
                raise PluginError("Instantiating step %s in workflow %s of task %s where step type %s specified in workflow is not compatible with core step type %s" % (self.stepName, self.workflowName, self.task.name, childType, templateStep.sType))
        self._localInit(sType = childType or templateStep.sType)

    # Only difference from parent: can't undo.
    
    def undo(self, annotSet, **kw):
        raise PluginError("can't undo signal step")

    def getCGIMetadata(self):
        d = BasicStep.getCGIMetadata(self)
        d["stepType"] = "signal"
        return d

# Because of the way our tools work, zone and token annotations can't be
# added by hand.

class AnnotationStep(BasicStep):

    DEFAULT_ANNOTATOR = "unknown human"

    def _localInit(self, sType = None, setsRequired = None,
                   setsForbidden = None, setsModified = None, setsAdded = None,
                   setsDesired = None):
        # I have to make sure this exists, in order to collect _managedSets.
        self.task.getAnnotationTypeRepository()
        if sType is None:
            sType = "mixed"
        self.sType = sType
        self.setsRequired = self._categoriesToSets(setsRequired)
        self.setsForbidden = self._categoriesToSets(setsForbidden)
        self.setsModified = self._categoriesToSets(setsModified)
        self.setsAdded = self._categoriesToSets(setsAdded)
        self.setsDesired = self._categoriesToSets(setsDesired)
        self.managedSets = (self.setsAdded | self.setsModified) & self.task._managedSets
        # This is the set of things that are mentioned last
        # in the sequence in this step. By default, it's all the
        # added and modified steps. When workflows are constructed,
        # this will be filtered.
        self.finalSetMentions = self.setsAdded | self.setsModified
        for s in self.finalSetMentions:
            if (self.task._annotationSetsToCategories.get(s) in ("zone", "token")) and (sType != "auto"):
                raise PluginError("hand-annotatable step %s of task %s adds or modifies a set which is a zone or token set" % (self.stepName, self.task.name))
        # It's an error to mix managed and unmanaged sets - this
        # causes the wrong thing to happen in workspace autotag,
        # because retagging requires undo in the unmanaged case but
        # shouldn't in the managed case (because of the ultimate
        # possibility of partial annotation).
        if self.managedSets and (self.managedSets < self.setsAdded):
            raise PluginError("step %s in task %s adds a mix of managed and unmanaged sets" % (self.stepName, self.task.name))
        if self.setsModified & self.setsAdded:
            raise PluginError("step %s in task %s adds and modifies the same set" % (self.stepName, self.task.name))
        if self.setsRequired & self.setsForbidden:
            raise PluginError("step %s in task %s both requires and forbids the same set" % (self.stepName, self.task.name))

    def _localClone(self, templateStep, childType):
        if childType is not None:
            # mixed -> auto-with-correction, hand, auto
            # auto-with-correction -> auto
            # Nothing else works.
            if (childType != templateStep.sType) and (templateStep.sType != "mixed"):
                # mixed can be specialized to anything.
                if templateStep.sType == "auto-with-correction":
                    if childType != "auto":
                        raise PluginError("Instantiating step %s in workflow %s of task %s where step type %s specified in workflow is not compatible with core step type %s" % (self.stepName, self.workflowName, self.task.name, childType, templateStep.sType))

                else:
                    raise PluginError("Instantiating step %s in workflow %s of task %s where step type %s specified in workflow is not compatible with core step type %s" % (self.stepName, self.workflowName, self.task.name, childType, templateStep.sType))
        self._localInit(sType = childType or templateStep.sType,
                        setsRequired = templateStep.setsRequired,
                        setsForbidden = templateStep.setsForbidden,
                        setsModified = templateStep.setsModified,
                        setsAdded = templateStep.setsAdded,
                        setsDesired = templateStep.setsDesired)

    # In most cases, if an operation is done, it can't be redone.
    # However, in certain exceptional cases (redoable taggers) this
    # is not true. See ToolChain.py.

    # For right now, this just checks in ToolChain.py whether you can
    # apply the step or not, based on whether the added sets have
    # been added. We're not yet going to worry about the
    # preconditions - that comes later.

    # If the step is managed, make sure the SEGMENTs are there, and
    # make sure they don't contradict what the stepsDone says (just
    # in case).

    # The modified sets can affect stepCanBeDone in only two ways:
    # any modified sets that are managed and marked gold cause the
    # step not to be able to be done, and when we try to turn off
    # stepCanBeDone when we're checking if steps are attributed,
    # the modify sets shouldn't be able to do that.
    
    def stepCanBeDone(self, annotSet):
        stepCanBeDone = not self.setsAdded
        if not stepCanBeDone:
            # It can be done if the setsAdded aren't all in the steps done.
            stepCanBeDone = not (self.setsAdded <= set(annotSet.getSetsDone()))
        if self.managedSets:
            # First, see if there are SEGMENTs. If there are, see what they
            # tell you, and see if it says the step isn't done if it is.
            # If there are no SEGMENTs, add them appropriately.
            segs = [s for s in annotSet.orderAnnotations(["SEGMENT"])
                    if s.get("set") in self.managedSets]
            if segs:
                if not self.engine:
                    # No engine. It's a hand step, and for the moment,
                    # we're not supporting partial hand annotation. So
                    # just make sure that the segments all have an annotator.
                    if not stepCanBeDone:
                        for seg in segs:
                            if seg.get("annotator") is None:
                                seg["annotator"] = self.DEFAULT_ANNOTATOR
                else:
                    noneSegCount = 0
                    machineSegCount = 0
                    attributedSegCount = 0
                    attributedSegs = []
                    for seg in segs:
                        ann = seg.get("annotator")
                        if ann is None:
                            noneSegCount += 1
                        elif ann == "MACHINE":
                            machineSegCount += 1
                        else:
                            attributedSegCount += 1
                            attributedSegs.append(seg)
                    totalSegCount = len(segs)
                    
                    # First, let's make sure that the current setting for
                    # stepCanBeDone is consistent with the segment marking.
                    if (not stepCanBeDone) and (noneSegCount == totalSegCount):
                        # There are annotations, but none of the segments
                        # are attributed. Attribute all the segments. We should
                        # never be in this state, unless perhaps we're reading
                        # an old doc.
                        for seg in segs:
                            seg["annotator"] = self.DEFAULT_ANNOTATOR
                    elif stepCanBeDone and (attributedSegCount == totalSegCount):
                        # There are no annotations, but all the segments are
                        # attributed. If any of these are gold modify steps,
                        # the step can't be done. Otherwise, modify steps
                        # can't count toward this.
                        modifiedSegs = [seg for seg in attributedSegs if seg.get("set") in self.setsModified]
                        goldModifiedSegs = [seg for seg in modifiedSegs if seg.get("status") in ("human gold", "reconciled")]
                        if goldModifiedSegs or (attributedSegCount > len(modifiedSegs)):
                            stepCanBeDone = False
                    else:
                        # If the engine is retaggable, consider steps whose
                        # annotators are None or "MACHINE", otherwise, just
                        # the steps which are None. If the engine is partially taggable,
                        # at least one such segment must exist; if not,
                        # all of the segments must be like this.
                        if not stepCanBeDone:
                            if self.engine.isRetaggable():
                                retaggableSegCount = noneSegCount + machineSegCount                                
                            else:
                                retaggableSegCount = noneSegCount
                            if self.engine.isPartiallyTaggable() and (retaggableSegCount > 0):
                                stepCanBeDone = True
                            elif totalSegCount == retaggableSegCount:
                                # I doubt we'll ever get here.
                                stepCanBeDone = True                        
            else:
                # There are no segs. Make sure they're added.
                # If the step can't be done, it's because it's already done,
                # so we put a real annotator in. Otherwise, None.
                self._createSegments(annotSet, ((not stepCanBeDone) and self.DEFAULT_ANNOTATOR) or None, "non-gold")
                    
        return stepCanBeDone

    def _createSegments(self, annotSet, annotator, status):
        zones = annotSet.orderAnnotations(self.task.getAnnotationTypesByCategory("zone"))
        # If there are no zones, make a big annotation.
        for mSet in self.managedSets:
            if not zones:
                annotSet.createAnnotation(0, len(annotSet.signal), "SEGMENT",
                                          {"annotator": annotator,
                                           "set": mSet,
                                           "status": status})
            else:
                lastZone = None
                for zone in zones:
                    # Better not overlap, dammit.
                    if lastZone and lastZone.end > zone.start:
                        raise PluginError("Can't add SEGMENTs because zones overlap")

                    annotSet.createAnnotation(zone.start, zone.end, "SEGMENT",
                                              {"annotator": annotator,
                                               "set": mSet,
                                               "status": status})
                    lastZone = zone

    # Used in Workspace.py, during import, to ensure that any step that's
    # at least partially done has SEGMENTs.

    def _ensureSegments(self, annotSet, status, annotator):
        if self.managedSets:
            segs = [s for s in annotSet.orderAnnotations(["SEGMENT"])
                    if s.get("set") in self.managedSets]
            if not segs:
                self._createSegments(annotSet, annotator, status)

    # Return True if you can determine that the step has been done,
    # independently of whether the phase is recorded.
    
    def getSetsDone(self, annotSet):
        if self.setsAdded:
            return self.task.getAnnotationTypeRepository().filterSetsPresent(annotSet, self.setsAdded)
        else:
            return []

    def undo(self, annotSet, **kw):
        # Undo the basic step, which undoes the step implementation if there's a special undo.
        BasicStep.undo(self, annotSet, **kw)
        if self.setsAdded:
            repo = self.task.getAnnotationTypeRepository()
            labList, ignore, attrDict, ignore = repo.getAnnotationSetDetails(self.setsAdded, attributeDetails = True)
            if labList:
                annotSet.removeAnnotations(labList)
            if attrDict:
                # There's no good way of doing this.
                for lab, attrs in attrDict.items():
                    annots = annotSet.getAnnotations([lab])
                    for attr in attrs:
                        attrObj = None
                        for annot in annots:
                            # This has the unfortunate side effect of
                            # extending the attr list, even if it's not
                            # currently set, UNLESS I check the presence of a value first.
                            # NOTE: because the initial creation sets defaults, whether
                            # or not the attributes can be set in the step which
                            # creates the annotation, any defaults should be
                            # set here, too.
                            if annot.get(attr) is not None:
                                if annot.atype.hasDefaults:
                                    if attrObj is None:
                                        attrObj = annot.atype.attr_list[annot.atype.attr_table[attr]]
                                    if attrObj.hasDefault:
                                        annot[attr] = attrObj._getAttributeDefault(annot)
                                    else:
                                        annot[attr] = None
                                else:
                                    annot[attr] = None
        if self.managedSets:            
            # Make sure to unmark the SEGMENTs. Really,
            # they should be removed entirely.
            # Otherwise, I would have done this:
            # annot["status"] = "non-gold"
            # annot["annotator"] = None
            # Actually, we want to remove the setsAdded ones, and mark the
            # others non-gold.
            segs = [s for s in annotSet.getAnnotations(["SEGMENT"])
                    if s.get("set") in self.managedSets]
            if self.setsModified:
                for s in segs:
                    if s["set"] in self.setsModified:
                        s["status"] = "non-gold"
            if self.setsAdded:
                annotSet.removeAnnotationGroup([s for s in segs if s["set"] in self.setsAdded])

    def recordStep(self, doc):
        if self.setsAdded:
            for c in self.setsAdded:
                doc.recordSet(c)

    def stepUndone(self, doc):
        if self.setsAdded:
            for c in self.setsAdded:
                doc.setUndone(c)

    def getPhaseNames(self):
        if self.setsAdded:
            return self.setsAdded
        return None

    # If there are setsModified, it may be partially done
    # if some of those modified sets are managed and gold.
    # This has to be true if believedCompletelyDone can be true. 
    
    def believedAtLeastPartiallyDone(self, iData):
        if self.setsModified & self.managedSets:
            for s in iData.getAnnotations(["SEGMENT"]):
                if (s.get("set") in self.setsModified) and (s.get("status") in ("human gold", "reconciled")):
                    return True
        if not self.setsAdded:
            return False
        else:
            # If there are setsAdded, if some of them are done (started)
            # the step is partially done.
            return self.setsAdded & set(iData.getSetsDone())

    # If all of the modify sets are managed and gold,
    # that should mark it as completely done, too.
    
    def believedCompletelyDone(self, iData):
        if not self.setsAdded:
            # So it can't be if there are no sets added, unless
            # it's got modify steps and they're all managed and gold.
            if self.setsModified and (self.setsModified == self.managedSets):
                segs = [s for s in iData.getAnnotations(["SEGMENT"])
                        if s.get("set") in self.setsModified]
                if not segs:
                    return False
                else:
                    return set([s.get("status") for s in segs]) <= set(["human gold", "reconciled"])
            else:
                return False
        elif not (self.setsAdded <= set(iData.getSetsDone())):
            # And if the sets added aren't in the sets done list,
            # nope.
            return False
        elif self.sType == "auto":
            # If it's an auto step, and the sets added are all there,
            # it's done.
            return True
        elif not self.managedSets:
            # If there are no managed sets, it's not done, since
            # there's still the chance of adding hand annotations.
            return False
        else:
            # There have to be segments, and they all have to be
            # human gold or reconciled, including the modify steps.
            segs = [s for s in iData.getAnnotations(["SEGMENT"])
                    if s.get("set") in self.managedSets]
            if not segs:
                return False
            else:
                return set([s.get("status") for s in segs]) <= set(["human gold", "reconciled"])
        
    def do(self, annotSet, **kw):
        if self.sType == "hand":
            raise PluginError("can't apply hand step")
        annotSet = BasicStep.do(self, annotSet, **kw)
        if self.managedSets:
            # Make sure to unmark the SEGMENTs.
            for annot in [s for s in annotSet.getAnnotations(["SEGMENT"])
                          if s.get("set") in self.managedSets]:
                if annot.get("annotator") is None:
                    annot["annotator"] = "MACHINE"
        return annotSet

    def doBatch(self, iDataPairs, **kw):
        if self.sType == "hand":
            raise PluginError("can't apply hand step")
        iDataPairs = BasicStep.doBatch(self, iDataPairs, **kw)
        if self.managedSets:
            for fname, annotSet in iDataPairs:
                for annot in [s for s in annotSet.getAnnotations(["SEGMENT"])
                              if s.get("set") in self.managedSets]:
                    if annot.get("annotator") is None:
                        annot["annotator"] = "MACHINE"
        return iDataPairs

    def isTrainable(self):
        return self.engine and self.engine.isTrainable() and self.sType != "hand"

    def getDefaultModel(self, language = None):
        return self.engine.getDefaultModel(step = self, language = language)

    def getModelInfo(self, configName = None):
        return self.engine.getModelInfo(configName = configName, step = self)
    
    def markGold(self, doc):
        self.markCompleted(doc, self.DEFAULT_ANNOTATOR, "human gold")

    def markReconciled(self, doc):
        self.markCompleted(doc, "GOLD_STANDARD", "reconciled")

    # We can only do this on the final step that mentions this
    # step. Otherwise, it should silently fail.

    def markCompleted(self, annotSet, annotator, status):
        # Nothing should happen unless this step is managed.
        # I think that otherwise, it'll silently fail for the moment.
        # These segments will already be present, because stepCanBeDone
        # has been called on each of the docs - this is
        # only called from ToolChain.py.
        # Oops, false. stepCanBeDone is only called if the
        # element is in the steps list.
        if self.managedSets:
            updatableSets = self.managedSets & self.finalSetMentions
            otherSets = self.managedSets - updatableSets
            # Make sure to unmark the SEGMENTs.
            if updatableSets:
                self._markStatus(annotSet, updatableSets, annotator, status)
            if otherSets:
                self._markStatus(annotSet, updatableSets, annotator, "non-gold")

    def _markStatus(self, annotSet, managedSets, annotator, status):
        segs = [s for s in annotSet.getAnnotations(["SEGMENT"])
                if s.get("set") in managedSets]
        if segs:
            for annot in segs:
                annot["annotator"] = annotator
                annot["status"] = status
        else:
            self._createSegments(annotSet, annotator, status)

    def getCGIMetadata(self):
        d = BasicStep.getCGIMetadata(self)
        d["stepType"] = "annotation"
        d["annotationStepInfo"] = {
            "sType": self.sType,
            "setsRequired": list(self.setsRequired),
            "setsForbidden": list(self.setsForbidden),
            "setsModified": list(self.setsModified),
            "setsAdded": list(self.setsAdded),
            "setsDesired": list(self.setsDesired),
            "finalSetMentions": list(self.finalSetMentions)
            }
        return d
    
class TransformStep(BasicStep):

    def _localInit(self, setsRequired = None,
                   setsForbidden = None, setsDesired = None):
        self.setsRequired = self._categoriesToSets(setsRequired)
        self.setsForbidden = self._categoriesToSets(setsForbidden)
        self.setsDesired = self._categoriesToSets(setsDesired)

    def _localClone(self, templateStep, childType):
        if childType is not None:
            raise PluginError("transform step %s in workflow %s of task %s does not support step types" % (self.stepName, self.workflowName, self.task.name))
        self._localInit(setsRequired = templateStep.setsRequired,
                        setsForbidden = templateStep.setsForbidden,
                        setsDesired = templateStep.setsDesired)

    def undo(self, annotSet, **kw):
        raise PluginError("Can't undo transform step")
    
    def getCGIMetadata(self):
        d = BasicStep.getCGIMetadata(self)
        d["stepType"] = "transform"
        return d

# In MAT 2.0, stepList was a list of PluginStep objects. Now, it's a list of
# BasicStep objects (or subclasses). 

class PluginWorkflow(object):

    def __init__(self, name, descriptor, steps, uiSettings = None, **params):
        self.descriptor = descriptor
        self.name = name
        self.stepList = steps
        self.uiSettings = uiSettings
        # These are the DEFINED settings.
        self.params = params
        # Loop through the steps backward. Accumulate the
        # final set mentions and filter them as we go along.
        finalSetMentions = set()
        for i in range(len(self.stepList) - 1, -1, -1):            
            thisStep = self.stepList[i]
            if isinstance(thisStep, AnnotationStep):
                thisStep.finalSetMentions -= finalSetMentions
                finalSetMentions |= thisStep.finalSetMentions
        self._stepNamesToSteps = {}
        for s in self.stepList:
            self._stepNamesToSteps[s.stepName] = s
            if s.prettyName:
                self._stepNamesToSteps[s.prettyName] = s                

    # This replaces the global step ordering. Raise a
    # PluginError if the step isn't in't the workflow.
    # The name can be the step name or the step pretty name.
    
    def getStepAndSuccessors(self, name):
        # It may not even have the step.
        s = self._stepNamesToSteps.get(name)
        if not s:
            raise PluginError("workflow does not have step %s" % name)
        else:
            # Return steps including this step.
            return self.stepList[self.stepList.index(s):]

    def getStepByName(self, name):
        return self._stepNamesToSteps.get(name)

    def workspaceEligible(self):
        # A workflow is workspace eligible if it has
        # at least one human-mediated step.
        for step in self.stepList:
            if isinstance(step, AnnotationStep) and (step.sType != "auto"):
                return True
        return False

    def languageEligible(self, lg):
        for step in self.stepList:
            # If the step has an engine which is language-restricted, barf.
            if step.engine and (not step.engine.languageEligible(lg)):
                return False
        return True

    def getTrainableSteps(self):
        return dict([(s.stepName, s) for s in self.stepList if s.isTrainable()])

# The task steps are objects. I'm going to subclass a
# bunch of them here.

class PluginStep(Operation):

    # These are CLASS variables, so that you can ask the engine,
    # rather than having to wait for the class to be implemented.

    retaggable = False
    partiallyTaggable = False
    
    def __init__(self, wrapperStep, descriptor, workflow, runSettings = None,
                 uiSettings = None, **initSettings):
        stepName = wrapperStep.stepName
        Operation.__init__(self)
        self.descriptor = descriptor
        self.workflow = workflow
        self.stepName = stepName
        self.wrapperStep = wrapperStep
        self.initSettings = initSettings
        initSettings["name"] = stepName
        self.runSettings = runSettings or {}
        self.uiSettings = uiSettings or {}

    #
    # These are the main functions, the ones which really
    # need to be reimplemented for every child type. Actually,
    # it's only do() that really needs it, but if you have
    # special batch behavior, you can override doBatch.    
    #
    
    def do(self, annotSet, **kw):
        raise PluginError("not implemented")

    def doBatch(self, iDataPairs, **kw):
        # By default, loop through the iDataPairs and
        # call the do() step.
        pairsDone = []
        for fname, iData in iDataPairs:
            # I'd like to get the file into the error in
            # batch, too, but I'm not sure how to do that consistently.
            try:
                iData = self.do(iData, **kw)
            except Exception as e:
                if ExecutionContext._DEBUG:
                    raise
                else:
                    raise Error.MATError(self.stepName, str(e),
                                         show_tb = True, file = fname)
            pairsDone.append((fname, iData))
        return pairsDone

    # Utilities.

    def getCGIMetadata(self):
        # This should contain the name, the pretty name, and
        # whatever settings are present. Basically, we should
        # return the entire params dictionary, plus pretty_name.
        return {"initSettings": self.initSettings,
                "runSettings": self.runSettings,
                "uiSettings": self.uiSettings}

    def paramsSatisfactory(self, wfName, failureReasons, **params):
        return True

    def removeAnnotationsByCategory(self, annotSet, *cats):
        for cat in cats:
            annotSet.removeAnnotations(self.descriptor.getAnnotationTypesByCategory(cat))

    def getServerKeys(self):
        return None

    def createServer(self, serverName, serverLanguage):
        raise TaggerConfigurationError

    def createTaggerServices(self, **kw):
        raise TaggerConfigurationError

# The clean step is to make the document processable. The clean
# step isn't necessarily undoable - it can modify the signal.

class CleanStep(PluginStep):

    def undo(self, annotSet, **kw):
        raise PluginError("can't undo clean step")

    def truncateToUnixAscii(self, annotSet):
        # The guts of the important cleaning step. We remove
        # everything that isn't ASCII, and we also remove
        # some other things we recognize, such as vertical tab
        # and form feed.

        # Convert \r\n to \n.
        annotSet.signal = annotSet.signal.replace("\r\n", "\n")
        # And finally, remove all the non-breaking spaces -
        # they're latin1, but not ascii, and at the moment,
        # the tokenizer barfs on them. Actually, it's everything
        # in latin1 that has to go. 
        annotSet.signal = annotSet.signal.replace(u"\xa0", " ")
        annotSet.signal = annotSet.signal.replace(u"\x92", "'")
        # Replace formfeed with newline.
        annotSet.signal = annotSet.signal.replace("\x0c", "\n")
        # Replace vertical tab with newline.
        annotSet.signal = annotSet.signal.replace("\x0b", "\n")
        # Now, replace everything else with ?. But make it
        # Unicode again.
        annotSet.signal = annotSet.signal.encode('ascii', 'replace').decode('ascii')
        return annotSet

class AlignStep(PluginStep):

    args = OptionTemplate([OpArgument("ensure_gold", help = "mark the specified step(s) gold", hasArg = True),
                           OpArgument("ignore_sets", help = "ignore the specified annotation sets or categories while aligning",
                                      hasArg = True),
                           OpArgument("verbose", help = "prints a detailed summary of the alignment decisions made"),
                           OpArgument("half_verbose", help = "prints a less detailed summary of the alignment decisions made")])

    def doBatch(self, iDataPairs, ensure_gold = None, **kw):
        iDataPairs = PluginStep.doBatch(self, iDataPairs, **kw)
        if ensure_gold:
            if type(ensure_gold) in _strtypes:
                ensure_gold = [s.strip() for s in ensure_gold.split()]
                if ensure_gold == ['']:
                    ensure_gold = []
            for step in ensure_gold:
                try:
                    step = self.descriptor.getSteps()[step]
                except KeyError:
                    raise PluginError("Step '%s' in --ensure_gold unknown (must be a true step name, not a pretty name)" % step)
                for fname, d in iDataPairs:
                    step.markGold(d)
        return iDataPairs

    def do(self, annotSet, verbose = False, half_verbose = False, ignore_sets = None, **kw):
        annotSet.adjustTagsToTokens(self.descriptor, doReport = verbose or half_verbose,
                                    suppressReportDetails = half_verbose,
                                    ignoreSets = ignore_sets)
        return annotSet

    def undo(self, annotSet, **kw):
        raise PluginError("can't undo align step")

# The tokenization step. Soon, this will be a general
# command line step.

class CmdlinePluginStepMixin(object):

    pass

class CmdlinePluginStep(PluginStep, CmdlinePluginStepMixin):
    pass

class TokenizationStep(PluginStep):

    pass

class CmdlineTokenizationStep(TokenizationStep, CmdlinePluginStepMixin):

    pass

# General zoning.

class ZoneStep(PluginStep):

    def addZones(self, annotSet, tuples, **kw):
        # I don't do anything with the kw - just so we can pass it down without
        # having to predict what flags are needed.

        zType, rAttr, regionTypes = self.descriptor.getTrueZoneInfo()
        
        # Everything that isn't in a zone is untaggable. But we no
        # longer record that, except in the UI.
        # tuples are (start, end, rtype). Sort them
        # by start index.
        d = {}
        for t in tuples:
            if t[0] in d:
                raise PluginError("multiple regions with the same start index")
            d[t[0]] = t
        keys = sorted(d.keys())
        i = 0
        signalLen = len(annotSet.signal)
        for k in keys:
            start, end, rtype = d[k]
            if end < start:
                raise PluginError("region end is before region start")
            # We've overshot. Not sure how, but just drop the regions.
            if start > signalLen:
                break
            # We've overshot the end. Just truncate.
            if end > signalLen:
                end = signalLen
            if start < i:
                raise PluginError("region starting at %d is before previous end" % i)
            elif start > i:
                # Here, we would have created an untaggable annotation, but no longer.
                # Also at the end.
                pass
            if rAttr and rtype and (rtype in regionTypes):
                # When I create the annotation, I can only use the list
                # if I'm absolutely sure that the type has the rAttr.
                # If the doc happened not to be instantiated with the
                # task's annotation type repository, then we're
                # hosed. Use the dictionary API to be sure.
                annotSet.createAnnotation(start, end, zType, {rAttr: rtype})
            else:
                annotSet.createAnnotation(start, end, zType)
            i = end

class WholeZoneStep(ZoneStep):

    def do(self, annotSet, **kw):

        # Mark everything. Use the region info if available.
        zType, rAttr, regionTypes = self.descriptor.getTrueZoneInfo()
        rType = None
        if regionTypes:
            rType = regionTypes[0]
        
        self.addZones(annotSet, [(0, len(annotSet.signal), rType)],
                      **kw)
        
        return annotSet

class TagStep(PluginStep):
    
    def do(self, annotSet, language = None, **kw):
        raise PluginError("undefined tag step")
    
    # This was buried way down in CarafeTagger, in multiple places, and if I
    # want to play with pretagging, I have to have better control over this.
    # And since I now have a complex way of managing models and languages
    # in my tasks, I can't have this stuck in JavaCarafe.py, either.

    # If we're looking at choosing models for the server, we can't fail
    # if there are multiple models available and no language specified.
    
    def findModelTriple(self, tagger_model, language, failOnMultiple = True):
        task = self.descriptor
        wrapperStep = self.wrapperStep
        # Triple of model, language address, isDefault.
        modelTriples = []
        if tagger_model is None:
            if language:
                langs = [language]
            else:
                # Use list() to freeze dictionary keys in Python 3
                langs = list(task.getLanguages().keys())
            # See if the first language has a default model. If it does, then
            # all of them will have one. 
            defaultModel = wrapperStep.getDefaultModel(language = langs[0])
            if defaultModel:
                modelTriples.append((defaultModel, langs[0], True))
                if len(langs) > 0:
                    modelTriples += [(wrapperStep.getDefaultModel(language = l), l, True) for l in langs[1:]]
        else:
            # The model is already there. I need to specify the languageAddress.
            if language is None:
                # Grab all the languages.
                # If the tagger model was passed in, the language may still
                # be None. In that case, when I set up the services, I'll
                # need to make sure that None is filled by every language.
                langs = sorted(wrapperStep.engine.languages or task.getLanguages().keys())
                languageAddress = ",".join(langs)
            else:
                languageAddress = task.getLanguages()[language]["name"]
            modelTriples.append((tagger_model, languageAddress, False))

        # OK, we have the model triples. Now, we need the chosen triple.

        # At this point, we have to resolve the model and language. We're either
        # about to invoke a local step, or a proxy, and in either case, we
        # need to know what the address is.

        if len(modelTriples) == 0:
            return modelTriples, None, "no model specified"
        elif len(modelTriples) > 1:
            if failOnMultiple:
                return modelTriples, None, "default model requested, but language not provided"
            else:
                finalTriples = []
                for model, lg, isDefault in modelTriples:
                    modelPath = self.modelPathExists(model)
                    if modelPath:
                        finalTriples.append((modelPath, lg, isDefault))
                if finalTriples:
                    return finalTriples, None, None
                else:
                    return [], None, "none of the specified models were found"
        else:
            model, lg, isDefault = modelTriples[0]
            modelPath = self.modelPathExists(model)
            if modelPath:
                return [(modelPath, lg, isDefault)], (modelPath, lg, isDefault), None
            elif isDefault:
                return modelTriples, None, "default model specified, but doesn't exist yet"
            else:
                return modelTriples, None, "requested model file not found"    

    def modelPathExists(self, model):
        if os.path.exists(model):
            return model
        else:
            return None    

class CmdlineTagStep(TagStep):

    args = OptionTemplate([OpArgument("local", help = "don't try to contact a remote tagger server; rather, start up a local command."),
                           OpArgument("model", help = "provide a tagger model file. Obligatory if no model is specified in the task step.", hasArg = True)], use_activator_value_prefix = True)

    # Support for finding servers.

    def getServerKeys(self):
        # Remember, this is MATWeb, so the only possible settings for
        # model and locality come from the run settings.
        local = self.runSettings.get("local", False)
        model = self.runSettings.get("model", None)
        if local:
            # If the run settings say it's always local, don't
            # return any keys.
            return None
        else:
            # There was a bug in JavaCarafe which led to a failure to start up
            # servers when the step supported multiple languages. I fixed that, and I
            # should move all that here anyway. Or, more to the point, I should
            # be doing something with the available languages. Not all the languages
            # will be supported by the engine, so we'll have more of a filter later.
            return self.getServerKeysForStep()

    def getServerKeysForStep(self):
        # Here's the default.
        return [(self.stepName, lg) for lg in self.descriptor.getLanguages()]

    def createServer(self, serverName, lg):
        raise TaggerConfigurationError
        
# First, we need to load all the classes, and then we can
# load the task files. 

# The Plugins directory will now
# contain subdirectories, each representing a task. The
# structure of each subdirectory is as described in sample_app/README.
# the otherDirs should be similar directories.

# The plugin dictionary will be a mapping from the named
# tasks to their objects (remember, the named tasks are the
# ones which are visible to the UIs). It also implements
# toplevel collection of information, especially in those
# cases where the information is only needed at the toplevel.

class PluginDict(dict):

    def __init__(self):
        self._cssCache = self._truncatedCssCache = self._jsCache = self._truncatedJsCache = None
        self._cgiMetadataCache = None
        self._cmdlineMetadataCache = None
        self.rootTask = PluginTaskDescriptor("<root>", None)
        self.rootTask.visible = False        
        # The task is the root task. It already has SEGMENT and VOTE in it.
        # First, the annotation set.
        self.rootTask.localAnnotationSetDescriptors["descriptors"] += [{"category": "token",
                                                                        "name": "token",
                                                                        "annotations":
                                                                        [{"label": "lex",
                                                                          "all_attributes_known": True,
                                                                          "span": True}],
                                                                        "attributes": []},
                                                                       {"category": "zone",
                                                                        "name": "zone",
                                                                        "annotations":
                                                                        [{"label": "zone",
                                                                          "all_attributes_known": True,
                                                                          "span": True}],
                                                                        "attributes":
                                                                        [{"name": "region_type",
                                                                          "type": "string",
                                                                          "of_annotations": ["zone"],
                                                                          "default": "body",
                                                                          "choices": ["body"]}]}]
        self.rootTask.localAnnotationDisplays["labels"].update({"lex": {"name": "lex", "css": "border: 1px solid #CCCCCC"}})
        # I need this because the frontend filters on what's in the
        # tag order list.
        self.rootTask.localAnnotationDisplays["order"] += ["lex", "zone"]
        self.byDir = {}

    # Only the visible tasks will be recorded. The task root will
    # never be recorded, then. It's the only task which is permitted
    # not to have a value for webDir.
    def recordTaskDir(self, task):
        if task.webDir in self.byDir:
            fullDir, curTasks = self.byDir[task.webDir]
            if task.taskRoot != fullDir:
                raise PluginError("Task basename '%s' matches basename in a different directory (%s)" % (task.webDir, fullDir))
            self.byDir[task.webDir][1].append(task)
        else:
            self.byDir[task.webDir] = [task.taskRoot, [task]]

    def getRecorded(self, webDir):
        try:
            return self.byDir[webDir]
        except KeyError:
            return None

    def getRootTask(self):
        return self.rootTask

    def getTask(self, tName):
        try:
            return self[tName]
        except KeyError:
            return None

    def getAllTasks(self):
        # Use list() to freeze dictionary values in Python 3.
        return list(self.values())

    def pruneTask(self, task):
        # Don't prune the root task.
        if task is self.rootTask:
            raise PluginError("Can't prune the root task")
        if task.visible:
            try:
                del self[task.name]
            except KeyError:
                pass
        if task.parentObj and task in task.parentObj.children:
            task.parentObj.children.remove(task)

        # Prune it from the byDir list, too.
        if task in self.byDir[task.webDir][1]:
            self.byDir[task.webDir][1].remove(task)

    #
    # Web customizations
    #

    def getCSSFiles(self, truncated = False):
        if self._cssCache is None:
            # recurse on the plugins.
            self._cssCache = []
            self._jsCache = []
            self._truncatedCssCache = []
            self._truncatedJsCache = []
            self._collectWebFiles(self.rootTask)            
        if truncated:
            return self._truncatedCssCache
        else:
            return self._cssCache

    def getJSFiles(self, truncated = False):
        if self._jsCache is None:
            # recurse on the plugins.
            self._cssCache = []
            self._jsCache = []
            self._truncatedCssCache = []
            self._truncatedJsCache = []
            self._collectWebFiles(self.rootTask)
        if truncated:
            return self._truncatedJsCache
        else:
            return self._jsCache

    def _collectWebFiles(self, c):
        for a in c.localCSSFiles:
            p = os.path.join(c.taskRoot, a)
            if p not in self._cssCache:
                self._cssCache.append(p)
            p = os.path.join(os.path.basename(c.taskRoot), a)
            if p not in self._truncatedCssCache:
                self._truncatedCssCache.append(p)
        for a in c.localJSFiles:
            p = os.path.join(c.taskRoot, a)
            if p not in self._jsCache:
                self._jsCache.append(p)
            p = os.path.join(os.path.basename(c.taskRoot), a)
            if p not in self._truncatedJsCache:
                self._truncatedJsCache.append(p)
        for child in c.children:
            self._collectWebFiles(child)

    # CGI metadata is a hash from workflow names to
    # an element consisting of the display config, steps,
    # and other metadata. Within the scope of a named
    # element, you can't reuse a workflow; the mapping
    # must be unique. It's a little kludgy, because
    # getWorkflows() climbs back up the tree, but I
    # can live with that, because I don't know when
    # I'll be calling the CGI fetcher.

    def getCGIMetadata(self):
        if self._cgiMetadataCache is None:
            self._cgiMetadataCache = {}
            self._compileCGIMetadata(self.rootTask, None, None, None)
        return self._cgiMetadataCache

    def _compileCGIMetadata(self, c, visibleScope, accumulatedMetadata, displayConfig):

        # Reset accumulatedMetadata within each visible element.
        # Check to see, when you add a workflow, whether it's already
        # in accumulatedWorkflows. But the restriction only
        # holds for the leaves, which is the only place I'm going
        # to collect results.

        # We ALSO want to collect workspace operations, because
        # we may need the settings in the UI. Arrgh. Also, I must
        # forbid an invisible leaf child from resetting the display config.
        
        if visibleScope is None and c.visible:
            visibleScope = c.name
            accumulatedMetadata = {"workflows": {}, "steps": {}}
            displayConfig = c.getDisplayConfig()

        if c.visible:
            # For the moment.
            languages = c.getLanguages().values()
            similarityProfiles = c.similarityProfiles
            if similarityProfiles is not None:
                sorter = TaskAnnotationTypeSorter(c)
                similarityProfiles = [c._compileSimilarityProfile(profile, sorter) for profile in similarityProfiles]
            self._cgiMetadataCache[c.name] = {"workflows": accumulatedMetadata["workflows"],
                                              "steps": accumulatedMetadata["steps"],
                                              "setsToCategories": c._annotationSetsToCategories,
                                              "shortName": c.getShortWebName(),
                                              "longName": c.getLongWebName(),
                                              "displayConfig": displayConfig,
                                              "annotationSetRepository": c.getJSONAnnotationRepositoryForDisplay(),
                                              "tagOrder": c.getAnnotationDisplayInfo()["order"],
                                              "overlapOrder": c.getAnnotationDisplayInfo()["overlapOrder"],
                                              "extraGestures": c.getAnnotationDisplayInfo()["gestures"],
                                              "alphabetizeLabels": c._inheritValue("alphabetizeUILabels",
                                                                                   default = False),
                                              "languages": dict([(l["name"], l) for l in languages]),
                                              "similarityProfiles": similarityProfiles,
                                              "tagHierarchy": c.getTagHierarchy(),
                                              "settings": c.settings}
            c.enhanceCGIMetadata(self._cgiMetadataCache[c.name])
        
        if not c.children:
            # It's a leaf. But it may not be visible, because expected
            # children weren't loaded.
            if c.visible and (c.getDisplayConfig() != displayConfig):
                raise PluginError("display config reset within the scope of task '%s'" % visibleScope)
            if accumulatedMetadata is not None:
                acWf = accumulatedMetadata["workflows"]
                for wf, wfObj in c.getWorkflows().items():
                    if wf in acWf:
                        raise PluginError("workflow '%s' found more than once in the scope of task '%s'" % (wf, visibleScope))
                    acWf[wf] = {"steps": [step.getCGIMetadata() for step in wfObj.stepList],
                                "workflowData": c.getCGIWorkflowMetadata(wfObj)}
                acStep = accumulatedMetadata["steps"]
                for sName, step in c.getSteps().items():
                    if sName in acStep:
                        raise PluginError("step '%s' found more than once in the scope of task '%s'" % (sName, visibleScope))
                    acStep[sName] = step.getCGIMetadata()
                    
        else:
            # It's not a leaf.
            for child in c.children:
                self._compileCGIMetadata(child, visibleScope, accumulatedMetadata, displayConfig)

    # Cmdline metadata. Ignore the root - it will never
    # have anything.

    def formatCmdlineMetadata(self, task):
        return "\n".join(self._compileCmdlineMetadata(task)[0])

    def _compileCmdlineMetadata(self, c, inName = False, configCounter = 1):
        # We don't start collecting stuff until c is
        # names.
        l = []
        localName = False
        if not inName and c.visible:
            inName = True
            l += [c.name + " :"]
            # Reset the config counter.
            configCounter = 1
            localName = True
        if c.children:
            # Not a leaf.
            for child in c.children:
                lIncr, configCounter = self._compileCmdlineMetadata(child, inName, configCounter)
                l += lIncr
        else:
            # Leaf. If the name was local, don't print out the
            # configuration line or increment the counter.
            if localName:
                indent = ""
            else:
                l.append("  configuration %d (%s):" % (configCounter, c.name))
                configCounter += 1
                indent = "  "
            l.append(indent + "  available workflows:")
            for  k, v in c.getWorkflows().items():
                l.append(indent + "    " + k + " : steps " +
                         ", ".join([x.stepName for x in v.stepList]))
            for incr in c.getCmdlineTaskMetadata():
                l.append(indent + incr)
            s1 = []
        return l, configCounter

# So things can be loaded from two places: the _PluginNamespace,
# which is populated by LoadPlugins, and from MAT. The problem is
# that MAT isn't in the global namespace anymore, AND there
# are two potential sources, because the MAT library is
# being split between src/MAT/lib/mat/python/MAT and
# src/MAT/lib/mat/python/MATApplication. I've addressed this latter problem by
# monkeypatching the MATApplication.JavaCarafe subpackage into the MAT
# package in the __init__.py of the MATApplication class. There's just
# too many tasks out there relying on this to change the name globally, and
# in the future, I may change the way jCarafe is distributed anyway.

_PluginNamespace = {}

def FindPluginClass(cName, tName, tpe = "class", delayedPackages = None, parentClass = None):
    
    def _clsCheck(v):
        import inspect
        if not inspect.isclass(v):
            # It's not a class
            raise PluginError("found element in plugin named %s, but it's not a class" % cName)
        if (parentClass is not None) and (not issubclass(v, parentClass)):
            raise PluginError("task %s specifies %s %s which is not a subclass of %s" % (tName, tpe, cName, parentClass.__name__))            

    return FindPluginObject(cName, tName, tpe = tpe,
                            delayedPackages = delayedPackages,
                            wfcChecker = _clsCheck)

# I've isolated the lookup problem described above
# in this function. I need to handle "MAT" specially.

def FindPluginObject(cName, tName, tpe = "object", delayedPackages = None, wfcChecker = None, callableType = True):
    global _PluginNamespace
        
    try:
        if cName.startswith("MAT."):
            # This will always work
            import MAT
            o = eval(cName)
        else:
            o = eval(cName, globals(), _PluginNamespace)
        if wfcChecker is not None:
            wfcChecker(o)
        return o
    except (NameError, AttributeError):
        if (delayedPackages is not None) and (cName.split(".")[0] in delayedPackages):
            return LazyCallable(cName, tName, tpe, wfcChecker)
        else:
            raise PluginError("can't find plugin %s named %s for %s" % (tpe, cName, tName))

# Because of the issue with JavaCarafe possibly not being
# defined, but wanting to use tasks for a subset of the things
# the task can do (e.g., annotations but not workflows), I'm introducing
# the notion of a lazy class

class LazyCallable(object):

    def __init__(self, clsName, taskName, classType, wfcChecker):
        self.clsName = clsName
        self.clsType = classType
        self.taskName = taskName
        self.wfcChecker = wfcChecker

    def __call__(self, *args, **kw):
        # This is equivalent to calling the class.
        return FindPluginObject(self.clsName, self.taskName, tpe = self.clsType, wfcChecker = self.wfcChecker)(*args, **kw)

    def __getattr__(self, attr):
        return getattr(FindPluginObject(self.clsName, self.taskName, tpe = self.clsType, wfcChecker = self.wfcChecker), attr)

#
# Plugin directory manager.
#

# The lazyClassLoad is to deal with situations where we don't want the
# classes in the task to be immediately located.

class PluginLocationMgr(object):

    def __init__(self, lazyClassLoad = False):

        self.matDir = os.path.dirname(os.path.abspath(__file__))
        self.matLoc = os.path.dirname(self.matDir)
        self.pluginRecord = self.pluginDir = None
        self.dirPairs = []
        self.webDirs = {}
        self.lazyClassLoad = lazyClassLoad

    # The plugin record isn't always used for the location manager.
    def _ensurePluginRecord(self):
        if self.pluginRecord is None:
            pluginRecord = MATConfig.get("MAT_PLUGIN_RECORD")
            if not pluginRecord:
                raise PluginError("No plugin record declared; add to MAT config using the MAT_PLUGIN_RECORD key")
            self.pluginRecord = pluginRecord
            self.pluginDir = os.path.dirname(self.pluginRecord)        

    #
    # core utilities
    #

    def pluginRecordExists(self):
        self._ensurePluginRecord()
        return os.path.exists(self.pluginRecord)

    def createPluginRecordDir(self, verbose = False):
        self._ensurePluginRecord()
        if not os.path.isdir(self.pluginDir):
            if verbose:
                print("### Creating", self.pluginDir, "...")
            os.makedirs(self.pluginDir)

    def readPluginRecord(self, verbose = False):
        # Create an ordered list, and a hash.
        self.dirPairs = []
        self.webDirs = {}
        if self.pluginRecordExists():
            if verbose:
                print("### Reading plugin record...")
            fp = codecs.open(self.pluginRecord, "r", "utf-8")
            for line in [line.strip() for line in fp.readlines()]:
                if not line:
                    continue
                elif line[0] == "#":
                    # Remove all the characters at the beginning which
                    # are comments, and then strip again.
                    while line[0] == "#":
                        line = line[1:].strip()
                    self.dirPairs.append(["#", line])
                else:
                    self.dirPairs.append(["", line])
                    self.webDirs[os.path.basename(line)] = line
            fp.close()

    def writePluginRecord(self, verbose = False):
        if not self.pluginRecordExists():
            self.createPluginRecordDir(verbose = verbose)
        if verbose:
            print("### Writing plugin record...")
        fp = codecs.open(self.pluginRecord, "w", "utf-8")
        for prefix, line in self.dirPairs:
            fp.write(prefix + line + "\n")
        fp.close()

    def add(self, appDir, verbose = False, testOnly = False):
        if os.path.basename(appDir) in self.webDirs:
            if appDir == self.webDirs[os.path.basename(appDir)]:
                print("Task basename '%s': directory already installed (%s)" % \
                      (os.path.basename(appDir), appDir))
            else:
                print("Task basename '%s' matches basename in a different directory (%s)" % \
                      (os.path.basename(appDir), self.webDirs[os.path.basename(appDir)]))
            return False            

        fullPath = os.path.realpath(os.path.abspath(appDir))

        try:
            i = self.dirPairs.index(["", fullPath])
            if verbose:
                print("Task %s already registered. Skipping." % fullPath)
            return False
        except ValueError:
            if not testOnly:
                try:
                    i = self.dirPairs.index(["#", fullPath])
                    # there, just commented out.
                    self.dirPairs[i][0] = ""
                except ValueError:
                    self.dirPairs.append(["", fullPath])
                self.webDirs[os.path.basename(appDir)] = fullPath
            return True

    def remove(self, appDir, commentOut = False, verbose = False):
        # If the directory you ask to remove has moved, you want to check
        # for everything. Use samefile() if the basename is there.
        # And if it's been removed, make sure that the literal
        # file is removed.
        fullPath = os.path.abspath(appDir)

        pathToRemove = None
        try:
            curPath = self.webDirs[os.path.basename(appDir)]
            if os.path.exists(curPath) and os.path.exists(fullPath) and os.path.samefile(curPath, fullPath):
                pathToRemove = curPath
            elif curPath == fullPath:
                pathToRemove = curPath
        except KeyError:
            pass

        if pathToRemove:
            i = self.dirPairs.index(["", pathToRemove])
            if commentOut:
                self.dirPairs[i][0] = "#"
            else:
                # Delete it.
                self.dirPairs[i:i+1] = []
            try:
                del self.webDirs[os.path.basename(appDir)]
            except:
                pass
        elif verbose:
            print("Task %s not registered. Skipping." % fullPath)

    #
    # toplevel operations
    #

    def validatePluginDir(self, appDir, verbose = False):

        global _PluginNamespace

        # Normalize this.
        appDir = os.path.realpath(os.path.abspath(appDir))

        taskFile = os.path.join(appDir, "task.xml")

        if not os.path.exists(taskFile):
            raise PluginError("%s is not a proper task directory" % appDir)

        if os.path.basename(appDir) in self.webDirs:
            if appDir != self.webDirs[os.path.basename(appDir)]:
                raise PluginError("Task basename '%s' matches basename in a different directory (%s)" % \
                                  (os.path.basename(appDir), self.webDirs[os.path.basename(appDir)]))

        # In order to truly do this, we have to load this dir in the
        # context of the existing dirs, and then check the CGI metadata.

        origPluginNamespace = _PluginNamespace.copy()
        try:            
            pdir = self.loadPlugins(appDir)
            # Now, for each one task in the pdir that has appDir
            # as a true directory, get the CGI metadata, to confirm
            # that it can be generated. Actually, that's not how
            # things work. I get the cgi metadata for the whole dir.
            # So I need to trim down the plugindict to just the
            # apps that are defined plus their parents. 
            allTasks = pdir.getAllTasks()
            relevantTasks = [t for t in allTasks if t.taskRoot == appDir]
            for t in set(allTasks) - set(relevantTasks):
                pdir.pruneTask(t)
            pdir.getCGIMetadata()
        finally:
            _PluginNamespace = origPluginNamespace

    def installPluginDir(self, appDir, verbose = False):

        # Normalize this.
        appDir = os.path.realpath(os.path.abspath(appDir))

        taskFile = os.path.join(appDir, "task.xml")

        if not os.path.exists(taskFile):
            raise PluginError("%s is not a proper task directory" % appDir)

        self.readPluginRecord(verbose = verbose)

        if not self.add(appDir, verbose = verbose, testOnly = True):
            # It was already registered.
            return

        # First, validate it. It'll throw an error if it
        # can't be validated. This will do some extra work,
        # but won't modify the global context.
        self.validatePluginDir(appDir, verbose = verbose)

        if os.path.exists(taskFile):

            settingsPath = os.path.join(appDir, "MAT_settings.config")
            if os.path.exists(settingsPath):
                MATConfig.configureFromFile(settingsPath, augment = True)
                        
            pluginPy = glob.glob(os.path.join(appDir, "python", "*.py"))

            if verbose:
                print("### Compiling Python files ...")

            # Since code might be spread out over multiple directories,
            # we need to have all the plugin directories in sys.path.

            # I want to ensure that each module is imported, so the
            # code is compiled at configuration time, JUST IN CASE
            # the installation directory can't be written by the
            # person running the code - otherwise, Python will hang for
            # a bit trying to save the compiled file.

            import imp

            # Originally, I was copying sys.modules and then reestablishing
            # it, but it turns out that copying sys.modules doesn't
            # do the right thing in the imp package, which points directly
            # to what sys.modules is supposed to be pointing to. So what
            # I actually need to do is check the keys in sys.modules, and
            # then restore that key set.

            # Use list() to freeze dictionary keys in Python 3.
            origSysModuleKeys = list(sys.modules.keys())
            origSysPath = sys.path

            # Now, because I haven't yet augmented self.dirPairs with the
            # new directory (because I don't trust it yet) I have to
            # add it temporarily

            dirPairs = self.dirPairs + [["", appDir]]

            try:
                for p in pluginPy:
                    print("Compiling", p, "...")
                    # I need to copy sys.path for the moment.
                    sys.path = [self.matLoc] + [os.path.join(pth, "python") for prefix, pth in dirPairs if prefix == ""] + origSysPath
                    modName = os.path.splitext(os.path.basename(p))[0]
                    try:
                        fp, pathname, desc = imp.find_module(modName)
                        imp.load_module(modName, fp, pathname, desc)
                    except ImportError as e:
                        raise PluginError("Couldn't load plugin from %s: %s" % (p, str(e)))
                    if fp:
                        fp.close()
            finally:
                for key in set(sys.modules.keys()) - set(origSysModuleKeys):                    
                    del sys.modules[key]
                sys.path = origSysPath

        # No errors were thrown.
        self.add(appDir, verbose = verbose)

        self.writePluginRecord(verbose = verbose)

    def uninstallPluginDir(self, appDir, commentOut = False, verbose = False):

        if self.pluginRecordExists():

            self.readPluginRecord(verbose = verbose)
            self.remove(appDir, commentOut = commentOut, verbose = verbose)
            self.writePluginRecord(verbose = verbose)

    def getPluginDirs(self):
        # Preserve the order of the file.
        return [x[1] for x in self.dirPairs if x[0] == ""]

    #
    # And now, the ugly job of actually loading the plugins.
    #

    def loadPlugins(self, *otherDirs):

        self.readPluginRecord()

        # Make sure the dirs are canonicalized and only
        # added if not present.

        allDirs = self.getPluginDirs()
        for dir in otherDirs:
            dir = os.path.realpath(os.path.abspath(dir))
            if dir not in allDirs:
                allDirs.append(dir)

        return self.loadPluginsFromDirs(allDirs)

    def loadPluginsFromDirs(self, allDirs):

        global _PluginNamespace

        # Originally, I just did execfile, but let's try loading the
        # modules, instead, and rooting around in them. We'll have to
        # use imp. But if I do that, I'd better make sure that the
        # apps have been compiled. It's not enough to tell it to load
        # a source module, because it'll try to compile it and
        # there's no way to block it.

        import imp

        xmlDescs = []

        #
        # Second step: load all the Python modules, and all the
        # XML descriptions. All the directories have to be available
        # in the Python path, because otherwise, if dependencies among
        # tasks are scattered across task directories, you'll
        # be hosed. 
        #

        allPyDirs = [os.path.join(x, "python") for x in allDirs]

        for pdirName in allDirs:

            xmlPath = os.path.join(pdirName, "task.xml")
            settingsPath = os.path.join(pdirName, "MAT_settings.config")
            if os.path.exists(settingsPath):
                MATConfig.configureFromFile(settingsPath, augment = True)

            if not os.path.exists(xmlPath):
                raise PluginError("%s contains no task.xml file" % pdirName)

            if os.path.exists(xmlPath):
                xmlDescs.append((pdirName, LoadPluginTask(xmlPath)))

                # It's not enough to pass a path to imp.find_module;
                # you have to set sys.path for load_module to work
                # if the loaded module has any other local dependencies.

                pyDir = os.path.join(pdirName, "python")

                # There may not be any extensions

                if not os.path.isdir(pyDir):
                    continue

                # Use all the paths, but put the current one first.

                oldSysPath = sys.path[:]        

                sys.path[0:0] = [pyDir] + [p for p in allPyDirs if p != pyDir]

                for file in os.listdir(pyDir):

                    p = os.path.join(pyDir, file)
                    if os.path.isfile(p) and os.path.splitext(file)[1] == ".py":
                        # It's a Python file. Load it AS A MODULE.
                        modName = os.path.splitext(file)[0]
                        if modName in sys.modules:
                            _PluginNamespace[modName] = sys.modules[modName]
                        else:
                            fp = None
                            try:
                                fp, pathname, desc = imp.find_module(modName)
                                m = imp.load_module(modName, fp, pathname, desc)
                                _PluginNamespace[modName] = m
                            except ImportError as e:                        
                                raise PluginError("Couldn't load plugin from %s: %s" % (p, str(e)))
                            if fp:
                                fp.close()

                sys.path = oldSysPath

        #
        # Third step: collect the classes for each description
        #

        pluginClasses = {}

        for pDirName, taskDescs in xmlDescs:
            for taskDesc in taskDescs:
                pAttrs = taskDesc.attrs
                if pAttrs["class"] is not None:
                    try:
                        c = FindPluginClass(pAttrs["class"], pAttrs["name"])
                        if not issubclass(c, PluginTaskDescriptor):
                            raise PluginError("task %s specifies descriptor class which is not a subclass of PluginTaskDescriptor" % pAttrs["name"])
                        pluginClasses[pAttrs["name"]] = c
                    except NameError:
                        raise PluginError("task %s specifies unknown descriptor class %s" % (pAttrs["name"], pAttrs["class"]))
                else:
                    pluginClasses[pAttrs["name"]] = PluginTaskDescriptor

        #
        # Fourth step: ensure that the parent relationships among
        # the XML descriptions parallel the relationships among
        # the referenced classes.
        #

        for pDirName, taskDescs in xmlDescs:
            for taskDesc in taskDescs:
                pAttrs = taskDesc.attrs
                if pAttrs["parent"] is not None:
                    if pAttrs["parent"] not in pluginClasses:
                        raise PluginError("the parent specified by task %s, %s, doesn't exist" % (pAttrs["name"], pAttrs["parent"]))
                    # I was checking to make sure that the parent/child relationship
                    # was mirrored in a superclass/subclass relationship, but now
                    # that doesn't work for the step proxies that I introduced (and later
                    # removed), and I'm not sure it's necessary anyway.
                elif not issubclass(pluginClasses[pAttrs["name"]], PluginTaskDescriptor):
                    raise PluginError("the class specified by task %s is not a subclass of PluginTaskDescriptor" % pAttrs["name"])

        #
        # Fifth step: actually build each element, and
        # then assign the parent relationships.
        #

        outputPDir = PluginDict()

        # Define a root task.

        rootTask = outputPDir.getRootTask()

        # Originally, we were digesting the XML before we
        # established the parent - child links. But more and
        # more, it seems we really need to do this in order.

        parentToChildren = {}

        for taskRoot, taskDescs in xmlDescs:
            for taskDesc in taskDescs:
                pAttrs = taskDesc.attrs
                # This might be None.
                parentName = pAttrs["parent"]
                try:
                    parentToChildren[parentName].append((taskRoot, taskDesc))
                except KeyError:
                    parentToChildren[parentName] = [(taskRoot, taskDesc)]

        # Don't forget to worry about cycles. First, let's
        # start with everything that doesn't have any parents,
        # and collect the ones that are eligible for the next round.

        # A list of tasks and task names.
        parentTasks = [(rootTask, None)]
        while True:
            newParentTasks = []
            for parentTask, parentTaskName in parentTasks:
                children = parentToChildren.get(parentTaskName)
                if children:
                    del parentToChildren[parentTaskName]
                    for childTaskRoot, childTaskDesc in children:
                        pAttrs = childTaskDesc.attrs
                        pClass = pluginClasses[pAttrs["name"]]
                        task = pClass(pAttrs["name"], childTaskRoot)
                        # Set up the links before we call fromXML.
                        parentTask.addChild(task)
                        task.fromXML(childTaskDesc, lazyClassLoad = self.lazyClassLoad)
                        outputPDir.recordTaskDir(task)
                        if task.visible:
                            outputPDir[pAttrs["name"]] = task
                            # We need to do a final check here. If the task has some bad dependencies, it
                            # won't get checked until you actually ask about the repository.
                            # I can do this here now, because I'm setting
                            # up the parent-child links.
                            try:
                                t = task.getAnnotationTypeRepository()
                                # Check the UI label constraints.
                                t.checkUILabelConstraints(task.name)
                            except Exception as e:
                                if ExecutionContext._DEBUG:
                                    raise
                                else:
                                    raise PluginError("For task %s: checking the annotation repository generated the following error: %s" % (task.name, str(e)))
                        newParentTasks.append((task, task.name))
            # Keep doing this until you run out of new parents to check.
            if not newParentTasks:
                break
            else:
                parentTasks = newParentTasks

        # Now, once you're done, there may still be stuff left in
        # the parentToChildren table, in which case, it represents a
        # circularity or a missing parent.

        if parentToChildren:
            raise PluginError("the following tasks could not be created, either because the parent is unknown or because the parent link would have created a circular reference: %s" %
                                ", ".join([", ".join(["'%s' (parent %s)" % (taskDesc.attrs["name"], parent)
                                                      for (taskRoot, taskDesc) in children])
                                           for (parent, children) in parentToChildren.items()]))

        # And we're done. Return ONLY the visible ones - we
        # can't refer to the invisible ones anyway.

        return outputPDir

# Backward compatibility.

def LoadPlugins(*otherDirs, **kw):
    
    # In Python 3, you can define keywards after defaults, but not in
    # Python 2. So I have to fake it. Doing this for backward
    # compatibility with previous uses of LoadPlugins().
    lazyClassLoad = False
    for k, v in kw.items():
        if k not in ["lazyClassLoad"]:
            raise TypeError("LoadPlugins() got an unexpected keyword argument '%s'" % k)
        else:
            lazyClassLoad = v

    return PluginLocationMgr(lazyClassLoad = lazyClassLoad).loadPlugins(*otherDirs)

def LoadPluginsFromDirs(allDirs, lazyClassLoad = False):
    return PluginLocationMgr(lazyClassLoad = lazyClassLoad).loadPluginsFromDirs(allDirs)

#
# Someday this will be more general. But right now, I'm using it only for
# inferred tasks, from MATReport and from the UI.

import xml.etree.ElementTree as ET

def taskJSONToXML(taskName, jsonD):
    root = ET.Element("task", {"name": taskName})
    if ("alphabetizeLabels" in jsonD) and (jsonD["alphabetizeLabels"] is False):
        ET.SubElement(root, "web_customization", {"alphabetize_labels": "no"})
    if "languages" in jsonD:
        lgNode = ET.SubElement(root, "languages")
        for lg, lgDesc in jsonD["languages"].items():
            lgattrs = {}
            if "code" in lgDesc:
                lgattrs["code"] = lgDesc["code"]
            if "name" in lgDesc:
                lgattrs["name"] = lgDesc["name"]
            if lgDesc.get("text_right_to_left", False):
                lgattrs["text_right_to_left"] = "yes"
            ET.SubElement(lgNode, "language", lgattrs)
    repo = jsonD.get("annotationSetRepository")
    if repo:
        annotsNode = ET.SubElement(root, "annotations")
        if repo.get("allAnnotationsKnown") == True:
            annotsNode.set("all_annotations_known", "yes")
        # If the tag order has been defined, use it.
        tagOrder = jsonD.get("tagOrder") or list(repo["types"].keys())
        for labName in tagOrder:
            if labName == "SEGMENT":
                continue
            lab = repo["types"][labName]
            if lab.get("hasSpan") is False:
                annotNode = ET.SubElement(annotsNode, "spanless", {"label": labName})
            else:
                annotNode = ET.SubElement(annotsNode, "span", {"label": labName})
            thisCategory = lab.get("category", "content")
            thisSetname = lab.get("set_name", "content")
            if thisCategory != "content":
                annotNode.set("of_category", thisCategory)
            if thisSetname != "content":
                annotNode.set("of_set", thisSetname)
            if lab.get("allAttributesKnown", False) is True:
                annotNode.set("all_attributes_known", "yes")
            if lab.get("processable", True) is False:
                annotNode.set("processable", "no")
            if lab.get("display", {}).get("css"):                
                annotNode.set("d_css", lab["display"]["css"])
            attrs = lab.get("attrs")
            if attrs:
                for attr in attrs:
                    attrType = attr["type"]
                    if attrType == "annotation":
                        attrType = "filler"
                    if attr.get("aggregation") and (attr.get("aggregation") != "none"):
                        attrType += "_" + attr["aggregation"]
                    attrNode = ET.SubElement(annotNode, attrType, {"name": attr["name"]})
                    if attr.get("category") and (attr["category"] != thisCategory):
                        attrNode.set("of_category", attr["category"])
                    if attr.get("set_name") and (attr["set_name"] != thisSetname):
                        attrNode.set("of_set", attr["set_name"])
                    if (attr["type"] == "annotation") and attr.get("label_restrictions"):
                        # There will be label restrictions, I hope.
                        attrNode.set("filler_types", ",".join(attr["label_restrictions"]))            
    if "steps" in jsonD:
        stepNode = ET.SubElement(root, "steps")
        for stepName, stepObj in jsonD["steps"].items():
            sAttrs = {"name": stepObj["initSettings"]["name"]}
            if "sType" in stepObj["annotationStepInfo"]:
                sAttrs["type"] = stepObj["annotationStepInfo"]["sType"]
            for k, attr in [("setsAdded", "sets_added"),
                            ("setsDesired", "sets_desired"),
                            ("setsForbidden", "sets_forbidden"),
                            ("setsRequired", "sets_required"),
                            ("setsModified", "sets_modified")]:
                if (k in stepObj["annotationStepInfo"]) and \
                   len(stepObj["annotationStepInfo"][k]) > 0:
                    sAttrs[attr] = ",".join(stepObj["annotationStepInfo"][k])
            if stepObj["stepType"] == "annotation":
                lab = "annotation_step"
            elif stepObj["stepType"] == "signal":
                lab = "signal_step"
            else:
                lab = "transform_step"
            ET.SubElement(stepNode, lab, sAttrs)
    if "workflows" in jsonD:
        wfNode = ET.SubElement(root, "workflows")
        for wfName, wfObj in jsonD["workflows"].items():
            wfAttrs = {"name": wfName}
            if ("workflowData" in wfObj) and (wfObj["workflowData"].get("undoable") == True):
                wfAttrs["undoable"] = "yes"
            thisWorkflow = ET.SubElement(wfNode, "workflow", wfAttrs)
            for step in wfObj.get("steps", []):
                stepName = step["initSettings"]["name"]
                globalStype = jsonD["steps"][stepName]["annotationStepInfo"]["sType"]
                thisStep = {"name": stepName}
                localStype = step.get("annotationStepInfo", {}).get("sType")
                if localStype and (localStype != globalStype):
                    thisStep["type"] = localStype
                if step.get("initSettings", {}).get("pretty_name"):
                    thisStep["pretty_name"] = step["initSettings"]["pretty_name"]
                ET.SubElement(thisWorkflow, "step", thisStep)
        
    # Oh, barf. No pretty printing with elementtree.
    import xml.dom.minidom
    return xml.dom.minidom.parseString(ET.tostring(root)).toprettyxml(indent = "  ")
        
#
# Task writer for the annotation reporter
#

# Part of this used to be in Document.py, but this portion depends on
# things here in the plugin manager, and the interdependencies just got to be too much.
# Also, DocumentIO depends on the core of the annotation reporter, and there's no
# reason for IT to depend on PluginMgr.

class TaskWritingAnnotationReporter(Document.AnnotationReporter):
    
    WEB_COLORS = [
        # Pink colors
        ["background:pink;color:black",
         "background:lightpink;color:black",
         "background:hotpink;color:white",
         "background:deeppink;color:white",
         "background:palevioletred;color:white",
         "background:mediumvioletred;color:white"],
        #Red colors
        ["background:lightsalmon;color:black",
         "background:salmon;color:black",
         "background:darksalmon;color:black",
         "background:lightcoral;color:black",
         "background:indianred;color:white",
         "background:crimson;color:white;color:white",
         "background:fireBrick;color:white",
         "background:darkred;color:white",
         "background:red;color:white"],
        # Orange colors
        ["background:orangered;color:white",
         "background:tomato;color:white",
         "background:coral;color:white",
         "background:darkorange;color:white",
         "background:orange;color:white",
         "background:gold;color:black"],
        # Yellow colors
        ["background:yellow;color:black",
         "background:lightyellow;color:black",
         "background:lemonchiffon;color:black",
         "background:lightgoldenrodyellow;color:black",
         "background:papayawhip;color:black",
         "background:moccasin;color:black",
         "background:peachpuff;color:black",
         "background:palegoldenrod;color:black",
         "background:khaki;color:black",
         "background:darkkhaki;color:black"],
        # Brown colors
        ["background:cornsilk;color:black",
         "background:blanchedalmond;color:black",
         "background:bisque;color:black",
         "background:navajowhite;color:black",
         "background:wheat;color:black",
         "background:burlywood;color:black",
         "background:tan;color:black",
         "background:rosybrown;color:black",
         "background:sandybrown;color:black",
         "background:goldenrod;color:black",
         "background:darkgoldenrod;color:white",
         "background:Peru;color:white",
         "background:chocolate;color:white",
         "background:saddlebrown;color:white",
         "background:sienna;color:white",
         "background:brown;color:white",
         "background:maroon;color:white"],
        # Green colors
        ["background:darkolivegreen;color:white",
         "background:olive;color:white",
         "background:olivedrab;color:white",
         "background:yellowgreen;color:black",
         "background:limegreen;color:black",
         "background:lime;color:black",
         "background:lawngreen;color:black",
         "background:chartreuse;color:black",
         "background:greenyellow;color:black",
         "background:springgreen;color:black",
         "background:mediumspringgreen;color:black",
         "background:lightgreen;color:black",
         "background:palegreen;color:black",
         "background:darkseagreen;color:black",
         "background:mediumseagreen;color:white",
         "background:seagreen;color:white",
         "background:forestgreen;color:white",
         "background:green;color:white",
         "background:darkgreen;color:white"],
        # Cyan colors
        ["background:mediumaquamarine;color:black",
         "background:aqua;color:black",
         "background:cyan;color:black",
         "background:lightcyan;color:black",
         "background:paleturquoise;color:black",
         "background:aquamarine;color:black",
         "background:turquoise;color:black",
         "background:mediumturquoise;color:black",
         "background:darkturquoise;color:white",
         "background:lightseagreen;color:white",
         "background:cadetblue;color:white",
         "background:darkcyan;color:white",
         "background:teal;color:white"],
        # Blue colors
        ["background:lightsteelblue;color:black",
         "background:powderblue;color:black",
         "background:lightblue;color:black",
         "background:skyblue;color:black",
         "background:lightskyblue;color:black",
         "background:deepskyblue;color:white",
         "background:dodgerblue;color:white",
         "background:cornflowerblue;color:white",
         "background:steelblue;color:white",
         "background:royalblue;color:white",
         "background:blue;color:white",
         "background:mediumblue;color:white",
         "background:darkblue;color:white",
         "background:navy;color:white",
         "background:midnightblue;color:white"],
        # Purple colors
        ["background:lavender;color:black",
         "background:thistle;color:black",
         "background:plum;color:black",
         "background:violet;color:black",
         "background:orchid;color:black",
         "background:fuchsia;color:white",
         "background:Magenta;color:white",
         "background:mediumorchid;color:white",
         "background:mediumpurple;color:white",
         "background:blueviolet;color:white",
         "background:darkviolet;color:white",
         "background:darkorchid;color:white",
         "background:darkmagenta;color:white",
         "background:purple;color:white",
         "background:indigo;color:white",
         "background:darkslateblue;color:white",
         "background:slateblue;color:white",
         "background:mediumslateblue;color:white"],
        # Gray/Grey/Black colors
        ["background:gainsboro;color:black",
         "background:lightgrey; color:black;",
         "background:silver;color:black",
         "background:darkgrey; color:black;",
         "background:gray;color:black",
         "background:dimgrey; color:white;",
         "background:lightslategrey; color:white;",
         "background:slategrey; color:white;",
         "background:darkslategrey; color:white;",
         "background:black;color:white"]]

    def writeTask(self, taskName, path):
        step = {
            "annotationStepInfo": {
                "setsAdded": ["content"],
                "setsDesired": [],
                "setsForbidden": [],
                "setsModified": [],
                "setsRequred": [],
                "sType": "hand",
                "finalSetMentions": ["content"]
            },
            "initSettings": {"name": "annotate"},
            "stepType": "annotation"
        }

        jsonD = {
            "languages": {
                "Left to right": {
                    "name": "Left to right",
                    "text_right_to_left": False,
                    "code": "ltr"
                    },
                "Right to left": {
                    "name": "Right to left",
                    "text_right_to_left": True,
                    "code": "rtl"
                    }
                },
            "alphabetizeLabels": False,
            "tagOrder": [],
            "steps": {
                "annotate": step
            },
            "workflows": {
                "Hand annotation": {
                    "steps": [step],
                    "workflowData": {
                        "undoable": True
                    },
                }
            },
            "annotationSetRepository": {
                "allAnnotationsKnown": False,
                "types": {}
            }
        }
        
        majorColorIndex = 0
        minorColorIndex = 0
        # Now, for each label in allAtypes, write out all the attributes you
        # can find, and assign a color.

        for k, v in self.allAtypes.items():

            # First, let's see if the annotation label ends up being both spanned and spanless.
            spanVals = set([atype.hasSpan for atype in v])
            if len(spanVals) > 1:
                print("WARNING: label %s appears as both spanned and spanless in reported documents; skipping label in rendering task" % k, file=sys.stderr)
                continue                
            
            # First, the color. That's the easy part.
            if majorColorIndex == len(self.WEB_COLORS):
                majorColorIndex = 0
                minorColorIndex += 1
            newAtype = {
                "display": {
                    "css": self.WEB_COLORS[majorColorIndex][minorColorIndex % len(self.WEB_COLORS[majorColorIndex])]
                },
                "set_name": "content",
                "category": "content",
                "attrs": []
            }
            majorColorIndex += 1
            jsonD["tagOrder"].append(k)

            if list(spanVals)[0]:
                newAtype["hasSpan"] = True
            else:
                newAtype["hasSpan"] = False
            jsonD["annotationSetRepository"]["types"][k] = newAtype

            # Now, let's loop through the attributes, and add them, one at a time.
            # If any of them clash, then skip it.

            typePairs = {}
            for atype in v:
                for attr in atype.attr_list:
                    prev = typePairs.get(attr.name)
                    if prev is not None:
                        prevType, prevAggr = prev
                        if (prevType != attr._typename_) or (prevAggr != attr.aggregation):
                            print("WARNING: attribute %s of label %s has internally inconsistent types in the reported documents; skipping in rendering task" % (attr.name, k), file=sys.stderr)
                    else:
                        typePairs[attr.name] = (attr._typename_, attr.aggregation)

            # Now, we've checked all the atypes and collected all the types.
            for attrName, (attrType, attrAggr) in typePairs.items():
                attr = {
                    "name": attrName,
                    "type": attrType
                }
                if attrAggr is not None:
                    attr["aggregation"] = attrAggr
                newAtype["attrs"].append(attr)
                if attrType == "annotation":
                    # List the label restrictions.
                    attr["label_restrictions"] = []
                    for lab in self.allAnnotAttrLabs[(k, attrName)]:
                        attr["label_restrictions"].append(lab)

        fp = codecs.open(path, "w", "utf8")
        fp.write(taskJSONToXML(taskName, jsonD))
        fp.close()
        
        

