# Copyright (C) 2008 - 2018 The MITRE Corporation. See the toplevel
# file LICENSE.txt for license terms.

from __future__ import print_function, absolute_import

import os

# This entry was previously in Workspace.py, but it would have caused
# a circularity if I'd been doing my imports cleanly. So I now have
# a registration method.

from .Operation import (Option, OpArgument)

from .CarafeTrain import (TestRun, TrainingRun, ExperimentEngine,
                          WorkspaceCorpusSet, WorkspaceCorpus, fromXML)
from .Workspace import (WorkspaceOp, WorkspaceToplevelOperation,
                        CMDLINE_AVAILABLE, WorkspaceError)
from . import Score

@WorkspaceOp()
class RunExperimentOperation(WorkspaceToplevelOperation):

    availability = CMDLINE_AVAILABLE

    name = "run_experiment"

    # Either we have an experiment_file and an workspace_binding, or we define test_users,
    # test_basename_sets, test_basename_patterns, test_document_statuses, test_exclude_unassigned,
    # and everything else is the remainder.
    args = [Option("--experiment_file", type="string",
                   help = "Specify an experiment file to use. If specified, --workspace_binding is also required. Either this or one of the --test_* parameters must be provided."),
            Option("--workspace_binding", type="string",
                   help = "variable in the workspace experiment file to which this workspace should be bound. Required if --experiment_file is present."),
            Option("--test_users", type="string",
                   metavar="user(,user...)",
                   help = "A comma-separated sequence of users to restrict the test corpus to. Not permitted if --experiment_file is provided."),
            Option("--test_basename_sets", type="string",
                   metavar="set(,set...)",
                   help = "A comma-separated sequence of basename set names to restrict the test corpus to. Not permitted if --experiment_file is provided."),
            Option("--test_basename_patterns", type="string",
                   metavar="pat(,pat...)",
                   help = "A comma-separated sequence of glob-style basename patterns to restrict the test corpus to. Not permitted if --experiment_file is provided."),
            Option("--test_step_statuses", type="string",
                   metavar="status(,status...)",
                   help = "A comma-separated sequence of step statuses in the target trainable step to restrict the test corpus to. The background corpus will already be restricted to 'partially gold,gold,reconciled'. Not permitted if --experiment_file is provided."),
            Option("--test_step", type="string",
                   metavar="step",
                   help = "The name of the trainable step in the workspace's workflow to target. Not permitted if --experiment_file is provided. Required if --experiment_file is absent and the workflow has more than one trainable step."),
            OpArgument("test_exclude_unassigned",
                       help = "If present, exclude unassigned documents from the test corpus. Not permitted if --experiment_file is provided."),
            OpArgument("csv_formula_output", hasArg = True,
                       help = "The format for the CSV output files. See the MATScore documentation for details.")]
               
    @classmethod
    def getUsage(cls, w):
        return """Usage: %prog [options] <dir> run_experiment [options]"""

    def do(self, experiment_file = None, workspace_binding = None, test_users = None,
           test_basename_sets = None, test_basename_patterns = None, test_step = None, test_step_statuses = None,
           test_exclude_unassigned = False, csv_formula_output = None):
        # Doesn't require a transaction.
        if experiment_file and \
           (test_users or test_basename_sets or test_basename_patterns or
            test_step_statuses or test_exclude_unassigned or test_step):
            raise WorkspaceError(("can't specify experiment file as well as any of the test_* parameters"))
        if (experiment_file is not None) != (workspace_binding is not None):
            raise WorkspaceError(("either both experiment_file and workspace_binding must be provided, or neither"))
        if (not experiment_file) and \
           (not (test_users or test_basename_sets or test_basename_patterns or
                 test_step_statuses or test_exclude_unassigned)):
            raise WorkspaceError(("no experiment_file, but also no specification of test corpus"))
        if not experiment_file:
            trainableSteps = self.workspace.workflow.getTrainableSteps()
            if test_step is None:
                if len(trainableSteps.values()) > 1:
                    raise WorkspaceError("no experiment_file, but no test_step and the workspace workflow has more than one trainable step")
                else:
                    test_step = trainableSteps[0].stepName
            else:
                found = False
                for t in trainableSteps.values():
                    if (t.stepName == test_step) or (t.prettyName == test_step):
                        test_step = t.stepName
                        found = True
                        break
                if not found:
                    raise WorkspaceError("no experiment_file, but unknown trainable step '%s' in workspace workflow" % test_step)
        # Create an experiment directory.
        if not os.path.exists(os.path.join(self.workspace.dir, "experiments")):
            os.makedirs(os.path.join(self.workspace.dir, "experiments"))
        import datetime
        now = datetime.datetime.now()
        dirName = os.path.join(self.workspace.dir, "experiments",
                               now.strftime("%Y%m%d_%H%M%S") + "_" + ("%06d" % now.microsecond))
        self.experimentDir = dirName
        if experiment_file:
            e = ExperimentEngine(**fromXML(experiment_file, dir = dirName,
                                           bindingDict = {workspace_binding: self.workspace.dir}))
        else:
            testCorpus = WorkspaceCorpus("test",
                                         stepStatuses = test_step_statuses,
                                         basenameSets = test_basename_sets,
                                         users = test_users,
                                         includeUnassigned = not test_exclude_unassigned,
                                         basenamePatterns = test_basename_patterns)
            trainCorpus = WorkspaceCorpus("train", useWorkspaceRemainder = True)
            e = ExperimentEngine(dir = dirName, task = self.workspace.task,
                                 workspaceCorpusSets = [WorkspaceCorpusSet(self.workspace.dir,
                                                                           step = test_step,
                                                                           stepStatuses = "partially gold,gold,reconciled",
                                                                           workspaceCorpora = [testCorpus, trainCorpus])],
                                 models = [TrainingRun("exp_model", trainingCorpora = [("train", None)],
                                                       trainableStep = test_step,
                                                       engineSettings = {"partial_training_on_gold_only": True})],
                                 runs = [TestRun("exp_run", model = "exp_model",
                                                 testCorpora = [("test", None)],
                                                 engineOptions = {"steps": test_step, "workflow": self.workspace.workflow.name},
                                                 scoreOptions = {"restrictRefToGoldSegments": True},
                                                 enginePrepOptions = {"undo_through": test_step,
                                                                      "workflow": self.workspace.workflow.name,
                                                                      "output_file_type": "mat-json"})])
        fmt = Score.ScoreFormat(csvFormulaOutput = csv_formula_output)
        e.run(format = fmt)

    def fnResult(self):
        return self.experimentDir
