#!/usr/bin/env python3

####################################################################################################
### -----------------------------------------  NOTICE  ----------------------------------------- ###
###                                                                                              ###
### This software (or technical data) was produced for the U.S. Government under contract        ###
### 2015-14120200002-002, and is subject to the Rights in Data-General Clause 52.227-14 , ALT    ###
### IV (MAY 2014) or (DEC 2007).                                                                 ###
###                                                                                              ###
### © 2019 The MITRE Corporation. All Rights Reserved.                                           ###
###                                                                                              ###
### -------------------------------------------------------------------------------------------- ###
####################################################################################################

"""
Diff's two BP documents.
"""

__author__  = "Leland E. Vakarian <lsperl@mitre.org>"
__date__    = "22 Aug, 2020"

import click

from lib.utils import CONTEXT_SETTINGS
from lib.utils import LOGGER


@click.command(context_settings=CONTEXT_SETTINGS)
@click.option('-v', '--verbose', is_flag=True, help="Show verbose output.")
@click.option('-f', '--force', is_flag=True, help="Skip asking to overwrite output.")
@click.option(
    '-l',
    '--save-logs',
    'save_logs',
    type=click.Path(),
    help="Capture logging output (at whatever verbosity level) to the specified path."
)
@click.option(
    '--no-validation',
    is_flag=True,
    help="Don't validate the output BP document."
)
@click.argument('in_path_a', type=click.Path(exists=True))
@click.argument('in_path_b', type=click.Path(exists=True))
def cli(verbose, force, save_logs, no_validation, in_path_a, in_path_b):
    """Diff's two BP documents.

    Output is written to the console verbosely. Save output by passing ``--save-logs``.

    This CLI wraps the ``lib.bp.BPDocument.diff(lib.bp.BPDocument)`` function. Wherever two values
    are shown side-by-side, the left value refers to the caller, and the right value refers to the
    callee.

    \f

    This function serves as the entrypoint for the program when called on the CLI.

    Args:
        verbose (bool): if True, pretty-print output as the command runs. Defaults to False.
        force (bool): if True, skip asking for permission to overwrite a file with output. Defaults
            to False.
        save_logs (str): a path to write captured logging output to.
        no_validation (bool): If True, skip validating constructed BP objects. Defaults to False.
        in_path_a (str): a path to an input BP document.
        in_path_b (str): a path to an input BP document.
    """
    # do this before anything else
    from lib.utils import capture_logs
    if save_logs:
        capture_logs(save_logs, True, force)

    from lib import bp

    global LOGGER

    if verbose:
        LOGGER.info("Verbose output is on!")
    if verbose and force:
        LOGGER.warning("Forced overwrite is on!")
    elif verbose and not force:
        LOGGER.info("Forced overwrite is off!")

    ## load docs and emit diff

    doc_a = bp.BPDocument.from_json(in_path_a, no_validation=no_validation)
    doc_b = bp.BPDocument.from_json(in_path_b, no_validation=no_validation)
    doc_a.diff(doc_b, verbose=verbose)

    if verbose:
        LOGGER.info(f"Documents are {'not ' if doc_a != doc_b else ''}equal.")


if __name__ == '__main__':

    ## program proper

    cli()
