#!/usr/bin/env python3

####################################################################################################
### -----------------------------------------  NOTICE  ----------------------------------------- ###
###                                                                                              ###
### This software (or technical data) was produced for the U.S. Government under contract        ###
### 2015-14120200002-002, and is subject to the Rights in Data-General Clause 52.227-14 , ALT    ###
### IV (MAY 2014) or (DEC 2007).                                                                 ###
###                                                                                              ###
### © 2019 The MITRE Corporation. All Rights Reserved.                                           ###
###                                                                                              ###
### -------------------------------------------------------------------------------------------- ###
####################################################################################################

"""
Mutates BP documents in various ways.
"""

__author__  = "Leland E. Vakarian <lsperl@mitre.org>, David S. Day <day@mitre.org>"
__date__    = "14 April, 2022"

import click

from lib.utils import CONTEXT_SETTINGS
from lib.utils import LOGGER
from lib.utils import MAT_DIALECT_CHOICES


DEBUG = False
FORCE = False
VERBOSE = False
"""Global flags handled by cli()."""

EXPECTED_ABSTRACT_TYPES = [
    "helpful-material",
    "helpful-verbal",
    "helpful-both",
    "harmful-material",
    "harmful-verbal",
    "harmful-both",
    "neutral-material",
    "neutral-verbal",
    "neutral-both",
]
EXTRA_ABSTRACT_TYPES = [
    "helpful-unk",
    "harmful-unk",
    "neutral-unk",
]
"""There's no constant in ``lib.utils`` for this construction, nor should there be."""

DIST_PROVIDED_DEFAULT = "72.0%"
DIST_PROVIDED_ANALYSIS_DEFAULT = "15.0%"
DIST_PROVIDED_DEVTEST_DEFAULT = "15.0%"
DEPTH_DEFAULT = 20000
"""Defaults for the deliverables."""


def __find_annotations(doc, dialect, no_check_coverage=False):
    """Maps instances of annotation types for a given dialect to Entry IDs.

    Args:
        doc (lib.bp.BPDocument): a document to map.
        dialect (str): the MAT task the BP document was created with. Must be one of
            ``utils.MAT_DIALECT_CHOICES``, which is verified by the Click decorator.
        no_check_coverage (bool): if True, skip checking that one of every event type for the
            chosen ``dialect`` can be found in the input document.

    Returns:
        dict: keyed with Entry IDs referring to an inner dict keyed with annotation types referring
        to the total number of times the annotation type can be found within the Entry.
    """

    import sys

    from lib.utils import BASIC_MAT_FAUX_ANCHOR_TYPES
    from lib.utils import BASIC_MAT_IGNORED_ANCHOR_TYPES
    from lib.utils import BASIC_MAT_KEY_ANCHOR_TYPES
    from lib.utils import GRANULAR_MAT_TEMPLATE_TYPES

    global DEBUG
    global EXPECTED_ABSTRACT_TYPES
    global EXTRA_ABSTRACT_TYPES
    global LOGGER
    global VERBOSE

    # Tracks which entries have examples of each annotation type, and how many examples there are
    # in that entry.
    enids_by_atype = {}
    # Tracks how many of each annotation type can be found per entry.
    atypes_by_enid = {}

    if dialect == "Abstract":

        expected_types = EXPECTED_ABSTRACT_TYPES

        for enid in doc.entries.value:

            atypes_by_enid[enid] = {k: 0 for k in expected_types}
            for type_ in EXTRA_ABSTRACT_TYPES:
                atypes_by_enid[enid][type_] = 0

            entry = doc.entries.value[enid]
            annot_sets = entry.annotation_sets.value
            if annot_sets:
                abstract_event_set = annot_sets.abstract_events.value
                if abstract_event_set:
                    events = abstract_event_set.events.value
                    if events:
                        for evid in events:
                            event = events[evid]
                            if event:
                                hh = event.helpful_harmful.value
                                mv = event.material_verbal.value
                                if hh and mv:
                                    if f"{hh}-{mv}" not in enids_by_atype:
                                        enids_by_atype[f"{hh}-{mv}"] = []
                                    enids_by_atype[f"{hh}-{mv}"].append(enid)
                                    atypes_by_enid[enid][f"{hh}-{mv}"] += 1

    elif dialect in ["Basic", "Granular",]:

        expected_types = BASIC_MAT_KEY_ANCHOR_TYPES + [
            k for k in GRANULAR_MAT_TEMPLATE_TYPES if dialect == "Granular"
        ]

        for enid in doc.entries.value:

            atypes_by_enid[enid] = {
                k: 0 for k in expected_types if k not in BASIC_MAT_FAUX_ANCHOR_TYPES
            }

            entry = doc.entries.value[enid]
            annot_sets = entry.annotation_sets.value
            if annot_sets:

                basic_event_set = annot_sets.basic_events.value
                if basic_event_set:

                    events = basic_event_set.events.value
                    if events:
                        for evid in events:

                            event = events[evid]
                            if event and event.event_type.value:

                                if event.event_type.value not in enids_by_atype:
                                    enids_by_atype[event.event_type.value] = []

                                enids_by_atype[event.event_type.value].append(enid)
                                atypes_by_enid[enid][event.event_type.value] += 1

                    gran_temps = basic_event_set.granular_templates.value
                    if dialect == "Granular" and gran_temps:
                        for gtid in gran_temps:

                            gt = gran_temps[gtid]
                            if gt and gt.type_.value:

                                if gt.type_.value not in enids_by_atype:
                                    enids_by_atype[gt.type_.value] = []

                                enids_by_atype[gt.type_.value].append(enid)
                                atypes_by_enid[enid][gt.type_.value] += 1

    miss = False
    for expected in expected_types:

        if dialect in ["Basic", "Granular",]:
            if expected in BASIC_MAT_FAUX_ANCHOR_TYPES or \
                expected in BASIC_MAT_IGNORED_ANCHOR_TYPES:

                continue

        elif expected not in enids_by_atype:
            miss = True
            enids_by_atype[expected] = []
            if not no_check_coverage:
                LOGGER.warning(("    "
                    f"Found {len(enids_by_atype[expected])} examples of '{expected}' in input "
                    "document."
                ))
                continue

        if DEBUG:
            LOGGER.debug((
                f"    Found {len(enids_by_atype[expected])} examples of '{expected}' in input "
                "document."
            ))

    if not no_check_coverage and miss == True:
        LOGGER.error(
            "Input document doesn't fully cover the annotation space! Pass "
            "'--no-ensure-indoc-coverage' to disregard."
        )
        sys.exit(1)

    return atypes_by_enid


def analyze_sentences(doc, showsents):
    """Gathers statistics about a BP document's sentence annotations.

    Args:
        doc (lib.bp.BPDocument): a document to analyze.
        showsents (bool): whether to show offending sentence details.

    Returns:
        collections.Counter: a :mod:`Counter` object that details information about Events with
        non-local SpanSets or Spans.
    """

    from collections import Counter
    import sys

    global DEBUG
    global VERBOSE
    global LOGGER

    counters = Counter()
    for enid in doc.entries.value:

        entry = doc.entries.value[enid]
        counters['entries'] += 1
        if DEBUG:
            LOGGER.debug(f"Processing {str(entry)}")

        segText = entry.segment_text.value
        if segText == None:
            LOGGER.error(f"Entry {str(entry)} has no segment-text!")
            sys.exit()

        sortedSegmentSections = sorted(entry.segment_sections.value, key=lambda x: x.start.value)

        for segSection in sortedSegmentSections:

            if segSection.structural_element.value == "Sentence":

                counters['sentences'] += 1
                sentStart = segSection.start.value
                sentEnd = segSection.end.value

                if DEBUG:
                    LOGGER.debug(("    "
                        f"analyzing sent: {sentStart}:{sentEnd} = <<{segText[sentStart:sentEnd]}>>"
                    ))

                annotSet = entry.annotation_sets.value
                eventSet = annotSet.basic_events.value
                events = eventSet.events.value
                spanSets = eventSet.span_sets.value
                for evtID, evt in events.items():

                    anchorsSpanSet = spanSets[evt.anchors.value]
                    anchorInSent = False
                    for span in anchorsSpanSet.spans.value:

                        start = span.start.value
                        end = span.end.value
                        if start >= sentStart and start < sentEnd and \
                            end > sentStart and end <= sentEnd:
                            anchorInSent = True

                    if anchorInSent:

                        counters['events'] += 1
                        if type(evt.ref_events.value) is list:
                            ref_events = [events[ref_evt_id] for ref_evt_id in evt.ref_events.value]
                        else:
                            ref_events = []
                        ref_anchor_ssids = [event.anchors.value for event in ref_events]

                        for ssid in evt.agents.value + evt.patients.value + ref_anchor_ssids:

                            counters['spansets'] += 1
                            spanSet = spanSets[ssid]
                            sentLocalSpan = False
                            for span in spanSet.spans.value:

                                start = span.start.value
                                end = span.end.value
                                if start >= sentStart and start < sentEnd and \
                                    end > sentStart and end <= sentEnd:
                                    sentLocalSpan = True

                            if not sentLocalSpan:

                                counters['events-with-nonlocal-spansets'] += 1
                                # As a last check, look for the exact string...
                                sentText = segText[sentStart:sentEnd]
                                sentLocalString = False
                                for span in spanSet.spans.value:

                                    if sentText.find(span.string_.value) >= 0:
                                        sentLocalString = True

                                if not sentLocalString:
                                    counters['events-with-nonlocal-spanstrings'] += 1

                                    if showsents:
                                        sentText = segText[sentStart:sentEnd].replace(
                                            "\r"," "
                                        ).replace("\n"," ")
                                        LOGGER.info(
                                            f"\nsentence {sentStart}:{sentEnd}=<<{sentText}>>"
                                        )
                                        LOGGER.info((
                                            f"event-id={evt.event_id.value} | anchors ssid="
                                            f"{evt.anchors.value}"
                                        ))
                                        for anchorSpan in spanSets[evt.anchors.value].spans.value:
                                            LOGGER.info(
                                                f"  anchor span: <<{anchorSpan.string_.value}>>"
                                            )
                                        LOGGER.info("non-local argument spans: ")
                                        for span in spanSet.spans.value:
                                            LOGGER.info((
                                                f"{span.start.value}:{span.end.value}=<<"
                                                f"{span.string_.value}>>"
                                            ))

    return counters


def dump_sentences(bp_path, out_path):
    """Generates a simple JSON representation of all the sentences in a BP document.

    Args:
        bp_path (str): a path to an input BP document.
        out_path (str): a path to output the JSON representation of the sentences.
    """

    import sys

    from lib import bp
    from lib.utils import write_json_blob

    global DEBUG
    global LOGGER
    global VERBOSE

    corpus = bp.BPDocument.from_json(bp_path)

    all_sents_list = []
    for enid in corpus.entries.value:
        if DEBUG:
            LOGGER.debug(f"processing BPDocumentEntry enid={enid}".format(enid))
        entry   = corpus.entries.value[enid]
        segText = entry.segment_text.value
        if segText == None:
            LOGGER.error(f"Entry {entry.entry_id.value} has no segment-text!")
            return
        entry_sents_list = []
        sortedSegmentSections = sorted(entry.segment_sections.value, key=lambda x: x.start.value)
        for segSection in sortedSegmentSections:
            if segSection.structural_element.value == "Sentence":
                sentText = segText[segSection.start.value:segSection.end.value]
                entry_sents_list.append(sentText)
        all_sents_list.append([enid, entry_sents_list])

    write_json_blob(all_sents_list, out_path, verbose=VERBOSE, force=FORCE)


def join(paths, corpus_id, no_validation=False):
    """Joins any number of BP documents into a single document.

    The documents are joined such that each Entry remains intact. If EntryIDs collide, a new unique
    ID will be created for the offending duplicate. Other information, like top-level provenance
    data, will be preserved. A new ``CORPUS_ID`` is required.

    \f

    Args:
        flat (bool): if True, emit a one-liner instead of indented JSON.
        no_recurse (bool): if True, don't look for files in subdirectories passed in ``mat_path``.
            Flag passed to ``lib.utils.resolve_paths()``, not used anywhere else.
        show_summary (bool): if True, printy summary statistics.
        allow_duplicates (bool): if True, don't dedupe the files.
        no_validation (bool): If True, skip validating constructed BP objects. Defaults to False.
        corpus_id (str): the new ID of the resulting BP document corpus.
        in_path (tuple): a list of paths to input BP documents or directories of BP documents to
            load.
        out_path (str): a path to output a new BP document.
    """
    from datetime import date
    import os
    import sys

    from lib import bp
    from lib.utils import MAT_DIALECT_CHOICES
    from lib.utils import TQDM_PREFIX
    from lib.utils import get_tqdm

    global DEBUG
    global LOGGER
    global VERBOSE

    tqdm = get_tqdm(VERBOSE, DEBUG)

    provenance = bp.BPDocumentProvenance(
        corpus_creation_date=str(date.today()),
        no_validation=no_validation
    )
    provenance.add_misc_key("join-sources", [])
    out_doc = bp.BPDocument(
        f"{corpus_id}", {}, provenance=provenance, no_validation=no_validation
    )

    if DEBUG:
        LOGGER.debug("Ingesting documents")

    collisions = {}
    for path in tqdm(paths, desc=f"{TQDM_PREFIX} Ingesting documents", unit="doc"):

        in_doc = bp.BPDocument.from_json(path, no_validation=no_validation)

        if not no_validation:
            if not in_doc.is_valid():
                LOGGER.warning(f"Document at path {path} failed validation.")
                continue

        entries = in_doc.entries.value
        local_cols = 0
        if entries:
            for enid in entries:

                if enid in out_doc.entries.value:

                    local_cols += 1

                    if enid not in collisions:
                        collisions[enid] = 1

                    enid += f"-{collisions[enid]}"
                    collisions[enid] += 1

                out_doc.entries.value[enid] = entries[enid]

            out_doc.provenance.value.misc_keys["join-sources"].append(os.path.basename(path))

            if DEBUG:
                LOGGER.debug(("    "
                    f"Ingested {len(entries.keys())} Entries ({local_cols} EnID collisions) from "
                    f"{str(in_doc)} ({path})"
                ))
        else:

            if DEBUG:
                LOGGER.debug(("    "
                    f"No Entries in {str(in_doc)} ({path})"
                ))

    return out_doc


def make_deliverables(
    in_doc,
    dialect,
    depth=DEPTH_DEFAULT,
    dist_context=None,
    dist_provided=DIST_PROVIDED_DEFAULT,
    dist_provided_analysis=DIST_PROVIDED_ANALYSIS_DEFAULT,
    dist_provided_devtest=DIST_PROVIDED_DEVTEST_DEFAULT,
    no_ensure_indoc_coverage=False,
    no_validation=False
):
    """Splits a BP document into the standard deliverable set of documents.

    Entries are left intact.

    Args:
        in_doc (lib.bp.BPDocument): an input document to split.
        dialect (str): the MAT task the BP document was created with. Must be one of
            ``utils.MAT_DIALECT_CHOICES``, which is verified by the Click decorator.
        depth (int): a positive integer representing the size of the random subspace to check while
            performing pseudo-optimization.
        dist_context (dict): a Distribution dictionary, which is a JSON blob of entry assignment
            information about a previous distribution. Entry IDs found in both the new document and
            the previous distribution will retain their previous assignments.
        dist_provided (str): the percentage of the input the provided partition should be, expressed
            as 'x.x%'.
        dist_provided_analysis (str): the percentage of the provided partition the provided-analysis
            partition should be, expressed as 'x.x%'.
        dist_provided_devtest (str): the percentage of the provided partition the provided-devtest
            partition should be, expressed as 'x.x%'.
        no_ensure_indoc_coverage (bool): if True, skip ensuring that one of every event type for the
            chosen Dialect can be found in the input document.
        no_validation (bool): If True, skip validating constructed BP objects. Defaults to False.

    Returns:
        tuple: a 4-tuple of :class:`lib.bp.BPDocument` objects representing hidden, provided-train,
        provided-analysis, provided-devtest partitions, respectively.

    .. Notes:
       - Distributions:
           - hidden -> defaults to 28% of input
           - provided -> defaults to 72% of input
               - provided-train -> defaults to 70% of provided
               - provided-analysis-ref -> defaults to 15% of provided
               - provided-analysis-empty -> clone of provided-analysis-ref, but without annotations
               - provided-devtest-ref -> defaults to 15% of provided
               - provided-devtest-empty -> clone of provided-devtest-ref, but without annotations
    """
    from copy import deepcopy
    from datetime import date
    import math
    import random
    import sys

    from lib import bp
    from lib.utils import MAT_DIALECT_CHOICES
    from lib.utils import TQDM_PREFIX
    from lib.utils import get_tqdm

    global DEBUG
    global LOGGER
    global VERBOSE

    ## setup

    def __find_widest_entry(remaining_enids, uncovered_atypes, enids_by_atype, atypes_by_enid):
        """Finds the Entry with the most annotations types. Not necessarily the one with the most
        instances of annotations. It must have at least 1 annotation of an uncovered type.
        """

        assert len(remaining_enids) > 0, "Can't call this with no remaining ids!"
        assert len(uncovered_atypes) > 0, "Can't call this with no remaining atypes!"

        widest = None
        max_width = 0
        for enid in remaining_enids:

            width = sum([1 for atype in uncovered_atypes if atypes_by_enid[enid][atype]])

            if width > max_width:
                max_width = width
                widest = enid

        return widest

    def __check_dist_context(dist_context):
        """Checks that this previous distribution context is valid."""

        assert "dialect" in dist_context, "Missing key 'dialect' from top-level!"
        assert "dist-creation-date" in dist_context, \
            "Missing key 'dist-creation-date' from top-level!"
        assert "input-corpus-id" in dist_context, "Missing key 'input-corpus-id' from top-level!"
        assert "input-entry-ids" in dist_context, "Missing key 'input-entry-ids' from top-level!"
        assert "partitions" in dist_context, "Missing key 'partitions' from top-level!"
        assert "pdist-provided" in dist_context, "Missing key 'pdist-provided' from top-level!"
        assert "pdist-provided-analysis" in dist_context, \
            "Missing key 'pdist-provided-analysis' from top-level!"
        assert "pdist-provided-devtest" in dist_context, \
            "Missing key 'pdist-provided-devtest' from top-level!"

        assert "hidden" in dist_context['partitions'], "Missing 'hidden' from the Partitions list!"
        assert "provided-analysis" in dist_context['partitions'], \
            "Missing 'provided-analysis' from the Partitions list!"
        assert "provided-devtest" in dist_context['partitions'], \
            "Missing 'provided-devtest' from the Partitions list!"
        assert "provided-train" in dist_context['partitions'], \
            "Missing 'provided-train' from the Partitions list!"

        assert "corpus-id" in dist_context['partitions']["hidden"], \
            "Missing 'corpus-id' from the 'hidden' Partition!"
        assert "corpus-id" in dist_context['partitions']["provided-analysis"], \
            "Missing 'corpus-id' from the 'provided-analysis' Partition!"
        assert "corpus-id" in dist_context['partitions']["provided-devtest"], \
            "Missing 'corpus-id' from the 'provided-devtest' Partition!"
        assert "corpus-id" in dist_context['partitions']["provided-train"], \
            "Missing 'corpus-id' from the 'provided-train' Partition!"

        assert "entry-ids" in dist_context['partitions']["hidden"], \
            "Missing 'entry-ids' from the 'hidden' Partition!"
        assert "entry-ids" in dist_context['partitions']["provided-analysis"], \
            "Missing 'entry-ids' from the 'provided-analysis' Partition!"
        assert "entry-ids" in dist_context['partitions']["provided-devtest"], \
            "Missing 'entry-ids' from the 'provided-devtest' Partition!"
        assert "entry-ids" in dist_context['partitions']["provided-train"], \
            "Missing 'entry-ids' from the 'provided-train' Partition!"

    tqdm = get_tqdm(VERBOSE, DEBUG)

    if not no_validation:
        if not in_doc.is_valid():
            sys.exit()

    assert dialect in MAT_DIALECT_CHOICES

    if dist_context:

        assert type(dist_context) == dict

        if VERBOSE:
            LOGGER.info("Parsing previous distribution's context data")

        __check_dist_context(dist_context)

        original_dist_context = deepcopy(dist_context)

        ## load dist percentages from previous data-delivery context

        pdist_provided = dist_context['pdist-provided']
        pdist_provided_analysis = dist_context['pdist-provided-analysis']
        pdist_provided_devtest = dist_context['pdist-provided-devtest']

        if DEBUG:
            LOGGER.debug(
                "    Overriding percentage distribution values with those from the context!"
            )
            LOGGER.debug(f"        pdist_provided: {pdist_provided}")
            LOGGER.debug(f"        pdist_provided_analysis: {pdist_provided_analysis}")
            LOGGER.debug(f"        pdist_provided_devtest: {pdist_provided_devtest}")

        ## process filename updates, if any

        if "enid-updates" in dist_context:

            if DEBUG:
                LOGGER.debug(
                    "    Updating EntryIDs"
                )

            updates = dist_context["enid-updates"]

            dist_context["input-entry-ids"] = [
                updates[enid] if enid in updates else enid \
                    for enid in dist_context["input-entry-ids"]
            ]

            dist_context["partitions"]["hidden"]["entry-ids"] = [
                updates[enid] if enid in updates else enid \
                    for enid in dist_context["partitions"]["hidden"]["entry-ids"]
            ]

            dist_context["partitions"]["provided-analysis"]["entry-ids"] = [
                updates[enid] if enid in updates else enid \
                    for enid in dist_context["partitions"]["provided-analysis"]["entry-ids"]
            ]

            dist_context["partitions"]["provided-devtest"]["entry-ids"] = [
                updates[enid] if enid in updates else enid \
                    for enid in dist_context["partitions"]["provided-devtest"]["entry-ids"]
            ]

            dist_context["partitions"]["provided-train"]["entry-ids"] = [
                updates[enid] if enid in updates else enid \
                    for enid in dist_context["partitions"]["provided-train"]["entry-ids"]
            ]

    else:

        ## convert percentages

        try:
            pdist_provided = float(dist_provided[:-1]) / 100
        except ValueError as err:
            LOGGER.error(f"dist_provided: {err}")
            sys.exit()

        try:
            pdist_provided_analysis = float(dist_provided_analysis[:-1]) / 100
        except ValueError as err:
            LOGGER.error(f"dist_provided_analysis: {err}")
            sys.exit()

        try:
            pdist_provided_devtest = float(dist_provided_devtest[:-1]) / 100
        except ValueError as err:
            LOGGER.error(f"dist_provided_devtest: {err}")
            sys.exit()

    assert pdist_provided < 1.0, "Provided distribution percentage too high!"
    assert pdist_provided_analysis < 0.5, "Analysis distribution percentage too high!"
    assert pdist_provided_devtest < 0.5, "Devtest distribution percentage too high!"

    ## ensure coverage, scan for annotation types

    if VERBOSE:
        LOGGER.info("Mapping input doc...")

    atypes_by_enid = __find_annotations(in_doc, dialect, no_ensure_indoc_coverage)

    ## First, decide which entries should go in the provided partition. They must have at least 1
    ## exemplar for each annotation type (type list decided by dialect).

    if DEBUG:
        LOGGER.debug("Assigning minimum required Entry IDs...")

    hidden_enids = []
    provided_train_enids = []
    provided_analysis_enids = []
    provided_devtest_enids = []

    assigned_enids = []
    total = len(atypes_by_enid.keys())
    assert total >= 4, "Not enough entries in the document to partition!"

    hidden_threshold = round(total * (1-pdist_provided))
    provided_train_threshold = round(
        total * pdist_provided * (1 - pdist_provided_analysis - pdist_provided_devtest)
    )
    provided_analysis_threshold = round(total * pdist_provided * pdist_provided_analysis)
    provided_devtest_threshold = round(total * pdist_provided * pdist_provided_devtest)

    thsum = sum([
        hidden_threshold,
        provided_train_threshold,
        provided_analysis_threshold,
        provided_devtest_threshold,
    ])
    if thsum < total:
        provided_train_threshold += 1
        thsum = sum([
            hidden_threshold,
            provided_train_threshold,
            provided_analysis_threshold,
            provided_devtest_threshold,
        ])

    assert thsum >= total, f"{thsum} != {total}"

    enids_by_atype = {}
    atype_appearance_totals = {}
    atype_totals = {}
    for enid in atypes_by_enid:
        for atype in atypes_by_enid[enid]:

            if atype not in atype_appearance_totals:
                atype_appearance_totals[atype] = 0
            if atype not in atype_totals:
                atype_totals[atype] = 0

            if atypes_by_enid[enid][atype]:
                atype_appearance_totals[atype] += 1
            if atypes_by_enid[enid][atype]:
                atype_totals[atype] += atypes_by_enid[enid][atype]

            if atype not in enids_by_atype:
                enids_by_atype[atype] = {}

            if enid not in enids_by_atype[atype]:
                enids_by_atype[atype][enid] = atypes_by_enid[enid][atype]

    ## apply existing assignments from previous distribution context, if any

    if dist_context:

        if DEBUG:
            LOGGER.debug("    Loading assignments from previous distribution context")

        for enid in dist_context["partitions"]["hidden"]["entry-ids"]:
            if enid in atypes_by_enid:
                hidden_enids.append(enid)
                assigned_enids.append(enid)
                if DEBUG:
                    LOGGER.debug(f"        Placed '{enid}' in hidden.")

        for enid in dist_context["partitions"]["provided-analysis"]["entry-ids"]:
            if enid in atypes_by_enid:
                provided_analysis_enids.append(enid)
                assigned_enids.append(enid)
                if DEBUG:
                    LOGGER.debug(f"        Placed '{enid}' in provided-analysis.")

        for enid in dist_context["partitions"]["provided-devtest"]["entry-ids"]:
            if enid in atypes_by_enid:
                provided_devtest_enids.append(enid)
                assigned_enids.append(enid)
                if DEBUG:
                    LOGGER.debug(f"        Placed '{enid}' in provided-devtest.")

        for enid in dist_context["partitions"]["provided-train"]["entry-ids"]:
            if enid in atypes_by_enid:
                provided_train_enids.append(enid)
                assigned_enids.append(enid)
                if DEBUG:
                    LOGGER.debug(f"        Placed '{enid}' in provided-train.")

        if DEBUG:
            LOGGER.debug(f"        Loaded {len(assigned_enids)} assignments.")

    if not sorted(assigned_enids) == sorted(list(atypes_by_enid.keys())):

        ## search for atypes that only appear in a small # of Entries ("minimum assignment")

        if DEBUG:
            LOGGER.debug("    Looking for atypes with only 1 example.")

        for atype in atype_appearance_totals:

            # these _must_ go in provided_training
            if atype_appearance_totals[atype] == 1:
                for enid in enids_by_atype[atype]:
                    if enids_by_atype[atype][enid]:

                        if dist_context and enid in assigned_enids:

                            if DEBUG:
                                LOGGER.debug(("        "
                                    f"Entry with only '{atype}' example already assigned by "
                                    "previous dist context! Skipping!"
                                ))
                                if enid in hidden_enids:
                                    LOGGER.debug(f"            {enid} -> hidden")
                                elif enid in provided_train_enids:
                                    LOGGER.debug(f"            {enid} -> provided-train")
                                elif enid in provided_analysis_enids:
                                    LOGGER.debug(f"            {enid} -> provided-analysis")
                                elif enid in provided_devtest_enids:
                                    LOGGER.debug(f"            {enid} -> provided-devtest")
                                else:
                                    raise RuntimeError

                        elif enid not in assigned_enids:
                            provided_train_enids.append(enid)
                            assigned_enids.append(enid)
                            if DEBUG:
                                LOGGER.debug(("        "
                                    f"Placed '{enid}' in provided-train since it has the only "
                                    f"example of {atype} in the input doc."
                                ))
                        elif DEBUG:
                            LOGGER.debug(("        "
                                f"Attempted to place '{enid}' in provided-train since it has the "
                                f"only example of {atype} in the input doc, but it's already "
                                "present."
                            ))

        if len(provided_train_enids) > provided_train_threshold:
            LOGGER.warning(("        "
                f"Event IDs assigned to provided-train ({len(provided_train_enids)}) exceeds "
                f"percentage threshold ({provided_train_threshold})!"
            ))

        if DEBUG:
            LOGGER.debug("    Looking for atypes with only 2 examples.")

        for atype in atype_appearance_totals:

            # place 1 in provided_training and another in hidden
            if atype_appearance_totals[atype] == 2:

                temp_enids = []
                for enid in enids_by_atype[atype]:
                    if enids_by_atype[atype][enid]:

                        if dist_context and enid in assigned_enids:

                            if DEBUG:
                                LOGGER.debug(("        "
                                    f"Entry with one of only two examples for '{atype}' already "
                                    "assigned by previous dist context! Skipping!"
                                ))
                                if enid in hidden_enids:
                                    LOGGER.debug(f"            {enid} -> hidden")
                                elif enid in provided_train_enids:
                                    LOGGER.debug(f"            {enid} -> provided-train")
                                elif enid in provided_analysis_enids:
                                    LOGGER.debug(f"            {enid} -> provided-analysis")
                                elif enid in provided_devtest_enids:
                                    LOGGER.debug(f"            {enid} -> provided-devtest")
                                else:
                                    raise RuntimeError

                        else:

                            temp_enids.append(enid)

                if len(temp_enids) == 2:

                    # assign "heavier" entry to the provided partition
                    weight_1 = sum(
                        [atypes_by_enid[temp_enids[0]][a] for a in atypes_by_enid[temp_enids[0]]]
                    )
                    weight_2 = sum(
                        [atypes_by_enid[temp_enids[1]][a] for a in atypes_by_enid[temp_enids[1]]]
                    )
                    if weight_1 >= weight_2:

                        if temp_enids[0] not in assigned_enids:
                            provided_train_enids.append(temp_enids[0])
                            assigned_enids.append(temp_enids[0])
                            if DEBUG:
                                LOGGER.debug(("        "
                                    f"Placed '{temp_enids[0]}' in provided-train since it has one "
                                    f"of only two examples of {atype} in the input doc."
                                ))
                        else:
                            if DEBUG:
                                LOGGER.debug(("        "
                                    f"Attempted to place '{temp_enids[0]}' in provided-train since "
                                    f"it has one of only two examples of {atype} in the input doc, "
                                    "but it's already assigned!"
                                ))

                            # there's only one place where it could be if not in provided-train
                            if temp_enids[0] not in provided_train_enids:
                                hidden_enids.remove(temp_enids[0])
                                provided_train_enids.append(temp_enids[0])
                                if DEBUG:
                                    LOGGER.debug(("            "
                                        f"Removed '{temp_enids[0]}' from hidden and placed in "
                                        f"provided-train!"
                                    ))
                            elif DEBUG:
                                LOGGER.debug("            Already in provided-train.")

                        if temp_enids[1] not in assigned_enids:
                            hidden_enids.append(temp_enids[1])
                            assigned_enids.append(temp_enids[1])
                            if DEBUG:
                                LOGGER.debug(("        "
                                    f"Placed '{temp_enids[1]}' in hidden since it has one of only "
                                    f"two examples of {atype} in the input doc."
                                ))
                        elif DEBUG:
                            LOGGER.debug(("        "
                                f"Attempted to place '{temp_enids[1]}' in hidden since it has one "
                                f"of only two examples of {atype} in the input doc, but it's "
                                "already assigned!"
                            ))
                    else:

                        if temp_enids[1] not in assigned_enids:
                            provided_train_enids.append(temp_enids[1])
                            assigned_enids.append(temp_enids[1])
                            if DEBUG:
                                LOGGER.debug(("        "
                                    f"Placed '{temp_enids[1]}' in provided-train since it has one "
                                    f"of only two examples of {atype} in the input doc."
                                ))
                        else:
                            if DEBUG:
                                LOGGER.debug(("        "
                                    f"Attempted to place '{temp_enids[1]}' in provided-train since "
                                    f"it has one of only two examples of {atype} in the input doc, "
                                    "but it's already assigned!"
                                ))
                            # there's only one place where it could be if not in provided-train
                            if temp_enids[1] not in provided_train_enids:
                                hidden_enids.remove(temp_enids[1])
                                provided_train_enids.append(temp_enids[1])
                                if DEBUG:
                                    LOGGER.debug(("            "
                                        f"Removed '{temp_enids[1]}' from hidden and placed in "
                                        f"provided-train!"
                                    ))
                            elif DEBUG:
                                LOGGER.debug("            Already in provided-train.")

                        if temp_enids[0] not in assigned_enids:
                            hidden_enids.append(temp_enids[0])
                            assigned_enids.append(temp_enids[0])
                            if DEBUG:
                                LOGGER.debug(("        "
                                    f"Placed '{temp_enids[0]}' in hidden since it has one of only "
                                    f"two examples of {atype} in the input doc."
                                ))
                        elif DEBUG:
                            LOGGER.debug(("        "
                                f"Attempted to place '{temp_enids[0]}' in hidden since it has one "
                                f"of only two examples of {atype} in the input doc, but it's "
                                "already assigned!"
                            ))

                elif len(temp_enids) == 1:

                    if temp_enids[0] not in provided_train_enids:

                        provided_train_enids.append(temp_enids[0])
                        assigned_enids.append(temp_enids[0])
                        if DEBUG:
                            LOGGER.debug(("        "
                                f"Placed '{temp_enids[0]}' in provided-train since provided-train "
                                f"had no examples of {atype} yet."
                            ))

                    else:

                        provided_train_enids.append(temp_enids[0])
                        assigned_enids.append(temp_enids[0])
                        if DEBUG:
                            LOGGER.debug(("        "
                                f"Placed '{temp_enids[0]}' in hidden since provided-train already "
                                f"had an example of {atype}."
                            ))

        if len(provided_train_enids) > provided_train_threshold:
            LOGGER.warning(("        "
                f"Number of Events assigned to provided-train ({len(provided_train_enids)}) exceeds"
                f" threshold ({provided_train_threshold})!"
            ))

        ## Check for coverage in provided-train after performing minimum assignment, insert more
        ## entries if necessary.

        if DEBUG:
            LOGGER.debug("    Looking for entries with remaining uncovered atypes.")

        covered_atypes = []
        for enid in assigned_enids:
            for atype in atypes_by_enid[enid]:
                if atypes_by_enid[enid][atype]:
                    covered_atypes.append(atype)

        uncovered_atypes = [k for k in enids_by_atype.keys() if k not in covered_atypes]
        remaining_enids = [k for k in atypes_by_enid.keys() if k not in assigned_enids]
        empties = [
            atype for atype in uncovered_atypes if sum(
                [enids_by_atype[atype][enid] for enid in remaining_enids]
            ) == 0
        ]
        uncovered_atypes = [k for k in uncovered_atypes if k not in empties]

        ## add more required entries to `provided-train` until covered

        while uncovered_atypes:

            if remaining_enids:

                ## Pick the 'widest' entry with this atype, or the one with the largest number of
                ## different atypes, not necessarily the one with the _most_ atypes.
                enid = __find_widest_entry(
                    remaining_enids, uncovered_atypes, enids_by_atype, atypes_by_enid
                )
                provided_train_enids.append(enid )
                assigned_enids.append(enid)
                remaining_enids.remove(enid)

                if DEBUG:
                    LOGGER.debug(("        "
                        f"Placed '{enid}' in provided-train since it has coverage of uncovered "
                        "atypes."
                    ))

                now_covered = []
                for atype in uncovered_atypes:
                    if atypes_by_enid[enid][atype]:
                        uncovered_atypes.remove(atype)
                        now_covered.append(atype)

                if DEBUG:
                    LOGGER.debug(("            "
                        f"Atypes now covered: {now_covered}"
                    ))

            else:
                raise RuntimeError(
                    "No Event IDs left to consider while attempting to ensure provided-train " + \
                    "coverage!"
                )

        ## distribute the rest randomly

        if remaining_enids:

            if DEBUG:
                LOGGER.debug("    Distributing the rest of the Entries randomly...")

            def __helper(part, thld, remaining, assigned):

                if remaining:

                    # The combinatinatoric space is often so large that we don't need to memoize
                    # states.
                    if len(part) < thld:
                        sample = random.sample(
                            remaining,
                            min(thld-len(part), len(remaining))
                        )
                        remaining = [k for k in remaining if k not in sample]
                        part.extend(sample)
                        assigned.extend(sample)

                return part, remaining, assigned

            def __delta_partition(part, pdist):

                deltas = []
                for atype in atype_totals:

                    atype_total = atype_totals[atype]
                    if atype_total:

                        # get delta_atype for this partition
                        deltas.append(
                            abs(
                                sum(
                                    [enids_by_atype[atype][enid] for enid in part]
                                ) / float(atype_total) - pdist
                            )
                        )

                    else:
                        deltas.append(0)

                return sum(deltas)

            pdist_hidden = 1 - pdist_provided
            pdist_provided_train = 1 - pdist_provided_devtest - pdist_provided_analysis
            lowest_delta = float('inf')
            best_assignments = {
                "hidden": None,
                "train": None,
                "devtest": None,
                "analysis": None,
            }

            for _ in tqdm(range(depth), desc=f"{TQDM_PREFIX}     Searching subspace", unit="state"):

                ## set up clones of state information

                state_hidden_enids = deepcopy(hidden_enids)
                state_provided_train_enids = deepcopy(provided_train_enids)
                state_provided_devtest_enids = deepcopy(provided_devtest_enids)
                state_provided_analysis_enids = deepcopy(provided_analysis_enids)
                state_remaining_enids = deepcopy(remaining_enids)
                state_assigned_enids = deepcopy(assigned_enids)

                ## make random assignments

                state_provided_train_enids, state_remaining_enids, state_assigned_enids = __helper(
                    state_provided_train_enids,
                    provided_train_threshold,
                    state_remaining_enids,
                    state_assigned_enids
                )
                state_provided_analysis_enids, state_remaining_enids, state_assigned_enids = \
                    __helper(
                        state_provided_analysis_enids,
                        provided_analysis_threshold,
                        state_remaining_enids,
                        state_assigned_enids
                    )
                state_provided_devtest_enids, state_remaining_enids, state_assigned_enids = \
                    __helper(
                        state_provided_devtest_enids,
                        provided_devtest_threshold,
                        state_remaining_enids,
                        state_assigned_enids
                    )
                state_hidden_enids, state_remaining_enids, state_assigned_enids = __helper(
                    state_hidden_enids,
                    hidden_threshold,
                    state_remaining_enids,
                    state_assigned_enids
                )

                ## calc aggregate delta

                delta_agg = sum([
                    __delta_partition(state_hidden_enids, pdist_hidden),
                    __delta_partition(state_provided_train_enids, pdist_provided_train),
                    __delta_partition(state_provided_devtest_enids, pdist_provided_devtest),
                    __delta_partition(state_provided_analysis_enids, pdist_provided_analysis),
                ])
                if delta_agg < lowest_delta:

                    if DEBUG:
                        LOGGER.debug(f"        Found new lowest state delta: {delta_agg}")

                    lowest_delta = delta_agg

                    best_assignments["hidden"] = deepcopy(state_hidden_enids)
                    best_assignments["train"] = deepcopy(state_provided_train_enids)
                    best_assignments["devtest"] = deepcopy(state_provided_devtest_enids)
                    best_assignments["analysis"] = deepcopy(state_provided_analysis_enids)

            ## apply assignments

            if DEBUG:
                LOGGER.debug(f"        Making assignments according to best subspace state found")

            for enid in best_assignments['hidden']:
                if enid not in hidden_enids:
                    hidden_enids.append(enid)
                    assigned_enids.append(enid)
                    remaining_enids.remove(enid)

            for enid in best_assignments['train']:
                if enid not in provided_train_enids:
                    provided_train_enids.append(enid)
                    assigned_enids.append(enid)
                    remaining_enids.remove(enid)

            for enid in best_assignments['devtest']:
                if enid not in provided_devtest_enids:
                    provided_devtest_enids.append(enid)
                    assigned_enids.append(enid)
                    remaining_enids.remove(enid)

            for enid in best_assignments['analysis']:
                if enid not in provided_analysis_enids:
                    provided_analysis_enids.append(enid)
                    assigned_enids.append(enid)
                    remaining_enids.remove(enid)

        assert len(assigned_enids) == total, f"{len(assigned_enids)} != {total}"
        assert len(remaining_enids) == 0, f"Unassigned EventIDs! {remaining_enids}"

    ## emit

    if VERBOSE:
        LOGGER.info("Spooling partition documents...")

    corpus_id_prefix = in_doc.corpus_id.value
    if corpus_id_prefix:
        corpus_id_prefix += ", "

    provenance = in_doc.provenance.value
    if provenance:
        provenance.corpus_creation_date.value = str(date.today())
        provenance.source.value = in_doc.corpus_id.value

    def __helper(cid, prov, no_v, enids):
        doc = bp.BPDocument(
            f"{corpus_id_prefix}{cid}", {}, provenance=prov, no_validation=no_v
        )
        for enid in enids:
            doc.entries.value[enid] = deepcopy(in_doc.entries.value[enid])
        return doc

    hidden = __helper("Hidden", provenance, no_validation, hidden_enids)
    provided_train = __helper("Provided Train", provenance, no_validation, provided_train_enids)
    provided_analysis = __helper(
        "Provided Analysis Ref", provenance, no_validation, provided_analysis_enids
    )
    provided_devtest = __helper(
        "Provided Devtest Ref", provenance, no_validation, provided_devtest_enids
    )

    if DEBUG:
        LOGGER.debug("Inserting Entries into partition documents")

    for enid in tqdm(
        atypes_by_enid,
        desc=f"{TQDM_PREFIX} Inserting Entries into partitions",
        unit="entry"
    ):

        if enid in provided_train_enids:

            if DEBUG:
                LOGGER.debug(f"    Placing entry '{enid}' in Provided-train.")
            provided_train.entries.value[enid] = in_doc.entries.value[enid]

        elif enid in provided_analysis_enids:

            if DEBUG:
                LOGGER.debug(f"    Placing entry '{enid}' in Provided-analysis.")
            provided_analysis.entries.value[enid] = in_doc.entries.value[enid]

        elif enid in provided_devtest_enids:

            if DEBUG:
                LOGGER.debug(f"    Placing entry '{enid}' in Provided-devtest.")
            provided_devtest.entries.value[enid] = in_doc.entries.value[enid]

        elif enid in hidden_enids:

            if DEBUG:
                LOGGER.debug(f"    Placing entry '{enid}' in Hidden.")
            hidden.entries.value[enid] = in_doc.entries.value[enid]

        else:

            if DEBUG:
                LOGGER.critical(f"    Found unassigned Entry ID '{enid}'!")
            raise RuntimeError

    ## make dist record

    dist = {
        "dialect": dialect,
        "dist-creation-date": str(date.today()),
        "input-corpus-id": in_doc.corpus_id.value,
        "input-entry-ids": assigned_enids,
        "partitions": {
            "hidden": {
                "corpus-id": hidden.corpus_id.value,
                "entry-ids": hidden_enids,
            },
            "provided-train": {
                "corpus-id": provided_train.corpus_id.value,
                "entry-ids": provided_train_enids,
            },
            "provided-analysis": {
                "corpus-id": provided_analysis.corpus_id.value,
                "entry-ids": provided_analysis_enids,
            },
            "provided-devtest": {
                "corpus-id": provided_devtest.corpus_id.value,
                "entry-ids": provided_devtest_enids,
            },
        },
        "pdist-provided": pdist_provided,
        "pdist-provided-analysis": pdist_provided_analysis,
        "pdist-provided-devtest": pdist_provided_devtest,
    }

    if dist_context:
        dist["dist-context"] = original_dist_context

    __check_dist_context(dist)

    return hidden, provided_train, provided_analysis, provided_devtest, dist


def split(
    in_doc,
    amount,
    dialect,
    no_ensure_indoc_coverage=False,
    no_validation=False
):
    """Splits a BP document into two documents.

    Entries are left intact. Some will go to document A, others will go to document B. The selection
    for which goes where is decided randomly.

    Args:
        in_doc (lib.bp.BPDocument): an input document to split.
        amount (str): the number of entries within the input BP document to reserve for the B
            partition, either represented as an integer or a percentage written as "x.x%". If an
            ``int`` is passed that is higher than the number of entries available, it will be
            treated as if ``100%`` was passed. If a percentage is passed, it will ``floor`` to the
            nearest whole entry.
        dialect (str): the MAT task the BP document was created with. Must be one of
            ``utils.MAT_DIALECT_CHOICES``, which is verified by the Click decorator.
        no_ensure_indoc_coverage (bool): if True, skip ensuring that one of every event type for the
            chosen Dialect can be found in the input document.
        no_validation (bool): If True, skip validating constructed BP objects. Defaults to False.

    Returns:
        tuple: a 2-tuple of :class:`lib.bp.BPDocument` values for docs A and B.
    """
    from copy import deepcopy
    from datetime import date
    import math
    import random
    import sys

    from lib import bp
    from lib.utils import MAT_DIALECT_CHOICES
    from lib.utils import TQDM_PREFIX
    from lib.utils import get_tqdm

    global DEBUG
    global LOGGER
    global VERBOSE

    ## setup

    tqdm = get_tqdm(VERBOSE, DEBUG)

    if not no_validation:
        if not in_doc.is_valid():
            sys.exit()

    assert dialect in MAT_DIALECT_CHOICES

    passed_amount = deepcopy(amount)
    corpus_id = in_doc.corpus_id.value
    provenance = in_doc.provenance.value
    if provenance:
        provenance.corpus_creation_date.value = str(date.today())
        provenance.source.value = corpus_id

    ## validate `amount`

    if not in_doc.entries.value:
        if VERBOSE:
            LOGGER.warning(f"Input document has no entries!")
        return doc_a, doc_b

    total = len(in_doc.entries.value)
    if DEBUG:
        LOGGER.debug(f"Found {total} entries in input document.")

    if amount.endswith("%"):

        if DEBUG:
            LOGGER.debug("    Got percentage value for amount!")

        try:
            pamount = float(amount[:-1]) / 100
        except ValueError as err:
            LOGGER.error(f"amount: {err}")
            sys.exit()

        if pamount < 0:
            pamount = 0
        elif pamount > 1:
            pamount = 1

        amount = int(math.ceil(len(in_doc.entries.value.keys())*pamount))

    else:

        if DEBUG:
            LOGGER.debug("    Got integer for amount!")

        try:
            amount = int(amount)
        except ValueError as err:
            LOGGER.error(f"amount: {err}")
            sys.exit()

        if amount > total:
            amount = total
        elif amount < 0:
            amount = 0

        pamount = float(amount) / total

    if DEBUG:
        LOGGER.debug(f"    Passed split amount {passed_amount} resolves to {pamount}.")

    ## ensure coverage of types

    if not no_ensure_indoc_coverage:

        if VERBOSE:
            LOGGER.info("Checking for annotation type coverage...")

    else:
        if VERBOSE:
            LOGGER.info("Mapping input doc...")

    __find_annotations(in_doc, dialect, no_ensure_indoc_coverage)

    ## split to docs A and B

    if amount == 0:

        if VERBOSE:
            LOGGER.warning(f"    Doc A will be identical to input doc and Doc B will be empty!")

    elif amount == total:

        if VERBOSE:
            LOGGER.warning(f"    Doc A will be empty and Doc B will be identical to input doc!")

    doc_a = bp.BPDocument(
        f"{corpus_id}-A", {}, provenance=provenance, no_validation=no_validation
    )
    doc_b = bp.BPDocument(
        f"{corpus_id}-B", {}, provenance=provenance, no_validation=no_validation
    )

    if VERBOSE:
        LOGGER.info(f"Building output documents ")

    b_dist = random.sample(in_doc.entries.value.keys(), amount)
    for enid in tqdm(
        in_doc.entries.value,
        desc=f"{TQDM_PREFIX} Splitting with {pamount*100}% random sample ",
        unit="entry"
    ):
        if enid in b_dist:
            if DEBUG:
                LOGGER.debug(f"        Placing entry '{enid}' in Doc B.")
            doc_b.entries.value[enid] = deepcopy(in_doc.entries.value[enid])
        else:
            if DEBUG:
                LOGGER.debug(f"        Placing entry '{enid}' in Doc A.")
            doc_a.entries.value[enid] = deepcopy(in_doc.entries.value[enid])

    return doc_a, doc_b


@click.group(context_settings=CONTEXT_SETTINGS)
@click.option('-v', '--verbose', count=True, help="Show verbose output. Pass again for debug.")
@click.option('-f', '--force', is_flag=True, help="Skip asking to overwrite output.")
@click.option(
    '-l',
    '--save-logs',
    'save_logs',
    type=click.Path(),
    help="Capture logging output (at whatever verbosity level) to the specified path."
)
def cli(verbose, force, save_logs):
    """Mutates BP documents in various ways.

    \f

    This function serves as the entrypoint for the program when called on the CLI. It also handles
    global flag ``verbose``.

    Args:
        verbose (int): if 1, pretty-print output as the command runs. If 2, print debugging output
            in addition to level 1 output. If 0, prints nothing as the command runs. Defaults to 0.
        force (bool): if True, skip asking for permission to overwrite a file with output. Defaults
            to False.
        save_logs (str): a path to write captured logging output to.
    """
    # do this before anything else
    from lib.utils import capture_logs
    if save_logs:
        capture_logs(save_logs, verbose, force)

    global DEBUG
    global FORCE
    global LOGGER
    global VERBOSE

    if verbose:
        LOGGER.info("Verbose output is on!")
        if verbose >= 1:
            VERBOSE = True
        if verbose > 1:
            DEBUG = True
            LOGGER.debug("Debug output is on!")

    if verbose and force:
        LOGGER.warning("Forced overwrite is on!")
    elif verbose and not force:
        LOGGER.info("Forced overwrite is off!")

    FORCE = force


@cli.command('analyze-sentences', context_settings=CONTEXT_SETTINGS)
@click.option(
    '-s',
    '--showsents',
    is_flag=True,
    help="Show sentences in which there are arguments without local spans."
)
@click.argument('in_path', type=click.Path(exists=True))
def analyze_sentences_cli(showsents, in_path):
    """Analyze sentence boundary information relative to event arguments.

    \f

    Args:
        showsents (bool): whether to show offending sentence details.
        in_path (str): a path to an input BP document.
    """

    import os
    import sys

    from tabulate import tabulate

    from lib import bp

    global FORCE
    global LOGGER
    global VERBOSE

    def __vpath(path):
        return f".../{os.path.basename(path)}" if os.path.basename(path) != path else path

    if DEBUG:
        LOGGER.debug("Invocation Arguments:")
        for line in tabulate([
            ["showsents", type(showsents), showsents,],
            ["in_path", type(in_path), __vpath(in_path),],
        ], headers=["Argument", "Type", "Value"], tablefmt="fancy_grid").split("\n"):
            LOGGER.debug(line)

    if VERBOSE:
        LOGGER.info("Loading BP document...")
    doc = bp.BPDocument.from_json(in_path)

    assert doc.is_valid()

    stats = analyze_sentences(doc, showsents)

    for name, value in stats.items():
        if VERBOSE:
            LOGGER.info(f"{name}: {value}")
        else:
            click.echo()


@cli.command('dedupe-args', context_settings=CONTEXT_SETTINGS)
@click.option('-f', '--flat', is_flag=True, help="Emit a one-line file instead of indented.")
@click.argument('in-path', type=click.Path(exists=True))
@click.argument('out-path', type=click.Path())
def dedupe_args_cli(flat, in_path, out_path):
    """Removes any duplicate values from event argument lists.

    Since documents with duplicate arguments are by definition invalid, all documents input to this
    function are loading with validation off.

    \f

    Args:
        flat (bool): if True, emit a one-liner instead of indented JSON.
        in_path (str): a path to an input BP document.
        out_path (str): a path to output a new BP document.
    """

    import os
    import sys

    from tabulate import tabulate

    from lib import bp

    global FORCE
    global LOGGER
    global VERBOSE

    def __vpath(path):
        return f".../{os.path.basename(path)}" if os.path.basename(path) != path else path

    if DEBUG:
        LOGGER.debug("Invocation Arguments:")
        for line in tabulate([
            ["flat", type(flat), flat,],
            ["in_path", type(in_path), __vpath(in_path),],
            ["out_path", type(out_path), __vpath(out_path),],
        ], headers=["Argument", "Type", "Value"], tablefmt="fancy_grid").split("\n"):
            LOGGER.debug(line)

    try:
        if os.path.dirname(out_path):
            assert os.path.exists(os.path.dirname(out_path))
    except AssertionError:
        LOGGER.error(f"Directory {os.path.dirname(out_path)} doesn't exist, can't write file!")
        sys.exit()

    doc = bp.BPDocument.from_json(in_path, no_validation=True)
    doc.dedupe_args()

    ## emit BP doc

    if VERBOSE:
        LOGGER.info("Serializing...")

    doc.serialize(out_path, flat=flat, verbose=VERBOSE, force=FORCE)

    if VERBOSE:
        LOGGER.info(f"New BP doc saved to {out_path}")


@cli.command('deidentify', context_settings=CONTEXT_SETTINGS)
@click.option('-f', '--flat', is_flag=True, help="Emit a one-line file instead of indented.")
@click.option(
    '-v',
    '--no-validation',
    is_flag=True,
    help="Don't validate the output BP document."
)
@click.argument('in-path', type=click.Path(exists=True))
@click.argument('out-path', type=click.Path())
def deidentify_cli(flat, no_validation, in_path, out_path):
    """Obfuscates Entry IDs and Doc IDs, and removes all Provenance objects.

    Obfuscation is performed using Python's built-in ``uuid`` module.

    \f

    Args:
        flat (bool): if True, emit a one-liner instead of indented JSON.
        no_validation (bool): If True, skip validating constructed BP objects. Defaults to False.
        in_path (str): a path to an input BP document.
        out_path (str): a path to output a new BP document.
    """

    import os
    import sys

    from tabulate import tabulate

    from lib import bp

    global FORCE
    global LOGGER
    global VERBOSE

    def __vpath(path):
        return f".../{os.path.basename(path)}" if os.path.basename(path) != path else path

    if DEBUG:
        LOGGER.debug("Invocation Arguments:")
        for line in tabulate([
            ["flat", type(flat), flat,],
            ["no_validation", type(no_validation), no_validation,],
            ["in_path", type(in_path), __vpath(in_path),],
            ["out_path", type(out_path), __vpath(out_path),],
        ], headers=["Argument", "Type", "Value"], tablefmt="fancy_grid").split("\n"):
            LOGGER.debug(line)

    try:
        if os.path.dirname(out_path):
            assert os.path.exists(os.path.dirname(out_path))
    except AssertionError:
        LOGGER.error(f"Directory {os.path.dirname(out_path)} doesn't exist, can't write file!")
        sys.exit()

    doc = bp.BPDocument.from_json(in_path, no_validation=no_validation)
    stats = doc.deidentify()

    ## emit BP doc

    if VERBOSE:
        LOGGER.info("Serializing...")

    doc.serialize(out_path, flat=flat, verbose=VERBOSE, force=FORCE)

    if VERBOSE:
        LOGGER.info(f"New BP doc saved to {out_path}")


@cli.command('dump-sents', context_settings=CONTEXT_SETTINGS)
@click.argument('in-path', type=click.Path(exists=True))
@click.argument('out-path', type=click.Path())
def dump_sentences_cli(in_path, out_path):
    """Generates a simple JSON representation of all the sentences in a BP document.

    \f

    Args:
        in_path (str): a path to an input BP document.
        out_path (str): a path to output the JSON representation of the sentences.
    """

    import os

    from tabulate import tabulate

    global FORCE
    global LOGGER
    global VERBOSE

    def __vpath(path):
        return f".../{os.path.basename(path)}" if os.path.basename(path) != path else path

    if DEBUG:
        LOGGER.debug("Invocation Arguments:")
        for line in tabulate([
            ["in_path", type(in_path), __vpath(in_path),],
            ["out_path", type(out_path), __vpath(out_path),],
        ], headers=["Argument", "Type", "Value"], tablefmt="fancy_grid").split("\n"):
            LOGGER.debug(line)

    dump_sentences(in_path, out_path)


@cli.command('join-bp-docs', context_settings=CONTEXT_SETTINGS)
@click.option('-f', '--flat', is_flag=True, help="Emit a one-line file instead of indented.")
@click.option('-r', '--no-recurse', is_flag=True, help="Don't look for files in subdirectories.")
@click.option(
    '-s',
    '--show-summary',
    count=True,
    help="Print summary statistics. Pass again for more statistics."
)
@click.option(
    '-u',
    '--allow-duplicates',
    is_flag=True,
    help="Don't dedupe the input files by content."
)
@click.option(
    '-v',
    '--no-validation',
    is_flag=True,
    help="Don't validate the output BP document."
)
@click.argument('corpus-id', type=str)
@click.argument('in-path', type=click.Path(exists=True), nargs=-1)
@click.argument('out-path', type=click.Path())
def join_cli(
    flat,
    no_recurse,
    show_summary,
    allow_duplicates,
    no_validation,
    corpus_id,
    in_path,
    out_path
):
    """Joins any number of BP documents into a single document.

    The documents are joined such that each Entry remains intact. If EntryIDs collide, a new unique
    ID will be created for the offending duplicate. Other information, like top-level provenance
    data, will be preserved. A new ``CORPUS_ID`` is required.

    \f

    Args:
        flat (bool): if True, emit a one-liner instead of indented JSON.
        no_recurse (bool): if True, don't look for files in subdirectories passed in ``mat_path``.
            Flag passed to ``lib.utils.resolve_paths()``, not used anywhere else.
        show_summary (int): if 0, print nothing. If 1, print top-level counts. If 2, print entry
            counts as well.
        allow_duplicates (bool): if True, don't dedupe the files.
        no_validation (bool): If True, skip validating constructed BP objects. Defaults to False.
        corpus_id (str): the new ID of the resulting BP document corpus.
        in_path (tuple): a list of paths to input BP documents or directories of BP documents to
            load.
        out_path (str): a path to output a new BP document.
    """
    import os
    import sys

    from tabulate import tabulate

    from lib.utils import resolve_paths

    global FORCE
    global LOGGER
    global VERBOSE

    def __vpath(p):
        str_ = ""
        if type(p) == tuple:
            p = list(p)
            if os.path.isfile(p[0]):
                str_ += f"...{os.path.basename(p[0])}" if os.path.basename(p[0]) != p[0] else p[0]
            else:
                if p[0].endswith("/") or p[0].endswith("\\"):
                    p[0] = p[0][:-1]
                str_ += f"...{os.path.normpath(p[0])}" if os.path.normpath(p[0]) != p[0] else p[0]
            if len(p) > 1:
                str_ += "\n(others)"
            return str_
        else:
            return f"...{os.path.basename(p)}" if os.path.basename(p) != p else p

    if DEBUG:
        LOGGER.debug("Invocation Arguments:")
        for line in tabulate([
            ["flat", type(flat), flat,],
            ["no_recurse", type(no_recurse), no_recurse,],
            ["show_summary", type(show_summary), show_summary,],
            ["allow_duplicates", type(allow_duplicates), allow_duplicates,],
            ["no_validation", type(no_validation), no_validation,],
            ["corpus_id", type(corpus_id), corpus_id,],
            ["in_path", type(in_path), __vpath(in_path),],
            ["out_path", type(out_path), __vpath(out_path),],
        ], headers=["Argument", "Type", "Value"], tablefmt="fancy_grid").split("\n"):
            LOGGER.debug(line)

    try:
        if os.path.dirname(out_path):
            assert os.path.exists(os.path.dirname(out_path))
    except AssertionError:
        LOGGER.error(f"Directory {os.path.dirname(out_path)} doesn't exist, can't write file!")
        sys.exit()

    paths = resolve_paths(
        in_path,
        ".bp.json",
        no_recurse=no_recurse,
        allow_duplicates=allow_duplicates,
        verbose=VERBOSE,
        debug=not DEBUG
    )

    if not paths:
        LOGGER.error("No BP documents found in input paths!")
        sys.exit(1)

    bp_doc = join(paths, corpus_id, no_validation=no_validation)

    ## emit BP doc

    if show_summary:
        if VERBOSE:
            LOGGER.info("Summary Results:")
            for line in bp_doc.stats(
                pretty=True,
                abstract_events=show_summary>1,
                basic_events=show_summary>1,
                granular_templates=show_summary>1
            ).split("\n"):
                LOGGER.info(line)
        else:
            click.echo(bp_doc.stats(
                abstract_events=show_summary>1,
                basic_events=show_summary>1,
                granular_templates=show_summary>1
            ))

    if VERBOSE:
        LOGGER.info("Serializing...")

    bp_doc.serialize(out_path, flat=flat, verbose=VERBOSE, force=FORCE)

    if VERBOSE:
        LOGGER.info(f"New BP doc saved to {out_path}")


@cli.command('make-deliverables', context_settings=CONTEXT_SETTINGS)
@click.option(
    '-d',
    '--depth',
    default=DEPTH_DEFAULT,
    help=(
        "The size (int) of the random subspace to search when pseudo-optimizing partition "
        f"assignments. Defaults to {DEPTH_DEFAULT}"
    )
)
@click.option(
    '-e',
    '--no-ensure-indoc-coverage',
    is_flag=True,
    help="Don't ensure that one of every Event type can be found in the input document."
)
@click.option('-f', '--flat', is_flag=True, help="Emit a one-line file instead of indented.")
@click.option(
    '-s',
    '--show-summary',
    count=True,
    help="Print summary statistics. Pass again for more statistics."
)
@click.option(
    '-v',
    '--no-validation',
    is_flag=True,
    help="Don't validate the output BP document."
)
@click.option(
    '--dist-context',
    type=click.Path(exists=True),
    help="Make new deliverables with the context of a previous distribution."
)
@click.option(
    '--dist-provided',
    default=DIST_PROVIDED_DEFAULT,
    help=(
        "Set the 'x.x%' percentage of the input the provided partition should be. "
        f"Defaults to '{DIST_PROVIDED_DEFAULT}'."
    )
)
@click.option(
    '--dist-provided-analysis',
    default=DIST_PROVIDED_ANALYSIS_DEFAULT,
    help=(
        "Set the 'x.x%' percentage of the provided partition the provided-analysis partition should"
        f" be. Defaults to '{DIST_PROVIDED_ANALYSIS_DEFAULT}'."
    )
)
@click.option(
    '--dist-provided-devtest',
    default=DIST_PROVIDED_DEVTEST_DEFAULT,
    help=(
        "Set the 'x.x%' percentage of the provided partition the provided-devtest partition should "
        f" be. Defaults to '{DIST_PROVIDED_DEVTEST_DEFAULT}'."
    )
)
@click.argument('dialect', type=click.Choice(MAT_DIALECT_CHOICES))
@click.argument('in-path', type=click.Path(exists=True))
@click.argument('out-dir', type=click.Path(exists=True))
def make_deliverables_cli(
    no_ensure_indoc_coverage,
    flat,
    show_summary,
    no_validation,
    dist_context,
    dist_provided,
    dist_provided_analysis,
    dist_provided_devtest,
    depth,
    dialect,
    in_path,
    out_dir
):
    """Splits a BP document into the four deliverable partitions.

    The document is split such that each Entry remains intact.

    You must pass either ``dist_context`` or ``dist_provided``, ``dist_provided_analysis``, and
    ``dist_provided_devtest``.

    Note that passing a ``dist_context`` will override values of ``dist_provided``,
    ``dist_provided_analysis``, and ``dist_provided_devtest``, regardless of whether they were
    passed in.

    \f

    Args:
        depth (int): the size of the random subspace to pseudooptimize partition assignments in.
        flat (bool): if True, emit one-liners instead of indented JSON.
        no_ensure_indoc_coverage (bool): if True, skip ensuring that one of every event type for the
            chosen Dialect can be found in the input document.
        show_summary (int): if 0, print nothing. If 1, print top-level counts. If 2, print entry
            counts as well.
        no_validation (bool): if True, skip validating BP objects. Defaults to False.
        dist_context (str): path to a Distribution file, which is a JSON blob of entry assignment
            information about a previous distribution.
        dist_provided (str): the percentage of the input the provided partition should be, expressed
            as 'x.x%'.
        dist_provided_analysis (str): the percentage of the provided partition the provided-analysis
            partition should be, expressed as 'x.x%'.
        dist_provided_devtest (str): the percentage of the provided partition the provided-devtest
            partition should be, expressed as 'x.x%'.
        dialect (str): the MAT task the BP document was created with. Must be one of
            ``utils.MAT_DIALECT_CHOICES``, which is verified by the Click decorator.
        in_path (str): a path to an input BP document.
        out_dir (str): a path to a directory to output the new documents.
    """
    from copy import deepcopy
    import os
    import sys

    from tabulate import tabulate

    from lib import bp
    from lib.utils import open_json_blob
    from lib.utils import write_json_blob

    global FORCE
    global LOGGER
    global VERBOSE

    if DEBUG:
        LOGGER.debug("Invocation Arguments:")
        def __vpath(path):
            if os.path.isdir(path):
                return f".../{os.path.basename((os.path.normpath(path)))}" if len(os.path.normpath(path)) > 1 else path
            else:
                return f".../{os.path.basename(path)}" if os.path.basename(path) != path else path
        for line in tabulate([
            ["depth", type(depth), depth,],
            ["no_ensure_indoc_coverage", type(no_ensure_indoc_coverage), no_ensure_indoc_coverage,],
            ["flat", type(flat), flat,],
            ["show_summary", type(show_summary), show_summary,],
            ["no_validation", type(no_validation), no_validation,],
            ["dist_provided", type(dist_provided), dist_provided,],
            ["dist_provided_analysis", type(dist_provided_analysis), dist_provided_analysis,],
            ["dist_provided_devtest", type(dist_provided_devtest), dist_provided_devtest,],
            ["dialect", type(dialect), dialect,],
            ["in_path", type(in_path), __vpath(in_path),],
            ["out_dir", type(out_dir), __vpath(out_dir),],
        ], headers=["Argument", "Type", "Value"], tablefmt="fancy_grid").split("\n"):
            LOGGER.debug(line)

    assert os.path.isdir(out_dir), f"Output path {out_dir} is not a directory!"

    assert dist_context or (dist_provided and dist_provided_devtest and dist_provided_analysis), \
        "Missing required values."

    ## ingest and split

    if VERBOSE:
        LOGGER.info("Loading input document...")
    in_doc = bp.BPDocument.from_json(in_path, no_validation)

    if not no_validation:
        if not in_doc.is_valid():
            sys.exit()

    if dist_context:
        dist_context = open_json_blob(dist_context)

    hidden, provided_train, provided_analysis, provided_devtest, dist = make_deliverables(
        in_doc,
        dialect,
        depth=depth,
        dist_context=dist_context,
        dist_provided=dist_provided,
        dist_provided_analysis=dist_provided_analysis,
        dist_provided_devtest=dist_provided_devtest,
        no_ensure_indoc_coverage=no_ensure_indoc_coverage,
        no_validation=no_validation
    )

    ## emit

    if show_summary:
        if VERBOSE:
            LOGGER.info("Summary Results:")

            ## easy stats

            hst = hidden.stats(
                pretty=False,
                abstract_events=show_summary>1 and dialect=="Abstract",
                basic_events=show_summary>1 and dialect in ["Basic", "Granular",],
                granular_templates=show_summary>1 and dialect=="Granular"
            )
            trst = provided_train.stats(
                pretty=False,
                abstract_events=show_summary>1 and dialect=="Abstract",
                basic_events=show_summary>1 and dialect in ["Basic", "Granular",],
                granular_templates=show_summary>1 and dialect=="Granular"
            )
            ast = provided_analysis.stats(
                pretty=False,
                abstract_events=show_summary>1 and dialect=="Abstract",
                basic_events=show_summary>1 and dialect in ["Basic", "Granular",],
                granular_templates=show_summary>1 and dialect=="Granular"
            )
            dst = provided_devtest.stats(
                pretty=False,
                abstract_events=show_summary>1 and dialect=="Abstract",
                basic_events=show_summary>1 and dialect in ["Basic", "Granular",],
                granular_templates=show_summary>1 and dialect=="Granular"
            )
            stats = [[f"# {k}", hst[k], trst[k], ast[k], dst[k],] for k in hst]

            for line in tabulate(
                stats,
                headers=[
                    "Stat", "Hidden", "Provided-Training", "Provided-Analysis", "Provided-Devtest",
                ],
                tablefmt="fancy_grid"
            ).split("\n"):
                LOGGER.info(line)
        else:
            click.echo(hidden.stats(
                abstract_events=show_summary>1 and dialect=="Abstract",
                basic_events=show_summary>1 and dialect in ["Basic", "Granular",],
                granular_templates=show_summary>1 and dialect=="Granular"
            ))
            click.echo(provided_train.stats(
                abstract_events=show_summary>1 and dialect=="Abstract",
                basic_events=show_summary>1 and dialect in ["Basic", "Granular",],
                granular_templates=show_summary>1 and dialect=="Granular"
            ))
            click.echo(provided_analysis.stats(
                abstract_events=show_summary>1 and dialect=="Abstract",
                basic_events=show_summary>1 and dialect in ["Basic", "Granular",],
                granular_templates=show_summary>1 and dialect=="Granular"
            ))
            click.echo(provided_devtest.stats(
                pretty=False,
                abstract_events=show_summary>1 and dialect=="Abstract",
                basic_events=show_summary>1 and dialect in ["Basic", "Granular",],
                granular_templates=show_summary>1 and dialect=="Granular"
            ))

    if VERBOSE:
        LOGGER.info("Serializing...")

    prefix = os.path.basename(in_path)
    if prefix.endswith('.json'):
        prefix = prefix[:-5]
    if prefix.endswith('.bp'):
        prefix = prefix[:-3]

    if dist:

        dist_provided = dist['pdist-provided'] * 100
        dist_provided_analysis = dist['pdist-provided-analysis'] * 100
        dist_provided_devtest = dist['pdist-provided-devtest'] * 100

        hpct = 100.0 - dist_provided
        ppct = 100.0 - hpct
        tpct = 100.0 - (dist_provided_analysis + dist_provided_devtest)
        apct = dist_provided_analysis
        dpct = dist_provided_devtest

    else:

        hpct = 100.0 - float(dist_provided[:-1])
        ppct = 100.0 - hpct
        tpct = 100.0 - (float(dist_provided_analysis[:-1]) + float(dist_provided_devtest[:-1]))
        apct = float(dist_provided_analysis[:-1])
        dpct = float(dist_provided_devtest[:-1])

    hname = f"{os.path.join(out_dir, prefix)}-hidden-{hpct}pct.bp.json"
    tname = f"{os.path.join(out_dir, prefix)}-provided-{ppct}pct.train-{tpct}pct.bp.json"
    aname = f"{os.path.join(out_dir, prefix)}-provided-{ppct}pct.analysis-{apct}pct.ref.bp.json"
    dname = f"{os.path.join(out_dir, prefix)}-provided-{ppct}pct.devtest-{dpct}pct.ref.bp.json"

    aname_empty = \
        f"{os.path.join(out_dir, prefix)}-provided-{ppct}pct.analysis-{apct}pct.empty.bp.json"
    dname_empty = \
        f"{os.path.join(out_dir, prefix)}-provided-{ppct}pct.devtest-{dpct}pct.empty.bp.json"

    tname_d = f"{os.path.join(out_dir, prefix)}-provided-{ppct}pct.train-{tpct}pct.d.bp.json"
    aname_d = f"{os.path.join(out_dir, prefix)}-provided-{ppct}pct.analysis-{apct}pct.ref.d.bp.json"
    aname_empty_d = \
        f"{os.path.join(out_dir, prefix)}-provided-{ppct}pct.analysis-{apct}pct.empty.d.bp.json"
    dname_d = f"{os.path.join(out_dir, prefix)}-provided-{ppct}pct.devtest-{dpct}pct.ref.d.bp.json"
    dname_empty_d = \
        f"{os.path.join(out_dir, prefix)}-provided-{ppct}pct.devtest-{dpct}pct.empty.d.bp.json"

    dist_name = f"{os.path.join(out_dir, prefix)}.dist.json"

    ## "normal" docs

    hidden.serialize(hname, flat=flat, verbose=VERBOSE, force=FORCE)
    if VERBOSE:
        LOGGER.info(f"    Hidden doc saved to {hname}")

    provided_train.serialize(tname, flat=flat, verbose=VERBOSE, force=FORCE)
    if VERBOSE:
        LOGGER.info(f"    Provided-Training doc saved to {tname}")
    provided_train_d = deepcopy(provided_train)

    provided_analysis.serialize(aname, flat=flat, verbose=VERBOSE, force=FORCE)
    if VERBOSE:
        LOGGER.info(f"    Provided-Analysis Reference doc saved to {aname}")
    provided_analysis_d = deepcopy(provided_analysis)

    provided_devtest.serialize(dname, flat=flat, verbose=VERBOSE, force=FORCE)
    if VERBOSE:
        LOGGER.info(f"    Provided-Devtest Reference doc saved to {dname}")
    provided_devtest_d = deepcopy(provided_devtest)

    ## empty versions of provided analysis and devtest

    provided_analysis.remove_annotation_sets()
    provided_analysis.corpus_id.value = provided_analysis.corpus_id.value[:-3] + "Empty"
    provided_analysis.serialize(aname_empty, flat=flat, verbose=VERBOSE, force=FORCE)
    if VERBOSE:
        LOGGER.info(f"    Provided-Analysis No-Annots doc saved to {aname_empty}")
    provided_devtest.remove_annotation_sets()
    provided_devtest.corpus_id.value = provided_devtest.corpus_id.value[:-3] + "Empty"
    provided_devtest.serialize(dname_empty, flat=flat, verbose=VERBOSE, force=FORCE)
    if VERBOSE:
        LOGGER.info(f"    Provided-Devtest No-Annots doc saved to {dname_empty}")

    ## deidentified versions of provided files

    provided_train_d.deidentify()
    provided_analysis_d.deidentify()
    provided_devtest_d.deidentify()

    provided_train_d.serialize(tname_d, flat=flat, verbose=VERBOSE, force=FORCE)
    if VERBOSE:
        LOGGER.info(f"    Provided-Training Deidentified doc saved to {tname_d}")
    provided_analysis_d.serialize(aname_d, flat=flat, verbose=VERBOSE, force=FORCE)
    if VERBOSE:
        LOGGER.info(f"    Provided-Analysis Reference Deidentified doc saved to {aname_d}")
    provided_devtest_d.serialize(dname_d, flat=flat, verbose=VERBOSE, force=FORCE)
    if VERBOSE:
        LOGGER.info(f"    Provided-Devtest Reference Deidentified doc saved to {dname_d}")

    provided_analysis_d.remove_annotation_sets()
    provided_devtest_d.remove_annotation_sets()

    provided_analysis_d.serialize(aname_empty_d, flat=flat, verbose=VERBOSE, force=FORCE)
    if VERBOSE:
        LOGGER.info(f"    Provided-Analysis No-Annots Deidentified doc saved to {aname_empty_d}")
    provided_devtest_d.serialize(dname_empty_d, flat=flat, verbose=VERBOSE, force=FORCE)
    if VERBOSE:
        LOGGER.info(f"    Provided-Devtest No-Annots Deidentified doc saved to {dname_empty_d}")

    ## distribution data

    write_json_blob(dist, dist_name, indent=4, verbose=VERBOSE, force=FORCE)
    if VERBOSE:
        LOGGER.info(f"    Distribution information saved to {dist_name}")


@cli.command('split-bp-doc', context_settings=CONTEXT_SETTINGS)
@click.option(
    '-e',
    '--no-ensure-indoc-coverage',
    is_flag=True,
    help="Don't ensure that one of every Event type can be found in the input document."
)
@click.option('-f', '--flat', is_flag=True, help="Emit a one-line file instead of indented.")
@click.option(
    '-s',
    '--show-summary',
    count=True,
    help="Print summary statistics. Pass again for more statistics."
)
@click.option(
    '-v',
    '--no-validation',
    is_flag=True,
    help="Don't validate the output BP document."
)
@click.argument('dialect', type=click.Choice(MAT_DIALECT_CHOICES))
@click.argument('in-path', type=click.Path(exists=True))
@click.argument('out-path-a', type=click.Path())
@click.argument('out-path-b', type=click.Path())
@click.argument('amount')
def split_bp_cli(
    no_ensure_indoc_coverage,
    flat,
    show_summary,
    no_validation,
    dialect,
    in_path,
    out_path_a,
    out_path_b,
    amount
):
    """Splits a BP document in two randomly.

    The document is split such that each Entry remains intact. Some will go to document A, others
    will go to document B. The selection for which goes where is decided randomly.

    ``AMOUNT`` is the number of entries within the input BP document to reserve for the B partition,
    either represented as an integer or a percentage written as "x.x%". If an integer is passed that
    is higher than the number of entries available, it will be treated as if ``100%`` was passed. If
    a percentage is passed, it will ``floor`` to the nearest whole entry.

    \f

    Args:
        no_ensure_indoc_coverage (bool): if True, skip ensuring that one of every event type for the
            chosen Dialect can be found in the input document.
        flat (bool): if True, emit one-liners instead of indented JSON.
        show_summary (int): if 0, print nothing. If 1, print top-level counts. If 2, print entry
            counts as well.
        no_validation (bool): If True, skip validating BP objects. Defaults to False.
        dialect (str): the MAT task the BP document was created with. Must be one of
            ``utils.MAT_DIALECT_CHOICES``, which is verified by the Click decorator.
        in_path (str): a path to an input BP document.
        out_path_a (str): a path to output the "A" partition BP document.
        out_path_b (str): a path to output the "B" partition BP document.
        amount (str): the number of entries within the input BP document to reserve for the B
            partition, either represented as an integer or a percentage written as "x.x%". If an
            ``int`` is passed that is higher than the number of entries available, it will be
            treated as if ``100%`` was passed. If a percentage is passed, it will ``floor`` to the
            nearest whole entry.
    """
    import os
    import sys

    from tabulate import tabulate

    from lib import bp

    global FORCE
    global LOGGER
    global VERBOSE

    if DEBUG:
        LOGGER.debug("Invocation Arguments:")
        def __vpath(path):
            return f".../{os.path.basename(path)}" if os.path.basename(path) != path else path
        for line in tabulate([
            ["flat", type(flat), flat,],
            ["no_ensure_indoc_coverage", type(no_ensure_indoc_coverage), no_ensure_indoc_coverage,],
            ["show_summary", type(show_summary), show_summary,],
            ["no_validation", type(no_validation), no_validation,],
            ["dialect", type(dialect), dialect,],
            ["in_path", type(in_path), __vpath(in_path),],
            ["out_path_a", type(out_path_a), __vpath(out_path_a),],
            ["out_path_b", type(out_path_b), __vpath(out_path_b),],
            ["amount", type(amount), amount,],
        ], headers=["Argument", "Type", "Value"], tablefmt="fancy_grid").split("\n"):
            LOGGER.debug(line)

    try:
        assert os.path.exists(os.path.dirname(out_path_a))
    except AssertionError:
        LOGGER.error(f"Directory {os.path.dirname(out_path_a)} doesn't exist, can't write file!")
        sys.exit()
    try:
        assert os.path.exists(os.path.dirname(out_path_b))
    except AssertionError:
        LOGGER.error(f"Directory {os.path.dirname(out_path_b)} doesn't exist, can't write file!")
        sys.exit()

    ## ingest and split

    if VERBOSE:
        LOGGER.info("Loading input document...")
    in_doc = bp.BPDocument.from_json(in_path, no_validation)

    if not no_validation:
        if not in_doc.is_valid():
            sys.exit()

    doc_a, doc_b = split(
        in_doc,
        amount,
        dialect,
        no_ensure_indoc_coverage=no_ensure_indoc_coverage,
        no_validation=no_validation
    )

    ## emit

    if show_summary:
        if VERBOSE:
            LOGGER.info("Summary Results:")

            ## easy stats

            ast = doc_a.stats(
                pretty=False,
                abstract_events=show_summary>1 and dialect=="Abstract",
                basic_events=show_summary>1 and dialect in ["Basic", "Granular",],
                granular_templates=show_summary>1 and dialect=="Granular"
            )
            bst = doc_b.stats(
                pretty=False,
                abstract_events=show_summary>1 and dialect=="Abstract",
                basic_events=show_summary>1 and dialect in ["Basic", "Granular",],
                granular_templates=show_summary>1 and dialect=="Granular"
            )
            stats = [[f"# {k}", ast[k], bst[k],] for k in ast]

            for line in tabulate(
                stats,
                headers=["Stat", "Doc A", "Doc B"],
                tablefmt="fancy_grid"
            ).split("\n"):
                LOGGER.info(line)
        else:
            click.echo(doc_a.stats(
                abstract_events=show_summary>1 and dialect=="Abstract",
                basic_events=show_summary>1 and dialect in ["Basic", "Granular",],
                granular_templates=show_summary>1 and dialect=="Granular"
            ))
            click.echo(doc_b.stats(
                abstract_events=show_summary>1 and dialect=="Abstract",
                basic_events=show_summary>1 and dialect in ["Basic", "Granular",],
                granular_templates=show_summary>1 and dialect=="Granular"
            ))

    if VERBOSE:
        LOGGER.info("Serializing...")

    doc_a.serialize(out_path_a, flat=flat, verbose=VERBOSE, force=FORCE)
    if VERBOSE:
        LOGGER.info(f"    New BP doc saved to {out_path_a}")
    doc_b.serialize(out_path_b, flat=flat, verbose=VERBOSE, force=FORCE)
    if VERBOSE:
        LOGGER.info(f"    New BP doc saved to {out_path_b}")


@cli.command('project-sentences', context_settings=CONTEXT_SETTINGS)
@click.argument('in_path', type=click.Path(exists=True))
@click.argument('out_path', type=click.Path())
def project_sentences_cli(in_path, out_path):
    """Projects BP sentences to MAT document format."""

    import os

    import MAT

    from lib import bp

    mat_io = MAT.DocumentIO.getDocumentIO("mat-json")

    bp_doc = bp.BPDocument.from_json(in_path)

    if bp_doc.entries.value:
        for entry_id, entry in bp_doc.entries.value.items():

            mat_doc = MAT.Document.AnnotatedDoc(signal=entry.segment_text.value)

            if entry.segment_sections.value:
                for segsec in entry.segment_sections.value:
                    if segsec.structural_element.value == "Sentence":
                        mat_doc.createAnnotation(segsec.start.value, segsec.end.value, "S")

            name = entry_id
            if not name.endswith(".json"):
                name += ".json"
            mat_io.writeToTarget(mat_doc, os.path.join(out_path, name))


if __name__ == '__main__':

    ## program proper

    cli()
