####################################################################################################
### -----------------------------------------  NOTICE  ----------------------------------------- ###
###                                                                                              ###
### This software (or technical data) was produced for the U.S. Government under contract        ###
### 2015-14120200002-002, and is subject to the Rights in Data-General Clause 52.227-14 , ALT    ###
### IV (MAY 2014) or (DEC 2007).                                                                 ###
###                                                                                              ###
### © 2019 The MITRE Corporation. All Rights Reserved.                                           ###
###                                                                                              ###
### -------------------------------------------------------------------------------------------- ###
####################################################################################################

"""
Utility library and settings/constants for the bp_lib repository.
"""

__author__  = "Leland E. Vakarian <lsperl@mitre.org>"
__date__    = "22 June, 2022"

import logging
import os
from string import punctuation

import click


## Logging


class ColoredLogFormatter(logging.Formatter):


    def __init__(self, msg):
        logging.Formatter.__init__(self, msg)


    def format(self, record):
        import click

        if record.levelname == "DEBUG":
            record.levelname = f"[{click.style(record.levelname, fg='bright_magenta')}]   "
        elif record.levelname == "INFO":
            record.levelname = f"[{click.style(record.levelname, fg='bright_green')}]    "
        elif record.levelname == "WARNING":
            record.levelname = f"[{click.style(record.levelname, fg='bright_yellow')}] "
        elif record.levelname == "ERROR":
            record.levelname = f"[{click.style(record.levelname, fg='bright_red')}]   "
        elif record.levelname == "CRITICAL":
            record.levelname = f"[{click.style(record.levelname, fg='red')}]"

        return logging.Formatter.format(self, record)


class ColorlessLogFormatter(logging.Formatter):


    def __init__(self, msg):
        logging.Formatter.__init__(self, msg)


    def format(self, record):

        if record.levelname == "DEBUG":
            record.levelname = f"[{record.levelname}]   "
        elif record.levelname == "INFO":
            record.levelname = f"[{record.levelname}]    "
        elif record.levelname == "WARNING":
            record.levelname = f"[{record.levelname}] "
        elif record.levelname == "ERROR":
            record.levelname = f"[{record.levelname}]   "
        elif record.levelname == "CRITICAL":
            record.levelname = f"[{record.levelname}]"

        return logging.Formatter.format(self, record)


class FormattedLogger(logging.Logger):


    def __init__(self, name):

        logging.Logger.__init__(self, name, logging.DEBUG)

        console = logging.StreamHandler()
        console.setFormatter(ColoredLogFormatter("%(levelname)s %(message)s"))
        self.addHandler(console)

    def add_fh(self, path):
        fh = logging.FileHandler(path)
        fh.setFormatter(ColorlessLogFormatter("%(levelname)s %(message)s"))
        self.addHandler(fh)


logging.setLoggerClass(FormattedLogger)


## Constants


LOGGER = logging.getLogger("bp_lib")

"""dict: Overrides the default "help" options for a ``click`` command.

Rather than only including ``--help`` for the "help" option, includes ``-h`` as well.
"""
CONTEXT_SETTINGS = dict(help_option_names=['-h', '--help'], max_content_width=100)

"""str: the path to the root of this repository."""
BASEPATH = os.path.dirname(os.path.dirname(os.path.realpath(__file__)))

"""Prefix to include in tqdm descriptions."""
TQDM_PREFIX = "[" + click.style("INFO", fg='bright_green') + "]    "

"""list: the only valid options for a MAT dialect (task)."""
MAT_DIALECT_CHOICES = ['Abstract', 'Basic', 'Granular',]
MAT_HEAD_TYPES = ["Head",]
MAT_NO_HEAD_TYPES = ["No-Head",]
MAT_SPAN_SYNCLASS_TYPES = [
    "name",
    "nominal",
    "pronoun",
    "event-anchor",
    "template-anchor",
    "time-mention",
    "duration-mention",
]

"""Possible choices for annotator classes."""
BP_ANNOTATOR_CLASS_TYPES = ["crowd", "mitre", "system", "pretagger", "arlis", "nist"]

"""Each of these lists are MAT annotation types for the Abstract dialect."""
ABSTRACT_MAT_ANCHOR_TYPES = ["AbstractEvent", "Additional_Anchor",]
ABSTRACT_MAT_ADDITIONAL_ANCHOR_TYPES = ["Additional_Anchor",]
ABSTRACT_MAT_KEY_MENTION_TYPES = ["Agent-or-Patient", "Nominal", "Name", "Pronoun",]
ABSTRACT_MAT_ADDITIONAL_MENTION_TYPES = [
    "Nominal_Mention",
    "Name_Mention",
    "Pronoun_Mention",
    "Additional_Mention",
]

"""Each of these lists are MAT annotation types for the Basic dialect."""
BASIC_MAT_KEY_MENTION_TYPES = ["Nominal-Arg-Mention", "Name-Arg-Mention", "Pronoun-Arg-Mention",]
BASIC_MAT_ADDITIONAL_MENTION_TYPES = [
    "Addtl-Nominal-Arg-Mention",
    "Addtl-Name-Arg-Mention",
    "Addtl-Pronoun-Arg-Mention",
]
BASIC_MAT_NOMINAL_MENTION_TYPES = ["Nominal-Arg-Mention", "Addtl-Nominal-Arg-Mention",]
BASIC_MAT_NAME_MENTION_TYPES = ["Name-Arg-Mention", "Addtl-Name-Arg-Mention",]
BASIC_MAT_PRONOUN_MENTION_TYPES = ["Pronoun-Arg-Mention", "Addtl-Pronoun-Arg-Mention",]
BASIC_MAT_SENTENCE_TYPES = ["Sentence", "S",]
BASIC_MAT_STRUCTURAL_TYPES = BASIC_MAT_SENTENCE_TYPES + [
    "Headline", "Byline", "Dateline", "Story-Lead", "Section_Header", "Delete-This-Text", "Ignore"
]
BASIC_MAT_FAUX_ANCHOR_TYPES = [
    # Faux Communicate-Category events generated with "replace-with" properties
    "Communicate-Report",
    "Communicate-Express-Optimism",
    "Communicate-Express-Pessimism",
    "Communicate-Apologize",
    "Communicate-Request",
    "Communicate-Demand",
    "Communicate-Accuse",
    "Communicate-Praise",
    "Communicate-Deny",
    "Communicate-Confirm",
    "Communicate-Wuzza",
]
BASIC_MAT_KEY_ANCHOR_TYPES = BASIC_MAT_FAUX_ANCHOR_TYPES + [
    "Aid-Needs",
    "Apply-NPI",
    "Award-Contract",
    "Bribery",
    "Business-Event-or-SoA",
    "Change-of-Govt",
    "Change-Repayment",
    "Close-Schools",
    "Communicate-Event",
    "Conduct-Diplomatic-Talks",
    "Conduct-Medical-Research",
    "Conduct-Meeting",
    "Conduct-Protest",
    "Conduct-Violent-Protest",
    "Conspiracy",
    "Construct-Project",
    "Coordinated-Comm",
    "Corruption",
    "Coup",
    "Cull-Livestock",
    "Cyber-Crime-Attack",
    "Cyber-Crime-Other",
    "Cybersecurity-Measure", # new name for "Cybersecurity-Response"
    "Cybersecurity-Response", # original name for "Cybersecurity-Measure".
    "Death-from-Crisis-Event",
    "Declare-Emergency",
    "Disease-Exposes",
    "Disease-Infects",
    "Disease-Kills",
    "Disease-Outbreak",
    "Disease-Recovery",
    "Dismiss-Workers",
    "Distribute-PPE",
    "Economic-Event-or-SoA",
    "Employ-Workers",
    "Environmental-Event-or-SoA",
    "Establish-Project",
    "Evacuate",
    "Expel",
    "Extortion",
    "Famine-Event-or-SoA",
    "Financial-Crime",
    "Financial-Loss",
    "Fiscal-or-Monetary-Action",
    "Fund-Project",
    "Hospitalize",
    "Identify-Vulnerability",
    "Illegal-Entry",
    "Impose-Quarantine",
    "Information-Release",
    "Information-Theft",
    "Infrastructure-Operation",
    "Interrupt-Construction",
    "Interrupt-Operations",
    "Judicial-Acquit",
    "Judicial-Convict",
    "Judicial-Indict",
    "Judicial-Other",
    "Judicial-Plead",
    "Judicial-Prosecute",
    "Judicial-Seize",
    "Judicial-Sentence",
    "Kidnapping",
    "Law-Enforcement-Arrest",
    "Law-Enforcement-Extradite",
    "Law-Enforcement-Investigate",
    "Law-Enforcement-Other",
    "Leave-Job",
    "Legislative-Action",
    "Lift-PPE-Requirements",
    "Lift-Quarantine",
    "Loosen-Business-Restrictions",
    "Loosen-Travel-Restrictions",
    "Make-Repayment",
    "Migrant-Detain",
    "Migrant-Relocation",
    "Migrant-Smuggling",
    "Migration-Blocked",
    "Migration-Impeded-Failed",
    "Military-Attack",
    "Military-Declare-War",
    "Military-Other",
    "Missing-from-Crisis-Event",
    "Monitor-Disease",
    "Natural-Disaster-Event-or-SoA",
    "Natural-Phenomenon-Event-or-SoA",
    "Open-Schools",
    "Organize-Protest",
    "Other-Crime",
    "Other-Government-Action",
    "Pay-Ransom",
    "Persecution",
    "Political-Election-Event",
    "Political-Event-or-SoA",
    "Political-Other",
    "Propose-Project",
    "Provide-Aid",
    "Refugee-Movement",
    "Repair",
    "Require-PPE",
    "Rescue",
    "Restrict-Business",
    "Restrict-Travel",
    "Sign-Agreement",
    "Suppress-Communication",
    "Suppression-of-Free-Speech",
    "Suppress-Meeting",
    "Suppress-or-Breakup-Protest",
    "Test-Patient",
    "Treat-Patient",
    "Vaccinate",
    "Violence",
    "Violence-Attack",
    "Violence-Bombing",
    "Violence-Damage",
    "Violence-Kill",
    "Violence-Other",
    "Violence-Set-Fire",
    "Violence-Wound",
    "War-Event-or-SoA",
    "Weather-Event-or-SoA",
    "Weather-or-Environmental-Damage",
    "Wounding-from-Crisis-Event",
]
BASIC_MAT_ADDITIONAL_ANCHOR_TYPES = ["Additional-Event-Anchor",]
# Filter out ignored event types
# https://3.basecamp.com/3910605/buckets/14032182/messages/3970853259#__recording_3997535741
BASIC_MAT_IGNORED_ANCHOR_TYPES = BASIC_MAT_FAUX_ANCHOR_TYPES + [
    'Conspiracy',
    "Distribute-PPE",
    'Famine-Event-or-SoA',
    'Law-Enforcement-Extradite',
    'Lift-PPE-Requirements',
    "Migration-Impeded-Failed",
    'Natural-Disaster-Event-or-SoA',
    'Persecution',
    "Propose-Project",
    'Suppress-Meeting',
    'Suppress-or-Breakup-Protest',
    'Weather-Event-or-SoA',
]
BASIC_SLICE_EXPERIMENT_IGNORED_ANCHOR_TYPES = [
    'Communicate-Report', 'Communicate-Express-Optimism', 'Communicate-Express-Pessimism',
    'Communicate-Apologize', 'Communicate-Request', 'Communicate-Demand', 'Communicate-Accuse',
    'Communicate-Praise', 'Communicate-Deny', 'Communicate-Confirm', 'Communicate-Wuzza',
    'Aid-Needs', 'Apply-NPI', 'Bribery', 'Business-Event-or-SoA', 'Change-of-Govt', 'Close-Schools',
    'Communicate-Event', 'Conduct-Diplomatic-Talks', 'Conduct-Medical-Research', 'Conduct-Meeting',
    'Conduct-Protest', 'Conduct-Violent-Protest', 'Conspiracy', 'Coordinated-Comm', 'Corruption',
    'Coup', 'Cull-Livestock', 'Cybersecurity-Response', 'Death-from-Crisis-Event',
    'Declare-Emergency', 'Disease-Exposes', 'Disease-Infects', 'Disease-Kills', 'Disease-Outbreak',
    'Disease-Recovery', 'Distribute-PPE', 'Economic-Event-or-SoA', 'Environmental-Event-or-SoA',
    'Evacuate', 'Expel', 'Extortion', 'Famine-Event-or-SoA', 'Financial-Crime',
    'Fiscal-or-Monetary-Action', 'Hospitalize', 'Illegal-Entry', 'Impose-Quarantine',
    'Judicial-Acquit', 'Judicial-Convict', 'Judicial-Indict', 'Judicial-Other', 'Judicial-Plead',
    'Judicial-Prosecute', 'Judicial-Seize', 'Judicial-Sentence', 'Kidnapping',
    'Law-Enforcement-Arrest', 'Law-Enforcement-Extradite', 'Law-Enforcement-Investigate',
    'Law-Enforcement-Other', 'Leave-Job', 'Lift-PPE-Requirements', 'Lift-Quarantine',
    'Loosen-Business-Restrictions', 'Loosen-Travel-Restrictions', 'Migrant-Detain',
    'Migrant-Relocation', 'Migrant-Smuggling', 'Migration-Blocked', 'Migration-Impeded-Failed',
    'Military-Attack', 'Military-Declare-War', 'Military-Other', 'Missing-from-Crisis-Event',
    'Monitor-Disease', 'Natural-Disaster-Event-or-SoA', 'Natural-Phenomenon-Event-or-SoA',
    'Open-Schools', 'Organize-Protest', 'Other-Crime', 'Other-Government-Action', 'Persecution',
    'Political-Election-Event', 'Political-Event-or-SoA', 'Political-Other', 'Propose-Project',
    'Provide-Aid', 'Refugee-Movement', 'Repair', 'Require-PPE', 'Rescue', 'Restrict-Business',
    'Restrict-Travel', 'Suppress-Communication', 'Suppression-of-Free-Speech', 'Suppress-Meeting',
    'Suppress-or-Breakup-Protest', 'Test-Patient', 'Treat-Patient', 'Vaccinate', 'Violence',
    'Violence-Attack', 'Violence-Bombing', 'Violence-Damage', 'Violence-Kill', 'Violence-Other',
    'Violence-Set-Fire', 'Violence-Wound', 'War-Event-or-SoA', 'Weather-Event-or-SoA',
    'Weather-or-Environmental-Damage', 'Wounding-from-Crisis-Event',
]
# uncomment for the Phase 3 slice experiement on phase-3-only events
# BASIC_MAT_IGNORED_ANCHOR_TYPES = BASIC_MAT_FAUX_ANCHOR_TYPES + BASIC_SLICE_EXPERIMENT_IGNORED_ANCHOR_TYPES

# uncomment for the Phase 3 slice experiement on phase-1-and-2-only events
# BASIC_MAT_IGNORED_ANCHOR_TYPES = BASIC_MAT_FAUX_ANCHOR_TYPES + \
#     [x for x in BASIC_MAT_KEY_ANCHOR_TYPES if x not in BASIC_SLICE_EXPERIMENT_IGNORED_ANCHOR_TYPES]


"""Each of these lists/dicts are MAT annotation types for the Granular dialect."""
GRANULAR_MAT_TEMPLATE_TYPES = {
    "Protestplate": { # template-type
        # MAT annotation property name : BPDocumentGranularTemplate member name
        "arrested": "arrested",
        "imprisoned": "imprisoned",
        "killed": "killed",
        "occupy": "occupy",
        "over-time": "over_time",
        "organizer": "organizer",
        "outcome-averted": "outcome_averted",
        "outcome-occurred": "outcome_occurred",
        "outcome-hypothetical": "outcome_hypothetical",
        "protest-against": "protest_against",
        "protest-event": "protest_event",
        "protest-for": "protest_for",
        "when": "when",
        "where": "where",
        "who": "who",
        "wounded":"wounded",
    },
    "Corruplate": {
        "charged-with": "charged_with",
        "corrupt-event": "corrupt_event",
        "judicial-actions": "judicial_actions",
        "fine": "fine",
        "over-time": "over_time",
        "outcome-averted": "outcome_averted",
        "outcome-occurred": "outcome_occurred",
        "outcome-hypothetical": "outcome_hypothetical",
        "prison-term": "prison_term",
        "where": "where",
        "who": "who",
    },
    "Terrorplate": {
        "blamed-by": "blamed_by",
        "claimed-by": "claimed_by",
        "completion": "completion",
        "coordinated": "coordinated",
        "killed": "killed",
        "kidnapped": "kidnapped",
        "named-perp": "named_perp",
        "named-perp-org": "named_perp_org",
        "named-organizer": "named_organizer",
        "over-time": "over_time",
        "outcome-averted": "outcome_averted",
        "outcome-occurred": "outcome_occurred",
        "outcome-hypothetical": "outcome_hypothetical",
        "perp-captured": "perp_captured",
        "perp-killed": "perp_killed",
        "perp-objective": "perp_objective",
        "perp-wounded": "perp_wounded",
        "target-human": "target_human",
        "target-physical": "target_physical",
        "terror-event": "terror_event",
        "type": "terrorism_type",
        "weapon": "weapon",
        "when": "when",
        "where": "where",
        "wounded": "wounded",
    },
    "Epidemiplate": {
        "disease": "disease",
        "exposed-count": "exposed_count",
        "exposed-cumulative": "exposed_cumulative",
        "exposed-individuals": "exposed_individuals",
        "hospitalized-count": "hospitalized_count",
        "hospitalized-cumulative": "hospitalized_cumulative",
        "hospitalized-individual": "hospitalized_individuals",
        "infected-count": "infected_count",
        "infected-cumulative": "infected_cumulative",
        "infected-individuals": "infected_individuals",
        "killed-count": "killed_count",
        "killed-cumulative": "killed_cumulative",
        "killed-individuals": "killed_individuals",
        "NPI-Events": "NPI_Events",
        "outbreak-event": "outbreak_event",
        "tested-count": "tested_count",
        "tested-cumulative": "tested_cumulative",
        "tested-individuals": "tested_individuals",
        "recovered-count": "recovered_count",
        "recovered-cumulative": "recovered_cumulative",
        "recovered-individuals": "recovered_individuals",
        "vaccinated-count": "vaccinated_count",
        "vaccinated-cumulative": "vaccinated_cumulative",
        "vaccinated-individuals": "vaccinated_individuals",
        "where": "where",
    },
    "Disasterplate": {
        "affected-cumulative-count": "affected_cumulative_count",
        "announce-disaster-warnings": "announce_disaster_warnings",
        "assistance-needed": "assistance_needed",
        "assistance-provided": "assistance_provided",
        "damage": "damage",
        "declare-emergency": "declare_emergency",
        "disease-outbreak-events": "disease_outbreak_events",
        "human-displacement-events": "human_displacement_events",
        "individuals-affected": "individuals_affected",
        "injured-count": "injured_count",
        "killed-count": "killed_count",
        "major-disaster-event": "major_disaster_event",
        "missing-count": "missing_count",
        "outcome": "outcome",
        "over-time": "over_time",
        "rescue-events": "rescue_events",
        "rescued-count": "rescued_count",
        "related-natural-phenomena": "related_natural_phenomena",
        "repair": "repair",
        "responders": "responders",
        "when": "when",
        "where": "where",
    },
    "Displacementplate": {
        "Assistance-provided": "assistance_provided",
        "Assistance-needed": "assistance_needed",
        "blocked-migration-count": "blocked_migration_count",
        "current-location": "current_location",
        "destination": "destination",
        "detained-count": "detained_count",
        "event-or-SoA-at-origin": "event_or_soa_at_origin",
        "group-identity": "group_identity",
        "human-displacement-event": "human_displacement_event",
        "injured-count": "injured_count",
        "killed-count": "killed_count",
        "missing-count": "missing_count",
        "origin": "origin",
        "outcome": "outcome",
        "over-time": "over_time",
        "settlement-status-event-or-SoA": "settlement_status_event_or_soa",
        "total-displaced-count": "total_displaced_count",
        "Transitory-events": "transitory_events",
        "transiting-location": "transiting_location",
        "when": "when",
    },
    "Etiplate": {
        "agreement-length": "agreement_length",
        "contract-amount": "contract_amount",
        "contract-awardee": "contract_awardee",
        "contract-awarder": "contract_awarder",
        "etip-event": "etip_event",
        "funding-amount": "funding_amount",
        "funding-recipient": "funding_recipient",
        "funding-source": "funding_source",
        "outcome": "outcome",
        "over-time": "over_time",
        "overall-project-value": "overall_project_value",
        "project-location": "project_location",
        "project-name": "project_name",
        "project-type": "project_type",
        "signatories": "signatories",
        "when": "when",
    },
    "Cybercrimeplate": {
        "cybercrime-event": "cybercrime_event",
        "information-stolen": "information_stolen",
        "outcome": "outcome",
        "over-time": "over_time",
        "perpetrator": "perpetrator",
        "related-crimes": "related_crimes",
        "response": "response",
        "victim": "victim",
        "victim-impact": "victim_impact",
        "when": "when",
    },
}
GRANULAR_MAT_ATTACHMENT_TYPES = {
    # MAT annotation label : BPDocumentGranularTemplateAttachment type
    "Time-attachment": "Time",
}
GRANULAR_MAT_COREF_TYPES = {
    # MAT annotation label : BPDocumentGranularTemplateCoref type
    "Coref-Thing": "Thing",
    "Coref-Person": "Person",
    "Coref-Org": "Org",
    "Coref-Place": "Place",
    "Coref-Country": "Country",
    "Coref-Date": "Date",
    "Coref-Disease": "Disease",
}
GRANULAR_MAT_IRREALIS_STATUSES = [
    "counterfactual", "hypothetical", "future", "unconfirmed", "unspecified", "non-occurrence"
]

ARBAGRAN_MAT_SENTENCE_TYPES = ["S",]

"""Scoring policy constants."""
SCORING_HEAD_MATCH_VALUE_DEFAULT = 0.5
SCORING_INCORRECT_SPAN_REDUCTION_RATE_DEFAULT = 0.05
SCORING_PARTIAL_MATCH_TYPES = ["characters", "tokens"]
SCORING_PARTIAL_MATCH_TYPE_DEFAULT = "characters"
SCORING_ARG_MATCH_TYPES = [
    'strict-head', 'strict-full', 'weighted-head', 'weighted-full', 'loose', 'ignore'
]
SCORING_ARG_MATCH_TYPE_DEFAULT = 'weighted-full'
SCORING_QUADCLASS_MATCH_TYPES = ['strict', 'loose', 'ignore',]
SCORING_QUADCLASS_MATCH_TYPE_DEFAULT = 'ignore'
SCORING_ANCHOR_MATCH_TYPES = ['strict', 'loose', 'ignore']
SCORING_ANCHOR_MATCH_TYPE_DEFAULT = 'ignore'
SCORING_EVENT_TYPE_MATCH_TYPES = ['strict', 'loose', 'ignore',]
SCORING_EVENT_TYPE_MATCH_TYPE_DEFAULT = 'loose'
SCORING_REF_ARG_SCORING_TYPES = ["anchors-as-string-args", "event-id-mapping",]
SCORING_REF_ARG_SCORING_TYPE_DEFAULT = "anchors-as-string-args"
SCORING_EVENT_MATCH_INSIDE_SENTENCE_DEFAULT = True
SCORING_IGNORED_SLOTS = [
    "corrupt_event", "terror_event", "protest_event", "outbreak_event", "etip_event",
    "cybercrime_event",

    # ignored in Phase 1
    # "hospitalized_individuals", "perp_captured",

    # ignored as of Phase 2
    "outcome_averted", "outcome_hypothetical", "outcome_occurred",
]

"""Standardize detected punctuation characters, tokens, etc."""
PUNCTUATION = punctuation + \
    """”“–’€´§¨«¬®¯°»×™¤£¥‹›“”„‚‘’¢฿₤—–Ω⅓⅔Ⅲ↑→∫⌘⌥┗▣▼◀◎◘☆☑☺♡✤✧✩❄❒❞❤➚➺⦁⬛…•©""" + \
    """،؍؎؏؛؞؟٭۔۝۞۩۽﴾﴿""" # Arabic punctuation symbols

"""Currency symbols, both left-to-right and right-to-left.
See: https://www.compart.com/en/unicode/category/Sc
"""
LTR_CURRENCY = "$¢£¤¥৲৳৻૱௹฿៛₠₡₢₣₤₥₦₧₨₩₪₫€₭₮₯₰₱₲₳₴₵₶₷₸₹₺₻₼₽₾₿꠸﹩＄￠￡￥￦"
RTL_CURRENCY = "؋﷼߿߾"

"""UTF-8 representation of the byte-order mark, which is unecessary in UTF-8."""
BOM = '\ufeff'

"""Default spaCy model to use, when necessary."""
SPACY_MODEL_NAME_DEFAULT = "en_core_web_lg"


## Utility Functions


annot_in_bounds = lambda annot, bounds : annot.start >= bounds[0] and annot.end <= bounds[1]
"""Determines whether a MAT annotation is completely within the bounds of the 2-tuple."""


annot_window = lambda start, end, annots : [
    annot for annot in annots if annot.start >= start and annot.end <= end
]
"""Subselects only annotations within the specified range in a MAT document."""


def capture_logs(path, verbose=False, force=False):
    """Captures logging output to the specified path.

    Args:
        path (str): a path to write the JSON blob.
        verbose (bool): print verbose output if True.
        force (bool): skip overwrite check and assert "OK" if True.
    """
    import logging
    import os

    global LOGGER

    if os.path.dirname(path):
        assert os.path.exists(os.path.dirname(path))

    if user_allows_overwrite(path, verbose, force):

        if os.path.exists(path):
            os.remove(path)

        LOGGER.add_fh(path)

    elif verbose:

        LOGGER.info(f"    User chose to not overwrite existing log output file.")


def __check_unsolvability(matrix):
    """Checks additional conditions to see if an input Munkres cost matrix is unsolvable.

    Munkres v1.1.4 suffers from an infinite loop given some edge conditions, when it should,
    instead, emit a ``munkres.UnsolvableMatrix`` error. This function attempts to identify those
    conditions.

    Args:
        matrix (list): of lists of numbers, representing a cost matrix (or a profit matrix) suitable
            for Munkres optimization.

    Raises:
        munkres.UnsolvableMatrix: if the matrix is unsolvable.

    .. Notes

       Hilariously, [this issue is unresolved](https://github.com/bmc/munkres/issues/28). I think I
       see the pattern to check for, though. If there exists more than 1 row where all of the values
       are ``DISALLOWED`` except for one, and the one non-DISALLOWED value has the same index in at
       least two of those rows, then it's unsolvable.

       I don't know if the same is true column-wise. I may try that as well, if I run into similar
       problems with columns. If I do, I could simply transpose the matrix and recurse once. Maybe.
    """

    from munkres import DISALLOWED
    from munkres import UnsolvableMatrix

    non_disallowed_indices = [] # (in possibly-offending rows)

    for row in matrix:

        ## check to see if all but 1 cell in the row are DISALLOWED

        indices = [i for i in range(len(row)) if type(row[i]) != type(DISALLOWED)]

        if len(indices) == 1:

            if indices[0] in non_disallowed_indices:
                raise UnsolvableMatrix(
                    "Encountered edge condition that the Munkres library doesn't check for!"
                )

            non_disallowed_indices.append(indices[0])


def detect_content_in_segsecs(mat_doc):
    """Detects illegal content annotations within the bounds of Segment Sections in MAT documents.
    Only works for Basic and Granular dialects.

    Args:
        mat_doc (MAT.Document.AnnotatedDoc): a document to parse.

    Returns:
        dict: keyed with segment sections that shouldn't have annotations whose value is a list of
        2-tuples of content anontations that shouldn't appear within their boundaries and "type" of
        overlap.
    """

    from MAT.Annotation import SpanlessAnnotation

    global BASIC_MAT_SENTENCE_TYPES
    global BASIC_MAT_STRUCTURAL_TYPES

    skippable_annot_types = ["SEGMENT",]
    legal_segsec_types = BASIC_MAT_SENTENCE_TYPES + ["Headline",]
    illegal_segsegs = mat_doc.getAnnotations(
        atypes=[type_ for type_ in BASIC_MAT_STRUCTURAL_TYPES if type_ not in legal_segsec_types]
    )
    annots = mat_doc.getAnnotations() # all annotations

    results = {}
    for segsec in illegal_segsegs:

        results[segsec] = []

        for annot in annots:

            if annot in illegal_segsegs:
                continue

            if annot.atype.lab in skippable_annot_types:
                continue

            if type(annot) == SpanlessAnnotation:
                continue

            starts_within = annot.start >= segsec.start and annot.start < segsec.end
            ends_within = annot.end > segsec.start and annot.end <= segsec.end
            duplicate = annot.start == segsec.start and annot.end == segsec.end

            if annot_in_bounds(annot, (segsec.start, segsec.end)):

                results[segsec].append((annot, "encompassed by segsec"))

            elif starts_within:

                results[segsec].append((annot, "starts within segsec"))

            elif ends_within:

                results[segsec].append((annot, "ends within segsec"))

            elif duplicate:

                results[segsec].append((annot, "shares same bounds as segsec"))

    return results


def detect_overlap_mat(annot_a, annot_b):
    """Determines whether two MAT annotations overlap, and in what way.

    Args:
        annot_a (MAT.Annotation.Annotation): an annotation to check.
        annot_b (MAT.Annotation.Annotation): another annotation to check.

    Returns:
        str: the type of overlap, or ``None`` if the annotations don't overlap. If not ``None``, the
        string values are meant to be read as "A {type_of_overlap} B."
    """
    starts_within = annot_a.start >= annot_b.start and annot_a.start < annot_b.end
    ends_within = annot_a.end > annot_b.start and annot_a.end <= annot_b.end
    bounded_by = annot_a.start >= annot_b.start and annot_a.end <= annot_b.end
    encompasses = annot_a.start <= annot_b.start and annot_a.end >= annot_b.end
    duplicate = annot_a.start == annot_b.start and annot_a.end == annot_b.end

    if duplicate:

        return "duplicate_of"

    elif bounded_by:

        return "bounded_by"

    elif encompasses:

        return "encompasses"

    elif starts_within:

        return "starts_within"

    elif ends_within:

        return "ends_within"

    else:

        return None


def detect_overlapping_bp_sents(doc, verbose=False):
    """Detects overlapping sentence annotations within a BP document.

    Args:
        doc (cp.BPDocument): a BP document to parse.
        verbose (bool): print verbose output if True.

    Returns:
        dict: keys "duplicates", which is a dict keyed with 2-tuples of starting-stopping indices
        that points to a list of annotation objects (unique-ified), and "overlaps" which points to a
        list of 2-tuples of annotation ovbjects that overlap each other in some way.
    """

    global TQDM_PREFIX

    tqdm = get_tqdm(verbose, False)

    results = {}

    if doc.entries.value:
        for entry_id, entry in tqdm(
            doc.entries.value.items(),
            desc=f"{TQDM_PREFIX} Detecting overlappying sentences",
            unit=" entry"
        ):
            if entry.segment_sections.value:

                results[entry_id] = {
                    "duplicates": {},
                    "overlaps": [],
                }

                for segsec in entry.segment_sections.value:
                    if segsec.structural_element.value == "Sentence":

                        for segsec_ in entry.segment_sections.value:
                            if segsec_.structural_element.value == "Sentence":

                                if segsec is segsec_:
                                    continue

                                starts_within = segsec.start.value >= segsec_.start.value and \
                                    segsec.start.value < segsec_.end.value

                                ends_within = segsec.end.value > segsec_.start.value and \
                                    segsec.end.value <= segsec_.end.value

                                bounded_by = segsec.start.value >= segsec_.start.value and \
                                    segsec.end.value <= segsec_.end.value

                                encompasses = segsec.start.value <= segsec_.start.value and \
                                    segsec.end.value >= segsec_.end.value

                                duplicate = segsec.start.value == segsec_.start.value and \
                                    segsec.end.value == segsec_.end.value

                                if duplicate:

                                    if (segsec.start.value, segsec.end.value) not in \
                                        results[entry_id]["duplicates"]:

                                        results[entry_id]["duplicates"][
                                            (segsec.start.value, segsec.end.value)
                                        ] = []

                                    key = results[entry_id]["duplicates"][
                                        (segsec.start.value, segsec.end.value)
                                    ]

                                    if id(segsec) not in key:
                                        key.append(id(segsec))
                                    if id(segsec_) not in key:
                                        key.append(id(segsec_))

                                elif bounded_by:

                                    if (segsec, segsec_, "A bounded_by B") not in \
                                        results[entry_id]["overlaps"]:

                                        results[entry_id]["overlaps"].append(
                                            (segsec, segsec_, "A bounded_by B")
                                        )

                                elif encompasses:

                                    if (segsec, segsec_, "A encompasses B") not in \
                                        results[entry_id]["overlaps"]:

                                        results[entry_id]["overlaps"].append(
                                            (segsec, segsec_, "A encompasses B")
                                        )

                                elif starts_within:

                                    if (segsec, segsec_, "A starts within B") not in \
                                        results[entry_id]["overlaps"]:

                                        results[entry_id]["overlaps"].append(
                                            (segsec, segsec_, "A starts within B")
                                        )

                                elif ends_within:

                                    if (segsec, segsec_, "A ends within B") not in \
                                        results[entry_id]["overlaps"]:

                                        results[entry_id]["overlaps"].append(
                                            (segsec, segsec_, "A ends within B")
                                        )

    return results


def detect_overlapping_mat_annots(mat_doc, type_):
    """Detects overlapping annotations of a particular type within a MAT document.

    Args:
        mat_doc (MAT.Document.AnnotatedDoc): a document to parse.
        type_ (str): a MAT annotation type to parse within the document.

    Returns:
        dict: keys "duplicates", which is a dict keyed with 2-tuples of starting-stopping indices
        that points to a list of annotation objects (unique-ified), and "overlaps" which points to a
        list of 2-tuples of annotation ovbjects that overlap each other in some way.
    """

    from MAT.Annotation import SpanlessAnnotation

    results = {
        "duplicates": {},
        "overlaps": [],
    }

    aset = mat_doc.getAnnotations(atypes=[type_,])

    for annot in aset:

        if type(annot) == SpanlessAnnotation:
            continue

        for annot_ in aset:

            if annot == annot_:
                continue

            if type(annot_) == SpanlessAnnotation:
                continue

            overlap = detect_overlap_mat(annot, annot_)

            if overlap == "duplicate_of":

                if (annot.start, annot.end) not in results["duplicates"]:
                    results["duplicates"][(annot.start, annot.end)] = []

                key = results["duplicates"][(annot.start, annot.end)]

                if annot not in key:
                    key.append(annot)
                if annot_ not in key:
                    key.append(annot_)

            elif overlap == "bounded_by":

                results["overlaps"].append(
                    (annot, annot_, "A bounded_by B")
                )

            elif overlap == "encompasses":

                results["overlaps"].append(
                    (annot, annot_, "A encompasses B")
                )

            elif overlap == "starts_within":

                results["overlaps"].append(
                    (annot, annot_, "A starts within B")
                )

            elif overlap == "ends_within":

                results["overlaps"].append(
                    (annot, annot_, "A ends within B")
                )

    return results


def detect_overlapping_mat_events(mat_doc, dialect):
    """Detects overlapping annotations of a particular type within a MAT document.

    Args:
        mat_doc (MAT.Document.AnnotatedDoc): a document to parse.
        dialect (str): the MAT annotation dialect. One of ``MAT_DIALECT_CHOICES``.

    Returns:
        dict: keys "duplicates", which is a dict keyed with 2-tuples of starting-stopping indices
        that points to a list of annotation objects (unique-ified), and "overlaps" which points to a
        list of 2-tuples of annotation ovbjects that overlap each other in some way.
    """

    global BASIC_MAT_KEY_ANCHOR_TYPES

    if dialect in ["Basic", "Granular",]:
        atypes = BASIC_MAT_KEY_ANCHOR_TYPES
    elif dialect == "Abstract":
        atypes = ABSTRACT_MAT_ANCHOR_TYPES
    else:
        raise ValueError(f"'{dialect}' is not one of {MAT_DIALECT_CHOICES}!")

    results = {
        "duplicates": {},
        "overlaps": [],
    }

    aset = mat_doc.getAnnotations(atypes=atypes)

    for annot in aset:

        for annot_ in aset:

            if annot == annot_:
                continue

            overlap = detect_overlap_mat(annot, annot_)

            if overlap == "duplicate_of":

                if (annot.start, annot.end) not in results["duplicates"]:
                    results["duplicates"][(annot.start, annot.end)] = []

                key = results["duplicates"][(annot.start, annot.end)]

                if annot not in key:
                    key.append(annot)
                if annot_ not in key:
                    key.append(annot_)

            elif overlap == "bounded_by":

                results["overlaps"].append(
                    (annot, annot_, "A bounded_by B")
                )

            elif overlap == "encompasses":

                results["overlaps"].append(
                    (annot, annot_, "A encompasses B")
                )

            elif overlap == "starts_within":

                results["overlaps"].append(
                    (annot, annot_, "A starts within B")
                )

            elif overlap == "ends_within":

                results["overlaps"].append(
                    (annot, annot_, "A ends within B")
                )

    return results


def detect_shared_addtl_achors(mat_doc, dialect):
    """Detects additional-anchor annotations that are referred to by multiple Event annotations, of
    any type, within MAT documents.

    Args:
        mat_doc (MAT.Document.AnnotatedDoc): a document to parse.
        dialect (str): the MAT annotation dialect. One of ``MAT_DIALECT_CHOICES``.

    Returns:
        dict: keyed with string representations of additional-anchor annotations, whose values are
        lists of string representations of anchor annotations that point to the associated
        additional-anchor.
    """

    global ABSTRACT_MAT_ADDITIONAL_ANCHOR_TYPES
    global BASIC_MAT_ADDITIONAL_ANCHOR_TYPES
    global BASIC_MAT_KEY_ANCHOR_TYPES
    global MAT_DIALECT_CHOICES

    ## first, gather all additional anchor annotations

    if dialect in ["Basic", "Granular",]:
        atypes = BASIC_MAT_ADDITIONAL_ANCHOR_TYPES
    elif dialect == "Abstract":
        atypes = ABSTRACT_MAT_ADDITIONAL_ANCHOR_TYPES
    else:
        raise ValueError(f"'{dialect}' is not one of {MAT_DIALECT_CHOICES}!")

    aset = mat_doc.getAnnotations(atypes=atypes)
    results = {pretty_annot(annot, indent=None, doc=mat_doc): [] for annot in aset}

    ## next, identify which anchors point to the additional-anchor

    if dialect in ["Basic", "Granular",]:
        atypes = BASIC_MAT_KEY_ANCHOR_TYPES
    elif dialect == "Abstract":
        atypes = ["AbstractEvent"]

    aset = mat_doc.getAnnotations(atypes=atypes)
    for annot in aset:
        if annot.get("addtl_anchors") is not None:
            for addtl_annot in annot.get("addtl_anchors"):

                addtl_annot_pretty = pretty_annot(addtl_annot, indent=None, doc=mat_doc)

                assert addtl_annot_pretty in results, \
                    f"Additional anchor {addtl_annot.describe()} not indexed in results!"

                results[addtl_annot_pretty].append(
                    pretty_annot(annot, indent=None, doc=mat_doc)
                )

    return results


def detect_unsentenced_annots(mat_doc):
    """Finds all annotations in a MAT document that aren't bounded by sentences but should be.

    Args:
        mat_doc (MAT.Document.AnnotatedDoc): a document to parse.

    Returns:
        list: of annotations that should be bounded by sentences by aren't.
    """

    global BASIC_MAT_STRUCTURAL_TYPES

    aset = mat_doc.getAnnotations(
        BASIC_MAT_KEY_MENTION_TYPES + \
        BASIC_MAT_ADDITIONAL_MENTION_TYPES + \
        BASIC_MAT_KEY_ANCHOR_TYPES + \
        BASIC_MAT_FAUX_ANCHOR_TYPES + \
        BASIC_MAT_ADDITIONAL_ANCHOR_TYPES
    )
    sentences = mat_doc.getAnnotations(["S", "Headline",])

    bad = []
    for annot in aset:

        bounded = False
        for sentence in sentences:

            overlap = detect_overlap_mat(annot, sentence)
            if overlap == "bounded_by":
                bounded = True

        if not bounded:
            bad.append(annot)

    return bad


def detect_unsentenced_text(mat_doc):
    """Detects text not bounded by an annotation of the given type within the given MAT doc.

    Args:
        mat_doc (MAT.Document.AnnotatedDoc): a document to parse.
        type_ (str): a MAT annotation type to parse within the document.

    Returns:
        list: of "Unbounded" annotations.
    """

    aset = mat_doc.getAnnotations(atypes=["S", "Headline",])

    start = None
    end = None
    found = False
    for i in range(len(mat_doc.signal)):

        overlap = False
        for annot in aset:
            if annot.start <= i and annot.end >= i:
                overlap = True

        if not overlap and not found:
            start = i
            found = True

        if overlap and found:
            end = i
            mat_doc.createAnnotation(start, end, "Unbounded")
            found = False
            start = None
            end = None

    if found:
        mat_doc.createAnnotation(start, i, "Unbounded")

    return mat_doc.getAnnotations(atypes=["Unbounded",])


def dict_get_icase(dict_: dict, key: str, default=None):
    """Attempts to find a case-insensitive match for the given key within the given dictionary and
    return the values at that key.

    The dictionary may have non-string keys, but the given key to search for _must_ be a string.

    Consider inputs before using this function! If any keys would match when forcing lowercase,
    you'll run into collisions. For example:

    >>> x = { 'A': 1, 'a': 2, }
    >>> y = dict_get_icase(x, "a") # raises KeyError

    Args:
        dict_ (dict): a dictionary to search within.
        key (str): a key to find (case-insensitive match) within the dictionary.
        default (*): a default value to return if no match is found.

    Raises:
        KeyError: if there are any duplicate keys when forcing all keys to be lowercase.

    Returns:
        Any: the value at the given key within the dictionary, otherwise the default value.
    """

    candidates = {}
    for k, v in dict_.items():
        if type(k) == str and k.lower() == key.lower():
            candidates[k] = v

    if len(candidates) > 1:
        raise KeyError(f"Collisions with case-insensitive search for '{key}': {candidates}")

    elif len(candidates) == 0:
        return default

    else:

        return candidates[list(candidates.keys())[0]]


def find_sentence_regions(mat_doc):
    """Finds the sentence boundaries in a MAT document.

    This is *not* a sentence segmenter. It detects sentence boundaries within MAT documents already
    formatted to convey that information.

    Args:
        mat_doc (MAT.Document.AnnotatedDoc): a MAT document to look for sentences in.

    Returns:
        list: of 2-tuples of ``int`` values indicating starting and stopping indices of sentences
        within the document.
    """

    def __next_nl(text, start):
        nl = text.find("\n", start)
        cr = text.find("\r", start)
        if (nl >= 0):
            if (cr >= 0):
                return min(nl, cr)
            else:
                return nl
        else:
            return cr

    bounds = []
    prev_idx = 0
    while prev_idx >= 0 and prev_idx+1 < len(mat_doc.signal):

        next_idx = __next_nl(mat_doc.signal, prev_idx+1)

        if next_idx-1 > prev_idx:

            window = mat_doc.signal[prev_idx:next_idx]
            if len(window.strip()) > 0:
                bounds.append([prev_idx, next_idx])

        prev_idx = next_idx

    if len(bounds) == 0:

        # check if file empty
        if len(mat_doc.signal.strip()) == 0:
            return bounds
        else:
            bounds = [(0, len(mat_doc.signal)-1)]
            return bounds

    else:

        # Get region following final newline
        last_bound = bounds[-1]
        trailing_text = mat_doc.signal[last_bound[1]:].strip()

        if len(trailing_text) == 0:
            # No additional signal following final newline character
            return bounds
        else:
            # there IS additional signal following final newline character
            bounds.append([last_bound[1], len(mat_doc.signal) - 1])
            return bounds


def fix_synclass(dialect, label):
    """Fixes inconsistencies bewteen synclasses in MAT tasks.

    Args:
        dialect (str): the MAT annotation dialect, used to determine *how* to "fix" the synclass
            label.
        label (str): a potentially "broken" synclass label.

    Returns:
        str: a "fixed" synclass label.

    .. P.S.: Leland hates this function has to exist.
    """

    if dialect == "Abstract":
        if label.lower() in ["nominal_mention", "nominal",]:
            return "nominal"
        elif label.lower() in ["name_mention", "name",]:
            return "name"
        elif label.lower() in ["pronoun_mention", "pronoun",]:
            return "pronoun"
        else:
            return None

    elif dialect == "Basic" or dialect == "Granular":
        if label.lower() in BASIC_MAT_NOMINAL_MENTION_TYPES:
            return "nominal"
        elif label.lower() in BASIC_MAT_PRONOUN_MENTION_TYPES:
            return "pronoun"
        elif label.lower() in BASIC_MAT_NAME_MENTION_TYPES:
            return "name"
        else:
            return None


def get_aset(doc, type_):
    """Gets an annotation set (``aset``) from the given MAT doc with the given type, if it exists.

    Args:
        doc (dict): a MAT JSON document, created with ``json.loads()``.
        type_ (str): the type of annotation to find in the document.

    Returns:
        dict: the annotation set of the specified type, or ``None`` if it doesn't exist.
    """
    assert 'asets' in doc, "Document is not valid MAT JSON!"
    for aset in doc['asets']:
        if aset['type'] == type_:
            return aset
    return None


def get_tqdm(verbose, debug):
    """Determines whether the ``tqdm`` function prints anything.

    Debug mode assumes verbose mode, in addition to debug-only output.

    If in verbose mode, print the ``tqdm`` progress bar as normal. If in quiet or debug mode, print
    nothing.

    Args:
        verbose (bool): whether in verbose mode or not.
        debug (bool): whether in debug mode or not.

    Returns:
        Either the standard ``tqdm.tqdm`` progress bar printer, or an override function that prints
        nothing.
    """

    def __tqdm(iterable, *args, **kwargs):
        return iterable

    if not verbose or (verbose and debug):
        tqdm = __tqdm
    else:
        from tqdm import tqdm

    return tqdm


def is_overlapped_by(annot, type_, mat_doc):
    """Detects whether an annotation overlaps any other annotation of the given type.

    Args:
        annot (MAT.Annotation.Annotation): an annotation to check with.
        type_ (str): a MAT annotation type to parse within the document.

    Returns:
        bool: whether the given annotation overlaps any other annotation of the given type.
    """

    aset = mat_doc.getAnnotations(atypes=[type_,])

    for annot_ in aset:
        overlap = detect_overlap_mat(annot, annot_)
        if overlap:
            return True

    return False


def levenshtein(str_a, str_b):
    """A simple Levenshtein distance function.

    Written by `Christopher P. Matthews <christophermatthews1985@gmail.com>`_ for `Wikipedia`_.
    Lightly modified by Leland E. Vakarian.

    Args:
        str_a (str): A string to compare to ``str_b``.
        str_b (str): A string to compare to ``str_a``.

    Returns:
        int: the Levenshtein distance between the two string inputs.

    .. _Wikipedia:
       https://en.wikibooks.org/wiki/Algorithm_Implementation/Strings/Levenshtein_distance#Python
    """

    assert type(str_a) == str
    assert type(str_b) == str

    # simple O(1) conditionals
    if str_a == str_b:
        return 0
    elif len(str_a) == 0:
        return len(str_b)
    elif len(str_b) == 0:
        return len(str_a)

    vector_0 = [None] * (len(str_b) + 1)
    vector_1 = [None] * (len(str_b) + 1)

    for i in range(len(vector_0)):

        vector_0[i] = i

    for i in range(len(str_a)):

        vector_1[0] = i + 1

        for j in range(len(str_b)):

            cost = 0 if str_a[i] == str_b[j] else 1
            vector_1[j+1] = min(vector_1[j] + 1, vector_0[j+1] + 1, vector_0[j] + cost)

        for j in range(len(vector_0)):
            vector_0[j] = vector_1[j]

    return vector_1[len(str_b)]


def open_json_blob(path):
    """Opens a JSON document in a consistent manner, using the standard ``json`` module.

    Args:
        path (str): a path to a document to open.

    Returns:
        dict: a dictionary representing a JSON object.
    """
    import json

    with open(path, 'r', encoding='utf-8') as f:
        doc = json.loads(f.read())

    return doc


def pretty_annot(annot, indent=0, doc=None):
    """Makes a pretty-printable string representation of a MAT Annotation.

    Args:
        annot (MAT.Annotation.Annotation): an annotation to prettify.
        indent (int): if greater than 0, format such that "Location" and "Signal" appear on their
            own lines below the annotation type, indented to the left by ``indent`` "levels".
        doc (MAT.Document.AnnotatedDoc): if passed, include the signal text this annotation
            encompasses. If not, don't include signal text.

    Returns:
        str: a pretty-printable representation of the annotation.
    """
    import MAT

    if annot.id:
        string_ = f"[{annot.id}]{annot.atype.lab}"
    else:
        string_ = f"{annot.atype.lab}"

    if isinstance(annot, MAT.Annotation.Annotation):

        if indent is None:

            string_ += f"<{annot.start}:{annot.end}>"

            if doc:
                string_ += f" ('{signal(doc, annot)}')"

        elif indent == 0:

            string_ += f" | Location: {annot.start}:{annot.end}"

            if doc:
                string_ += f" | Signal: '{signal(doc, annot)}'"

        else:

            lead = ">" * (indent * 4 - 1) # string * -1 = ""
            if lead:
                lead += " "

            string_ += "\n"
            string_ += lead
            string_ += f"Location: {annot.start}:{annot.end}"

            if doc:
                string_ += "\n" + lead
                string_ += f"Signal: '{signal(doc, annot)}'"

    return string_


def pretty_munkres_matrix(matrix, index_pairs=[]):
    """Builds a pretty-printable list of strings from a Munkres cost (or profit) matrix.

    It's expected that the matrix is square.

    Args:
        matrix (list): a list of lists of elements (ints, floats, etc), representing a Munkres cost
            (or profit) matrix.
        index_pairs (list): a list of 2-tuples of ints, representing selected cells chosen by the
            Munkres algorithm. Including this value will highlight the selected cells, while
            omitting this value will leave every cell unhighlighted.

    Returns:
        list: a list of pretty-printable string representations of rows from the input matrix.
    """

    from click import style
    from tabulate import tabulate

    vdata = []
    for r in range(len(matrix)):
        row = []
        for c in range(len(matrix[r])):
            if (r, c) in index_pairs:
                row.append(style(str(matrix[r][c]), fg='bright_green'))
            else:
                row.append(str(matrix[r][c]))
        vdata.append(row)

    return [line for line in tabulate(vdata, tablefmt="fancy_grid", numalign="center").split("\n")]


def resolve_paths(
    paths,
    type_,
    no_recurse=False,
    allow_duplicates=False,
    verbose=False,
    debug=False
):
    """Builds a list of unique file paths.

    If any files have the same contents, the first file listed will be kept, and all others will be
    ignored.

    Args:
        paths (list): a list of path names (str) to either a specific file or a directory of files
            to include. All files found in each directory of the specified type will be included.
        type_ (str): a file extension to look for, such as ".mat.json". Include the leading period.
        no_recurse (bool): if True, don't look for files in subdirectories of dirs passed in
            ``paths``.
        allow_duplicates (bool): if True, don't dedupe the files.
        verbose (bool): whether in verbose mode or not.
        debug (bool): whether in debug mode or not.

    Returns:
        list: string paths to unique files.
    """
    import filecmp
    import os

    global LOGGER

    tqdm = get_tqdm(verbose, debug)

    assert type(paths) == list or type(paths) == tuple

    if debug:
        LOGGER.debug(f"Input paths: {paths}")

    candidates = []
    for path in tqdm(paths, desc=f"{TQDM_PREFIX} Flattening paths ", unit="path"):

        if os.path.isfile(path):

            if path.lower().endswith(type_):
                candidates.append(path)
            elif debug:
                LOGGER.debug(f"Encountered file with different extension: '{path}'")

        elif os.path.isdir(path):

            if debug:
                LOGGER.debug(f"Found dir '{path}', recursing.")

            for root, dirs, files in os.walk(path):

                for file_ in files:

                    if file_.lower().endswith(type_):
                        candidates.append(os.path.join(root, file_))
                    elif debug:
                        LOGGER.debug(f"Encountered file with different extension: '{file_}'")

                if no_recurse:
                    break
        elif debug:
            LOGGER.debug(f"Skipping '{path}', not a file or a directory.")

    if debug:
        LOGGER.debug(f"Found {len(candidates)} files with the '{type_}' extension.")

    if allow_duplicates:

        return candidates

    files = []

    for path in tqdm(candidates, desc=f"{TQDM_PREFIX} Deduping files ", unit="file"):

        exists = False
        for file_ in files:
            if filecmp.cmp(path, file_, shallow=False):
                exists = True
                if verbose:
                    LOGGER.warning(
                        f"File `{path}` is a duplicate of `{file_}`, which was already added to the"
                        f" list. Skipping!"
                    )
        if not exists:
            files.append(path)

    return files


def signal(mat_doc, annot):
    """Gets a slice of a MAT annotation's signal string.

    Args:
        mat_doc (MAT.Document.AnnotatedDoc): the document the ``annot`` hails from.
        annot (MAT.Annotation.Annotation): an annotation to extract boundary information from.
    """
    import MAT
    assert isinstance(mat_doc, MAT.Document.AnnotatedDoc), \
        f"Value for `mat_doc` is not a MAT.Document.AnnotatedDoc: {type(annot)}"
    assert isinstance(annot, MAT.Annotation.Annotation), \
        f"Value for `annot` is not a MAT.Annotation.Annotation: {type(annot)}"
    assert mat_doc.signal, f"Document Signal isn't populated: `{mat_doc.signal}`"
    return mat_doc.signal[annot.start:annot.end]


def user_allows_overwrite(path, verbose=False, force=False):
    """Interactively checks to see if the user wants to overwrite a file at the given path.

    Set ``force`` to ``True`` to bypass interactively asking on the CLI, and assume "it's OK to
    overwrite".

    Args:
        path (str): a string containing a file location.
        verbose (bool): print verbose output if True.
        force (bool): skip this check and assert "OK" if True.

    Returns:
        bool: indication of whether writing a file to this path is "OK". If no file existed at the
        path, the assumption is "OK".
    """
    import os
    import sys

    import click

    global LOGGER

    warning = "[" + click.style("WARNING", fg='bright_yellow') + "]"

    if os.path.isfile(path):

        if not force:

            if verbose:
                ans = click.confirm((
                    f"{warning}  A file exists at the path '{path}'. Do you really want to "
                    "overwrite it?"
                ))
            else:
                ans = click.confirm(f"Overwrite '{path}'?")

            if ans:
                if verbose:
                    LOGGER.warning("    Overwriting file!")
            else:
                if verbose:
                    LOGGER.info("    Canceling due to denied overwrite.")
                return False

    return True


def validate_dist(doc, verbose=False):
    """Validates a ``dist.json`` file at ``path`` against the standard schema.

    Args:
        doc (dict): a dictionary constructed from reading-in a``dist.json`` file.
        verbose (bool): print verbose output if True.

    Returns:
        bool: ``True`` if the document conforms to the JSON schema, ``False`` otherwise.
    """

    import json
    import os

    from jsonschema import validate
    from jsonschema.exceptions import ValidationError

    from lib.utils import BASEPATH
    from lib.utils import open_json_blob

    global LOGGER

    schema = open_json_blob(os.path.join(BASEPATH, "schema", "dist.schema.json"))

    try:

        validate(doc, schema)

    except ValidationError as err:

        if verbose:
            LOGGER.error(err)
        else:
            LOGGER.error(str(err)[:str(err).find("\n")])
        return False

    return True


def validate_tasks(doc, verbose=False):
    """Validates a ``tasks.json`` file at ``path`` against the standard schema.

    Args:
        doc (dict): a dictionary constructed from reading-in a``tasks.json`` file.
        verbose (bool): print verbose output if True.

    Returns:
        bool: ``True`` if the document conforms to the JSON schema, ``False`` otherwise.
    """

    import json
    import os

    from jsonschema import validate
    from jsonschema.exceptions import ValidationError

    from lib.utils import BASEPATH
    from lib.utils import open_json_blob

    global LOGGER

    schema = open_json_blob(os.path.join(BASEPATH, "schema", "tasks.schema.json"))

    try:

        validate(doc, schema)

    except ValidationError as err:

        if verbose:
            LOGGER.error(err)
        else:
            LOGGER.error(str(err)[:str(err).find("\n")])
        return False

    return True


def write_json_blob(doc, path, indent=2, verbose=False, force=False, nosort=False):
    """Writes a JSON document in a consistent manner, using the standard ``json`` module.

    Args:
        doc (dict): a dictionary to translate into JSON.
        path (str): a path to write the JSON blob.
        indent (int): the number of spaces to insert at each indendation level. Set to ``0`` to make
            a one-line file.
        verbose (bool): print verbose output if True.
        force (bool): skip overwrite check and assert "OK" if True.
        nosort (bool): if True, disable automatic key sorting in output.
    """
    import json
    import os
    import sys

    global LOGGER

    try:
        if os.path.dirname(path):
            assert os.path.exists(os.path.dirname(path))
    except AssertionError:
        LOGGER.error(f"Directory {os.path.dirname(path)} doesn't exist!")
        sys.exit()

    if not indent:
        indent = None

    if user_allows_overwrite(path, verbose, force):
        with open(path, 'w', encoding='utf-8') as f:
            sort = not nosort
            blob = json.dumps(doc, indent=indent, sort_keys=sort, ensure_ascii=False)
            f.write(blob)

