#!/usr/bin/env python3

####################################################################################################
### -----------------------------------------  NOTICE  ----------------------------------------- ###
###                                                                                              ###
### This software (or technical data) was produced for the U.S. Government under contract        ###
### 2015-14120200002-002, and is subject to the Rights in Data-General Clause 52.227-14 , ALT    ###
### IV (MAY 2014) or (DEC 2007).                                                                 ###
###                                                                                              ###
### © 2019 The MITRE Corporation. All Rights Reserved.                                           ###
###                                                                                              ###
### -------------------------------------------------------------------------------------------- ###
####################################################################################################

"""
A command-line utility script that edits MAT JSON files. This technological terror has yet to be
rigorously tested.

Formatted to 100 cols, generally following the Google Python style guide (with some of my
modifications).

19 Aug, 2020

Leland E. Vakarian
Copyright (C) 2020 The MITRE Corporation.
"""

__author__  = "Leland E. Vakarian <lsperl@mitre.org>"
__status__  = "production"
__date__    = "20 Aug, 2020"

import click

from lib.utils import CONTEXT_SETTINGS
from lib.utils import LOGGER

DEBUG = False
FORCE = False
VERBOSE = False
"""bool: Global flags handled by cli()."""


def __validate_mat_json(mat_file):
    """Loads and validates the MAT JSON file.

    Here, 'validation' means verifying that the MAT document has a ``signal`` and ``asets`` keys.

    Args:
        mat_file (str): a string containing a file location.

    Returns:
        dict: a dictionary representing the loaded MAT JSON file.

    ---

    The checks made here are very simple; there's likely a better way to ensure that the format is
    correct, but this is sufficient for now.
    """
    import json
    import sys

    global DEBUG
    global FORCE
    global LOGGER
    global VERBOSE

    # file existence handled by ``click`` in ``prepare``
    with open(mat_file, 'r') as f:
        mat_json = json.loads(f.read())

    ## simple checks

    try:
        mat_json["signal"]
    except KeyError:
        if VERBOSE:
            LOGGER.error("MAT JSON file is missing 'signal'! Aborting!")
        else:
            click.echo("MAT JSON file is missing 'signal'!", err=True)
        sys.exit(131)

    try:
        mat_json["asets"]
    except KeyError:
        if VERBOSE:
            LOGGER.error("MAT JSON file is missing 'asets'! Aborting!")
        else:
            click.echo("MAT JSON file is missing 'asets'!", err=True)
        sys.exit(131)

    return mat_json


def __update_references(mat_json, deleted_id):
    """Updates the annotation references of MAT annotations affected by deletions.

    Args:
        mat_json (dict): a dictionary representing MAT JSON data.
        deleted_id (int): the id of the most recently-deleted annotation

    .. Notes:
       Asets with the attribute "hasID: true" indicates that the third value of each annotation
       instance is the ID value, which can either be an integer (as a string) or a None.
    """
    global DEBUG
    global LOGGER
    global VERBOSE

    deletions = 0

    for aset in mat_json['asets']:

        attr_offset = 0
        if "hasSpan" in aset and aset["hasSpan"]:
            attr_offset += 2
        if "hasID" in aset and aset["hasID"]:
            attr_offset += 1

        if "annots" in aset and type(aset["annots"]) == list:
            annots = aset["annots"]
        else:
            continue # need annots to process

        if "attrs" in aset and type(aset["attrs"]) == list:
            for i in range(len(aset["attrs"])):
                attr = aset["attrs"][i]
                if "type" in attr and attr["type"] == "annotation":
                    for annot in annots:
                        if "aggregation" in attr and attr["aggregation"] == "set":

                            if len(annot) <= i + attr_offset:
                                continue # mat compression leaves outlist elements
                            if annot[i+attr_offset] == None:
                                continue # no elements of the aggregation set means no references

                            assert type(annot[i+attr_offset]) == list, (
                                    f"aggregation set isn't a list! {aset['type']} -> "
                                    f"{attr['name']} (index {i+attr_offset}): {annot}"
                                )

                            indices_to_delete = []
                            for j in range(len(annot[i+attr_offset])):
                                ref = annot[i+attr_offset][j]
                                if ref == deleted_id:
                                    indices_to_delete.append(j)

                            for idx in indices_to_delete:
                                if DEBUG:
                                    LOGGER.debug(("        "
                                        f"Deleting reference in {aset['type']} -> {attr['name']} "
                                    ))
                                annot[i+attr_offset].pop(idx)
                                deletions += 1
                            if not len(annot[i+attr_offset]):
                                annot[i+attr_offset] = None

                        else:
                            if annot[i+attr_offset] == deleted_id:
                                if DEBUG:
                                    LOGGER.debug(("        "
                                        f"Deleting reference in {aset['type']} -> {attr['name']} "
                                    ))
                                annot[i+attr_offset] = None
                                deletions += 1
        else:
            continue # no attributes means no annotation references

    if DEBUG:
        LOGGER.debug(f"        Deleted {deletions} references")


def update_annotation_indices(mat_json, index, new=None, old=None):
    """Updates the indices of each annotation in the given mat_json object.

    Presumably, the mat_json has been validated before passing to this function.

    To be clear, the lengths described in the arguments are for the added or deleted portions of the
    string, _not_ the whole signal string.

    Args:
        mat_json (dict): a dictionary representing MAT JSON data.
        index (int): the index in signal where the change was made.
        new (int): the length of the new string if one was added.
        old (int): the length of the old string if one was deleted.

    Returns:
        int: the number of deleted annotations, if any; -1 if neither new nor old were passed.

    .. Notes:
       MAT JSON files handle annotation references _positionally_, within each list at the "annots"
       key for a particular annotation type. So, if one gets deleted by this function, every MAT
       annotion with a property that refers to another MAT annotation defined positionally _after_
       the deleted annotation will need to be updated to reflect that change. This is currently
       _not_ something this function handles. Presumably, the annotations you're deleting (like the
       "Delete-This-Text" annotations) have no other annotations within their bounds, so the problem
       never presents itself.
       Be careful. The indices for each MAT annotation are updated because of modification to the
       signal text, but _that's it_. Nothing is done to handle annotation reference modifications.
       You've been warned.
    """
    global DEBUG
    global LOGGER
    global VERBOSE

    update_type = ""
    if new is not None and old is not None:
        offset = new - old
        update_type = "replace"
    elif new is not None:
        offset = new
        update_type = "insert"
        old = 0
    elif old is not None:
        offset = old * -1
        new = 0
        update_type = "delete"
    else:
        return -1

    if DEBUG:
        LOGGER.debug(f"Got update type '{update_type}'.")

    new_end_index = index + new # convert from length to index after modified value
    old_end_index = index + old

    if offset == 0:
        return 0

    deleted = 0
    for aset in mat_json['asets']:

        counter = 0
        deletion_queue = []

        for annot in aset['annots']:

            # try to detect spanless annotations, and skip them
            try:
                if type(annot[0]) != int:
                    counter += 1
                    continue
            except IndexError:
                counter += 1
                continue

            if update_type == 'replace':

                if annot[0] < index and annot[1] <= index:
                    # annot before old string
                    pass
                elif annot[0] >= old_end_index and annot[1] > old_end_index:
                    # annot after old string
                    annot[0] += offset
                    annot[1] += offset
                if new >= old: # if replacement is the same or greater length
                    if annot[0] < index and annot[1] > index and annot[1] < old_end_index:
                        # annot crosses into and ends before old string end
                        pass
                    elif annot[0] < index and annot[1] > index and annot[1] == old_end_index:
                        # annot crosses into and ends at old string end
                        annot[1] += offset
                    elif annot[0] >= index and annot[0] < old_end_index and \
                        annot[1] > old_end_index:
                        # annot starts in old string and ends after
                        annot[1] += offset
                    elif annot[0] >= index and annot[0] <= old_end_index and \
                        annot[1] > index and annot[1] <= old_end_index:
                        # annot within old string
                        annot[1] += offset
                    elif annot[0] < index and annot[1] > old_end_index:
                        # annot begins before and ends after old string
                        annot[1] += offset
                else: # if replacement is less length
                    if annot[0] < index and annot[1] > index and annot[1] <= old_end_index:
                        # annot crosses into and ends before old string end
                        if annot[1] >= new_end_index:
                            annot[1] = new_end_index
                    elif annot[0] >= index and annot[0] < old_end_index and \
                        annot[1] > old_end_index:
                        # annot starts in old string and ends after
                        annot[0] += offset
                        annot[1] += offset
                        if annot[0] < index:
                            annot[0] = index
                    elif annot[0] >= index and annot[0] <= old_end_index and \
                        annot[1] > index and annot[1] <= old_end_index:
                        # annot within old string
                        if (annot[1] + offset) - annot[0] <= 0:
                            deletion_queue.append(counter)
                        else:
                            annot[1] += offset
                    elif annot[0] < index and annot[1] > old_end_index:
                        # annot begins before and ends after old string
                        annot[1] += offset

            elif update_type == 'insert':

                if annot[0] < index and annot[1] <= index:
                    pass # do nothing
                elif annot[0] >= old_end_index and annot[1] > old_end_index:
                    annot[0] += offset
                    annot[1] += offset
                elif annot[0] <= index and annot[1] > old_end_index:
                    annot[1] += offset

            elif update_type == 'delete':

                if annot[0] < index and annot[1] <= index:
                    pass # do nothing

                elif annot[0] >= old_end_index and annot[1] > old_end_index:
                    annot[0] += offset
                    annot[1] += offset

                # annot crosses into and ends before old string end
                elif annot[0] < index and annot[1] > index and annot[1] < old_end_index:

                    if annot[0] == index - 1: # annot would become length 0
                        deletion_queue.append(counter)
                    else:
                        annot[1] = index - 1

                # annot starts in old string and ends after
                elif annot[0] >= index and annot[0] < old_end_index and annot[1] > old_end_index:

                    if annot[1] + offset <= annot[0] - 1: # annot would become length 0
                        deletion_queue.append(counter)
                    else:
                        annot[0] -= 1
                        annot[1] += offset

                # annot within old string
                elif annot[0] >= index and annot[0] <= old_end_index and \
                    annot[1] > index and annot[1] <= old_end_index:

                    deletion_queue.append(counter) # annot length would be negative

                # annot begins before and ends after old string
                elif annot[0] < index and annot[1] > old_end_index:

                    annot[1] += offset

            if annot[0] < 0: # dumb
                annot[0] = 0

            counter += 1

        deleted += len(deletion_queue)
        deletion_queue.reverse()
        for flagged in deletion_queue:

            if DEBUG:
                LOGGER.debug(f"Deleting {aset['type']} {aset['annots'][flagged]}")

            if "hasID" in aset and aset["hasID"]:
                if DEBUG:
                    LOGGER.debug(f"    Deleting references to annotation with id")
                if "hasSpan" in aset and aset["hasSpan"]:
                    __update_references(mat_json, aset['annots'][flagged][2])
                else:
                    __update_references(mat_json, aset['annots'][flagged][0])

            aset['annots'].pop(flagged)

    return deleted


def write_mat_json(doc, path, force=True):
    """Dumps a JSON blob, representing MAT data, to disk.

    This is _not_ the ordinary way of handling MAT data.

    Args:
        doc (dict): a dictionary representing MAT JSON data.
        path (str): a string containing a file location.
        force (bool): if True, don't ask for permission when overwriting files. Defaults to True.

    ---

    The goal here is to write the file if overwrite checks are passed.
    """
    import json
    import os
    import sys

    from lib.utils import get_aset
    from lib.utils import write_json_blob

    global DEBUG
    global FORCE
    global LOGGER
    global VERBOSE

    # check if SEGMENT exists, and delete it if it does. avoids duplicate signal string on render.
    segment = get_aset(doc, "SEGMENT")
    if segment:
        doc['asets'].remove(segment)

    write_json_blob(doc, path, indent=0, verbose=VERBOSE, force=force)


@click.group(context_settings=CONTEXT_SETTINGS)
@click.option('-v', '--verbose', count=True, help="Show verbose output. Pass again for debug.")
@click.option('-f', '--force', is_flag=True, help="Skip asking to overwrite output.")
@click.option(
    '-l',
    '--save-logs',
    'save_logs',
    type=click.Path(),
    help="Capture logging output (at whatever verbosity level) to the specified path."
)
def cli(verbose, force, save_logs):
    """MAT JSON signal editor.

    Allows users to edit the signal in a MAT JSON file while safely preserving annotations. When
    the signal is edited, all existing annotations will be adjusted to preserve their location
    anchors.

    \b
    +------------------------------------------------------------+
    | This technological terror has yet to be rigorously tested. |
    +------------------------------------------------------------+

    \f

    This function serves as the entrypoint for the program when called on the CLI. It also handles
    global flags ``verbose`` and ``copy``.

    Args:
        verbose (bool): if True, pretty-print output as the command runs. If False, prints nothing
            as the command runs. Defaults to False.
        force (bool): if True, don't ask for permission when overwriting files. Defaults to False.
    """

   # do this before anything else
    from lib.utils import capture_logs
    if save_logs:
        capture_logs(save_logs, verbose, force)

    global DEBUG
    global FORCE
    global LOGGER
    global VERBOSE

    if verbose:
        LOGGER.info("Verbose output is on!")
        if verbose >= 1:
            VERBOSE = True
        if verbose > 1:
            DEBUG = True
            LOGGER.debug("Debug output is on!")

    if verbose and force:
        LOGGER.warning("Forced overwrite is on!")
    elif verbose and not force:
        LOGGER.info("Forced overwrite is off!")

    FORCE = force


@cli.command(context_settings=CONTEXT_SETTINGS)
@click.argument('mat_file', type=click.Path(exists=True))
@click.argument('out_file', type=click.Path())
@click.argument('index', type=int)
@click.argument('old_string')
@click.argument('new_string')
def replace(mat_file, in_file, index, old_string, new_string):
    """Replace an OLD_STRING at the given INDEX with a NEW_STRING.

    All annotation indices remaining in the signal will be updated to preserve offsets.

    \b
    Args:
      MAT_FILE    Path name to an input MAT JSON file.
      OUT_FILE    Path name to output a MAT JSON file.
      INDEX       Location in the signal (0-based index) at which to replacement.
      OLD_STRING  The string at INDEX you want to remove. As a check, if the string at INDEX does
                  not match OLD_STRING, the script will abort.
      NEW_STRING  The new value to insert into the signal at INDEX.

    \f

    Args:
        mat_file (str): path name to the MAT JSON file to edit.
        out_file (str): path name to the MAT JSON file to write out.
        index (int): the 0-based index of the string to replace.
        old_string (str): the string to replace. It need not be the same length as the new string.
        new_string (str): the string to insert.
    """
    import sys

    global DEBUG
    global FORCE
    global LOGGER
    global VERBOSE

    if index < 0:
        LOGGER.error("INDEX must be at least 0! Aborting!")
        sys.exit(131)

    ## load json data as python dict

    mat_json = __validate_mat_json(mat_file)

    if index >= len(mat_json['signal']):
        LOGGER.error("INDEX must be at less than the length of the signal! Aborting!")
        sys.exit(131)

    ## check that the ``old_string`` is at the given index

    if mat_json['signal'][index:index+len(old_string)] == old_string:

        ## replace old string with new string

        mat_json['signal'] = "%s%s%s" % (
            mat_json['signal'][:index],
            new_string,
            mat_json['signal'][index+len(old_string):]
        )

        ## calulate offset and update annotation indices

        deleted = update_annotation_indices(
            mat_json,
            index,
            new=len(new_string),
            old=len(old_string),
        )

        if VERBOSE:
            LOGGER.info(f"{deleted} annotations deleted due to modifications.")

        ## write new file

        write_mat_json(mat_json, out_file, force=FORCE)

    else:

        LOGGER.error("OLD_STRING does not match the signal at the given INDEX. Aborting!")
        sys.exit(131)


@cli.command(context_settings=CONTEXT_SETTINGS)
@click.argument('mat_file', type=click.Path(exists=True))
@click.argument('out_file', type=click.Path())
@click.argument('index', type=int)
@click.argument('string')
def delete(mat_file, out_file, index, string):
    """Delete an OLD_STRING at the given INDEX.

    All annotation indices remaining in the signal will be updated to preserve offsets.

    \b
    Args:
      MAT_FILE  Path name to an input MAT JSON file.
      OUT_FILE  Path name to output a MAT JSON file.
      INDEX     Location in the signal (0-based index) at which to delete.
      STRING    The string at INDEX you want to remove. As a check, if the string at INDEX does not
                match STRING, the script will abort.

    \f

    Args:
        mat_file (str): path name to the MAT JSON file to edit.
        out_file (str): path name to the MAT JSON file to write out.
        index (int): the 0-based index of the string to remove.
        string (str): the string to remove.
    """
    import sys

    global DEBUG
    global FORCE
    global LOGGER
    global VERBOSE

    if index < 0:
        LOGGER.error("INDEX must be at least 0! Aborting!")
        sys.exit(131)

    ## load json data as python dict

    mat_json = __validate_mat_json(mat_file)

    if index >= len(mat_json['signal']):
        LOGGER.error("INDEX must be at less than the length of the signal! Aborting!")
        sys.exit(131)

    ## check that the ``string`` is at the given index

    if mat_json['signal'][index:index+len(string)] == string:

        ## replace old string with new string

        mat_json['signal'] = "%s%s" % (
            mat_json['signal'][:index],
            mat_json['signal'][index+len(string):]
        )

        ## calulate offset and update annotation indices

        deleted = update_annotation_indices(
            mat_json,
            index,
            old=len(string),
        )

        if VERBOSE:
            LOGGER.info(f"{deleted} annotations deleted due to modifications.")

        ## write new file

        write_mat_json(mat_json, out_file, force=FORCE)

    else:

        LOGGER.error("OLD_STRING does not match the signal at the given INDEX. Aborting!")
        sys.exit(131)


@cli.command(context_settings=CONTEXT_SETTINGS)
@click.argument('mat_file', type=click.Path(exists=True))
@click.argument('out_file', type=click.Path())
@click.argument('index', type=int)
@click.argument('new_string')
def insert(mat_file, out_file, index, new_string):
    """Insert a NEW_STRING at the given INDEX.

    All annotation indices remaining in the signal will be updated to preserve offsets.

    \b
    Args:
      MAT_FILE    Path name to an input MAT JSON file.
      OUT_FILE    Path name to output a MAT JSON file.
      INDEX       Location in the signal (0-based index) at which to insert.
      NEW_STRING  The new value to insert into the signal at INDEX.

    \f

    Args:
        mat_file (str): path name to the MAT JSON file to edit.
        out_file (str): path name to the MAT JSON file to write out.
        index (int): the 0-based index of the string to replace.
        new_string (str): the string to insert.
    """
    import sys

    global DEBUG
    global FORCE
    global LOGGER
    global VERBOSE

    if index < 0:
        LOGGER.error("INDEX must be at least 0! Aborting!")
        sys.exit(131)

    ## load json data as python dict

    mat_json = __validate_mat_json(mat_file)

    if index >= len(mat_json['signal']):
        LOGGER.error("INDEX must be at less than the length of the signal! Aborting!")
        sys.exit(131)

    ## replace old string with new string

    mat_json['signal'] = "%s%s%s" % (
        mat_json['signal'][:index],
        new_string,
        mat_json['signal'][index:]
    )

    ## calulate offset and update annotation indices

    deleted = update_annotation_indices(
        mat_json,
        index,
        new=len(new_string),
    )

    ## write new file

    write_mat_json(mat_json, out_file, force=FORCE)


@cli.command(context_settings=CONTEXT_SETTINGS)
@click.option('-l', '--length', default=100, help="Print LENGTH characters.")
@click.argument('mat_file', type=click.Path(exists=True))
@click.argument('index', type=int)
def show(length, mat_file, index):
    """Print LENGTH characters from the MAT_FILE, starting at INDEX.

    \b
    Args:
      MAT_FILE    Path name to an input MAT JSON file.
      INDEX       Location in the signal (0-based index) at which to insert.

    \f

    Args:
        length (int): the number of characters to print to the console.
        mat_file (str): path name to the MAT JSON file to edit.
        index (int): the 0-based index of the string to replace.
    """
    import sys

    global DEBUG
    global FORCE
    global LOGGER
    global VERBOSE

    if index < 0:
        LOGGER.error("INDEX must be at least 0! Aborting!")
        sys.exit(131)

    if length < 1:
        LOGGER.error("LENGTH must be at least 1! Aborting!")
        sys.exit(131)

    ## load json data as python dict

    mat_json = __validate_mat_json(mat_file)

    if index >= len(mat_json['signal']):
        LOGGER.error("INDEX must be at less than the length of the signal! Aborting!")
        sys.exit(131)

    ## print the chars

    if VERBOSE:
        LOGGER.info(mat_json['signal'][index:index+length])
    else:
        click.echo(mat_json['signal'][index:index+length])


@cli.command(context_settings=CONTEXT_SETTINGS)
@click.argument('mat_file', type=click.Path(exists=True))
@click.argument('search_phrase')
def search(mat_file, search_phrase):
    """Search for a SEARCH_PHRASE in a MAT_FILE.

    Returns the index of the first occurrence of the phrase, if it exists.

    \b
    Args:
      MAT_FILE       Path name to an input MAT JSON file.
      SEARCH_PHRASE  The term to look for.

    \f

    Args:
        mat_file (str): path name to the MAT JSON file to edit.
        search_phrase (str): the string to search for.
    """
    import sys

    global DEBUG
    global FORCE
    global LOGGER
    global VERBOSE

    ## load json data as python dict

    mat_json = __validate_mat_json(mat_file)

    ## search

    index = mat_json['signal'].find(search_phrase)
    if index > -1:
        if VERBOSE:
            start = index - 50 if index >= 50 else 0
            LOGGER.info((
                f"Found at index {index}:\n"
                f"{mat_json['signal'][start:index]}"
                f"{click.style(search_phrase, bg='yellow', fg='bright_cyan')}"
                f"{mat_json['signal'][index+len(search_phrase):index+len(search_phrase)+50]}"
            ))
        else:
            click.echo(str(index))
    else:
        click.echo("Phrase not found.")


@cli.command(context_settings=CONTEXT_SETTINGS)
@click.argument('mat_file', type=click.Path(exists=True))
@click.argument('out_file', type=click.Path())
@click.argument('annotation_type')
def delete_by_annot(mat_file, out_file, annotation_type):
    """Delete every portion of the signal bounded by ANNOTATION_TYPE.

    All annotation indices remaining in the signal will be updated to preserve offsets.

    \b
    Args:
      MAT_FILE         Path name to an input MAT JSON file.
      OUT_FILE         Path name to output a MAT JSON file.
      ANNOTATION_TYPE  A type of annotation found within the MAT JSON file.

    \f

    Args:
        mat_file (str): path name to the MAT JSON file to edit.
        out_file (str): path name to the MAT JSON file to write out.
        annotation_type (str): a type of annotation found within the MAT JSON file.
    """
    import sys

    from lib.utils import get_aset

    global DEBUG
    global FORCE
    global LOGGER
    global VERBOSE

    ## load json data as python dict

    mat_json = __validate_mat_json(mat_file)

    ## check that the ``annotation_type`` exists in the MAT JSON file:

    aset = get_aset(mat_json, annotation_type)
    if aset:

        total_deleted_annots = 0
        total_loops = 0

        while aset['annots']:

            annot = aset['annots'][0]

            ## replace old string with new string

            mat_json['signal'] = "%s%s" % (
                mat_json['signal'][:annot[0]],
                mat_json['signal'][annot[1]:]
            )

            ## calulate offset and update annotation indices

            total_deleted_annots += update_annotation_indices(
                mat_json,
                annot[0],
                old=annot[1]-annot[0],
            )

            total_loops += 1

        if VERBOSE:
            LOGGER.info((
                f"{total_deleted_annots} annotations deleted due to modifications, {total_loops} of"
                f" which were '{annotation_type}'."
            ))

        ## write new file

        write_mat_json(mat_json, out_file, force=FORCE)

    else:
        if VERBOSE:
            LOGGER.warning(f"{annotation_type} does not exist. Writing without modification!")
        else:
            click.echo(f"{annotation_type} does not exist. Writing without modification", err=True)
        write_mat_json(mat_json, out_file, force=FORCE)


@cli.command(context_settings=CONTEXT_SETTINGS)
@click.argument('mat_file', type=click.Path(exists=True))
@click.argument('out_file', type=click.Path())
def delete_bom(mat_file, out_file):
    """Delete every byte-order mark from the signal.

    All annotation indices remaining in the signal will be updated to preserve offsets.

    \b
    Args:
      MAT_FILE         Path name to an input MAT JSON file.
      OUT_FILE         Path name to output a MAT JSON file.

    \f

    Args:
        mat_file (str): path name to the MAT JSON file to edit.
        out_file (str): path name to the MAT JSON file to write out.
    """
    from lib.utils import BOM

    import sys

    global DEBUG
    global FORCE
    global LOGGER
    global VERBOSE

    ## load json data as python dict

    mat_json = __validate_mat_json(mat_file)

    if BOM in mat_json['signal']:

        index = mat_json['signal'].find(BOM)

        ## replace old string with new string

        mat_json['signal'] = "%s%s" % (
            mat_json['signal'][:index],
            mat_json['signal'][index+len(BOM):]
        )

        ## calulate offset and update annotation indices

        deleted = update_annotation_indices(
            mat_json,
            index,
            old=len(BOM),
        )

        if VERBOSE:
            LOGGER.info(f"{deleted} annotations deleted due to modifications.")

        ## write new file

        write_mat_json(mat_json, out_file, force=FORCE)


if __name__ == '__main__':

    ## program proper

    cli()