#!/usr/bin/env python3

####################################################################################################
### -----------------------------------------  NOTICE  ----------------------------------------- ###
###                                                                                              ###
### This software (or technical data) was produced for the U.S. Government under contract        ###
### 2015-14120200002-002, and is subject to the Rights in Data-General Clause 52.227-14 , ALT    ###
### IV (MAY 2014) or (DEC 2007).                                                                 ###
###                                                                                              ###
### © 2019 The MITRE Corporation. All Rights Reserved.                                           ###
###                                                                                              ###
### -------------------------------------------------------------------------------------------- ###
####################################################################################################

"""
Shows stats for a BP or MAT document.
"""

__author__  = "Leland E. Vakarian <lsperl@mitre.org>"
__date__    = "14 April, 2021"

import click

from lib.utils import BP_ANNOTATOR_CLASS_TYPES
from lib.utils import CONTEXT_SETTINGS
from lib.utils import LOGGER
from lib.utils import MAT_DIALECT_CHOICES

DEBUG = False
FORCE = False
VERBOSE = False
"""bool: Global flags handled by cli()."""


def __detect_dupl_str(arg: list) -> str:
    seen = {}
    dupes = []
    for x in arg:
        if x not in seen:
            seen[x] = 1
        else:
            if seen[x] == 1:
                dupes.append(x)
            seen[x] += 1
    return ", ".join(dupes)


@click.group(context_settings=CONTEXT_SETTINGS)
@click.option('-v', '--verbose', count=True, help="Show verbose output. Pass again for debug.")
@click.option('-f', '--force', is_flag=True, help="Skip asking to overwrite output.")
@click.option(
    '-l',
    '--save-logs',
    'save_logs',
    type=click.Path(),
    help="Capture logging output (at whatever verbosity level) to the specified path."
)
def cli(verbose, force, save_logs):
    """Inspects BP and MAT documents.

    \f

    This function serves as the entrypoint for the program when called on the CLI. It also handles
    global flag ``verbose``.

    Args:
        verbose (int): if 1, pretty-print output as the command runs. If 2, print debugging output
            in addition to level 1 output. If 0, prints nothing as the command runs. Defaults to 0.
        force (bool): if True, skip asking for permission to overwrite a file with output. Defaults
            to False.
        save_logs (str): a path to write captured logging output to.
    """
    # do this before anything else
    from lib.utils import capture_logs
    if save_logs:
        capture_logs(save_logs, verbose, force)

    global DEBUG
    global FORCE
    global LOGGER
    global VERBOSE

    if verbose:
        LOGGER.info("Verbose output is on!")
        if verbose >= 1:
            VERBOSE = True
        if verbose > 1:
            DEBUG = True
            LOGGER.debug("Debug output is on!")

    if verbose and force:
        LOGGER.warning("Forced overwrite is on!")
    elif verbose and not force:
        LOGGER.info("Forced overwrite is off!")

    FORCE = force


@cli.command('bp-json', context_settings=CONTEXT_SETTINGS)
@click.option(
    '-v',
    '--no-validation',
    is_flag=True,
    help="Don't validate the input BP document."
)
@click.option(
    '-a',
    '--abstract',
    'abstract',
    is_flag=True,
    help="Show Abstract stats."
)
@click.option(
    '-b',
    '--basic',
    'basic',
    is_flag=True,
    help="Show Basic stats."
)
@click.option(
    '-g',
    '--granular',
    'granular',
    is_flag=True,
    help="Show Granular stats."
)
@click.option(
    '--filter-by-annotator-class',
    type=click.Choice(BP_ANNOTATOR_CLASS_TYPES),
    help=\
        "Only consider entries whose Provenance in the reference matches the given annotator class."
)
@click.option(
    '--filter-by-dataset-label',
    type=str,
    help=\
        "Only consider entries whose Provenance in the reference matches the given dataset label."
)
@click.argument('path', type=click.Path(exists=True))
def stats_cli(
    no_validation,
    abstract,
    basic,
    granular,
    filter_by_annotator_class,
    filter_by_dataset_label,
    path
):
    """Shows stats about a BP document.

    \f

    This function serves as the entrypoint for the program when called on the CLI. It also handles
    global flag ``verbose``.

    Args:
        no_validation (bool): If True, skip validating constructed BP objects. Defaults to False.
        abstract (bool): if True, count how many types of each Abstract Event type there are.
        basic (bool): if True, count how many types of each Basic Event type there are.
        granular (bool): if True, count how many types of each Granular Template type there are.
        filter_by_annotator_class (str): if passed, only consider entries whose Provenance in the
            reference matches the given annotator class.
        filter_by_dataset_label (str): Only consider entries whose Provenance in the reference
            matches the given dataset label.
        path (str): a path to an input BP document.
    """

    from lib import bp

    global LOGGER
    global VERBOSE

    ## load doc and emit stats

    bp_doc = bp.BPDocument.from_json(path, no_validation=no_validation)

    if VERBOSE:
        LOGGER.info("Summary Results:")
        for line in bp_doc.stats(
            pretty=True,
            abstract_events=abstract,
            basic_events=basic,
            granular_templates=granular,
            filter_by_annotator_class=filter_by_annotator_class,
            filter_by_dataset_label=filter_by_dataset_label
        ).split("\n"):
            LOGGER.info(line)
    else:
        click.echo(bp_doc.stats(
            abstract_events=abstract,
            basic_events=basic,
            granular_templates=granular,
            filter_by_annotator_class=filter_by_annotator_class,
            filter_by_dataset_label=filter_by_dataset_label
        ))


@cli.command('bp-circular-ref-events', context_settings=CONTEXT_SETTINGS)
@click.option("--no-validate", is_flag=True, help="Don't validate the input BP document.")
@click.argument('dialect', type=click.Choice(MAT_DIALECT_CHOICES))
@click.argument('in_path', type=click.Path(exists=True))
@click.argument('out_path', type=click.Path())
def bp_circular_ref_events_cli(no_validate, dialect, in_path, out_path):
    """Detects instances when values in ref_events refer to the parent event.

    Emits an Excel spreadsheet. Every row added to the spreadsheet indicates an Event with a
    duplicate arg _somewhere_.

    \f

    Args:
        no_validate (bool): if True, don't validate the input BP document.
        dialect (str): the task the document was created with. Must be one of
            ``utils.MAT_DIALECT_CHOICES``, which is verified by the Click decorator.
        in_path (str): a path to an input BP document.
        out_path (str): a path to output a PNG plot.
    """
    import sys

    from openpyxl import Workbook
    from openpyxl.comments import Comment

    from lib import bp
    from lib.utils import get_tqdm
    from lib.utils import user_allows_overwrite
    from lib.utils import TQDM_PREFIX as TP

    global FORCE
    global LOGGER
    global VERBOSE
    global DEBUG

    tqdm = get_tqdm(VERBOSE, DEBUG)

    if dialect not in ["Basic", "Granular",]:
        LOGGER.error("Only Basic events have the `ref_events` field. Nothing to do!")
        sys.exit(0)

    if VERBOSE:
        LOGGER.info("Loading BP doc...")
    doc = bp.BPDocument.from_json(in_path, no_validation=no_validate)

    wb = Workbook()
    ws = wb.active
    ws.title = "Circular ref_events"
    ws.freeze_panes = 'A2'

    ws['A1'].value = "EntryID"
    ws['B1'].value = "EventID"
    ws['C1'].value = "Ref Events"
    ws["C1"].comment = Comment(
        (
            "Values in this column represent the entire value of the event's ref_events property, "
            "not just the offending values in ref_events."
        ), 'bp_lib::stats'
    )
    ws["C1"].comment.width = 300
    ws["C1"].comment.height = 80
    ws.column_dimensions["C"].width = 20

    counter = 0
    if doc.entries.value:
        for entry_id, entry in tqdm(doc.entries.value.items(), desc=f"{TP} Searching Entries"):

            aset = entry.get_event_set(dialect)
            if aset:

                for event_id, event in aset.events.value.items():

                    if event.ref_events.value and event_id in event.ref_events.value:

                        ws.append([entry_id, event_id, str(event.ref_events.value)])
                        counter += 1

    ## render Excel report

    if user_allows_overwrite(out_path, VERBOSE, FORCE):

        wb.save(out_path)

    elif VERBOSE:
        LOGGER.info(f"User chose to not overwrite existing Excel file.")

    if VERBOSE:
        LOGGER.info(f"Found {counter} instances.")


@cli.command('bp-duplicate-args', context_settings=CONTEXT_SETTINGS)
@click.option("--no-validate", is_flag=True, help="Don't validate the input BP document.")
@click.argument('dialect', type=click.Choice(MAT_DIALECT_CHOICES))
@click.argument('in_path', type=click.Path(exists=True))
@click.argument('out_path', type=click.Path())
def bp_duplicate_args_cli(no_validate, dialect, in_path, out_path):
    """Detects instances of duplicate arguments in events.

    Emits an Excel spreadsheet. Every row added to the spreadsheet indicates an Event with a
    duplicate arg _somewhere_.

    \f

    Args:
        no_validate (bool): if True, don't validate the input BP document.
        dialect (str): the task the document was created with. Must be one of
            ``utils.MAT_DIALECT_CHOICES``, which is verified by the Click decorator.
        in_path (str): a path to an input BP document.
        out_path (str): a path to output a PNG plot.
    """
    from openpyxl import Workbook

    from lib import bp
    from lib.utils import get_tqdm
    from lib.utils import user_allows_overwrite
    from lib.utils import TQDM_PREFIX as TP

    global FORCE
    global LOGGER
    global VERBOSE
    global DEBUG

    tqdm = get_tqdm(VERBOSE, DEBUG)

    if VERBOSE:
        LOGGER.info("Loading BP doc...")
    doc = bp.BPDocument.from_json(in_path, no_validation=no_validate)

    wb = Workbook()
    ws = wb.active
    ws.title = "Duplicate Args"
    ws.freeze_panes = 'A2'

    ws['A1'].value = "EntryID"
    ws['B1'].value = "EventID"
    ws['C1'].value = "Duplicate Agents"
    ws.column_dimensions["C"].width = 16
    ws['D1'].value = "Duplicate Patients"
    ws.column_dimensions["D"].width = 17

    if dialect in ["Basic", "Granular",]:
        ws['E1'].value = "Duplicate Money"
        ws.column_dimensions["E"].width = 16
        ws['F1'].value = "Duplicate Ref Events"
        ws.column_dimensions["F"].width = 19

    if doc.entries.value:
        for entry_id, entry in tqdm(doc.entries.value.items(), desc=f"{TP} Searching Entries"):

            aset = entry.get_event_set(dialect)
            if aset:

                for event_id, event in aset.events.value.items():

                    row = [
                        entry_id,
                        event_id,
                        __detect_dupl_str(event.agents.value),
                        __detect_dupl_str(event.patients.value),
                    ]

                    if dialect in ["Basic", "Granular",]:
                        row.append(__detect_dupl_str(event.money.value))
                        row.append(__detect_dupl_str(event.ref_events.value)) # Ref Events
                        if row[2] or row[3] or row[4] or row[5]:
                            ws.append(row)
                    else:
                        if row[2] or row[3]:
                            ws.append(row)

    ## render Excel report

    if user_allows_overwrite(out_path, VERBOSE, FORCE):

        wb.save(out_path)

    elif VERBOSE:
        LOGGER.info(f"User chose to not overwrite existing Excel file.")


@cli.command('bp-entity-density', context_settings=CONTEXT_SETTINGS)
@click.option("--no-validate", is_flag=True, help="Don't validate the input BP document.")
@click.argument('dialect', type=click.Choice(MAT_DIALECT_CHOICES))
@click.argument('in_path', type=click.Path(exists=True))
@click.argument('out_path', type=click.Path())
def bp_entity_density_cli(no_validate, dialect, in_path, out_path):
    """Plots the density of Spans per Span Set.

    Emits a PNG of the rendered plot.

    \f

    Args:
        no_validate (bool): if True, don't validate the input BP document.
        dialect (str): the task the document was created with. Must be one of
            ``utils.MAT_DIALECT_CHOICES``, which is verified by the Click decorator.
        in_path (str): a path to an input BP document.
        out_path (str): a path to output a PNG plot.
    """

    import matplotlib
    matplotlib.use('Agg') # render to in-memory locations, without a windowing toolkit
    import matplotlib.pyplot as plt

    from lib.bp import BPDocument
    from lib.utils import get_tqdm
    from lib.utils import TQDM_PREFIX
    from lib.utils import user_allows_overwrite

    global DEBUG
    global FORCE
    global VERBOSE
    global LOGGER

    tqdm = get_tqdm(VERBOSE, DEBUG)

    doc = BPDocument.from_json(in_path, no_validation=no_validate)

    plt.rcParams["figure.figsize"] = 5,2

    y = []
    if doc.entries.value:

        iterator = tqdm(
            doc.entries.value.items(),
            desc=f"{TQDM_PREFIX} Parsing document",
            unit="entry"
        )
        for entry_id, entry in iterator:

            if dialect == "Basic" or dialect == "Granular":
                span_sets = entry.get_event_set("Basic").span_sets.value
            elif dialect == "Abstract":
                span_sets = entry.get_event_set("Abstract").span_sets.value

            for ssid, span_set in span_sets.items():
                y.append(len(span_set.spans.value))

    y = [y]
    fig, ax = plt.subplots()
    im = ax.imshow(y, cmap="plasma", aspect="auto")
    ax.set_yticks([])
    ax.set_xlabel("SpanSets")
    ax.set_title("Spans per SpanSet")
    plt.colorbar(im)
    plt.tight_layout()
    if user_allows_overwrite(out_path, VERBOSE, FORCE):
        fig.savefig(out_path, dpi=600)
    plt.close()


@cli.command('bp-dump-sentences', context_settings=CONTEXT_SETTINGS)
@click.argument('in_path', type=click.Path(exists=True))
@click.argument('out_path', type=click.Path())
def bp_dump_sentences(in_path, out_path):
    """Dumps all sentence tags in a BP document to Excel."""

    from openpyxl import Workbook

    from lib import bp
    from lib.utils import get_tqdm
    from lib.utils import user_allows_overwrite

    global FORCE
    global LOGGER
    global VERBOSE
    global DEBUG

    tqdm = get_tqdm(VERBOSE, DEBUG)

    if VERBOSE:
        LOGGER.info("Loading BP doc...")
    doc = bp.BPDocument.from_json(in_path)

    wb = Workbook()
    ws = wb.active
    ws.title = "Sentences"
    ws.freeze_panes = 'A2'

    ws['A1'].value = "EntryID"
    ws['B1'].value = "Start"
    ws['C1'].value = "Stop"
    ws['D1'].value = "Text"

    if doc.entries.value:
        for entry_id, entry in doc.entries.value.items():

            if entry.segment_sections.value:
                for segsec in entry.segment_sections.value:
                    if segsec.structural_element.value == "Sentence":

                        ws.append([
                            entry_id,
                            segsec.start.value,
                            segsec.end.value,
                            entry.segment_text.value[segsec.start.value:segsec.end.value]
                        ])

    ## render Excel report

    if user_allows_overwrite(out_path, VERBOSE, FORCE):

        wb.save(out_path)

    elif VERBOSE:
        LOGGER.info(f"User chose to not overwrite existing Excel file.")


@cli.command('bp-noarg-ratio', context_settings=CONTEXT_SETTINGS)
@click.argument('in_path', type=click.Path(exists=True))
@click.argument('out_path', type=click.Path())
def bp_noarg_ratio_cli(in_path, out_path):
    """Computes the ratio of events without arguments to all events.

    Emits an Excel spreadsheet report.

    \f

    Args:
        in_path (str): a path to an input BP document to load.
        out_path (str): a path to output an excel report.
    """

    from openpyxl import Workbook

    from lib import bp
    from lib.utils import get_tqdm
    from lib.utils import user_allows_overwrite

    global FORCE
    global LOGGER
    global VERBOSE
    global DEBUG

    tqdm = get_tqdm(VERBOSE, DEBUG)

    if VERBOSE:
        LOGGER.info("Loading BP doc...")
    doc = bp.BPDocument.from_json(in_path)

    ## count no-arg events

    abstract_results = []
    basic_results = []
    if doc.entries.value:
        for entry_id, entry in tqdm(
            doc.entries.value.items(), desc="Processing Entries", unit="entry"
        ):

            event_set = entry.get_event_set("Abstract")
            if event_set and event_set.events.value:

                total = 0
                noarg = 0
                for _, event in event_set.events.value.items():

                    total += 1
                    if not (event.agents.value or event.patients.value):
                        noarg += 1

                abstract_results.append((entry_id, noarg, total, noarg/total))

            event_set = entry.get_event_set("Basic")
            if event_set and event_set.events.value:

                total = 0
                noarg = 0
                for _, event in event_set.events.value.items():

                    total += 1
                    if not (event.agents.value or event.patients.value or event.ref_events.value):
                        noarg += 1

                basic_results.append((entry_id, noarg, total, noarg/total))

    ## render Excel report

    if user_allows_overwrite(out_path, VERBOSE, FORCE):

        wb = Workbook()
        ws = wb.active
        ws.title = "Abstract No-Arg Ratios"
        ws.freeze_panes = 'A2'

        ws['A1'].value = "EntryID"
        ws['B1'].value = "# No-Arg Events"
        ws['C1'].value = "# Events"
        ws['D1'].value = "No-Arg Ratio"

        for result in abstract_results:
            ws.append(result)

        ws = wb.create_sheet("Basic No-Arg Ratios")
        ws.freeze_panes = 'A2'

        ws['A1'].value = "EntryID"
        ws['B1'].value = "# No-Arg Events"
        ws['C1'].value = "# Events"
        ws['D1'].value = "No-Arg Ratio"

        for result in basic_results:
            ws.append(result)

        wb.save(out_path)

    elif VERBOSE:
        LOGGER.info(f"User chose to not overwrite existing Excel file.")


@cli.command('bp-sent-overlap', context_settings=CONTEXT_SETTINGS)
@click.argument('in_path', type=click.Path(exists=True))
@click.argument('out_path', type=click.Path())
def bp_sent_overlap_cli(in_path, out_path):
    """Detects overlapping annotations within BP documents.

    Emits an Excel spreadsheet report.

    \f

    Args:
        in_path (str): a path to an input BP document.
        out_path (str): a path to output an excel report.
    """

    from openpyxl import Workbook
    import sys

    from MAT.DocumentIO import getDocumentIO

    from lib import bp
    from lib.utils import detect_overlapping_bp_sents
    from lib.utils import get_tqdm
    from lib.utils import resolve_paths
    from lib.utils import user_allows_overwrite

    global FORCE
    global LOGGER
    global VERBOSE
    global DEBUG

    tqdm = get_tqdm(VERBOSE, DEBUG)

    if VERBOSE:
        LOGGER.info("Loading BP doc...")
    doc = bp.BPDocument.from_json(in_path)
    if VERBOSE:
        LOGGER.info("Checking sentence boundaries...")

    ## detect sentence overlaps

    results = detect_overlapping_bp_sents(doc, verbose=VERBOSE)

    ## render Excel report

    if user_allows_overwrite(out_path, VERBOSE, FORCE):

        wb = Workbook()
        ws = wb.active
        ws.title = "Exact Duplicates"
        ws.freeze_panes = 'A2'

        ws['A1'].value = "Entry ID"
        ws['B1'].value = "Start"
        ws['C1'].value = "Stop"
        ws['D1'].value = "# of Duplicate Tags"

        for entry_id, result in results.items():

            duplicates = result["duplicates"]

            for dupe in duplicates:
                ws.append([entry_id, dupe[0], dupe[1], len(duplicates[dupe])])

        ws = wb.create_sheet('Overlaps')
        ws.freeze_panes = 'A2'
        ws['A1'].value = "Entry ID"
        ws['B1'].value = "Sent A Start"
        ws['C1'].value = "Sent A Stop"
        ws['D1'].value = "Sent B Start"
        ws['E1'].value = "Sent B Stop"
        ws['F1'].value = "Overlap Type"

        for entry_id, result in results.items():

            overlaps = result["overlaps"]

            for segsec, segsec_, type_ in overlaps:
                ws.append([
                    entry_id,
                    segsec.start.value, segsec.end.value,
                    segsec_.start.value, segsec_.end.value,
                    type_,
                ])

        wb.save(out_path)

    elif VERBOSE:
        LOGGER.info(f"User chose to not overwrite existing Excel file.")


@cli.command('bp-span-density', context_settings=CONTEXT_SETTINGS)
@click.option("--no-validate", is_flag=True, help="Don't validate the input BP document.")
@click.argument('dialect', type=click.Choice(MAT_DIALECT_CHOICES))
@click.argument('in_path', type=click.Path(exists=True))
@click.argument('out_path', type=click.Path())
def bp_span_density_cli(no_validate, dialect, in_path, out_path):
    """Plots the density of spans per sentence.

    Emits a PNG of the rendered plot.

    \f

    Args:
        no_validate (bool): if True, don't validate the input BP document.
        dialect (str): the task the document was created with. Must be one of
            ``utils.MAT_DIALECT_CHOICES``, which is verified by the Click decorator.
        in_path (str): a path to an input BP document.
        out_path (str): a path to output a PNG plot.
    """

    import matplotlib
    matplotlib.use('Agg') # render to in-memory locations, without a windowing toolkit
    import matplotlib.pyplot as plt

    from lib.bp import BPDocument
    from lib.utils import get_tqdm
    from lib.utils import TQDM_PREFIX
    from lib.utils import user_allows_overwrite

    global DEBUG
    global FORCE
    global VERBOSE
    global LOGGER

    tqdm = get_tqdm(VERBOSE, DEBUG)

    doc = BPDocument.from_json(in_path, no_validation=no_validate)

    plt.rcParams["figure.figsize"] = 5,2

    y = []
    if doc.entries.value:

        iterator = tqdm(
            doc.entries.value.items(),
            desc=f"{TQDM_PREFIX} Parsing document",
            unit="entry"
        )
        for entry_id, entry in iterator:

            seen_spans = []

            if dialect == "Basic" or dialect == "Granular":

                # Basic and Granular have `Sentence` segment-sections.

                span_sets = entry.get_event_set("Basic").span_sets.value

                if entry.segment_sections.value:
                    for segsec in entry.segment_sections.value:
                        if segsec.structural_element.value == "Sentence":

                            spans = []
                            start = segsec.start.value
                            end = segsec.end.value

                            for ssid, span_set in span_sets.items():
                                for span in span_set.spans.value:
                                    if span.start.value >= start and span.end.value <= end:
                                        if span not in seen_spans:
                                            spans.append(span)
                                            seen_spans.append(span)

                            y.append(len(spans))

            elif dialect == "Abstract":

                # Abstract doesn't have sentence segment-sections, but instead has sentence-level
                # Entries.

                if entry.segment_type.value != "sentence":
                    continue

                span_sets = entry.get_event_set("Abstract").span_sets.value

                for ssid, span_set in span_sets.items():
                    for span in span_set.spans.value:
                        if span.start.value >= start and span.end.value <= end:
                            if span not in seen_spans:
                                seen_spans.append(span)

                y.append(len(seen_spans))

    y = [y]
    fig, ax = plt.subplots()
    im = ax.imshow(y, cmap="plasma", aspect="auto")
    ax.set_yticks([])
    ax.set_xlabel("Sentences")
    ax.set_title("Spans per Sentence")
    plt.colorbar(im)
    plt.tight_layout()
    if user_allows_overwrite(out_path, VERBOSE, FORCE):
        fig.savefig(out_path, dpi=600)
    plt.close()


@cli.command('mat-bad-annots', context_settings=CONTEXT_SETTINGS)
@click.option(
    "--allow-duplicates", is_flag=True,
    help="Don't ensure a unique set of input MAT files."
)
@click.argument('in_path', type=click.Path(exists=True), nargs=-1)
@click.argument('out_path', type=click.Path())
def mat_bad_annots_cli(allow_duplicates, in_path, out_path):
    """Detects bad annotations within MAT documents.

    Specifically, this function checks for event and argument annotations within the bounds of
    illegal segment-sections.

    Ensures that each input MAT file is unique, by their _contents_, before proceeding. Skips this
    step if `--allow-duplicates` is passed.

    Emits an Excel spreadsheet report.

    \f

    Args:
        allow_duplicates (bool): if True, skip uniqufying the list of input MAT files.
        in_path (list): a list of paths to input MAT documents or directories of MAT documents to
            load.
        out_path (str): a path to output an excel report.
    """

    from openpyxl import Workbook
    import sys

    from MAT.DocumentIO import getDocumentIO

    from lib.utils import detect_content_in_segsecs
    from lib.utils import get_tqdm
    from lib.utils import resolve_paths
    from lib.utils import TQDM_PREFIX
    from lib.utils import user_allows_overwrite

    global FORCE
    global LOGGER
    global VERBOSE
    global DEBUG

    tqdm = get_tqdm(VERBOSE, DEBUG)

    paths = resolve_paths(
        in_path,
        ".mat.json",
        no_recurse=False,
        allow_duplicates=allow_duplicates,
        verbose=VERBOSE,
        debug=DEBUG
    )

    if not paths:
        LOGGER.error("No MAT documents found in input paths!")
        sys.exit(1)

    ## detect illegal content

    results = []
    for path in tqdm(paths, desc=f"{TQDM_PREFIX} Checking MAT Docs", unit=" doc"):

        mat_io = getDocumentIO("mat-json")
        doc = mat_io.readFromSource(path)
        results.append((path, detect_content_in_segsecs(doc)))

    ## render Excel report

    if user_allows_overwrite(out_path, VERBOSE, FORCE):

        wb = Workbook()
        ws = wb.active
        ws.title = "Annots in Illegal SegSecs"
        ws.freeze_panes = 'A2'

        ws['A1'].value = "Document Path"
        ws['B1'].value = "SegSec Type"
        ws['C1'].value = "SegSec Start"
        ws['D1'].value = "Segsec Stop"
        ws['E1'].value = "Annot Type"
        ws['F1'].value = "Annot Start"
        ws['G1'].value = "Annot Stop"
        ws['H1'].value = "Type of Overlap"

        for result in results:

            path = result[0]

            for segsec in result[1]:
                for annot, type_ in result[1][segsec]:
                    ws.append([
                        path,
                        segsec.atype.lab, segsec.start, segsec.end,
                        annot.atype.lab, annot.start, annot.end,
                        type_
                    ])

        wb.save(out_path)

    elif VERBOSE:
        LOGGER.info(f"User chose to not overwrite existing Excel file.")


@cli.command('mat-event-overlap', context_settings=CONTEXT_SETTINGS)
@click.option(
    "--allow-duplicates", is_flag=True,
    help="Don't ensure a unique set of input MAT files."
)
@click.argument('in_path', type=click.Path(exists=True), nargs=-1)
@click.argument('dialect', type=click.Choice(MAT_DIALECT_CHOICES))
@click.argument('out_path', type=click.Path())
def mat_event_overlap_cli(allow_duplicates, in_path, dialect, out_path):
    """Detects overlapping Event annotations, of any type, within MAT documents.

    Ensures that each input MAT file is unique, by their _contents_, before proceeding. Skips this
    step if `--allow-duplicates` is passed.

    Emits an Excel spreadsheet report.

    \f

    Args:
        allow_duplicates (bool): if True, skip uniqufying the list of input MAT files.
        in_path (list): a list of paths to input MAT documents or directories of MAT documents to
            load.
        dialect (str): the task the document was created with. Must be one of
            ``utils.MAT_DIALECT_CHOICES``, which is verified by the Click decorator.
        out_path (str): a path to output an excel report.
    """

    import os
    import sys

    from MAT.DocumentIO import getDocumentIO
    from openpyxl import Workbook

    from lib.utils import detect_overlapping_mat_events
    from lib.utils import get_tqdm
    from lib.utils import resolve_paths
    from lib.utils import TQDM_PREFIX
    from lib.utils import user_allows_overwrite

    global FORCE
    global LOGGER
    global VERBOSE
    global DEBUG

    tqdm = get_tqdm(VERBOSE, DEBUG)

    paths = resolve_paths(
        in_path,
        ".mat.json",
        no_recurse=False,
        allow_duplicates=allow_duplicates,
        verbose=VERBOSE,
        debug=DEBUG
    )

    if not paths:
        LOGGER.error("No MAT documents found in input paths!")
        sys.exit(1)

    ## detect overlapping events

    results = {}
    for path in tqdm(paths, desc=f"{TQDM_PREFIX} Checking MAT Docs", unit=" doc"):

        mat_io = getDocumentIO("mat-json")
        doc = mat_io.readFromSource(path)
        entry_id = os.path.splitext(os.path.basename(path))[0]
        results[entry_id] = [path, detect_overlapping_mat_events(doc, dialect)]

    ## render Excel report

    if user_allows_overwrite(out_path, VERBOSE, FORCE):

        wb = Workbook()
        ws = wb.active
        ws.title = "Event Duplicates"
        ws.freeze_panes = 'A2'

        ws['A1'].value = "Document Path"
        ws['B1'].value = "Entry ID"
        ws['C1'].value = "Start"
        ws['D1'].value = "Stop"
        ws['E1'].value = "# of Duplicate Tags"

        for entry_id, result in results.items():

            path = result[0]
            duplicates = result[1]["duplicates"]

            for dupe in duplicates:
                ws.append([path, entry_id, dupe[0], dupe[1], len(duplicates[dupe]),])

        ws = wb.create_sheet('Event Overlaps')
        ws.freeze_panes = 'A2'
        ws['A1'].value = "Document Path"
        ws['B1'].value = "Entry ID"
        ws['C1'].value = "Annot A Type"
        ws['D1'].value = "Annot A Start"
        ws['E1'].value = "Annot A Stop"
        ws['F1'].value = "Annot B Type"
        ws['G1'].value = "Annot B Start"
        ws['H1'].value = "Annot B Stop"
        ws['I1'].value = "Overlap Type"

        for entry_id, result in results.items():

            path = result[0]
            overlaps = result[1]["overlaps"]

            for annot1, annot2, type_ in overlaps:
                ws.append([
                    path, entry_id,
                    annot1.atype.lab, annot1.start, annot1.end,
                    annot2.atype.lab, annot2.start, annot2.end,
                    type_
                ])

        wb.save(out_path)

    elif VERBOSE:
        LOGGER.info(f"User chose to not overwrite existing Excel file.")


@cli.command('mat-overlap', context_settings=CONTEXT_SETTINGS)
@click.option(
    "--allow-duplicates", is_flag=True,
    help="Don't ensure a unique set of input MAT files."
)
@click.argument('in_path', type=click.Path(exists=True), nargs=-1)
@click.argument('atype', type=str)
@click.argument('out_path', type=click.Path())
def mat_overlap_cli(allow_duplicates, in_path, atype, out_path):
    """Detects overlapping annotations of the same type within MAT documents.

    Ensures that each input MAT file is unique, by their _contents_, before proceeding. Skips this
    step if `--allow-duplicates` is passed.

    Emits an Excel spreadsheet report.

    \f

    Args:
        allow_duplicates (bool): if True, skip uniqufying the list of input MAT files.
        in_path (list): a list of paths to input MAT documents or directories of MAT documents to
            load.
        atype (str): an annotation type to parse from the MAT document.
        out_path (str): a path to output an excel report.
    """

    from openpyxl import Workbook
    import sys

    from MAT.DocumentIO import getDocumentIO

    from lib.utils import detect_overlapping_mat_annots
    from lib.utils import get_tqdm
    from lib.utils import resolve_paths
    from lib.utils import TQDM_PREFIX
    from lib.utils import user_allows_overwrite

    global FORCE
    global LOGGER
    global VERBOSE
    global DEBUG

    tqdm = get_tqdm(VERBOSE, DEBUG)

    paths = resolve_paths(
        in_path,
        ".mat.json",
        no_recurse=False,
        allow_duplicates=allow_duplicates,
        verbose=VERBOSE,
        debug=DEBUG
    )

    if not paths:
        LOGGER.error("No MAT documents found in input paths!")
        sys.exit(1)

    ## detect duplicates

    results = []
    for path in tqdm(paths, desc=f"{TQDM_PREFIX} Checking MAT Docs", unit=" doc"):

        mat_io = getDocumentIO("mat-json")
        doc = mat_io.readFromSource(path)
        results.append((path, detect_overlapping_mat_annots(doc, atype)))

    ## render Excel report

    if user_allows_overwrite(out_path, VERBOSE, FORCE):

        wb = Workbook()
        ws = wb.active
        ws.title = "Exact Duplicates"
        ws.freeze_panes = 'A2'

        ws['A1'].value = "Document Path"
        ws['B1'].value = "Start"
        ws['C1'].value = "Stop"
        ws['D1'].value = "# of Duplicate Tags"

        for result in results:

            path = result[0]
            duplicates = result[1]["duplicates"]

            for dupe in duplicates:
                ws.append([path, dupe[0], dupe[1], len(duplicates[dupe]),])

        ws = wb.create_sheet('Overlaps')
        ws.freeze_panes = 'A2'
        ws['A1'].value = "Document Path"
        ws['B1'].value = "Annot A Start"
        ws['C1'].value = "Annot A Stop"
        ws['D1'].value = "Annot B Start"
        ws['E1'].value = "Annot B Stop"
        ws['F1'].value = "Overlap Type"

        for result in results:

            path = result[0]
            overlaps = result[1]["overlaps"]

            for annot1, annot2, type_ in overlaps:
                ws.append([path, annot1.start, annot1.end, annot2.start, annot2.end, type_])

        wb.save(out_path)

    elif VERBOSE:
        LOGGER.info(f"User chose to not overwrite existing Excel file.")


@cli.command('mat-shared-anchors', context_settings=CONTEXT_SETTINGS)
@click.option(
    "--allow-duplicates", is_flag=True,
    help="Don't ensure a unique set of input MAT files."
)
@click.argument('in_path', type=click.Path(exists=True), nargs=-1)
@click.argument('dialect', type=click.Choice(MAT_DIALECT_CHOICES))
@click.argument('out_path', type=click.Path())
def mat_shared_anchors(allow_duplicates, in_path, dialect, out_path):
    """Detects additional-anchor annotations that are referred to by multiple Event annotations, of
    any type, within MAT documents.

    Ensures that each input MAT file is unique, by their _contents_, before proceeding. Skips this
    step if `--allow-duplicates` is passed.

    Emits an Excel spreadsheet report.

    \f

    Args:
        allow_duplicates (bool): if True, skip uniqufying the list of input MAT files.
        in_path (list): a list of paths to input MAT documents or directories of MAT documents to
            load.
        dialect (str): the task the document was created with. Must be one of
            ``utils.MAT_DIALECT_CHOICES``, which is verified by the Click decorator.
        out_path (str): a path to output an excel report.
    """

    import os
    import sys

    from MAT.DocumentIO import getDocumentIO
    from openpyxl import Workbook

    from lib.utils import detect_shared_addtl_achors
    from lib.utils import get_tqdm
    from lib.utils import resolve_paths
    from lib.utils import TQDM_PREFIX
    from lib.utils import user_allows_overwrite

    global FORCE
    global LOGGER
    global VERBOSE
    global DEBUG

    tqdm = get_tqdm(VERBOSE, DEBUG)

    paths = resolve_paths(
        in_path,
        ".mat.json",
        no_recurse=False,
        allow_duplicates=allow_duplicates,
        verbose=VERBOSE,
        debug=DEBUG
    )

    if not paths:
        LOGGER.error("No MAT documents found in input paths!")
        sys.exit(1)

    ## detect shared additional anchors

    results = {}
    for path in tqdm(paths, desc=f"{TQDM_PREFIX} Checking MAT Docs", unit=" doc"):

        mat_io = getDocumentIO("mat-json")
        doc = mat_io.readFromSource(path)
        entry_id = os.path.splitext(os.path.basename(path))[0]
        results[entry_id] = [path, detect_shared_addtl_achors(doc, dialect)]

    ## render Excel report

    if user_allows_overwrite(out_path, VERBOSE, FORCE):

        wb = Workbook()
        ws = wb.active
        ws.title = "Shared Addtl. Anchors"
        ws.freeze_panes = 'A2'

        ws['A1'].value = "Document Path"
        ws['B1'].value = "Entry ID"
        ws['C1'].value = "Addtl. Anchor"
        ws['D1'].value = "# of Referencing Events"
        ws['E1'].value = "Referencing Event Anchors"

        for entry_id, result in results.items():

            path = result[0]
            addtl_anchors = result[1]

            for addtl_anchor in addtl_anchors:

                if len(addtl_anchors[addtl_anchor]) < 2:
                    continue

                ws.append([
                    path,
                    entry_id,
                    addtl_anchor,
                    len(addtl_anchors[addtl_anchor]),
                    ", ".join([anchor for anchor in addtl_anchors[addtl_anchor]]),
                ])

        wb.save(out_path)

    elif VERBOSE:
        LOGGER.info(f"User chose to not overwrite existing Excel file.")


@cli.command('mat-unsentenced-annots', context_settings=CONTEXT_SETTINGS)
@click.option(
    "--allow-duplicates", is_flag=True,
    help="Don't ensure a unique set of input MAT files."
)
@click.argument('in_path', type=click.Path(exists=True), nargs=-1)
@click.argument('out_path', type=click.Path())
def mat_unsentenced_annots_cli(allow_duplicates, in_path, out_path):
    """Finds all annotations in a MAT document that aren't bounded by sentences but should be.

    Ensures that each input MAT file is unique, by their _contents_, before proceeding. Skips this
    step if `--allow-duplicates` is passed.

    Emits an Excel spreadsheet report.

    \f

    Args:
        allow_duplicates (bool): if True, skip uniqufying the list of input MAT files.
        in_path (list): a list of paths to input MAT documents or directories of MAT documents to
            load.
        out_path (str): a path to output an excel report.
    """
    from openpyxl import Workbook
    import sys

    from MAT.DocumentIO import getDocumentIO

    from lib.utils import detect_unsentenced_annots
    from lib.utils import get_tqdm
    from lib.utils import resolve_paths
    from lib.utils import signal
    from lib.utils import TQDM_PREFIX
    from lib.utils import user_allows_overwrite

    global FORCE
    global LOGGER
    global VERBOSE
    global DEBUG

    tqdm = get_tqdm(VERBOSE, DEBUG)

    paths = resolve_paths(
        in_path,
        ".mat.json",
        no_recurse=False,
        allow_duplicates=allow_duplicates,
        verbose=VERBOSE,
        debug=DEBUG
    )

    if not paths:
        LOGGER.error("No MAT documents found in input paths!")
        sys.exit(1)

    ## detect duplicates

    results = []
    for path in tqdm(paths, desc=f"{TQDM_PREFIX} Checking MAT Docs", unit=" doc"):

        mat_io = getDocumentIO("mat-json")
        doc = mat_io.readFromSource(path)
        results.append((path, doc, detect_unsentenced_annots(doc)))

    ## render Excel report

    if user_allows_overwrite(out_path, VERBOSE, FORCE):

        wb = Workbook()
        ws = wb.active
        ws.title = "Unsentenced Annots"
        ws.freeze_panes = 'A2'

        ws['A1'].value = "Document Path"
        ws['B1'].value = "Start"
        ws['C1'].value = "Stop"
        ws['D1'].value = "Annotation Type"
        ws['E1'].value = "Text"

        for result in results:

            path = result[0]
            doc = result[1]

            for annot in result[2]:

                ws.append([path, annot.start, annot.end, annot.atype.lab, signal(doc, annot)])

        wb.save(out_path)

    elif VERBOSE:
        LOGGER.info(f"User chose to not overwrite existing Excel file.")


@cli.command('mat-unsentenced-text', context_settings=CONTEXT_SETTINGS)
@click.option(
    "--allow-duplicates", is_flag=True,
    help="Don't ensure a unique set of input MAT files."
)
@click.argument('in_path', type=click.Path(exists=True), nargs=-1)
@click.argument('out_path', type=click.Path())
def mat_unsentenced_text_cli(allow_duplicates, in_path, out_path):
    """Finds all text in a MAT document that isn't bounded by Sentence tags or Headlines.

    Ensures that each input MAT file is unique, by their _contents_, before proceeding. Skips this
    step if `--allow-duplicates` is passed.

    Emits an Excel spreadsheet report.

    \f

    Args:
        allow_duplicates (bool): if True, skip uniqufying the list of input MAT files.
        in_path (list): a list of paths to input MAT documents or directories of MAT documents to
            load.
        out_path (str): a path to output an excel report.
    """

    from openpyxl import Workbook
    import sys

    from MAT.DocumentIO import getDocumentIO

    from lib.utils import detect_unsentenced_text
    from lib.utils import get_tqdm
    from lib.utils import resolve_paths
    from lib.utils import signal
    from lib.utils import TQDM_PREFIX
    from lib.utils import user_allows_overwrite

    global FORCE
    global LOGGER
    global VERBOSE
    global DEBUG

    tqdm = get_tqdm(VERBOSE, DEBUG)

    paths = resolve_paths(
        in_path,
        ".mat.json",
        no_recurse=False,
        allow_duplicates=allow_duplicates,
        verbose=VERBOSE,
        debug=DEBUG
    )

    if not paths:
        LOGGER.error("No MAT documents found in input paths!")
        sys.exit(1)

    ## detect duplicates

    results = []
    for path in tqdm(paths, desc=f"{TQDM_PREFIX} Checking MAT Docs", unit=" doc"):

        mat_io = getDocumentIO("mat-json")
        doc = mat_io.readFromSource(path)
        results.append((path, doc, detect_unsentenced_text(doc)))

    ## render Excel report

    if user_allows_overwrite(out_path, VERBOSE, FORCE):

        wb = Workbook()
        ws = wb.active
        ws.title = "Unsentenced Text"
        ws.freeze_panes = 'A2'

        ws['A1'].value = "Document Path"
        ws['B1'].value = "Start"
        ws['C1'].value = "Stop"
        ws['D1'].value = "Text"

        for result in results:

            path = result[0]
            doc = result[1]

            for annot in result[2]:

                ws.append([path, annot.start, annot.end, signal(doc, annot)])

        wb.save(out_path)

    elif VERBOSE:
        LOGGER.info(f"User chose to not overwrite existing Excel file.")


@cli.command('score-analysis', context_settings=CONTEXT_SETTINGS)
@click.option('--csv', 'csv_input', is_flag=True, help="Accept CSV input instead of TSV.")
@click.argument('in_path', type=click.Path(exists=True))
@click.argument('out_path', type=click.Path())
def score_analysis(csv_input, in_path, out_path):
    """Consolidates TSV scoring information post-evaluation into a single Excel workbook.

    The data should have been generated with ``--extras`` passed to the scorer. It also expects
    file names to have one of the following two formats:

        {team_label}.formal.{run_label}.{system_name}.tsv

    OR

        {team_label}.{execution_label}.{system_name}.{run_label}.tsv

    \f

    Args:
        csv_input (bool): accept CSV files as input instead of TSV files.
        in_path (str): path to a directory of TSV files, each representing a run to include.
        out_path (str): path to an output file.
    """
    import csv
    import os
    import sys

    import openpyxl

    from lib.utils import get_tqdm
    from lib.utils import resolve_paths
    from lib.utils import user_allows_overwrite
    from lib.utils import TQDM_PREFIX

    global FORCE
    global LOGGER
    global VERBOSE
    global DEBUG

    tqdm = get_tqdm(VERBOSE, DEBUG)

    paths = resolve_paths(
        [in_path,],
        ".csv" if csv_input else ".tsv",
        no_recurse=True,
        allow_duplicates=True,
        verbose=VERBOSE,
        debug=DEBUG
    )

    if DEBUG:
        LOGGER.debug("Resolved paths:")
        for path in paths:
            LOGGER.debug(f"    {path}")

    if not paths:
        if VERBOSE:
            LOGGER.error("No files found!")
        sys.exit()

    headers = []
    data = []

    def __dtype_format(v):
        try:
            v = float(v)
        except ValueError:
            try:
                v = int(v)
            except ValueError:
                pass
        return v

    for path in tqdm(paths, desc=f"{TQDM_PREFIX} Parsing", unit=" csv" if csv_input else " tsv"):

        basename = os.path.basename(path)

        with open(path, 'r',  newline='') as f:

            reader = csv.reader(f, delimiter=',' if csv_input else '\t')

            last_row = None
            first_row = None
            for row in reader:
                if not first_row:
                    first_row = row
                last_row = row

            if not headers:
                headers = first_row
                if DEBUG:
                    LOGGER.debug(f"Found headers: {headers}")
            else:
                if headers != first_row:
                    if VERBOSE:
                        LOGGER.warning(f"Skipping {basename} with mismatched headers.")
                        LOGGER.warning(f"    {first_row}")
                    continue

            last_row.insert(0, basename)
            data.append(last_row)

    headers.insert(0, "system")

    if DEBUG:
        LOGGER.debug(f"Found systems: {[row[0] for row in data]}")

    wb = openpyxl.Workbook()
    ws = wb.active
    ws.title = "Combined Scores"
    ws.freeze_panes = 'B2'

    ws.append(headers)
    for row in data:
        ws.append([__dtype_format(cell) for cell in row])

    if user_allows_overwrite(out_path, verbose=VERBOSE, force=FORCE):
        wb.save(out_path)


@cli.command('inspect-bp', context_settings=CONTEXT_SETTINGS)
@click.option("--no-validate", is_flag=True, help="Don't validate the input BP document.")
@click.argument('dialect', type=click.Choice(MAT_DIALECT_CHOICES))
@click.argument('in_path', type=click.Path(exists=True))
@click.argument('out_path', type=click.Path())
def inspect_bp(no_validate, dialect, in_path, out_path):
    """Runs the standard gauntlet of inspection functions on a single BP file.

    \b
    Inspections:
    - Detect circular ref_events in Basic Events
    - Detect duplicate values in Argument lists in Abstract and Basic Events
    - Detect overlapping Sentence segment-sections

    \f

    Args:
        no_validate (bool): if True, don't validate the input BP document.
        dialect (str): the task the document was created with. Must be one of
            ``utils.MAT_DIALECT_CHOICES``, which is verified by the Click decorator.
        in_path (str): a path to an input BP document.
        out_path (str): a path to output Excel report with the results.
    """
    from openpyxl import Workbook
    from openpyxl.comments import Comment

    from lib import bp
    from lib.utils import detect_overlapping_bp_sents
    from lib.utils import get_tqdm
    from lib.utils import user_allows_overwrite
    from lib.utils import TQDM_PREFIX as TP

    global FORCE
    global LOGGER
    global VERBOSE
    global DEBUG

    ## spool

    tqdm = get_tqdm(VERBOSE, DEBUG)

    if VERBOSE:
        LOGGER.info("Loading BP doc...")
    doc = bp.BPDocument.from_json(in_path, no_validation=no_validate)

    if not user_allows_overwrite(out_path, VERBOSE, FORCE):
        LOGGER.info(f"User chose to not overwrite existing Excel file.")
        sys.exit(0)

    wb = Workbook()

    ## detect circular ref_events

    if dialect in ["Basic", "Granular",]:

        ws = wb.active
        ws.title = "Circular ref_events"
        ws.freeze_panes = 'A2'

        ws['A1'].value = "EntryID"
        ws['B1'].value = "EventID"
        ws['C1'].value = "Ref Events"
        ws["C1"].comment = Comment(
            (
                "Values in this column represent the entire value of the event's ref_events "
                "property, not just the offending values in ref_events."
            ), 'bp_lib::stats'
        )
        ws["C1"].comment.width = 300
        ws["C1"].comment.height = 80
        ws.column_dimensions["C"].width = 20

        if doc.entries.value:
            for entry_id, entry in tqdm(
                doc.entries.value.items(), desc=f"{TP} Detecting circular ref_events"
            ):

                aset = entry.get_event_set(dialect)
                if aset:

                    for event_id, event in aset.events.value.items():

                        if event.ref_events.value and event_id in event.ref_events.value:

                            ws.append([entry_id, event_id, str(event.ref_events.value)])

    ## detect duplicate args

    ws = wb.create_sheet("Duplicate Args")
    ws.freeze_panes = 'A2'

    ws['A1'].value = "EntryID"
    ws['B1'].value = "EventID"
    ws['C1'].value = "Duplicate Agents"
    ws.column_dimensions["C"].width = 16
    ws['D1'].value = "Duplicate Patients"
    ws.column_dimensions["D"].width = 17

    if dialect in ["Basic", "Granular",]:
        ws['E1'].value = "Duplicate Money"
        ws.column_dimensions["E"].width = 16
        ws['F1'].value = "Duplicate Ref Events"
        ws.column_dimensions["F"].width = 19

    if doc.entries.value:
        for entry_id, entry in tqdm(
            doc.entries.value.items(), desc=f"{TP} Detecting duplicate args"
        ):

            aset = entry.get_event_set(dialect)
            if aset:

                for event_id, event in aset.events.value.items():

                    row = [
                        entry_id,
                        event_id,
                        __detect_dupl_str(event.agents.value),
                        __detect_dupl_str(event.patients.value),
                    ]

                    if dialect in ["Basic", "Granular",]:
                        row.append(__detect_dupl_str(event.money.value))
                        row.append(__detect_dupl_str(event.ref_events.value)) # Ref Events
                        if row[2] or row[3] or row[4] or row[5]:
                            ws.append(row)
                    else:
                        if row[2] or row[3]:
                            ws.append(row)

    ## detect overlapping sentence segsecs

    results = detect_overlapping_bp_sents(doc, verbose=VERBOSE)

    ws = wb.create_sheet("Exact Sentence Duplicates")
    ws.freeze_panes = 'A2'

    ws['A1'].value = "Entry ID"
    ws['B1'].value = "Start"
    ws['C1'].value = "Stop"
    ws['D1'].value = "# of Duplicate Tags"

    for entry_id, result in results.items():

        duplicates = result["duplicates"]

        for dupe in duplicates:
            ws.append([entry_id, dupe[0], dupe[1], len(duplicates[dupe])])

    ws = wb.create_sheet('Sentence Overlaps')
    ws.freeze_panes = 'A2'
    ws['A1'].value = "Entry ID"
    ws['B1'].value = "Sent A Start"
    ws['C1'].value = "Sent A Stop"
    ws['D1'].value = "Sent B Start"
    ws['E1'].value = "Sent B Stop"
    ws['F1'].value = "Overlap Type"

    for entry_id, result in results.items():

        overlaps = result["overlaps"]

        for segsec, segsec_, type_ in overlaps:
            ws.append([
                entry_id,
                segsec.start.value, segsec.end.value,
                segsec_.start.value, segsec_.end.value,
                type_,
            ])

    ## write

    wb.save(out_path)
    LOGGER.info(f"Wrote workbook to '{out_path}'.")


@cli.command('inspect-mat', context_settings=CONTEXT_SETTINGS)
@click.option(
    "--allow-duplicates", is_flag=True,
    help="Don't ensure a unique set of input MAT files."
)
@click.argument('dialect', type=click.Choice(MAT_DIALECT_CHOICES))
@click.argument('in_path', type=click.Path(exists=True), nargs=-1)
@click.argument('out_path', type=click.Path())
def inspect_mat(allow_duplicates, dialect, in_path, out_path):
    """Runs the standard gauntlet of inspection functions on a collection of MAT files.

    \b
    Inspections:
    - Detect Event and Argument annotations within the bounds of illegal segment-sections
    - Detect overlapping Sentence tags
    - Detect Mention, Event, and Template annotations not completely within Sentence or Headline
      tags
    - Detect signal text not within Sentence or Headline tags
    - Detect event overlap across all event types (not just a single event type)

    \f

    Args:
        no_validate (bool): if True, don't validate the input BP document.
        dialect (str): the task the document was created with. Must be one of
            ``utils.MAT_DIALECT_CHOICES``, which is verified by the Click decorator.
        in_path (str): a path to an input BP document.
        out_path (str): a path to output Excel report with the results.
    """
    import os
    import sys

    from MAT.DocumentIO import getDocumentIO
    from openpyxl import Workbook

    from lib.utils import detect_content_in_segsecs
    from lib.utils import detect_overlapping_mat_annots
    from lib.utils import detect_overlapping_mat_events
    from lib.utils import detect_shared_addtl_achors
    from lib.utils import detect_unsentenced_annots
    from lib.utils import detect_unsentenced_text
    from lib.utils import get_tqdm
    from lib.utils import resolve_paths
    from lib.utils import signal
    from lib.utils import TQDM_PREFIX
    from lib.utils import user_allows_overwrite

    global FORCE
    global LOGGER
    global VERBOSE
    global DEBUG

    ## spool

    tqdm = get_tqdm(VERBOSE, DEBUG)

    paths = resolve_paths(
        in_path,
        ".mat.json",
        no_recurse=False,
        allow_duplicates=allow_duplicates,
        verbose=VERBOSE,
        debug=DEBUG
    )

    if not paths:
        LOGGER.error("No MAT documents found in input paths!")
        sys.exit(1)

    if not user_allows_overwrite(out_path, VERBOSE, FORCE):
        LOGGER.info(f"User chose to not overwrite existing Excel file.")
        sys.exit(0)

    wb = Workbook()

    ## detect illegal annots

    results = []
    for path in tqdm(paths, desc=f"{TQDM_PREFIX} Detecting illegal annots", unit=" doc"):

        mat_io = getDocumentIO("mat-json")
        doc = mat_io.readFromSource(path)
        results.append((path, detect_content_in_segsecs(doc)))

    ws = wb.active
    ws.title = "Annots in Illegal SegSecs"
    ws.freeze_panes = 'A2'

    ws['A1'].value = "Document Path"
    ws['B1'].value = "SegSec Type"
    ws['C1'].value = "SegSec Start"
    ws['D1'].value = "Segsec Stop"
    ws['E1'].value = "Annot Type"
    ws['F1'].value = "Annot Start"
    ws['G1'].value = "Annot Stop"
    ws['H1'].value = "Type of Overlap"

    for result in results:

        path = result[0]

        for segsec in result[1]:
            for annot, type_ in result[1][segsec]:
                ws.append([
                    path,
                    segsec.atype.lab, segsec.start, segsec.end,
                    annot.atype.lab, annot.start, annot.end,
                    type_
                ])

    ## detect overlapping sentences

    results = []
    for path in tqdm(paths, desc=f"{TQDM_PREFIX} Detecting sentence overlap", unit=" doc"):

        mat_io = getDocumentIO("mat-json")
        doc = mat_io.readFromSource(path)
        results.append((path, detect_overlapping_mat_annots(doc, "S")))

    ws = wb.create_sheet("Exact Sentence Duplicates")
    ws.freeze_panes = 'A2'

    ws['A1'].value = "Document Path"
    ws['B1'].value = "Start"
    ws['C1'].value = "Stop"
    ws['D1'].value = "# of Duplicate Tags"

    for result in results:

        path = result[0]
        duplicates = result[1]["duplicates"]

        for dupe in duplicates:
            ws.append([path, dupe[0], dupe[1], len(duplicates[dupe]),])

    ws = wb.create_sheet('Sentence Overlaps')
    ws.freeze_panes = 'A2'

    ws['A1'].value = "Document Path"
    ws['B1'].value = "Annot A Start"
    ws['C1'].value = "Annot A Stop"
    ws['D1'].value = "Annot B Start"
    ws['E1'].value = "Annot B Stop"
    ws['F1'].value = "Overlap Type"

    for result in results:

        path = result[0]
        overlaps = result[1]["overlaps"]

        for annot1, annot2, type_ in overlaps:
            ws.append([path, annot1.start, annot1.end, annot2.start, annot2.end, type_])

    ## detect annots outside of sentences

    results = []
    for path in tqdm(paths, desc=f"{TQDM_PREFIX} Detecting unsentenced annots", unit=" doc"):

        mat_io = getDocumentIO("mat-json")
        doc = mat_io.readFromSource(path)
        results.append((path, doc, detect_unsentenced_annots(doc)))

    ws = wb.create_sheet("Unsentenced Annots")
    ws.freeze_panes = 'A2'

    ws['A1'].value = "Document Path"
    ws['B1'].value = "Start"
    ws['C1'].value = "Stop"
    ws['D1'].value = "Annotation Type"
    ws['E1'].value = "Text"

    for result in results:

        path = result[0]
        doc = result[1]

        for annot in result[2]:

            ws.append([path, annot.start, annot.end, annot.atype.lab, signal(doc, annot)])

    ## detect text outside of sentences

    results = []
    for path in tqdm(paths, desc=f"{TQDM_PREFIX} Detecting unsentenced text", unit=" doc"):

        mat_io = getDocumentIO("mat-json")
        doc = mat_io.readFromSource(path)
        results.append((path, doc, detect_unsentenced_text(doc)))

    ws = wb.create_sheet("Unsentenced Text")
    ws.freeze_panes = 'A2'

    ws['A1'].value = "Document Path"
    ws['B1'].value = "Start"
    ws['C1'].value = "Stop"
    ws['D1'].value = "Text"

    for result in results:

        path = result[0]
        doc = result[1]

        for annot in result[2]:

            ws.append([path, annot.start, annot.end, signal(doc, annot)])

    ## detect overlapping event annotations (of any type)

    results = {}
    for path in tqdm(paths, desc=f"{TQDM_PREFIX} Detecting overlapping Events", unit=" doc"):

        mat_io = getDocumentIO("mat-json")
        doc = mat_io.readFromSource(path)
        entry_id = os.path.splitext(os.path.basename(path))[0]
        results[entry_id] = [path, detect_overlapping_mat_events(doc, dialect)]

    ws = wb.create_sheet("Event Duplicates")
    ws.freeze_panes = 'A2'

    ws['A1'].value = "Document Path"
    ws['B1'].value = "Entry ID"
    ws['C1'].value = "Start"
    ws['D1'].value = "Stop"
    ws['E1'].value = "# of Duplicate Tags"

    for entry_id, result in results.items():

        path = result[0]
        duplicates = result[1]["duplicates"]

        for dupe in duplicates:
            ws.append([path, entry_id, dupe[0], dupe[1], len(duplicates[dupe]),])

    ws = wb.create_sheet('Event Overlaps')
    ws.freeze_panes = 'A2'
    ws['A1'].value = "Document Path"
    ws['B1'].value = "Entry ID"
    ws['C1'].value = "Annot A Type"
    ws['D1'].value = "Annot A Start"
    ws['E1'].value = "Annot A Stop"
    ws['F1'].value = "Annot B Type"
    ws['G1'].value = "Annot B Start"
    ws['H1'].value = "Annot B Stop"
    ws['I1'].value = "Overlap Type"

    for entry_id, result in results.items():

        path = result[0]
        overlaps = result[1]["overlaps"]

        for annot1, annot2, type_ in overlaps:
            ws.append([
                path, entry_id,
                annot1.atype.lab, annot1.start, annot1.end,
                annot2.atype.lab, annot2.start, annot2.end,
                type_
            ])

    ## detect shared additional anchors

    results = {}
    for path in tqdm(paths, desc=f"{TQDM_PREFIX} Detecting shared additional anchors", unit=" doc"):

        mat_io = getDocumentIO("mat-json")
        doc = mat_io.readFromSource(path)
        entry_id = os.path.splitext(os.path.basename(path))[0]
        results[entry_id] = [path, detect_shared_addtl_achors(doc, dialect)]

    ws = wb.create_sheet("Shared Addtl. Anchors")
    ws.freeze_panes = 'A2'

    ws['A1'].value = "Document Path"
    ws['B1'].value = "Entry ID"
    ws['C1'].value = "Addtl. Anchor"
    ws['D1'].value = "# of Referencing Events"
    ws['E1'].value = "Referencing Event Anchors"

    for entry_id, result in results.items():

        path = result[0]
        addtl_anchors = result[1]

        for addtl_anchor in addtl_anchors:

            if len(addtl_anchors[addtl_anchor]) < 2:
                continue

            ws.append([
                path,
                entry_id,
                addtl_anchor,
                len(addtl_anchors[addtl_anchor]),
                ", ".join([anchor for anchor in addtl_anchors[addtl_anchor]]),
            ])

    ## write

    wb.save(out_path)
    LOGGER.info(f"Wrote workbook to '{out_path}'.")


if __name__ == '__main__':

    ## program proper

    cli()
