####################################################################################################
### -----------------------------------------  NOTICE  ----------------------------------------- ###
###                                                                                              ###
### This software (or technical data) was produced for the U.S. Government under contract        ###
### 2015-14120200002-002, and is subject to the Rights in Data-General Clause 52.227-14 , ALT    ###
### IV (MAY 2014) or (DEC 2007).                                                                 ###
###                                                                                              ###
### © 2019 The MITRE Corporation. All Rights Reserved.                                           ###
###                                                                                              ###
### -------------------------------------------------------------------------------------------- ###
####################################################################################################

"""
Test functions for the ``lib/bp.py`` library. To use, call `pyest` from the root of the project.

Every function must start with ``test_``, and every class must start with ``Test``. For
convenvience, testing classes will test the correspinding class in the BP library.

The ``BPProperty`` class will be tested on validation and comparison methods. Each BPDocument class
will be tested on validation and dictionary conversion, where applicable; the ``is_valid()``
function of each BPDocumnet class is sufficient for simple "assert each property is valid" routines.
For cases with more complex logic, as is the case with ``BPDocumentSpan.is_valid()``, cases will be
written to ensure ``is_valid()`` is performing correctly.

Since all of the BPDocument classes in the BP library test equivalence _exactly the same way_, and
all of them rely on ``BPProperty`` for the comparison testing, I've written test cases for the
``BPProperty`` comparison methods but not for the comparison methods of the other classes.
"""

__author__  = "Leland E. Vakarian <lsperl@mitre.org>"
__date__    = "29 June 2020"

import pytest

from lib import bp


class TestBPProperty(object):


    ## validation


    def test_min_args(self):
        """Minimally, a BPProperty is valid with a name and default args."""
        prop = bp.BPProperty("test_min_args")
        assert prop.is_valid(), repr(prop)


    def test_null_when_required(self):
        """A BPProperty is invalid when ``value`` isn't set and ``required`` is True."""
        prop = bp.BPProperty("test_null_when_required", required=True)
        with pytest.raises(AssertionError):
            assert prop.is_valid()


    def test_val_when_required_without_type(self):
        """A BPProperty is invalid when ``value`` is set and ``type_`` is not. This also tests type
        mismatch between ``value`` and ``type_`` since ``type_`` defaults to ``NoneType``.
        """
        prop = bp.BPProperty(
            "test_val_when_required_without_type",
            value="test me"
        )
        with pytest.raises(AssertionError):
            assert prop.is_valid()


    def test_val_when_required_with_type(self):
        """A BPProperty is valid when both ``value`` and ``type_`` are set and the types match."""
        prop = bp.BPProperty(
            "test_val_when_required_with_type",
            value="test me",
            type_=str
        )
        assert prop.is_valid()


    def test_default_when_required_without_type(self):
        """A BPProperty is invalid when ``default`` is set and ``type_`` is not. This also tests
        type mismatch between ``default`` and ``type_`` since ``type_`` defaults to ``NoneType``.
        """
        prop = bp.BPProperty(
            "test_default_when_required_without_type",
            default="test me"
        )
        with pytest.raises(AssertionError):
            assert prop.is_valid()


    def test_default_when_required_with_type(self):
        """A BPProperty is valid when both ``default`` and ``type_`` are set and the types match."""
        prop = bp.BPProperty(
            "test_default_when_required_with_type",
            default="test me",
            type_=str
        )
        assert prop.is_valid()


    def test_correct_choices(self):
        """A BPProperty is valid when the types of the elements in choices match. They must
        all match the value of ``type_``, which defaults to ``NoneType``.
        """
        prop = bp.BPProperty(
            "test_correct_choices",
            type_=int,
            choices=[0, 1, 2, 3,]
        )


    def test_incorrect_choices(self):
        """A BPProperty is invalid when the types of the elements in choices mismatch. They
        must all match the value of ``type_``, which defaults to ``NoneType``.
        """
        with pytest.raises(AssertionError):
            prop = bp.BPProperty(
                "test_incorrect_choices",
                choices=[1, "chicken", None,]
            )


    ## comparison testing


    def test_expecting_eq(self):
        """For a BPProperty to equal another, all corresponding constructor values must evaluate to
        be equal to one another.
        """
        a = bp.BPProperty(
            "prop",
            required=True,
            value="chicken",
            type_=str,
            default="chicken",
            choices=["chicken", "chips", "ice cream",]
        )
        assert a.is_valid()

        b = bp.BPProperty(
            "prop",
            required=True,
            value="chicken",
            type_=str,
            default="chicken",
            choices=["chicken", "chips", "ice cream",]
        )
        assert b.is_valid()

        assert a == b


    def test_expecting_ne_name(self):
        """For a BPProperty to equal another, all corresponding constructor values must evaluate to
        be equal to one another.
        """
        a = bp.BPProperty(
            "some name",
            required=True,
            value="chicken",
            type_=str,
            default="chicken",
            choices=["chicken", "chips", "ice cream",]
        )
        assert a.is_valid()

        b = bp.BPProperty(
            "some other name",
            required=True,
            value="chicken",
            type_=str,
            default="chicken",
            choices=["chicken", "chips", "ice cream",]
        )
        assert b.is_valid()

        assert a != b


    def test_expecting_ne_required(self):
        """For a BPProperty to equal another, all corresponding constructor values must evaluate to
        be equal to one another.
        """
        a = bp.BPProperty(
            "some name",
            required=True,
            value="chicken",
            type_=str,
            default="chicken",
            choices=["chicken", "chips", "ice cream",]
        )
        assert a.is_valid()

        b = bp.BPProperty(
            "some name",
            required=False,
            value="chicken",
            type_=str,
            default="chicken",
            choices=["chicken", "chips", "ice cream",]
        )
        assert b.is_valid()

        assert a != b


    def test_expecting_ne_value(self):
        """For a BPProperty to equal another, all corresponding constructor values must evaluate to
        be equal to one another.
        """
        a = bp.BPProperty(
            "some name",
            required=True,
            value="chicken",
            type_=str,
            default="chicken",
            choices=["chicken", "chips", "ice cream",]
        )
        assert a.is_valid()

        b = bp.BPProperty(
            "some name",
            required=True,
            value="chips",
            type_=str,
            default="chicken",
            choices=["chicken", "chips", "ice cream",]
        )
        assert b.is_valid()

        assert a != b


    def test_expecting_ne_type(self):
        """For a BPProperty to equal another, all corresponding constructor values must evaluate to
        be equal to one another.
        """
        a = bp.BPProperty(
            "some name",
            value="",
            type_=str
        )
        assert a.is_valid()

        b = bp.BPProperty(
            "some name",
            value=0,
            type_=int
        )
        assert b.is_valid()

        assert a != b


    def test_expecting_ne_default(self):
        """For a BPProperty to equal another, all corresponding constructor values must evaluate to
        be equal to one another.
        """
        a = bp.BPProperty(
            "some name",
            required=True,
            value="chicken",
            type_=str,
            default="chicken",
            choices=["chicken", "chips", "ice cream",]
        )
        assert a.is_valid()

        b = bp.BPProperty(
            "some name",
            required=True,
            value="chicken",
            type_=str,
            default="chips",
            choices=["chicken", "chips", "ice cream",]
        )
        assert b.is_valid()

        assert a != b


    def test_expecting_ne_choices(self):
        """For a BPProperty to equal another, all corresponding constructor values must evaluate to
        be equal to one another.
        """
        a = bp.BPProperty(
            "some name",
            value="chicken",
            type_=str,
            choices=["chicken", "chips", "donuts",]
        )
        assert a.is_valid()

        b = bp.BPProperty(
            "some name",
            value="chicken",
            type_=str,
            choices=["chicken", "chips", "ice cream",]
        )
        assert b.is_valid()

        assert a != b


class TestBPDocumentSpan(object):


    ## validation


    def valid_span(self):
        """Null hypothesis."""
        span = bp.BPDocumentSpan(
            "some string",
            start=0, end=11,
            hstring="some", hstart=0, hend=4,
            synclass="name"
        )
        with pytest.raises(AssertionError):
            assert not span.is_valid(), "with Head info and synclass"

        span = bp.BPDocumentSpan(
            "some string",
            start=0, end=11,
            no_head=True,
            synclass="nominal"
        )
        with pytest.raises(AssertionError):
            assert not span.is_valid(), "with No-Head info and synclass"

        span = bp.BPDocumentSpan(
            "some string",
            start=0, end=11,
            synclass="pronoun",
            is_anchor=True
        )
        with pytest.raises(AssertionError):
            assert not span.is_valid(), "without Head info, with is_anchor and synclass"


    def test_rogue_hstring_within_string(self):
        """A BPDocumentSpan is invalid when ``hstring`` cannot be found within ``string_``.
        """

        span = bp.BPDocumentSpan("some string", hstring="foo bar")
        with pytest.raises(AssertionError):
            assert span.is_valid()


    def test_no_head_with_hstring(self):
        """A BPDocumentSpan is invalid when ``no_head`` is True and any of ``hstring``, ``hstart``,
        or ``hend`` are specified.
        """

        span = bp.BPDocumentSpan("some string", no_head=True, hstring="string")
        with pytest.raises(AssertionError):
            assert span.is_valid()


    def test_no_head_with_hstart(self):
        """A BPDocumentSpan is invalid when ``no_head`` is True and any of ``hstring``, ``hstart``,
        or ``hend`` are specified.
        """

        span = bp.BPDocumentSpan("some string", no_head=True, hstart=0)
        with pytest.raises(AssertionError):
            assert span.is_valid()


    def test_no_head_with_hend(self):
        """A BPDocumentSpan is invalid when ``no_head`` is True and any of ``hstring``, ``hstart``,
        or ``hend`` are specified.
        """

        span = bp.BPDocumentSpan("some string", no_head=True, hend=5)
        with pytest.raises(AssertionError):
            assert span.is_valid()


    def test_span_trim(self):
        """A BPDocumentSpan should safely trim leading and trailing whitespace and punctuation
        symbols.
        """

        from lib.utils import LOGGER

        span = bp.BPDocumentSpan(
            ', Clifford the (Big Red) dog!', start=20, end=49,
            hstring=' (Big Red)', hstart=34, hend=44,
            no_head=False, synclass='name',
            is_anchor=False, hinferred=False,
            no_validation=False
        )

        expected = bp.BPDocumentSpan(
            'Clifford the (Big Red) dog', start=22, end=48,
            hstring='Big Red', hstart=36, hend=43,
            no_head=False, synclass='name',
            is_anchor=False, hinferred=False,
            no_validation=False
        )

        span.trim()

        assert span == expected, f"Unexpected trim! -> {expected.diff(span)}"


    def test_span_trim_currency(self):
        """A BPDocumentSpan should not trim leading or trailing currency symbols."""

        from copy import deepcopy

        from lib.utils import LOGGER

        span = bp.BPDocumentSpan(
            '$99.99', start=0, end=6,
            no_head=True, synclass='nominal',
        )
        expected = deepcopy(span)

        span.trim()

        assert span == expected, f"Unexpected trim! -> {expected.diff(span)}"


    def test_span_trim_percents(self):
        """A BPDocumentSpan should not trim trailing percent signs."""

        from copy import deepcopy

        from lib.utils import LOGGER

        span = bp.BPDocumentSpan(
            '50%', start=0, end=3,
            no_head=True, synclass='nominal',
        )
        expected = deepcopy(span)

        span.trim()

        assert span == expected, f"Unexpected trim! -> {expected.diff(span)}"


class TestBPDocumentAbstractEventSet(object):


    ## validation


    def test_anchor_ssid_not_in_span_sets(self):
        """A BPDocumentAbstractEventSet is invalid when an SSID in the ``anchors`` field cannot be
        found within the span_sets for that Abstract Event.
        """

        event_set = bp.BPDocumentAbstractEventSet(
            {
                "ev-1": bp.BPDocumentAbstractEvent(
                    "ev-1",
                    "ss-3",
                    [],
                    [],
                    "neutral",
                    "unk"
                ), 
            },
            {
                "ss-1": bp.BPDocumentSpanSet(
                    "ss-1",
                    [
                        bp.BPDocumentSpan("some string"),
                        bp.BPDocumentSpan("some other string"),
                    ]
                ),
                "ss-2": bp.BPDocumentSpanSet(
                    "ss-2",
                    [
                        bp.BPDocumentSpan("how now brown cow"),
                        bp.BPDocumentSpan("blarg honk"),
                    ]
                ),
            }
        )

        with pytest.raises(AssertionError):
            assert event_set.is_valid()


    def test_agent_ssid_not_in_span_sets(self):
        """A BPDocumentAbstractEventSet is invalid when an SSID in the ``agents`` field cannot be
        found within the span_sets for that Abstract Event.
        """

        event_set = bp.BPDocumentAbstractEventSet(
            {
                "ev-1": bp.BPDocumentAbstractEvent(
                    "ev-1",
                    "ss-1",
                    ["ss-3"],
                    [],
                    "neutral",
                    "unk"
                ), 
            },
            {
                "ss-1": bp.BPDocumentSpanSet(
                    "ss-1",
                    [
                        bp.BPDocumentSpan("some string"),
                        bp.BPDocumentSpan("some other string"),
                    ]
                ),
                "ss-2": bp.BPDocumentSpanSet(
                    "ss-2",
                    [
                        bp.BPDocumentSpan("how now brown cow"),
                        bp.BPDocumentSpan("blarg honk"),
                    ]
                ),
            }
        )

        with pytest.raises(AssertionError):
            assert event_set.is_valid()


    def test_patient_ssid_not_in_span_sets(self):
        """A BPDocumentAbstractEventSet is invalid when an SSID in the ``patients`` field cannot be
        found within the span_sets for that Abstract Event.
        """

        event_set = bp.BPDocumentAbstractEventSet(
            {
                "ev-1": bp.BPDocumentAbstractEvent(
                    "ev-1",
                    "ss-1",
                    [],
                    ["ss-3"],
                    "neutral",
                    "unk"
                ), 
            },
            {
                "ss-1": bp.BPDocumentSpanSet(
                    "ss-1",
                    [
                        bp.BPDocumentSpan("some string"),
                        bp.BPDocumentSpan("some other string"),
                    ]
                ),
                "ss-2": bp.BPDocumentSpanSet(
                    "ss-2",
                    [
                        bp.BPDocumentSpan("how now brown cow"),
                        bp.BPDocumentSpan("blarg honk"),
                    ]
                ),
            }
        )

        with pytest.raises(AssertionError):
            assert event_set.is_valid()


class TestBPDocumentBasicEventSet(object):


    ## validation


    def test_anchor_ssid_not_in_span_sets(self):
        """A BPDocumentBasicEventSet is invalid when an SSID in the ``anchors`` field cannot be
        found within the span_sets for that Basic Event.
        """

        event_set = bp.BPDocumentBasicEventSet(
            {
                "ev-1": bp.BPDocumentBasicEvent(
                    "ev-1",
                    "ss-3",
                    [], # agents
                    [], # patients
                    [], # ref_events
                    "Corruption"
                ), 
            },
            {
                "ss-1": bp.BPDocumentSpanSet(
                    "ss-1",
                    [
                        bp.BPDocumentSpan("some string"),
                        bp.BPDocumentSpan("some other string"),
                    ]
                ),
                "ss-2": bp.BPDocumentSpanSet(
                    "ss-2",
                    [
                        bp.BPDocumentSpan("how now brown cow"),
                        bp.BPDocumentSpan("blarg honk"),
                    ]
                ),
            }
        )

        with pytest.raises(AssertionError):
            assert event_set.is_valid()


    def test_agent_ssid_not_in_span_sets(self):
        """A BPDocumentBasicEventSet is invalid when an SSID in the ``agents`` field cannot be
        found within the span_sets for that Basic Event.
        """

        event_set = bp.BPDocumentBasicEventSet(
            {
                "ev-1": bp.BPDocumentBasicEvent(
                    "ev-1",
                    "ss-1",
                    ["ss-3"],
                    [],
                    [],
                    "Coup"
                ), 
            },
            {
                "ss-1": bp.BPDocumentSpanSet(
                    "ss-1",
                    [
                        bp.BPDocumentSpan("some string"),
                        bp.BPDocumentSpan("some other string"),
                    ]
                ),
                "ss-2": bp.BPDocumentSpanSet(
                    "ss-2",
                    [
                        bp.BPDocumentSpan("how now brown cow"),
                        bp.BPDocumentSpan("blarg honk"),
                    ]
                ),
            }
        )

        with pytest.raises(AssertionError):
            assert event_set.is_valid()


    def test_patient_ssid_not_in_span_sets(self):
        """A BPDocumentBasicEventSet is invalid when an SSID in the ``patients`` field cannot be
        found within the span_sets for that Basic Event.
        """

        event_set = bp.BPDocumentBasicEventSet(
            {
                "ev-1": bp.BPDocumentBasicEvent(
                    "ev-1",
                    "ss-1",
                    [],
                    ["ss-3"],
                    [],
                    "Expel"
                ), 
            },
            {
                "ss-1": bp.BPDocumentSpanSet(
                    "ss-1",
                    [
                        bp.BPDocumentSpan("some string"),
                        bp.BPDocumentSpan("some other string"),
                    ]
                ),
                "ss-2": bp.BPDocumentSpanSet(
                    "ss-2",
                    [
                        bp.BPDocumentSpan("how now brown cow"),
                        bp.BPDocumentSpan("blarg honk"),
                    ]
                ),
            }
        )

        with pytest.raises(AssertionError):
            assert event_set.is_valid()


    def test_ref_events_ssid_not_in_span_sets(self):
        """A BPDocumentBasicEventSet is invalid when an event ID in the ``ref_events`` field cannot
        be found within the events for that Basic Event Set.
        """

        event_set = bp.BPDocumentBasicEventSet(
            {
                "ev-1": bp.BPDocumentBasicEvent(
                    "ev-1",
                    "ss-1",
                    [],
                    [],
                    ["ev-2",],
                    "Bribery",
                ),
            },
            {
                "ss-1": bp.BPDocumentSpanSet(
                    "ss-1",
                    [
                        bp.BPDocumentSpan("some string"),
                        bp.BPDocumentSpan("some other string"),
                    ]
                ),
                "ss-2": bp.BPDocumentSpanSet(
                    "ss-2",
                    [
                        bp.BPDocumentSpan("how now brown cow"),
                        bp.BPDocumentSpan("blarg honk"),
                    ]
                ),
            }
        )

        with pytest.raises(AssertionError):
            assert event_set.is_valid()


    def test_money_ssid_not_in_span_sets(self):
        """A BPDocumentBasicEventSet is invalid when an SSID in the ``money`` field cannot be
        found within the span_sets for that Basic Event.
        """

        event_set = bp.BPDocumentBasicEventSet(
            {
                "ev-1": bp.BPDocumentBasicEvent(
                    "ev-1",
                    "ss-1",
                    [],
                    [],
                    [],
                    "Leave-Job",
                    money=["ss-3",]
                ),
            },
            {
                "ss-1": bp.BPDocumentSpanSet(
                    "ss-1",
                    [
                        bp.BPDocumentSpan("some string"),
                        bp.BPDocumentSpan("some other string"),
                    ]
                ),
                "ss-2": bp.BPDocumentSpanSet(
                    "ss-2",
                    [
                        bp.BPDocumentSpan("how now brown cow"),
                        bp.BPDocumentSpan("blarg honk"),
                    ]
                ),
            }
        )

        with pytest.raises(AssertionError):
            assert event_set.is_valid()


class TestBPDocumentEntry(object):


    def test_language_id_is_valid(self):
        """A BPDocumentEntry is valid when its language code is valid or empty."""

        entry = bp.BPDocumentEntry("some-entry-id", "some-doc-id", lang="eng")
        assert entry.is_valid()

        entry = bp.BPDocumentEntry("some-entry-id", "some-doc-id")
        assert entry.is_valid()

        entry = bp.BPDocumentEntry("some-entry-id", "some-doc-id", lang="ARB")
        assert entry.is_valid()


    def test_language_id_is_invalid(self):
        """A BPDocumentEntry is invalid when its language code isn't in the ISO-639-3 format, or
        when it's not a real language code.
        """

        entry = bp.BPDocumentEntry("some-entry-id", "some-doc-id", lang="xxx")
        with pytest.raises(AssertionError):
            assert entry.is_valid()

        entry = bp.BPDocumentEntry("some-entry-id", "some-doc-id", lang="i'm not an iso code!")
        with pytest.raises(AssertionError):
            assert entry.is_valid()


class TestBPDocument(object):


    ## object tree reconstruction


    def test_from_json_determinism(self):
        """A BPDocument object should be deterministically reconstructable from a valid JSON
        file.
        """

        # a valid doc written by hand
        handwritten_doc = bp.BPDocument.from_json("tests/fixtures/tester.handwritten.bp.json")

        # a valid doc generated by loading the hand-written doc and serializing it
        generated_doc = bp.BPDocument.from_json("tests/fixtures/tester.generated.bp.json")

        assert handwritten_doc == generated_doc, \
            "Documents reconstructed from JSON files aren't identical!"


    def test_from_dict_determinism(self):
        """A BPDocument object should be deterministically reconstructable from a valid JSON
        dictionary.
        """
        from lib.utils import open_json_blob

        # a valid doc written by hand
        handwritten_doc = bp.BPDocument.from_dict(
            open_json_blob("tests/fixtures/tester.handwritten.bp.json")
        )

        # a valid doc generated by loading the hand-written doc and serializing it
        generated_doc = bp.BPDocument.from_dict(
            open_json_blob("tests/fixtures/tester.generated.bp.json")
        )

        assert handwritten_doc == generated_doc, \
            "Documents reconstructed from dictionaries aren't identical!"


    def test_eq_1(self):

        doc_a = bp.BPDocument.from_json("tests/fixtures/abstract.en.test-1.bp.json")
        doc_b = bp.BPDocument.from_json("tests/fixtures/abstract.en.test-1.bp.json")

        assert doc_a == doc_b, "Abstract document doesn't resolve to being identical to itself!"


    def test_eq_2(self):

        doc_a = bp.BPDocument.from_json("tests/fixtures/tester.handwritten.bp.json")
        doc_b = bp.BPDocument.from_json("tests/fixtures/tester.handwritten.bp.json")

        assert doc_a == doc_b, \
            "Handwritten Basic document doesn't resolve to being identical to itself!"


    def test_eq_3(self):

        doc_a = bp.BPDocument.from_json("tests/fixtures/tester.generated.bp.json")
        doc_b = bp.BPDocument.from_json("tests/fixtures/tester.generated.bp.json")

        assert doc_a == doc_b, \
            "Generated Basic document doesn't resolve to being identical to itself!"

