#!/usr/bin/env python3

####################################################################################################
### -----------------------------------------  NOTICE  ----------------------------------------- ###
###                                                                                              ###
### This software (or technical data) was produced for the U.S. Government under contract        ###
### 2015-14120200002-002, and is subject to the Rights in Data-General Clause 52.227-14 , ALT    ###
### IV (MAY 2014) or (DEC 2007).                                                                 ###
###                                                                                              ###
### © 2019 The MITRE Corporation. All Rights Reserved.                                           ###
###                                                                                              ###
### -------------------------------------------------------------------------------------------- ###
####################################################################################################

"""
Scores BP JSON results from performer (and reference) systems, for Abstract, Basic, and Granular
tasks.
"""

__author__  = "Leland E. Vakarian <lsperl@mitre.org>, David S. Day <day@mitre.org>"
__date__    = "28 June, 2022"

import click
import sys

from lib.utils import BP_ANNOTATOR_CLASS_TYPES
from lib.utils import CONTEXT_SETTINGS
from lib.utils import SCORING_EVENT_MATCH_INSIDE_SENTENCE_DEFAULT
from lib.utils import LOGGER
from lib.utils import MAT_DIALECT_CHOICES
from lib.utils import SCORING_ANCHOR_MATCH_TYPES
from lib.utils import SCORING_ANCHOR_MATCH_TYPE_DEFAULT
from lib.utils import SCORING_ARG_MATCH_TYPES
from lib.utils import SCORING_ARG_MATCH_TYPE_DEFAULT
from lib.utils import SCORING_EVENT_TYPE_MATCH_TYPES
from lib.utils import SCORING_EVENT_TYPE_MATCH_TYPE_DEFAULT
from lib.utils import SCORING_HEAD_MATCH_VALUE_DEFAULT
from lib.utils import SCORING_INCORRECT_SPAN_REDUCTION_RATE_DEFAULT
from lib.utils import SCORING_PARTIAL_MATCH_TYPES
from lib.utils import SCORING_PARTIAL_MATCH_TYPE_DEFAULT
from lib.utils import SCORING_QUADCLASS_MATCH_TYPES
from lib.utils import SCORING_QUADCLASS_MATCH_TYPE_DEFAULT
from lib.utils import SCORING_REF_ARG_SCORING_TYPES
from lib.utils import SCORING_REF_ARG_SCORING_TYPE_DEFAULT

DEBUG = False
FORCE = False
VERBOSE = False
"""bool: Global flags handled by cli()."""

ABSTRACT_TYPES = [
    "helpful_material",
    "helpful_verbal",
    "helpful_both",
    "harmful_material",
    "harmful_verbal",
    "harmful_both",
    "neutral_material",
    "neutral_verbal",
    "neutral_both",
    "helpful_unk",
    "harmful_unk",
    "neutral_unk",
    "None_None" # for systems that don't provide event types
]
"""There's no constant in ``lib.utils`` for this construction, nor should there be."""

SCORING_ANCHOR_MATCH_TYPE = SCORING_ANCHOR_MATCH_TYPE_DEFAULT
SCORING_ARG_MATCH_TYPE = SCORING_ARG_MATCH_TYPE_DEFAULT
SCORING_EVENT_TYPE_MATCH_TYPE = SCORING_EVENT_TYPE_MATCH_TYPE_DEFAULT
SCORING_PARTIAL_MATCH_TYPE = SCORING_PARTIAL_MATCH_TYPE_DEFAULT
SCORING_QUADCLASS_MATCH_TYPE = SCORING_QUADCLASS_MATCH_TYPE_DEFAULT
SCORING_REF_ARG_SCORING_TYPE = SCORING_REF_ARG_SCORING_TYPE_DEFAULT
SCORING_EVENT_MATCH_INSIDE_SENTENCE = SCORING_EVENT_MATCH_INSIDE_SENTENCE_DEFAULT
SCORING_INCORRECT_SPAN_REDUCTION_RATE = SCORING_INCORRECT_SPAN_REDUCTION_RATE_DEFAULT
SCORING_HEAD_MATCH_VALUE = SCORING_HEAD_MATCH_VALUE_DEFAULT
"""Global constants set (later) by the CLI, with defaults."""


## helper classes


class ArgumentAlignments(object):
    """Keeps track of a pair of arguments (span sets). Does not determine how to match arguments for
    scoring.
    """

    def __init__(self, type_):

        self.ssid_pairs = []
        self.measures = Measures(type_)

    def __eq__(self, other):

        if not isinstance(other, self.__class__):
            return NotImplemented

        __eq = lambda prop : getattr(self, prop) == getattr(other, prop)

        bools = []
        for prop in ["ssid_pairs", "measures",]:
            bools.append(__eq(prop))

        return all(bools)

    def __ne__(self, other):
        return not self == other


class EntryPair(object):
    """Keeps track of a pair of entries. Does not determine how to match entries for scoring."""

    def __init__(self, ref_entry, sys_entry, type_):

        from lib.utils import BASIC_MAT_KEY_ANCHOR_TYPES
        from lib.utils import GRANULAR_MAT_TEMPLATE_TYPES
        global ABSTRACT_TYPES

        self.ref_entry = ref_entry
        self.sys_entry = sys_entry
        self.type_ = type_ # dialect
        assert type_ in ["Abstract", "Basic", "Granular"], "Not a valid EntryPair type_!"

        # Pairs are matched by their `doc-id` and `sent-id` values, so it's possible for their
        # `entryid` values to not match. Here, a warning is printed whenever this is the case.
        if VERBOSE:
            if self.ref_entry and self.sys_entry:
                if self.ref_entry.entry_id.value != self.sys_entry.entry_id.value:
                    LOGGER.warning("EntryPair formed without matching Entry IDs!")
                    LOGGER.warning(f"- Reference Entry: {self.ref_entry.entry_id.value}")
                    LOGGER.warning(f"    - Doc ID: {self.ref_entry.doc_id.value}")
                    if self.type_ == "Abstract":
                        LOGGER.warning(f"    - Sent ID: {self.ref_entry.sent_id.value}")
                    LOGGER.warning(f"- System Entry: {self.sys_entry.entry_id.value}")
                    LOGGER.warning(f"    - Doc ID: {self.sys_entry.doc_id.value}")
                    if self.type_ == "Abstract":
                        LOGGER.warning(f"    - Sent ID: {self.sys_entry.sent_id.value}")
                    if self.type_ in ["Basic", "Granular",]:
                        condition = \
                            self.ref_entry.segment_text.value == self.sys_entry.segment_text.value
                        LOGGER.warning(f"Segment texts are{' not' if condition else ''} equal.")

        if ref_entry:
            if self.type_ == "Abstract":
                event_set = ref_entry.get_event_set("Abstract")
            elif self.type_ in ["Basic", "Granular",]:
                event_set = ref_entry.get_event_set("Basic")
            if event_set:
                self.ref_events = event_set.events.value or {}
                self.ref_span_sets = event_set.span_sets.value or {}
                if self.type_ in "Granular":
                    self.ref_templates = event_set.granular_templates.value or {}
            else:
                self.ref_events = {}
                self.ref_span_sets = {}
                self.ref_templates = {}
            self.ref_entry_id = ref_entry.entry_id.value
        else:
            self.ref_events = {}
            self.ref_span_sets = {}
            self.ref_entry_id = None
            self.ref_templates = {}

        if sys_entry:
            if self.type_ == "Abstract":
                event_set = sys_entry.get_event_set("Abstract")
            elif self.type_ in ["Basic", "Granular",]:
                event_set = sys_entry.get_event_set("Basic")
            if event_set:
                self.sys_events = event_set.events.value or {}
                self.sys_span_sets = event_set.span_sets.value or {}
                if self.type_ in "Granular":
                    self.sys_templates = event_set.granular_templates.value or {}
            else:
                self.sys_events = {}
                self.sys_span_sets = {}
                self.sys_templates = {}
            self.sys_entry_id = sys_entry.entry_id.value
        else:
            self.sys_events = {}
            self.sys_span_sets = {}
            self.sys_entry_id = None
            self.sys_templates = {}

        if self.type_ == "Abstract":
            self.scores = ScoreStruct()
            for evtype in ABSTRACT_TYPES:
                self.scores.evtype_match_measures[evtype] = Measures(evtype)
        elif self.type_ in ["Basic", "Granular",]:
            self.scores = ScoreStruct()
            for evtype in BASIC_MAT_KEY_ANCHOR_TYPES:
                self.scores.evtype_match_measures[evtype] = Measures(evtype)
            if self.type_ == "Granular":
                self.gran_scores = GranScoreStruct()
                for ttype in GRANULAR_MAT_TEMPLATE_TYPES:
                    self.gran_scores.ttype_match_measures[ttype] = Measures(ttype)

    def __eq__(self, other):

        if not isinstance(other, self.__class__):
            return NotImplemented

        __eq = lambda prop : getattr(self, prop) == getattr(other, prop)

        bools = []
        for prop in ["ref_entry", "sys_entry",]:
            bools.append(__eq(prop))

        return all(bools)

    def __ne__(self, other):
        return not self == other

    def __str__(self):
        if self.ref_entry:
            return self.ref_entry.entry_id.value
        else:
            return "<Missing Ref Entry>"

    def __repr__(self):
        return f"EntryPair({repr(self.ref_entry)}, {repr(self.sys_entry)}"

    def describe(self):

        global DEBUG

        if DEBUG:

            LOGGER.debug(f"Reference: {self.ref_entry_id}")
            if self.ref_span_sets:
                LOGGER.debug(f"    SpanSets: {self.ref_span_sets.keys()}")
            else:
                LOGGER.debug(f"    SpanSets: {self.ref_span_sets}")
            if self.ref_events:
                LOGGER.debug(f"    Events: {self.ref_events.keys()}")
            else:
                LOGGER.debug(f"    Events: {self.ref_events}")
            if self.ref_templates:
                LOGGER.debug(f"    Templates: {self.ref_templates.keys()}")
            else:
                LOGGER.debug(f"    Templates: {self.ref_templates}")

            LOGGER.debug(f"System: {self.sys_entry_id}")
            if self.sys_span_sets:
                LOGGER.debug(f"    SpanSets: {self.sys_span_sets.keys()}")
            else:
                LOGGER.debug(f"    SpanSets: {self.sys_span_sets}")
            if self.sys_events:
                LOGGER.debug(f"    Events: {self.sys_events.keys()}")
            else:
                LOGGER.debug(f"    Events: {self.sys_events}")
            if self.sys_templates:
                LOGGER.debug(f"    Templates: {self.sys_templates.keys()}")
            else:
                LOGGER.debug(f"    Templates: {self.sys_templates}")


class GranScoreStruct(object):
    """Score structs are used to maintain counts and scores at multiple levels, including corpus,
    entry-to-entry and event-to-event scores.

    This ScoreStruct is specific to Granular Templates.
    """

    def __init__(self):

        self.template_alignment = None
        self.template_pair_scores = {}
        self.slot_alignments = {}
        # measure obects that separately track event matchs, arg matches and their sum
        self.template_measures = Measures("template")
        self.slotmatch_measures = Measures("slotmatch")
        self.combined_measures = Measures("combined")
        # the product of event and argmatch f1
        self.combined_score = 0.0
        # other statistics for analyzing results (so called "extras")
        self.irrealis_measures = Measures("irrealis")
        self.time_attachment_measures = Measures("time-attachment")
        self.ttype_match_measures = {}

    def __eq__(self, other):

        if not isinstance(other, self.__class__):
            return NotImplemented

        __eq = lambda prop : getattr(self, prop) == getattr(other, prop)

        bools = []
        for prop in [
            "template_alignment", "template_pair_scores", "slot_alignments", "template_measures",
            "slotmatch_measures", "combined_measures", "ttype_match_measures", "combined_score",
            "irrealis_measures", "time_attachment_measures",
        ]:
            bools.append(__eq(prop))

        return all(bools)

    def __ne__(self, other):
        return not self == other

    def __str__(self):
        return str(self.combined_score)

    def compute_score(self, tab=0):

        self.template_measures.compute_metrics(tab=tab)
        self.slotmatch_measures.compute_metrics(tab=tab)
        # let's make sure that the combined counts and their 'value' are reflecting the latest
        # template and slotmatch info
        self.update_combined(tab=tab)
        self.combined_measures.compute_metrics(tab=tab)
        self.combined_score = self.template_measures.f1 * self.slotmatch_measures.f1

        # Extras
        self.irrealis_measures.compute_metrics(tab=tab)
        self.time_attachment_measures.compute_metrics(tab=tab)
        for ttype in self.ttype_match_measures:
            self.ttype_match_measures[ttype].compute_metrics(tab=tab)

        return self.combined_score

    def descr(self, msg_before="", msg_after="\n"):
        """Compact debugging."""

        global DEBUG
        global LOGGER

        if DEBUG:
            LOGGER.debug(msg_before)

            LOGGER.debug(f"combscr={self.combined_score} ")
            LOGGER.debug(f"combval={self.combined_measures.value} ")
            LOGGER.debug(f"combvalf1={self.combined_measures.f1} ")

            self.template_measures.descr(" evt:", "")
            self.slotmatch_measures.descr(" slot:", "")
            self.combined_measures.descr(" comb:", "")

            LOGGER.debug(msg_after)

    def describe(self, msg=""):
        """For debugging purposes."""
        from tabulate import tabulate

        global LOGGER

        if msg:
            LOGGER.debug(f"====== {msg} ======")

        for line in tabulate(
            [
                [
                    "template_measures",
                    self.template_measures.miss,
                    self.template_measures.false_alarm,
                    self.template_measures.hit,
                    self.template_measures.match,
                    self.template_measures.value,
                    self.template_measures.recall,
                    self.template_measures.precision,
                    self.template_measures.f1,
                ],
                [
                    "slotmatch_measures",
                    self.slotmatch_measures.miss,
                    self.slotmatch_measures.false_alarm,
                    self.slotmatch_measures.hit,
                    self.slotmatch_measures.match,
                    self.slotmatch_measures.value,
                    self.slotmatch_measures.recall,
                    self.slotmatch_measures.precision,
                    self.slotmatch_measures.f1,
                ],
                [
                    "combined_measures",
                    self.combined_measures.miss,
                    self.combined_measures.false_alarm,
                    self.combined_measures.hit,
                    self.combined_measures.match,
                    self.combined_measures.value,
                    self.combined_measures.recall,
                    self.combined_measures.precision,
                    self.combined_measures.f1,
                ],
            ],
            headers=[
                "Member", "Miss", "FA", "Hit", "Match", "OptValue", "Recall", "Precision", "F1",
            ],
            tablefmt="fancy_grid"
        ).split("\n"):
            LOGGER.debug(line)

        for line in tabulate(
            [
                ["combined_value", self.combined_measures.value],
                ["combined_f1", self.combined_measures.f1],
                ["combined_score", self.combined_score]
            ],
            headers=["Member", "Value"],
            tablefmt="fancy_grid"
        ).split("\n"):
            LOGGER.debug(line)

    def update(self, other, tab=0):

        self.template_measures.update(other.template_measures, tab=tab)
        self.slotmatch_measures.update(other.slotmatch_measures, tab=tab)

        self.irrealis_measures.update(other.irrealis_measures, tab=tab)
        self.time_attachment_measures.update(other.time_attachment_measures, tab=tab)

        # the combined measures are gathered from both the template and the slot measures
        # of the OTHER score structure passed in
        self.update_combined(tab=tab)

        keys = set(self.ttype_match_measures.keys()).union(set(other.ttype_match_measures.keys()))

        for evtype in keys:

            if (evtype in self.ttype_match_measures) and (evtype in other.ttype_match_measures):

                self.ttype_match_measures[evtype].update(
                    other.ttype_match_measures[evtype], tab=tab
                )

        self.compute_score(tab=tab)

    def update_combined(self, tab=0):

        # The combined measures are derived from the template and slot match measures.
        self.combined_measures.miss = self.template_measures.miss + self.slotmatch_measures.miss
        self.combined_measures.false_alarm = self.template_measures.false_alarm + \
            self.slotmatch_measures.false_alarm
        self.combined_measures.hit = self.template_measures.hit + self.slotmatch_measures.hit
        self.combined_measures.match = self.template_measures.match + self.slotmatch_measures.match
        self.combined_measures.compute_metrics(tab=tab)


class Measures(object):
    """Keeps track of measurements."""

    def __init__(self, type_):

        self.type_ = type_
        self.miss = 0
        self.false_alarm = 0
        self.hit = 0
        self.match = 0.0
        self.value = 0.0
        self.recall = 0.0
        self.precision = 0.0
        self.f1 = 0.0

    def __eq__(self, other):

        if not isinstance(other, self.__class__):
            return NotImplemented

        __eq = lambda prop : getattr(self, prop) == getattr(other, prop)

        bools = []
        for prop in [
            "type_", "miss", "false_alarm", "hit", "match", "value", "recall", "precision", "f1",
        ]:
            bools.append(__eq(prop))

        return all(bools)

    def __ne__(self, other):
        return not self == other

    def __str__(self):
        return (
            f"Type_: {self.type_} | Miss: {self.miss} | False Alarm: {self.false_alarm} | "
            f"Hit: {self.hit} | Match: {self.match} | Precision: {self.precision} | "
            f"Recall: {self.recall} | F1: {self.f1}"
        )

    def __repr__(self):
        return str(self)

    def increment_miss(self, increment, tab=0):
        self.miss += increment
        if DEBUG:
            LOGGER.debug(f"{' '*(tab*4)}[{self.type_}] Adding {increment} to Miss ({self.miss}).")

    def increment_false_alarm(self, increment, tab=0):
        self.false_alarm += increment
        if DEBUG:
            LOGGER.debug(
                f"{' '*(tab*4)}[{self.type_}] Adding {increment} to False Alarm "
                f"({self.false_alarm})."
            )

    def increment_match(self, increment, tab=0):
        self.match += increment
        if DEBUG:
            LOGGER.debug(f"{' '*(tab*4)}[{self.type_}] Adding {increment} to Match ({self.match}).")

    def increment_hit(self, increment, tab=0):
        self.hit += increment
        if DEBUG:
            LOGGER.debug(f"{' '*(tab*4)}[{self.type_}] Adding {increment} to Hit ({self.hit}).")

    def compute_metrics(self, tab=0):

        # match cannot be negative.
        self.match = max(0.0, self.match)

        ## Version 1:
        ## self.value = self.match - (self.false_alarm + self.miss)
        ## Version 2:
        self.value = self.match - self.false_alarm

        if float(self.hit + self.match + self.false_alarm + self.miss) == 0.0:

            # if DEBUG:
            #     LOGGER.debug(
            #         f"{' '*(tab*4)}[{self.type_}] Init, all basic values were 0."
            #     )

            # Give 100% credit if no false alarms or misses are encountered -- either the system was
            # perfect or there were no items in the reference set.
            self.precision = 1.0
            self.recall = 1.0
            self.f1 = 1.0

        else:

            if self.hit + self.false_alarm > 0:
                self.precision = float(self.match) / float(self.hit + self.false_alarm)

            if self.hit + self.miss > 0:
                self.recall = float(self.match) / float(self.hit + self.miss)

            if self.precision + self.recall != 0.0:
                self.f1 = 2.0 * ((self.precision * self.recall) / (self.precision + self.recall))

            # if DEBUG:
            #     LOGGER.debug(f"{' '*(tab*4)}[{self.type_}] Scores updated! {str(self)}")

    def update(self, other, tab=0):
        self.miss += other.miss
        self.false_alarm += other.false_alarm
        self.hit += other.hit
        self.match += other.match
        self.compute_metrics(tab=tab)

    def descr(self, msg_before="", msg_after="\n"):
        """Compact debugging purposes."""
        global DEBUG
        global LOGGER

        if DEBUG:
            LOGGER.debug(msg_before)
            LOGGER.debug(f"[mis={self.miss}")
            LOGGER.debug(f" fal={self.false_alarm}")
            LOGGER.debug(f" hit={self.hit}")
            LOGGER.debug(f" mch={self.match}")
            LOGGER.debug(f" val={self.value}")
            LOGGER.debug(f" rec={self.recall}")
            LOGGER.debug(f" pre={self.precision}")
            LOGGER.debug(f" f1={self.f1}]")
            LOGGER.debug(msg_after)

    def describe(self, comment=""):
        """Describes the state of this Measures object. For debugging purposes."""

        global LOGGER

        LOGGER.info(
            "comment=<<{7}>> miss={0} fa={1} hit={2} match={3} rec={4} pre={5} f1={6}\n".format(
                self.miss, self.false_alarm, self.hit, self.match, self.recall, self.precision,
                self.f1, comment
            )
        )


class SlotAlignments(object):
    """Keeps track of a pair of Granluar slots. Does not determine how to match arguments for
    scoring.
    """

    def __init__(self, type_):

        self.pairs = []
        self.measures = Measures(type_)

    def __eq__(self, other):

        if not isinstance(other, self.__class__):
            return NotImplemented

        __eq = lambda prop : getattr(self, prop) == getattr(other, prop)

        bools = []
        for prop in ["pairs", "measures",]:
            bools.append(__eq(prop))

        return all(bools)

    def __ne__(self, other):
        return not self == other


class ScoreStruct(object):
    """Score structs are used to maintain counts and scores at multiple levels, including corpus,
    entry-to-entry and event-to-event scores.
    """

    def __init__(self):

        self.event_alignment = None
        self.event_pair_scores = {}
        self.arg_alignments = {}
        # measure obects that separately track event matchs, arg matches and their sum
        self.event_measures = Measures("event")
        self.argmatch_measures = Measures("argmatch")
        self.combined_measures = Measures("combined")
        # the product of event and argmatch f1
        self.combined_score = 0.0
        # other statistics for analyzing results (so called "extras")
        self.agent_match_measures = Measures("agent")
        self.patient_match_measures = Measures("patient")
        self.money_match_measures = Measures("money")
        self.evtype_match_measures = {}

    def __eq__(self, other):

        if not isinstance(other, self.__class__):
            return NotImplemented

        __eq = lambda prop : getattr(self, prop) == getattr(other, prop)

        bools = []
        for prop in [
            "event_alignment", "event_pair_scores", "arg_alignments", "event_measures",
            "argmatch_measures", "combined_measures", "agent_match_measures",
            "patient_match_measures", "money_match_measures", "evtype_match_measures",
            "combined_score",
        ]:
            bools.append(__eq(prop))

        return all(bools)

    def __ne__(self, other):
        return not self == other

    def __str__(self):
        return str(self.combined_score)

    def compute_score(self, tab=0):

        self.event_measures.compute_metrics(tab=tab)
        self.argmatch_measures.compute_metrics(tab=tab)
        # let's make sure that the combined counts and their 'value' are reflecting the latest event
        # and argmatch info
        self.update_combined(tab=tab)
        self.combined_measures.compute_metrics(tab=tab)
        self.combined_score = self.event_measures.f1 * self.argmatch_measures.f1

        # Extras
        self.agent_match_measures.compute_metrics(tab=tab)
        self.patient_match_measures.compute_metrics(tab=tab)
        self.money_match_measures.compute_metrics(tab=tab)
        for evtype in self.evtype_match_measures:
            self.evtype_match_measures[evtype].compute_metrics(tab=tab)
        return self.combined_score

    def descr(self, msg_before="", msg_after="\n"):
        """Compact debugging."""

        global DEBUG
        global LOGGER

        if DEBUG:
            LOGGER.debug(msg_before)

            LOGGER.debug(f"combscr={self.combined_score} ")
            LOGGER.debug(f"combval={self.combined_measures.value} ")
            LOGGER.debug(f"combvalf1={self.combined_measures.f1} ")

            self.event_measures.descr(" evt:", "")
            self.argmatch_measures.descr(" arg:", "")
            self.combined_measures.descr(" comb:", "")

            LOGGER.debug(msg_after)

    def describe(self, msg=""):
        """For debugging purposes."""
        from tabulate import tabulate

        global LOGGER

        if msg:
            LOGGER.debug(f"====== {msg} ======")

        for line in tabulate(
            [
                [
                    "event_measures",
                    self.event_measures.miss,
                    self.event_measures.false_alarm,
                    self.event_measures.hit,
                    self.event_measures.match,
                    self.event_measures.value,
                    self.event_measures.recall,
                    self.event_measures.precision,
                    self.event_measures.f1,
                ],
                [
                    "argmatch_measures",
                    self.argmatch_measures.miss,
                    self.argmatch_measures.false_alarm,
                    self.argmatch_measures.hit,
                    self.argmatch_measures.match,
                    self.argmatch_measures.value,
                    self.argmatch_measures.recall,
                    self.argmatch_measures.precision,
                    self.argmatch_measures.f1,
                ],
                [
                    "combined_measures",
                    self.combined_measures.miss,
                    self.combined_measures.false_alarm,
                    self.combined_measures.hit,
                    self.combined_measures.match,
                    self.combined_measures.value,
                    self.combined_measures.recall,
                    self.combined_measures.precision,
                    self.combined_measures.f1,
                ],
            ],
            headers=[
                "Member", "Miss", "FA", "Hit", "Match", "OptValue", "Recall", "Precision", "F1",
            ],
            tablefmt="fancy_grid"
        ).split("\n"):
            LOGGER.debug(line)

        for line in tabulate(
            [["combined_value", self.combined_measures.value],
             ["combined_f1", self.combined_measures.f1],
             ["combined_score", self.combined_score]],
            headers=["Member", "Value"],
            tablefmt="fancy_grid").split("\n"):
                LOGGER.debug(line)

    def update(self, other, tab=0):

        self.event_measures.update(other.event_measures, tab=tab)
        self.argmatch_measures.update(other.argmatch_measures, tab=tab)

        # the combined measures are gathered from both the event and the arg measures
        # of the OTHER score structure passed in
        self.update_combined(tab=tab)

        self.agent_match_measures.update(other.agent_match_measures, tab=tab)
        self.patient_match_measures.update(other.patient_match_measures, tab=tab)
        self.money_match_measures.update(other.money_match_measures, tab=tab)

        keys = set(self.evtype_match_measures.keys()).union(set(other.evtype_match_measures.keys()))

        for evtype in keys:

            if (evtype in self.evtype_match_measures) and (evtype in other.evtype_match_measures):

                self.evtype_match_measures[evtype].update(
                    other.evtype_match_measures[evtype], tab=tab
                )

        self.compute_score(tab=tab)

    def update_combined(self, tab=0):

        # The combined measures are derived from the event and arg match measures.
        self.combined_measures.miss = self.event_measures.miss + self.argmatch_measures.miss
        self.combined_measures.false_alarm = self.event_measures.false_alarm + \
            self.argmatch_measures.false_alarm
        self.combined_measures.hit = self.event_measures.hit + self.argmatch_measures.hit
        self.combined_measures.match = self.event_measures.match + self.argmatch_measures.match
        self.combined_measures.compute_metrics(tab=tab)


## helper functions


def __aggregate_slot_scores(entry_pair):
    """Aggregates slot scores by template type and slot label for a given Entry pair.

    Args:
        entry_pair (:class:EntryPair): a pair of Entries to process.

    Returns:
        dict: keyed with a combination of template type and slot label
            (``f"{ttype}::{slot_label}"``) whose values are :class:Measures objects.
    """

    from lib.utils import GRANULAR_MAT_TEMPLATE_TYPES

    aggs = {}
    for ttype in GRANULAR_MAT_TEMPLATE_TYPES:
        for slot_label in GRANULAR_MAT_TEMPLATE_TYPES[ttype]:
            header = f"{ttype}::{GRANULAR_MAT_TEMPLATE_TYPES[ttype][slot_label]}"
            aggs[header] = Measures(header)

    if entry_pair.gran_scores.template_alignment:
        for i in range(len(entry_pair.gran_scores.template_alignment[0])):

            ref_tid = entry_pair.gran_scores.template_alignment[0][i]
            sys_tid = entry_pair.gran_scores.template_alignment[1][i]

            if ref_tid:
                ttype = entry_pair.ref_templates[ref_tid].type_.value
            else:
                ttype = entry_pair.sys_templates[sys_tid].type_.value
            slot_alignments = entry_pair.gran_scores.slot_alignments[(ref_tid, sys_tid)]

            for slot_label in slot_alignments:
                aggs[f"{ttype}::{slot_label}"].update(slot_alignments[slot_label].measures)

    return aggs


def __get_aligned_arg_strings(ref_ssid, sys_ssid, ref_span_sets, sys_span_sets, include_ssid=True):

    ref_arg_rendering = None
    if ref_ssid != None:
        ref_arg_rendering = __render_spanset_compact(
            ref_ssid, ref_span_sets, include_ssid=include_ssid
        )

    sys_arg_rendering = None
    if sys_ssid != None:
        sys_arg_rendering = __render_spanset_compact(
            sys_ssid, sys_span_sets, include_ssid=include_ssid
        )

    return (sys_arg_rendering, ref_arg_rendering)


def __get_events_in_sent_region(
    entry,
    entry_events,
    ref_sent,
    source_of_events,
    ignored_event_types
):
    """ Note that we only expect and require that there be sentence structural elements
        provided by the reference data -- systems need not provide sentences.  However,
        systems ARE required to indicate document-level character offsets, start and end,
        for each event anchor span provided.  """

    global DEBUG
    global LOGGER
    global VERBOSE

    events_in_sent = []

    if ref_sent == None and DEBUG:
        LOGGER.debug((
            f"NOTICE: without a reference sentence in {entry.entry_id.value}, "
            "all events from the whole document are being returned (__get_events_in_sent_region)"
        ))
        return entry_events.items()

    if entry_events:

        event_set = entry.get_event_set("Basic")
        spansets = event_set.span_sets.value
        if not event_set:
            if DEBUG:
                LOGGER.debug((
                    f"Nothing returned from get_event_set for entry {entry.entry_id.value}!"
                    "Skipping!"
                ))
            return events_in_sent

        for eventid, event in entry_events.items():

            if event.event_type.value in ignored_event_types:
                continue

            anchor_ssid = event.anchors.value
            anchor_spanset = spansets[anchor_ssid]

            for anchor_span in anchor_spanset.spans.value:
                if anchor_span.start.value != None and anchor_span.end.value != None:

                    if anchor_span.start.value >= ref_sent.start.value and \
                        anchor_span.start.value <  ref_sent.end.value   and \
                        anchor_span.end.value   >  ref_sent.start.value and \
                        anchor_span.start.value <= ref_sent.end.value:

                        # Multiple spans for an event will likely be in the same sentence, but we
                        # need/want only one event-id for each event in the list of events to be
                        # aligned.
                        if eventid not in events_in_sent:
                            events_in_sent.append(eventid)

                elif (anchor_span.string_.value != None):

                    if DEBUG:
                        LOGGER.debug(f"no anchor offsets, so looking for sub-string search...")
                    ref_sent_string = \
                        entry.segment_text.value[ref_sent.start.value:ref_sent.end.value]
                    if anchor_span.string_.value in ref_sent_string:
                        if eventid not in events_in_sent:
                            if DEBUG:
                                LOGGER.debug(
                                    "SUBSTRING FOUND.  anchor string: <<{0}>>".format(
                                        anchor_span.string_.value
                                    )
                                )
                                LOGGER.debug("ref_sent_string: <<{0}>>".format(ref_sent_string))
                            events_in_sent.append(eventid)
                elif DEBUG:
                    LOGGER.debug((
                        "no anchor offsets, no anchor span string, thus it is impossible to find "
                        "events constrained to be within a sentence ..."
                    ))

        if DEBUG:
            LOGGER.debug((
                f"found these {source_of_events}={events_in_sent} in this sentence: "
                f"{ref_sent.start.value}:{ref_sent.end.value}"
            ))

    return events_in_sent


def __get_ref_sents(ref_entry):
    global LOGGER
    global VERBOSE

    ref_sents = []
    if ref_entry.segment_sections.value == None:
        if VERBOSE:
            LOGGER.error((
                f"ERROR: The reference entry {ref_entry.entry_id.value} "
                "has no sentence segments (or any segment_sections at all)"
            ))
        return ref_sents

    for segment_section in ref_entry.segment_sections.value:
        if segment_section.structural_element.value == "Sentence":
            ref_sents.append(segment_section)

    if len(ref_sents) == 0 and VERBOSE:
        LOGGER.error(
            f"ERROR: The reference entry {ref_entry.entry_id.value} has no sentence segments"
        )

    return ref_sents


def __mention_str(span, type_):

    if type_ == 'head':
        if span.hstring.value:
            return span.hstring.value
        elif span.no_head.value:
            return ""
        else:
            return span.string_.value

    elif type_ == 'full':
        if span.string_.value:
            return span.string_.value
        elif span.hstring.value:
            return span.hstring.value
        else:
            return ""

    elif type_ == "string":
        return span.string_.value

    elif type_ == "hstring":
        return span.hstring.value

    else:
        return ""


def __render_aligned_entry(sys_entry_id, ref_entry_id, entry_pair):
    global LOGGER
    global VERBOSE
    global SCORING_EVENT_MATCH_INSIDE_SENTENCE

    if SCORING_EVENT_MATCH_INSIDE_SENTENCE:
        for ref_sent in __get_ref_sents(entry_pair.ref_entry):

            ref_sent_string = entry_pair.ref_entry.segment_text.value[
                ref_sent.start.value:ref_sent.end.value
            ]
            ref_sent_string = ref_sent_string.replace("\n"," ").replace("\r"," ")

            if VERBOSE:
                LOGGER.info(
                    f"\nsentence: <<{ref_sent_string}>> "
                    f"[{ref_sent.start.value}:{ref_sent.end.value}])"
                )

            sys_event_ids = __get_events_in_sent_region(
                entry_pair.sys_entry, entry_pair.sys_events, ref_sent, "sys events")
            ref_event_ids = __get_events_in_sent_region(
                entry_pair.ref_entry, entry_pair.ref_events, ref_sent, "ref events")

            for ref_sys_id_pair, score_str in entry_pair.scores.event_pair_scores.items():
                sys_event_id = ref_sys_id_pair[1]
                ref_event_id = ref_sys_id_pair[0]

                if sys_event_id in sys_event_ids or ref_event_id in ref_event_ids:

                    # Render the event pairs organized by sentence
                    if sys_event_id == None:
                        __render_miss_event(ref_event_id, entry_pair, score_str)
                    elif ref_event_id == None:
                        __render_false_alarm_event(sys_event_id, entry_pair, score_str)
                    else:
                        __render_aligned_events(sys_event_id, ref_event_id, entry_pair, score_str)

    else:

        for ref_sys_id_pair, score_str in entry_pair.scores.event_pair_scores.items():
            sys_event_id = ref_sys_id_pair[1]
            ref_event_id = ref_sys_id_pair[0]
            # Render the event pairs organized by sentence
            if sys_event_id == None:
                __render_miss_event(ref_event_id, entry_pair, score_str)
            elif ref_event_id == None:
                __render_false_alarm_event(sys_event_id, entry_pair, score_str)
            else:
                __render_aligned_events(sys_event_id, ref_event_id, entry_pair, score_str)


def __render_aligned_events(sys_event_id, ref_event_id, entry_pair, score_str):
    global LOGGER
    global VERBOSE

    sys_entry = entry_pair.sys_entry
    sys_span_sets = entry_pair.sys_span_sets
    sys_event = entry_pair.sys_events[sys_event_id]
    sys_event_type = sys_event.event_type.value
    sys_anchor = __render_spanset_compact(sys_event.anchors.value, sys_span_sets)

    ref_entry = entry_pair.ref_entry
    ref_span_sets = entry_pair.ref_span_sets
    ref_event = entry_pair.ref_events[ref_event_id]
    ref_event_type = ref_event.event_type.value
    ref_anchor = __render_spanset_compact(ref_event.anchors.value, ref_span_sets)

    if VERBOSE:
        LOGGER.info(f"  type: {sys_event_type}({sys_event_id}) :: {ref_event_type}({ref_event_id})")
        LOGGER.info(f"    anchors: {sys_anchor} :: {ref_anchor}")

    for arg_field in ["agents", "patients", "ref_events", "money",]:

        if arg_field in score_str.arg_alignments:

            for ssid_pair in score_str.arg_alignments[arg_field].ssid_pairs:

                ref_ssid = ssid_pair[0]
                sys_ssid = ssid_pair[1]
                (sys_arg_rendering, ref_arg_rendering) = \
                    __get_aligned_arg_strings(ref_ssid, sys_ssid, ref_span_sets, sys_span_sets)

                if VERBOSE:
                    LOGGER.info(f"    {arg_field}: {sys_arg_rendering} :: {ref_arg_rendering}")


def __render_entry_alignment(entry_pair):
    global LOGGER
    global VERBOSE

    sys_entry = entry_pair.sys_entry
    sys_entry_id = None
    if sys_entry != None:
        sys_entry_id = sys_entry.entry_id.value

    ref_entry = entry_pair.ref_entry
    ref_entry_id = None
    if ref_entry != None:
        ref_entry_id = ref_entry.entry_id.value

    if VERBOSE:
        LOGGER.info((
            f"\n----------- sys_entry = {sys_entry_id} :: {ref_entry_id} = ref_entry "
            "-------------------"
        ))

    if sys_entry == None:
        __render_miss_entry(ref_entry_id, entry_pair)
    elif ref_entry == None:
        __render_false_alarm_entry(sys_entry_id, entry_pair)
    else:
        __render_aligned_entry(sys_entry_id, ref_entry_id, entry_pair)


def __render_false_alarm_entry(sys_entry_id, entry_pair):
    global LOGGER
    global VERBOSE
    global SCORING_EVENT_MATCH_INSIDE_SENTENCE

    sys_event_ids = [sys_event for sys_event in entry_pair.sys_events]
    if SCORING_EVENT_MATCH_INSIDE_SENTENCE:
        for ref_sent in __get_ref_sents(entry_pair.ref_entry):

            ref_sent_string = entry_pair.ref_entry.segment_text.value[
                ref_sent.start.value:ref_sent.end.value
            ]
            ref_sent_string = ref_sent_string.replace("\n"," ").replace("\r"," ")

            if VERBOSE:
                LOGGER.info(
                    f"\nsentence: <<{ref_sent_string}>> "
                    f"[{ref_sent.start.value}:{ref_sent.end.value}])"
                )

            sys_event_ids = __get_events_in_sent_region(
                entry_pair.sys_entry, entry_pair.sys_events, ref_sent, "sys events")

    for ref_sys_id_pair, score_str in entry_pair.scores.event_pair_scores.items():
        sys_event_id = ref_sys_id_pair[1]
        if sys_event_id in sys_event_ids:
            __render_false_alarm_event(sys_event_id, entry_pair, score_str)


def __render_false_alarm_event(sys_event_id, entry_pair, score_str):
    global LOGGER
    global VERBOSE

    sys_entry = entry_pair.sys_entry
    sys_span_sets = entry_pair.sys_span_sets
    sys_event = entry_pair.sys_events[sys_event_id]
    sys_event_type = sys_event.event_type.value
    sys_anchor = __render_spanset_compact(sys_event.anchors.value, sys_span_sets)

    if VERBOSE:
        LOGGER.info(f"  type: {sys_event_type}({sys_event_id}) :: None")
        LOGGER.info(f"    anchors: {sys_anchor} :: None")

    for arg_field in ["agents", "patients", "ref_events", "money",]:

        if (arg_field in score_str.arg_alignments):

            for ssid_pair in score_str.arg_alignments[arg_field].ssid_pairs:

                ref_ssid = ssid_pair[0]
                sys_ssid = ssid_pair[1]
                (sys_arg_rendering, ref_arg_rendering) = \
                    __get_aligned_arg_strings(ref_ssid, sys_ssid, {}, sys_span_sets)

                if VERBOSE:
                    LOGGER.info(f"    {arg_field}: {sys_arg_rendering} :: {ref_arg_rendering}")


def __render_headers(scores, granular=False, extras=False):
    """Generates a list of headers for tabular data output.

    Args:
        granular (bool): whether the scorer was run in Granular mode or not.
        extras (bool): whether to include fine-grain details as well.

    Returns:
        list: a list of strings, representing the headers to emit.
    """
    from lib.utils import GRANULAR_MAT_TEMPLATE_TYPES

    headers = ["sys-entry-id", "ref-entry-id",]
    std_headers = [
        "misses", "false-alarms", "alignments", "match-qual", "optimizer-value", "precision",
        "recall", "fmeasure",
    ]
    extra_headers = [
        "misses", "false-alarms", "alignments", "match-qual", "precision", "recall", "fmeasure",
    ]

    if granular:
        # event match counts and summary stats
        headers.extend([f"tmplt-{col}" for col in std_headers])
        # arg match counts and summary stats
        headers.extend([f"slot-{col}" for col in std_headers])
    else:
        # event match counts and summary stats
        headers.extend([f"evt-{col}" for col in std_headers])
        # arg match counts and summary stats
        headers.extend([f"arg-{col}" for col in std_headers])

    headers.extend([f"comb-{col}" for col in std_headers])
    headers.append("combined-score")

    if extras:

        if granular:
            extra_labels = sorted(list(scores.ttype_match_measures.keys()))
            extra_labels.extend(["irrealis", "time-attachments"])
            for ttype in GRANULAR_MAT_TEMPLATE_TYPES:
                for slot_label in GRANULAR_MAT_TEMPLATE_TYPES[ttype]:
                    extra_labels.append(
                        f"{ttype}::{slot_label}"
                    )
        else:
            extra_labels = ["agents", "patients", "money",] + sorted(
                list(scores.evtype_match_measures.keys())
            )
        for label in extra_labels:
            for header in extra_headers:
                headers.append(f"{label}-{header}")

    return headers


def __render_miss_entry(ref_entry_id, entry_pair):
    global LOGGER
    global VERBOSE
    global SCORING_EVENT_MATCH_INSIDE_SENTENCE

    ref_event_ids = [ref_event for ref_event in entry_pair.ref_events]
    if SCORING_EVENT_MATCH_INSIDE_SENTENCE:
        for ref_sent in __get_ref_sents(entry_pair.ref_entry):

            ref_sent_string = entry_pair.ref_entry.segment_text.value[
                ref_sent.start.value:ref_sent.end.value
            ]
            ref_sent_string = ref_sent_string.replace("\n"," ").replace("\r"," ")

            if VERBOSE:
                LOGGER.info(
                    f"\nsentence: <<{ref_sent_string}>> "
                    f"[{ref_sent.start.value}:{ref_sent.end.value}])"
                )

            ref_event_ids = __get_events_in_sent_region(
                entry_pair.ref_entry, entry_pair.ref_events, ref_sent, "ref events")

    for ref_sys_id_pair, score_str in entry_pair.scores.event_pair_scores.items():
        ref_event_id = ref_sys_id_pair[0]
        if ref_event_id in ref_event_ids:
            # Render the event pairs organized by sentence
            __render_miss_event(ref_event_id, entry_pair, score_str)


def __render_miss_event(ref_event_id, entry_pair, score_str):
    global LOGGER
    global VERBOSE

    ref_entry = entry_pair.ref_entry
    ref_span_sets = entry_pair.ref_span_sets
    ref_event = entry_pair.ref_events[ref_event_id]
    ref_event_type = ref_event.event_type.value
    ref_anchor = __render_spanset_compact(ref_event.anchors.value, ref_span_sets)

    if VERBOSE:
        LOGGER.info(f"  type: None :: {ref_event_type}({ref_event_id})")
        LOGGER.info(f"    anchors: None :: {ref_anchor}")

    for arg_field in ["agents", "patients", "ref_events", "money",]:

        if arg_field in score_str.arg_alignments:

            for ssid_pair in score_str.arg_alignments[arg_field].ssid_pairs:

                ref_ssid = ssid_pair[0]
                sys_ssid = ssid_pair[1]
                (sys_arg_rendering, ref_arg_rendering) = \
                    __get_aligned_arg_strings(ref_ssid, sys_ssid, ref_span_sets, {})

                if VERBOSE:
                    LOGGER.info(f"    {arg_field}: {sys_arg_rendering} :: {ref_arg_rendering}")


def __render_spanset_compact(ssid, span_sets, include_ssid=True):

    if ssid == "5-merged-2":
        LOGGER.debug(span_sets.keys())

    parts = []
    spans = span_sets[ssid].spans.value
    for span in spans:
        span_string = span.string_.value.replace("\n"," ").replace("\r"," ")
        head_string = ""
        if (span.hstring.value != None):
            head_string = span.hstring.value.replace("\n"," ").replace("\r"," ")
            if span.synclass.value == None:
                span_depiction = f"<<{span_string}|head={head_string}>>"
            else:
                span_depiction = f"<<{span_string}|head={head_string}//{span.synclass.value}>>"
        else:
            if span.synclass.value == None:
                span_depiction = f"<<{span_string}>>"
            else:
                span_depiction = f"<<{span_string}//{span.synclass.value}>>"
        if (span_depiction not in parts):
            parts.append (span_depiction)
            parts.append(",")
    ## remove the trailing comma
    parts = parts[:-1]
    if include_ssid:
        parts.append(f"({ssid})")
    spanset_compact = ''.join(parts)
    return spanset_compact


def __score_abstract_event_pair(
    ref_eid,
    sys_eid,
    entry_pair,
    event_pair_score_cache,
    ignore_synclass=False,
    extras=False,
):

    global ABSTRACT_TYPES
    global DEBUG
    global SCORING_QUADCLASS_MATCH_TYPE

    if DEBUG:
        LOGGER.debug(f">>> Scoring Entry Pair: {entry_pair}")

    if (ref_eid, sys_eid) in event_pair_score_cache:

        if DEBUG:
            LOGGER.debug((
                "        Pair already in score cache! Score: "
                f"{event_pair_score_cache[(ref_eid, sys_eid)].combined_score}"
                "        Value: "
                f"{event_pair_score_cache[(ref_eid, sys_eid)].combined_measures.value}"
            ))
        ## return event_pair_score_cache[(ref_eid, sys_eid)].combined_score
        return event_pair_score_cache[(ref_eid, sys_eid)].combined_measures.value

    pair_scores = ScoreStruct()
    for evtype in ABSTRACT_TYPES:
        pair_scores.evtype_match_measures[evtype] = Measures(evtype)

    if not sys_eid:

        if DEBUG:
            LOGGER.debug((
                "        Adding 1 miss to evt_measures since there's no system entry in the score "
                "cache."
            ))
        pair_scores.event_measures.increment_miss(1, tab=3)

        if extras:
            ref_event = entry_pair.ref_events[ref_eid]
            evtype = f"{ref_event.hh.value}_{ref_event.mv.value}"
            pair_scores.evtype_match_measures[evtype].increment_miss(1, tab=3)

    elif not ref_eid:

        if DEBUG:
            LOGGER.debug((
                "        Adding 1 false alarm to evt_measures since there's no reference entry in "
                "the score cache."
            ))
        pair_scores.event_measures.increment_false_alarm(1, tab=3)

        if extras:
            sys_event = entry_pair.sys_events[sys_eid]
            evtype = f"{sys_event.hh.value}_{sys_event.mv.value}"
            pair_scores.evtype_match_measures[evtype].increment_false_alarm(1, tab=3)

    else:

        sys_event = entry_pair.sys_events[sys_eid]
        ref_event = entry_pair.ref_events[ref_eid]

        if extras:
            # using the reference QC for the extras label
            evtype = f"{ref_event.hh.value}_{ref_event.mv.value}"

        if SCORING_QUADCLASS_MATCH_TYPE == 'strict':

            if not ref_event.mv.value and not sys_event.mv.value and not ref_event.hh.value and \
                not sys_event.hh.value:

                # Both dimensions of QuadClass are empty in both sys and ref
                if DEBUG:
                    LOGGER.debug((
                        f"        [{SCORING_QUADCLASS_MATCH_TYPE}] "
                        "Adding 1 hit (1.0 match) to evt_measures since both QC are empty."
                    ))
                pair_scores.event_measures.increment_match(1.0, tab=3)
                pair_scores.event_measures.increment_hit(1, tab=3)

                if extras:
                    pair_scores.evtype_match_measures[evtype].increment_match(1.0, tab=3)
                    pair_scores.evtype_match_measures[evtype].increment_hit(1, tab=3)

            elif ref_event.mv.value and ref_event.hh.value and sys_event.mv.value and \
                sys_event.hh.value and ref_event.mv.value == sys_event.mv.value and \
                ref_event.hh.value == sys_event.hh.value:

                # Both dimensions of QuadClass match
                if DEBUG:
                    LOGGER.debug((
                        f"        [{SCORING_QUADCLASS_MATCH_TYPE}] "
                        "Adding 1 hit (1.0 match) to evt_measures since both QC are the same."
                    ))
                pair_scores.event_measures.increment_match(1.0, tab=3)
                pair_scores.event_measures.increment_hit(1, tab=3)

                if extras:
                    pair_scores.evtype_match_measures[evtype].increment_match(1.0, tab=3)
                    pair_scores.evtype_match_measures[evtype].increment_hit(1, tab=3)

            else:
                # An event with this specific reference quadclass wasn't generated by system
                if DEBUG:
                    LOGGER.debug((
                        f"        [{SCORING_QUADCLASS_MATCH_TYPE}] "
                        "Adding 1 miss to evt_measures since the ref QC is missing in sys."
                    ))
                pair_scores.event_measures.increment_miss(1, tab=3)

                # An event with this specific system quadclass had no counterpart in reference
                if DEBUG:
                    LOGGER.debug((
                        f"        [{SCORING_QUADCLASS_MATCH_TYPE}] "
                        "Adding 1 false alarm to evt_measures since the sys QC is missing in ref."
                    ))
                pair_scores.event_measures.increment_false_alarm(1, tab=3)

                if extras:
                    pair_scores.evtype_match_measures[evtype].increment_miss(1, tab=3)
                    pair_scores.evtype_match_measures[evtype].increment_false_alarm(1, tab=3)

        elif SCORING_QUADCLASS_MATCH_TYPE == 'loose':

            mv_match = 0.0
            if (not ref_event.mv.value and not sys_event.mv.value) or (ref_event.mv.value and \
                sys_event.mv.value and ref_event.mv.value == sys_event.mv.value):
                if DEBUG:
                    LOGGER.debug((
                        f"        [{SCORING_QUADCLASS_MATCH_TYPE}] "
                        "Material-Verbal 1.0 match since both values are identical."
                    ))
                mv_match = 1.0

            elif ref_event.mv.value and sys_event.mv.value and \
                (ref_event.mv.value == "both" or sys_event.mv.value == "both"):
                if DEBUG:
                    LOGGER.debug((
                        f"        [{SCORING_QUADCLASS_MATCH_TYPE}] "
                        "Material-Verbal 0.5 match since neither is None and one of the values is "
                        "'both'."
                    ))
                mv_match = 0.5

            else:
                if DEBUG:
                    LOGGER.debug((
                        f"        [{SCORING_QUADCLASS_MATCH_TYPE}] "
                        "Material-Verbal 0.0 match since the values don't match, neither is None "
                        "and neither is Both."
                    ))
                mv_match = 0.0

            hh_match = 0.0
            if (not ref_event.hh.value and not sys_event.hh.value) or (ref_event.hh.value and \
                sys_event.hh.value and ref_event.hh.value == sys_event.hh.value):
                if DEBUG:
                    LOGGER.debug((
                        f"        [{SCORING_QUADCLASS_MATCH_TYPE}] "
                        "Helpful-Harmful 1.0 match since both values match and the values aren't "
                        "'both' or None."
                    ))
                hh_match = 1.0

            else:
                if DEBUG:
                    LOGGER.debug((
                        f"        [{SCORING_QUADCLASS_MATCH_TYPE}] "
                        "Helpful-Harmful 0.0 match since the values don't match, or the both values"
                        " are None."
                    ))
                hh_match = 0.0

            qc_match = (mv_match + hh_match) / 2.0
            if DEBUG:
                LOGGER.debug((
                    f"        [{SCORING_QUADCLASS_MATCH_TYPE}] "
                    f"Adding match value {qc_match} because of QuadClass match to evt_measures."
                ))
                LOGGER.debug((
                    f"        [{SCORING_QUADCLASS_MATCH_TYPE}] "
                    f"Adding 1 hit to evt_measures."
                ))

            # There was a fractional match
            pair_scores.event_measures.increment_match(qc_match, tab=3)

            # Out of this total possible hit count
            pair_scores.event_measures.increment_hit(1, tab=3)

            if extras:
                pair_scores.evtype_match_measures[evtype].increment_match(qc_match, tab=3)
                pair_scores.evtype_match_measures[evtype].increment_hit(1, tab=3)

        elif SCORING_QUADCLASS_MATCH_TYPE == 'ignore':

            # We give credit without consulting actual quadclass values
            if DEBUG:
                LOGGER.debug((
                    f"        [{SCORING_QUADCLASS_MATCH_TYPE}] "
                    f"Adding 1.0 match to evt_measures."
                ))
                LOGGER.debug((
                    f"        [{SCORING_QUADCLASS_MATCH_TYPE}] "
                    f"Adding 1 hit to evt_measures."
                ))
            pair_scores.event_measures.increment_match(1.0, tab=3)
            pair_scores.event_measures.increment_hit(1, tab=3)

            if extras:
                pair_scores.evtype_match_measures[evtype].increment_match(1.0, tab=3)
                pair_scores.evtype_match_measures[evtype].increment_hit(1, tab=3)

    __score_arguments(
        entry_pair,
        (ref_eid, sys_eid,),
        pair_scores,
        "agents",
        ignore_synclass=ignore_synclass,
        extras=extras
    )
    __score_arguments(
        entry_pair,
        (ref_eid, sys_eid,),
        pair_scores,
        "patients",
        ignore_synclass=ignore_synclass,
        extras=extras
    )

    pair_scores.compute_score()
    event_pair_score_cache[(ref_eid, sys_eid)] = pair_scores

    return pair_scores.combined_measures.value


def __score_arguments(
    entry_pair,
    ref_sys_pair,
    pair_scores,
    arg_field,
    ignore_synclass=False,
    extras=False
):

    import itertools

    from munkres import make_cost_matrix
    from munkres import Munkres

    ref_eid = ref_sys_pair[0]
    sys_eid = ref_sys_pair[1]
    ref_ssids = []
    sys_ssids = []
    ref_event = None
    sys_event = None

    if DEBUG:
        LOGGER.debug((
            f">>>>>>> Scoring {arg_field} values for Entry Pair: {entry_pair}, ref EventID: "
            f"{ref_eid}, sys EventID: {sys_eid} "
        ))

    if ref_eid in entry_pair.ref_events:
        ref_event = entry_pair.ref_events[ref_eid]
        ref_ssids = getattr(ref_event, arg_field).value or []

    if sys_eid in entry_pair.sys_events:
        sys_event = entry_pair.sys_events[sys_eid]
        sys_ssids = getattr(sys_event, arg_field).value or []

    if arg_field == "anchors":
        ref_ssids = [ref_ssids]
        sys_ssids = [sys_ssids]

    elif arg_field == "ref_events":
        if ref_event:
            for eid in ref_event.ref_events.value:
                assert eid in entry_pair.ref_events, \
                    f"Can't find EventID {eid} in reference events!"
            ref_ssids = [
                entry_pair.ref_events[eid].anchors.value for eid in ref_event.ref_events.value
            ]
        if sys_event:
            for eid in sys_event.ref_events.value:
                assert eid in entry_pair.sys_events, \
                    f"Can't find EventID {eid} in system events!"
            sys_ssids = [
                entry_pair.sys_events[eid].anchors.value for eid in sys_event.ref_events.value
            ]

    if DEBUG:
        LOGGER.debug(f"            ref_event: {ref_event}")
        LOGGER.debug(f"            ref_ssids: {ref_ssids}")
        LOGGER.debug(f"            sys_event: {sys_event}")
        LOGGER.debug(f"            sys_ssids: {sys_ssids}")

    pair_scores.arg_alignments[arg_field] = ArgumentAlignments(arg_field)

    ## check for easy edge conditions

    if not sys_event or not len(sys_ssids) or not sys_ssids[0]:

        if not ref_event or not len(ref_ssids) or not ref_ssids[0]:
            # Giving no positive value for the correct "decision" to create zero arguments
            if DEBUG:
                LOGGER.debug("            No sys or ref spans found! Nothing to do.")

        else:

            if DEBUG:
                LOGGER.debug((
                    f"            No sys spans found, but {len(ref_ssids)} ref spans found! "
                    f"Adding {len(ref_ssids)} misses to arg alignments."
                ))

            for ref_ssid in ref_ssids:
                pair_scores.arg_alignments[arg_field].ssid_pairs.append([ref_ssid, None])
                pair_scores.arg_alignments[arg_field].measures.increment_miss(1, tab=4)

                if extras and arg_field == "agents":
                    pair_scores.agent_match_measures.increment_miss(1, tab=4)
                elif extras and arg_field == "patients":
                    pair_scores.patient_match_measures.increment_miss(1, tab=4)
                elif extras and arg_field == "money":
                    pair_scores.money_match_measures.increment_miss(1, tab=4)

            pair_scores.arg_alignments[arg_field].measures.compute_metrics(tab=4)

            if extras and arg_field == "agents":
                pair_scores.agent_match_measures.compute_metrics(tab=4)
            elif extras and arg_field == "patients":
                pair_scores.patient_match_measures.compute_metrics(tab=4)
            elif extras and arg_field == "money":
                pair_scores.money_match_measures.compute_metrics(tab=4)

        pair_scores.argmatch_measures.update(pair_scores.arg_alignments[arg_field].measures, tab=4)

        return

    elif not ref_event or not len(ref_ssids) or not ref_ssids[0]:

        if DEBUG:
            LOGGER.debug((
                f"            No ref spans found, but {len(sys_ssids)} sys spans found! "
                f"Adding {len(sys_ssids)} false alarms to arg_alignments."
            ))

        for sys_ssid in sys_ssids:
            pair_scores.arg_alignments[arg_field].ssid_pairs.append([None, sys_ssid])
            pair_scores.arg_alignments[arg_field].measures.increment_false_alarm(1, tab=4)

            if extras and arg_field == "agents":
                pair_scores.agent_match_measures.increment_false_alarm(1, tab=4)
            elif extras and arg_field == "patients":
                pair_scores.patient_match_measures.increment_false_alarm(1, tab=4)
            elif extras and arg_field == "money":
                pair_scores.money_match_measures.increment_false_alarm(1, tab=4)

        pair_scores.arg_alignments[arg_field].measures.compute_metrics(tab=4)

        if extras and arg_field == "agents":
            pair_scores.agent_match_measures.compute_metrics(tab=4)
        elif extras and arg_field == "patients":
            pair_scores.patient_match_measures.compute_metrics(tab=4)
        elif extras and arg_field == "money":
            pair_scores.money_match_measures.compute_metrics(tab=4)

        pair_scores.argmatch_measures.update(pair_scores.arg_alignments[arg_field].measures, tab=4)

        return

    ## Execute Kuhn-Munkres optimizing search for best match between system and reference spans

    if len(ref_ssids) < len(sys_ssids):

        if DEBUG:
            LOGGER.debug((
                f"            Padding ref ssids with {len(sys_ssids) - len(ref_ssids)} blanks since"
                f" sys_ssids is longer."
            ))

        # pad with `None`s
        ref_ssids = ref_ssids + [x for x in itertools.repeat(None, len(sys_ssids)-len(ref_ssids))]

    elif len(sys_ssids) < len(ref_ssids):

        if DEBUG:
            LOGGER.debug((
                f"            Padding sys ssids with {len(ref_ssids) - len(sys_ssids)} blanks since"
                f" ref_ssids is longer."
            ))

        # pad with `None`s
        sys_ssids = sys_ssids + [x for x in itertools.repeat(None, len(ref_ssids)-len(sys_ssids))]

    match_cache = {}
    matrix = make_cost_matrix(
        [
            [
                __score_ss_pair(
                    sys_ssid,
                    ref_ssid,
                    entry_pair,
                    match_cache,
                    ignore_synclass=ignore_synclass
                ) for sys_ssid in sys_ssids
            ] for ref_ssid in ref_ssids
        ],
        lambda cost: 1.0 - cost
    )
    index_pairs = Munkres().compute(matrix)

    for row, column in index_pairs:

        ref_ssid = ref_ssids[row]
        sys_ssid = sys_ssids[column]

        if not sys_ssid and not ref_ssid:

            if DEBUG:
                LOGGER.debug(f"            Both sys and ref ssids are blank; skipping.")

        elif not sys_ssid:

            if DEBUG:
                LOGGER.debug((
                    f"            Adding 1 miss to arg_alignments since ref spanset {ref_ssid} has "
                    "no sys counterpart."
                ))

            pair_scores.arg_alignments[arg_field].measures.increment_miss(1)
            if extras and arg_field == "agents":
                pair_scores.agent_match_measures.increment_miss(1, tab=4)
            elif extras and arg_field == "patients":
                pair_scores.patient_match_measures.increment_miss(1, tab=4)
            elif extras and arg_field == "money":
                pair_scores.money_match_measures.increment_miss(1, tab=4)
            pair_scores.arg_alignments[arg_field].ssid_pairs.append([ref_ssid, sys_ssid])

        elif not ref_ssid:

            if DEBUG:
                LOGGER.debug((
                    "            Adding 1 false alarm to arg_alignments since sys spanset "
                    f"{sys_ssid} has no ref counterpart."
                ))

            pair_scores.arg_alignments[arg_field].measures.increment_false_alarm(1)
            if extras and arg_field == "agents":
                pair_scores.agent_match_measures.increment_false_alarm(1, tab=4)
            elif extras and arg_field == "patients":
                pair_scores.patient_match_measures.increment_false_alarm(1, tab=4)
            elif extras and arg_field == "money":
                pair_scores.money_match_measures.increment_false_alarm(1, tab=4)
            pair_scores.arg_alignments[arg_field].ssid_pairs.append([ref_ssid, sys_ssid])

        else:

            cached_match = match_cache[(ref_ssid, sys_ssid)]

            if cached_match > 0.0:

                if DEBUG:
                    LOGGER.debug((
                        f"            Adding 1 hit ({cached_match} match) to arg_alignments since "
                        f"sys spanset {sys_ssid} aligns with ref spanset {ref_ssid}."
                    ))

                pair_scores.arg_alignments[arg_field].measures.increment_match(cached_match, tab=4)
                pair_scores.arg_alignments[arg_field].measures.increment_hit(1, tab=4)
                if extras and arg_field == "agents":
                    pair_scores.agent_match_measures.increment_match(cached_match, tab=4)
                    pair_scores.agent_match_measures.increment_hit(1, tab=4)
                elif extras and arg_field == "patients":
                    pair_scores.patient_match_measures.increment_match(cached_match, tab=4)
                    pair_scores.patient_match_measures.increment_hit(1, tab=4)
                elif extras and arg_field == "money":
                    pair_scores.money_match_measures.increment_match(cached_match, tab=4)
                    pair_scores.money_match_measures.increment_hit(1, tab=4)
                pair_scores.arg_alignments[arg_field].ssid_pairs.append([ref_ssid, sys_ssid])

            else:

                if DEBUG:
                    LOGGER.debug((
                        f"            Adding 1 miss and 1 false alarm to arg_alignments since sys "
                        f"spanset {sys_ssid} does not align with ref spanset {ref_ssid}."
                    ))

                pair_scores.arg_alignments[arg_field].measures.increment_miss(1, tab=4)
                if extras and arg_field == "agents":
                    pair_scores.agent_match_measures.increment_miss(1, tab=4)
                elif extras and arg_field == "patients":
                    pair_scores.patient_match_measures.increment_miss(1, tab=4)
                elif extras and arg_field == "money":
                    pair_scores.money_match_measures.increment_miss(1, tab=4)
                pair_scores.arg_alignments[arg_field].ssid_pairs.append([ref_ssid, None])

                pair_scores.arg_alignments[arg_field].measures.increment_false_alarm(1, tab=4)
                if extras and arg_field == "agents":
                    pair_scores.agent_match_measures.increment_false_alarm(1, tab=4)
                elif extras and arg_field == "patients":
                    pair_scores.patient_match_measures.increment_false_alarm(1, tab=4)
                elif extras and arg_field == "money":
                    pair_scores.money_match_measures.increment_false_alarm(1, tab=4)
                pair_scores.arg_alignments[arg_field].ssid_pairs.append([None, sys_ssid])

            pair_scores.arg_alignments[arg_field].measures.compute_metrics(tab=4)
            if extras and arg_field == "agents":
                pair_scores.agent_match_measures.compute_metrics(tab=4)
            elif extras and arg_field == "patients":
                pair_scores.patient_match_measures.compute_metrics(tab=4)
            elif extras and arg_field == "money":
                pair_scores.money_match_measures.compute_metrics(tab=4)

    pair_scores.argmatch_measures.update(pair_scores.arg_alignments[arg_field].measures, tab=4)


def __score_basic_event_pair(
    ref_eid,
    sys_eid,
    entry_pair,
    event_pair_score_cache,
    ignore_synclass=False,
    extras=False
):

    from lib.utils import BASIC_MAT_KEY_ANCHOR_TYPES

    global DEBUG
    global SCORING_EVENT_TYPE_MATCH_TYPE
    global SCORING_REF_ARG_SCORING_TYPE

    if DEBUG:
        LOGGER.debug(f">>> Scoring Entry Pair: {entry_pair}")

    if (ref_eid, sys_eid) in event_pair_score_cache:

        if DEBUG:
            LOGGER.debug((
                "        Pair already in score cache! Score: "
                f"{event_pair_score_cache[(ref_eid, sys_eid)].combined_score}"
                "        Value: "
                f"{event_pair_score_cache[(ref_eid, sys_eid)].combined_measures.value}"
            ))
        return event_pair_score_cache[(ref_eid, sys_eid)].combined_measures.value

    pair_scores = ScoreStruct()
    for evtype in BASIC_MAT_KEY_ANCHOR_TYPES:
        pair_scores.evtype_match_measures[evtype] = Measures(evtype)

    if not sys_eid:

        if DEBUG:
            LOGGER.debug((
                "        Adding 1 miss to evt_measures since there's no system entry in the score "
                "cache."
            ))
        pair_scores.event_measures.increment_miss(1, tab=3)

        if extras:
            ref_event = entry_pair.ref_events[ref_eid]
            pair_scores.evtype_match_measures[ref_event.event_type.value].increment_miss(1, tab=3)

    elif not ref_eid:

        if DEBUG:
            LOGGER.debug((
                "        Adding 1 false alarm to evt_measures since there's no reference entry in "
                "the score cache."
            ))
        pair_scores.event_measures.increment_false_alarm(1, tab=3)

        if extras:
            sys_event = entry_pair.sys_events[sys_eid]
            pair_scores.evtype_match_measures[sys_event.event_type.value].increment_false_alarm(
                1, tab=3
            )

    else:

        sys_event = entry_pair.sys_events[sys_eid]
        ref_event = entry_pair.ref_events[ref_eid]

        evtype = ref_event.event_type.value

        if SCORING_EVENT_TYPE_MATCH_TYPE == 'strict':

            if ref_event.event_type.value and sys_event.event_type.value and \
                ref_event.event_type.value == sys_event.event_type.value:

                if DEBUG:
                    LOGGER.debug((
                        f"        [{SCORING_EVENT_TYPE_MATCH_TYPE}] "
                        "Adding 1 hit (1.0 match) to evt_measures since both ET are the same and "
                        f"not None."
                    ))
                pair_scores.event_measures.increment_match(1.0, tab=3)
                pair_scores.event_measures.increment_hit(1, tab=3)

                if extras:
                    pair_scores.evtype_match_measures[evtype].increment_match(1.0, tab=3)
                    pair_scores.evtype_match_measures[evtype].increment_hit(1, tab=3)

                __score_arguments(
                    entry_pair,
                    (ref_eid, sys_eid,),
                    pair_scores,
                    "agents",
                    ignore_synclass=ignore_synclass,
                    extras=extras
                )
                __score_arguments(
                    entry_pair,
                    (ref_eid, sys_eid,),
                    pair_scores,
                    "patients",
                    ignore_synclass=ignore_synclass,
                    extras=extras
                )
                __score_arguments(
                    entry_pair,
                    (ref_eid, sys_eid,),
                    pair_scores,
                    "money",
                    ignore_synclass=ignore_synclass,
                    extras=extras
                )

                if SCORING_REF_ARG_SCORING_TYPE == "anchors-as-string-args":
                    __score_arguments(
                        entry_pair,
                        (ref_eid, sys_eid,),
                        pair_scores,
                        "ref_events",
                        ignore_synclass=ignore_synclass
                    )

            else:

                # An event with this specific reference event-type wasn't generated by system
                if DEBUG:
                    LOGGER.debug((
                        f"        [{SCORING_EVENT_TYPE_MATCH_TYPE}] "
                        "Adding 1 miss to evt_measures since the ref ET is missing in sys."
                    ))
                pair_scores.event_measures.increment_miss(1, tab=3)

                __score_arguments(
                    entry_pair,
                    (ref_eid, None,),
                    pair_scores,
                    "agents",
                    ignore_synclass=ignore_synclass,
                    extras=extras
                )
                __score_arguments(
                    entry_pair,
                    (ref_eid, None,),
                    pair_scores,
                    "patients",
                    ignore_synclass=ignore_synclass,
                    extras=extras
                )
                __score_arguments(
                    entry_pair,
                    (ref_eid, None,),
                    pair_scores,
                    "money",
                    ignore_synclass=ignore_synclass,
                    extras=extras
                )

                ## score ref_events

                if SCORING_REF_ARG_SCORING_TYPE == "anchors-as-string-args":
                    __score_arguments(
                        entry_pair,
                        (ref_eid, None,),
                        pair_scores,
                        "ref_events",
                        ignore_synclass=ignore_synclass
                    )

                # An event with this specific system event-type had no counterpart in reference
                if DEBUG:
                    LOGGER.debug((
                        f"        [{SCORING_EVENT_TYPE_MATCH_TYPE}] "
                        "Adding 1 false alarm to evt_measures since the sys ET is missing in ref."
                    ))
                pair_scores.event_measures.increment_false_alarm(1, tab=3)

                __score_arguments(
                    entry_pair,
                    (None, sys_eid,),
                    pair_scores,
                    "agents",
                    ignore_synclass=ignore_synclass,
                    extras=extras
                )
                __score_arguments(
                    entry_pair,
                    (None, sys_eid,),
                    pair_scores,
                    "patients",
                    ignore_synclass=ignore_synclass,
                    extras=extras
                )
                __score_arguments(
                    entry_pair,
                    (None, sys_eid,),
                    pair_scores,
                    "money",
                    ignore_synclass=ignore_synclass,
                    extras=extras
                )

                if SCORING_REF_ARG_SCORING_TYPE == "anchors-as-string-args":
                    __score_arguments(
                        entry_pair,
                        (None, sys_eid,),
                        pair_scores,
                        "ref_events",
                        ignore_synclass=ignore_synclass
                    )

                if extras:
                    pair_scores.evtype_match_measures[
                        ref_event.event_type.value
                    ].increment_miss(1, tab=3)

                    try:
                        pair_scores.evtype_match_measures[
                            sys_event.event_type.value
                        ]
                    except KeyError as err:
                        if VERBOSE:
                            LOGGER.error((
                                "Found an invalid Event Type in System Event: "
                                f"{sys_event.event_type.value}"
                            ))
                            LOGGER.error(str(err), exc_info=True)
                        else:
                            click.echo(str(err), err=True)
                    pair_scores.evtype_match_measures[
                        sys_event.event_type.value
                    ].increment_false_alarm(1, tab=3)

        elif SCORING_EVENT_TYPE_MATCH_TYPE == 'loose':

            if (ref_event.event_type.value == None and sys_event.event_type.value == None) or \
                (ref_event.event_type.value and sys_event.event_type.value and \
                ref_event.event_type.value == sys_event.event_type.value):
                if DEBUG:
                    LOGGER.debug((
                        f"        [{SCORING_EVENT_TYPE_MATCH_TYPE}] "
                        "EventType 1.0 match since both values match and the values aren't None."
                    ))
                match = 1.0

            else:

                if DEBUG:
                    LOGGER.debug((
                        f"        [{SCORING_EVENT_TYPE_MATCH_TYPE}] "
                        "EventType 0.0 match since the values don't match, or the both values"
                        " are None."
                    ))
                match = 0.0

            if DEBUG:
                LOGGER.debug((
                    f"        [{SCORING_EVENT_TYPE_MATCH_TYPE}] "
                    f"Adding match value {match} because of EventType match to evt_measures."
                ))
                LOGGER.debug((
                    f"        [{SCORING_EVENT_TYPE_MATCH_TYPE}] "
                    f"Adding 1 hit to evt_measures."
                ))

            # There was a fractional match
            pair_scores.event_measures.increment_match(match, tab=3)

            # Out of this total possible hit count
            pair_scores.event_measures.increment_hit(1, tab=3)

            __score_arguments(
                entry_pair,
                (ref_eid, sys_eid,),
                pair_scores,
                "agents",
                ignore_synclass=ignore_synclass,
                extras=extras
            )
            __score_arguments(
                entry_pair,
                (ref_eid, sys_eid,),
                pair_scores,
                "patients",
                ignore_synclass=ignore_synclass,
                extras=extras
            )
            __score_arguments(
                entry_pair,
                (ref_eid, sys_eid,),
                pair_scores,
                "money",
                ignore_synclass=ignore_synclass,
                extras=extras
            )

            if SCORING_REF_ARG_SCORING_TYPE == "anchors-as-string-args":
                __score_arguments(
                    entry_pair,
                    (ref_eid, sys_eid,),
                    pair_scores,
                    "ref_events",
                    ignore_synclass=ignore_synclass
                )

            if extras:
                pair_scores.evtype_match_measures[evtype].increment_match(match, tab=3)
                pair_scores.evtype_match_measures[evtype].increment_hit(1, tab=3)

        elif SCORING_EVENT_TYPE_MATCH_TYPE == 'ignore':

            # We give credit without consulting actual event type values
            if DEBUG:
                LOGGER.debug((
                    f"        [{SCORING_EVENT_TYPE_MATCH_TYPE}] "
                    f"Adding 1.0 match to evt_measures."
                ))
                LOGGER.debug((
                    f"        [{SCORING_EVENT_TYPE_MATCH_TYPE}] "
                    f"Adding 1 hit to evt_measures."
                ))
            pair_scores.event_measures.increment_match(1.0, tab=3)
            pair_scores.event_measures.increment_hit(1, tab=3)

            if extras:
                pair_scores.evtype_match_measures[evtype].increment_match(1.0, tab=3)
                pair_scores.evtype_match_measures[evtype].increment_hit(1, tab=3)

            __score_arguments(
                entry_pair,
                (ref_eid, sys_eid,),
                pair_scores,
                "agents",
                ignore_synclass=ignore_synclass,
                extras=extras
            )
            __score_arguments(
                entry_pair,
                (ref_eid, sys_eid,),
                pair_scores,
                "patients",
                ignore_synclass=ignore_synclass,
                extras=extras
            )
            __score_arguments(
                entry_pair,
                (ref_eid, sys_eid,),
                pair_scores,
                "money",
                ignore_synclass=ignore_synclass,
                extras=extras
            )

            ## score ref_events

            if SCORING_REF_ARG_SCORING_TYPE == "anchors-as-string-args":
                __score_arguments(
                    entry_pair,
                    (ref_eid, sys_eid,),
                    pair_scores,
                    "ref_events",
                    ignore_synclass=ignore_synclass
                )

    ## Testing a counting approach

    pair_scores.compute_score()
    event_pair_score_cache[(ref_eid, sys_eid)] = pair_scores
    ## return pair_scores.combined_score
    return pair_scores.combined_measures.value


def __score_gran_template_pair(
    ref_tid,
    sys_tid,
    entry_pair,
    template_pair_score_cache,
    extras=False
):
    from lib.utils import GRANULAR_MAT_TEMPLATE_TYPES
    from lib.utils import SCORING_IGNORED_SLOTS

    global DEBUG

    if DEBUG:
        LOGGER.debug(f">>> Scoring Ref '{ref_tid}' to Sys '{sys_tid}'")
        LOGGER.debug(f"    Loaded Basic event alignments: {entry_pair.scores.event_alignment}")

    if (ref_tid, sys_tid) in template_pair_score_cache:

        if DEBUG:
            LOGGER.debug((
                "    Pair already in score cache! Score: "
                f"{template_pair_score_cache[(ref_tid, sys_tid)].combined_score}"
                "    Value: "
                f"{template_pair_score_cache[(ref_tid, sys_tid)].combined_measures.value}"
            ))
        return template_pair_score_cache[(ref_tid, sys_tid)].combined_measures.value

    pair_scores = GranScoreStruct()
    for ttype in GRANULAR_MAT_TEMPLATE_TYPES:
        pair_scores.ttype_match_measures[ttype] = Measures(ttype)

    if not ref_tid and not sys_tid:

        if DEBUG:
            LOGGER.debug("    Got 'None' on both sides. Skipping.")

        return 0.0

    elif not sys_tid:

        if DEBUG:
            LOGGER.debug(
                "    Adding 1 miss to template_measures since there's no system template."
            )

        m = GranScoreStruct()
        m.template_measures.increment_miss(1, tab=2)
        ref_template = entry_pair.ref_templates[ref_tid]
        for _, slot_label in GRANULAR_MAT_TEMPLATE_TYPES[ref_template.type_.value].items():
            slot = getattr(ref_template, slot_label).value
            if slot and slot_label not in SCORING_IGNORED_SLOTS:
                if DEBUG:
                    LOGGER.debug(f"        Adding misses for slot {slot_label}: {slot}")
                if type(slot) == list:
                    m.slotmatch_measures.miss += len(slot)
                else:
                    m.slotmatch_measures.miss += 1

        pair_scores.update(m)

        if extras:
            ref_template = entry_pair.ref_templates[ref_tid]
            pair_scores.ttype_match_measures[ref_template.type_.value].increment_miss(
                1, tab=2
            )

        template_pair_score_cache[(ref_tid, sys_tid)] = pair_scores
        return 0.0

    elif not ref_tid:

        if DEBUG:
            LOGGER.debug((
                "    Adding 1 false alarm to template_measures since there's no reference template."
            ))

        fa = GranScoreStruct()
        fa.template_measures.increment_false_alarm(1, tab=2)
        sys_template = entry_pair.sys_templates[sys_tid]
        for _, slot_label in GRANULAR_MAT_TEMPLATE_TYPES[sys_template.type_.value].items():
            slot = getattr(sys_template, slot_label).value
            if slot and slot_label not in SCORING_IGNORED_SLOTS:
                if DEBUG:
                    LOGGER.debug(f"        Adding false-alarms for slot {slot_label}: {slot}")
                if type(slot) == list:
                    fa.slotmatch_measures.false_alarm += len(slot)
                else:
                    fa.slotmatch_measures.false_alarm += 1

        pair_scores.update(fa)

        if extras:
            sys_template = entry_pair.sys_templates[sys_tid]
            pair_scores.ttype_match_measures[
                sys_template.type_.value
            ].increment_false_alarm(1, tab=2)

        template_pair_score_cache[(ref_tid, sys_tid)] = pair_scores
        return 0.0

    else:

        sys_template = entry_pair.sys_templates[sys_tid]
        ref_template = entry_pair.ref_templates[ref_tid]

        assert ref_template.type_.value == sys_template.type_.value, "Got template type mismatch!"

        ttype = ref_template.type_.value

        if DEBUG:
            LOGGER.debug(("    "
                "Adding 1 hit (1.0 match) to template_measures since both are the same type "
                f"('{ref_template.type_.value}') and not None."
            ))
        pair_scores.template_measures.increment_match(1.0, tab=2)
        pair_scores.template_measures.increment_hit(1, tab=2)

        for _, slot_label in GRANULAR_MAT_TEMPLATE_TYPES[ttype].items():
            if slot_label not in SCORING_IGNORED_SLOTS:
                __score_slot(
                    entry_pair,
                    (ref_tid, sys_tid,),
                    pair_scores,
                    slot_label,
                    extras=extras
                )

        if extras:
            pair_scores.ttype_match_measures[ttype].increment_match(1.0, tab=2)
            pair_scores.ttype_match_measures[ttype].increment_hit(1, tab=2)

        pair_scores.compute_score()
        template_pair_score_cache[(ref_tid, sys_tid)] = pair_scores
        return pair_scores.combined_measures.value


def __score_slot(entry_pair, ref_sys_pair, pair_scores, slot_label, extras=False):

    ref_tid = ref_sys_pair[0]
    sys_tid = ref_sys_pair[1]
    ref_slot = None
    sys_slot = None
    ref_template = None
    sys_template = None

    if DEBUG:
        LOGGER.debug((
            f">>>>>>> Scoring {slot_label} values for Entry Pair: {entry_pair}, ref TemplID: "
            f"{ref_tid}, sys TemplID: {sys_tid} "
        ))
        eqcls = entry_pair.ref_entry.get_event_set('Basic').template_filler_coref_events.value or []
        LOGGER.debug(f"        Loaded equivalence classes: {eqcls}")

    if ref_tid in entry_pair.ref_templates:
        ref_template = entry_pair.ref_templates[ref_tid]
        ref_slot_element = getattr(ref_template, slot_label)
        ref_slot = ref_slot_element.value
        ref_slot_type = ref_slot_element.type_
    else:
        ref_template = None
        ref_slot_element = None
        ref_slot = None
        ref_slot_type = None

    if sys_tid in entry_pair.sys_templates:
        sys_template = entry_pair.sys_templates[sys_tid]
        sys_slot_element = getattr(sys_template, slot_label)
        sys_slot = sys_slot_element.value
        sys_slot_type = sys_slot_element.type_
    else:
        sys_template = None
        sys_slot_element = None
        sys_slot = None
        sys_slot_type = None

    if DEBUG:
        LOGGER.debug(f"        ref_template: {ref_template}")
        LOGGER.debug(f"        ref_slot: {ref_slot}")
        LOGGER.debug(f"        sys_template: {sys_template}")
        LOGGER.debug(f"        sys_slot: {sys_slot}")

    pair_scores.slot_alignments[slot_label] = SlotAlignments(slot_label)

    if not sys_template or sys_slot == None or sys_slot == []:

        if not ref_template or ref_slot == None or ref_slot == []:

            # Giving no positive value for the correct "decision" to create zero arguments
            if DEBUG:
                LOGGER.debug("        No sys or ref slot elements found! Nothing to do.")

        else:

            if DEBUG:
                LOGGER.debug((
                    f"        No sys slot element found, but {ref_slot} ref slot element found! "
                    f"Adding misses to slot alignments."
                ))

            if ref_slot_type == list:
                for tprop in ref_slot:
                    if tprop.get_type() == "spanset":
                        pair_scores.slot_alignments[slot_label].pairs.append(
                            [tprop.ssid.value, None]
                        )
                    elif tprop.get_type() == "event":
                        pair_scores.slot_alignments[slot_label].pairs.append(
                            [tprop.event_id.value, None]
                        )
                    else:
                        pair_scores.slot_alignments[slot_label].pairs.append([tprop, None])
                    pair_scores.slot_alignments[slot_label].measures.increment_miss(1, tab=3)
            else:
                pair_scores.slot_alignments[slot_label].pairs.append([ref_slot, None])
                pair_scores.slot_alignments[slot_label].measures.increment_miss(1, tab=3)

    elif not ref_template or ref_slot == None or ref_slot == []:

        if sys_template and sys_slot != None and sys_slot != []:

            if DEBUG:
                LOGGER.debug((
                    f"        No ref slot element found, but {sys_slot} sys slot element found! "
                    f"Adding false alarms to slot alignments."
                ))

            if sys_slot_type == list:
                for tprop in sys_slot:
                    if tprop.get_type() == "spanset":
                        pair_scores.slot_alignments[slot_label].pairs.append(
                            [tprop.ssid.value, None]
                        )
                    elif tprop.get_type() == "event":
                        pair_scores.slot_alignments[slot_label].pairs.append(
                            [tprop.event_id.value, None]
                        )
                    else:
                        pair_scores.slot_alignments[slot_label].pairs.append([tprop, None])
                    pair_scores.slot_alignments[slot_label].measures.increment_false_alarm(1, tab=3)
            else:
                pair_scores.slot_alignments[slot_label].pairs.append([None, sys_slot])
                pair_scores.slot_alignments[slot_label].measures.increment_false_alarm(1, tab=3)

    elif ref_slot_type != sys_slot_type:

        if DEBUG:
            LOGGER.debug((
                f"        Ref slot type {ref_slot_type} != sys_slot_type {sys_slot_type}! "
                f"Adding misses false alarms to slot alignments."
            ))

        if ref_slot_type == list:
            for ref_element in ref_slot:
                pair_scores.slot_alignments[slot_label].pairs.append([ref_element, None])
                pair_scores.slot_alignments[slot_label].measures.increment_miss(1, tab=3)
        else:
            pair_scores.slot_alignments[slot_label].pairs.append([ref_slot, None])
            pair_scores.slot_alignments[slot_label].measures.increment_miss(1, tab=3)

        if sys_slot_type == list:
            for sys_element in sys_slot:
                pair_scores.slot_alignments[slot_label].pairs.append([None, sys_element])
                pair_scores.slot_alignments[slot_label].measures.increment_false_alarm(1, tab=3)
        else:
            pair_scores.slot_alignments[slot_label].pairs.append([None, sys_slot])
            pair_scores.slot_alignments[slot_label].measures.increment_false_alarm(1, tab=3)

    elif ref_slot_type == list:

        # Can be a mixture of SSIDs and EventIDs. Parse them out, and run munkres on groups of
        # span-sets and events only. Each element of the list is a
        # ``BPDocumentGranularTemplateProperty``.

        if DEBUG:
            LOGGER.debug("        Slot is a list of Entities or Events.")

        ref_entities = []
        sys_entities = []
        ref_events = []
        sys_events = []

        for tprop in ref_slot:
            if tprop.get_type() == "both" and DEBUG:
                LOGGER.debug(
                    f"            Found reference TemplateProperty of type 'both': {tprop}"
                )
            elif tprop.get_type() == "spanset":
                ref_entities.append(tprop)
            elif tprop.get_type() == "event":
                ref_events.append(tprop)
            elif DEBUG:
                LOGGER.debug(
                    f"            Found reference TemplateProperty without a type: {tprop}"
                )

        for tprop in sys_slot:
            if tprop.get_type() == "both" and DEBUG:
                LOGGER.debug(f"            Found system TemplateProperty of type 'both': {tprop}")
            elif tprop.get_type() == "spanset":
                sys_entities.append(tprop)
            elif tprop.get_type() == "event":
                sys_events.append(tprop)
            elif DEBUG:
                LOGGER.debug(f"            Found system TemplateProperty without a type: {tprop}")

        __score_slot_events(ref_events, sys_events, entry_pair, pair_scores, slot_label)
        __score_slot_entities(ref_entities, sys_entities, entry_pair, pair_scores, slot_label)

    else:

        if DEBUG:
            LOGGER.debug("        Slot is a primitive value.")

        if ref_slot_type == str:

            # we allow case-insensitivity when comparing string values
            ref_slot = ref_slot.lower()
            sys_slot = sys_slot.lower()

        if ref_slot == sys_slot:
            pair_scores.slot_alignments[slot_label].measures.increment_match(1.0, tab=3)
            pair_scores.slot_alignments[slot_label].measures.increment_hit(1, tab=3)
            pair_scores.slot_alignments[slot_label].pairs.append([ref_slot, sys_slot])
        elif type(ref_slot) == type(sys_slot) and ref_slot != sys_slot:
            pair_scores.slot_alignments[slot_label].measures.increment_match(0.0, tab=3)
            pair_scores.slot_alignments[slot_label].measures.increment_hit(1, tab=3)
            pair_scores.slot_alignments[slot_label].pairs.append([ref_slot, sys_slot])
        else:
            pair_scores.slot_alignments[slot_label].measures.increment_miss(1, tab=3)
            pair_scores.slot_alignments[slot_label].measures.increment_false_alarm(1, tab=3)
            pair_scores.slot_alignments[slot_label].pairs.append([ref_slot, None])
            pair_scores.slot_alignments[slot_label].pairs.append([None, sys_slot])

    if DEBUG:
        LOGGER.debug(
            f"        Final slot measures: {pair_scores.slot_alignments[slot_label].measures}"
        )
    pair_scores.slotmatch_measures.update(
        pair_scores.slot_alignments[slot_label].measures, tab=2
    )
    pair_scores.compute_score()


def __score_slot_entities(ref_entities, sys_entities, entry_pair, pair_scores, slot_label):
    """Scores Entities (SpanSets) within a Template slot.

    Args:
        ref_entities (list): of ``lib.bp.BPDocumentGranularTemplateProperty``, each referencing a
            SpanSet from the reference data.
        sys_entities (list): of ``lib.bp.BPDocumentGranularTemplateProperty``, each referencing a
            SpanSet from the system data.
        entry_pair (EntryPair): the object wrapping both the reference and system entries.
        pair_scores (GranScoreStruct): contains scoring information for the entry pair.
        slot_label (str): the name of the template slot.
    """

    import itertools

    from munkres import make_cost_matrix
    from munkres import Munkres

    global DEBUG
    global LOGGER
    global VERBOSE

    if DEBUG:
        LOGGER.debug("            Starting Munkres alignment of Entities within slot.")
        LOGGER.debug(("            Subscores before Entity alignment: "
            f"{pair_scores.slot_alignments[slot_label].measures}"
        ))
        LOGGER.debug(f"            Reference entities: {ref_entities}")
        LOGGER.debug(f"            System entities: {sys_entities}")

    if len(ref_entities) < len(sys_entities):

        if DEBUG:
            LOGGER.debug((
                f"            Padding ref ssids with {len(sys_entities) - len(ref_entities)} blanks"
                f" since sys_entities is longer."
            ))

        # pad with `None`s
        ref_entities = ref_entities + [
            x for x in itertools.repeat(None, len(sys_entities)-len(ref_entities))
        ]

    elif len(sys_entities) < len(ref_entities):

        if DEBUG:
            LOGGER.debug((
                f"            Padding sys ssids with {len(ref_entities) - len(sys_entities)} blanks"
                f" since ref_entities is longer."
            ))

        # pad with `None`s
        sys_entities = sys_entities + [
            x for x in itertools.repeat(None, len(ref_entities)-len(sys_entities))
        ]

    sys_ssids = []
    for entity in sys_entities:
        if entity:
            sys_ssids.append(entity.ssid.value)
        else:
            sys_ssids.append(None)

    ref_ssids = []
    for entity in ref_entities:
        if entity:
            ref_ssids.append(entity.ssid.value)
        else:
            ref_ssids.append(None)

    match_cache = {}
    matrix = make_cost_matrix(
        [
            [
                __score_ss_pair(
                    sys_ssid,
                    ref_ssid,
                    entry_pair,
                    match_cache,
                    ignore_synclass=True
                ) for sys_ssid in sys_ssids
            ] for ref_ssid in ref_ssids
        ],
        lambda cost: 1.0 - cost
    )
    if matrix:
        index_pairs = Munkres().compute(matrix)
    else:
        index_pairs = []

    if DEBUG:
        LOGGER.debug("            Evaluating Munkres alignments.")

    for row, column in index_pairs:

        ref_entity = ref_entities[row]
        sys_entity = sys_entities[column]

        if not sys_entity and not ref_entity:

            if DEBUG:
                LOGGER.debug(f"                Both sys and ref ssids are blank; skipping.")

        elif not sys_entity:

            if DEBUG:
                LOGGER.debug(("                "
                    f"Adding 1 miss to slot_alignments since ref spanset {ref_entity} "
                    "has no sys counterpart."
                ))

            pair_scores.slot_alignments[slot_label].measures.increment_miss(1, tab=5)
            pair_scores.slot_alignments[slot_label].pairs.append(
                [ref_entity.ssid.value, None]
            )
            if ref_entity.irrealis.value:
                pair_scores.irrealis_measures.increment_miss(1, tab=5)
            if ref_entity.time_attachments.value:
                pair_scores.time_attachment_measures.increment_miss(
                    len(ref_entity.time_attachments.value), tab=5
                )

        elif not ref_entity:
            if DEBUG:
                LOGGER.debug((
                    "                Adding 1 false alarm to slot_alignments since sys spanset "
                    f"{sys_entity} has no ref counterpart."
                ))

            pair_scores.slot_alignments[slot_label].measures.increment_false_alarm(1, tab=5)
            pair_scores.slot_alignments[slot_label].pairs.append(
                [None, sys_entity.ssid.value]
            )
            if sys_entity.irrealis.value:
                pair_scores.irrealis_measures.increment_false_alarm(1, tab=5)
            if sys_entity.time_attachments.value:
                pair_scores.time_attachment_measures.increment_false_alarm(
                    len(sys_entity.time_attachments.value), tab=5
                )

        else:

            ss_score = match_cache[(ref_entity.ssid.value, sys_entity.ssid.value)]
            irrealis =  __score_slot_irrealis(ref_entity, sys_entity, entry_pair)
            ta = __score_slot_timeatt(ref_entity, sys_entity, entry_pair)

            if ss_score == 0:

                # Don't weight irrealis and time-attachments unless the spans actually match, other-
                # wise spans with different strings could score at 0.25.
                if DEBUG:
                    LOGGER.debug((
                        "                Adding 1 false alarm and 1 miss to slot_alignments since "
                        f"sys spanset {sys_entity} and ref spanset {ref_entity} don't align."
                    ))

                pair_scores.slot_alignments[slot_label].measures.increment_miss(1, tab=5)
                pair_scores.slot_alignments[slot_label].measures.increment_false_alarm(1, tab=5)
                pair_scores.slot_alignments[slot_label].pairs.append(
                    [ref_entity.ssid.value, None]
                )
                pair_scores.slot_alignments[slot_label].pairs.append(
                    [None, sys_entity.ssid.value]
                )

            else:

                match_score = (ss_score * 0.75) + irrealis + ta

                if DEBUG:
                    LOGGER.debug((f"                Reference entity '{ref_entity.ssid.value}' "
                        f"aligns with System entity '{sys_entity.ssid.value}'."
                    ))
                    LOGGER.debug(f"                    SpanSet Alignment Score: {ss_score}")
                    LOGGER.debug(f"                    Irrealis: {irrealis}")
                    LOGGER.debug(f"                    Time-Attachments Alignment Score: {ta}")
                    LOGGER.debug((
                        f"                Adding 1 hit ({match_score} match) to slot_alignments."
                    ))

                pair_scores.slot_alignments[slot_label].measures.increment_match(match_score, tab=5)
                pair_scores.slot_alignments[slot_label].measures.increment_hit(1, tab=5)
                pair_scores.slot_alignments[slot_label].pairs.append(
                    [ref_entity.ssid.value, sys_entity.ssid.value]
                )

    if DEBUG:
        LOGGER.debug(("            Subscores after Entity alignment: "
            f"{pair_scores.slot_alignments[slot_label].measures}"
        ))


def __score_slot_events(ref_events, sys_events, entry_pair, pair_scores, slot_label):
    """Scores Entities (spansets) within a Template slot.

    Args:
        ref_events (list): of ``lib.bp.BPDocumentGranularTemplateProperty``, each referencing an
            Event from the reference data.
        sys_events (list): of ``lib.bp.BPDocumentGranularTemplateProperty``, each referencing an
            Event from the system data.
        entry_pair (EntryPair): the object wrapping both the reference and system entries.
        pair_scores (GranScoreStruct): contains scoring information for the entry pair.
        slot_label (str): the name of the template slot.
    """
    from lib.utils import BASIC_MAT_IGNORED_ANCHOR_TYPES

    global DEBUG
    global LOGGER
    global VERBOSE

    def __normalize(x):
        if type(x) == int:
            return f"Equivalence Class at Index {x}"
        else:
            return x.event_id.value

    equivalence_classes = \
        entry_pair.ref_entry.get_event_set("Basic").template_filler_coref_events.value or []

    # generated from running the Basic scorer.
    # shape: list of lists, outer[0] is list of ref evids, outer[1] is list of sys evids, matching
    # inner indices indicate alignment.
    alignments = entry_pair.scores.event_alignment

    assert len(alignments[0]) == len(alignments[1]), \
        "Uneven Basic event alignments!"

    # replaces event-ids with indices of equivalence classes if event belongs to an EqCl
    expanded_ref_events = []
    for ref_event in ref_events:

        found = []
        for i in range(len(equivalence_classes)):
            if ref_event.event_id.value in equivalence_classes[i]:
                found.append(i)

        assert len(found) <= 1, (
            f"Event {ref_event.event_id.value} belongs to multiple equivalence classes!\n"
            f">>> Possible classes: {equivalence_classes}\n"
            f">>> Detected Indices: {found}"
        )

        if found:
            expanded_ref_events.append(found[0])
        else:
            expanded_ref_events.append(ref_event)

    if DEBUG:
        LOGGER.debug(f"            Ref Events within slot: {expanded_ref_events}")
        LOGGER.debug(f"            Sys Events within slot: {sys_events}")

    mapped_ref_events = []
    mapped_sys_events = []

    for ref_event in expanded_ref_events:

        if ref_event in mapped_ref_events:
            continue

        # if the ref_event does not belong to an EqCl
        if type(ref_event) != int:

            if DEBUG:
                LOGGER.debug(
                    f"            ref event {ref_event} doesn't belong to an equivalence class."
                )

            try:

                index = alignments[0].index(ref_event.event_id.value)

            except ValueError:

                subject = entry_pair.ref_events[ref_event.event_id.value]
                if subject.event_type.value in BASIC_MAT_IGNORED_ANCHOR_TYPES:

                    LOGGER.warning(
                        f"            ref event {ref_event} missing from entrypair alignments "
                        f"because it's an ignored type: {subject.event_type.value}"
                    )
                    mapped_ref_events.append(ref_event)
                    continue

                else:

                    raise ValueError(
                        f"ref event {ref_event.event_id.value} missing from entrypair alignments! "
                        f"Is it not within sentence bounds in {entry_pair}?"
                    )

            if DEBUG:
                LOGGER.debug((
                    f"                Got alignment index {index}, which refers to system event "
                    f"{alignments[1][index]}."
                ))

            sys_counterpart = None
            for sys_event in sys_events:
                if alignments[1][index] == sys_event.event_id.value:
                    sys_counterpart = sys_event
                    break

            if sys_counterpart:

                if DEBUG:
                    LOGGER.debug(f"                Found system counterpart: {sys_counterpart}")

                match_score = 0.75 +  __score_slot_irrealis(ref_event, sys_counterpart, entry_pair)
                match_score += __score_slot_timeatt(ref_event, sys_counterpart, entry_pair)

                pair_scores.slot_alignments[slot_label].measures.increment_match(match_score, tab=4)
                pair_scores.slot_alignments[slot_label].measures.increment_hit(1, tab=4)
                pair_scores.slot_alignments[slot_label].pairs.append(
                    [__normalize(ref_event), __normalize(sys_counterpart)]
                )
                mapped_sys_events.append(sys_counterpart)

            else:

                if DEBUG:
                    LOGGER.debug(f"                System counterpart not found in system slot!")

                pair_scores.slot_alignments[slot_label].measures.increment_miss(1, tab=4)
                pair_scores.slot_alignments[slot_label].pairs.append([__normalize(ref_event), None])

                if ref_event.irrealis.value:
                    pair_scores.irrealis_measures.increment_miss(1, tab=5)
                if ref_event.time_attachments.value:
                    pair_scores.time_attachment_measures.increment_miss(
                        len(ref_event.time_attachments.value), tab=5
                    )

        # if the ref_event _is_ an EqCl
        else:

            if DEBUG:
                LOGGER.debug("            "
                    f"Processing equivalence class {ref_event}:"
                )

            best_score = 0
            best_counterpart = None
            all_counterparts = []
            eqcl_event_types = []
            for ref_evid in equivalence_classes[ref_event]:

                subject = entry_pair.ref_events[ref_evid]
                eqcl_event_types.append(subject.event_type.value)

                try:

                    index = alignments[0].index(ref_evid)

                except ValueError:

                    if subject.event_type.value in BASIC_MAT_IGNORED_ANCHOR_TYPES:

                        LOGGER.warning(
                            f"            ref event {ref_evid} in equivalence class {ref_event} "
                            "missing from entrypair alignments because it's an ignored type: "
                            f"{subject.event_type.value}"
                        )
                        continue

                    else:

                        raise ValueError(
                            f"ref event {ref_evid} in equivalence class {ref_event} missing from "
                            f"entrypair alignments!"
                        )

                sys_counterpart = None
                for sys_event in sys_events:
                    if alignments[1][index] == sys_event.event_id.value:
                        sys_counterpart = sys_event
                        break

                if sys_counterpart:

                    if DEBUG:
                        LOGGER.debug(f"                Found system counterpart: {sys_counterpart}")

                    match_score = 0.75 +  __score_slot_irrealis(
                        ref_events[expanded_ref_events.index(ref_event)],
                        sys_counterpart, entry_pair
                    )
                    match_score += __score_slot_timeatt(
                        ref_events[expanded_ref_events.index(ref_event)],
                        sys_counterpart, entry_pair
                    )
                    all_counterparts.append(sys_counterpart)

                    if match_score > best_score:
                        best_score = match_score
                        best_counterpart = sys_counterpart

                else:

                    if DEBUG:
                        LOGGER.debug(f"                No system counterpart found in system slot!")

            if best_counterpart:

                pair_scores.slot_alignments[slot_label].measures.increment_match(best_score, tab=4)
                pair_scores.slot_alignments[slot_label].measures.increment_hit(1, tab=4)
                pair_scores.slot_alignments[slot_label].pairs.append(
                    [__normalize(ref_event), __normalize(best_counterpart)]
                )
                mapped_sys_events.extend(all_counterparts)

            else:

                # only increment miss for an equivalence class if the type of events in the class
                # were not ignored.

                if not all([type_ in BASIC_MAT_IGNORED_ANCHOR_TYPES for type_ in eqcl_event_types]):

                    pair_scores.slot_alignments[slot_label].measures.increment_miss(1, tab=4)
                    pair_scores.slot_alignments[slot_label].pairs.append(
                        [__normalize(ref_event), None]
                    )

                    # find the ref event in the eqcl that hurts the composite score the _least_
                    cheapest_ir = 0
                    cheapest_ta = float('inf')
                    for evid in equivalence_classes[ref_event]:

                        event = None
                        for x in ref_events:
                            if x.event_id == evid:
                                event = x

                        if not event:
                            continue

                        if event.irrealis.value:
                            cheapest_ir = 1 # there's only ever 1
                        if event.time_attachments.value:
                            x = len(equivalence_classes[ref_event].time_attachments.value)
                            if x < cheapest_ta:
                                cheapest_ta = x

                    if cheapest_ta == float('inf'):
                        cheapest_ta = 0
                    pair_scores.irrealis_measures.increment_miss(cheapest_ir, tab=5)
                    pair_scores.time_attachment_measures.increment_miss(cheapest_ta, tab=5)

        mapped_ref_events.append(ref_event)

    ## produce misses from leftover unmapped reference events

    if DEBUG:
        LOGGER.debug("            Producing misses for all unmapped Reference slot elements.")

    for ref_event in [event for event in expanded_ref_events if event not in mapped_ref_events]:

        pair_scores.slot_alignments[slot_label].measures.increment_miss(1, tab=4)
        pair_scores.slot_alignments[slot_label].pairs.append([__normalize(ref_event), None])
        if ref_event.irrealis.value:
            pair_scores.irrealis_measures.increment_miss(1, tab=5)
        if ref_event.time_attachments.value:
            pair_scores.time_attachment_measures.increment_miss(
                len(ref_event.time_attachments.value), tab=5
            )

    ## produce false alarms from leftover unmapped system events

    if DEBUG:
        LOGGER.debug(
            "            Producing false alarms for all unmapped System slot elements."
        )

    for sys_event in [event for event in sys_events if event not in mapped_sys_events]:

        # Check to see if it maps to a ref_event that's part of an equivalence class we've already
        # scored. If it is, don't count it as a false alarm.

        try:

            index = alignments[1].index(sys_event.event_id.value)

        except ValueError:

            subject = entry_pair.sys_events[sys_event.event_id.value]
            if subject.event_type.value in BASIC_MAT_IGNORED_ANCHOR_TYPES:

                LOGGER.warning(
                    f"            sys event {sys_event} missing from entrypair alignments "
                    f"because it's an ignored type: {subject.event_type.value}"
                )
                mapped_sys_events.append(sys_event)
                continue

            else:

                raise ValueError(
                    f"sys event {sys_event} missing from entrypair alignments!"
                )

        mapped = False
        for ref_event in mapped_ref_events:

            if type(ref_event) == int: # if the ref_event _is_ an EqCl

                ref_counterpart = alignments[0][index]
                if ref_counterpart in equivalence_classes[ref_event]:
                    mapped = True

        if not mapped:

            pair_scores.slot_alignments[slot_label].measures.increment_false_alarm(1, tab=4)
            pair_scores.slot_alignments[slot_label].pairs.append([None, __normalize(sys_event)])
            if sys_event.irrealis.value:
                pair_scores.irrealis_measures.increment_false_alarm(1, tab=5)
            if sys_event.time_attachments.value:
                pair_scores.time_attachment_measures.increment_false_alarm(
                    len(sys_event.time_attachments.value), tab=5
                )

        elif DEBUG:
            LOGGER.debug("                "
                f"System slot element {__normalize(sys_event)} part of mapped equivalence class. "
                "Skipping!"
            )

    if DEBUG:
        LOGGER.debug((f"            Final event-match subscore within slot: "
            f"{pair_scores.slot_alignments[slot_label].measures}"
        ))


def __score_slot_irrealis(ref_slot_element, sys_slot_element, entry_pair):
    """Adds a slot score modification based on the irrealis of the slot pair.

    Args:
        ref_slot_element (lib.bp.BPDocumentGranularTemplateProperty): the reference template slot
            element.
        sys_slot_element (lib.bp.BPDocumentGranularTemplateProperty): the system-output template
            slot element.
        entry_pair (EntryPair): the object wrapping both the reference and system entries.

    Returns:
        float: 0.125 if the irrealis values match, or -0.125 if they don't.
    """

    if ref_slot_element.irrealis.value == None and sys_slot_element.irrealis.value == None:
        score_mod = 0.125
    elif ref_slot_element.irrealis.value and sys_slot_element.irrealis.value == None:
        entry_pair.gran_scores.irrealis_measures.increment_miss(1, tab=4)
        score_mod = -0.125
    elif sys_slot_element.irrealis.value and ref_slot_element.irrealis.value == None:
        entry_pair.gran_scores.irrealis_measures.increment_false_alarm(1, tab=4)
        score_mod = -0.125
    elif ref_slot_element.irrealis.value == sys_slot_element.irrealis.value:
        entry_pair.gran_scores.irrealis_measures.increment_hit(1, tab=4)
        score_mod = 0.125
        entry_pair.gran_scores.irrealis_measures.increment_match(1, tab=4)
    else:
        entry_pair.gran_scores.irrealis_measures.increment_hit(1, tab=4)
        entry_pair.gran_scores.irrealis_measures.increment_match(0, tab=4)
        score_mod = -0.125

    return score_mod


def __score_slot_timeatt(ref_slot_element, sys_slot_element, entry_pair):
    """Adds a slot score modification based on the time-attachments of the slot pair.

    Args:
        ref_slot_element (lib.bp.BPDocumentGranularTemplateProperty): the reference template slot
            element.
        sys_slot_element (lib.bp.BPDocumentGranularTemplateProperty): the system-output template
            slot element.
        entry_pair (EntryPair): the object wrapping both the reference and system entries.

    Returns:
        float: 0.125 if the time-attachments align perfectly, or -0.125 if nothing aligns.
    """
    import itertools

    from munkres import make_cost_matrix
    from munkres import Munkres

    global DEBUG
    global LOGGER
    global VERBOSE

    ref_ssids = ref_slot_element.time_attachments.value or []
    sys_ssids = sys_slot_element.time_attachments.value or []

    if len(ref_ssids) < len(sys_ssids):

        if DEBUG:
            LOGGER.debug((
                f"            Padding ref ssids with {len(sys_ssids) - len(ref_ssids)} blanks"
                f" since sys_ssids is longer."
            ))

        # pad with `None`s
        ref_ssids = ref_ssids + [
            x for x in itertools.repeat(None, len(sys_ssids)-len(ref_ssids))
        ]

    elif len(sys_ssids) < len(ref_ssids):

        if DEBUG:
            LOGGER.debug((
                f"            Padding sys ssids with {len(ref_ssids) - len(sys_ssids)} blanks"
                f" since ref_ssids is longer."
            ))

        # pad with `None`s
        sys_ssids = sys_ssids + [
            x for x in itertools.repeat(None, len(ref_ssids)-len(sys_ssids))
        ]

    match_cache = {}
    matrix = make_cost_matrix(
        [
            [
                __score_ss_pair(
                    sys_ssid,
                    ref_ssid,
                    entry_pair,
                    match_cache,
                    ignore_synclass=True
                ) for sys_ssid in sys_ssids
            ] for ref_ssid in ref_ssids
        ],
        lambda cost: 1.0 - cost
    )
    if matrix:
        index_pairs = Munkres().compute(matrix)
    else:
        index_pairs = []

    measures = Measures("time-attachment")
    for row, column in index_pairs:

        ref_entity = ref_ssids[row]
        sys_entity = sys_ssids[column]

        if not sys_entity and not ref_entity:

            if DEBUG:
                LOGGER.debug(
                    f"                Both sys and ref time-attachment ssids are blank; skipping."
                )

        elif not sys_entity:

            if DEBUG:
                LOGGER.debug(("                "
                    "Adding 1 miss to time-attachment alignment within slot since ref entity "
                    f"{ref_entity} has no sys counterpart."
                ))

            measures.increment_miss(1, tab=4)

        elif not ref_entity:

            if DEBUG:
                LOGGER.debug(("                "
                    "Adding 1 false alarm to time-attachment alignment within slot since sys entity"
                    f" {sys_entity} has no ref counterpart."
                ))

            measures.increment_false_alarm(1, tab=4)

        else:

            match_score = match_cache[(ref_entity, sys_entity)]
            measures.increment_hit(1, tab=4)
            measures.increment_match(match_score, tab=4)
            if DEBUG:
                LOGGER.debug(("               "
                    f"Adding 1 hit ({match_score} match) to time-attachment alignment within "
                    f"slot since sys entity {sys_entity} aligns with ref entity {ref_entity}."
                ))

    entry_pair.gran_scores.time_attachment_measures.update(measures, tab=4)
    # required to set the F1 within `measures` to 1.0 if there are no time-attachment values
    measures.compute_metrics()

    return -0.125 + measures.f1 * 0.25


def __score_ss_pair(sys_ssid, ref_ssid, entry_pair, match_cache, ignore_synclass=False):

    from lib.utils import PUNCTUATION

    global DEBUG
    global SCORING_ARG_MATCH_TYPE
    global SCORING_INCORRECT_SPAN_REDUCTION_RATE

    if DEBUG:
        LOGGER.debug((
            f"            Scoring span set values for Entry Pair: {entry_pair}, ref SSID: "
            f"{ref_ssid}, sys SSID: {sys_ssid} "
        ))

    def __synclass_weight(best, second_best, third_best, ref_span):

        if ref_span.synclass.value == "event-anchor" or ref_span.is_anchor.value:

            if DEBUG:
                LOGGER.debug(("                    "
                    f"Reference span {ref_span} is an event anchor! Weighting 1.0."
                ))
            return 1.0

        elif ref_span.synclass.value:

            if ref_span.synclass.value == best:
                if DEBUG:
                    LOGGER.debug(("                    "
                        f"matched on {best} (best)"
                    ))
                return 1.0
            elif ref_span.synclass.value == second_best:
                if DEBUG:
                    LOGGER.debug(("                    "
                        f"matched on {second_best} (second best)"
                    ))
                return 0.5
            elif ref_span.synclass.value == third_best:
                if DEBUG:
                    LOGGER.debug(("                    "
                        f"matched on {third_best} (third best)"
                    ))
                return 0.25
            else:
                # no synclasses were found, so we don't weight it
                if DEBUG:
                    LOGGER.debug(("                    "
                        f"Didn't match anything"
                    ))
                return 1.0

        else:

            if DEBUG:
                LOGGER.debug(("                    "
                    f"Reference span {ref_span} has no synclass! Weighting 1.0."
            ))

            return 1.0

    def __strict_full(sys_span, ref_span):

        if sys_span != ref_span and (sys_span in [None, "",] or ref_span in [None, "",]):
            return 0.0

        ref_str = __mention_str(ref_span, 'full').lower()
        sys_str = __mention_str(sys_span, 'full').lower()

        if sys_str == ref_str:
            return 1.0

        else:

            # 2020/02/24: Replaced in response to consultation with team
            # and Tim McKinnon (sponsor CETA) on Redmine.  No longer
            # attempting to ignore inclusion/exclusion of articles,
            # determiners or honorifics.

            sys_str = sys_str.strip().strip(PUNCTUATION).strip()
            ref_str = ref_str.strip().strip(PUNCTUATION).strip()

            # Has the punct stripping created equivalent strings?
            # Make sure the punct stripping hasn't reduced the mention to nothing
            if sys_str == ref_str and len(ref_str) > 0:
                return 1.0
            else:
                return 0.0

    def __strict_head(sys_span, ref_span):

        if sys_span != ref_span and (sys_span in [None, "",] or ref_span in [None, "",]):
            return 0.0

        if ref_span.no_head.value:

            if sys_span.no_head.value:
                return 1.0

            # System did not identify this as having "no-head"; determine if system 'head' (or
            # 'full' if head not marked) appears anywhere within the 'full' reference, or vice
            # versa.
            ref_full = __mention_str(ref_span, 'full').lower()
            sys_head = __mention_str(sys_span, 'head').lower()

            if ref_full.strip().strip(PUNCTUATION).strip().find(
                sys_head.strip().strip(PUNCTUATION).strip()
            ) >= 0:
                return 1.0

        else:

            if sys_span.no_head.value:

                sys_full = __mention_str(sys_span, 'full').lower()
                ref_head = __mention_str(sys_span, 'head').lower()

                if sys_full.strip().strip(PUNCTUATION).strip().find(
                    ref_head.strip().strip(PUNCTUATION).strip()
                ) >= 0:
                    return 1.0
                else:
                    return 0.0

            else:

                ref_head = __mention_str(ref_span, 'head').lower()
                sys_head = __mention_str(sys_span, 'head').lower()

                if sys_head == ref_head:
                    return 1.0
                else:
                    sys_head = sys_head.strip().strip(PUNCTUATION).strip()
                    ref_head = ref_head.strip().strip(PUNCTUATION).strip()
                    # Has the punct stripping created equivalent strings?
                    # Make sure the punct stripping hasn't reduced the mention to nothing
                    if sys_head == ref_head and len(ref_head) > 0:
                        return 1.0
                    else:
                        return 0.0

    def __weighted_full(sys_span, ref_span):

        import re
        from lib.utils import levenshtein

        global SCORING_HEAD_MATCH_VALUE
        global SCORING_PARTIAL_MATCH_TYPE

        if sys_span != ref_span and (sys_span in [None, "",] or ref_span in [None, "",]):
            return 0.0

        sys_full = __mention_str(sys_span, 'full').lower().strip().strip(PUNCTUATION).strip()
        ref_head = __mention_str(ref_span, 'head').lower().strip().strip(PUNCTUATION).strip()

        # if the refMent is marked as not having a head ("no-head":true), then we consider that to
        # satisfy the minimum criteria of "matching a head". If this isn't a "headless" mention,
        # then we see if the reference mention's head string occurs anywhere within the system's
        # full string. If either of these conditions hold, then we proceed to evaluate character
        # overlap for a weighted score.
        if ref_span.no_head.value or sys_full.find(ref_head) >= 0:

            ref_full = __mention_str(ref_span, 'full').lower().strip().strip(PUNCTUATION).strip()
            sys_head = __mention_str(sys_span, 'head').lower().strip().strip(PUNCTUATION).strip()

            if SCORING_PARTIAL_MATCH_TYPE == "tokens":

                ref_tokens_len = len(ref_full.split())
                sys_tokens_len = len(sys_full.split())
                ref_head_tokens_len = len(ref_head.split())
                sys_head_tokens_len = len(sys_head.split())

                match_score  = 1.0
                if ref_head_tokens_len < ref_tokens_len:

                    ref_prenom_len = ref_tokens_len - ref_head_tokens_len
                    sys_prenom_len = sys_tokens_len - sys_head_tokens_len

                    if ref_prenom_len == 0:
                        match_score = 1.0
                    elif sys_prenom_len == 0:
                        match_score = SCORING_HEAD_MATCH_VALUE
                    else:
                        match_score = \
                            float(sys_prenom_len) / float(ref_prenom_len) * \
                            (1 - SCORING_HEAD_MATCH_VALUE) + \
                            SCORING_HEAD_MATCH_VALUE

                return match_score

            elif SCORING_PARTIAL_MATCH_TYPE == "characters":

                ref_full = ''.join([c for c in ref_full if re.compile(r"\w").match(c)])
                sys_full = ''.join([c for c in sys_full if re.compile(r"\w").match(c)])
                if len(ref_full) == 0:
                    if len(sys_full) == 0:
                        # Both are empty strings, so head = full and the non-head match quality is
                        # 1.0
                        return 1.0
                    else:
                        # A zero-length non-head string match has 0% overlap with a string with some
                        # characters
                        return 0.0
                if len(sys_full) == 0:
                    # A zero-length non-head string match has 0% overlap with a string with some
                    # characters
                    return 0.0
                return 1.0 - float(levenshtein(ref_full, sys_full)) / \
                    float(max(len(ref_full),len(sys_full)))

            else:

                raise NotImplementedError(
                    f"Unknown Partial Match Type set: '{SCORING_PARTIAL_MATCH_TYPE}'"
                )

        else:

            return 0.0

    def __weighted_head(sys_span, ref_span):

        from lib.utils import levenshtein

        if sys_span != ref_span and (sys_span in [None, "",] or ref_span in [None, "",]):
            return 0.0

        ref_head = __mention_str(ref_span, 'head').lower().strip().strip(PUNCTUATION).strip()
        sys_head = __mention_str(sys_span, 'head').lower().strip().strip(PUNCTUATION).strip()

        if len(ref_head) == 0:
            if len(sys_head) == 0:
                # Both are empty strings
                return 1.0
            else:
                # A zero-length string match has 0% overlap with a string with some characters
                return 0.0
        if len(sys_head) == 0:
            # A zero-length string match has 0% overlap with a string with some characters
            return 0.0
        return 1.0 - float(levenshtein(ref_head, sys_head)) / \
            float(max(len(ref_head),len(sys_head)))

    def __loose(sys_span, ref_span):

        if sys_span != ref_span and (sys_span in [None, "",] or ref_span in [None, "",]):
            return 0.0

        ref_full = __mention_str(ref_span, 'full').lower().strip().strip(PUNCTUATION).strip()
        sys_full = __mention_str(sys_span, 'full').lower().strip().strip(PUNCTUATION).strip()

        if sys_full.find(ref_full) >= 0:
            return 1.0
        elif ref_full.find(sys_full) >= 0:
            return 1.0
        else:
            return 0.0

    if (ref_ssid, sys_ssid) in match_cache:
        if DEBUG:
            LOGGER.debug("                Already in match cache! Nothing to do.")
        return match_cache[(ref_ssid, sys_ssid)]

    if not sys_ssid:
        if DEBUG:
            LOGGER.debug("                Sys SSID is `None`! Setting match to 0.0.")
        match_cache[(ref_ssid, sys_ssid)] = 0.0
        return match_cache[(ref_ssid, sys_ssid)]

    if not ref_ssid:
        if DEBUG:
            LOGGER.debug("                Ref SSID is `None`! Setting match to 0.0.")
        match_cache[(ref_ssid, sys_ssid)] = 0.0
        return match_cache[(ref_ssid, sys_ssid)]

    if ref_ssid not in entry_pair.ref_span_sets:

        if DEBUG:
            LOGGER.debug((
                "                "
                "Ref SSID isn't in this Entry's span set list! Setting match to 0.0."
            ))
        match_cache[(ref_ssid, sys_ssid)] = 0.0
        return match_cache[(ref_ssid, sys_ssid)]

    if sys_ssid not in entry_pair.sys_span_sets:
        if DEBUG:
            LOGGER.debug((
                "                "
                "Sys SSID isn't in this Entry's span set list! Setting match to 0.0."
            ))
        match_cache[(ref_ssid, sys_ssid)] = 0.0
        return match_cache[(ref_ssid, sys_ssid)]

    sys_span_set = entry_pair.sys_span_sets[sys_ssid]
    ref_span_set = entry_pair.ref_span_sets[ref_ssid]

    if DEBUG:
        LOGGER.debug("                Available reference spans:")

    name_present = False
    nominal_present = False
    pronoun_present = False
    is_event_anchor = False
    for span in ref_span_set.spans.value:

        if DEBUG:
            LOGGER.debug(f"                    {repr(span)}")

        if span.synclass.value == "name":
            name_present = True
        elif span.synclass.value == "nominal":
            nominal_present = True
        elif span.synclass.value == "pronoun":
            pronoun_present = True
        elif span.synclass.value == "event-anchor" or span.is_anchor.value:
            is_event_anchor = True

    if DEBUG:
        LOGGER.debug((
            f"                Reference Classes found:{' Name' if name_present else ''}"
            f"{' Nominal' if nominal_present else ''}{' Pronoun' if pronoun_present else ''}"
            f"{' Event Anchor' if is_event_anchor else ''}"
        ))

    if DEBUG:
        LOGGER.debug("                Available system spans:")
        for span in sys_span_set.spans.value:
            if DEBUG:
                LOGGER.debug(f"                    {repr(span)}")

    ## determine weights

    w_synclass_best = None
    w_synclass_second_best = None
    w_synclass_third_best = None

    if name_present:
        w_synclass_best = 'name'
        if nominal_present:
            w_synclass_second_best = 'nominal'
            if pronoun_present:
                w_synclass_third_best = 'pronoun'
        elif pronoun_present:
            w_synclass_second_best = 'pronoun'
    elif nominal_present:
        w_synclass_best = 'nominal'
        if pronoun_present:
            w_synclass_second_best = 'pronoun'
    elif pronoun_present:
        w_synclass_best = 'pronoun'
    else:
        if DEBUG and not ignore_synclass and not is_event_anchor:
            LOGGER.warning(("                "
                f"No synclasses in reference non-Event-anchor SpanSet {ref_span_set}!"
            ))

    if DEBUG and not ignore_synclass:
        LOGGER.debug(("                "
            f"synclass availability: {w_synclass_best} > {w_synclass_second_best} > "
            f"{w_synclass_third_best}"
        ))

    ## assign span pair matches

    best_match_score = 0.0
    best_ref_span = None
    best_sys_span = None

    for ref_span in ref_span_set.spans.value:

        for sys_span in sys_span_set.spans.value:

            if SCORING_ARG_MATCH_TYPE == 'ignore':
                match_score = 1.0
            elif SCORING_ARG_MATCH_TYPE == 'strict-full':
                match_score = __strict_full(sys_span, ref_span)
            elif SCORING_ARG_MATCH_TYPE == 'strict-head':
                match_score = __strict_head(sys_span, ref_span)
            elif SCORING_ARG_MATCH_TYPE == 'weighted-full':
                match_score = __weighted_full(sys_span, ref_span)
            elif SCORING_ARG_MATCH_TYPE == 'weighted-head':
                match_score = __weighted_head(sys_span, ref_span)
            elif SCORING_ARG_MATCH_TYPE == 'loose':
                match_score = __loose(sys_span, ref_span)
            else:
                raise RuntimeError(
                    f"Bad value for SCORING_ARG_MATCH_TYPE: {SCORING_ARG_MATCH_TYPE}"
                )

            if ref_span and not ignore_synclass:
                match_score *= __synclass_weight(
                    w_synclass_best,
                    w_synclass_second_best,
                    w_synclass_third_best,
                    ref_span
                )

            if match_score >= best_match_score:

                if DEBUG and match_score > best_match_score:
                    if best_match_score == 0.0:
                        LOGGER.debug((
                            f"                [{SCORING_ARG_MATCH_TYPE}] "
                            f"Found first non-zero match score: {match_score}."
                        ))
                    else:
                        LOGGER.debug((
                            f"                [{SCORING_ARG_MATCH_TYPE}] "
                            f"Found better match score: {match_score}."
                        ))

                # if the scores are the same, the most-recently-seen spans are selected
                best_match_score = match_score
                best_ref_span = ref_span
                best_sys_span = sys_span

    if DEBUG:
        if best_ref_span:
            LOGGER.debug(f"                Best Ref Span: '{best_ref_span.string_.value}'")
        else:
            LOGGER.debug(f"                Best Ref Span: None")
        if best_sys_span:
            LOGGER.debug(f"                Best Sys Span: '{best_sys_span.string_.value}'")
        else:
            LOGGER.debug(f"                Best Sys Span: None")
        LOGGER.debug(f"                Best Match score: {best_match_score}")

    if best_ref_span:
        if DEBUG:
            LOGGER.debug(f"                Removing best ref span from ref span sets.")
            LOGGER.debug(f"                  {ref_span_set.spans.value}")
        ref_spans = [span for span in ref_span_set.spans.value if span != best_ref_span]
        if DEBUG:
            LOGGER.debug(f"                  {ref_spans}")
    else:
        ref_spans = ref_span_set.spans.value

    if best_sys_span != None:
        if DEBUG:
            LOGGER.debug(f"                Removing best sys span from sys span sets.")
            LOGGER.debug(f"                  {sys_span_set.spans.value}")
        sys_spans = [span for span in sys_span_set.spans.value if span != best_sys_span]
        if DEBUG:
            LOGGER.debug(f"                  {sys_spans}")
    else:
        sys_spans = sys_span_set.spans.value

    span_matches = {
        "matches": [[best_match_score, best_sys_span, best_ref_span]],
        "remaining_sys_spans": sys_spans,
        "remaining_ref_spans": ref_spans,
        "overall_score": best_match_score
    }

    if DEBUG:
        LOGGER.debug(f"                Overall score: {span_matches['overall_score']}")

    match_cache[(ref_ssid, sys_ssid)] = span_matches["overall_score"]

    return match_cache[(ref_ssid, sys_ssid)]


def __to_cost(profit):
    """Converts a profit value (float) into a cost value.

    Q: Why didn't we use sys.maxvalue like the Munkres docs suggest?
    A: Our optimization algorithm allows negative profit values. If we use the maximum possible
      allowed value as our ceiling, subtracting a negative profit from it would result in the same
      number, breaking the cost matrix construction.

    Args:
        profit (float|munkres.DISALLOWED_OBJ): either a floating-point number or a DISALLOWED
            constant reflective of a profit.

    Returns:
        float or munkres.DISALLOWED_OBJ: a value reflective of a cost, given the input.
    """

    if type(profit) == int or type(profit) == float:
        return 1000000000.0 - profit
    else:
        return profit


## Surface Functions


def dump_alignments(entry_pairs, path, dialect, ignore_synclass=False):
    """Writes alignment information to an Excel workbook.

    Args:
        entry_pairs (list): of :class:`EntryPair` objects, each containing score information
            for each pair of entries in the documents.
        path (str): a path to write the data to.
        dialect (str): the task the document was created with. Must be one of
            ``utils.MAT_DIALECT_CHOICES``.
        ignore_synclass (bool): if True, don't consider the syntactic classes of argument
            mentions when scoring.
    """
    import json

    from openpyxl import Workbook

    from lib.utils import get_tqdm
    from lib.utils import MAT_DIALECT_CHOICES
    from lib.utils import TQDM_PREFIX
    from lib.utils import user_allows_overwrite

    global DEBUG
    global FORCE
    global LOGGER
    global VERBOSE

    assert dialect in MAT_DIALECT_CHOICES, f"Bad dialect choice: {dialect}"

    tqdm = get_tqdm(VERBOSE, DEBUG)

    if path and user_allows_overwrite(path, VERBOSE, FORCE):

        wb = Workbook()

        ## Abstract/Basic

        event_ws = wb.active
        event_ws.title = "Event Alignments"
        event_ws.freeze_panes = 'C2'

        event_ws.append([
            "System Entry ID", "Reference Entry ID",
            "System Event ID", "Reference Event ID",
            "System Event Type", "Reference Event Type",
            "Alignment Score",
            "System Anchor SSID", "Reference Anchor SSID",
            "System Anchor Spans", "Reference Anchor Spans",
        ])

        arg_ws = wb.create_sheet('Arg Alignments')
        arg_ws.freeze_panes = 'C2'

        arg_ws.append([
            "System Entry ID", "Reference Entry ID",
            "System Event ID", "Reference Event ID",
            "Arg Type",
            "System SSID", "Reference SSID",
            "Alignment Score",
            "System Spans", "Reference Spans",
        ])

        iterator = tqdm(entry_pairs, desc=f"{TQDM_PREFIX} Dumping Event alignments", unit="entpr")
        for entry_pair in iterator:

            # When filtering by annotator class or dataset label, this value is never initialized.
            if entry_pair.scores.event_alignment == None:
                continue

            for i in range(len(entry_pair.scores.event_alignment[0])):

                ref_evid = entry_pair.scores.event_alignment[0][i]
                sys_evid = entry_pair.scores.event_alignment[1][i]

                if ref_evid:
                    ref_event = entry_pair.ref_events[ref_evid]
                else:
                    ref_event = None

                if sys_evid:
                    sys_event = entry_pair.sys_events[sys_evid]
                else:
                    sys_event = None

                scores = entry_pair.scores.event_pair_scores[(ref_evid, sys_evid)]

                if dialect == "Abstract":
                    if ref_event:
                        ref_event_type = f"{ref_event.hh.value}-{ref_event.mv.value}"
                    else:
                        ref_event_type = "N/A"
                    if sys_event:
                        sys_event_type = f"{sys_event.hh.value}-{sys_event.mv.value}"
                    else:
                        sys_event_type = "N/A"
                elif dialect == "Basic" or dialect == "Granular":
                    if ref_event:
                        ref_event_type = ref_event.event_type.value
                    else:
                        ref_event_type = "N/A"
                    if sys_event:
                        sys_event_type = sys_event.event_type.value
                    else:
                        sys_event_type = "N/A"

                event_ws.append([
                    entry_pair.sys_entry_id, entry_pair.ref_entry_id,
                    sys_evid, ref_evid,
                    sys_event_type, ref_event_type,
                    scores.event_measures.match,
                    sys_event.anchors.value if sys_event else None,
                    ref_event.anchors.value if ref_event else None,
                    __render_spanset_compact(
                        sys_event.anchors.value, entry_pair.sys_span_sets, include_ssid=False
                    ) if sys_event else "N/A",
                    __render_spanset_compact(
                        ref_event.anchors.value, entry_pair.ref_span_sets, include_ssid=False
                    ) if ref_event else "N/A",
                ])

                for arg_field in ["agents", "patients", "ref_events", "money"]:
                    if arg_field in scores.arg_alignments:
                        for ssid_pair in scores.arg_alignments[arg_field].ssid_pairs:

                            ref_ssid = ssid_pair[0]
                            sys_ssid = ssid_pair[1]
                            sys_spanset, ref_spanset = __get_aligned_arg_strings(
                                ref_ssid, sys_ssid,
                                entry_pair.ref_span_sets, entry_pair.sys_span_sets,
                                include_ssid=False
                            )

                            arg_ws.append([
                                entry_pair.sys_entry_id, entry_pair.ref_entry_id,
                                sys_evid, ref_evid,
                                arg_field,
                                sys_ssid, ref_ssid,
                                __score_ss_pair(
                                    sys_ssid, ref_ssid, entry_pair, {},
                                    ignore_synclass=ignore_synclass
                                ),
                                sys_spanset, ref_spanset
                            ])

        ## Granular

        if dialect == "Granular":

            template_ws = wb.create_sheet('Template Alignments')
            template_ws.freeze_panes = 'C2'

            template_ws.append([
                "System Entry ID", "Reference Entry ID",
                "System Template ID", "Reference Template ID",
                "System Template Type", "Reference Template Type",
                "Alignment Score",
            ])

            slot_ws = wb.create_sheet('Slot Alignments')
            slot_ws.freeze_panes = 'C2'

            slot_ws.append([
                "System Entry ID", "Reference Entry ID",
                "System Template ID", "Reference Template ID",
                "Slot Type", "Slot Alignment Value Type"
                "System ID", "Reference ID",
                "Alignment Score",
                "System Spans (If SpanSet)", "Reference Spans (If SpanSet)",
            ])

            iterator = tqdm(
                entry_pairs, desc=f"{TQDM_PREFIX} Dumping Template alignments", unit="entpr"
            )
            for entry_pair in iterator:

                # When filtering by annotator class or dataset label, this value is never
                # initialized.
                if entry_pair.gran_scores.template_alignment == None:
                    continue

                equivalence_classes = \
                    entry_pair.ref_entry.get_event_set("Basic").template_filler_coref_events.value \
                        or []

                for i in range(len(entry_pair.gran_scores.template_alignment[0])):

                    ref_tid = entry_pair.gran_scores.template_alignment[0][i]
                    sys_tid = entry_pair.gran_scores.template_alignment[1][i]

                    if ref_tid:
                        ref_template = entry_pair.ref_templates[ref_tid]
                    else:
                        ref_template = None

                    if sys_tid:
                        sys_template = entry_pair.sys_templates[sys_tid]
                    else:
                        sys_template = None

                    scores = entry_pair.gran_scores.template_pair_scores[(ref_tid, sys_tid)]

                    template_ws.append([
                        entry_pair.sys_entry_id, entry_pair.ref_entry_id,
                        sys_tid, ref_tid,
                        sys_template.type_.value if sys_template else None,
                        ref_template.type_.value if ref_template else None,
                        scores.template_measures.match,
                    ])

                    slot_alignments = entry_pair.gran_scores.slot_alignments[(ref_tid, sys_tid)]
                    for slot_label in slot_alignments:

                        pairs = slot_alignments[slot_label].pairs
                        measures = slot_alignments[slot_label].measures

                        for pair in pairs:

                            ref_id = pair[0]
                            sys_id = pair[1]

                            alignment_type = None

                            # determine whether it's an event or a spanset in this slot
                            if sys_id:

                                if sys_id in entry_pair.sys_span_sets:

                                    if not alignment_type:
                                        alignment_type = "span"

                                elif sys_id in entry_pair.sys_events:

                                    sys_event = entry_pair.sys_events[sys_id]
                                    if not alignment_type:
                                        alignment_type = "event"

                                elif sys_id is not None:

                                    if not alignment_type:
                                        alignment_type = f"primitive ({type(sys_id)})"

                            if ref_id:

                                if ref_id in entry_pair.ref_span_sets:

                                    if not alignment_type:
                                        alignment_type = "span"
                                    sys_spanset, ref_spanset = __get_aligned_arg_strings(
                                        ref_id, sys_id,
                                        entry_pair.ref_span_sets, entry_pair.sys_span_sets,
                                        include_ssid=False
                                    )
                                    score = __score_ss_pair(
                                        sys_id, ref_id, entry_pair, {},
                                        ignore_synclass=ignore_synclass
                                    )

                                elif ref_id in entry_pair.ref_events:

                                    ref_event = entry_pair.ref_events[ref_id]
                                    if not alignment_type:
                                        alignment_type = "event"
                                    score = __score_basic_event_pair(
                                        ref_id, sys_id, entry_pair, {},
                                        ignore_synclass=ignore_synclass
                                    )

                                elif type(ref_id) == str and \
                                    "Equivalence Class at Index " in ref_id:

                                    index = int(ref_id.replace("Equivalence Class at Index ", ""))
                                    if not alignment_type:
                                        alignment_type = f"event (equivalence class)"

                                    try:
                                        eqcl = equivalence_classes[index]
                                    except IndexError:
                                        eqcl = None

                                    if eqcl:

                                        best = 0
                                        for evid in eqcl:
                                            score = __score_basic_event_pair(
                                                evid, sys_id, entry_pair, {},
                                                ignore_synclass=ignore_synclass
                                            )
                                            if score > best:
                                                best = score
                                        score = best

                                elif ref_id is not None:

                                    if not alignment_type:
                                        alignment_type = f"primitive ({type(ref_id)})"
                                    score = 1 if ref_id == sys_id else 0

                            slot_ws.append([
                                entry_pair.sys_entry_id, entry_pair.ref_entry_id,
                                sys_tid, ref_tid,
                                slot_label, alignment_type,
                                sys_id, ref_id,
                                score,
                                sys_spanset if alignment_type == "span" else None,
                                ref_spanset if alignment_type == "span" else None,
                            ])

        if VERBOSE:
            LOGGER.info("Writing dumped alignments...")
        wb.save(path)


def render_csv(
    scores,
    entry_pairs,
    sys_path,
    ref_path,
    path,
    tabbed=False,
    granular=False,
    extras=False
):
    """Writes score information to a CSV file.

    Args:
        scores (ScoreStruct|GranScoreStruct): contains scores for the whole system-generated
            document.
        entry_pairs (list): of :class:`EntryPair` objects, each containing score information
            for each pair of entries in the documents.
        sys_path (str): a path to a BP JSON document created by an IE system (typically named
            "results.json").
        ref_path (str): a path to a BP JSON document containing gold-standard annotations.
        path (str): a path to write the data to.
        tabbed (bool): if True, write tab-delimited values instead of comma-delimited. Defaults to
            False.
        granular (bool): whether the scorer was run in Granular mode or not.
        extras (bool): if True, include extra scoring info for agents, patients, and event-types.
    """
    import csv
    import os

    from lib.utils import GRANULAR_MAT_TEMPLATE_TYPES
    from lib.utils import user_allows_overwrite

    global FORCE
    global LOGGER
    global VERBOSE

    if VERBOSE:
        LOGGER.info(f"Writing {'TSV' if tabbed else 'CSV'} file to {path}.")

    if path:

        if user_allows_overwrite(path, VERBOSE, FORCE):

            with open(path, 'w', encoding='UTF-8') as f:

                if tabbed:
                    stream = csv.writer(f, delimiter="\t")
                else:
                    stream = csv.writer(f)

                headers = __render_headers(scores, granular=granular, extras=extras)

                stream.writerow(headers)

                doc_aggs = {}
                for ttype in GRANULAR_MAT_TEMPLATE_TYPES:
                    for slot_label in GRANULAR_MAT_TEMPLATE_TYPES[ttype]:
                        key = f"{ttype}::{GRANULAR_MAT_TEMPLATE_TYPES[ttype][slot_label]}"
                        doc_aggs[key] = Measures(key)
                for entry_pair in entry_pairs:
                    if granular:
                        row = [
                            entry_pair.sys_entry_id,
                            entry_pair.ref_entry_id,
                            ## event match stats
                            entry_pair.gran_scores.template_measures.miss,
                            entry_pair.gran_scores.template_measures.false_alarm,
                            entry_pair.gran_scores.template_measures.hit,
                            entry_pair.gran_scores.template_measures.match,
                            entry_pair.gran_scores.template_measures.value,
                            entry_pair.gran_scores.template_measures.precision,
                            entry_pair.gran_scores.template_measures.recall,
                            entry_pair.gran_scores.template_measures.f1,
                            ## argmatch stats
                            entry_pair.gran_scores.slotmatch_measures.miss,
                            entry_pair.gran_scores.slotmatch_measures.false_alarm,
                            entry_pair.gran_scores.slotmatch_measures.hit,
                            entry_pair.gran_scores.slotmatch_measures.match,
                            entry_pair.gran_scores.slotmatch_measures.value,
                            entry_pair.gran_scores.slotmatch_measures.precision,
                            entry_pair.gran_scores.slotmatch_measures.recall,
                            entry_pair.gran_scores.slotmatch_measures.f1,
                            ## combined stats
                            entry_pair.gran_scores.combined_measures.miss,
                            entry_pair.gran_scores.combined_measures.false_alarm,
                            entry_pair.gran_scores.combined_measures.hit,
                            entry_pair.gran_scores.combined_measures.match,
                            entry_pair.gran_scores.combined_measures.value,
                            entry_pair.gran_scores.combined_measures.precision,
                            entry_pair.gran_scores.combined_measures.recall,
                            entry_pair.gran_scores.combined_measures.f1,
                            ## program-defined product of eventmatch F1 and argmatch F1
                            entry_pair.gran_scores.combined_score,
                        ]
                    else:
                        row = [
                            entry_pair.sys_entry_id,
                            entry_pair.ref_entry_id,
                            ## event match stats
                            entry_pair.scores.event_measures.miss,
                            entry_pair.scores.event_measures.false_alarm,
                            entry_pair.scores.event_measures.hit,
                            entry_pair.scores.event_measures.match,
                            entry_pair.scores.event_measures.value,
                            entry_pair.scores.event_measures.precision,
                            entry_pair.scores.event_measures.recall,
                            entry_pair.scores.event_measures.f1,
                            ## argmatch stats
                            entry_pair.scores.argmatch_measures.miss,
                            entry_pair.scores.argmatch_measures.false_alarm,
                            entry_pair.scores.argmatch_measures.hit,
                            entry_pair.scores.argmatch_measures.match,
                            entry_pair.scores.argmatch_measures.value,
                            entry_pair.scores.argmatch_measures.precision,
                            entry_pair.scores.argmatch_measures.recall,
                            entry_pair.scores.argmatch_measures.f1,
                            ## combined stats
                            entry_pair.scores.combined_measures.miss,
                            entry_pair.scores.combined_measures.false_alarm,
                            entry_pair.scores.combined_measures.hit,
                            entry_pair.scores.combined_measures.match,
                            entry_pair.scores.combined_measures.value,
                            entry_pair.scores.combined_measures.precision,
                            entry_pair.scores.combined_measures.recall,
                            entry_pair.scores.combined_measures.f1,
                            ## program-defined product of eventmatch F1 and argmatch F1
                            entry_pair.scores.combined_score,
                        ]

                    if extras:
                        if not granular:
                            row.extend([
                                entry_pair.scores.agent_match_measures.miss,
                                entry_pair.scores.agent_match_measures.false_alarm,
                                entry_pair.scores.agent_match_measures.hit,
                                entry_pair.scores.agent_match_measures.match,
                                entry_pair.scores.agent_match_measures.precision,
                                entry_pair.scores.agent_match_measures.recall,
                                entry_pair.scores.agent_match_measures.f1,

                                entry_pair.scores.patient_match_measures.miss,
                                entry_pair.scores.patient_match_measures.false_alarm,
                                entry_pair.scores.patient_match_measures.hit,
                                entry_pair.scores.patient_match_measures.match,
                                entry_pair.scores.patient_match_measures.precision,
                                entry_pair.scores.patient_match_measures.recall,
                                entry_pair.scores.patient_match_measures.f1,

                                entry_pair.scores.money_match_measures.miss,
                                entry_pair.scores.money_match_measures.false_alarm,
                                entry_pair.scores.money_match_measures.hit,
                                entry_pair.scores.money_match_measures.match,
                                entry_pair.scores.money_match_measures.precision,
                                entry_pair.scores.money_match_measures.recall,
                                entry_pair.scores.money_match_measures.f1,
                            ])
                            for evtype in sorted(
                                list(entry_pair.scores.evtype_match_measures.keys())
                            ):
                                row.extend([
                                    entry_pair.scores.evtype_match_measures[evtype].miss,
                                    entry_pair.scores.evtype_match_measures[evtype].false_alarm,
                                    entry_pair.scores.evtype_match_measures[evtype].hit,
                                    entry_pair.scores.evtype_match_measures[evtype].match,
                                    entry_pair.scores.evtype_match_measures[evtype].precision,
                                    entry_pair.scores.evtype_match_measures[evtype].recall,
                                    entry_pair.scores.evtype_match_measures[evtype].f1,
                                ])
                        else:
                            for ttype in sorted(
                                list(entry_pair.gran_scores.ttype_match_measures.keys())
                            ):
                                row.extend([
                                    entry_pair.gran_scores.ttype_match_measures[ttype].miss,
                                    entry_pair.gran_scores.ttype_match_measures[ttype].false_alarm,
                                    entry_pair.gran_scores.ttype_match_measures[ttype].hit,
                                    entry_pair.gran_scores.ttype_match_measures[ttype].match,
                                    entry_pair.gran_scores.ttype_match_measures[ttype].precision,
                                    entry_pair.gran_scores.ttype_match_measures[ttype].recall,
                                    entry_pair.gran_scores.ttype_match_measures[ttype].f1,
                                ])
                            row.extend([
                                entry_pair.gran_scores.irrealis_measures.miss,
                                entry_pair.gran_scores.irrealis_measures.false_alarm,
                                entry_pair.gran_scores.irrealis_measures.hit,
                                entry_pair.gran_scores.irrealis_measures.match,
                                entry_pair.gran_scores.irrealis_measures.precision,
                                entry_pair.gran_scores.irrealis_measures.recall,
                                entry_pair.gran_scores.irrealis_measures.f1,
                            ])
                            row.extend([
                                entry_pair.gran_scores.time_attachment_measures.miss,
                                entry_pair.gran_scores.time_attachment_measures.false_alarm,
                                entry_pair.gran_scores.time_attachment_measures.hit,
                                entry_pair.gran_scores.time_attachment_measures.match,
                                entry_pair.gran_scores.time_attachment_measures.precision,
                                entry_pair.gran_scores.time_attachment_measures.recall,
                                entry_pair.gran_scores.time_attachment_measures.f1,
                            ])
                            entry_pair_aggs = __aggregate_slot_scores(entry_pair)
                            for ttype in GRANULAR_MAT_TEMPLATE_TYPES:
                                for slot_label in GRANULAR_MAT_TEMPLATE_TYPES[ttype]:
                                    measure = entry_pair_aggs[
                                        f"{ttype}::{GRANULAR_MAT_TEMPLATE_TYPES[ttype][slot_label]}"
                                    ]
                                    row.extend([
                                        measure.miss, measure.false_alarm, measure.hit,
                                        measure.match, measure.precision, measure.recall,
                                        measure.f1,
                                    ])
                                    doc_aggs[
                                        f"{ttype}::{GRANULAR_MAT_TEMPLATE_TYPES[ttype][slot_label]}"
                                    ].update(measure)

                    stream.writerow(row)

                if granular:
                    row = [
                        os.path.basename(sys_path),
                        os.path.basename(ref_path),
                        # event match
                        scores.template_measures.miss,
                        scores.template_measures.false_alarm,
                        scores.template_measures.hit,
                        scores.template_measures.match,
                        scores.template_measures.value,
                        scores.template_measures.precision,
                        scores.template_measures.recall,
                        scores.template_measures.f1,
                        # arg match
                        scores.slotmatch_measures.miss,
                        scores.slotmatch_measures.false_alarm,
                        scores.slotmatch_measures.hit,
                        scores.slotmatch_measures.match,
                        scores.slotmatch_measures.value,
                        scores.slotmatch_measures.precision,
                        scores.slotmatch_measures.recall,
                        scores.slotmatch_measures.f1,
                        # combined
                        scores.combined_measures.miss,
                        scores.combined_measures.false_alarm,
                        scores.combined_measures.hit,
                        scores.combined_measures.match,
                        scores.combined_measures.value,
                        scores.combined_measures.precision,
                        scores.combined_measures.recall,
                        scores.combined_measures.f1,
                        ## f1 product score
                        scores.combined_score,
                    ]
                else:
                    row = [
                        os.path.basename(sys_path),
                        os.path.basename(ref_path),
                        # event match
                        scores.event_measures.miss,
                        scores.event_measures.false_alarm,
                        scores.event_measures.hit,
                        scores.event_measures.match,
                        scores.event_measures.value,
                        scores.event_measures.precision,
                        scores.event_measures.recall,
                        scores.event_measures.f1,
                        # arg match
                        scores.argmatch_measures.miss,
                        scores.argmatch_measures.false_alarm,
                        scores.argmatch_measures.hit,
                        scores.argmatch_measures.match,
                        scores.argmatch_measures.value,
                        scores.argmatch_measures.precision,
                        scores.argmatch_measures.recall,
                        scores.argmatch_measures.f1,
                        # combined
                        scores.combined_measures.miss,
                        scores.combined_measures.false_alarm,
                        scores.combined_measures.hit,
                        scores.combined_measures.match,
                        scores.combined_measures.value,
                        scores.combined_measures.precision,
                        scores.combined_measures.recall,
                        scores.combined_measures.f1,
                        ## f1 product score
                        scores.combined_score,
                    ]

                if extras:
                    if not granular:
                        row.extend([
                            scores.agent_match_measures.miss,
                            scores.agent_match_measures.false_alarm,
                            scores.agent_match_measures.hit,
                            scores.agent_match_measures.match,
                            scores.agent_match_measures.precision,
                            scores.agent_match_measures.recall,
                            scores.agent_match_measures.f1,

                            scores.patient_match_measures.miss,
                            scores.patient_match_measures.false_alarm,
                            scores.patient_match_measures.hit,
                            scores.patient_match_measures.match,
                            scores.patient_match_measures.precision,
                            scores.patient_match_measures.recall,
                            scores.patient_match_measures.f1,

                            scores.money_match_measures.miss,
                            scores.money_match_measures.false_alarm,
                            scores.money_match_measures.hit,
                            scores.money_match_measures.match,
                            scores.money_match_measures.precision,
                            scores.money_match_measures.recall,
                            scores.money_match_measures.f1,
                        ])
                        for evtype in sorted(
                            list(scores.evtype_match_measures.keys())
                        ):
                            row.extend([
                                scores.evtype_match_measures[evtype].miss,
                                scores.evtype_match_measures[evtype].false_alarm,
                                scores.evtype_match_measures[evtype].hit,
                                scores.evtype_match_measures[evtype].match,
                                scores.evtype_match_measures[evtype].precision,
                                scores.evtype_match_measures[evtype].recall,
                                scores.evtype_match_measures[evtype].f1,
                            ])
                    else:
                        for ttype in sorted(
                            list(scores.ttype_match_measures.keys())
                        ):
                            row.extend([
                                scores.ttype_match_measures[ttype].miss,
                                scores.ttype_match_measures[ttype].false_alarm,
                                scores.ttype_match_measures[ttype].hit,
                                scores.ttype_match_measures[ttype].match,
                                scores.ttype_match_measures[ttype].precision,
                                scores.ttype_match_measures[ttype].recall,
                                scores.ttype_match_measures[ttype].f1,
                            ])
                        row.extend([
                            scores.irrealis_measures.miss,
                            scores.irrealis_measures.false_alarm,
                            scores.irrealis_measures.hit,
                            scores.irrealis_measures.match,
                            scores.irrealis_measures.precision,
                            scores.irrealis_measures.recall,
                            scores.irrealis_measures.f1,
                        ])
                        row.extend([
                            scores.time_attachment_measures.miss,
                            scores.time_attachment_measures.false_alarm,
                            scores.time_attachment_measures.hit,
                            scores.time_attachment_measures.match,
                            scores.time_attachment_measures.precision,
                            scores.time_attachment_measures.recall,
                            scores.time_attachment_measures.f1,
                        ])
                        for ttype in GRANULAR_MAT_TEMPLATE_TYPES:
                            for slot_label in GRANULAR_MAT_TEMPLATE_TYPES[ttype]:
                                measure = doc_aggs[
                                    f"{ttype}::{GRANULAR_MAT_TEMPLATE_TYPES[ttype][slot_label]}"
                                ]
                                row.extend([
                                    measure.miss, measure.false_alarm, measure.hit,
                                    measure.match, measure.precision, measure.recall,
                                    measure.f1,
                                ])

                stream.writerow(row)

        elif VERBOSE:

            LOGGER.info(
                f"    User chose to not overwrite existing {'TSV' if tabbed else 'CSV'} file."
            )

    else:

        raise RuntimeError(f"Missing {'TSV' if tabbed else 'CSV'} output path!")


def render_excel(
    scores,
    entry_pairs,
    sys_path,
    ref_path,
    path,
    color_scale=False,
    granular=False,
    extras=False
):
    """Writes score information to an Excel file.

    Args:
        scores (ScoreStruct|GranScoreStruct): containing scores for the whole system-generated
            document.
        entry_pairs (list): of :class:`EntryPair` objects, each containing score information
            for each pair of entries in the documents.
        sys_path (str): a path to a BP JSON document created by an IE system (typically named
            "results.json").
        ref_path (str): a path to a BP JSON document containing gold-standard annotations.
        path (str): a path to write the data to.
        color_scale (bool): if True, colorize floating point data with a range scale. Defaults to
            False.
        granular (bool): whether the scorer was run in Granular mode or not.
        extras (bool): if True, include extra scoring info for agents, patients, and event-types.
    """

    import os

    from openpyxl import Workbook
    from openpyxl.formatting.rule import ColorScale
    from openpyxl.formatting.rule import FormatObject
    from openpyxl.formatting.rule import Rule
    from openpyxl.styles import Alignment
    from openpyxl.styles import Border
    from openpyxl.styles import Color
    from openpyxl.styles import Font
    from openpyxl.styles import Side
    from openpyxl.utils import get_column_letter

    from lib.utils import user_allows_overwrite

    global HEADERS
    global FORCE
    global VERBOSE

    if VERBOSE:
        LOGGER.info(f"Writing Excel file to {path}.")

    if path:
        if user_allows_overwrite(path, VERBOSE, FORCE):

            ## init workbook, add basic formatting

            wb = Workbook()
            ws = wb.active
            ws.title = "Results"
            ws.freeze_panes = 'C3'

            font = Font(
                name='Calibri',
                size=11,
                bold=True,
                color="FFFFFF"
            )

            alignmnet = Alignment(horizontal="center", vertical="center")

            border = Border(right=Side(border_style="thin", color="000000"))

            # see https://materializecss.com/color.html for palette
            colors = [
                Color('FF8A80'), # min, red accent-1
                Color('FFF59D'), # mid, yellow lighten-3
                Color('B9F6CA'), # max, green accent-1
            ]

            cmap = Rule(
                type='colorScale',
                colorScale=ColorScale(
                    cfvo=[
                        FormatObject(type='min', val=0),
                        FormatObject(type='num', val=0.5),
                        FormatObject(type='max', val=1.0),
                    ],
                    color=colors
                )
            )

            ## set up header rows

            ws.column_dimensions['A'].width = 37
            ws.column_dimensions['B'].width = 37
            ws.column_dimensions['AA'].width = 15

            ws.merge_cells('A1:A2')
            ws.merge_cells('B1:B2')
            ws.merge_cells('AA1:AA2')

            ws.merge_cells('C1:J1')
            ws.merge_cells('K1:R1')
            ws.merge_cells('S1:Z1')

            ws['A1'].style = "Check Cell"
            ws['B1'].style = "Check Cell"
            ws['AA1'].style = "Check Cell"
            ws['C1'].style = "Check Cell"
            ws['K1'].style = "Check Cell"
            ws['S1'].style = "Check Cell"

            ws['A1'].font = font
            ws['B1'].font = font
            ws['AA1'].font = font
            ws['C1'].font = font
            ws['K1'].font = font
            ws['S1'].font = font

            ws['A1'].alignment = alignmnet
            ws['B1'].alignment = alignmnet
            ws['AA1'].alignment = alignmnet
            ws['C1'].alignment = alignmnet
            ws['K1'].alignment = alignmnet
            ws['S1'].alignment = alignmnet

            ws['B1'].border = border
            ws['I1'].border = border
            ws['I2'].border = border
            ws['P1'].border = border
            ws['P2'].border = border
            ws['W1'].border = border
            ws['W2'].border = border

            # header values
            ws['A1'].value = "System Entry ID"
            ws['B1'].value = "Reference Entry ID"
            ws['C1'].value = "Templates" if granular else "Events"
            ws['K1'].value = "Slots" if granular else "Arguments"
            ws['S1'].value = "Combined"
            ws['AA1'].value = "Combined Score"

            ws['C2'].value = "False Negatives"
            ws['D2'].value = "False Positives"
            ws['E2'].value = f"Aligned {'Templates' if granular else 'Events'}"
            ws['F2'].value = "Weighted Match Score"
            ws['G2'].value = "Match Value"
            ws['H2'].value = "Precision"
            ws['I2'].value = "Recall"
            ws['J2'].value = "F1"

            ws['K2'].value = "False Negatives"
            ws['L2'].value = "False Positives"
            ws['M2'].value = f"Aligned {'Slots' if granular else 'Arguments'}"
            ws['N2'].value = "Weighted Match Score"
            ws['O2'].value = "Match Value"
            ws['P2'].value = "Precision"
            ws['Q2'].value = "Recall"
            ws['R2'].value = "F1"

            ws['S2'].value = "False Negatives"
            ws['T2'].value = "False Positives"
            ws['U2'].value = "Aligned (Combined)"
            ws['V2'].value = "Weighted Match Score"
            ws['W2'].value = "Match Value"
            ws['X2'].value = "Precision"
            ws['Y2'].value = "Recall"
            ws['Z2'].value = "F1"

            for i in range(24):
                letter = get_column_letter(i+3) # C thru Z

                # set all cols to a reasonable width for floats
                ws.column_dimensions[letter].width = 15

                ws[f'{letter}2'].style = "Check Cell" # style built into Excel
                ws[f'{letter}2'].font = font
                ws[f'{letter}2'].alignment = alignmnet

            if color_scale:
                # evt-pr/re/f1 cols
                ws.conditional_formatting.add(f"H3:J{len(entry_pairs)+3}", cmap)
                # arg-pr/re/f1 cols
                ws.conditional_formatting.add(f"P3:R{len(entry_pairs)+3}", cmap)
                # combined-score cols
                ws.conditional_formatting.add(f"X3:Z{len(entry_pairs)+3}", cmap)

            ## insert data

            i = 3
            for entry_pair in entry_pairs:
                if granular:
                    ws.append([
                        entry_pair.sys_entry_id,
                        entry_pair.ref_entry_id,

                        entry_pair.gran_scores.template_measures.miss,
                        entry_pair.gran_scores.template_measures.false_alarm,
                        entry_pair.gran_scores.template_measures.hit,
                        entry_pair.gran_scores.template_measures.match,
                        entry_pair.gran_scores.template_measures.value,
                        entry_pair.gran_scores.template_measures.precision,
                        entry_pair.gran_scores.template_measures.recall,
                        entry_pair.gran_scores.template_measures.f1,

                        entry_pair.gran_scores.slotmatch_measures.miss,
                        entry_pair.gran_scores.slotmatch_measures.false_alarm,
                        entry_pair.gran_scores.slotmatch_measures.hit,
                        entry_pair.gran_scores.slotmatch_measures.match,
                        entry_pair.gran_scores.slotmatch_measures.value,
                        entry_pair.gran_scores.slotmatch_measures.precision,
                        entry_pair.gran_scores.slotmatch_measures.recall,
                        entry_pair.gran_scores.slotmatch_measures.f1,

                        entry_pair.gran_scores.combined_measures.miss,
                        entry_pair.gran_scores.combined_measures.false_alarm,
                        entry_pair.gran_scores.combined_measures.hit,
                        entry_pair.gran_scores.combined_measures.match,
                        entry_pair.gran_scores.combined_measures.value,
                        entry_pair.gran_scores.combined_measures.precision,
                        entry_pair.gran_scores.combined_measures.recall,
                        entry_pair.gran_scores.combined_measures.f1,
                        entry_pair.gran_scores.combined_score,
                    ])
                else:
                    ws.append([
                        entry_pair.sys_entry_id,
                        entry_pair.ref_entry_id,

                        entry_pair.scores.event_measures.miss,
                        entry_pair.scores.event_measures.false_alarm,
                        entry_pair.scores.event_measures.hit,
                        entry_pair.scores.event_measures.match,
                        entry_pair.scores.event_measures.value,
                        entry_pair.scores.event_measures.precision,
                        entry_pair.scores.event_measures.recall,
                        entry_pair.scores.event_measures.f1,

                        entry_pair.scores.argmatch_measures.miss,
                        entry_pair.scores.argmatch_measures.false_alarm,
                        entry_pair.scores.argmatch_measures.hit,
                        entry_pair.scores.argmatch_measures.match,
                        entry_pair.scores.argmatch_measures.value,
                        entry_pair.scores.argmatch_measures.precision,
                        entry_pair.scores.argmatch_measures.recall,
                        entry_pair.scores.argmatch_measures.f1,

                        entry_pair.scores.combined_measures.miss,
                        entry_pair.scores.combined_measures.false_alarm,
                        entry_pair.scores.combined_measures.hit,
                        entry_pair.scores.combined_measures.match,
                        entry_pair.scores.combined_measures.value,
                        entry_pair.scores.combined_measures.precision,
                        entry_pair.scores.combined_measures.recall,
                        entry_pair.scores.combined_measures.f1,
                        entry_pair.scores.combined_score,
                    ])
                ws[f'B{i}'].border = border
                ws[f'J{i}'].border = border
                ws[f'R{i}'].border = border
                ws[f'Z{i}'].border = border
                i += 1

            if granular:
                ws.append([
                    os.path.basename(sys_path),
                    os.path.basename(ref_path),

                    scores.template_measures.miss,
                    scores.template_measures.false_alarm,
                    scores.template_measures.hit,
                    scores.template_measures.match,
                    scores.template_measures.value,
                    scores.template_measures.precision,
                    scores.template_measures.recall,
                    scores.template_measures.f1,

                    scores.slotmatch_measures.miss,
                    scores.slotmatch_measures.false_alarm,
                    scores.slotmatch_measures.hit,
                    scores.slotmatch_measures.match,
                    scores.slotmatch_measures.value,
                    scores.slotmatch_measures.precision,
                    scores.slotmatch_measures.recall,
                    scores.slotmatch_measures.f1,

                    scores.combined_measures.miss,
                    scores.combined_measures.false_alarm,
                    scores.combined_measures.hit,
                    scores.combined_measures.match,
                    scores.combined_measures.value,
                    scores.combined_measures.precision,
                    scores.combined_measures.recall,
                    scores.combined_measures.f1,
                    scores.combined_score,
                ])
            else:
                ws.append([
                    os.path.basename(sys_path),
                    os.path.basename(ref_path),

                    scores.event_measures.miss,
                    scores.event_measures.false_alarm,
                    scores.event_measures.hit,
                    scores.event_measures.match,
                    scores.event_measures.value,
                    scores.event_measures.precision,
                    scores.event_measures.recall,
                    scores.event_measures.f1,

                    scores.argmatch_measures.miss,
                    scores.argmatch_measures.false_alarm,
                    scores.argmatch_measures.hit,
                    scores.argmatch_measures.match,
                    scores.argmatch_measures.value,
                    scores.argmatch_measures.precision,
                    scores.argmatch_measures.recall,
                    scores.argmatch_measures.f1,

                    scores.combined_measures.miss,
                    scores.combined_measures.false_alarm,
                    scores.combined_measures.hit,
                    scores.combined_measures.match,
                    scores.combined_measures.value,
                    scores.combined_measures.precision,
                    scores.combined_measures.recall,
                    scores.combined_measures.f1,
                    scores.combined_score,
                ])

            # final styles
            ws[f'A{len(entry_pairs)+3}'].style = "Calculation"
            ws[f'B{len(entry_pairs)+3}'].style = "Calculation"
            ws[f'AA{len(entry_pairs)+3}'].style = "Calculation"

            for i in range(24):
                ws[f'{get_column_letter(i+3)}{len(entry_pairs)+3}'].style = "Calculation"

            ws[f'B{i}'].border = border
            ws[f'J{i}'].border = border
            ws[f'R{i}'].border = border
            ws[f'Z{i}'].border = border

            ## handle extras

            if extras:

                ws = wb.create_sheet('Extra Statistics')
                ws.freeze_panes = 'C3'

                ## set up header rows

                ws.column_dimensions['A'].width = 37
                ws.column_dimensions['B'].width = 37

                ws.merge_cells('A1:A2')
                ws.merge_cells('B1:B2')
                ws.merge_cells('C1:I1') # agents
                ws.merge_cells('J1:P1') # patients

                ws['A1'].style = "Check Cell"
                ws['B1'].style = "Check Cell"
                ws['C1'].style = "Check Cell"
                ws['J1'].style = "Check Cell"

                ws['A1'].font = font
                ws['B1'].font = font
                ws['C1'].font = font
                ws['J1'].font = font

                ws['A1'].alignment = alignmnet
                ws['B1'].alignment = alignmnet
                ws['C1'].alignment = alignmnet
                ws['J1'].alignment = alignmnet

                ws['B1'].border = border
                ws['I1'].border = border
                ws['I2'].border = border
                ws['P1'].border = border
                ws['P2'].border = border
                ws['Q2'].border = border

                # header values
                ws['A1'].value = "System Entry ID"
                ws['B1'].value = "Reference Entry ID"
                ws['C1'].value = "Agents"
                ws['J1'].value = "Patients"
                ws['Q1'].value = "Money"

                if granular:
                    labels = sorted(list(scores.ttype_match_measures.keys()))
                    labels.extend(["Irrealis", "Time-Attachments"])
                else:
                    labels = ["Agents", "Patients", "Money",] + sorted(
                        list(scores.evtype_match_measures.keys())
                    )

                col_ptr = 3
                for type_ in labels:
                    ws.merge_cells(f'{get_column_letter(col_ptr)}1:{get_column_letter(col_ptr+6)}1')
                    ws[f'{get_column_letter(col_ptr)}1'].style = "Check Cell"
                    ws[f'{get_column_letter(col_ptr)}1'].font = font
                    ws[f'{get_column_letter(col_ptr)}1'].alignment = Alignment(
                        horizontal="center", vertical="center"
                    )
                    ws[f'{get_column_letter(col_ptr)}1'] = type_

                    ws[f'{get_column_letter(col_ptr)}2'].value = "False Negatives"
                    ws[f'{get_column_letter(col_ptr+1)}2'].value = "False Positives"
                    ws[f'{get_column_letter(col_ptr+2)}2'].value = "True Positives"
                    ws[f'{get_column_letter(col_ptr+3)}2'].value = "Weighted Match Score"
                    ws[f'{get_column_letter(col_ptr+4)}2'].value = "Precision"
                    ws[f'{get_column_letter(col_ptr+5)}2'].value = "Recall"
                    ws[f'{get_column_letter(col_ptr+6)}2'].value = "F1"

                    for i in range(7):
                        letter = get_column_letter(col_ptr+i)
                        ws.column_dimensions[letter].width = 15
                        ws[f'{letter}2'].style = "Check Cell" # style built into Excel
                        ws[f'{letter}2'].font = font
                        ws[f'{letter}2'].alignment = alignmnet

                    col_ptr += 7

                i = 3
                for entry_pair in entry_pairs:
                    row = [
                        entry_pair.sys_entry_id,
                        entry_pair.ref_entry_id,
                    ]
                    ws[f'B{i}'].border = border
                    if granular:
                        col_ptr = 3
                        for ttype in sorted(list(scores.ttype_match_measures.keys())):
                            row.extend([
                                entry_pair.gran_scores.ttype_match_measures[ttype].miss,
                                entry_pair.gran_scores.ttype_match_measures[ttype].false_alarm,
                                entry_pair.gran_scores.ttype_match_measures[ttype].hit,
                                entry_pair.gran_scores.ttype_match_measures[ttype].match,
                                entry_pair.gran_scores.ttype_match_measures[ttype].precision,
                                entry_pair.gran_scores.ttype_match_measures[ttype].recall,
                                entry_pair.gran_scores.ttype_match_measures[ttype].f1,
                            ])
                            ws[f'{get_column_letter(col_ptr+6)}{i}'].border = border
                            col_ptr += 7
                        row.extend([
                            entry_pair.gran_scores.irrealis_measures.miss,
                            entry_pair.gran_scores.irrealis_measures.false_alarm,
                            entry_pair.gran_scores.irrealis_measures.hit,
                            entry_pair.gran_scores.irrealis_measures.match,
                            entry_pair.gran_scores.irrealis_measures.precision,
                            entry_pair.gran_scores.irrealis_measures.recall,
                            entry_pair.gran_scores.irrealis_measures.f1,
                        ])
                        ws[f'{get_column_letter(col_ptr+6)}{i}'].border = border
                        col_ptr += 7
                        row.extend([
                            entry_pair.gran_scores.time_attachment_measures.miss,
                            entry_pair.gran_scores.time_attachment_measures.false_alarm,
                            entry_pair.gran_scores.time_attachment_measures.hit,
                            entry_pair.gran_scores.time_attachment_measures.match,
                            entry_pair.gran_scores.time_attachment_measures.precision,
                            entry_pair.gran_scores.time_attachment_measures.recall,
                            entry_pair.gran_scores.time_attachment_measures.f1,
                        ])
                        ws[f'{get_column_letter(col_ptr+6)}{i}'].border = border
                        col_ptr += 7
                    else:
                        row.extend([
                            entry_pair.scores.agent_match_measures.miss,
                            entry_pair.scores.agent_match_measures.false_alarm,
                            entry_pair.scores.agent_match_measures.hit,
                            entry_pair.scores.agent_match_measures.match,
                            entry_pair.scores.agent_match_measures.precision,
                            entry_pair.scores.agent_match_measures.recall,
                            entry_pair.scores.agent_match_measures.f1,

                            entry_pair.scores.patient_match_measures.miss,
                            entry_pair.scores.patient_match_measures.false_alarm,
                            entry_pair.scores.patient_match_measures.hit,
                            entry_pair.scores.patient_match_measures.match,
                            entry_pair.scores.patient_match_measures.precision,
                            entry_pair.scores.patient_match_measures.recall,
                            entry_pair.scores.patient_match_measures.f1,

                            entry_pair.scores.money_match_measures.miss,
                            entry_pair.scores.money_match_measures.false_alarm,
                            entry_pair.scores.money_match_measures.hit,
                            entry_pair.scores.money_match_measures.match,
                            entry_pair.scores.money_match_measures.precision,
                            entry_pair.scores.money_match_measures.recall,
                            entry_pair.scores.money_match_measures.f1,
                        ])
                        ws[f'I{i}'].border = border
                        ws[f'P{i}'].border = border
                        ws[f'Q{i}'].border = border
                        col_ptr = 24
                        for evtype in sorted(list(scores.evtype_match_measures.keys())):
                            row.extend([
                                entry_pair.scores.evtype_match_measures[evtype].miss,
                                entry_pair.scores.evtype_match_measures[evtype].false_alarm,
                                entry_pair.scores.evtype_match_measures[evtype].hit,
                                entry_pair.scores.evtype_match_measures[evtype].match,
                                entry_pair.scores.evtype_match_measures[evtype].precision,
                                entry_pair.scores.evtype_match_measures[evtype].recall,
                                entry_pair.scores.evtype_match_measures[evtype].f1,
                            ])
                            ws[f'{get_column_letter(col_ptr+6)}{i}'].border = border
                            col_ptr += 7
                    ws.append(row)
                    i += 1

                row = [
                    os.path.basename(sys_path),
                    os.path.basename(ref_path),
                ]
                ws[f'B{i}'].border = border
                if granular:
                    col_ptr = 3
                    for ttype in sorted(list(scores.ttype_match_measures.keys())):
                        row.extend([
                            scores.ttype_match_measures[ttype].miss,
                            scores.ttype_match_measures[ttype].false_alarm,
                            scores.ttype_match_measures[ttype].hit,
                            scores.ttype_match_measures[ttype].match,
                            scores.ttype_match_measures[ttype].precision,
                            scores.ttype_match_measures[ttype].recall,
                            scores.ttype_match_measures[ttype].f1,
                        ])
                        ws[f'{get_column_letter(col_ptr+6)}{i}'].border = border
                        col_ptr += 7
                    row.extend([
                        scores.irrealis_measures.miss,
                        scores.irrealis_measures.false_alarm,
                        scores.irrealis_measures.hit,
                        scores.irrealis_measures.match,
                        scores.irrealis_measures.precision,
                        scores.irrealis_measures.recall,
                        scores.irrealis_measures.f1,
                    ])
                    ws[f'{get_column_letter(col_ptr+6)}{i}'].border = border
                    col_ptr += 7
                    row.extend([
                        scores.time_attachment_measures.miss,
                        scores.time_attachment_measures.false_alarm,
                        scores.time_attachment_measures.hit,
                        scores.time_attachment_measures.match,
                        scores.time_attachment_measures.precision,
                        scores.time_attachment_measures.recall,
                        scores.time_attachment_measures.f1,
                    ])
                    ws[f'{get_column_letter(col_ptr+6)}{i}'].border = border
                    col_ptr += 7
                else:
                    row.extend([
                        scores.agent_match_measures.miss,
                        scores.agent_match_measures.false_alarm,
                        scores.agent_match_measures.hit,
                        scores.agent_match_measures.match,
                        scores.agent_match_measures.precision,
                        scores.agent_match_measures.recall,
                        scores.agent_match_measures.f1,

                        scores.patient_match_measures.miss,
                        scores.patient_match_measures.false_alarm,
                        scores.patient_match_measures.hit,
                        scores.patient_match_measures.match,
                        scores.patient_match_measures.precision,
                        scores.patient_match_measures.recall,
                        scores.patient_match_measures.f1,

                        scores.money_match_measures.miss,
                        scores.money_match_measures.false_alarm,
                        scores.money_match_measures.hit,
                        scores.money_match_measures.match,
                        scores.money_match_measures.precision,
                        scores.money_match_measures.recall,
                        scores.money_match_measures.f1,
                    ])
                    ws[f'I{i}'].border = border
                    ws[f'P{i}'].border = border
                    ws[f'Q{i}'].border = border
                    col_ptr = 24
                    for evtype in sorted(list(scores.evtype_match_measures.keys())):
                        row.extend([
                            scores.evtype_match_measures[evtype].miss,
                            scores.evtype_match_measures[evtype].false_alarm,
                            scores.evtype_match_measures[evtype].hit,
                            scores.evtype_match_measures[evtype].match,
                            scores.evtype_match_measures[evtype].precision,
                            scores.evtype_match_measures[evtype].recall,
                            scores.evtype_match_measures[evtype].f1,
                        ])
                        ws[f'{get_column_letter(col_ptr+6)}{i}'].border = border
                        col_ptr += 7
                ws.append(row)
                i += 1

                if color_scale:
                    col_ptr = 3
                    for _ in labels:
                        # evtype-pr/re/f1 and combined-score cols
                        ws.conditional_formatting.add(
                            (
                                f"{get_column_letter(col_ptr+4)}3:"
                                f"{get_column_letter(col_ptr+6)}{i}"
                            ),
                            cmap
                        )
                        col_ptr += 7

            ## desc colorscale

            if color_scale:
                ws = wb.create_sheet("Color Scale")

                ws.column_dimensions['A'].width = 37
                for i in range(101):
                    letter = get_column_letter(i+2) # B thru CX
                    ws.column_dimensions[letter].width = 2
                ws.conditional_formatting.add("B1:CX1", cmap)

                row = ["Color scale is linear, from 0.0 to 1.0.", ]
                row.extend([x*.01 for x in range(0, 101)])
                ws.append(row)

                ws['A2'].value = "Low (0.0)"
                ws['A2'].alignment = Alignment(horizontal="right", vertical="center")
                ws[f"CY2"] = "High (1.0)"

            wb.save(path)

        elif VERBOSE:
            LOGGER.info(
                f"    User chose to not overwrite existing Excel file."
            )
    else:
        raise RuntimeError(f"Missing Excel output path!")


def score_abstract(
    sys_doc,
    ref_doc,
    ignore_synclass=False,
    no_validation=False,
    extras=False,
    filter_by_annotator_class=None,
    filter_by_dataset_label=None,
    filter_by_language=None,
    excluded_entries=[]
):
    """Scores BP JSON results from performer systems against reference data, in the Abstract
    dialect.

    Args:
        sys_doc (bp.BPDocument): a :class:`BPDocument` containing inferred results emitted from a
            system.
        ref_doc (bp.BPDocument): a :class:`BPDocument` containing gold-standard data.
        ignore_synclass (bool): if True, don't consider the syntactic classes of argument
            mentions when scoring.
        no_validation (bool): If True, skip validating BP objects. Defaults to False.
        extras (bool): if True, include extra scoring info for agents, patients, and event-types.
        filter_by_annotator_class (str): if passed, only consider entries whose Provenance in the
            reference matches the given annotator class.
        filter_by_dataset_label (str): Only consider entries whose Provenance in the reference
            matches the given dataset label.
        filter_by_language (str): Only consider entries whose language matches the given ISO-639-3
            code.
        excluded_entries (list) of strings, indicating the Entries (by entry_id) that the scorer
            should skip.

    Returns:
        2-tuple: element 0 is a :class:`ScoreStruct` containing scores for the whole
        system-generated document. Element 1 is a list of :class:`EntryPair` objects, each
        containing score information for each pair of entries in the documents.
    """
    import os
    import sys

    from munkres import make_cost_matrix
    from munkres import Munkres
    import pycountry

    from lib.bp import BPDocument
    from lib.utils import get_tqdm
    from lib.utils import TQDM_PREFIX

    global ABSTRACT_TYPES
    global DEBUG
    global LOGGER
    global VERBOSE

    ## setup

    tqdm = get_tqdm(VERBOSE, DEBUG)

    assert filter_by_annotator_class in BP_ANNOTATOR_CLASS_TYPES+[None,], \
        f"Bad annotator label: {repr(filter_by_annotator_class)}."

    assert type(filter_by_dataset_label) in [type(None), type(""),], \
        f"Bad dataset label: {repr(filter_by_dataset_label)}"

    if filter_by_language:
        try:
            pycountry.languages.get(alpha_3=filter_by_language)
        except KeyError:
            raise KeyError(f"'{filter_by_langauge}' is not a valid ISO-639-3 language code!")

    if not no_validation:
        if not sys_doc.is_valid():
            sys.exit()
        if not ref_doc.is_valid():
            sys.exit()

    stats_summary = ref_doc.stats(pretty=True)
    stats_summary = sys_doc.stats(pretty=True)
    if DEBUG:
        LOGGER.debug(f"Summary of the REFERENCE BP JSON document: \n{stats_summary}\n\n")
        LOGGER.debug(f"Summary of the SYSTEM    BP JSON document: \n{stats_summary}\n\n")

    ref_entries = ref_doc.entries.value
    sys_entries = sys_doc.entries.value

    entry_pairs = []
    scores = ScoreStruct()

    for evtype in ABSTRACT_TYPES:
        scores.evtype_match_measures[evtype] = Measures(evtype)

    if DEBUG:
        LOGGER.debug("======= Starting Abstract Scoring Run =======")

    ## Construct EntryPairs for this EntrySet

    ref_entries_with_events = 0
    for ref_eid in ref_entries:

        if ref_eid in excluded_entries:
            continue

        ref_entry = ref_entries[ref_eid]

        if ref_entry.segment_type.value == "sentence":

            if ref_entry.get_event_set("Abstract") and \
                ref_entry.get_event_set("Abstract").events.value:

                ref_entries_with_events += 1

            candidates = [
                sys_entries[target_eid] for target_eid in sys_entries \
                    if ref_entry.doc_id.value == sys_entries[target_eid].doc_id.value and \
                    ref_entry.sent_id.value == sys_entries[target_eid].sent_id.value
                ]

            if candidates:
                entry_pairs.append(EntryPair(ref_entry, candidates[0], "Abstract"))
                if DEBUG:
                    LOGGER.debug(f"Constructed EntryPair with: {entry_pairs[-1]}")
            else:
                entry_pairs.append(EntryPair(ref_entry, None, "Abstract"))
                if DEBUG:
                    LOGGER.debug(f"Constructed EntryPair with: {entry_pairs[-1]}")

    sys_entries_with_events = 0
    for sys_eid in sys_entries:

        if sys_eid in excluded_entries:
            continue

        sys_entry = sys_entries[sys_eid]

        if sys_entry.segment_type.value == "sentence":

            if sys_entry.get_event_set("Abstract") and \
                sys_entry.get_event_set("Abstract").events.value:

                sys_entries_with_events += 1

            candidates = [
                ref_entries[target_eid] for target_eid in ref_entries \
                    if sys_entry.doc_id.value == ref_entries[target_eid].doc_id.value and \
                    sys_entry.sent_id.value == ref_entries[target_eid].sent_id.value
                ]

            if not candidates:
                entry_pairs.append(EntryPair(None, sys_entry, "Abstract"))
                if DEBUG:
                    LOGGER.debug(f"Constructed EntryPair with: {entry_pairs[-1]}")

    if not ref_entries_with_events and not sys_entries_with_events:
        if DEBUG:
            LOGGER.debug(
                "Both the reference and system files have no populated entries! Nothing to do!"
            )
        scores.compute_score(tab=1)
        return (scores, entry_pairs)

    ## evaluate Abstract Events

    for entry_pair in tqdm(entry_pairs, desc=f"{TQDM_PREFIX} Scoring EntryPairs", unit=" entpr"):

        if filter_by_annotator_class and entry_pair.ref_entry:

            provenance = entry_pair.ref_entry.provenance.value
            if not provenance:
                raise ValueError(
                    "Filtering by annotator class enabled, but reference entry "
                    f"{entry_pair.ref_entry} has no Provenance object!"
                )

            if provenance.annotator_class.value != filter_by_annotator_class:
                if DEBUG:
                    LOGGER.info((
                        f"Skipping entry_pair with reference {entry_pair.ref_entry} because its "
                        f"annotator_class ({provenance.annotator_class.value}) doesn't match!"
                    ))
                continue

        if filter_by_dataset_label and entry_pair.ref_entry:

            provenance = entry_pair.ref_entry.provenance.value
            if not provenance:
                raise ValueError(
                    "Filtering by dataset label enabled, but reference entry "
                    f"{entry_pair.ref_entry} has no Provenance object!"
                )

            if "dataset-label" not in provenance.misc_keys:
                if DEBUG:
                    LOGGER.info((
                        f"Skipping entry_pair with reference {entry_pair.ref_entry} because its "
                        f"has no dataset label!"
                    ))
                continue

            if provenance.misc_keys["dataset-label"] != filter_by_dataset_label:
                if DEBUG:
                    LOGGER.info((
                        f"Skipping entry_pair with reference {entry_pair.ref_entry} because its "
                        f"dataset label ({provenance.misc_keys['dataset-label']}) doesn't "
                        "match!"
                    ))
                continue

        if filter_by_language and entry_pair.ref_entry:

            ref_lang = entry_pair.ref_entry.lang.value
            if not ref_lang:
                raise ValueError(
                    "Filtering by language code enabled, but reference entry "
                    f"{entry_pair.ref_entry} has no language code!"
                )

            if ref_lang != filter_by_language:
                if DEBUG:
                    LOGGER.info((
                        f"Skipping entry_pair with reference {entry_pair.ref_entry} because its "
                        f"langauge code ({ref_lang}) doesn't match the filter "
                        f"'{filter_by_langauge}'!"
                    ))
                continue

        if filter_by_language and entry_pair.sys_entry:

            sys_lang = entry_pair.sys_entry.lang.value
            if not sys_lang:
                raise ValueError(
                    "Filtering by language code enabled, but system entry "
                    f"{entry_pair.sys_entry} has no language code!"
                )

            if sys_lang != filter_by_language:
                if DEBUG:
                    LOGGER.info((
                        f"Skipping entry_pair with system {entry_pair.sys_entry} because its "
                        f"langauge code ({sys_lang}) doesn't match the filter "
                        f"'{filter_by_langauge}'!"
                    ))
                continue

        if not entry_pair.ref_entry:

            if DEBUG:
                LOGGER.debug(f"Found pair with all false-alarms: {entry_pair}")

            num_agents = 0
            num_patients = 0
            for _, event in entry_pair.sys_events.items():
                num_agents += len(getattr(event, "agents").value)
                num_patients += len(getattr(event, "patients").value)

                falarmEvtSS = ScoreStruct()
                falarmEvtSS.event_measures.false_alarm = 1
                falarmEvtSS.argmatch_measures.false_alarm = num_agents + num_patients
                falarmEvtSS.compute_score()
                falarmEvtSS.descr("event falarm  ")

            entry_pair.scores.update(falarmEvtSS)
            if extras and entry_pair.sys_events:
                for event_id, event in entry_pair.sys_events.items():
                    evtype = f"{event.hh.value}_{event.mv.value}"
                    entry_pair.scores.evtype_match_measures[evtype].false_alarm += 1
                entry_pair.scores.agent_match_measures.false_alarm = num_agents
                entry_pair.scores.patient_match_measures.false_alarm = num_patients

            entry_pair.scores.compute_score(tab=1)
            entry_pair.scores.event_alignment = [
                [None for eid in entry_pair.sys_events],
                [eid for eid in entry_pair.sys_events],
            ]

        elif not entry_pair.sys_entry:

            if DEBUG:
                LOGGER.debug(f"Found pair with all misses: {entry_pair}")

            num_agents = 0
            num_patients = 0
            for _, event in entry_pair.ref_events.items():
                num_agents += len(event.agents.value)
                num_patients += len(event.patients.value)

                missedEvtSS = ScoreStruct()
                missedEvtSS.event_measures.miss = 1
                missedEvtSS.argmatch_measures.miss = num_agents + num_patients
                missedEvtSS.compute_score()
                missedEvtSS.descr("event miss    ")

            entry_pair.scores.update(missedEvtSS)
            if extras and entry_pair.ref_events:
                for event_id, event in entry_pair.ref_events.items():

                    evtype = f"{event.hh.value}_{event.mv.value}"
                    entry_pair.scores.evtype_match_measures[evtype].miss += 1

                entry_pair.scores.agent_match_measures.miss = num_agents
                entry_pair.scores.patient_match_measures.miss = num_patients

            entry_pair.scores.compute_score(tab=1)
            entry_pair.scores.event_alignment = [
                [eid for eid in entry_pair.ref_events],
                [None for eid in entry_pair.ref_events],
            ]

        else:

            entry_pair.scores.event_alignment = [[], []]

            # Entry Match with Events to Evaluate (with Munkres algorithm)
            sys_eids = [eid for eid in entry_pair.sys_events]
            ref_eids = [eid for eid in entry_pair.ref_events]

            # Fill out with Nones for misses/false-alarms
            if len(sys_eids) > len(ref_eids):
                fa_delta = len(sys_eids) - len(ref_eids)
                ref_eids.extend([None,]*fa_delta)
            else:
                fa_delta = 0

            if len(ref_eids) > len(sys_eids):
                m_delta = len(ref_eids) - len(sys_eids)
                sys_eids.extend([None,]*(m_delta))
            else:
                m_delta = 0

            if DEBUG:
                LOGGER.debug(
                    f"Found pair with {fa_delta} false-alarms and {m_delta} misses: {entry_pair}"
                )

            event_pair_score_cache = {}

            matrix = make_cost_matrix(
                [
                    [
                        __score_abstract_event_pair(
                            ref_eid,
                            sys_eid,
                            entry_pair,
                            event_pair_score_cache,
                            ignore_synclass=ignore_synclass,
                            extras=extras
                        ) for sys_eid in sys_eids
                    ] for ref_eid in ref_eids
                ],
                ## lambda cost: 1.0 - cost
                lambda cost: cost * -1.0
            )
            if not matrix:
                continue
            index_pairs = Munkres().compute(matrix)

            # Since we have padded the sys and ref id lists with None values, and the scoreEventPair
            # function handles one of the event IDs being a None, then we should have a completely
            # filled square matrix.

            for row, column in index_pairs:
                this_pair_cost = matrix[row][column]
                ref_eid = ref_eids[row]
                sys_eid = sys_eids[column]

                alignedScoreStr = event_pair_score_cache[(ref_eid, sys_eid)]

                alignedScoreStr.compute_score()
                alignedScoreStr.descr("event aligned ", "")

                entry_pair.scores.event_pair_scores[(ref_eid, sys_eid)] = alignedScoreStr
                entry_pair.scores.update(alignedScoreStr, tab=1)
                entry_pair.scores.compute_score(tab=1)

                entry_pair.scores.event_alignment[0].append(ref_eid)
                entry_pair.scores.event_alignment[1].append(sys_eid)

                # Capture the argument-level alignments for this particular pair of aligned events
                entry_pair.scores.arg_alignments[(ref_eid, sys_eid)] = \
                    alignedScoreStr.arg_alignments

            entry_pair.scores.compute_score(tab=1)

        ref_docid = None
        ref_sentid = None
        if entry_pair.ref_entry != None:
            ref_docid = entry_pair.ref_entry.doc_id.value
            ref_sentid = entry_pair.ref_entry.sent_id.value
        sys_docid = None
        sys_sentid = None
        if entry_pair.sys_entry != None:
            sys_docid = entry_pair.sys_entry.doc_id.value
            sys_sentid = entry_pair.sys_entry.sent_id.value

        if DEBUG:
            LOGGER.debug(f"Incrementing scores.")

        scores.update(entry_pair.scores, tab=1)

    scores.compute_score(tab=1)
    scores.descr("corpus scores ")
    return (scores, entry_pairs)


def score_basic(
    sys_doc,
    ref_doc,
    ignore_synclass=False,
    no_validation=False,
    extras=False,
    filter_by_annotator_class=None,
    filter_by_dataset_label=None,
    filter_by_language=None,
    granular=False,
    excluded_entries=[]
):
    """Scores BP JSON results from performer systems against reference data, in the Basic dialect.

    Args:
        sys_doc (bp.BPDocument): a :class:`BPDocument` containing inferred results emitted from a
            system.
        ref_doc (bp.BPDocument): a :class:`BPDocument` containing gold-standard data.
        ignore_synclass (bool): if True, don't consider the syntactic classes of argument
            mentions when scoring.
        no_validation (bool): If True, skip validating BP objects. Defaults to False.
        extras (bool): if True, include extra scoring info for agents, patients, and event-types.
        filter_by_annotator_class (str): if passed, only consider entries whose Provenance in the
            reference matches the given annotator class.
        filter_by_dataset_label (str): Only consider entries whose Provenance in the reference
            matches the given dataset label.
        filter_by_language (str): Only consider entries whose language matches the given ISO-639-3
            code.
        granular (bool): whether the scorer was run in Granular mode or not.
        excluded_entries (list) of strings, indicating the Entries (by entry_id) that the scorer
            should skip.

    Returns:
        2-tuple: element 0 is a :class:`ScoreStruct` containing scores for the whole
        system-generated document. Element 1 is a list of :class:`EntryPair` objects, each
        containing score information for each pair of entries in the documents.
    """
    import os
    import sys

    from munkres import make_cost_matrix
    from munkres import Munkres
    import pycountry

    from lib.bp import BPDocument
    from lib.utils import BP_ANNOTATOR_CLASS_TYPES
    from lib.utils import BASIC_MAT_KEY_ANCHOR_TYPES
    from lib.utils import BASIC_MAT_IGNORED_ANCHOR_TYPES
    from lib.utils import get_tqdm
    from lib.utils import TQDM_PREFIX

    global DEBUG
    global LOGGER
    global SCORING_REF_ARG_SCORING_TYPE
    global VERBOSE

    tqdm = get_tqdm(VERBOSE, DEBUG)

    ## setup

    assert filter_by_annotator_class in BP_ANNOTATOR_CLASS_TYPES+[None,], \
        f"Bad annotator label: {repr(filter_by_annotator_class)}."

    assert type(filter_by_dataset_label) in [type(None), type(""),], \
        f"Bad dataset label: {repr(filter_by_dataset_label)}"

    if filter_by_language:
        try:
            pycountry.languages.get(alpha_3=filter_by_language)
        except KeyError:
            raise KeyError(f"'{filter_by_langauge}' is not a valid ISO-639-3 language code!")

    if not no_validation:
        if not sys_doc.is_valid():
            sys.exit()
        if not ref_doc.is_valid():
            sys.exit()

    ref_entries = ref_doc.entries.value
    sys_entries = sys_doc.entries.value

    entry_pairs = []
    scores = ScoreStruct()

    for evtype in BASIC_MAT_KEY_ANCHOR_TYPES:
        scores.evtype_match_measures[evtype] = Measures(evtype)

    if DEBUG:
        LOGGER.debug("======= Starting Basic Scoring Run =======")

    ## Construct EntryPairs for this EntrySet

    assert len(ref_entries) <= len(sys_entries), (
        "System document does not have the same number of entries! Aborting!\n"
        f"Reference entries: {len(ref_entries)}"
        f"System entries: {len(sys_entries)}"
    )

    ref_entries_with_events = 0
    for ref_eid in ref_entries:

        if ref_eid in excluded_entries:
            continue

        ref_entry = ref_entries[ref_eid]

        if ref_entry.segment_type.value == "document":

            if ref_entry.get_event_set("Basic") and ref_entry.get_event_set("Basic").events.value:
                ref_entries_with_events += 1

            candidates = [
                sys_entries[target_eid] for target_eid in sys_entries \
                    if ref_entry.doc_id.value == sys_entries[target_eid].doc_id.value and \
                    ref_entry.segment_text.value == sys_entries[target_eid].segment_text.value
                ]

            if candidates:
                entry_pairs.append(
                    EntryPair(ref_entry, candidates[0], "Granular" if granular else "Basic")
                )
            else:
                entry_pairs.append(EntryPair(ref_entry, None, "Granular" if granular else "Basic"))
            if DEBUG:
                LOGGER.debug(f"Constructed EntryPair with: {entry_pairs[-1]}")

    sys_entries_with_events = 0
    for sys_eid in sys_entries:

        if sys_eid in excluded_entries:
            continue

        sys_entry = sys_entries[sys_eid]

        if sys_entry.segment_type.value == "document":

            if sys_entry.get_event_set("Basic") and sys_entry.get_event_set("Basic").events.value:
                sys_entries_with_events += 1

            candidates = [
                ref_entries[target_eid] for target_eid in ref_entries \
                    if sys_entry.doc_id.value == ref_entries[target_eid].doc_id.value and \
                    sys_entry.segment_text.value == ref_entries[target_eid].segment_text.value
                ]

            if not candidates:
                entry_pairs.append(EntryPair(None, sys_entry, "Granular" if granular else "Basic"))
                if DEBUG:
                    LOGGER.debug(f"Constructed EntryPair with: {entry_pairs[-1]}")

    if not ref_entries_with_events and not sys_entries_with_events:
        if DEBUG:
            LOGGER.debug(
                "Both the reference and system files have no populated entries! Nothing to do!"
            )
        scores.compute_score(tab=1)
        return (scores, entry_pairs)

    ## evaluate Basic Events

    for entry_pair in tqdm(entry_pairs, desc=f"{TQDM_PREFIX} Scoring EntryPairs", unit=" ent_pr"):

        if filter_by_annotator_class and entry_pair.ref_entry:

            provenance = entry_pair.ref_entry.provenance.value
            if not provenance:
                raise ValueError(
                    "Filtering by annotator class enabled, but reference entry "
                    f"{entry_pair.ref_entry} has no Provenance object!"
                )

            if provenance.annotator_class.value != filter_by_annotator_class:
                if DEBUG:
                    LOGGER.info((
                        f"Skipping entry_pair with reference {entry_pair.ref_entry} because its "
                        f"annotator_class ({provenance.annotator_class.value}) doesn't match!"
                    ))
                continue

        if filter_by_dataset_label and entry_pair.ref_entry:

            provenance = entry_pair.ref_entry.provenance.value
            if not provenance:
                raise ValueError(
                    "Filtering by dataset label enabled, but reference entry "
                    f"{entry_pair.ref_entry} has no Provenance object!"
                )

            if "dataset-label" not in provenance.misc_keys:
                if DEBUG:
                    LOGGER.info((
                        f"Skipping entry_pair with reference {entry_pair.ref_entry} because its "
                        f"has no dataset label!"
                    ))
                continue

            if provenance.misc_keys["dataset-label"] != filter_by_dataset_label:
                if DEBUG:
                    LOGGER.info((
                        f"Skipping entry_pair with reference {entry_pair.ref_entry} because its "
                        f"dataset label ({provenance.misc_keys['dataset-label']}) doesn't "
                        "match!"
                    ))
                continue

        if filter_by_language and entry_pair.ref_entry:

            ref_lang = entry_pair.ref_entry.lang.value
            if not ref_lang:
                raise ValueError(
                    "Filtering by language code enabled, but reference entry "
                    f"{entry_pair.ref_entry} has no language code!"
                )

            if ref_lang != filter_by_language:
                if DEBUG:
                    LOGGER.info((
                        f"Skipping entry_pair with reference {entry_pair.ref_entry} because its "
                        f"langauge code ({ref_lang}) doesn't match the filter "
                        f"'{filter_by_langauge}'!"
                    ))
                continue

        if filter_by_language and entry_pair.sys_entry:

            sys_lang = entry_pair.sys_entry.lang.value
            if not sys_lang:
                raise ValueError(
                    "Filtering by language code enabled, but system entry "
                    f"{entry_pair.sys_entry} has no language code!"
                )

            if sys_lang != filter_by_language:
                if DEBUG:
                    LOGGER.info((
                        f"Skipping entry_pair with system {entry_pair.sys_entry} because its "
                        f"langauge code ({sys_lang}) doesn't match the filter "
                        f"'{filter_by_langauge}'!"
                    ))
                continue

        if not entry_pair.ref_entry:

            if DEBUG:
                LOGGER.debug(f"Found pair with all false-alarms: {entry_pair}")

            entry_pair.scores.event_measures.false_alarm = len(entry_pair.sys_events)

            num_agents = 0
            num_patients = 0
            num_ref_events = 0
            num_money = 0
            for _, event in entry_pair.sys_events.items():

                if extras:
                    entry_pair.scores.evtype_match_measures[event.event_type.value].false_alarm += \
                        len(entry_pair.sys_events)

                num_agents += len(event.agents.value)
                num_patients += len(event.patients.value)
                num_ref_events += len(event.ref_events.value)
                num_money += len(event.money.value)

                falarmEvtSS = ScoreStruct()
                falarmEvtSS.argmatch_measures.false_alarm = num_agents + num_patients + \
                    num_ref_events + num_money
                falarmEvtSS.compute_score()
                falarmEvtSS.descr("event falarm  ")

                entry_pair.scores.update(falarmEvtSS)

            if extras:
                entry_pair.scores.agent_match_measures.false_alarm = num_agents
                entry_pair.scores.patient_match_measures.false_alarm = num_patients
                entry_pair.scores.money_match_measures.false_alarm = num_money

            entry_pair.scores.compute_score(tab=1)
            entry_pair.scores.event_alignment = [
                [None for eid in entry_pair.sys_events],
                [eid for eid in entry_pair.sys_events],
            ]

        elif not entry_pair.sys_entry:

            if DEBUG:
                LOGGER.debug(f"Found pair with all misses: {entry_pair}")

            entry_pair.scores.event_measures.miss = len(entry_pair.ref_events)

            num_agents = 0
            num_patients = 0
            num_ref_events = 0
            num_money = 0
            for _, event in entry_pair.ref_events.items():

                if extras:
                    entry_pair.scores.evtype_match_measures[event.event_type.value].miss += \
                        len(entry_pair.ref_events)

                num_agents += len(event.agents.value)
                num_patients += len(event.patients.value)
                num_ref_events += len(event.ref_events.value)
                num_money += len(event.money.value)

                missedEvtSS = ScoreStruct()
                missedEvtSS.argmatch_measures.miss = num_agents + num_patients + num_ref_events + \
                    num_money
                missedEvtSS.compute_score()
                missedEvtSS.descr("event miss    ")

                entry_pair.scores.update(missedEvtSS)

            if extras:
                entry_pair.scores.agent_match_measures.miss = num_agents
                entry_pair.scores.patient_match_measures.miss = num_patients
                entry_pair.scores.money_match_measures.miss = num_money

            entry_pair.scores.compute_score(tab=1)
            entry_pair.scores.event_alignment = [
                [eid for eid in entry_pair.ref_events],
                [None for eid in entry_pair.ref_events],
            ]

        else:

            first_iteration = True
            entry_pair.scores.event_alignment = [[], []]
            for ref_sent in __get_ref_sents(entry_pair.ref_entry):

                ref_eids = []
                sys_eids = []
                if not SCORING_EVENT_MATCH_INSIDE_SENTENCE:
                    ## Policy has been changed to allow system events to be aligned with
                    ## any reference event anywhere within the document.
                    if first_iteration:
                        ## We use the first pseudo-sentence-level iteration to perform
                        ## system-to-reference event alignment search across ALL the events,
                        ## irrespective of sentence location.  For subsequent iterations we
                        ## do nothing, since all the work has been done.
                        ref_eids = [
                            eid for eid in entry_pair.ref_events \
                                if entry_pair.ref_events[eid].event_type.value not in \
                                BASIC_MAT_IGNORED_ANCHOR_TYPES
                        ]
                        sys_eids = [
                            eid for eid in entry_pair.sys_events \
                                if entry_pair.sys_events[eid].event_type.value not in \
                                BASIC_MAT_IGNORED_ANCHOR_TYPES
                        ]
                        first_iteration = False
                else:
                    ## This is the default behavior, which restricts system-to-reference event
                    ## alignments to those that are located in the same sentence.  Location of
                    ## an event is determined by the offsets of any of the event anchors. Each
                    ## iteration the alignment process is performed for individual reference
                    ## sentences.
                    ref_eids = __get_events_in_sent_region(
                        entry_pair.ref_entry, entry_pair.ref_events, ref_sent, "ref events",
                        BASIC_MAT_IGNORED_ANCHOR_TYPES
                    )
                    sys_eids = __get_events_in_sent_region(
                        entry_pair.sys_entry, entry_pair.sys_events, ref_sent, "sys events",
                        BASIC_MAT_IGNORED_ANCHOR_TYPES
                    )

                # Fill out with Nones for misses/false-alarms
                if len(sys_eids) > len(ref_eids):
                    fa_delta = len(sys_eids) - len(ref_eids)
                    ref_eids.extend([None,]*fa_delta)
                else:
                    fa_delta = 0

                if len(ref_eids) > len(sys_eids):
                    m_delta = len(ref_eids) - len(sys_eids)
                    sys_eids.extend([None,]*(m_delta))
                else:
                    m_delta = 0

                if DEBUG:
                    LOGGER.debug((
                        f"Found pair with {fa_delta} false-alarms and {m_delta} misses: "
                        f"{entry_pair}"
                    ))
                event_pair_score_cache = {}

                matrix = make_cost_matrix(
                    [
                        [
                            __score_basic_event_pair(
                                ref_eid,
                                sys_eid,
                                entry_pair,
                                event_pair_score_cache,
                                ignore_synclass=ignore_synclass,
                                extras=extras
                            ) for sys_eid in sys_eids
                        ] for ref_eid in ref_eids
                    ],
                    lambda cost: cost * -1.0
                )

                if not matrix:
                    continue
                index_pairs = Munkres().compute(matrix)

                for row, column in index_pairs:
                    this_pair_cost = matrix[row][column]
                    ref_eid = ref_eids[row]
                    sys_eid = sys_eids[column]

                    alignedScoreStr = event_pair_score_cache[(ref_eid, sys_eid)]
                    alignedScoreStr.compute_score()
                    alignedScoreStr.descr("event aligned ","")

                    ref_event = entry_pair.ref_events[ref_eid] if ref_eid else None
                    sys_event = entry_pair.sys_events[sys_eid] if sys_eid else None

                    # check if this alignment needs to be broken apart
                    if ref_event and sys_event and \
                        ref_event.event_type.value != sys_event.event_type.value and \
                        alignedScoreStr.event_measures.miss == 1 and \
                        alignedScoreStr.event_measures.false_alarm == 1 and \
                        alignedScoreStr.event_measures.hit == 0 and \
                        alignedScoreStr.event_measures.match == 0:

                        missed_struct = ScoreStruct()
                        falarm_struct = ScoreStruct()

                        missed_measures = Measures(ref_event.event_type.value)
                        falarm_measures = Measures(sys_event.event_type.value)

                        missed_measures.increment_miss(1)
                        falarm_measures.increment_false_alarm(1)

                        missed_struct.event_measures = missed_measures
                        falarm_struct.event_measures = falarm_measures

                        entry_pair.scores.event_pair_scores[(ref_eid, None)] = missed_struct
                        entry_pair.scores.event_alignment[0].append(ref_eid)
                        entry_pair.scores.event_alignment[1].append(None)
                        entry_pair.scores.event_pair_scores[(None, sys_eid)] = falarm_struct
                        entry_pair.scores.event_alignment[0].append(None)
                        entry_pair.scores.event_alignment[1].append(sys_eid)

                    else:

                        entry_pair.scores.event_alignment[0].append(ref_eid)
                        entry_pair.scores.event_alignment[1].append(sys_eid)
                        entry_pair.scores.event_pair_scores[(ref_eid, sys_eid)] = alignedScoreStr

                    entry_pair.scores.update(alignedScoreStr, tab=1)
                    entry_pair.scores.compute_score(tab=1)

                    # Capture the argument-level alignments for this particular pair of aligned
                    # events
                    entry_pair.scores.arg_alignments[(ref_eid, sys_eid)] = \
                        alignedScoreStr.arg_alignments

                if SCORING_REF_ARG_SCORING_TYPE == "event-id-mapping":

                    sys_by_ref = {}

                    if entry_pair.scores.event_alignment:
                        ref_evids = entry_pair.scores.event_alignment[0]
                        sys_evids = entry_pair.scores.event_alignment[1]
                    else:
                        ref_evids = []
                        sys_evids = []
                    for i in range(len(ref_evids)):
                        sys_by_ref[ref_evids[i]] = sys_evids[i]

                    for ref_evid, sys_evid in sys_by_ref.items():

                        if ref_evid == None:
                            if entry_pair.sys_events[sys_evid].ref_events.value != None:
                                for _ in entry_pair.sys_events[sys_evid].ref_events.value:
                                    entry_pair.scores.argmatch_measures.increment_false_alarm(1)

                        elif (sys_evid == None):
                            if entry_pair.ref_events[sys_evid].ref_events.value != None:
                                for _ in entry_pair.ref_events[ref_evid].ref_events.value:
                                    entry_pair.scores.argmatch_measures.increment_miss(1)

                        else:

                            if entry_pair.ref_events[ref_evid].ref_events.value != None:
                                for evid in entry_pair.ref_events[ref_evid].ref_events.value:

                                    if entry_pair.sys_events[sys_evid].ref_events.value != None \
                                        and sys_by_ref[evid] in \
                                        entry_pair.sys_events[sys_evid].ref_events.value:

                                        entry_pair.scores.argmatch_measures.increment_match(1.0)
                                        entry_pair.scores.argmatch_measures.increment_hit(1)
                                        entry_pair.sys_events[sys_evid].ref_events.value.remove(
                                            sys_by_ref[evid]
                                        )

                                    else:
                                        entry_pair.scores.argmatch_measures.increment_miss(1)

                            if entry_pair.sys_events[sys_evid].ref_events.value != None:
                                for _ in entry_pair.sys_events[sys_evid].ref_events.value:
                                    entry_pair.scores.argmatch_measures.increment_false_alarm(1)

            entry_pair.scores.compute_score(tab=1)

        if DEBUG:
            LOGGER.debug(f"Incrementing scores.")

        scores.update(entry_pair.scores, tab=1)

    scores.compute_score(tab=1)
    scores.descr("corpus scores")
    return (scores, entry_pairs)


def score_granular(
    sys_doc,
    ref_doc,
    no_validation=False,
    extras=False,
    filter_by_annotator_class=None,
    filter_by_dataset_label=None,
    filter_by_language=None,
    excluded_entries=[]
):
    """Scores BP JSON results from performer systems against reference data, in the Granular
    dialect.

    Args:
        sys_doc (bp.BPDocument): a :class:`BPDocument` containing inferred results emitted from a
            system.
        ref_doc (bp.BPDocument): a :class:`BPDocument` containing gold-standard data.
        no_validation (bool): If True, skip validating BP objects. Defaults to False.
        extras (bool): if True, include extra scoring info for agents, patients, and event-types.
        filter_by_annotator_class (str): if passed, only consider entries whose Provenance in the
            reference matches the given annotator class.
        filter_by_dataset_label (str): Only consider entries whose Provenance in the reference
            matches the given dataset label.
        filter_by_language (str): Only consider entries whose language matches the given ISO-639-3
            code.
        excluded_entries (list) of strings, indicating the Entries (by entry_id) that the scorer
            should skip.

    Returns:
        2-tuple: element 0 is a :class:`ScoreStruct` containing scores for the whole
        system-generated document. Element 1 is a list of :class:`EntryPair` objects, each
        containing score information for each pair of entries in the documents.
    """
    import csv
    import os
    import sys

    from munkres import make_cost_matrix
    from munkres import Munkres
    import pycountry

    from lib.bp import BPDocument
    from lib.utils import BP_ANNOTATOR_CLASS_TYPES
    from lib.utils import get_tqdm
    from lib.utils import GRANULAR_MAT_TEMPLATE_TYPES
    from lib.utils import SCORING_IGNORED_SLOTS
    from lib.utils import TQDM_PREFIX

    global DEBUG
    global LOGGER
    global VERBOSE

    tqdm = get_tqdm(VERBOSE, DEBUG)

    ## setup

    assert filter_by_annotator_class in BP_ANNOTATOR_CLASS_TYPES+[None,], \
        f"Bad annotator label: {repr(filter_by_annotator_class)}."

    assert type(filter_by_dataset_label) in [type(None), type(""),], \
        f"Bad dataset label: {repr(filter_by_dataset_label)}"

    if filter_by_language:
        try:
            pycountry.languages.get(alpha_3=filter_by_language)
        except KeyError:
            raise KeyError(f"'{filter_by_langauge}' is not a valid ISO-639-3 language code!")

    if not no_validation:
        if not sys_doc.is_valid():
            sys.exit()
        if not ref_doc.is_valid():
            sys.exit()

    if VERBOSE:
        LOGGER.info("Running Basic scorer first to align entries and events.")

    debug_setting = DEBUG
    DEBUG = False
    basic_scores, entry_pairs = score_basic(
        sys_doc,
        ref_doc,
        ignore_synclass=True,
        no_validation=no_validation,
        extras=extras,
        filter_by_annotator_class=filter_by_annotator_class,
        filter_by_dataset_label=filter_by_dataset_label,
        granular=True,
        excluded_entries=excluded_entries
    )
    DEBUG = debug_setting

    ref_entries = ref_doc.entries.value
    sys_entries = sys_doc.entries.value

    scores = GranScoreStruct()

    for ttype in GRANULAR_MAT_TEMPLATE_TYPES:
        scores.ttype_match_measures[ttype] = Measures(ttype)

    if DEBUG:
        LOGGER.debug("======= Starting Granular Scoring Run =======")

    ## evaluate Granular Templates

    for entry_pair in tqdm(entry_pairs, desc=f"{TQDM_PREFIX} Evaluating Templates", unit=" entpr"):

        if DEBUG:
            LOGGER.debug(f"Processing EntryPair {entry_pair}")
            entry_pair.describe()

        if filter_by_annotator_class and entry_pair.ref_entry:

            provenance = entry_pair.ref_entry.provenance.value
            if not provenance:
                raise ValueError(
                    "Filtering by annotator class enabled, but reference entry "
                    f"{entry_pair.ref_entry} has no Provenance object!"
                )

            if provenance.annotator_class.value != filter_by_annotator_class:
                if DEBUG:
                    LOGGER.info((
                        f"Skipping entry_pair with reference {entry_pair.ref_entry} because its "
                        f"annotator_class ({provenance.annotator_class.value}) doesn't match!"
                    ))
                continue

        if filter_by_dataset_label and entry_pair.ref_entry:

            provenance = entry_pair.ref_entry.provenance.value
            if not provenance:
                raise ValueError(
                    "Filtering by dataset label enabled, but reference entry "
                    f"{entry_pair.ref_entry} has no Provenance object!"
                )

            if "dataset-label" not in provenance.misc_keys:
                if DEBUG:
                    LOGGER.info((
                        f"Skipping entry_pair with reference {entry_pair.ref_entry} because its "
                        f"has no dataset label!"
                    ))
                continue

            if provenance.misc_keys["dataset-label"] != filter_by_dataset_label:
                if DEBUG:
                    LOGGER.info((
                        f"Skipping entry_pair with reference {entry_pair.ref_entry} because its "
                        f"dataset label ({provenance.misc_keys['dataset-label']}) doesn't "
                        "match!"
                    ))
                continue

        if filter_by_language and entry_pair.ref_entry:

            ref_lang = entry_pair.ref_entry.lang.value
            if not ref_lang:
                raise ValueError(
                    "Filtering by language code enabled, but reference entry "
                    f"{entry_pair.ref_entry} has no language code!"
                )

            if ref_lang != filter_by_language:
                if DEBUG:
                    LOGGER.info((
                        f"Skipping entry_pair with reference {entry_pair.ref_entry} because its "
                        f"langauge code ({ref_lang}) doesn't match the filter "
                        f"'{filter_by_langauge}'!"
                    ))
                continue

        if filter_by_language and entry_pair.sys_entry:

            sys_lang = entry_pair.sys_entry.lang.value
            if not sys_lang:
                raise ValueError(
                    "Filtering by language code enabled, but system entry "
                    f"{entry_pair.sys_entry} has no language code!"
                )

            if sys_lang != filter_by_language:
                if DEBUG:
                    LOGGER.info((
                        f"Skipping entry_pair with system {entry_pair.sys_entry} because its "
                        f"langauge code ({sys_lang}) doesn't match the filter "
                        f"'{filter_by_langauge}'!"
                    ))
                continue

        if not entry_pair.ref_entry:

            if DEBUG:
                LOGGER.debug(f"    Found pair with all false-alarms: {entry_pair}")

            for _, template in entry_pair.sys_templates.items():

                fa = GranScoreStruct()
                fa.template_measures.false_alarm = 1
                for _, slot_label in GRANULAR_MAT_TEMPLATE_TYPES[template.type_.value].items():

                    # extras
                    entry_pair.gran_scores.ttype_match_measures[template.type_.value].false_alarm \
                        += 1
                    entry_pair.gran_scores.irrealis_measures.false_alarm
                    entry_pair.gran_scores.time_attachment_measures.false_alarm

                    slot = getattr(template, slot_label).value
                    if slot and slot_label not in SCORING_IGNORED_SLOTS:
                        if DEBUG:
                            LOGGER.debug(
                                f"        Adding false-alarms for slot {slot_label}: {slot}"
                            )
                        if type(slot) == list:
                            fa.slotmatch_measures.false_alarm += len(slot)
                            for element in slot:
                                if element.irrealis.value:
                                    pair_scores.irrealis_measures.false_alarm += 1
                                if element.time_attachments.value:
                                    pair_scores.time_attachment_measures.false_alarm += \
                                        len(element.time_attachments.value)
                        else:
                            fa.slotmatch_measures.false_alarm += 1

                entry_pair.gran_scores.update(fa)

            entry_pair.gran_scores.template_alignment = [
                [None for tid in entry_pair.sys_templates],
                [tid for tid in entry_pair.sys_templates],
            ]

        elif not entry_pair.sys_entry:

            if DEBUG:
                LOGGER.debug(f"    Found pair with all misses: {entry_pair}")

            for _, template in entry_pair.ref_templates.items():

                m = GranScoreStruct()
                m.template_measures.miss = 1
                for _, slot_label in GRANULAR_MAT_TEMPLATE_TYPES[template.type_.value].items():

                    entry_pair.gran_scores.ttype_match_measures[template.type_.value].miss += 1

                    slot = getattr(template, slot_label).value
                    if slot and slot_label not in SCORING_IGNORED_SLOTS:
                        if DEBUG:
                            LOGGER.debug(f"        Adding misses for slot {slot_label}: {slot}")
                        if type(slot) == list:
                            m.slotmatch_measures.miss += len(slot)
                            for element in slot:
                                if element.irrealis.value:
                                    pair_scores.irrealis_measures.miss += 1
                                if element.time_attachments.value:
                                    pair_scores.time_attachment_measures.miss += \
                                        len(element.time_attachments.value)
                        else:
                            m.slotmatch_measures.miss += 1

                entry_pair.gran_scores.update(m)

            entry_pair.gran_scores.template_alignment = [
                [tid for tid in entry_pair.ref_templates],
                [None for tid in entry_pair.ref_templates],
            ]

        else:

            entry_pair.gran_scores.template_alignment = [[],[]]
            for template_type in GRANULAR_MAT_TEMPLATE_TYPES:

                if DEBUG:
                    LOGGER.debug(f"=== Evaluating {template_type} Templates ===")

                # get all system and reference templates of this type
                sys_tids = [
                    tid for tid in entry_pair.sys_templates or [] \
                        if entry_pair.sys_templates[tid].type_.value == template_type
                ]
                ref_tids = [
                    tid for tid in entry_pair.ref_templates or [] \
                        if entry_pair.ref_templates[tid].type_.value == template_type
                ]

                # Pad with Nones for misses/false-alarms
                if len(sys_tids) > len(ref_tids):
                    fa_delta = len(sys_tids) - len(ref_tids)
                    ref_tids.extend([None,]*fa_delta)
                else:
                    fa_delta = 0

                if len(ref_tids) > len(sys_tids):
                    m_delta = len(ref_tids) - len(sys_tids)
                    sys_tids.extend([None,]*(m_delta))
                else:
                    m_delta = 0

                if DEBUG:
                    LOGGER.debug((
                        f"    Found {fa_delta} template false-alarms and {m_delta} template misses "
                        "when constructing Munkres template matrix."
                    ))
                    LOGGER.debug("    Starting Munkres asignment search to align templates.")

                template_pair_score_cache = {}
                profit_matrix = [
                    [
                        __score_gran_template_pair(
                            ref_tid,
                            sys_tid,
                            entry_pair,
                            template_pair_score_cache,
                            extras=extras
                        ) for sys_tid in sys_tids
                    ] for ref_tid in ref_tids
                ]

                matrix = make_cost_matrix(
                    profit_matrix,
                    lambda profit : __to_cost(profit)
                )

                if DEBUG:
                    from pprint import pformat
                    LOGGER.debug(f"    Profit matrix: {pformat(profit_matrix, indent=2)}")
                    LOGGER.debug(f"    Cost matrix:   {pformat(matrix, indent=2)}")

                if not matrix:
                    continue
                index_pairs = Munkres().compute(matrix)

                if DEBUG:
                    LOGGER.debug("    Resolving Munkres template alignments.")

                for row, column in index_pairs:
                    this_pair_cost = matrix[row][column]
                    ref_tid = ref_tids[row]
                    sys_tid = sys_tids[column]
                    if DEBUG:
                        LOGGER.debug(f"        Ref {ref_tid} aligns with Sys {sys_tid}")

                    entry_pair.gran_scores.template_alignment[0].append(ref_tid)
                    entry_pair.gran_scores.template_alignment[1].append(sys_tid)

                    pair_scores = template_pair_score_cache[(ref_tid, sys_tid)]
                    entry_pair.gran_scores.slot_alignments[(ref_tid, sys_tid)] = \
                        pair_scores.slot_alignments

                    if DEBUG:
                        LOGGER.debug("            Slot alignments:")
                        for slot_label in pair_scores.slot_alignments:
                            pairs = pair_scores.slot_alignments[slot_label].pairs
                            measures = pair_scores.slot_alignments[slot_label].measures
                            LOGGER.debug(f"                {slot_label}: {pairs}")
                            LOGGER.debug(f"                    {measures}")
                        LOGGER.debug(("                "
                            f"Aggregate Slot Metrics: {pair_scores.slotmatch_measures}"
                        ))

                    entry_pair.gran_scores.template_pair_scores[(ref_tid, sys_tid)] = \
                        pair_scores

                    entry_pair.gran_scores.update(pair_scores, tab=3)

        if DEBUG:
            LOGGER.debug(f"    Updating global scores.")

        scores.update(entry_pair.gran_scores, tab=2)

    scores.compute_score(tab=1)

    if DEBUG:

        LOGGER.debug("Final alignments:")
        for entry_pair in entry_pairs:
            LOGGER.debug(f"    Entry Pair {entry_pair}")
            LOGGER.debug(
                f"      EntryPair Template Metrics: {entry_pair.gran_scores.template_measures}"
            )
            LOGGER.debug(("      "
                f"Aggregate EntryPair Slot Metrics: {entry_pair.gran_scores.slotmatch_measures}"
            ))
            for i in range(len(entry_pair.gran_scores.template_alignment[0])):
                ref_tid = entry_pair.gran_scores.template_alignment[0][i]
                sys_tid = entry_pair.gran_scores.template_alignment[1][i]
                LOGGER.debug(f"        Ref template {ref_tid} aligns with Sys template {sys_tid}")
                slot_alignments = entry_pair.gran_scores.slot_alignments[(ref_tid, sys_tid)]
                agg = Measures(f"{ref_tid} -> {sys_tid}")
                for slot_label in slot_alignments:
                    pairs = slot_alignments[slot_label].pairs
                    measures = slot_alignments[slot_label].measures
                    LOGGER.debug(f"            {slot_label}: {pairs}")
                    LOGGER.debug(f"                {measures}")
                    agg.update(measures)
                LOGGER.debug(("        "
                        f"Subtotal Slot Metrics: {agg}"
                    ))

        scores.describe("corpus scores")

    return (scores, entry_pairs)


## CLI Functions


@click.command(context_settings=CONTEXT_SETTINGS)
@click.option('-v', '--verbose', count=True, help="Show verbose output. Pass again for debug.")
@click.option('-f', '--force', is_flag=True, help="Skip asking to overwrite output.")
@click.option(
    '-l',
    '--save-logs',
    'save_logs',
    type=click.Path(),
    help="Capture logging output (at whatever verbosity level) to the specified path."
)
@click.option(
    '--validate-sys',
    'validate_sys',
    is_flag=True,
    help="Validate the system-generated file against the BP schema (disabled by default)."
)
@click.option(
    '-c',
    '--csv',
    'csv_path',
    type=click.Path(),
    help="Output a CSV spreadsheet of score data to the specified path."
)
@click.option(
    '-t',
    '--tsv',
    'tsv_path',
    type=click.Path(),
    help="Output a TSV spreadsheet of score data to the specified path."
)
@click.option(
    '-e',
    '--excel',
    'excel_path',
    type=click.Path(),
    help="Output an Excel spreadsheet of score data to the specified path."
)
@click.option(
    '--color-scale',
    'color_scale',
    is_flag=True,
    help="Colorize emitted Excel workbook."
)
@click.option(
    '--partial-match-policy',
    type=click.Choice(SCORING_PARTIAL_MATCH_TYPES),
    default=SCORING_PARTIAL_MATCH_TYPE_DEFAULT,
    help=(
        f"Defaults to '{SCORING_PARTIAL_MATCH_TYPE_DEFAULT}'. Only matters for the 'weighted-full' "
        "arg-match-policy."
    )
)
@click.option(
    '--arg-match-policy',
    type=click.Choice(SCORING_ARG_MATCH_TYPES),
    default=SCORING_ARG_MATCH_TYPE_DEFAULT,
    help=f"Defaults to '{SCORING_ARG_MATCH_TYPE_DEFAULT}'."
)
@click.option(
    '--quadclass-match-policy',
    type=click.Choice(SCORING_QUADCLASS_MATCH_TYPES),
    default=SCORING_QUADCLASS_MATCH_TYPE_DEFAULT,
    help=f"Abstract only. Defaults to '{SCORING_QUADCLASS_MATCH_TYPE_DEFAULT}'."
)
@click.option(
    '--event-type-match-policy',
    type=click.Choice(SCORING_EVENT_TYPE_MATCH_TYPES),
    default=SCORING_EVENT_TYPE_MATCH_TYPE_DEFAULT,
    help=f"Basic and Granular only. Defaults to '{SCORING_EVENT_TYPE_MATCH_TYPE_DEFAULT}'."
)
@click.option(
    '--event-can-match-outside-sentence', 'event_outside_sent',
    is_flag=True,
    help=("Allows system and reference events to match that aren't within the same sentence. "
        f"Basic only. Defaults to '{not SCORING_EVENT_MATCH_INSIDE_SENTENCE_DEFAULT}'."
    )
)
@click.option(
    '--anchor-match-policy',
    type=click.Choice(SCORING_ANCHOR_MATCH_TYPES),
    default=SCORING_ANCHOR_MATCH_TYPE_DEFAULT,
    help=f"Defaults to '{SCORING_ANCHOR_MATCH_TYPE_DEFAULT}'."
)
@click.option(
    '--ref-arg-scoring-type',
    type=click.Choice(SCORING_REF_ARG_SCORING_TYPES),
    default=SCORING_REF_ARG_SCORING_TYPE_DEFAULT,
    help=f"Basic and Granular only. Defaults to '{SCORING_REF_ARG_SCORING_TYPE_DEFAULT}'."
)
@click.option(
    '--wrong-span-penalty',
    type=float,
    default=SCORING_INCORRECT_SPAN_REDUCTION_RATE_DEFAULT,
    help=(
        "Multiplicative penalty for spans incorrectly present in a spanset. Defaults to "
        f"{SCORING_INCORRECT_SPAN_REDUCTION_RATE_DEFAULT}."
    )
)
@click.option(
    '--ignore-synclass',
    is_flag=True,
    help="Don't consider the syntactic class of Spans when scoring argument mentions."
)
@click.option(
    '--filter-by-annotator-class',
    type=click.Choice(BP_ANNOTATOR_CLASS_TYPES),
    help=\
        "Only consider entries whose Provenance in the reference matches the given annotator class."
)
@click.option(
    '--filter-by-dataset-label',
    type=str,
    help=\
        "Only consider entries whose Provenance in the reference matches the given dataset label."
)
@click.option(
    '--filter-by-language',
    type=str,
    help="Only consider entries whose language matches the given ISO-639-3 code."
)
@click.option(
    '--exclude-entry', 'excluded_entries',
    type=str, multiple=True,
    help="Skip scoring an entry with the given ID string."
)
@click.option(
    '--extras',
    is_flag=True,
    help="Include extra scoring info for arguments, event-types, templates, and slots."
)
@click.option(
    '--dump-alignments', 'dump_alignments_path',
    type=click.Path(),
    help="Output an Excel workbook with alignment information to the specified path."
)
@click.argument('dialect', type=click.Choice(MAT_DIALECT_CHOICES))
@click.argument('system', type=click.Path(exists=True))
@click.argument('reference', type=click.Path(exists=True))
def cli(
    verbose,
    force,
    save_logs,
    validate_sys,
    csv_path,
    tsv_path,
    excel_path,
    color_scale,
    partial_match_policy,
    arg_match_policy,
    quadclass_match_policy,
    event_type_match_policy,
    event_outside_sent,
    anchor_match_policy,
    ref_arg_scoring_type,
    wrong_span_penalty,
    ignore_synclass,
    filter_by_annotator_class,
    filter_by_dataset_label,
    filter_by_language,
    excluded_entries,
    extras,
    dump_alignments_path,
    dialect,
    system,
    reference
):
    """Scores BP JSON results from performer systems against reference data.

    SYSTEM and REFERENCE must be paths to BP JSON documents. The REFERENCE document is the
    "gold-standard" data that the SYSTEM document will be scored against.

    Tune the scoring policies with the various "match-policy" options. By default, syntactic classes
    are considered when scoring argument mentions; pass ``--ignore-synclass`` to avoid this
    behavior.

    Passing the ``--extras`` option includes other statistics in spreadsheet-like and verbose
    output. It **does not** affect the "combined-score" calculation.

    \f

    This function serves as the entrypoint for the program when called on the CLI. It also handles
    global flag ``verbose``.

    Args:
        verbose (int): if 1, pretty-print output as the command runs. If 2, print debugging output
            in addition to level 1 output. If 0, prints nothing as the command runs. Defaults to 0.
        force (bool): if True, skip asking for permission to overwrite a file with output. Defaults
            to False.
        save_logs (str): a path to write captured logging output to.
        validate_sys (bool): If True, construct the object with validation turned off. Defaults to
            False.
        csv_path (str): a path to write a CSV version of the score data. If None, a CSV file won't
            be written.
        tsv_path (str): a path to write a TSV version of the score data. If None, a TSV file won't
            be written.
        excel_path (str): a path to write an MS Excel version of the score data. If None, an Excel
            file won't be written.
        partial_match_policy (str): Defaults to ``lib.utils.SCORING_PARTIAL_MATCH_TYPE_DEFAULT``.
        arg_match_policy (str): Defaults to ``lib.utils.SCORING_ARG_MATCH_TYPE_DEFAULT``.
        quadclass_match_policy (str): Defaults to
            ``lib.utils.SCORING_QUADCLASS_MATCH_TYPE_DEFAULT``.
        event_type_match_policy (str): Defaults to
            ``lib.utils.SCORING_EVENT_TYPE_MATCH_TYPE_DEFAULT``.
        event_outside_sent (bool): Defaults to
            ``not lib.utils.SCORING_EVENT_MATCH_OUTSIDE_SENTENCE_DEFAULT``. If in Granular mode,
            sentence boundaries are always ignored, and this argument is ignored.
        anchor_match_policy (str): Defaults to ``lib.utils.SCORING_ANCHOR_MATCH_TYPE_DEFAULT``.
        ref_arg_scoring_type (str): Defaults to ``lib.utils.SCORING_REF_ARG_SCORING_TYPE_DEFAULT``.
        wrong_span_penalty (float): Defaults to ``SCORING_INCORRECT_SPAN_REDUCTION_RATE_DEFAULT``.
            For each span incorrectly added to a spanset, the subscore for that spanset's alignment
            will be multiplied by this value.
        ignore_synclass (bool): if True, don't consider the syntactic classes of argument
            mentions when scoring.
        filter_by_annotator_class (str): if passed, only consider entries whose Provenance in the
            reference matches the given annotator class.
        filter_by_dataset_label (str): Only consider entries whose Provenance in the reference
            matches the given dataset label.
        filter_by_language (str): Only consider entries whose language matches the given ISO-639-3
            code.
        excluded_entries (list) of strings, indicating the Entries (by entry_id) that the scorer
            should skip.
        extras (bool): if True, include extra scoring info for agents, patients, and event-types.
        dump_alignments_path (str): a path to write an Excel workbook with alignment information to.
            If None, a file won't be written.
        dialect (str): the task the document was created with. Must be one of
            ``utils.MAT_DIALECT_CHOICES``, which is verified by the Click decorator.
        system (str): a path to a BP JSON document created by an IE system (typically named
            "results.json").
        reference (str): a path to a BP JSON document containing gold-standard annotations.
    """
    # do this before anything else
    from lib.utils import capture_logs
    if save_logs:
        capture_logs(save_logs, verbose, force)

    import os
    import sys

    from tabulate import tabulate

    from lib.bp import BPDocument

    global DEBUG
    global FORCE
    global LOGGER
    global SCORING_PARTIAL_MATCH_TYPE
    global SCORING_ARG_MATCH_TYPE
    global SCORING_QUADCLASS_MATCH_TYPE
    global SCORING_ANCHOR_MATCH_TYPE
    global SCORING_REF_ARG_SCORING_TYPE
    global SCORING_EVENT_TYPE_MATCH_TYPE
    global SCORING_EVENT_MATCH_INSIDE_SENTENCE
    global SCORING_INCORRECT_SPAN_REDUCTION_RATE
    global VERBOSE

    if verbose:
        LOGGER.info("Verbose output is on!")
        if verbose >= 1:
            VERBOSE = True
        if verbose > 1:
            DEBUG = True
            LOGGER.debug("Debug output is on!")

    if verbose and force:
        LOGGER.warning("Forced overwrite is on!")
        if DEBUG and not any([csv_path, tsv_path, excel_path,]):
            LOGGER.debug(
                "    ...but there are no output files specified, so 'force' doesn't apply."
            )
    elif verbose and not force:
        LOGGER.info("Forced overwrite is off!")

    FORCE = force

    SCORING_PARTIAL_MATCH_TYPE = partial_match_policy
    SCORING_ARG_MATCH_TYPE = arg_match_policy
    SCORING_QUADCLASS_MATCH_TYPE = quadclass_match_policy
    SCORING_EVENT_TYPE_MATCH_TYPE = event_type_match_policy
    SCORING_ANCHOR_MATCH_TYPE = anchor_match_policy
    SCORING_REF_ARG_SCORING_TYPE = ref_arg_scoring_type
    SCORING_EVENT_MATCH_INSIDE_SENTENCE = not event_outside_sent
    SCORING_INCORRECT_SPAN_REDUCTION_RATE = wrong_span_penalty

    ## debug stats

    if DEBUG:
        LOGGER.debug("Invocation Arguments:")
        def __vpath(path):
            return f".../{os.path.basename(path)}" if os.path.basename(path) != path else path
        for line in tabulate([
            ["validate_sys", type(validate_sys).__name__, validate_sys,],
            ["csv_path", type(csv_path).__name__, csv_path,],
            ["tsv_path", type(tsv_path).__name__, tsv_path,],
            ["excel_path", type(excel_path).__name__, excel_path,],
            ["partial_match_policy", type(partial_match_policy).__name__, partial_match_policy,],
            ["arg_match_policy", type(arg_match_policy).__name__, arg_match_policy,],
            [
                "quadclass_match_policy",
                type(quadclass_match_policy).__name__,
                quadclass_match_policy,
            ],
            [
                "event_type_match_policy",
                type(event_type_match_policy).__name__,
                event_type_match_policy,
            ],
            ["event_outside_sent", type(event_outside_sent).__name__, event_outside_sent,],
            ["anchor_match_policy", type(anchor_match_policy).__name__, anchor_match_policy,],
            ["ref_arg_scoring_type", type(ref_arg_scoring_type).__name__, ref_arg_scoring_type,],
            ["wrong_span_penalty", type(wrong_span_penalty).__name__, wrong_span_penalty,],
            ["ignore_synclass", type(ignore_synclass).__name__, ignore_synclass,],
            [
                "filter_by_annotator_class",
                type(filter_by_annotator_class).__name__,
                filter_by_annotator_class,
            ],
            [
                "filter_by_dataset_label",
                type(filter_by_dataset_label).__name__,
                filter_by_dataset_label,
            ],
            ["filter_by_language", type(filter_by_language).__name__, filter_by_language,],
            ["excluded_entries", type(excluded_entries).__name__, excluded_entries,],
            ["extras", type(extras).__name__, extras,],
            ["dialect", type(dialect).__name__, dialect,],
            ["system", type(system).__name__, __vpath(system),],
            ["reference", type(reference).__name__, __vpath(reference),],
        ], headers=["Argument", "Type", "Value"], tablefmt="fancy_grid").split("\n"):
            LOGGER.debug(line)

    ## load docs

    # we _may not want_ to validate the system-generated document, since some errors count against
    # systems.
    system_doc = BPDocument.from_json(system, no_validation=(not validate_sys))

    # we _want_ to validate the reference document, always.
    reference_doc = BPDocument.from_json(reference, no_validation=True)

    ## score the system document
    # note that the policies set by this CLI are globally modified earlier.

    if dialect == "Abstract":

        scores, entry_pairs = score_abstract(
            system_doc,
            reference_doc,
            ignore_synclass=ignore_synclass,
            no_validation=(not validate_sys),
            extras=extras,
            filter_by_annotator_class=filter_by_annotator_class,
            filter_by_dataset_label=filter_by_dataset_label,
            filter_by_language=filter_by_language,
            excluded_entries=excluded_entries
        )

    elif dialect == "Basic":

        scores, entry_pairs = score_basic(
            system_doc,
            reference_doc,
            ignore_synclass=ignore_synclass,
            no_validation=(not validate_sys),
            extras=extras,
            filter_by_annotator_class=filter_by_annotator_class,
            filter_by_dataset_label=filter_by_dataset_label,
            filter_by_language=filter_by_language,
            excluded_entries=excluded_entries
        )

    elif dialect == "Granular":

        scores, entry_pairs = score_granular(
            system_doc,
            reference_doc,
            no_validation=(not validate_sys),
            extras=extras,
            filter_by_annotator_class=filter_by_annotator_class,
            filter_by_dataset_label=filter_by_dataset_label,
            filter_by_language=filter_by_language,
            excluded_entries=excluded_entries
        )

    ## generate output data (in various formats)

    if VERBOSE:

        LOGGER.info("Summary Results:")
        if dialect == "Granular":

            for line in tabulate([
                ["tmplt-misses", scores.template_measures.miss,],
                ["tmplt-false-alarms", scores.template_measures.false_alarm,],
                ["tmplt-alignments", scores.template_measures.hit,],
                ["tmplt-match-qual", scores.template_measures.match,],
                ["tmplt-optimizer-value", scores.template_measures.value,],
                ["tmplt-precision", scores.template_measures.precision,],
                ["tmplt-recall", scores.template_measures.recall,],
                ["tmplt-fmeasure", scores.template_measures.f1,],
                ["slot-misses", scores.slotmatch_measures.miss,],
                ["slot-false-alarms", scores.slotmatch_measures.false_alarm,],
                ["slot-alignments", scores.slotmatch_measures.hit,],
                ["slot-match-qual", scores.slotmatch_measures.match,],
                ["slot-match-optimizer-value", scores.slotmatch_measures.value,],
                ["slot-precision", scores.slotmatch_measures.precision,],
                ["slot-recall", scores.slotmatch_measures.recall,],
                ["slot-fmeasure", scores.slotmatch_measures.f1,],
                ["combined-score", scores.combined_score,],
            ], headers=["Field", "Value"], tablefmt="fancy_grid", numalign="right").split("\n"):
                LOGGER.info(line)
            if extras:
                LOGGER.info("Extras:")
                vdata = []
                for ttype in scores.ttype_match_measures:
                    vdata.extend([
                        [
                            f"[{ttype}] template-type-misses",
                            scores.ttype_match_measures[ttype].miss,
                        ],
                        [
                            f"[{ttype}] template-type-false-alarms",
                            scores.ttype_match_measures[ttype].false_alarm,
                        ],
                        [
                            f"[{ttype}] template-type-alignments",
                            scores.ttype_match_measures[ttype].hit,
                        ],
                        [
                            f"[{ttype}] template-type-match-qual",
                            scores.ttype_match_measures[ttype].match,
                        ],
                        [
                            f"[{ttype}] template-type-precision",
                            scores.ttype_match_measures[ttype].precision,
                        ],
                        [
                            f"[{ttype}] template-type-recall",
                            scores.ttype_match_measures[ttype].recall,
                        ],
                        [
                            f"[{ttype}] template-type-fmeasure",
                            scores.ttype_match_measures[ttype].f1,
                        ],
                    ])
                vdata.extend([
                    [f"irrealis-misses", scores.irrealis_measures.miss],
                    [f"irrealis-false-alarms", scores.irrealis_measures.false_alarm],
                    [f"irrealis-alignments", scores.irrealis_measures.hit],
                    [f"irrealis-match-qual", scores.irrealis_measures.match],
                    [f"irrealis-precision", scores.irrealis_measures.precision],
                    [f"irrealis-recall", scores.irrealis_measures.recall],
                    [f"irrealis-fmeasure", scores.irrealis_measures.f1],
                ])
                vdata.extend([
                    [f"time-attachment-misses", scores.time_attachment_measures.miss],
                    [f"time-attachment-false-alarms", scores.time_attachment_measures.false_alarm],
                    [f"time-attachment-alignments", scores.time_attachment_measures.hit],
                    [f"time-attachment-match-qual", scores.time_attachment_measures.match],
                    [f"time-attachment-precision", scores.time_attachment_measures.precision],
                    [f"time-attachment-recall", scores.time_attachment_measures.recall],
                    [f"time-attachment-fmeasure", scores.time_attachment_measures.f1],
                ])
                for line in tabulate(
                    vdata, headers=["Field", "Value"], tablefmt="fancy_grid", numalign="right"
                ).split("\n"):
                    LOGGER.info(line)

        else:

            for line in tabulate([
                ["evt-misses", scores.event_measures.miss,],
                ["evt-false-alarms", scores.event_measures.false_alarm,],
                ["evt-alignments", scores.event_measures.hit,],
                ["evt-match-qual", scores.event_measures.match,],
                ["evt-optimizer-value", scores.event_measures.value,],
                ["evt-precision", scores.event_measures.precision,],
                ["evt-recall", scores.event_measures.recall,],
                ["evt-fmeasure", scores.event_measures.f1,],
                ["arg-misses", scores.argmatch_measures.miss,],
                ["arg-false-alarms", scores.argmatch_measures.false_alarm,],
                ["arg-alignments", scores.argmatch_measures.hit,],
                ["arg-match-qual", scores.argmatch_measures.match,],
                ["arg-match-optimizer-value", scores.argmatch_measures.value,],
                ["arg-precision", scores.argmatch_measures.precision,],
                ["arg-recall", scores.argmatch_measures.recall,],
                ["arg-fmeasure", scores.argmatch_measures.f1,],
                ["combined-score", scores.combined_score,],
            ], headers=["Field", "Value"], tablefmt="fancy_grid", numalign="right").split("\n"):
                LOGGER.info(line)
            if extras:
                LOGGER.info("Extras:")
                vdata = [
                    ["agent-misses", scores.agent_match_measures.miss,],
                    ["agent-false-alarms", scores.agent_match_measures.false_alarm,],
                    ["agent-alignments", scores.agent_match_measures.hit,],
                    ["agent-match-qual", scores.agent_match_measures.match,],
                    ["agent-precision", scores.agent_match_measures.precision,],
                    ["agent-recall", scores.agent_match_measures.recall,],
                    ["agent-fmeasure", scores.agent_match_measures.f1,],

                    ["patient-misses", scores.patient_match_measures.miss,],
                    ["patient-false-alarms", scores.patient_match_measures.false_alarm,],
                    ["patient-alignments", scores.patient_match_measures.hit,],
                    ["patient-match-qual", scores.patient_match_measures.match,],
                    ["patient-precision", scores.patient_match_measures.precision,],
                    ["patient-recall", scores.patient_match_measures.recall,],
                    ["patient-fmeasure", scores.patient_match_measures.f1,],

                    ["money-misses", scores.money_match_measures.miss,],
                    ["money-false-alarms", scores.money_match_measures.false_alarm,],
                    ["money-alignments", scores.money_match_measures.hit,],
                    ["money-match-qual", scores.money_match_measures.match,],
                    ["money-precision", scores.money_match_measures.precision,],
                    ["money-recall", scores.money_match_measures.recall,],
                    ["money-fmeasure", scores.money_match_measures.f1,],
                ]
                for evtype in scores.evtype_match_measures:
                    vdata.extend([
                        [
                            f"[{evtype}] event-type-misses",
                            scores.evtype_match_measures[evtype].miss,
                        ],
                        [
                            f"[{evtype}] event-type-false-alarms",
                            scores.evtype_match_measures[evtype].false_alarm,
                        ],
                        [
                            f"[{evtype}] event-type-alignments",
                            scores.evtype_match_measures[evtype].hit,
                        ],
                        [
                            f"[{evtype}] event-type-match-qual",
                            scores.evtype_match_measures[evtype].match,
                        ],
                        [
                            f"[{evtype}] event-type-precision",
                            scores.evtype_match_measures[evtype].precision,
                        ],
                        [
                            f"[{evtype}] event-type-recall",
                            scores.evtype_match_measures[evtype].recall,
                        ],
                        [
                            f"[{evtype}] event-type-fmeasure",
                            scores.evtype_match_measures[evtype].f1,
                        ],
                    ])
                for line in tabulate(
                    vdata, headers=["Field", "Value"], tablefmt="fancy_grid", numalign="right"
                ).split("\n"):
                    LOGGER.info(line)

    else:

        click.echo(scores.combined_score)

    if csv_path:
        render_csv(
            scores, entry_pairs, system, reference, csv_path, granular=dialect=="Granular",
            extras=extras
        )

    if tsv_path:
        render_csv(
            scores, entry_pairs, system, reference, tsv_path, tabbed=True,
            granular=dialect=="Granular", extras=extras
        )

    if excel_path:
        render_excel(
            scores, entry_pairs, system, reference, excel_path, color_scale=color_scale,
            granular=dialect=="Granular", extras=extras
        )

    if dump_alignments_path:
        dump_alignments(entry_pairs, dump_alignments_path, dialect, ignore_synclass=ignore_synclass)


if __name__ == '__main__':

    ## program proper

    cli()

